#include "Group.h"

#include "Unit.h"
#include "UtilFunctions.h"

#include <vector>
#include <assert.h>
#include <math.h>
#include "Order.h"
#include "Player.h"

Group::Group(const std::string& name,unsigned int playerID) : MapObject(name,playerID) {
	maxSpeed = 100000.0;
	radius = 0.0;
	maxAttackRange = 0.0;
	this->lastDistanceMoved = 0.0;
	currentOrder = NULL;
}

Group::~Group(){
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		for(unsigned int i=0;i<unitTypeVector->size();i++){
			Unit* currentUnit = (*unitTypeVector)[i];
			delete currentUnit;
		}
		delete unitTypeVector;
	}
	if(currentOrder != NULL){
		delete currentOrder;
		if(numIncompleteTasks != NULL){
			(*numIncompleteTasks)--;
		}
		if(conditional != NULL){
			SDL_mutex* mutex = SDL_CreateMutex();
			SDL_mutexP(mutex);
			SDL_CondBroadcast(conditional);
			SDL_mutexV(mutex);
			SDL_DestroyMutex(mutex);
		}
	}
}

Group::Group(const Group& b) : MapObject((MapObject&)b){
	if(b.getOrder() != NULL){
		currentOrder = new Order(*b.getOrder());
	}else{
		currentOrder = NULL;
	}
	
	maxSpeed = b.maxSpeed;
	maxAttackRange = b.maxAttackRange;
	lastDistanceMoved = b.lastDistanceMoved;
	
	UnitMapVector unitMap = b.units;

	UnitMapVector::iterator it;
	for(it = unitMap.begin();it != unitMap.end(); ++it){
		std::vector<Unit*>* unitsOfTypeVector = it->second;
		for(unsigned int i=0;i<unitsOfTypeVector->size();++i){
			Unit* newUnit = new Unit(*((*unitsOfTypeVector)[i])); // copy the unit's attributes
			addUnit(it->first,newUnit);
		}
	}

	lastShotTimes = b.lastShotTimes;
	radius = b.radius;
	
	//  these pointers should be set to NULL, not copied, (I think)
	conditional = NULL;
	numIncompleteTasks = NULL;
}

Group* Group::clone() const{
	return (new Group(*this));
	
}

/*
Group& Group::operator=(const Group& b){
	if (this == &b) return *this;
	
	return *this;
}
*/

bool Group::withinRange(Group* otherGroup,const std::string& unitType){
	if((position - otherGroup->getPosition()).length() <= (radius + otherGroup->getExtents()+unitMap[unitType]->getAttackRange())){
		return true;
	}else{
		return false;
	}
}
bool Group::withinRange(Group* otherGroup){
	if((position - otherGroup->getPosition()).length() <= (radius + otherGroup->getExtents()+otherGroup->getMaxAttackRange())){
		return true;
	}else{
		return false;
	}
}

// take X amount of damage on given unit type
bool Group::takeDamage(const double& damage,const std::string& unitType){
	wasHit = true;
	UnitMapVector::iterator it;
	it = units.find(unitType);
	if(it == units.end()){ // we should never get here, but you never know
		print("WARNING: shouldn't be here in takeDamage function");
	}else{
		for(unsigned int i=0;i<(it->second)->size();i++){
			double hitPoints = (*it->second)[i]->getHitPoints();
			if(hitPoints > 0){
				(*it->second)[i]->setHitPoints(hitPoints-std::max(damage-(*it->second)[i]->getArmor(),1.0)); // always do at least 1 dmg, no matter what
				return true;
			}
		}
	}
	return false;
}
bool Group::eliminateDead(){
	if(!wasHit) return false;
	wasHit = false;
	std::vector<std::string> unitTypesDestroyed;
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		assert(unitTypeVector);
		std::vector<Unit*>::iterator it2;
		if(unitTypeVector != NULL && unitTypeVector->size() > 0){ // this should ALWAYS be true, otherwise, something is wrong
			it2= unitTypeVector->begin();
			while(it2 != unitTypeVector->end()){
				Unit* currentUnit = *it2;
				if(currentUnit->getHitPoints() <= 0){ // if unit is dead
					delete currentUnit;
					it2 = unitTypeVector->erase(it2);
				}else{
					++it2;
				}
			}
			if(unitTypeVector->empty()){
				unitTypesDestroyed.push_back(it->first);
			}
		}
	}
	// now that we have deleted all the relevant units, we need to remove any unit types if all units of that type have been destroyed
	for(unsigned int i=0;i<unitTypesDestroyed.size();i++){
		units.erase(unitTypesDestroyed[i]);
	}
	// now for a final check, check if we have ANY units left..
	if(units.empty()){
		return true;
	}
	recalculateMaxValues();
	recalculateRadius();
	
	return false;
}

std::string Group::getMostDamagingUnit(){
	std::string bestUnit = "NONE";
	double bestDamage = 0.0;
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		assert(unitTypeVector);
		if(unitTypeVector != NULL && unitTypeVector->size() > 0){ // this should ALWAYS be true, otherwise, something is wrong
			for(unsigned int i=0;i<unitTypeVector->size();i++){
				Unit* currentUnit = (*unitTypeVector)[i];
				if(currentUnit->getHitPoints() > 0){ // if unit is still alive
					double unitDmg = currentUnit->getAttackPower() / currentUnit->getAttackRate();
					if(unitDmg > bestDamage){
						bestDamage = unitDmg;
						bestUnit = it->first;
						break;
					}
				}
			}
		}
	}
	return bestUnit;
}

void Group::recalculateRadius(){
	// radius of group depends on number of units, which determines the area.  then use r = sqrt(A/PI) to get result
	this->radius = sqrt((double)getNumUnits()/M_PI) + MIN_GROUP_SIZE;
}

void Group::getChargedUnits(std::vector<std::string>& unitTypeList,double currentTime){
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		assert(unitTypeVector);
		if(unitTypeVector != NULL && unitTypeVector->size() > 0){ // this should ALWAYS be true, otherwise, something is wrong
			// if enough time has passed between shots
			if(currentTime-lastShotTimes[it->first] >= 1000.0*(*unitTypeVector)[0]->getAttackRate()){
				unitTypeList.push_back(it->first);
			}
		}
	}
}

double Group::getNextShotTime(std::vector<std::string>& unitsInRange,double currentTime){
	//UnitMapVector::iterator it;
	double minimumTime = DOUBLE_MAX;
	//for(it = units.begin();it != units.end(); ++it){
	//std::vector<Unit*>* unitTypeVector = (it->second);
	for(unsigned int i=0;i<unitsInRange.size();++i){
		std::vector<Unit*>* unitTypeVector = units[unitsInRange[i]];
		assert(unitTypeVector);
		// FIXME this function is wrong!!!! need to take into account which guys can reach
		if(unitTypeVector != NULL && unitTypeVector->size() > 0){ // this should ALWAYS be true, otherwise, something is wrong
			// if enough time has passed between shots
			double attackRate = 1000.0*(*unitTypeVector)[0]->getAttackRate();
			double timeToNextShot = (lastShotTimes[unitsInRange[i]] + attackRate);
			print("last shot time " + convertToString(lastShotTimes[unitsInRange[i]]));
			print("attackRate " + convertToString(attackRate));
			print("next shot time " + convertToString(timeToNextShot));
			
			if(timeToNextShot < currentTime){ // if we can shoot already, return current time
				return currentTime;
			}else if(timeToNextShot < minimumTime){ // otherwise, see when we can shoot next
				minimumTime = timeToNextShot;
			}
		}
	}
	return minimumTime;
}

void Group::draw(){
	
	// here, draw the attack ranges of all the different units in the group...
	unsigned int count = 0;
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		assert(unitTypeVector);
		if(unitTypeVector != NULL && unitTypeVector->size() > 0){ // this should ALWAYS be true, otherwise, something is wrong
			drawCircle(position,(*unitTypeVector)[0]->getAttackRange()+radius,CIRCLE_DETAIL,transparentBlue);
			
			// this whole thing is really hacky... but there is no easy way to make this work, and I don't want to put that much effort
			// just to make it look pretty
			if(displayObjectInfo){
				std::string unitsString = (it->first) +": " + convertToString(unitTypeVector->size());
				if(draw3DTextInfo){
					draw3DString(position.x-8,position.y-8-count*(2+cameraPos.z/-200),unitsString,color,1.0/(-cameraPos.z/4.0)); 
				}else{
					draw2DString(position.x-8,position.y-8-count*(2+cameraPos.z/-200),unitsString,color,true);
				}
			}
		}else{
			printToScreen("SOMETHING TERRIBLY WRONG IN GROUP: " + name);
		}
		count++;
	}
	
	// draw the unit itself...
	drawCircle(position,radius,CIRCLE_DETAIL,color);
	
}

void Group::drawInSelectionMode(){
	glPushName(GROUP_TYPE_ID);
	
	MapObject::drawInSelectionMode();
	
	glPopName();
}

void Group::addUnit(const std::string& name,Unit* newUnit){
	UnitMapVector::iterator it;
	it = units.find(name);
	if(it == units.end()){
		units[name] = new std::vector<Unit*>;
		units[name]->push_back(newUnit);
		recalculateMaxValues();
	}else{
		units[name]->push_back(newUnit);
	}
	
	//hasShot(name);
	recalculateRadius();
}

void Group::recalculateMaxValues(){
	maxSpeed = 1000000.0;
	maxAttackRange = 0.0;
	
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		// recalculate attack range
		if((*unitTypeVector)[0]->getAttackRange() > maxAttackRange){
			maxAttackRange = (*unitTypeVector)[0]->getAttackRange();
		}
		// now do the speed..
		if((*unitTypeVector)[0]->getMoveSpeed() < maxSpeed){
			maxSpeed = (*unitTypeVector)[0]->getMoveSpeed();
		}
	}
	
}

std::vector<Unit*>* Group::getUnitsOfType(const std::string& unitType){
	return units[unitType];
}

double Group::getAttackRangeForType(const std::string& unitType){
	if(units[unitType]->empty()){
		print("SHOULD NEVER GET HERE IN Group::getAttackRange");
	}else{
		return (*(units[unitType]))[0]->getAttackRange() + radius;
	}
	return 0.0;
}

void Group::printUnits(){
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		for(unsigned int i=0;i<unitTypeVector->size();i++){
			Unit* currentUnit = (*unitTypeVector)[i];
			print(currentUnit->toString());
		}
	}
}

double Group::getExtents() const{
	return radius;
}

void Group::hasShot(const std::string& unitType,double time){
	lastShotTimes[unitType] = time;
}

unsigned int Group::getNumUnits(){
	unsigned int count = 0;
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		count += unitTypeVector->size();
	}
	return count;
}
unsigned int Group::getNumUnitsOfType(const std::string& unitType){
	return units[unitType]->size();
}

// FIXME unfortunately, this method of saving will heal all injured units... no easy way to fix this, except to print out the stats for ALL units...
// not sure what to do about it right now, just gonna leave it...
std::string Group::printUnitComposition(){
	unsigned int count = 0;
	std::ostringstream compositionStream;
	
	UnitMapVector::iterator it;
	for(it = units.begin();it != units.end(); ++it){
		std::vector<Unit*>* unitTypeVector = (it->second);
		assert(unitTypeVector);
		count = unitTypeVector->size();
		if(count > 0){ // this should ALWAYS BE TRUE
			compositionStream << (it->first) << " " << count << "\n";
		}else{
			print("COUNT OF ZERO DETECTED IN GROUP, THIS IS WRONG!");
		}
	}
	
	return compositionStream.str();
}

void Group::setOrder(Order* newOrder,SDL_cond* conditional,unsigned int* numIncompleteTasks){
	// if we had a previous order, get rid of it
	if(currentOrder != NULL){
		delete currentOrder;
		if(this->numIncompleteTasks != NULL) (*this->numIncompleteTasks)--;
		if(this->conditional != NULL){
			SDL_mutex* mutex = SDL_CreateMutex();
			SDL_mutexP(mutex);
			SDL_CondBroadcast(conditional);
			SDL_mutexV(mutex);
			SDL_DestroyMutex(mutex);
		}
	}
	this->conditional = conditional;
	this->numIncompleteTasks = numIncompleteTasks;
	
	if(this->numIncompleteTasks != NULL) (*this->numIncompleteTasks)++;
	
	currentOrder = newOrder;
	
	// FIXME the estimated times are slightly off of the actual time.... they shouldn't be, but it could be okay, since it happens for all units?
	//double startTime = (double)time;
	//print("unit " + this->getName() + " order start time: "+convertToString(startTime));
	
	//double estimatedTime = (currentOrder->getDestination() - this->getPosition()).length() / ((double)currentOrder->getSpeed()/1000.0);

	//print("unit " + this->getName() + " order estimated end time: "+convertToString(estimatedTime+startTime));
	
}

Order* Group::getOrder() const{
	return currentOrder;
}

void Group::executeOrders(double timeElapsed){
	if(currentOrder == NULL){
		lastDistanceMoved = 0.0;
		return;
	}
	Vector3D direction = currentOrder->getDestination() - position;
	
	double adjustedMaxSpeed = currentOrder->getSpeed()*(timeElapsed/1000.0);
	
	if(direction.length() <= adjustedMaxSpeed){
		// if we have reached our current destination order...
		position = currentOrder->getDestination();
		//print("REACHED GOAL");
		// if we have reached our destination, then we can execute our order action...
		executeOrderAction(currentOrder->getActionType());
	}else{
		position+=(direction|adjustedMaxSpeed);
		//print("position is: " + convertToString(position.x) + ", " + convertToString(position.y));
	}
	if(direction.length() == 0.0){
		lastDistanceMoved = 0.0;
	}else{
		lastDistanceMoved = (direction|adjustedMaxSpeed).length();
	}
}

void Group::executeOrderAction(const std::string& actionType){
	if(actionType == "MOVE"){ // do nothing at the end, since all we were doing is moving...

	}else if(actionType == "JOIN"){
		Player* owningPlayer = playerMap[playerID];
		std::vector<Group*> groupsAtLocation;
		owningPlayer->getGroupsAtLocation(this,groupsAtLocation);
		for(unsigned int i=0;i<groupsAtLocation.size();++i){
			join(groupsAtLocation[i]);
			
		}
		
	}else if(actionType == "ATTACK"){
		if(enemyInRange()) return; // if there are still enemies in range, don't end this order...
		
	}
	
	cleanupOrder();
}

void Group::cleanupOrder(){
	if(currentOrder == NULL){
		print("Trying to cleanup an empty order, this should not happen!");
		return;
	}
	
	// some debug stuff here for now...
	
	print("unit " + this->getName() + " order end time: "+convertToString(glutGet(GLUT_ELAPSED_TIME)));
	
	delete currentOrder;
	currentOrder = NULL;
	
	// if we are done with this order...
	if(numIncompleteTasks != NULL){
		(*numIncompleteTasks)--;
	}
	if(conditional != NULL){
		SDL_mutex* mutex = SDL_CreateMutex();
		SDL_mutexP(mutex);
		SDL_CondBroadcast(conditional);
		SDL_mutexV(mutex);
		SDL_DestroyMutex(mutex);
	}
	
}

void Group::join(Group* groupToJoin){
	this->name = this->name + "_" + groupToJoin->getName(); // concatenate the names...
	UnitMapVector::iterator it;
	for(it = groupToJoin->units.begin();it != groupToJoin->units.end(); ++it){
		std::vector<Unit*>* unitType = it->second;
		for(unsigned int i=0;i<unitType->size();++i){
			UnitRecharges::iterator it2 = lastShotTimes.find(it->first);
			if(it2 != lastShotTimes.end()){
				if(lastShotTimes[it->first] < (groupToJoin->getLastShotTimes())[it->first]){
					lastShotTimes[it->first] =  (groupToJoin->getLastShotTimes())[it->first];
				}
			}else{
				lastShotTimes[it->first] = (groupToJoin->getLastShotTimes())[it->first];
			}
			
			addUnit((it->first),(*unitType)[i]);
		}
		unitType->clear();
	}
	groupToJoin->units.clear();
	groupToJoin->wasHit = true; // pretend like it was hit in combat, so it is removed from the sim
}

bool Group::enemyInRange(){
	PlayerMap::iterator playerIterator;
	Player* owningPlayer = playerMap[playerID];
	
	for(playerIterator= playerMap.begin();playerIterator != playerMap.end(); ++playerIterator){
		if(owningPlayer == playerIterator->second) continue; // don't check your own objects...
			
		PlayerObjectMap* playerObjects =  (playerIterator->second)->getObjects();
		PlayerObjectMap::iterator it;
		for(it = playerObjects->begin();it != playerObjects->end(); ++it){
			if((it->second)->withinRange(this)){
				return true;
			}
		}
	}
	return false;
}

void Group::enemyInRangeOfUnitType(std::vector<std::string>& unitTypes){
	PlayerMap::iterator playerIterator;
	Player* owningPlayer = playerMap[playerID];
	
	
	for(playerIterator= playerMap.begin();playerIterator != playerMap.end(); ++playerIterator){
		if(owningPlayer == playerIterator->second) continue; // don't check your own objects...
		PlayerObjectMap* playerObjects =  (playerIterator->second)->getObjects();
		PlayerObjectMap::iterator it;
		for(it = playerObjects->begin();it != playerObjects->end(); ++it){
			
			UnitMapVector::iterator it2;
			for(it2 = units.begin();it2 != units.end(); ++it2){
				if((it->second)->withinRange(this,it2->first)){
					unitTypes.push_back(it2->first);
				}
			}
		}
	}
}


double Group::getLastDistanceMoved() const{
	return lastDistanceMoved;
}

double Group::getMaxSpeed() const{
	return maxSpeed;
}

double Group::getMaxAttackRange() const{
	return maxAttackRange;
}

UnitRecharges Group::getLastShotTimes() const{
	return lastShotTimes;
}

std::string Group::toString(){
	std::ostringstream groupStream;
	
	groupStream << "group: "  << MapObject::toString();
	
	groupStream << printUnitComposition();
	
	groupStream << "end" << "\n";
	
	return groupStream.str();
}



