#include "Application.H"
#include "Helper.H"

#include "GameStateModule.H"
#include "GameChanges.H"
#include "GameObj.H"
#include "Game.H"

class MyApplication : public Application
{
    bool				firstView;
    std::vector<Unit::ptr>		mySoldiers;
    std::vector<Unit::ptr>		myBuildings;
    std::vector<Unit::const_ptr>	enemySoldiers;
    std::vector<Unit::const_ptr>	enemyBuildings;

	
public:
    MyApplication() : firstView(true) {}

    void OnNewUnit(Unit::ptr unit);
    void OnUnitDeath(Unit::ptr unit);
    void OnReceivedView();
};

static Unit::const_ptr PickClosest(
    const Unit & unit, 
    const std::vector<Unit::const_ptr> & candidates);

void MyApplication::OnNewUnit(Unit::ptr unit)
{
    // Store it depending on its relationship to this client
    if(unit->IsOwned())
    {
        if(unit->CanMove())
        {
            mySoldiers.push_back(unit);
        }
        else
        {
            myBuildings.push_back(unit);
        }
    }
    else if(unit->IsEnemy())
    {
        if(unit->CanMove())
        {
            enemySoldiers.push_back(unit);
        }
        else
        {
            enemyBuildings.push_back(unit);
        }
    }
}

void MyApplication::OnUnitDeath(Unit::ptr unit)
{
    // Quit storing the unit
    quick_erase(mySoldiers,unit);
    quick_erase(myBuildings,unit);
    quick_erase(enemySoldiers,unit);
    quick_erase(enemyBuildings,unit);
}

void MyApplication::OnReceivedView()
{
    if(firstView)
    {
        // TODO: Any one-time initialisation or planning
        firstView = false;
    }

    // EXAMPLE: First target enemy units, then target enemy buildings

    // Iterate through all mobile units owned by this client
    for(std::vector<Unit::ptr>::const_iterator it(mySoldiers.begin()), 
        end(mySoldiers.end()); it!=end; ++it)
    {
        Unit & unit(**it);

        // If the unit has a weapon, look for targets
        if(unit.HasWeapon())
        {
            // Draw the range of this unit
            DrawDebugCircle(unit.GetPosition(), 
                            unit.GetWeaponRange(), 
                            Color(1,1,1));

            // Choose the best target
            Unit::const_ptr bestTarget;
            if(!bestTarget) bestTarget = PickClosest(unit, enemySoldiers);
            if(!bestTarget) bestTarget = PickClosest(unit, enemyBuildings);
            if(bestTarget)
            {
                const Unit & target(*bestTarget);

                // If we are close enough to attack
                if(unit.GetDistanceTo(target) <= unit.GetWeaponRange())
                {
                    unit.Attack(target); // Attack the enemy with our weapons

                    // Draw a red line, symbolising attacking
                    DrawDebugLine(unit.GetPosition(), 
                                  target.GetPosition(), 
                                  Color(1,0,0));
                }
                // Otherwise, if we aren't already moving, move to the target
                else if(!unit.IsMoving())
                {
                    // Schedule pathfinding to take us to within firing range 
                    // of the enemy.
                    unit.PathTo(Movement::AttackObject(target.GetGameObjPtr()));

                    // Draw a green line, symbolising movement
                    DrawDebugLine(unit.GetPosition(), 
                                  target.GetPosition(), 
                                  Color(0,1,0));
                }
            }
        }
    }
}

Unit::const_ptr PickClosest(const Unit & unit, 
                            const std::vector<Unit::const_ptr> & candidates)
{
    Unit::const_ptr bestCandidate;
    real8 bestDistance(std::numeric_limits<float>::infinity());
    for(std::vector<Unit::const_ptr>::const_iterator it(candidates.begin()), 
        end(candidates.end()); it!=end; ++it)
    {
        const Unit::const_ptr & candidate(*it);
        real8 distance(unit.GetDistanceTo(*candidate));
        if(distance < bestDistance)
        {
            bestCandidate	= candidate;
            bestDistance	= distance;
        }
    }
    return bestCandidate;
}

int main(int argc, char * argv[])
{
    try
    {
        MyApplication app;
        return app.Run(argc, argv);
    }
    catch(...) 
    {
        return -1;
    }
}
