// $Id: Net_Service.C 5745 2007-08-23 17:59:09Z krysta $

// (c) Michael Buro 2003, mburo@cs.ualberta.ca, licensed under the GPL

#include <sstream>

#include "Net_Service.H"
using namespace std;
using namespace boost;

static const bool debug = true;

Net_Service::Net_Service(string& server, int port, string& alias,
                         string& pass,   string& dbFile, string& rootNick_, string& rootPass_, bool verbose_) :
  Net_GGSClient(alias, pass, server, port, true)
{

  invitList = new Net_InvitList;
  db = new Net_DB(dbFile);

  q_mut = SDL_CreateMutex();
  client_mut = SDL_CreateMutex();

  serve_sem = SDL_CreateSemaphore(0);

  rootNick = rootNick_;
  rootPass = rootPass_;

  // add root to the database
  DB_Player root(rootNick, rootPass,  ROOT);
  db->put(root);

  verbose = verbose_;
}


Net_Service::~Net_Service() {
  delete invitList;
  delete db;
  SDL_DestroySemaphore(serve_sem);
  SDL_DestroyMutex(client_mut);
  SDL_DestroyMutex(q_mut);
}
/**
   prints the given string if running on verbose mode
 */
void Net_Service::print(const string& out){
  if(verbose){
    time_t t;
    time(&t);
    char ts[27];
    asctime_r(localtime(&t), ts);
    ts[strlen(ts) - 1] = 0; // delete the new line
    cout << "SERVICE: " << out << " " << ts << endl;
  }
}


/**
 * it is run as a thread
 * blocks waiting for a message
 * once a message is recieved it creats a thread of dispatchRequest
 */
sint4 Net_Service::listen()
{
  string request;

  while (!shut_down) {

    this->recv(request);

    if (request.size() > 0) {
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION
      if (SDL_mutexP(q_mut)) ERR("mutex 1");
      // put request
      //cout << "# Received request:" << request << endl;
      queue.push_front(request);
      if (SDL_mutexV(q_mut)) ERR("mutex 2");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      SDL_SemPost(serve_sem); // signal read that there is data available
    }
  }

  return 0;
  //  exit(0); // this thread is done
}


int dispatch(void * data)
{
  // message parsers
  Invite invite;
  Join join;
  Leave leave;
  Ready ready;
  Unready unready;
  Ban ban;
  Abort abort_;
  Tell tell;
  Chat chat;
  Join_Team join_team;
  Game_Start game_start;
  Game_End game_end;
  
  Register register_;
  Login login;
  List list;

  Buddy buddy;
  Worker_Ready wready;

  Down down;

  GGS_Notify notify;
  Finger finger;

  Error error;

  Worker_Available wavailable;

  Net_Service &service = * static_cast<Net_Service *>(data);

  service.print("[------------DISPATCH------------]");

  string request;
  Net_Message m;

  while (!shut_down ) {

    // wait for a request
    sint4 sw;
    while ((sw = SDL_SemWait(service.serve_sem))) {
      REM("NetService dispatch loop awoken prematurely");
    }

    // <<<<<<<<<<<<<<<< START CRITICAL SECTION
    if (SDL_mutexP(service.q_mut)) ERR("mutex 1");
    // get request form queue
    if (service.queue.size() <= 0) {
      if (SDL_mutexV(service.q_mut)) ERR("mutex 2");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION
      continue;
    }
    request = service.queue.back();
    service.queue.pop_back();
    if (SDL_mutexV(service.q_mut)) ERR("mutex 2");
    // >>>>>>>>>>>>>>>> END CRITICAL SECTION

    m.setMessage(request);
   
    if (m.getTo() == ".chat") { continue; }
    if (chat.parse(m))       { service.serve_chat(chat);              continue; }
    if (down.parse(m))       { service.serve_down(down);              continue; }
    if (buddy.parse(m))      { service.serve_buddy(buddy);            continue; }
    if (notify.parse(m))     { service.serve_notify(notify);          continue; }
    if (invite.parse(m))     { service.serve_invite(invite);          continue; }
    if (game_start.parse(m)) { service.serve_game_start(game_start);  continue; }
    if (game_end.parse(m))   { service.serve_game_end(game_end);      continue; }
    if (join.parse(m))       { service.serve_join(join);              continue; }
    if (join_team.parse(m))  { service.serve_join_team(join_team);    continue; }
    if (ban.parse(m))        { service.serve_ban(ban);                continue; }
    if (leave.parse(m))      { service.serve_leave(leave);            continue; }
    if (ready.parse(m))      { service.serve_ready(ready);            continue; }
    if (unready.parse(m))    { service.serve_unready(unready);        continue; }
    if (abort_.parse(m))     { service.serve_abort(abort_);           continue; }
    if (register_.parse(m))  { service.serve_register(register_);     continue; }
    if (login.parse(m))      { service.serve_login(login);            continue; }
    if (wready.parse(m))     { service.serve_worker_ready(wready);    continue; }
//    if (error.parse(m)) { service.serve_error(error); continue;}
    if (wavailable.parse(m)){ service.serve_worker_available(wavailable); continue; }

    if (list.parse(m)) {
      if(list.get_list_name()== INVITE_LIST){ service.serve_invitations(list); continue; }
      if(list.get_list_name()== BUDDY_LIST) { service.serve_buddies(list); continue; }
      if(list.get_list_name()== GAME_LIST)  { continue; }
      if(list.get_list_name()== WHO_LIST)   { service.serve_who(list); continue; }
      continue;
    }

    if (tell.parse(m)) { service.serve_tell(tell); continue; }

    if (m.getFrom() == "G") { continue; }

    if (m.getArgument(0) == "print"){
      service.db->print();
      continue;
    }

    service.print("Message not parsed : " + m.toString());
    if ( m.getArgument(0) == "list") {
      cout << "**THIS INITIAL GGS LIST MESSAGE IS IGNORED!\n**"; //fixme, hack to get rid of error message
      continue;
    }
    // invalid
    if (m.getFrom() != "/os")  //fixme hack
      service.tellClient(m.getFrom(), Error::compose("Invalid command"));
  }

  return 0;
  // exit(0); // this thread is done
}


/**
 * starts listening for service requests
 * returns -1 if the starting failed
 */
sint4 Net_Service::start()
{
  FORS (i, MAX_THREADS) {
    if (SDL_CreateThread(dispatch, this) < 0) return -1;
  }

  listen();
  return 1;
}

void Net_Service::shutDown()
{
  print("Service is shutting down...");
  this->close();
  db->close();
  //exit(0);
}


bool Net_Service::isOnline(const string& name){
  return !(allPlayers.find(name) == allPlayers.end());
}

void Net_Service::tell_worker_info(){
  string msg, rest = "";
  Net_Message m;
  Worker_Start ws;
  FORALL(gameQ, i) {
    msg = *i;
    m.setMessage("U: " + msg);
    if(!ws.parse(m))
      cout << "iid is: "<< ws.getIID() << endl;
    rest += ( to_string(ws.getIID()) + " ");
  }
  tellAll( Winfo::compose(to_string(num_avail_workers()), rest) );
}

/**
   picks a worker, starts the game
 */
void Net_Service::startGame(Net_Invitation& inv)
{
  print("[------------------i am starting a game ------------------]");

  // generate passes
  Vector<string> ips;
  Vector<string> passes;
  generatePasses(inv, ips, passes);

  // assemble team string and id's string
  char ch = 'a' - 1;
  stringstream teamstr;
  teamstr << "-team ";

  int numTeamsLeft = inv.getTeamCount();
  for (int i=0; numTeamsLeft > 0; i++) {
    FORS (j, inv.getPlayerCount(i+1)) {
      if (j==0) {
	--numTeamsLeft;
	++ch;
      }
      teamstr << ch;
    } 
  }

  string idstr = inv.getidstring(); 
    
  if(workerQ.size() > 0) {
    string& workerName =(workerQ.front()).first;
    string& workerHost =(workerQ.front()).second.first;
    string& workerPort =(workerQ.front()).second.second;

    cout << "Starting game with options: " << inv.getOptions() << endl;    
    tellClient(workerName, Worker_Start::compose( inv.getIID(), inv.getOptions(), inv.getReadyCount(), ips, passes, teamstr.str(), idstr, workerPort, workerHost));

    const Vector<string> &all = inv.getPlayers();
    const Vector<uint1>  present = inv.getPresent();

    // <<<<<<<<<<<<<<<< START CRITICAL SECTION
    if (SDL_mutexP(client_mut)) ERR("mutex 1");
    string info;
	      	  
    int j = 0;
    FORU(i, all.size()){
      if(present[i]){    	  
        info = Game_Info::compose(workerHost, passes[j], inv.getIID(), teamstr.str(), "-port " + workerPort);
        j++;
        this->send("tell " + all[i] + " " + info);
      }
    }

    readyGameQ[workerName] = inv;

    inv.setStarted(true);

    workerQ.pop_front();
    print("*** POPPING from workerQ");    
    
  } else{
    print("*** WORKER QUEUE EMPTY! pushing to gameQ");
    gameQ.push_back(Worker_Start::compose(inv.getIID() , inv.getOptions(), inv.getPlayerCount(), ips, passes, teamstr.str(), idstr, "", "" )); /** fixme NO WORKER PORT???? */
    
  }
  tell_worker_info();
  
}


/**
   generates some random passwords for the player
   @todo: generate random passwords :)
 */
void Net_Service::generatePasses(const Net_Invitation& inv, Vector<string>& ips, Vector<string>& passes){
  int num = inv.getPlayerCount();

  for(int i = 0; i < num; i++){
    ips.push_back("dummy");
    passes.push_back("dummy");
  }

}

/*-------------------------------------------- Service Functions -------------------------------*/

/**
   creates a game invitation, and notifies
   ( all online players if it is public or
   buddies if it is invisible )
   that an invitation has been created.
   joins the creater to the game
 */

void Net_Service::serve_chat(Chat& chat)
{
  print("[------------CHAT--------------]");

  if (!isOnline(chat.get_sender())) {
    cout << "### Player (" << chat.get_sender() <<
    ") is not logged in, returning from serve_chat. " << endl;
    return;
  }

  string name = chat.get_sender();
  
  sint4 iid = chat.get_iid();
  string from = chat.get_from();
  string msg = chat.get_msg();
  
  Net_Invitation * n = invitList->get(iid);

  tellInvitation( *n, Chat::compose(iid, from, msg) );
}



/**
   add/remove the given player from senders buddy list
 */
void Net_Service::serve_buddy(Buddy& buddy)
{
  print("[------------BUDDY---------------]");

  if (!isOnline(buddy.get_sender())) return;

  if(buddy.get_buddy_name() == buddy.get_sender()){ // cant add your self to your own buddy list
    tellClient(buddy.get_sender(),Error::compose( "You cannot add yourself as a buddy" ));
    return;
  }

  DB_Player me  (buddy.get_sender());
  DB_Player him (buddy.get_buddy_name());

  if (db->get(me)  < 0) return;

  if (db->get(him)  < 0){
    tellClient(buddy.get_sender(),Error::compose(buddy.get_buddy_name() + " is not a registered player" ));
    return;
  }

  if (buddy.get_add()) { // add
    me.addBuddy(buddy.get_buddy_name());
    tellClient(buddy.get_sender(), Added_Buddy::compose(buddy.get_buddy_name(), isOnline(buddy.get_buddy_name())));
  } else {                // remove
    me.remBuddy(buddy.get_buddy_name());
    tellClient(buddy.get_sender(), Remed_Buddy::compose(buddy.get_buddy_name()));
  }

  db->put(me);

  db->print(); //--
}

void Net_Service::serve_buddies(List& list){
  print("[------------BUDDIES-------------]");

  if (!isOnline(list.get_sender())) return;

  DB_Player player(list.get_sender());

  if (db->get(player) < 0){
    print("Could not retrieve record of " + list.get_sender() + \
          "\n\tHow is he/she logged in ?!");
    return;
  }

  Vector<string>*buddyList = player.getBuddies();
  Vector<uint1>   onlineList;

  FORALL(*buddyList, i){
    if(isOnline(*i)){
      onlineList.push_back(true);
    } else{
      onlineList.push_back(false);
    }
  }

  tellClient(list.get_sender(), Buddies::compose(*buddyList, onlineList));
}


/**

 */
void Net_Service::serve_who(List& list){
  print("[------------WHO-----------------]");
  string who;

  if(!isOnline(list.get_sender())){ return; }
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(client_mut)) ERR("mutex 1");
  FORALL (allPlayers, i){
    who += " ";
    who += (*i).first;
  }
  if (SDL_mutexV(client_mut)) ERR("mutex 2");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION

  tellClient(list.get_sender(), Who_List::compose(who));
}



/**
   if the sender is logged in and has level ROOT
   shut down the service
 */
void Net_Service::serve_down(Down& down)
{
  PlayerList::iterator root;
  root = allPlayers.find(down.get_sender());
  // todo : db->close();
  if (root == allPlayers.end()) return;

  if((*root).second.second == (int) ROOT){
    shutDown();
  }

  signal_handler_shut_down(0);
  signal_handler_exit(0);
}



/**
   creates a game invitation, and notifies
   ( all online players if it is public or
   buddies if it is invisible )
   that an invitation has been created.
   joins the creater to the game
 */

void Net_Service::serve_invite(Invite& invite)
{
  print("[------------INVITE--------------]");

  // Commented out, because it's done in ggs, we don't need another level of authentication (lanctot 2007-06-19)
  // authenticate : player is logged in
  if (!isOnline(invite.get_sender())) {
    cout << "### Player (" << invite.get_sender() <<
    ") is not logged in, returning from serve_invite. " << endl;
    return;
  }

  string name = invite.get_sender();

  Net_Invitation * n = new Net_Invitation(invite);
  
  cout << "Game type is " << n->getGameType() << endl;

  // "alias" may be null because when someone created the game and selected
  // "anonymized ids" (creategame.ui.h) they did not enter an alias. In this case
  // we will use their username.
  string alias;
  if( invite.get_alias() != "null" ) alias = invite.get_alias();
  else alias = invite.get_sender();

  sint4 pid = n->addPlayer( invite.get_sender(), alias );
  invitList->add(n);
  fprintf(stderr, "___*** Invite Created with ID: %d\n", n->getIID());

  if(n->getInvisible()){
    DB_Player player(invite.get_sender());

    if(db->get(player)>0){
      tellBuddies(player, Invitations::compose(n->getInfo()));
    } else{
      ERR("Client " + invite.get_sender() + " is online but not in the database !!!");
    }

  } else{
    // send list of matches
    tellAll(Invitations::compose(n->getInfo()));
  }

  // disabled.. join is enough
  //string sender = invite.get_sender();
  //string msg = Player_List::compose(n->getIID(), n->getPresent(), n->getAliases(), n->getReady(), n->getTeamAssignments());
  //tellClient(sender, msg);

  tellInvitation( *n, Joined::compose(n->getIID(), pid, n->getAnony(), alias) );

  invitList->print();
  
}




/**
   checks if the player is the owner
   kills the game
   notifies all members that it has been killed
 */

void Net_Service::serve_abort(Abort& abort)
{
  // authenticate : player is the owner
  //                player is logged in
  print("[------------ABORT---------------]");

  if (!isOnline(abort.get_sender())) return;

  Net_Invitation * inv = invitList->get((uint4)abort.get_iid());
  if (!inv){
    char s[256];
    sprintf(s,"Cannot abort:\n\tGame with id %d could not be found",abort.get_iid());
    tellClient(abort.get_sender(),Error::compose(string(s)));
    return;
  }

  if (abort.get_sender() == inv->getOwner()) {
    invitList->del(abort.get_iid());
    fprintf(stderr, "___*** Invite Deleted with ID: %d\n", abort.get_iid());
    tellAll(Aborted::compose(abort.get_iid()));
  } else{
    char s[256];
    sprintf(s,"Cannot abort:\n\tYou are not the owner of game %d",abort.get_iid());
    tellClient(abort.get_sender(),Error::compose(string(s)));
  }

  invitList->print(); //--
  tell_worker_info();
}





/**
   checks if the player is the owner
   bans the reqested member
   notifies all members
 */

void Net_Service::serve_ban(Ban& ban)
{
  print("[------------BAN----------------]");
  // authenticate : player is the owner
  //                player is logged in

  if (ban.get_pid() == 0) return; // can't ban yourself... i donno why, just dont like it

  if (!isOnline(ban.get_sender())) return;

  Net_Invitation * inv = invitList->get((uint4)ban.get_iid());

  if (inv == NULL) {
    stringstream msg;
    msg << "invitation with iid " << ban.get_iid() << " not found";
    print(msg.str());
    tellClient(ban.get_sender(), Error::compose(msg.str()));
    return;
  }

  if (ban.get_sender() != inv->getOwner()) {
    string msg = ban.get_sender() + " is not the owner\n\tthe owner is " + inv->getOwner();
    print(msg);
    tellClient(ban.get_sender(), Error::compose(msg));
  } else {
    tellInvitation(*inv, Banned::compose(inv->getIID(), ban.get_pid()));
    inv->banPlayer(ban.get_pid());
    inv->removePlayer(ban.get_pid());
  }
  inv->print();
}




/**
   checks if player is registered
   adds player to the game
   notifies all members
 */

void Net_Service::serve_join(Join& join)
{
  print("[------------JOIN----------------]");
  // authenticate : player is logged in
  string name = join.get_sender();
  
  if (!isOnline(join.get_sender())) return;

  cout << name << " is joining iid=" << join.get_iid() << endl; 

  Net_Invitation * inv = invitList->get((uint4)join.get_iid());
  fprintf(stderr, "___*** Join ID: %d\n", join.get_iid());

  if (inv) {
    if( inv->isPlayerBanned(join.get_sender()) ){
      // do not allow banned player to rejoin
      tellClient(join.get_sender(), Error::compose("You are banned from this match!"));
      return;
    }
    
    if( join.get_passprotected() ){
      if(join.get_pass() != inv->getPass() /*&& join.get_passprotected()*/ ){
        tellClient(join.get_sender(), Error::compose("Incorrect password for this game"));
        cout << "reallPass |" << inv->getPass()  << "|" << endl;
        cout << "givenPass |" << join.get_pass() << "|" << endl;
        return;
      }
    }
    
    if (inv->getStarted())
    {
      //game is already started :( 
      cout << "Sorry, game has already started. " << endl;
      tellClient(join.get_sender(), Error::compose("join1"));
      return;
    }
      

    // "alias" may be null because they did not enter an anonymizing id
    // when prompted to in AnonymizingDialog (anonymizingdialog.ui.h).
    // In this case we will use their username.
    string alias;
    if( join.get_alias() != "null") alias = join.get_alias();
    else alias = join.get_sender();

    int pid = inv->addPlayer(join.get_sender(), alias);
    
    cout << "join: pid = " << pid << endl;
    
    if (pid >= 0) {
      
      if( join.get_anony_ids() ){

        tellClient(join.get_sender(), Player_List::compose(inv->getIID(), inv->getPresent(), 
          inv->getAliases(), inv->getReady(), inv->getTeamAssignments()));
      }
      else{
        alias = join.get_sender();
        tellClient(join.get_sender(), Player_List::compose(inv->getIID(), inv->getPresent(), 
          inv->getPlayers(), inv->getReady(), inv->getTeamAssignments()));
      }

      tellInvitation( *inv, Joined::compose(inv->getIID(), pid, inv->getAnony(), alias) );
    }
  }

  if (inv) inv->print();
}

void Net_Service::serve_join_team(Join_Team& join_team)
{
  print("[------------JOINTEAM----------------]");
  // authenticate : player is logged in
  string name = join_team.get_sender();

  cout << "name is " << name << endl; 

  if (!isOnline(name)) return;
  
  int iid = join_team.get_iid();
  int teamNum = join_team.get_teamNum();
  
  Net_Invitation * inv = invitList->get((uint4)iid);
  //fprintf(stderr, "___*** Join ID: %d\n", join.get_iid());
  
  bool success = inv->joinTeam(name, teamNum);
  
  if (success) 
  {
    if (teamNum > 0)
      cout << "iid=" << iid << ", " << name << " has joined team " <<  teamNum << endl;
    else
      cout << "iid=" << iid << ", " << name << " has left them team!" << endl;

    //tellClient(join.get_sender(), Join_Team::compose(iid, teamNum));
    tellInvitation( *inv,  Join_Team::compose(iid, teamNum, name));
  }
  else
  {
    cout << "iid=" << iid << ", " << name << " has failed to join team " <<  teamNum << endl;
    tellClient(join_team.get_sender(), Join_Team::compose(iid, -1, name));
  } 
}



/**
   if player is in the game
   removes player
   notifies all members
 */

void Net_Service::serve_leave(Leave& leave)
{
  print("[------------LEAVE---------------]");
  // authenticate : the sender must match the player alias
  //                player is logged in

  if (!isOnline(leave.get_sender())) return;

  Net_Invitation * inv = invitList->get((uint4)leave.get_iid());

  if (inv) {

    sint4 pid = inv->removePlayer(leave.get_sender());
    
    cout << "Player (" << leave.get_sender() << ") leaving his/her team" << endl; 
    inv->leaveTeam(leave.get_sender());    

    if(pid < 0) {
      char s[256];
      sprintf(s,"You are not a member of game %d", leave.get_iid());
      tellClient(leave.get_sender(), Error::compose(string(s)));
      return;
    }

    tellInvitation(*inv, Left::compose(inv->getIID(), pid));
    tellClient(leave.get_sender(), Left::compose(inv->getIID(), pid));

    if (inv->getPlayerCount() == 0 /* || pid == 0 */) { 
       // everybody gone :(... kill game
      invitList->del(leave.get_iid());
      fprintf(stderr, "___*** Invite Deleted with ID: %d\n", leave.get_iid());
      tellAll(Aborted::compose(leave.get_iid()));
      return;
    }
    inv->print(); //--
  }
}

/**
   if player is in the game
   player status is readied
   all players are notified
 */
void Net_Service::serve_ready(Ready& ready)
{
  print("[------------READY---------------]");
  //authenticate : player is logged in
  //               player is in game

  if (!isOnline(ready.get_sender())) {
    cout << ready.get_sender() << " is not online! returning..." << endl;
    return;
  }

  Net_Invitation * inv = invitList->get((uint4)ready.get_iid());

  if (inv) {
    
    if (inv->allReady()) {
    	cout << "iid = " << inv->getIID() << ", # readied players has reached # maximum players... sorry" << endl;
      tellClient(ready.get_sender(), Error::compose("readied1 " + to_string(inv->getIID())));
      return;
    }
     
    sint4 alias = inv->setReady(ready.get_sender(), true);
    if(alias < 0){
      stringstream msg;
      msg << "not currently in game with iid " << ready.get_iid();
      print( "Player " + ready.get_sender() + " is " + msg.str());
      tellClient(ready.get_sender(), Error::compose("You are " + msg.str()));
      return;
    }
    tellInvitation(*inv, Readied::compose(inv->getIID(), alias));
    
    // handled in serve_start
    //if (inv->allReady()) {
    //  startGame(*inv);
    //}
  }

  if(inv) inv->print(); //--
  
 }

void Net_Service::serve_game_start(Game_Start& game_start)
{
  print("[------------GAME_START---------------]");
  if (!isOnline(game_start.get_sender())) {
    cout << game_start.get_sender() << " is not online! returning..." << endl;
    return;
  }
  
  Net_Invitation * inv = invitList->get((uint4)game_start.get_iid());  
  
  
  if (inv->getGameType() == 1 && inv->getReadyCount() != 1)
  {
      cout << "Game type is " << inv->getGameType() << " and ready count != 1" << endl;
      stringstream msg;
      msg << "" << inv->getIID();
      tellClient(game_start.get_sender(), Error::compose("gstart1 " + msg.str()));      
  }
  else if (inv->getGameType() != 1 && !inv->readyToStart())
  {
      cout << "Game type is " << inv->getGameType() << " and team count < 2" << endl;
      stringstream msg;
      msg << "" << inv->getIID();
      tellClient(game_start.get_sender(), Error::compose("gstart1 " + msg.str()));      
  }
  else
  {
    cout << "Everyone is ready, I'm starting the game!!" << endl;   
    startGame(*inv);
  }
}


/*
 *  Notifies all players of this game that the game
 *  has ended. If it ended successfully, results are
 *  passed on as well; if not, an error message is passed.
 */
void Net_Service::serve_game_end(Game_End& game_end) {
  print("[------------GAME_END-------------]");
  //tell the clients that this game has ended.
  //have a bool whether the results are real or not

  Net_Invitation * inv = invitList->get((uint4)game_end.get_iid());
  if(inv){
    tellInvitation(*inv, Game_End::compose(inv->getIID(), game_end.getIfSuccess(), game_end.get_results(), game_end.get_ip(), game_end.get_port()));
  }

  // start any waiting games
  if (gameQ.size() > 0) {
    
    popFromGameQ(game_end.get_sender(), game_end.get_ip(), to_string(game_end.get_port()));

  } else {
    workerQ.push_back(make_pair(game_end.get_sender(), make_pair(game_end.get_ip(),to_string(game_end.get_port()))));
    print("*** GAME QUEUE EMPTY! pushing to workerQ");
    cout << "*** Adding worker: " << game_end.get_sender() << " at " << game_end.get_ip() << " at port " << game_end.get_port() << endl;
  }
  tell_worker_info();
}

void Net_Service::popFromGameQ(string workerName, string host, string port){
  print("*** POPPING from gameQ");

  Worker_Start ws;
  ws.parse("U: " + gameQ.front());
  string new_ws_msg = ws.recompose(host, port); 
  
  Net_Invitation *inv = invitList->get((uint4)ws.getIID());
  if(inv){

    readyGameQ[workerName] = *inv;

    string info;
    const Vector<string> &all = (*inv).getPlayers();
    const Vector<uint1>  present = (*inv).getPresent();
    Vector<string> passes = (ws).get_passv();
    
    int j = 0;
    FORU(i, all.size()){
      if(present[i]){       
        info = Game_Info::compose(host, passes[j], (*inv).getIID(), ws.getTeamStr(), "-port " + port);
        j++;
        this->send("tell " + all[i] + " " + info);
      }
    }
    cout << "NEW WORKER_START STRING IS: " << new_ws_msg << endl;


    tellClient(workerName, new_ws_msg);     
    gameQ.pop_front();
  }
}

/**
   if player is in the game
   player status is unreadied
   all players are notified
 */

void Net_Service::serve_unready(Unready& unready)
{
  print("[------------UNREADY-------------]");
  // authenticate : player is in game
  //                player is logged in

  if (!isOnline(unready.get_sender())) return;

  Net_Invitation * inv = invitList->get((uint4)unready.get_iid());
  if (inv){
    sint4 alias = inv->setReady(unready.get_sender(), false);
    tellInvitation(*inv, Unreadied::compose(inv->getIID(), alias));
  }
  if(inv) inv->print(); //--
}

/**
 * Only use GGS notify messages for log off info.
 * Login info for the orts.net service is provided
 * by the login message when players specifically
 * login to orts, not just to GGS.
 */
void Net_Service::serve_notify(GGS_Notify& notify)
{
  print("[------------NOTIFY--------------]");
  cout << "Received notify from " << notify.get_sender() 
  << ", get_online = " << notify.get_online() 
  << ", get_name = " << notify.get_name() 
  << endl;

  if (!notify.get_online()) {

    cout << "Player is logging out." << endl; 

    // not using this
    //DB_Player player(notify.get_name());
    //if (db->get(player) < 0) return;
    bool isWorker = false; 
   
    PlayerList::iterator pl = allPlayers.find(notify.get_name());
    if( pl == allPlayers.end()) {
      cout << "Not an Orts player..." << endl;
      return;
    }

    if(((*pl).second).second == WORKER){
      isWorker = true;
    }

    // remove player from online list
    allPlayers.erase(notify.get_name()); //could just erase the iterator at this point too KM07  


    // tell buddies player went offline
    tellAll(Login_Notify::compose(false, notify.get_name()));

    // remove from worker queue if worker:
    list< pair<string, pair<string, string> > >::iterator it = workerQ.begin(), end = workerQ.end();
    for( ; it != end; ++it) {
      if(it->first == notify.get_name()) {
        cout << "Removing worker " << notify.get_name() << " from workerQ!\n";
        workerQ.erase(it);


        tell_worker_info();
        return;  // if worker, don't need to go through all the invitations
      }
    }
    if(isWorker) {
      //catches the workers who aren't in the workerQ
      //must send the game_end message to the games which they were servicing
      ReadyGameQ::iterator g;
      g = readyGameQ.find(notify.get_name());
      if(g != readyGameQ.end()){
        const Net_Invitation& inv = (*g).second;
        tellInvitation(inv, Game_End::compose(inv.getIID(), false, Game_End::INVALID_RES, "", -1));
      }
      return;
    }
    // remove player from all invitations
    Net_Invitation * inv;

    cout << "Iterating through invitations..." << endl;

    for(uint4 i = 0; i < invitList->getSize(); i++){
      inv = invitList->get(i);      

      if(inv){

        cout << "  iid=" << inv->getIID() << endl;

        int pid = inv->removePlayer(notify.get_name());

        if(pid < 0){
          cout << "   pid < 0, next!" << endl;
          continue;
        }

        tellInvitation(*inv, Left::compose(i, pid));
        inv->print();

        if (inv->getPlayerCount() == 0 /*|| pid == 0*/ ) { // everybody gone :(... kill game

          cout << "  Everybody's gone, killing the game!" << endl;

          invitList->del(i);
          fprintf(stderr, "___*** Invite Deleted with ID: %d\n", i);
          tellAll(Aborted::compose(i));
        }
      }
    }
  }
}




/**
   root can register super
   super can register general
   check user (name, pass, level)
   add the new registration to the database
 */

void Net_Service::serve_register(Register& reg)
{
  print("[------------REGISTER------------]");
  // add player to the database

  // case 1 : new entry
  //          level = general || worker        open to everybody
  //          level = super                    open to root only

  // case 2 : existing entry (changing password)
  //          level = general || worker        open to super,root, and the record owner
  //          level = super                    open to super,and root
  //          level = root                     open to root

  // retrieve record
  DB_Player player(reg.get_name());

  bool exists = (db->get(player) > 0);
  bool allow  = false;


  PlayerList::iterator i = allPlayers.find(reg.get_sender());

  if(i != allPlayers.end() && ((*i).second).second == ROOT){
    allow = true;
  }

  if(i == allPlayers.end() && exists){
    tellClient(reg.get_sender(), Error::compose("User name " + reg.get_name() + " already exists"));
    return;
  }

  if(exists && !allow){

    if( i != allPlayers.end() ){
      // if the entry exists allow it if
      //    level is not changing && sender is the entry owner
      //    sender has a higher level && requested level is < sender

      if(reg.get_level() == player.getLevel()){     // level not changing
        allow = true;
      } else{
        tellClient(reg.get_sender(), Error::compose("You cannot change your registration level"));
        return;
      }

      if( !((((*i).second).second >  reg.get_level()) || // u have a higher level
            (reg.get_sender() == player.getName()   )) ){ // u are the owner
        tellClient(reg.get_sender(), Error::compose("You cannot change password unless you are the owner of the record or you have a higher level"));
        return;
      }
    }

  } else{
    if(!allow){
      // if the entry does not exist allow it if the user wants to register
      // as worker or general
      allow = (reg.get_level() == WORKER || reg.get_level() ==  GENERAL);
    }
  }

  if(!allow){
    tellClient(reg.get_sender(), Error::compose("This is an illegal request."));
    return;
  }

  /* vvvvvvvvvvvvvvvvvvvvv request is legal now add the entry vvvvvvvvvvvvvvvvvvvvvvvvvv */

  if(reg.get_level() ==  WORKER){
    DB_Player worker(reg.get_name(), reg.get_pass(),  WORKER);
    db->put(worker);
  } else{
    if(exists){
      player.setPass(reg.get_pass());
      db->put(player);
    } else{
      DB_Player entry(reg.get_name(), reg.get_pass(), (Level_)reg.get_level());
      db->put(entry);
    }
  }

}



/**
   authenticate the player
   (nick, pass, level)
   add them to the online players list
   tell buddies that player is online ---- fixme
   ask GGS to notify the service when this player goes offline
 */

void Net_Service::serve_login(Login& log)
{
  print("[------------LOGIN---------------]");
  
  
    if (isOnline(log.get_sender())){
    // player is already logged in
   
    // send list of matches
    FORU(i, invitList->getSize()){
      
      cout << "(top) Iterating through list of matches, i = " << i  << endl;
       
      Net_Invitation * n = invitList->get(i);
      
      if (n != NULL) {
        
        tellClient(log.get_sender(), Invitations::compose(n->getInfo()));
        
        if (n->isPlayerBanned(log.get_sender())) {
        
          tellClient(log.get_sender(), Error::compose("banned1 " + to_string(n->getIID())));
        }
                
      }
             
    }

    return;
  }

  // check password
  DB_Player player(log.get_sender());

  bool found = true;
  /*  db->get(player) > 0                   &&
     log.get_sender()  == player.getName() &&
     log.get_pass()  == player.getPass()   &&
     log.get_level() == player.getLevel(); */

  if(found){
    print(log.get_sender() + " logged in successfuly");
    
  } else{
    print(log.get_sender() + " failed to login");
  }

  if(found) {
    allPlayers[log.get_sender()] = make_pair(log.get_sender(), log.get_level());
    if(log.get_level() ==  WORKER){
      
      // notfy worker that the login was success ful
      tellClient(log.get_sender(), Login_Notify::compose(true, log.get_sender()));
     
      // start any waiting games
      if (gameQ.size() > 0) {
        popFromGameQ(log.get_sender(), log.get_ip(), to_string(log.get_port()));
//        print("*** POPPING from gameQ");
//        tellClient(log.get_sender(), gameQ.front());
//        gameQ.pop_front();
      } else {
        workerQ.push_back(make_pair(log.get_sender(), make_pair(log.get_ip(),to_string(log.get_port()))));
        
        
        tell_worker_info();
        print("*** GAME QUEUE EMPTY! pushing to workerQ");
        cout << "*** Adding worker: " << log.get_sender() << " at " << log.get_ip() << " at port " << log.get_port() << endl;
      }

    } else {
     // allPlayers[log.get_sender()] = make_pair(log.get_sender(), log.get_level());

      FORU(i, invitList->getSize()){
        cout << "(top) Iterating through list of matches, i = " << i  << endl;
        Net_Invitation * n = invitList->get(i);        
        if (n != NULL) {
          tellClient(log.get_sender(), Invitations::compose(n->getInfo()));
          if (n->isPlayerBanned(log.get_sender())) 
            tellClient(log.get_sender(), Error::compose("banned1 " + to_string(n->getIID())));
        }
      }

      // tellBuddies that player is online
      tellAll(Login_Notify::compose(true, log.get_sender()));
      // ask GGS to notify service when player goes off line
      tellGGS(GGS_Notify::compose(log.get_sender(), true));

    }
    
    //update player's count of number of available worker
    tell_worker_info();
      
    
  } else{ // not found
    tellClient(log.get_sender(), Error::compose("Login failed.\nInvalid information."));
  }
}

void Net_Service::serve_invitations(List& list){
  Net_Invitation * current;
  FORU(i, invitList->getSize()){
    current=invitList->get(i);
    if (current!=0) {
      if (!current->getInvisible()) { //fix so if on buddy list send anyways
        tellClient(list.get_sender(), Invitations::compose(current->getInfo()));
      }
    }
  }
}

void Net_Service::serve_tell(Tell& tell){
  print("[------------TELL----------------]");
  Net_Invitation * inv = invitList->get(tell.get_iid());
  if (!inv){
    stringstream msg;
    msg << "invitation with iid " << tell.get_iid() << " not found" << endl;
    print(msg.str());
    tellClient(tell.get_sender(), Error::compose(msg.str()));
    return;
  }
  tellInvitation(*inv, Tell::compose(tell.get_iid(), tell.get_pid(), tell.get_out_msg()));
}

void Net_Service::serve_worker_ready(Worker_Ready& wready){
  print("[------------WREADY--------------]");
  ReadyGameQ::iterator i;

  i = readyGameQ.find(wready.get_sender());

  if(i == readyGameQ.end()){
    tellClient(wready.get_sender(),Error::compose("Worker_Ready message invalid: No pending game for you"));
    return;
  }

  const Net_Invitation& inv = (*i).second;
  tellInvitation((*i).second, Game_Start::compose((int)inv.getIID()));
}

void Net_Service::serve_worker_available(Worker_Available& wavailable) {

  print("[------------WAVAILABLE----------]");

  workerQ.push_back(make_pair(wavailable.get_sender(), make_pair(wavailable.get_ip(), wavailable.get_port())));
  
  
  tell_worker_info();
  print("*** Pushing to workerQ, IP: " + wavailable.get_ip());

}

/*--------------------------------------- tell functions -------------------------------------------- */

void Net_Service::tellInvitation(const Net_Invitation& inv, const string& message)
{
  if (debug)
    cout << "Telling invitation (" << inv.getIID() << ") message: " << message << endl;

  const Vector<string> &all = inv.getPlayers();
  const Vector<uint1>  present = inv.getPresent();

  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(client_mut)) ERR("mutex 1");
  FORU(i, all.size()){
    if(present[i]){
      this->send("tell " + all[i] + " " + message);
    }
  }
  if (SDL_mutexV(client_mut)) ERR("mutex 2");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION
}

void Net_Service::tellAll(const string& message)
{
  if (debug)
    cout << "Telling all message: " << message << endl;

  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(client_mut)) ERR("mutex 1");
  FORALL (allPlayers, i) {
    if((*i).second.second != WORKER) //KM07
      this->send("tell " + (*i).first + " " + message);
  }
  if (SDL_mutexV(client_mut)) ERR("mutex 2");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION
}

void Net_Service::tellBuddies(DB_Player& player, const string& message)
{
  if (debug)
    cout << "Telling all player " << player.getName() << "'s buddies message: " << message << endl;

  const Vector<string> &buddies = *player.getBuddies();

  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(client_mut)) ERR("mutex 1");
  FORALL (buddies, i) this->send("tell " + *i + " " + message);
  if (SDL_mutexV(client_mut)) ERR("mutex 2");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION
}



