#ifndef Net_Message_Service_H
#define Net_Message_Service_H

// $Id: Net_Service.H 5743 2007-08-23 16:22:46Z krysta $

// (c) Michael Buro 2003, mburo@cs.ualberta.ca, licensed under the GPL

/** the SERVICE
 * manages online game creation:
 *  - players create invitations (join/leave/ready/unready/ban)
 *  - creates a game when all spots are filled and all players are ready
 *
 * maintains a list of workers and balances work loads between them
 *  - coordinates game starting and facilites connection between players and workers
 *  - workers log(on/off)
 *
 * maintains player and worker records
 *  - player and worker passwords
 *  - player rating
 *  - players buddy list
 */

#include "Global.H"
#include "ShutDown.H"
#include "Net_Protocol.H"
#include "Net_Message.H"
#include "Net_InvitList.H"
#include "Net_Invitation.H"
#include "Net_DB.H"
#include "Net_DBTypes.H"
#include "Net_GGSClient.H"


class Net_Service : public Net_GGSClient {

public:

  Net_Service(std::string& server,
              int port,
              std::string& alias,
              std::string& pass,
              std::string& dbFile,
              std::string& rootNick_,
              std::string& rootPass_,
              bool verbose_);

  virtual ~ Net_Service();

  /**
   * connects to the GGS server
   * and starts litening for service requests
   * returns -1 if the starting failed
   */
  sint4 start();

protected:

  enum { MAX_THREADS = 1 };    /** the largest number of allowed threads   */
  std::list<std::string> queue;         /** requests waiting to be handeled          */

  Net_InvitList * invitList;   /** list of pending invitations             */
  Net_DB * db;                 /** the data base for players and wokers    */

  SDL_mutex * q_mut;           /** protects requst queue access            */
  SDL_mutex * client_mut;      /** protects the GGSClient                  */

  std::string rootNick;            /** the GGS nick that root should use       */
  std::string rootPass;            /** the pass to log root in                 */

  bool verbose;               /** mode if quite or verbose                */

  typedef __gnu_cxx::hash_map<std::string, std::pair<std::string, int>, StringHash> PlayerList;
  PlayerList allPlayers;


  std::list<std::pair<std::string, std::pair<std::string, std::string> > >  workerQ;      /** worker are added to this queue             */
  typedef __gnu_cxx::hash_map<std::string, Net_Invitation, StringHash> ReadyGameQ;
  // ^--panInfo--            <workerName, pendingGame    >

  ReadyGameQ readyGameQ;

  std::list<std::string>    gameQ;       /** games that are ready, waiting for a worker */

  SDL_sem * serve_sem;        /** semaphore for number of unserved requests  */

  /** it is run as a thread
   * spins, probing for messages
   * once a message is recieved it creats a thread of dispatchRequest
   */
  sint4 listen();

  /** gets the number of available workers!*/
  int num_avail_workers() { return (int)workerQ.size(); } //KM07
  
  /** tells all clients the updated worker info/gameQ status */
  void tell_worker_info(); //KM07
  
  
  /** shuts down the service */
  void shutDown();

  /**
     prints the given std::string if running on verbose mode
   */
  void print(const std::string& print);

  /**
   * it is run as a thread
   * dipatches the request to the appropriate function to serve it
   */
  friend int::dispatch(void * data);

  /**
     picks a worker, starts the game
   */
  void startGame(Net_Invitation& inv);
  void generatePasses(const Net_Invitation& inv, Vector<std::string>& ips, Vector<std::string>& passes);

  /**
   * Pops a game from the gameQ (list of games waiting for workers)
   * and starts it on worker with name @param workerName
   */
  void popFromGameQ(std::string workerName, std::string host, std::string port);
  
  inline bool isOnline(const std::string& name);

  // servers
  void serve_buddy(Buddy& buddy);
  void serve_chat(Chat& down);
  void serve_down(Down& down);
  void serve_invite(Invite& invite);
  void serve_abort(Abort& abort);
  void serve_ban(Ban& ban);
  void serve_game_start(Game_Start& game_start);
  void serve_game_end(Game_End& game_end);
  void serve_join(Join& join);
  void serve_join_team(Join_Team& join_team);
  void serve_leave(Leave& leave);
  void serve_ready(Ready& ready);
  void serve_unready(Unready& unready);
  void serve_invitations();
  void serve_games();
  void serve_tell(Tell& tell);
  void serve_notify(GGS_Notify& notify);
  void serve_register(Register& reg);
  void serve_login(Login& log);
  void serve_worker_ready(Worker_Ready& wready);
  void serve_worker_available(Worker_Available& wavailable);

  void serve_buddies(List& list);
  void serve_who(List& list);
  void serve_games(List& list);
  void serve_invitations(List& list);

  /**
     specialized teller funcions extention to the tell functions
     inherited fro Net_GGSClient
   */
  void tellInvitation(const Net_Invitation& inv, const std::string& message);
  void tellAll(const std::string& message);
  void tellBuddies(DB_Player& player, const std::string& message);

};

#endif

