#include <algorithm>
#include "Global.H"

#include "NUS_SimulationState.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameObj.H"

#include "NUS_Order.h"

static const bool debug = false;

namespace NUS_Simul{

	SimulationState::SimulationState(const SimulationState& b){
		//if (debug) std::cout <<  "SimulationState: start cloning of players" << std::endl;
		
		FORALL(b.players,it){
			if ((it->second)->getGroups().empty()) continue;
			
			players[it->first] = (it->second)->clone();
		}

		//if (debug) std::cout <<  "SimulationState: done cloning of players" << std::endl;

		groupSize = b.groupSize;
		smartJoinSize = b.smartJoinSize;
		smartJoin = b.smartJoin;
		useMoveForAttack = b.useMoveForAttack;
		findspot = b.findspot;
		simAI = b.simAI;
	}

	SimulationState::~SimulationState(){
		cleanup();
	}

	SimulationState& SimulationState::operator= (const SimulationState& b){
		if (this == &b) return *this;

		cleanup();

		FORALL(b.players,it){
			players[it->first] = (it->second)->clone();
		}
		
		groupSize = b.groupSize;
		smartJoinSize = b.smartJoinSize;
		smartJoin = b.smartJoin;
		useMoveForAttack = b.useMoveForAttack;
		findspot = b.findspot;
		simAI = b.simAI;
		
		return *this;
	}

	SimulationState* SimulationState::clone(){
		return new SimulationState(*this);
	}

	void SimulationState::cleanup(){
		FORALL(players,it){
			delete it->second;
			it->second = NULL;
		}
		players.clear();
	}

	Unit* SimulationState::getNearestEnemyUnitOfGroup(uint4 playerID,const UnitVector& units){
		real8 nearest = 2e8;
		Unit* nearestUnit = NULL;

		FORALL(players,it) {
			if((it->second)->getID() == playerID) continue; // don't check our own units
			FORALL((it->second)->getOrtsUnits(),it2){
				FORALL(units,it3){
					real8 dist = (*it3)->distance_to(*(it2->first));
					if(dist < nearest) {
						nearestUnit = it2->first;
						nearest = dist;
					}
				}
			}
				
		}
		return nearestUnit;
	}
	
	//get the nearest unit to enemy unit
	Unit* SimulationState::getNearestUnitToEnemy(uint4 playerID,const UnitVector& units){
		real8 nearest = 2e8;
		Unit* nearestUnit = NULL;

		FORALL(players,it) {
			if((it->second)->getID() == playerID) continue; // don't check our own units
			FORALL((it->second)->getOrtsUnits(),it2){
				FORALL(units,it3){
					real8 dist = (*it3)->distance_to(*(it2->first));
					if(dist < nearest) {
						nearestUnit = *it3;
						nearest = dist;
					}
				}
			}
		}
		return nearestUnit;
	}
	
	

	Unit* SimulationState::getNearestEnemyUnitOfCGOfGroup(uint4 playerID,const UnitVector& units){
		real8 nearest = 2e8;
		Unit* nearestUnit = NULL;

		NUS_TerrainBase::Loc CG(0,0);
		// compute CG
		FORALL(units,uit){
			CG.x += (*uit)->getPosition().x;
			CG.y += (*uit)->getPosition().y;
		}
		
		CG.x /= (float) units.size();
		CG.y /= (float) units.size();
		
		FORALL(players,it) {
			if((it->second)->getID() == playerID) continue; // don't check our own units
			FORALL((it->second)->getOrtsUnits(),it2){
				real8 dist = (it2->first)->distance_to(CG);
				if(dist < nearest) {
					nearestUnit = it2->first;
					nearest = dist;
				}
			}
				
		}
		return nearestUnit;
	}
		
	Group* SimulationState::getNearestGroupOfGroup(uint4 playerID, const Group* group){
		real8 nearest = 2e8;
		Group* nearestGroup = NULL;

		FORALL(players,it) {
			if((it->second)->getID() != playerID) continue; 
			FORALL((it->second)->getGroups(),it2){
				if((*it2) == group) continue;
					real8 dist = group->distance_to(*it2);
					if(dist < nearest) {
						nearestGroup = *it2;
						nearest = dist;
				}
			}
				
		}
		return nearestGroup;
	}
	
	Group* SimulationState::getNearestGroupOfGroup(uint4 playerID, const Group* group, GroupVector& groups){
		real8 nearest = 2e8;
		Group* nearestGroup = NULL;

		FORALL(players,it) {
			if((it->second)->getID() != playerID) continue; 
			FORALL(groups,it2){
				if((*it2) == group) continue;
					real8 dist = group->distance_to(*it2);
					if(dist < nearest) {
						nearestGroup = *it2;
						nearest = dist;
				}
			}
				
		}
		return nearestGroup;
	}
	
	Group* SimulationState::getStrongestFriendlyGroup(uint4 playerID, const Group* group){
		real4 strongest = -2e8;
		Group* strongestGroup = NULL;

		FORALL(players,it) {
			if((it->second)->getID() != playerID) continue; 
			FORALL((it->second)->getGroups(),it2){
				if((*it2) == group) continue;
				NUS_TerrainBase::Loc loc((*it2)->getPosition().x, (*it2)->getPosition().y);
				real4 strength = getFriendlyInfluence(playerID, loc);
				if(strength > strongest) {
					strongestGroup = *it2;
					strongest = strength;
				}else if(strength == strongest ){
					if(group->distance_to(strongestGroup) > group->distance_to(*it2)){
						strongestGroup = *it2;
						strongest = strength;
					}
				}
			}
				
		}
		return strongestGroup;
	}

	Group* SimulationState::getStrongerFriendlyGroup(uint4 playerID, const Group* group, GroupVector &groups){
		real8 distanceToStrongerGroup = 2e8;
		Group* nearestStrongerGroup = NULL;
		
		NUS_TerrainBase::Loc groupPosition(group->getPosition().x, group->getPosition().y);
		
		real4 myStrength = getFriendlyInfluence(playerID, groupPosition);

		FORALL(players,it) {
			if((it->second)->getID() != playerID) continue; 
			FORALL(groups,it2){
				if((*it2) == group) continue;
				NUS_TerrainBase::Loc loc((*it2)->getPosition().x, (*it2)->getPosition().y);
				real4 strength = getFriendlyInfluence(playerID, loc);
				
				if(strength >= myStrength){
					
					real8 dist = (group->getPosition()-(*it2)->getPosition()).length();
					
					if (dist < distanceToStrongerGroup) {
						nearestStrongerGroup = *it2;
						distanceToStrongerGroup = dist;
					}
				}
			}	
		}
		return nearestStrongerGroup;
	}
		
	//get nearest enemy unit with weaker influence
	Unit* SimulationState::getNearestWeakEnemyUnitOfGroup(uint4 playerID,const UnitVector& units, uint4 groupsize){
		real8 nearest = 2e8;
		Unit* nearestUnit = NULL;
		Player* p;
		FORALL(players,it) {
			if((it->second)->getID() == playerID) continue; // don't check our own units
			FORALL((it->second)->getOrtsUnits(),it2){
				if((it->second)->getInfluence((it2->first)->getPosition()) <= groupsize)
					FORALL(units,it3){
						real8 dist = (*it3)->distance_to(*(it2->first));
						if(dist < nearest){
							nearestUnit = it2->first;
							nearest = dist;
						}
					}
			}
				
		}
		return nearestUnit;
	}

	Unit* SimulationState::getAverageNearestEnemyUnitOfGroup(uint4 playerID,const UnitVector& units){
		real8 nearest = 2e8;
		Unit* nearestUnit = NULL;

		FORALL(players,it) {
			if((it->second)->getID() == playerID) continue; // don't check our own units
			FORALL((it->second)->getOrtsUnits(),it2){
				real8 totalDistance = 0.0;
				FORALL(units,it3){
					totalDistance += (*it3)->distance_to(*(it2->first));
				}
				if(totalDistance < nearest){
					nearest = totalDistance;
					nearestUnit = it2->first;
				}
			}
		}
		return nearestUnit;
	}
	
	Base* SimulationState::getNearestEnemyBaseOfGroup(uint4 playerID,const UnitVector& units){
		real8 nearest = 2e8;
		Base* nearestBase = NULL;

		FORALL(players,it) {
			if((it->second)->getID() == playerID) continue; // don't check our own units
			FORALL((it->second)->getOrtsBases(),it2){
				FORALL(units,it3){
					real8 dist = (*it3)->distance_to(*(it2->first));
					if(dist < nearest) {
						nearestBase = it2->first;
						nearest = dist;
					}
				}
			}
		}
		return nearestBase;
	}

	Base* SimulationState::getNearestEnemyBaseOfGroup(uint4 playerID,const Group *group){
		real8 nearest = 2e8;
		Base* nearestBase = NULL;

		Vector3D groupPosition(group->getPosition());
		
		FORALL(players,it) {
			if((it->second)->getID() == playerID) continue; // don't check our own units
			FORALL((it->second)->getOrtsBases(),it2){
				real8 dist = (it2->first->getPosition() - groupPosition).length();
				if(dist < nearest) {
					nearestBase = it2->first;
					nearest = dist;
				}
			}
		}
		return nearestBase;
	}
	
	real4 SimulationState::getEnemyInfluence(uint4 playerID, NUS_TerrainBase::Loc loc){
		real4 influence = 0.0;
	//	if(useInfluence){
			FORALL(players, it){
				if((it->second)->getID() == playerID) continue; // don't check our own units
				influence += (it->second)->getInfluence(loc);
			}
			return influence;
	//		}else ERR("This strategy need to set influence");
	}
	
	
	real4 SimulationState::getFriendlyInfluence(uint4 playerID, NUS_TerrainBase::Loc loc){
//		if(useInfluence){
			FORALL(players, it){
				if((it->second)->getID() == playerID) 
					return (it->second)->getInfluence(loc);
			}
//		}else ERR("This strategy need to set influence");
		
		ERR("getFriendlyInfluence: Player not found");
		return 0.0;
	}


	NUS_TerrainBase::Loc SimulationState::getCenterOfObjects(const UnitVector& units){

		sint4 maxX = min_sint4;
		sint4 maxY = min_sint4;
		sint4 minX = max_sint4;
		sint4 minY = max_sint4;

		FORALL(units, it) {
			sint4 x = (*it)->position.x;
			sint4 y = (*it)->position.y;

			if (x > maxX) maxX = x;
			if (y > maxY) maxY = y;
			if (x < minX) minX = x;
			if (y < minY) minY = y;
		}

		return NUS_TerrainBase::Loc((maxX + minX)/2, (maxY+minY)/2);
	}

	Unit* SimulationState::getNearestObject(const UnitVector& units,const NUS_TerrainBase::Loc& loc){
		if(units.empty()) return NULL;
		real8 closestDistance = 2e8;
		Unit* closestObject = NULL;
		FORALL(units,it){

			real8 distance = sqrt(pow(((*it)->position.x - loc.x),2) + pow(((*it)->position.y - loc.y),2));
			if(distance < closestDistance){
				closestDistance = distance;
				closestObject = *it;
			}
		}
		return closestObject;
	}

	Unit* SimulationState::getNearestObject(const UnitVector& unitVector,Unit* unit){
		if(unit == NULL) ERR("can't find nearest group to a NULL group");

		sint4 x = unit->position.x;
		sint4 y = unit->position.y;

		return getNearestObject(unitVector,NUS_TerrainBase::Loc(x,y));
	}

	Unit* SimulationState::getNearestObject(const UnitVector& unitVector,Base* base){
		if(base == NULL) ERR("can't find nearest group to a NULL base");

		sint4 x = base->getPosition().x;
		sint4 y = base->getPosition().y;

		return getNearestObject(unitVector,NUS_TerrainBase::Loc(x,y));
	}

	void SimulationState::getNearestObjectsToObject(\
		UnitVector& closestUnits,\
		UnitVector originalUnits,\
		Unit* unit,\
	uint4 numUnits){

		// for more robustness, if we want more objects than there are available in the vector, just select all we can.
		if(numUnits > originalUnits.size()){
			//ERR("can't select more objects than we have in our vector!");
			numUnits = originalUnits.size();
		}
		if(numUnits <= 0) return;

		closestUnits.clear();

		while(closestUnits.size() < numUnits){
			Unit* mostNearestUnit = getNearestObject(originalUnits,unit);
			closestUnits.push_back(mostNearestUnit);
			removeUnitFromVector(mostNearestUnit,originalUnits);
		}
	}

	void SimulationState::getNearestObjectsToObject(\
		UnitVector& closestUnits,\
		UnitVector originalUnits,\
		Base* base,\
	uint4 numUnits){

		// for more robustness, if we want more objects than there are available in the vector, just select all we can.
		if(numUnits > originalUnits.size()){
			//ERR("can't select more objects than we have in our vector!");
			numUnits = originalUnits.size();
		}
		if(numUnits <= 0) return;

		closestUnits.clear();

		while(closestUnits.size() < numUnits){
			Unit* mostNearestUnit = getNearestObject(originalUnits,base);
			closestUnits.push_back(mostNearestUnit);
			removeUnitFromVector(mostNearestUnit,originalUnits);
		}
	}

	bool SimulationState::removeUnitFromVector(Unit* unit, UnitVector& vector){
		Vector<Unit*>::iterator it = std::find(vector.begin(), vector.end(), unit);

		if (it != vector.end()) {
			vector.erase(it);
			return true;
		}

		return false;
	}

	bool SimulationState::subtractVector(const UnitVector& vectorToSubtract,UnitVector& vector){
		bool succesfullyRemovedAllObjects = true;
		FORALL(vectorToSubtract,it){
			if(!removeUnitFromVector(*it,vector)){
				succesfullyRemovedAllObjects = false;
			}
		}
		return succesfullyRemovedAllObjects;
	}

	bool SimulationState::enemyInRange(Group* group){
		Player* owningPlayer = getPlayer(group->getPlayerID());
		assert(owningPlayer != NULL);

		FORALL(players,playerIterator){
			if(owningPlayer->getID() == (playerIterator->second)->getID()) continue; // don't check your own objects...

			BaseVector* playerBases =  (playerIterator->second)->getBases();

			FORALL(*playerBases,it){
				if((*it)->withinRange(group)){
					return true;
				}
			}
			GroupVector playerGroups =  (playerIterator->second)->getGroups();

			FORALL(playerGroups,it){
				if((*it)->withinRange(group)){
					return true;
				}
			}
		}
		return false;
	}

	Player* SimulationState::getPlayer(uint4 playerID){
		PlayerMap::iterator it;
		it = players.find(playerID);
		if(it != players.end()){
			return it->second;
		}else{
			return NULL;
		}
	}

	void SimulationState::updateObjects(real8 lastMoveUpdateMillis, real8 currUpdateMillis){

		if (debug) std::cout <<  "SimulationState::updateObjects: moving objects" << std::endl;
		
		moveObjects(currUpdateMillis - lastMoveUpdateMillis);

		if (debug) std::cout <<  "SimulationState::updateObjects: resolving collisions" << std::endl;

		resolveCollisions(currUpdateMillis);

		// now we delete any objects that have < 0 hitpoints
		//deleteDeadObjects();

		if (debug) std::cout <<  "SimulationState::updateObjects: update orders" << std::endl;

		updateOrders();

		if (debug) std::cout <<  "SimulationState::updateObjects: update strategies" << std::endl;

		updateStrategies();
	}

	TaskCreateFunctionPtr SimulationState::findBestStrategy(
		uint4 playerID,
		real8 currentTime,
		real8 allowedComputationTime,
		real8 timeToSimulate,
		SimulationResult& simResult)
	{
		
		simResult.isDone = false;

		//double simStartTime = (double)glutGet(GLUT_ELAPSED_TIME);
		real8 simStartTime = getCurrentTime();

		TaskCreateFunctionPtr bestStrategy = NULL; //simResult.bestStrategy;
		uint4 bestStrategyIndex = simResult.bestStrategyIndex;		
		real8 bestTotalValue = simResult.bestValue;
		real8 currentTotalValue = simResult.currentValue;

		for(uint4 ours = simResult.ourStrategyIndex; ours <= NUM_STRATEGIES; ++ours){

			if (debug) std::cout <<  "trying out " << ours << std::endl;
			
			for(uint4 theirs = simResult.theirStrategyIndex; theirs <= NUM_STRATEGIES; ++theirs){
				if (debug) std::cout <<  "trying out theirs " << theirs << std::endl;

				// before doing any simulation.. check if we still have some computation time left to do it...
				// if(((double)glutGet(GLUT_ELAPSED_TIME) -simStartTime) > allowedComputationTime){
				if((getCurrentTime() - simStartTime) > allowedComputationTime){

					if (debug) std::cout <<  "time limit exceeded" << std::endl;
					// if we don't have any time left save the current state
					simResult.ourStrategyIndex = ours;
					simResult.theirStrategyIndex = theirs;
					simResult.bestValue = bestTotalValue;
					simResult.currentValue = currentTotalValue;
					simResult.bestStrategyIndex = bestStrategyIndex;

					return NULL;
				}

				SimulationState* tempState = this->clone();

				if (debug) std::cout <<  "copying player info" << std::endl;
				
				// now we copy the tasks for the players IFF their current task matches the task we will be simulating
				/*
				FORALL(players,it){
					// NOTE: since tasks and orders are not copied by default, we copy them here when we need to
					if((it->second)->getID() == playerID){
						if((it->second)->getStrategy() != NULL && (it->second)->getStrategy()->getFunction() == strategyList[ours]){
							copyPlayerInfo(it->second,tempState);
						}
					}else{
						if((it->second)->getStrategy() != NULL && (it->second)->getStrategy()->getFunction() == strategyList[theirs]){
							copyPlayerInfo(it->second,tempState);
						}
					}
				}
				*/

				if (debug) std::cout <<  "trying to find strategy value" << std::endl;
				
				// for now, count up the total value of this strategy vs ALL of theirs
				real8 combinationValue =  tempState->getStrategyValue(playerID, strategyList[ours],strategyList[theirs],currentTime,timeToSimulate);

				currentTotalValue += combinationValue;

				delete tempState;

				/*
				PRINT(boost::format("ours:[%u] %p theirs:[%u] %p value: %lf ") % ours % (void*)strategyList[ours] % theirs % (void*)strategyList[theirs] % combinationValue);
				*/
			}

			// PRINT(boost::format("-->total value of strategy [%u] %p  is: %lf") % ours % (void*)strategyList[ours] % currentTotalValue);


			if(currentTotalValue > bestTotalValue){
				bestTotalValue = currentTotalValue;
				bestStrategy = strategyList[ours];
				
				bestStrategyIndex = ours;
			}

			currentTotalValue = 0.0;
			// after the first saved iteration, we reset this.  only first iteration needs to be resumed afterall
			simResult.theirStrategyIndex = 1;
		}

		if (debug) std::cout <<  "findBestStrategy done" << std::endl;
		if (debug) std::cout <<  "bestTotalValue: " << bestTotalValue << std::endl;
		if (debug) std::cout <<  "bestStrategy: " << bestStrategyIndex << std::endl;

		if (bestStrategy == NULL)
		{
			if (debug) std::cout <<  "error? bestStrategy is NULL" << std::endl;
			if (debug) std::cout <<  "NUM_STRATEGIES:" << NUM_STRATEGIES << std::endl;
		}

		// if we are done, reset all the simulation values to their
		// defaults... and then return the best strategy
		simResult.isDone = true;
		simResult.ourStrategyIndex = 0;
		simResult.theirStrategyIndex = 0;
		simResult.bestValue = -2e30;
		simResult.currentValue = 0.0;
		simResult.bestStrategyIndex = bestStrategyIndex;

		//PRINT(boost::format("---->best in search: %p\n") % (void*)bestStrategy);
		return strategyList[bestStrategyIndex];
	}

	void SimulationState::copyPlayerInfo(Player* origPlayer,SimulationState* tempState){
		Player* tempPlayer = tempState->getPlayer(origPlayer->getID());
			// first copy the task
		tempPlayer->setStrategy(origPlayer->getStrategy()->clone(tempState,NULL));
			// then copy the orders
			/*
		FORALL(*origPlayer->getGroups(),it2){
			Group* tempGroup = tempPlayer->getGroup((it2->second)->getID());

			Order* currOrder = (it2->second)->getOrder();
			if(currOrder != NULL){
				tempGroup->setOrder(currOrder->clone());
			}
		}
			*/
	}

	TaskCreateFunctionPtr SimulationState::searchStrategy(
		uint4 playerID,
		real8 currentTime,
		real8 allowedComputationTime,
		real8 timeToSimulate,
		SimulationResult &simResult)
	{
		// set up local variables needed before calling negamax
		// do the first level here 
		// root is somewhat special since we actually only need the strategy to play at this level
		simResult.isDone = false;

		real8 simStartTime = getCurrentTime();

		TaskCreateFunctionPtr bestStrategy = NULL; 
		real8 bestTotalValue = simResult.bestValue;
		real8 currentTotalValue = simResult.currentValue;
		uint4 bestStrategyIndex = 0;
		
		// magic number for search depth
		// note: this should be even as we want to evaluate the opponent's replies
		const int depth = 2;

		real8 alpha = -999999.9;
		real8 beta  =  999999.9;

		for(uint4 ours = simResult.ourStrategyIndex; ours < NUM_STRATEGIES; ++ours) {

			if (debug) std::cout <<  "Trying strategy " << ours << std::endl;

			// before doing any simulation.. check if we still have some computation time left to do it...
			if((getCurrentTime() - simStartTime) > allowedComputationTime){				
				if (debug) std::cout <<  "time limit exceeded" << std::endl;
				// if we don't have any time left save the current state
				// FIXME: save states correctly
				simResult.ourStrategyIndex = ours;
				simResult.bestValue = bestTotalValue;
				simResult.currentValue = currentTotalValue;
				simResult.bestStrategyIndex = bestStrategyIndex;

				return NULL;
			}

			simResult.searchStrategyIndex[depth] = ours;

			real8 value = -negamax(this, playerID, depth-1, currentTime, simStartTime, allowedComputationTime, timeToSimulate, -beta, -alpha, simResult);

			if (value > alpha) {
				bestStrategy = strategyList[ours];
				bestStrategyIndex = ours;
				alpha = value;

				if (alpha >= beta) 
					if (debug) std::cout <<  "Potential Error: value returned " << value << " is larger than beta in root" << std::endl;
			}		

		}

		if (bestStrategy == NULL)
			std::cerr << "error? bestStrategy is NULL" << std::endl;

		// if we are done, reset all the simulation values to their
		// defaults... and then return the best strategy
		simResult.isDone = true;
		simResult.ourStrategyIndex = 0;
		simResult.theirStrategyIndex = 0;
		simResult.bestValue = -2e30;
		simResult.currentValue = 0.0;
		simResult.bestStrategyIndex = 0;

		return strategyList[bestStrategyIndex];
	}

	real8 SimulationState::negamax(
		SimulationState *currentState,
		uint4 playerID,
		uint8 depth,
		real8 currentTime,
		real8 simStartTime,
		real8 allowedComputationTime,
		real8 timeToSimulate,
		real8 alpha,
		real8 beta,
		SimulationResult &simResult)
	{
		if((getCurrentTime() - simStartTime) > allowedComputationTime) 
			return -999999.9;

		// parity determines whether it is our turn to move (even) 
		// or opponent's turn to move (odd)
		if (depth % 2 == 0) {
			for(uint4 ours = simResult.ourStrategyIndex; ours < NUM_STRATEGIES; ++ours) {
				simResult.searchStrategyIndex[depth] = ours;

				real8 value = -negamax(currentState, playerID, depth-1, currentTime, simStartTime, allowedComputationTime, timeToSimulate, -beta, -alpha, simResult);

				if (value >= alpha) {
					alpha = value;

					if (alpha >= beta) {
						return beta;
					}
				}
			}
		}
		else {
			for(uint4 theirs = simResult.theirStrategyIndex; theirs < NUM_STRATEGIES; ++theirs){
				simResult.searchStrategyIndex[depth] = theirs;

				SimulationState* tempState = currentState->clone();

				uint4 ours = simResult.searchStrategyIndex[depth+1];

				// now we copy the tasks for the players IFF their current task matches the task we will be simulating
				FORALL(players,it){
					// NOTE: since tasks and orders are not copied by default, we copy them here when we need to
					if((it->second)->getID() == playerID){
						if((it->second)->getStrategy() != NULL && (it->second)->getStrategy()->getFunction() == strategyList[ours]){
							copyPlayerInfo(it->second,tempState);
						}
					}else{
						if((it->second)->getStrategy() != NULL && (it->second)->getStrategy()->getFunction() == strategyList[theirs]){
							copyPlayerInfo(it->second,tempState);
						}
					}
				}

				real8 value = tempState->getStrategyValue(playerID, strategyList[ours],strategyList[theirs], currentTime, timeToSimulate);

				// whether we use this value depends if depth == 0, otherwise,
				// we will rely on negamax results
				if (depth-1 != 0) {
					value = -negamax(tempState, playerID, depth-1, currentTime+timeToSimulate, simStartTime, allowedComputationTime, timeToSimulate, -beta, -alpha, simResult);
				}

				if (value > alpha) {
					alpha = value;

					if (alpha >= beta) {
						delete tempState;				
						return beta;
					}
				}

				delete tempState;								
			}
		}

		return alpha;
	}

	real8 SimulationState::advanceTime(real8 lastTime){
		real8 lastMoveUpdateMillis = lastTime;

		if (debug) std::cout <<  "advanceTime: get next interesting event time" << std::endl;

		lastTime = getNextTime(lastTime);

		// this can happen sometime, so when next time is = to current time,
		// make sure we advance at least a little bit

		//FIXME we shouldn't need this, get rid of it if possible
		if(lastTime == lastMoveUpdateMillis){
			lastTime+=50;
		}

		if (debug) std::cout <<  "advanceTime: Updating objects" << std::endl;
		updateObjects(lastMoveUpdateMillis,lastTime);

		return lastTime;
	}

	bool SimulationState::isGameOver(){
		FORALL(players,it){
			if((it->second)->getVictoryValue() <= 0.0 || (it->second)->getGroups().size() == 0)
			{
				if ((it->second)->getVictoryValue() <= 0.0) {
					if (debug) std::cout <<  "isGameOver: player has victory value " << (it->second)->getVictoryValue() << " <= 0.0" << std::endl;
				}
				
				if ((it->second)->getGroups().size() == 0) {
					if (debug) std::cout <<  "Player has no groups" << std::endl;
				}
				
				return true;
			}
		}
		return false;
	}

	bool SimulationState::hasWon(uint4 playerID){
		FORALL(players,it){
			if((it->second)->getID() == playerID) continue;

			if((it->second)->getVictoryValue() > 0.0){
				return false;
			}
		}
		return true;
	}

	// FIXME for now, just return the total value of our bases... want to make this much more complex!
	real8 SimulationState::evaluate(uint4 playerID){
		real8 value = 0.0;
		FORALL(players,it){
			real8 playerValue = (it->second)->getVictoryValue();
			//if(playerValue <= 0.0) playerValue -= VICTORY_BONUS; // if player has no bases left, we have to make our evaluation very low!!!
			//if((it->second)->getGroups()->size() == 0) playerValue -= VICTORY_BONUS*10.0; // FIXME this is just arbitrary right now

			if((it->second)->getID() == playerID){
				value += playerValue;
			}else{
				value -= playerValue;
			}
		}
		return value;
	}

	real8 SimulationState::getNextTime(real8 currentTime){
		real8 minTime = 2e30;

		if (debug) std::cout <<  "getNextTime: getNextCollisionTime" << std::endl;
		
		real8 collisionTime = getNextCollisionTime(currentTime);
		if(collisionTime < minTime) minTime = collisionTime;

		if (debug) std::cout <<  "getNextTime: getNextOrderCompletionTime" << std::endl;

		real8 orderCompletionTime = getNextOrderCompletionTime(currentTime);
		if(orderCompletionTime < minTime) minTime = orderCompletionTime;

		//real8 shootingTime = getNextShootingTime(currentTime);
		//if(shootingTime < minTime) minTime = shootingTime;

		//real8 timeoutTime = getNextOrderTimeoutTime(currentTime);
		//  if(timeoutTime < minTime) minTime = timeoutTime;

		if(minTime == 2e30){
			return currentTime;
		}

		if(minTime < currentTime){
			//print("error, mintime of " + convertToString(minTime) + " is less than the currentTime of: " + convertToString(currentTime));
			return currentTime;
		}

		//print("new minTime = " + convertToString(minTime));

		// IMPORTANT: the +0.01 is just to prevent some rounding off problems which can be inefficient, since some units don't quite reach their goal sometimes
		// it's equivalent to 1/100 milliseconds so not significant at all
		return minTime;
	}

	void SimulationState::updateOrders(){
		FORALL(players,it){
			(it->second)->updateOrders(*this);
		}
	}

	void SimulationState::moveObjects(real8 timeElapsed){
		FORALL(players,it){
			(it->second)->executeOrders(timeElapsed);
		}
	}

	bool SimulationState::attackBasesInRange(Vector<Base*> &basesInRange, Group *currentGroup, real8 currentTime) {
		// IDEA: Time to destroy building?
		// get base hp and number of units

		// what do we return here??! just a bool, but we want to simulate destroying the buildings
		// so do we simulate the time it takes? or that it instantly is destroyed?

		if (basesInRange.empty()) return false;
		
		//if (debug) std::cout <<  "SimulationState::attackBasesInRange: start" << std::endl;

		// simplest solution: remove basesInRange
		uint4 ownStrength = currentGroup->getNumUnits();
		double damage = (currentGroup->getUnits()[0]->getObject()->component("weapon")->get_int("max_damage"));
		double ownDamage = (ownStrength * damage);

		for (Vector<Base*>::iterator it = basesInRange.begin(); it != basesInRange.end(); it++) {
			//(*it)->hitPoints -= ownDamage;
			
			if ((*it)->hitPoints < 0) 
				basesInRangeMap[*it]->removeOrtsBase(*it);
		}

		//if (debug) std::cout <<  "SimulationState::attackBasesInRange: end" << std::endl;

		return true;
	}

	bool SimulationState::attackUnitsInRange(Vector<Group*> &groupsInRange, Group *currentGroup, real8 currentTime) {
		// IDEA: winner depends on size of groups

		if (groupsInRange.empty()) return false;
		
		//if (debug) std::cout <<  "SimulationState::attackUnitsInRange: start" << std::endl;
		
		uint4 ownStrength = currentGroup->getNumUnits();
		uint4 theirStrength = 0;		
		for (Vector<Group*>::iterator it = groupsInRange.begin(); it != groupsInRange.end(); it++ ) {
			theirStrength += (*it)->getNumUnits();
		}

		// simplest solution, every unit is subtracted a fraction of the total fire power
		double damage = (currentGroup->getUnits()[0]->getObject()->component("weapon")->get_int("max_damage"));
		//double ownDamage = (theirStrength * damage)/(double)ownStrength;
		double theirDamage = (ownStrength * damage)/(double)theirStrength;
		/*
		for (Vector<Unit*>::iterator it = currentGroup->getUnits().begin(); it != currentGroup->getUnits().end(); it++) {
			(*it)->hitPoints -= ownDamage;
			
			if ((*it)->hitPoints < 0) 
			{				
				if (groupsInRangeMap.find(currentGroup) == groupsInRangeMap.end()) ERR("No Player?");
				groupsInRangeMap[currentGroup]->removeOrtsUnit(*it);
			}
		}*/
		
		for (Vector<Group*>::iterator it = groupsInRange.begin(); it != groupsInRange.end(); it++ ) {
			UnitVector unitVector((*it)->getUnits());
			for (Vector<Unit*>::iterator it2 = unitVector.begin(); it2 != unitVector.end(); it2++) {
				(*it2)->hitPoints -= theirDamage;
				
				if ((*it2)->hitPoints < 0) 
					groupsInRangeMap[*it]->removeOrtsUnit(*it2);
			}
		}
		
		
		//if (debug) std::cout <<  "SimulationState::attackUnitsInRange: end" << std::endl;

		return true;
	}

	void SimulationState::resolveCollisions(real8 currentTime){
		// yes,, it is n^2, and could be made a bit better by doing both
		// evaluations for collisions, but total computing time is around
		// the same...(i think)

		FORALL (players, currentPlayerIterator) {

			groupsInRangeMap.clear();
			basesInRangeMap.clear();

			FORALL ((currentPlayerIterator->second)->getGroups(), it) {
				Group* currentGroup =(*it);
				groupsInRangeMap[currentGroup] = currentPlayerIterator->second;
			}
			
			FORALL ((currentPlayerIterator->second)->getGroups(), it) {
				Group* currentGroup =(*it);

				Vector<Base*> basesInRange;
				Vector<Group*> groupsInRange;

				FORALL(players,secondPlayerIterator){
					if((currentPlayerIterator->second)->getID() == (secondPlayerIterator->second)->getID()){
						continue;
					}

					FORALL(*(secondPlayerIterator->second)->getBases(), it2) {
						if((*it2)->withinRange(currentGroup)){
							basesInRange.push_back(*it2);
							//basesInRangeMap[*k].push_back(it2->second);

							basesInRangeMap[*it2] = secondPlayerIterator->second;
						}
					}

					FORALL((secondPlayerIterator->second)->getGroups(), it2) {
						if((*it2)->withinRange(currentGroup)){
							groupsInRange.push_back(*it2);
							//groupsInRangeMap[*k].push_back(it2->second);

							groupsInRangeMap[*it2] = secondPlayerIterator->second;
						}
					}
				}
				
				//if (debug) std::cout <<  "SimulationState::resolveCollisions attacking!" << std::endl;
				// if no objects in range, we can continue
				if(basesInRange.empty() && groupsInRange.empty()) continue;

				if(attackUnitsInRange(groupsInRange,currentGroup,currentTime)){

				}else if(attackBasesInRange(basesInRange,currentGroup,currentTime)){

				}else{
					// no targets found, should never get here!
					//print("no targets found in collisions while trying to deal damage, this should NOT happen!");
				}

				/*
				// if we are targetting bases first...
				if(currentGroup->getBaseFirstAttackPriority()){
					if(attackBasesInRange(basesInRangeMap,currentGroup,currentTime)){

					}else if(attackUnitsInRange(groupsInRangeMap,currentGroup,currentTime)){

					}else{
						// no targets found, should never get here!
						// print("no targets found in collisions while trying to deal damage, this should NOT happen!");
					}
				}else{
					if(attackUnitsInRange(groupsInRangeMap,currentGroup,currentTime)){

					}else if(attackBasesInRange(basesInRangeMap,currentGroup,currentTime)){

					}else{
						// no targets found, should never get here!
						//print("no targets found in collisions while trying to deal damage, this should NOT happen!");
					}
				}
				*/
			}
		}

		if (debug) std::cout <<  "SimulationState::resolveCollisions: done attacking" << std::endl;
		
	}

	// function originally from Collision2D.C from ORTS

	/** Solves \f$ at^2 + bt = c; a > 0 \f$
	@return true iff solution exists

		robust solution (numerical recipies)
	*/

	bool SimulationState::solveQuadratic(real8 a, real8 b, real8 c, real8& t1, real8& t2){

		if (a <= 0.0){
			//print("ERROR: Trying to solve quadratic with a <=0");
			return false;
		}
		// no need to do this, pass a negative c instead
		//c = -c;
		real8 squareRootPart = b*b-4.0*a*c;

		// can't have a negative square root, so return false if this is negative
		if (squareRootPart < 0.0){
			//print("error, squareRootPart is negative: "+convertToString(squareRootPart));
			return false;
		}
		// robust root computation (Numerical Recipies p.184)

		real8 q = -0.5L * (b + sgn(b) * sqrt(squareRootPart));
		if(q == 0.0){
			//print("error in quadratic, q was 0");
			return false;
		}
		t1 = q / a;
		t2 = c / q;

		if (t1 > t2){
			real8 temp = t2;
			t2 = t1;
			t1 = temp;
		}
		//assert(!(t1 > t2));

		real8 e1 = fabs(a*t1*t1+b*t1+c);
		real8 e2 = fabs(a*t2*t2+b*t2+c);
		if (e1 > 1.5e-5) ERR2("error in quadratic: ",e1);
		if (e2 > 1.5e-5)  ERR2("error in quadratic: ",e2);

		return true;
	}

	// originally from Collision2D.H from ORTS

	/** Compute collision time of a (moving) circle and a (moving) point.
	Used to check for collision with the endpoint of a line.
		@param circlePos     circle center
		@param radius    square of the circle's radius
		@param pointPos     point's location
		@param relativeHeading     circle's heading relative to the point
		@param startTime  start simulation at this time
	*/

		real8 SimulationState::getCirclePointCollisionTime(const Vector3D& circlePos,
		real8 radius,
		const Vector3D& pointPos,
		const Vector3D& relativeHeading,
		real8 startTime)
	{

		Vector3D positionDifference = circlePos - pointPos;
		real8 quadraticValueC =
			(radius*radius) -
			(positionDifference.x*positionDifference.x) -
			(positionDifference.y*positionDifference.y);

		//int c_h = c_p * h; // c dot h
		real8 differenceHeading = positionDifference.dot(relativeHeading);

		// if they are heading away from each other and not touching, then
		// no need to do any more work...

		if (differenceHeading > 0.0) {   	// moving away from each other
			if (quadraticValueC >= 0.0) { 	// they don't intersect
											//print("moving from each other");
			return -1.0;
			}
		}

		real8 t1= 0.0;
		real8 t2 = 0.0;

		real8 quadraticValueA =
			(relativeHeading.x*relativeHeading.x) + (relativeHeading.y*relativeHeading.y);
		real8 quadraticValueB = 2.0 * differenceHeading;
		bool isCollision =
			solveQuadratic(quadraticValueA,quadraticValueB,-quadraticValueC, t1, t2); // (a)t^2 + (b)t - c = 0

		if (!isCollision){
					//print("no collision");
			return -1.0;
		}

		t1+=startTime;
		t2+=startTime;
		if (t1 == t2){
					//print("t1 and t2 are equal!");
			return -1.0; // point intersects tangentially
		}

		if (t2 <= startTime){
					//print("t2 <= starttime");
			return -1.0;
		}
		if (t1 == startTime){
					//print("WARNING: time is equal to startTime");
			return -1.0;
					//return startTime; // intersecting at time
		}
		if(t1 < startTime || t2 < startTime){
					//print("WARNING: time of collision is less than start time, this is wrong!");
					//return startTime;
			return -1.0;
		}

		return t1;
}

real8 SimulationState::getNextCollisionTime(real8 currentTime){
	real8 minimumTime = 2e30;

	FORALL(players,currentPlayerIterator){
		GroupVector &currPlayerGroups( (currentPlayerIterator->second)->getGroups() );

		//if (debug) std::cout <<  "getNextCollisionTime: looking through groups of player" << std::endl;

		FORALL(currPlayerGroups,it){
			// FIXME later make this work for bases as well
			Group* currentObject = (*it);

			Vector3D currentObjectHeading(0.0,0.0,0.0);
			Order *currentOrder = currentObject->getOrder();

			if (currentOrder) {
				currentObjectHeading =
					(currentOrder->getDestination() - currentObject->getPosition()) |
					currentOrder->getSpeed();
			}

			//if (debug) std::cout <<  "getNextCollisionTime: current object heading:" << currentObjectHeading.toString() << std::endl;
			
			FORALL(players,otherPlayerIterator){
				// don't check your own objects...
				if(currentPlayerIterator == otherPlayerIterator) continue;

				GroupVector &otherPlayerGroups( (otherPlayerIterator->second)->getGroups() );

				FORALL(otherPlayerGroups,it2){
					Group* otherObject = (*it2);

						// if objects are already within each other's range, don't do them anymore.

					if((currentObject->getPosition() - otherObject->getPosition()).length() <=
						currentObject->getExtents() + currentObject->getMaxAttackRange() +
					otherObject->getExtents()) {
							//print("inside object!!!");
						continue;
					}

					Vector3D otherObjectHeading;

					if(otherObject->getOrder() == NULL){
						otherObjectHeading = Vector3D(0.0,0.0,0.0);
					}else{
						otherObjectHeading =
							(otherObject->getOrder()->getDestination()-otherObject->getPosition()) |
							otherObject->getOrder()->getSpeed();
					}

					Vector3D relativeHeading = currentObjectHeading - otherObjectHeading;

					real8 collisionTime =
						getCirclePointCollisionTime(currentObject->getPosition(),
						currentObject->getExtents()+
						currentObject->getMaxAttackRange()+
						otherObject->getExtents(),
						otherObject->getPosition(),
						relativeHeading,
						currentTime);

						//print("found time: "+convertToString(collisionTime));
					if(collisionTime == -1.0){
						continue;
					}

					if(collisionTime < minimumTime) minimumTime = collisionTime;
				}
			}
		}
	}
	return minimumTime;
}

real8 SimulationState::getNextOrderCompletionTime(real8 currentTime){
	real8 minCompletionTime = 2e30;

	FORALL(players,playerIterator){
		GroupVector &currPlayerGroups( (playerIterator->second)->getGroups() );

		FORALL(currPlayerGroups,it){
			Group* currentGroup = (*it);

			if(currentGroup->getOrder() == NULL) continue;

				// FIXME not sure if the next line should be left in or not.
			if(currentGroup->getPosition() == currentGroup->getOrder()->getDestination())
				continue;

				//print("unit " + currentGroup->getName() + " order start time: "+convertToString(currentTime));
			real8 estimatedTime = currentTime +
				(currentGroup->getOrder()->getDestination() - currentGroup->getPosition()).length() /
				((real8)currentGroup->getOrder()->getSpeed());

			if(estimatedTime == currentTime) continue;

				//print("unit " + currentGroup->getName() + " order estimated end time: "+convertToString(estimatedTime));

			if(estimatedTime < minCompletionTime){
				minCompletionTime = estimatedTime;
			}
		}
	}
	return minCompletionTime;
}

real8 SimulationState::getStrategyValue(
	uint4 playerID,
	TaskCreateFunctionPtr ourStrategy,
	TaskCreateFunctionPtr theirStrategy,
	real8 currentTime,
	real8 timeToSimulate)
{

	if (debug) std::cout <<  "getStrategyValue: start" << std::endl;
	
	FORALL (players, it) {
		GroupVector newVector = (it->second)->getGroups();

		// by default, give ALL the units of a player the same strategy, this can be changed
		/*
		FORALL ((it->second)->getOrtsUnits(), it2) {
			newVector.push_back(it2->first);
		}
		*/
		if(newVector.empty()){
			(it->second)->setStrategy(NULL);
			continue; // don't do anything if the player has no more groups
		}

		if((it->second)->getID() == playerID){
			if(ourStrategy == NULL) continue;
			if(((it->second)->getStrategy() == NULL || (it->second)->getStrategy()->getFunction() != ourStrategy)){
				(it->second)->setStrategy(ourStrategy((it->second)->getID(),this,newVector,0.0));
			}
		}else{
			if(theirStrategy == NULL) continue;
			if(((it->second)->getStrategy() == NULL || (it->second)->getStrategy()->getFunction() != theirStrategy)){
				(it->second)->setStrategy(theirStrategy((it->second)->getID(),this,newVector,0.0));
			}
		}
	}

	real8 value = 0.0;
	value = simulateForTime(playerID,currentTime,timeToSimulate);
	if (debug) std::cout <<  "getStrategyValue: value =" << value << std::endl;

	return value;
}

real8 SimulationState::simulateForTime(
	uint4 playerID,
	real8 currentTime,
	real8 timeToSimulate)
{
	real8 timeToSimulateTo = currentTime+timeToSimulate;
	real8 nextSimulationTime = currentTime;

	if (debug) std::cout <<  "simulateForTime: start time at " << currentTime << std::endl;
	while(nextSimulationTime < timeToSimulateTo && !isGameOver()){
		if (debug) std::cout <<  "simulateForTime: advancing time to " << nextSimulationTime << std::endl;
		
		nextSimulationTime = advanceTime(nextSimulationTime);
	}

	return evaluate(playerID);
}

bool SimulationState::isObstacle(sint4 x, sint4 y){
	GameTile* tile = &simAI->get_gsm().get_game().get_map().get_tile(x,y);
	sint4 height = tile->get_max_h();
	
	if(height != 0 || tile->get_typeW() == 1 || tile->get_typeW() == 3 )
		return true;
	else
		return false;
}

void SimulationState::updateStrategies(){
	FORALL(players,it){
		Task* currStrategy = (it->second)->getStrategy();
		if(currStrategy != NULL){
			bool isDone = currStrategy->update();

			/*if(isDone){
				(it->second)->setStrategy(NULL);
			}*/
		}
	}
}

}

