#ifndef AutoGatherAI_H
#define AutoGatherAI_H

#include "GameChanges.H"
#include "ClientModule.H"
#include "EventHandler.H"
#include "Game.H"

#include "GobPtrWrapper.H"
#include "TerrainModule.H"
#include "Game1_SimpleTerrain.H"

#include <set>
#include <map>
#include <string>
#include "boost/smart_ptr.hpp"

//#define USE_DEBUG_GFX

#ifdef USE_DEBUG_GFX
#include "DrawOnTerrain.H"
#include "GfxModule.H"
#endif


/// Automatically manage the resource gathering operations, assuming a single control center.
/** This is an implementation for an AI managing the resoure gathering operations.
    It is designed for the AIIDE06 tournamnent's game1, including:
    - Assuming there's only a single control center.
    - Assuming all minerals are known from the game's first view.
*/
class AutoGatherAI : public ClientModule
{
  //---------------------------------------------------------------------------
  /// constants
  //---------------------------------------------------------------------------
  //@{ 
private:
  static const sint4 FROM;                        ///< Identifier of this class when sending events

  static const std::string workerBPname;          ///< worker's blueprint name
  static const std::string mineralBPname;         ///< minerals' blueprint name
  static const std::string controlCenterBPname;   ///< control center's blueprint name

  /** Interval to wait before trying to reuse a mineral which cannot 
      be reached considering only static obstacles. */
  static const sint4 BLOCKED_MINERAL_RETEST_IN_TICKS = 360;
  /** Interval to wait before trying to reuse a mineral which cannot 
      be reached temporarily (due to mobile obstacles). */
  static const sint4 FAILED_MINERAL_SLEEP_TICKS = 20;
  /// How much time (milliseconds) to spend on testing whether a mineral is statically blocked
  static const sint4 TIME_FOR_SINGLE_STATIC_TEST = 60;

  /// \name Weights for selecting a mineral per worker
  //@{
  static const sint4 CUR_DIST_WEIGHT = 1;     ///< Weight for the distance from the worker's current location to the mineral
  static const sint4 STATIC_DIST_WEIGHT = 1;  ///< Weight for the length of the (statically computed) path between the mineral and the control center
  //@}

  /// Maximum worker/mineral distance which allows mining
  static const sint4 WORKER_MINERAL_DISTANCE_THRESHOLD = 2;
  /// Maximum worker/control-center distance which allows returning minerals
  static const sint4 WORKER_CC_DISTANCE_THRESHOLD = 3;

  /// Maximum workers to associate with each mineral patch simultaneously
  static const uint4 MAX_WORKERS_PER_MINERAL = 1;

  /// For clusters close to the CC don't use cluster_planning at all
  static const sint4 TOO_CLOSE_CLUSTER = 120;

  /** Maximum workers to associate with each mineral cluster simultaneously is
      the number of minerals multiplied by this value */
  static const real4 MAX_WORKERS_PER_CLUSTER_FACTOR;
  /// Minimal number of mineral clusters expected
  static const sint4 MIN_NUM_CLUSTERS = 2;
  /// Maximal number of mineral clusters expected
  static const sint4 MAX_NUM_CLUSTERS = 4;
  /// Consider paths that are longer by this factor more than the straight line distance.
  //static const real8 LEGAL_ELONGATION_FACTOR = 2.5;
  /// How far from a cluster center will the cached lanes start/end
  static const sint4 DEPLOYMENT_POINTS_DISTANCE_TO_CLUSTER = 60;
  /// The minimal time delay needed between workers leaving the CC towards the same cluster.
  static const sint4 SAME_CLUSTER_DELAY_TICKS = 10;

  /// Maximum ticks to allow a worker to be stuck without forcing it into an idle state
  static const sint4 MAX_TICKS_TO_FORCED_IDLE = 100;
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Internal bookkeeping typedefs
  //---------------------------------------------------------------------------
  //@{
public:

  //---------------------------------------------------------------------------
  // public to allow REGISTER_TYPEOF definitions
  //---------------------------------------------------------------------------
  struct MineralCluster {
    Vector<const GobPtrWrapper*>  mineralsInCluster;      ///< the minerals in this cluster
    Vector<TerrainBase::Path>     cachedPaths;            ///< N+1 already-planned paths from the CC, to each waypoint, and back, in this order
    bool                          usePathAroundCluster;   ///< In problematic cases, use a fall-back to plain pathfinding.
    sint4                         last_departure_from_CC; ///< the view frame when the last worker left the CC, to allow temporal spacing
    TerrainBase::Loc              centroid;               ///< the centroid of mineral locations

/* Version 1 of cluster-planning:
    Vector<TerrainBase::Loc>     pathAroundCluster; ///< a series of N waypoints for moving around the cluster, without the CC itself */

    MineralCluster() : usePathAroundCluster(false), last_departure_from_CC(-1) {};
    MineralCluster(const MineralCluster &other)
      : mineralsInCluster(other.mineralsInCluster), 
        cachedPaths(other.cachedPaths),
        usePathAroundCluster(other.usePathAroundCluster), 
        last_departure_from_CC(other.last_departure_from_CC),
        centroid(other.centroid) {};
    MineralCluster& operator=(const MineralCluster &other) {
      if( this != &other ) {
        mineralsInCluster = other.mineralsInCluster; 
        cachedPaths = other.cachedPaths;
        usePathAroundCluster = other.usePathAroundCluster; 
        last_departure_from_CC = other.last_departure_from_CC;
        centroid = other.centroid;
      }
      return *this;
    };
  };

  struct MineralState {
    sint4 dont_use_until;           ///< dont use this mineral until a certain view_frame
    sint4 clusterId;                ///< the cluster containing the mineral
    sint4 static_distance_from_cc;  ///< Length of the path from the CC to this mineral, planned while ignoring all mobile objects
    /** >0: the mineral cannot be reached even if ignoring all mobile objects;
        <0: was not tested yet; 
        =0: can be statically reached; */
    sint1 statically_blocked;

    //sint4 failed_moves;       ///< number of times we tried sending a worker to this mineral and failed

    MineralState() 
      : dont_use_until(-1), clusterId(-1), static_distance_from_cc(-1), statically_blocked(-1) /*, failed_moves(0)*/ {};
    MineralState(const MineralState &other) 
      : dont_use_until(other.dont_use_until), clusterId(other.clusterId), static_distance_from_cc(other.static_distance_from_cc), statically_blocked(other.statically_blocked) {};
  private:
    MineralState& operator=(const MineralState &other);
  };

  typedef Vector<MineralCluster> MineralClusters;
  //typedef std::map<const GobPtrWrapper*,const MineralCluster*> M2C;
  typedef std::map<const GobPtrWrapper*,MineralState> M2S;
  //---------------------------------------------------------------------------

private:
  struct WorkerState {
    enum { IDLE, MOVING, MINING, RETURNING_MINERALS } job;
    const GobPtrWrapper* mineral;
    sint4 clusterId;
    sint4 clusterPathWaypoint;
    bool initial_move;
    bool failedToArrive;
    TerrainBase::Loc last_pos;
    sint4 in_this_pos_since;

    WorkerState() : job(IDLE), mineral(NULL), clusterId(-1), clusterPathWaypoint(0),
                    initial_move(true), failedToArrive(false), in_this_pos_since(-1) {};
    WorkerState(const WorkerState &other)
      : job(other.job), mineral(other.mineral), clusterId(other.clusterId),
        clusterPathWaypoint(other.clusterPathWaypoint), initial_move(other.initial_move),
        failedToArrive(other.failedToArrive), in_this_pos_since(other.in_this_pos_since) {};
  private:
    WorkerState& operator=(const WorkerState&);
  };

  typedef std::set<GobPtrWrapper> GameObjects;
  typedef std::set<Worker>  Workers;
  typedef std::map<const Worker*,WorkerState> W2S;
  typedef std::multimap<const GobPtrWrapper*,Worker*> M2W;
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Variables
  //---------------------------------------------------------------------------
  //@{ 
private:
  Workers             workers;
  GameObjects         minerals;
  GameObjects         controlCenters;
  W2S                 workers2state;        ///< keep state info for each worker
  M2S                 minerals2state;       ///< keep state info for each mineral
  M2W                 minerals2workers;     ///< map a mineral patch to its mining workers
  MineralClusters     mineralClusters;      ///< list of mineral clusters
  //M2C                 minerals2clusters;    ///< map a mineral to its containing cluster
  sint4               map_width;            ///< Width in tile_points
  sint4               map_height;           ///< Height in tile_points

  GameStateModule &gsm;  
  TerrainModule &tm;
  Game1_SimpleTerrain::ST_Terrain &st;
  bool noClusterPlanning;                   ///< true to prevent planning based on mineral clusters
  
  //@}
  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------
  /// Internal bookkeeping
  //---------------------------------------------------------------------------
  //@{ 
  /// update world view with the given changes.
  void applyViewMessage(const GameChanges &changes);

  /// get the state of a given worker
  WorkerState& getState(const Worker& worker) { 
    W2S::iterator w2sIt = workers2state.find(&worker);
    assert(w2sIt != workers2state.end());
    return w2sIt->second;
  };
  /// get the state of a given worker
  const WorkerState& getState(const Worker& worker) const { 
    W2S::const_iterator w2sIt = workers2state.find(&worker);
    assert(w2sIt != workers2state.end());
    return w2sIt->second;
  };

  MineralState dummy_against_segfault;
  /// get the state of a given mineral
  MineralState& getMineralState(const GobPtrWrapper& mineral) { 
    M2S::iterator m2sIt = minerals2state.find(&mineral);
    //assert(m2sIt != minerals2state.end());
    if( m2sIt == minerals2state.end() ) return dummy_against_segfault;
    return m2sIt->second;
  };
  /// get the state of a given mineral
  const MineralState& getMineralState(const GobPtrWrapper& mineral) const { 
    M2S::const_iterator m2sIt = minerals2state.find(&mineral);
    //assert(m2sIt != minerals2state.end());
    if( m2sIt == minerals2state.end() ) return dummy_against_segfault;
    return m2sIt->second;
  };

  /// get the mineral associated with a given worker
  const GobPtrWrapper* getMineral(const Worker& worker) const { 
    return getState(worker).mineral;
  };

  /// remove given workers, minerals and control center(s) from the internal bookkeeping
  void removeWMC(const Vector<ScriptObj::WPtr>& objects);
  /// add given workers, minerals and control center(s) to the internal bookkeeping
  void insertWMC(const Vector<ScriptObj::WPtr>& objects);
  /** track changes in worker positions to find unexpectedly idle workers. 
      This is a fail-safe which should not happen in a bug-free client. */
  void updateWorkers(const Vector<ScriptObj::WPtr>& objects);
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Babysitting the workers
  //---------------------------------------------------------------------------
  //@{ 
  /// check if a worker is idle
  bool isIdle(const Worker& worker) const;
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Global planning
  //---------------------------------------------------------------------------
  //@{ 
private:
  /** Initialize the work plan. Should be called after the minerals and
      control center locations are known, ie after the first view msg was received. */
  void plan(void);
  /// Create a clusters map for the mineral locations
  void computeMineralClusters(void);

#if 0
  /** Create roundabout paths for each mineral cluster, going to one side
      of the cluster and returning from the other side. 
      Assuming that the clusters were already computed. */
  void computeWaypointsAroundClusters(void);
  /** Try PathFinding for the waypoints of each cluster. If possible, cache
      these paths for later use. */
  void computePathsAroundClusters(void);
#endif

  /** Compute a path to and a different path from each cluster. Attempt
      to minimize interferences between the lanes to different clusters. */
  void computeLanesToClusters(void);
  /** Compute a path to and a different path from a given cluster, with a
      given width. Return true iff successful, update the paths in the 
      given cluster's info. */
  bool computeLanesToCluster(MineralCluster &cluster, sint4 radius);

  /** Check (using PF) if this mineral is blocked or accessible ignoring
      all mobile obstacles, and update its state accordingly. */
  void testStaticBlock(const GobPtrWrapper *mineral);
  /// find a (single) statically-blocked mineral whose timeout passed and retest.
  void retestStaticBlock(void);
  /// check for static-blocks to all minerals.
  void markBlockedMinerals(void);

  /** Allocate some workers to mineral clusters, if possible, and all other
      workers to specific mineral pathces. */
  void InitialAllocationOfWorkers(void);

  /** choose and allocate a mineral patch to the given worker; 
      Optionally instructed to use a different patch than the one currently
      allocated, if at all possible. */
  void allocateMineralPatchFor(Worker& worker, bool forcedChange=false);
  /// select a mineral in the cluster associated to this worker
  void allocateWithClusterScheme(Worker& worker, bool forcedChange=false);
  /// select a mineral from those not in clusters that are handled in the cluster planning scheme
  void allocateWithSingleMineralScheme(Worker& worker, bool forcedChange=false, bool allowDynamicallyBlocked = false, bool allowSpill = false);
  /// A comparison functor used for prioritizing mineral patches for allocation to a given worker.
  struct CompareMineralsByDistanceToWorker; friend struct CompareMineralsByDistanceToWorker;

  /// perform the low-level association of a worker and amineral in the internal maps
  void associateWorkerAndMineral(Worker& worker, const GobPtrWrapper* mineral);

  /** Check if it is OK to allocate a given mineral to a given worker to use with
      the single-mineral planning mode. 
      @param allowDynamicallyBlocked  Don't reject minerals that are sleeping after a dynamic block (ie due to mobile obstacles).
      @param allowSpill               Don't enforce the max_workers_per_mineral constraint.
      */
  bool isMineralFreeFor(const GobPtrWrapper* mineral, const Worker& worker, bool allowDynamicallyBlocked = false, bool allowSpill = false) const;

  /// A predicate adaptor for the isMineralFreeFor() member function
  struct IsMineralFreeFor {
    const AutoGatherAI *agai;
    const Worker& worker;
    IsMineralFreeFor(const AutoGatherAI *agai_, const Worker& worker_) 
      : agai(agai_), worker(worker_) {};
    bool operator()(const GobPtrWrapper* mineral) {
      return agai->isMineralFreeFor(mineral, worker);
    };
  };

  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Access a worker's status
  //---------------------------------------------------------------------------
  //@{ 
private:
  /// Is the given worker located near the control center? (No type-checking is done)
  bool isNearTheCC(const Worker& worker) const;
  /// Is the given worker located near its allotted mineral patch? (No type-checking is done)
  bool isNearTheMineral(const Worker& worker) const;
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Set a worker's action
  //---------------------------------------------------------------------------
  //@{ 
private:
  /// Move to the given location, optionally using a cached path
  void moveTo(Worker& worker, const TerrainBase::Loc &location, const TerrainBase::Path *cached_path = NULL);
  /// Move to the given target
  void moveTo(Worker& worker, const GobPtrWrapper& target);
  /// Move back to the control center
  void moveToCC(Worker& worker);
  /// Move to the mineral patch allocated to this worker
  void moveToMineral(Worker& worker);
  /// Start mining (assuming the worker is near the mineral)
  void startMining(Worker& worker);
  /// Return resources (assuming the worker is near the control center)
  void returnMinerals(Worker& worker);
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Handlers for different events
  //---------------------------------------------------------------------------
  //@{
private:
  /// Handler for notifications of the PathFinder object
  class HandlerOfPathFinderEvents : public EventHandler {
    AutoGatherAI* parent;
  public:
    HandlerOfPathFinderEvents(AutoGatherAI* parent_) : parent(parent_) {};
    virtual bool handle_event(const Event &event);
  } handlerOfPathFinderEvents;

  /// Handler for GameStateModule's view_msg notifications
  class HandlerOfViewMsgEvents : public EventHandler {
    AutoGatherAI* parent;
  public:
    HandlerOfViewMsgEvents(AutoGatherAI* parent_) : parent(parent_) {};
    virtual bool handle_event(const Event &event);
  } handlerOfViewMsgEvents;
  //@}
  //---------------------------------------------------------------------------


private:
  /// Compute and apply actions to the workers
  void act();


  //---------------------------------------------------------------------------
  /// Public interface
  //---------------------------------------------------------------------------
  //@{
public:
#ifndef USE_DEBUG_GFX
  AutoGatherAI(GameStateModule &gsm_, TerrainModule &tm_, Game1_SimpleTerrain::ST_Terrain &st_);
#else
  AutoGatherAI(GameStateModule &gsm_, TerrainModule &tm_, Game1_SimpleTerrain::ST_Terrain &st_, GfxModule *gfxm_);
#endif
  ~AutoGatherAI();

  static void add_options(void);
  //@}
  //---------------------------------------------------------------------------

#ifdef USE_DEBUG_GFX
private:
  class DbgDraw : public EventHandler, public DrawOnTerrain
  {
    const AutoGatherAI *parent;
    bool enabled;

  public:
    DbgDraw(const AutoGatherAI *parent_, GfxModule &gfxm_);
    ~DbgDraw();

    bool is_enabled() const { return enabled; }
    void enable() { enabled = true; }
    void disable() { enabled = false; }
    void draw();
    bool handle_event(const Event &e);
  };
  boost::scoped_ptr<DbgDraw> dbg_draw;
#endif
};

REGISTER_TYPEOF(2901, AutoGatherAI::MineralClusters::iterator);
REGISTER_TYPEOF(2902, AutoGatherAI::MineralClusters::const_iterator);
//REGISTER_TYPEOF(2903, AutoGatherAI::M2C::iterator);
//REGISTER_TYPEOF(2904, AutoGatherAI::M2C::const_iterator);
REGISTER_TYPEOF(2905, AutoGatherAI::M2S::iterator);
REGISTER_TYPEOF(2906, AutoGatherAI::M2S::const_iterator);
#endif

