#include "Global.H"
#include "AutoGatherAI.H"
#include "Game1_accessories.H"
#include <numeric>

using namespace std;

typedef TerrainBase::Loc Loc;
typedef TerrainBase::Path Path;
typedef TerrainBase::Goal Goal;
typedef TerrainBase::Task Task;

//#define TEST(x)
#define TEST(x) x
#define TEST2(x)
//#define TEST2(x) x

/**************************************************************************/
static Loc get_center_loc(const Object *obj)
{
  Loc l;
  obj->get_center(l.x, l.y);
  return l;
}

/**************************************************************************/

bool AutoGatherAI::isMineralFreeFor(const GobPtrWrapper* mineral, const Worker& worker, bool allowDynamicallyBlocked, bool allowSpill) const
{
  if( !allowSpill ) {
    // This mineral must not be allocated to too many workers (counting correctly the 
    // case of worker already allocated to this mineral).
    if( minerals2workers.count(mineral) - ((getMineral(worker)==mineral)? 1:0) >= MAX_WORKERS_PER_MINERAL) {
      TEST2(cout << "(" << mineral->getPosition().x << "," << mineral->getPosition().y << ") full; "; );
      return false;
    }
  }

  const MineralState &mineralState = getMineralState(*mineral);

  // The path to the mineral must not be blocked (ie for inner minerals in a cluster)
  sint4 now = gsm.get_game().get_view_frame();
  if( mineralState.statically_blocked || (!allowDynamicallyBlocked && mineralState.dont_use_until > now) ) {
    TEST2(cout << "(" << mineral->getPosition().x << "," << mineral->getPosition().y << ") blocked; "; );
    return false;
  }

  // Apart from that, it must either not be allocated as part of cluster planning,
  // or else it must belong to the same cluster this worker is working on.
  if( mineralState.clusterId == -1 || mineralClusters[mineralState.clusterId].usePathAroundCluster == false )
    return true;
  const WorkerState &workerState = getState(worker);
  return (workerState.clusterId  == mineralState.clusterId);
}

/**************************************************************************/

/// A comparison functor used for sorting mineral patches according to distance from a worker
struct AutoGatherAI::CompareMineralsByDistanceToWorker {
  const AutoGatherAI& parent;
  const Worker& worker;
  CompareMineralsByDistanceToWorker(const AutoGatherAI &parent_, const Worker& worker_) 
    : parent(parent_), worker(worker_) {}
  bool operator()(const GobPtrWrapper* mineral1, const GobPtrWrapper* mineral2) {

    return Game::distance(*worker,**mineral1) < Game::distance(*worker,**mineral2);

    // "NEWER CRITERIA" is inconsistent and breaks sorting -> everything else

    // newer criteria - combine the distance from the current worker position
    // with the static path's length from the CC to the mineral :

    //const MineralState &state1 = parent.getMineralState(*mineral1);
    //const MineralState &state2 = parent.getMineralState(*mineral2);
    FIND(parent.minerals2state, state1it, mineral1);
    FIND(parent.minerals2state, state2it, mineral2);
    if(state1it != parent.minerals2state.end() && state2it != parent.minerals2state.end()) {
      const MineralState &state1 = state1it->second;
      const MineralState &state2 = state2it->second;

      // if any of the two is blocked
      if (state2.statically_blocked > 0) return true;  // 1 < 2
      if (state1.statically_blocked > 0) return false; // 1 !< 2

      // if no static info available
      if( state1.statically_blocked || state2.statically_blocked ) {
        return Game::distance(*worker,**mineral1) < Game::distance(*worker,**mineral2);
      } 

      // else, combine static and instantenous info.
      sint4 static_dist1 = state1.static_distance_from_cc;
      sint4 static_dist2 = state2.static_distance_from_cc;
      real8 cur_dist1 = Game::distance(*worker,**mineral1);
      real8 cur_dist2 = Game::distance(*worker,**mineral2); 
      return  (CUR_DIST_WEIGHT*cur_dist1 + STATIC_DIST_WEIGHT*static_dist1) < 
              (CUR_DIST_WEIGHT*cur_dist2 + STATIC_DIST_WEIGHT*static_dist2);
    } else {
      return Game::distance(*worker,**mineral1) < Game::distance(*worker,**mineral2);
    }
  }
};

void AutoGatherAI::allocateWithSingleMineralScheme(Worker& worker, bool forcedChange, bool allowDynamicallyBlocked, bool allowSpill)
{
  assert( getState(worker).clusterId == -1 && "tried single-mineral planning for a cluster-planning worker");

  const GobPtrWrapper* current_mineral = getMineral(worker);

/* we don't want workers getting stuck with distant minerals (which happens
   more often due to the dynamic blocking mechanism):
  // Is there a good currently allocated mineral ?
  if( !forcedChange && current_mineral != NULL && isMineralFreeFor(current_mineral, worker, allowDynamicallyBlocked, allowSpill) ) {
    return;
  }*/

  // find the closest mineral patch which is not allocated to too many workers

  // sort according to distance
  vector<const GameObjects::value_type*> sortable_minerals;

  FORALL(minerals, mIt) {
    const GobPtrWrapper& mineral = *mIt;
    sortable_minerals.push_back(&mineral);
  }
  CompareMineralsByDistanceToWorker comparator(*this, worker);
  sort(sortable_minerals.begin(), sortable_minerals.end(), comparator);

  // find the first free mineral, taking into account the 'forcedChange' instruction
  FORALL_B(sortable_minerals, freeMineralIt) {
    if( (!forcedChange || current_mineral != *freeMineralIt) && 
      isMineralFreeFor(*freeMineralIt, worker, allowDynamicallyBlocked, allowSpill) )
      break;
  }

  // if there is such a mineral, use it
  if( freeMineralIt != sortable_minerals.end() ) {
    const GobPtrWrapper* mineral = *freeMineralIt;
    associateWorkerAndMineral(worker, mineral);
  }
  // no available mineral
  else {
    // fall back to a dynamically blocked minerals
    if( !allowDynamicallyBlocked ) {
      allocateWithSingleMineralScheme(worker, forcedChange, true);
    }
    else {
      // fall back to minerals that already has many workers
      if( !allowSpill ) {
        allocateWithSingleMineralScheme(worker, forcedChange, true, true);
      }
      // really none available
      else {
        associateWorkerAndMineral(worker, NULL);
      }
    }
  }
}

/**************************************************************************/

/** A comparison functor used for sorting mineral patches 
    according to descending distance from a location */
struct CompareMineralsByDistanceToLocation {
  real8 x, y;
  CompareMineralsByDistanceToLocation(real8 x_, real8 y_) : x(x_), y(y_) {}
  bool operator() (const GobPtrWrapper* mineral1, const GobPtrWrapper* mineral2) {
    Loc l1 = get_center_loc(mineral1->get_GameObj());
    Loc l2 = get_center_loc(mineral2->get_GameObj());
    real8 dx1 = x - l1.x;
    real8 dy1 = y - l1.y;
    real8 dx2 = x - l2.x;
    real8 dy2 = y - l2.y;
    return ( dx1*dx1 + dy1*dy1 > dx2*dx2 + dy2*dy2 );
  }
};

void AutoGatherAI::allocateWithClusterScheme(Worker& worker, bool /*forcedChange*/)
{
  // POLICY: select 'outer' minerals in the cluster by sorting in descending
  // order of distance from the centroid.

  assert( getState(worker).clusterId != -1 && "tried cluster planning for a single-mineral-planning worker");

  MineralCluster &cluster = mineralClusters[getState(worker).clusterId];
  Vector<const GobPtrWrapper*> &minerals = cluster.mineralsInCluster;

  // sort according to descending distance from the centroid
  Vector<const GobPtrWrapper*> sortable_minerals;
  copy(minerals.begin(), minerals.end(), back_inserter(sortable_minerals));
  sort(sortable_minerals.begin(), sortable_minerals.end(), 
    CompareMineralsByDistanceToLocation(cluster.centroid.x, cluster.centroid.y));

  // find the first free mineral for this worker.
  typeof(sortable_minerals.begin()) freeMineralIt =
    find_if(sortable_minerals.begin(), sortable_minerals.end(), IsMineralFreeFor(this, worker));

  // if there is such a mineral, use it
  if( freeMineralIt != sortable_minerals.end() ) {
    const GobPtrWrapper* mineral = *freeMineralIt;
    associateWorkerAndMineral(worker, mineral);
  }
  // no available mineral
  else {
    // fall back to single-mineral planning
    getState(worker).clusterId = -1;
    allocateWithSingleMineralScheme(worker);
  }
}

/**************************************************************************/
void AutoGatherAI::allocateMineralPatchFor(Worker& worker, bool forcedChange)
{
  if( getState(worker).clusterId == -1 )
    allocateWithSingleMineralScheme(worker, forcedChange);
  else
    allocateWithClusterScheme(worker, forcedChange);
}

/**************************************************************************/
/**************************************************************************/
inline sint4 sumXPosition(sint4 v, const GobPtrWrapper *obj) {
  return v + get_center_loc(obj->get_GameObj()).x;
}

inline sint4 sumYPosition(sint4 v, const GobPtrWrapper *obj) {
  return v + get_center_loc(obj->get_GameObj()).y;
}

void AutoGatherAI::computeMineralClusters()
{
  // List all mineral locations
  Vector<ScalarPoint> mineralLocations;
  mineralLocations.reserve(minerals.size());
  FORALL(minerals, mIt) {
    mineralLocations.push_back(mIt->getPosition());
  }

  // Cluster
  Vector<sint4> clusterOf;
  simple_clustering(mineralLocations, MIN_NUM_CLUSTERS, MAX_NUM_CLUSTERS, clusterOf);

  // Store the results
  sint4 clusterCount = *max_element(clusterOf.begin(), clusterOf.end()) + 1;
  mineralClusters.clear(); mineralClusters.resize(clusterCount);
  int i=0; 
  FORALL(minerals, mIt) {
    const GobPtrWrapper &mineral = *mIt;
    mineralClusters[clusterOf[i]].mineralsInCluster.push_back(&mineral);
    getMineralState(mineral).clusterId = clusterOf[i];
    i++;
  }

  // Compute cluster centroids
  FORALL(mineralClusters, cluster) {
    Vector<const GobPtrWrapper*> &minerals = cluster->mineralsInCluster;
    real8 count = minerals.size();
    cluster->centroid.x = (sint4)(accumulate(minerals.begin(), minerals.end(), 0, sumXPosition) / count);
    cluster->centroid.y = (sint4)(accumulate(minerals.begin(), minerals.end(), 0, sumYPosition) / count);
  }

  TEST(
    cout << "mineral2cluster = [" << endl;
    FORU(i, mineralLocations.size()) 
      cout << mineralLocations[i].x << "," << mineralLocations[i].y << "," << clusterOf[i] << ";" << endl;
    cout << "];" << endl;
  );
}

/**************************************************************************/
ostream& operator<<(ostream& os, const Path& p)
{
  os << "[";
  FORALL(p.locs, w) {
    os<< w->x << "," << w->y << ";" << endl;
  }
  os << "];";
  return os;
}

/**************************************************************************/
bool AutoGatherAI::computeLanesToCluster(MineralCluster &cluster, sint4 radius)
{
  // A single leg of the composite path
  Path leg;
  // Goal for most legs
  Goal to_minerals_goal(Goal::LOCATION, Goal::VICINITY);   
  to_minerals_goal.distance = DEPLOYMENT_POINTS_DISTANCE_TO_CLUSTER;
  // Goal for 'go to the control center' leg
  Goal to_cc_goal(*controlCenters.begin(), Goal::TOUCH);
  // The actual length of each planned leg
  real8 length;
  // the two 'launching' points
  Loc fromMinerals, fromCC;
  // the "path" (actually a single waypoint) to the fromMinerals launching point
  Path launch;

  cluster.usePathAroundCluster = true;

  // Create a fake object for planning, start with some arbitrary worker which
  // is (assumed to be) initially located near the CC.
  Game1_SimpleTerrain::ObjectTranslocator obj(*workers.begin(), radius);

  // plan a path to the vicinity of the cluster's centroid.
  to_minerals_goal.loc = cluster.centroid;
  length = st.find_path(&obj, to_minerals_goal, &leg, TerrainBase::IGNORE_ALL_MOBILE_OBJS);

  if( length < 0 ) {
    // no path exists, try next cluster
    cluster.usePathAroundCluster = false;
    goto cleanup;
  }

  // use the final location found and plan from there to the CC
  fromMinerals = leg.locs.back();
  obj.set_position(fromMinerals.x, fromMinerals.y);
  length = st.find_path(&obj, to_cc_goal, &leg, TerrainBase::IGNORE_ALL_MOBILE_OBJS);

  if( length < 0 ) {
    // no path exists, try next cluster
    cluster.usePathAroundCluster = false;
    goto cleanup;
  }

  // cache this path
  cluster.cachedPaths.push_back(leg);

  // mark this path as blocked for finding the other path for this cluster
  st.block_path(&obj, leg);

  // now plan a path back from the CC to the vicinity of the cluster
  fromCC = leg.locs.back();
  obj.set_position(fromCC.x, fromCC.y);
  length = st.find_path(&obj, to_minerals_goal, &leg, TerrainBase::IGNORE_ALL_MOBILE_OBJS);

  if( length < 0 ) {
    // no path exists, try next cluster
    cluster.usePathAroundCluster = false;
    goto cleanup;
  }

  // cache this path too; keep the to-cluster path first in vector; then
  // a dummy path as a stand-in for the "go to the mineral itself"
  // movement; then the launching point; then the path back to the CC. 
  launch.locs.clear(); launch.locs.push_back(fromMinerals);
  cluster.cachedPaths.insert(cluster.cachedPaths.begin(), launch);
  cluster.cachedPaths.insert(cluster.cachedPaths.begin(), Path());
  cluster.cachedPaths.insert(cluster.cachedPaths.begin(), leg);

cleanup:
  obj.set_position(fromMinerals.x, fromMinerals.y);
  st.release_path(&obj, cluster.cachedPaths.back());

  return cluster.usePathAroundCluster;
}


/**************************************************************************/
void AutoGatherAI::computeLanesToClusters()
{
  assert(workers.size() >= 1 && 
         controlCenters.size() >= 1 &&
        "Must be called after the initial view msg was handled");

  const sint4 worker_radius = (*workers.begin()).get_GameObj()->get_radius();

  // FOR EACH CLUSTER
  FORALL(mineralClusters, cluster) {

    // don't use lanes for very close clusters
    if( Distance::distance(*(controlCenters.begin()->get_GameObj()), cluster->centroid.x, cluster->centroid.y)
          < TOO_CLOSE_CLUSTER )
      continue;

    // try creating wide lanes, fall-back to narrow lanes
    for( sint4 factor = 2 ; factor >= 1 ; factor-- ) {
      TEST(cout << "Trying to find a lane wide enough for " << factor << " workers" << endl;);
      if( computeLanesToCluster(*cluster, worker_radius * factor) )
        break;
    }
    TEST( ScalarPoint mp = cluster->mineralsInCluster.front()->getPosition();
          if( cluster->usePathAroundCluster ) {
            cout << "Lanes for cluster that includes the mineral at (" << mp.x << "," << mp.y << "):" << endl;
            FORALL(cluster->cachedPaths, p)
              cout << *p << endl;
          } else {
            cout << "No path around cluster that includes the mineral at (" << mp.x << "," << mp.y << ")" << endl;
          });
  }
}
/**************************************************************************/

void AutoGatherAI::InitialAllocationOfWorkers(void)
{
  // for each cluster that has a valid round-about path, allocate some 
  // (currently arbitrary) workers

  FORU(clusterId, mineralClusters.size()) {
    // if the cluster is intended for cluster-planning
    if( mineralClusters[clusterId].usePathAroundCluster ) {

      // allocate the requested number of workers
      sint4 count_workers = (sint4)(MAX_WORKERS_PER_CLUSTER_FACTOR * mineralClusters[clusterId].mineralsInCluster.size());
      FORU(i, count_workers) {

        // get a (currently) arbitrary free worker
        FORALL_B(workers, workerIt) {
          if( getState(*workerIt).mineral == NULL )
            break;
        }
        // if there's an available worker
        if( workerIt != workers.end() ) {
          // allocate it to the cluster
          getState(*workerIt).clusterId = clusterId;
          // and to some mineral in the cluster
          allocateMineralPatchFor(const_cast<Worker&>(*workerIt));

#if 0
          // and to some (currently arbitrary) mineral in the cluster
          FORALL_B(mineralClusters[clusterId].mineralsInCluster, mineralIt) {
            if( isMineralFreeFor(*mineralIt, *workerIt) ) {
              break;
            }
          }
          // if there's a free mineral
          if( mineralIt != mineralClusters[clusterId].mineralsInCluster.end() ) {
            associateWorkerAndMineral(const_cast<Worker&>(*workerIt), *mineralIt);
          } else {
            // fixme: what to do ??
          }
#endif
        }
        // no more workers
        else
          break;
      }
    }
  }

  // allocate all other workers an arbitrary mineral from the other clusters
  FORALL(workers, workerIt) {
    if( getState(*workerIt).mineral == NULL ) {
      allocateMineralPatchFor(const_cast<Worker&>(*workerIt));
    }
  }
}

/**************************************************************************/
void AutoGatherAI::testStaticBlock(const GobPtrWrapper *mineral)
{
  assert(workers.size() >= 1 && controlCenters.size() >= 1 && "Must be called after the initial view msg was handled");

  // Try planning a path from the center of the CC to this mineral.
  // if no path exists, then this mineral is blocked.

  // remove the CC itself from view 
  st.ignore_rectangles();

  // from center of CC
  Game1_SimpleTerrain::ObjectTranslocator worker(*workers.begin());
  Loc CC_center = get_center_loc(*controlCenters.begin());
  worker.set_position(CC_center.x, CC_center.y);

  // to the mineral
  Goal to_mineral(*mineral, Goal::TOUCH);

  // plan with a time limit
  st.set_find_path_max_time(TIME_FOR_SINGLE_STATIC_TEST);
  real8 length = st.find_path(&worker, to_mineral, NULL, TerrainBase::IGNORE_ALL_MOBILE_OBJS);
  st.set_find_path_max_time(0);

  // reconsider the CC
  st.consider_rectangles();

  // is blocked ?
  MineralState &state = getMineralState(*mineral);
  if( length < 0 ) {
    state.statically_blocked = 1;
    state.static_distance_from_cc = -1;
    sint4 now = gsm.get_game().get_view_frame();
    state.dont_use_until = now + BLOCKED_MINERAL_RETEST_IN_TICKS;
  } else {
    state.statically_blocked = 0;
    state.static_distance_from_cc = (sint4)length;
    state.dont_use_until = -1;
  }
}

void AutoGatherAI::markBlockedMinerals(void)
{
  FORALL(minerals, m) {
    testStaticBlock(&*m);
  }

  TEST( 
    cout << "mineral_is_blocked=[" << endl;
    FORALL(minerals2state, m2s) {
      cout << m2s->first->getPosition().x << "," << m2s->first->getPosition().y << "," << (sint4)(m2s->second.statically_blocked) << ";" << endl;
    }
    cout << "];" << endl;
  );
}

/**************************************************************************/
void AutoGatherAI::retestStaticBlock(void)
{
  sint4 now = gsm.get_game().get_view_frame();

  FORALL(minerals2state, m2s) {
    MineralState &state = m2s->second;
    if( state.statically_blocked && state.dont_use_until <= now ) {
      testStaticBlock(m2s->first);
      break;
    }
  }
}

/**************************************************************************/
void AutoGatherAI::plan()
{
  static bool firstTime = true;

  // initial planning
  if( firstTime ) {
    firstTime = false;

    computeMineralClusters();

    if( !noClusterPlanning ) {

      st.useLevelOfDetail(Game1_SimpleTerrain::ST_Terrain::MAXIMAL_DETAILS);
      computeLanesToClusters();
#if 0 // Version 1 of cluster-planning:
      computeWaypointsAroundClusters();
      computePathsAroundClusters();
#endif 
      st.useLevelOfDetail(Game1_SimpleTerrain::ST_Terrain::MINIMAL_DETAILS);
    }

    markBlockedMinerals();
    InitialAllocationOfWorkers();
  }
  
  // continuos planning
  else {
    // find a blocked mineral whose timeout passed and retest it
    retestStaticBlock();
  }
}



/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/

#if 0 // Version 1 of cluster-planning:

void AutoGatherAI::computeWaypointsAroundClusters()
{
  // fixme: there's probably a way to get worker/mineral radius from the blueprints.
  assert(workers.size() >= 1 && minerals.size() >= 1 && 
         "Must be called after the initial view msg was handled");
  const sint4 worker_radius = ((const GameObj*)(*workers.begin()))->get_radius();
  const sint4 mineral_radius = ((const GameObj*)(*minerals.begin()))->get_radius();


  // for each cluster, compute a roundabout path and blow it up a little
  FORALL(mineralClusters, cluster)
  {
    Vector<ScalarPoint> clusterPoints;
    clusterPoints.reserve(cluster->mineralsInCluster.size());
    FORALL(cluster->mineralsInCluster, mIt) {
      clusterPoints.push_back((*mIt)->getPosition());
    }
    Vector<ScalarPoint> roundabout;
    const GobPtrWrapper& controlCenter = *(controlCenters.begin());
    compute_roundabout(clusterPoints, controlCenter->getPosition(), roundabout);

    // close the polygon using the control center's position
    roundabout.push_back(controlCenter->getPosition());

    // enlarge it a little bit, so a worker can be mining close to the mineral 
    // and the path will still be clear for other workers.
    const sint4 blow_by = mineral_radius + WORKER_MINERAL_DISTANCE_THRESHOLD +
                          3 * worker_radius + 2;
    Vector<ScalarPoint> blownup;
    blowup_polygon(roundabout, blow_by, blownup);

    TEST( cout << "blownup = [" << endl;
          FORALL(blownup, j)  cout << j->x << "," << j->y << ";" << endl;
          cout << "];" << endl; );
    
    // copy to mineralClusters, remove the CC vertex of the path
    cluster->pathAroundCluster.clear();
    FORALL(blownup, vertexIt) {
      cluster->pathAroundCluster.push_back(Loc(vertexIt->x, vertexIt->y));
    }
    cluster->pathAroundCluster.pop_back();
  }
}

/**************************************************************************/

void AutoGatherAI::computePathsAroundClusters()
{
  assert(workers.size() >= 1 && 
         controlCenters.size() >= 1 &&
         map_width > 0 && map_height > 0 &&
        "Must be called after the initial view msg was handled");

  const sint4 worker_radius = ((const GameObj*)(*workers.begin()))->get_radius();
  
  // A single leg of the composite path
  Path leg;
  // The start point for the next leg
  Loc from;
  // Goal for most legs
  Goal goal(Goal::LOCATION, Goal::TOUCH);   
  // Goal for 'go to the control center' leg
  Goal to_cc_goal(Goal::OBJ, Goal::TOUCH);  
  to_cc_goal.obj = *controlCenters.begin();
  // Create a fake object for planning
  Game1_SimpleTerrain::ObjectTranslocator obj(*workers.begin());
  // The actual length of each planned leg
  real8 length;

  // FOR EACH CLUSTER
  FORALL(mineralClusters, cluster) {
    
    cluster->usePathAroundCluster = true;
    
    typeof(cluster->pathAroundCluster.begin()) waypoint = cluster->pathAroundCluster.begin();
    from = *waypoint++;

    // go over all waypoints
    for( ; waypoint != cluster->pathAroundCluster.end() ; waypoint++ ) {

      // if the waypoints are out-of-bounds, no path can exist
      if( from.x < worker_radius || from.x >= map_width-worker_radius || 
          from.y < worker_radius || from.y >= map_height-worker_radius ||
          waypoint->x < worker_radius || waypoint->x >= map_width-worker_radius || 
          waypoint->y < worker_radius || waypoint->y >= map_height-worker_radius ) {

        // no path exists, try next cluster
        cluster->usePathAroundCluster = false;
        goto cleanup;
      }

      // try PF to the next waypoint
      obj.set_position(from.x, from.y);
      goal.loc = *waypoint;
      length = st.find_path(&obj, goal, &leg, TerrainBase::IGNORE_ALL_MOBILE_OBJS);

      // If there's no path, or if it's too long, try next cluster
      if( length < 0 || length > from.distance(*waypoint)*LEGAL_ELONGATION_FACTOR ) {
        cluster->usePathAroundCluster = false;
        goto cleanup;
      }

      // cache the path and continue from the last point in it
      cluster->cachedPaths.push_back(leg);
      from = leg.locs.back();

      // mark this path as blocked for finding the other paths in this cluster
      st.block_path(&obj, leg);
    }

    // Now add paths to and from the CC

    // To the CC
    from = cluster->cachedPaths.back().locs.back();
    obj.set_position(from.x, from.y);
    length = st.find_path(&obj, to_cc_goal, &leg, TerrainBase::IGNORE_ALL_MOBILE_OBJS);
    // If there's no path, or if it's too long ?
    if( length < 0 || length > from.distance(get_center_loc(to_cc_goal.obj))*LEGAL_ELONGATION_FACTOR ) {
      // no path exists, try next cluster
      cluster->usePathAroundCluster = false;
      goto cleanup;
    }
    cluster->cachedPaths.push_back(leg);
    // mark this path as blocked when finding the last leg for this cluster
    st.block_path(&obj, leg);

    // From the CC
    from = cluster->cachedPaths.back().locs.back();
    obj.set_position(from.x, from.y);
    goal.loc = cluster->pathAroundCluster.front();
    length = st.find_path(&obj, goal, &leg, TerrainBase::IGNORE_ALL_MOBILE_OBJS);
    // If there's no path, or if it's too long ?
    if( length < 0 || length > from.distance(goal.loc)*LEGAL_ELONGATION_FACTOR ) {
      // no path exists, try next cluster
      cluster->usePathAroundCluster = false;
      goto cleanup;
    }
    // insert this leg as the first leg
    cluster->cachedPaths.insert(cluster->cachedPaths.begin(),leg);

cleanup:
    // release all blocked paths

    // if all went well, don't unblock the first leg, which is the path from the CC
    typeof(cluster->cachedPaths.begin()) p = cluster->cachedPaths.begin();
    if( cluster->usePathAroundCluster == true )
      p++;
    from = *cluster->pathAroundCluster.begin();
    for( ; p != cluster->cachedPaths.end() ; p++ ) {
      obj.set_position(from.x, from.y);
      st.release_path(&obj, *p);
      from = p->locs.back();
    }

    TEST( if(cluster->usePathAroundCluster) {
            cout << "path_around_cluster=[" << endl;
            FORALL(cluster->cachedPaths, p) {
              FORALL(p->locs, w)
                cout << w->x << "," << w->y << ";" << endl;
            }
            cout << "];" << endl;
          } else {
            cout << "No path around cluster whose first waypoint = " << cluster->pathAroundCluster.front() << endl;
          }
         );
  }
}
#endif
