#ifndef Game1_SimpleTerrain_H
#define Game1_SimpleTerrain_H

// $Id: Game1_SimpleTerrain.H 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

// simple grid-based terrain representation

#include "Global.H"
#include "Game1_TerrainBasicImp.H"

//typedef Game1_SimpleTerrain::TerrainBasicImp TerrainBasicImp; // Temp rerouting until after the tournament.

#include "Game1_ST_SimpleMap.H"
#include "Random.H"
#include "Coord.H"
#include "Distance.H"

#include "boost/shared_ptr.hpp"

//#define GAME1_SIMPLE_TERRAIN_NO_GFX // also define GAME1_NO_GFX in game1_main.C


namespace Game1_SimpleTerrain {

  typedef TerrainBase::Loc              Loc;
  typedef TerrainBase::Segment          Segment;  
  typedef TerrainBase::Task             Task;
  typedef TerrainBase::Path             Path;
  typedef TerrainBase::Goal             Goal;
  typedef TerrainBase::MoveCmd          MoveCmd;
  typedef TerrainBase::StatusMsg        StatusMsg;
  typedef TerrainBase::TaskId           TaskId;
  typedef TerrainBase::ConsiderObjects  ConsiderObjects;

  // The movement module - later this should probably made more general...
  class ST_BasicMove;
  class ST_ForceFieldMove;
  typedef ST_ForceFieldMove MoveEngine;

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
  /// make a CIRCLE Object appear to be at an arbitrary location / size
  struct ObjectTranslocator : public Object {
    const Object *obj;
    sint4 newx, newy, newr;

    ObjectTranslocator(const Object *obj_, sint4 radius_=-1) 
      : obj(obj_), newr( (radius_<0)? obj->get_radius() : radius_ )
    { 
      assert(obj->get_shape() == Object::CIRCLE);
      obj->get_center(newx, newy);
    };
    void set_position(sint4 newx_, sint4 newy_) { newx = newx_; newy = newy_; };

    // make it look like it's in a new location / size
    void  get_center(sint4 &x, sint4 &y, sint4 *px=0, sint4 *py=0) const
      { x = newx; y = newy; if(px) *px = newx; if(py) *py = newy; };
    real8 distance_to(const Object &other) const 
      { return Distance::distance(*this, other); };
    sint4 get_radius() const              { return newr; };

    // forward calls to the internal 'real' object
    Shape get_shape() const               { return obj->get_shape(); };
    sint4 get_max_speed() const           { return obj->get_max_speed(); };
    ZCat  get_zcat(ZCat *prev=0) const    { return obj->get_zcat(prev); };
    sint4 get_speed(sint4 *prev=0) const  { return obj->get_speed(prev); };
    bool  get_moving(bool *prev=0) const  { return obj->get_moving(prev); }; 
    bool is_pending_action(void) const    { return obj->is_pending_action(); };

    // not relevant for circles
    void get_p1(sint4 &, sint4 &, sint4 *, sint4 *) const { assert(0 && "Unimplemented for SimpleTerrain::ObjectTranslocator"); }; 
    void get_p2(sint4 &, sint4 &, sint4 *, sint4 *) const { assert(0 && "Unimplemented for SimpleTerrain::ObjectTranslocator"); };
  };

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
  // fixme: Transfer this functionality into ST_Task, this is a heritage from before
  // the TerrainBase interface.
  struct UnitPath
  {
    /// Path data
    //@{
    const Object *obj;
    Vector<Loc> path;
    Loc goal;
    int current_target; // index of current way point in path vector
    bool no_orders_yet; // true at the beginning, before the first move command is issued
    //@}

    /// Collision avoidance and resolution
    //@{
    // If adding fields - check ST_Task::set_path for needed updates.
    bool stopped_to_avoid_collision; // true if the object was stopped temporarily to avoid a collision
    int collision_step; // this is incremented while solving unexpected stops
    uint4 wakeup;       // if SDL_GetTicks() >= wakeup then this UnitPath is not
                        // sleeping ie not to be skipped over.
    Loc spot;           // spot of the last unexpected collision
    //@}

    UnitPath(const Object *obj_, const Vector<Loc> &path_);
    ~UnitPath() {};

    void mark_spot(const Loc &p) { spot = p; };
    void sleep(int time) { wakeup = SDL_GetTicks() + time; };
    bool sleeping() const { return wakeup > SDL_GetTicks(); };

    bool is_intersect(const Segment &) const;
  };

/*  // For use in ST_ForceFieldMove
  struct UnitPath_FF : public UnitPath
  {
    bool replanned;

    UnitPath_FF(const Object *obj_, const Vector<Loc> &path_) : UnitPath(obj_, path_), replanned(false){};
    ~UnitPath_FF() {};
  };
*/
  /// The internal data complementing Task data:
  struct ST_Task {
    Task task;                           ///< This is the actual task and is characterized by:
    bool isPending;                      ///< pending or yet to be planned
    bool isFailed;                       ///< failed, and try to replan 
    bool noPathFound;                    ///< could not find a path to that location 
    boost::shared_ptr<UnitPath> path;    // todo: replace with a cleaner implementation
    uint4 replanned;                     ///< Number of times this task was replanned due to failures

    // These are used to calculate the approximate priority of the task:
    /// When was the task created, as return by SDL_GetTicks()
    uint8 start_time;
    /// Straight-line length of the path, to prefer planning shorter paths first
    real8 path_dist;
 
    ST_Task(const Task& task_);
    ST_Task(const ST_Task& other) 
      : task(other.task), isPending(other.isPending),
        isFailed(other.isFailed), noPathFound(other.noPathFound), path(other.path),
        replanned(other.replanned), start_time(other.start_time), 
        path_dist(other.path_dist) {};
    void set_path(const Object *obj_, const Vector<Loc> &path_);
    bool isValid() const { return !isPending && !isFailed && !noPathFound && path; };
  private:
    ST_Task operator=(const ST_Task& other);
  };

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

  /// A Simple implementation of the TerrainBase interface.
  /** This TerrainBase implementation is using a relatively basic Astar algorithm,
      originally written by Sami Wagiaalla and adapted by David Deutscher. 
      Limitations:
      - Only considers a single Obj per Task (ignoring the second and on obj_ids).
      - Assumes all objects are either ON_LAND or IN_AIR (and that boundaries do not 
        apply to the latter, though this is currently always true so it's not really a
        limitation).
      - Using a target *object*:
        a) No pursuit is implemented. Instead, the target location at the moment 
           of pathfinding is used as the goal.
        b) If the object is removed, it's last known location is used as goal.
        Nonetheless, it is useful (mainly for static targets) since it allows a 
        'touch' condition - where the unit is not really at the exact goal location -
        to be reported as a successful arrival and not retried.
  */
  class ST_Terrain : public TerrainBasicImp<ST_Task>
  {
  public:
    ST_Terrain();
    virtual ~ST_Terrain() {};

    /// add command-line options specific to this implementation, using static methods of class Options
    static void add_options(void);

    //---------------------------------------------------------------------
    /// \name Implementation of relevant parts of the TerrainBase interface
    //---------------------------------------------------------------------
    //@{
    virtual void init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_);

    virtual void add_obj(const Object *obj);
    virtual void update_obj(const Object *obj);
    virtual void remove_obj(const Object *obj);
    virtual void add_segments(const Vector<Segment> &s);

    virtual void cancel_task(const Object *obj);
    virtual void remove_task(TaskId tid);

    virtual void execute_tasks(Vector<MoveCmd> &cmds, Vector<StatusMsg> &msgs);
    virtual bool plan_tasks(uint4 max_time=0);

    virtual real8 find_path(const Loc &l1, const Loc &l2, sint4 radius, Path *path, ConsiderObjects consider = CONSIDER_ALL);
    virtual real8 find_path(const Object *obj, const Goal &goal, Path *path, ConsiderObjects consider = CONSIDER_ALL);
    //@}
    //---------------------------------------------------------------------

    typedef std::set<const Object *> Objects;

    //---------------------------------------------------------------------
    /// \name Richer interfaces, candidates for adding to TerrainBase
    //---------------------------------------------------------------------
    //@{
    /** Mark the path as an obstacle to be avoided in future PFing.
        Expects a forward order of waypoints, as returned by TerrainBase::find_path(). */
    void block_path(const Object *obj, const Path &path, bool with_endpoints=false);
    /** Release a previously marked obstacle-path. 
        Releasing an unblocked-path might corrupt the inner world view.
        Expects a forward order of waypoints, as returned by TerrainBase::find_path(). */
    void release_path(const Object *obj, const Path &path, bool with_endpoints=false);
    /** Execute a previously planned path. The waypoints are assumed
        to be in forward order, as returned by TerrainBase::find_path().
        The caller is responsible that the given arguments are consistent. */
    TaskId execute_task(const Task &task, const Path &path);

    // temporary hack - we want to plan a path from the center of the CC (in case some
    // of its sides are blocked). For this we will order "ignore_rectangels()". 
    // A variation might be considered for the TerrainBase interface after the tournament.
    // These must be called in pairs, and consider_* must be restored before the next
    // view is processed.
    void ignore_rectangles(void);
    void consider_rectangles(void);

    // temporary hack - apply a time limit on "find_path()" calls.
    // instead of changing TerrainBase interface (and add another parameter) just
    // before the tournament, we do this:
    void set_find_path_max_time(uint4 find_path_max_time_) { find_path_max_time = find_path_max_time_; };
    //@}
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    /// \name Set operation modes/settings which are specifc to SimpleTerrain
    //---------------------------------------------------------------------
    //@{
    enum LevelOfDetail {MINIMAL_DETAILS,  ///< Lowest resolution <=> fast, less accurate path finding. This is the default.
                        MAXIMAL_DETAILS   ///< Highest resolution <=> slow, accurate path finding.
    };
    /// Setup the PF engine to use a specific level of detail until further notice.
    void useLevelOfDetail(LevelOfDetail lod);
    //@}
    //---------------------------------------------------------------------

  private:

    // prevent copying as we currently don't need it.
    ST_Terrain(const ST_Terrain& other);
    ST_Terrain& operator=(const ST_Terrain& other);

  protected:
 
    //---------------------------------------------------------------------
    /// \name Internal accessories
    //---------------------------------------------------------------------
    //@{
    /// Define the maximum amount of time that a long distance task can be put off
    static const uint4 MAXIMUM_TASK_AGE;

    /** Perform PF for a single pending task, and insert results to data structures. 
        Return true iff did some planning. */
    bool plan_pending_task(uint4 stop_time = 0);
    /** Perform PF for a single failed task, and insert results to data structures. 
        Return true iff did some planning. */
    bool plan_failed_task(uint4 stop_time = 0);
    /** Accessory function for planning either a pending or a failed task. */
    void plan_task(TIter tit, sint4 stop_time);
    /** Randomly choose a failed task. Return tasks.end() if no failed task exists. */
    TIter get_random_failed_task(void);
 
    /** Invalidate existing paths that intersect a given new boundary. */
    void invalidate_paths(const Vector<Segment> &segments);
    
    /** Find the best task to plan now.  This is based on the task's age and also
	how short its path is.*/
    std::deque<TIter>::iterator get_best_task(void);
    //@}
    //---------------------------------------------------------------------

    /// The mechanics of world representation and basic A* path-finding
    class PFEngine; friend class ST_Terrain::PFEngine;

    boost::shared_ptr<PFEngine> pfEngine;       ///< The internal Path Finding engine
    
    boost::shared_ptr<MoveEngine> moveEngine;   ///< The internal movement engine
    
    Objects objsToStop;                         ///< Objects for which we should set a stop-moving action
    Random rand;
    sint4 currentLevelOfDetail;                 ///< Which LOD to use for PFing

    uint4 find_path_max_time; // temporary hack, see "set_max_time()"

    friend class SimpleTerrainWidget;
    friend class SimpleTerrainOverlay;
  };

} // end of namespace SimpleTerrain

REGISTER_TYPEOF(2501, Vector<Game1_SimpleTerrain::ST_Task*>::iterator);
REGISTER_TYPEOF(2502, Vector<Game1_SimpleTerrain::ST_Task*>::const_iterator);
REGISTER_TYPEOF(2503, Vector<const Game1_SimpleTerrain::ST_Task*>::iterator);
REGISTER_TYPEOF(2504, Vector<const Game1_SimpleTerrain::ST_Task*>::const_iterator);
REGISTER_TYPEOFS_TerrainBasicImp_OneTime(2505, Game1_SimpleTerrain::ST_Task);
REGISTER_TYPEOFS_TerrainBasicImp_ForEach(2509, Game1_SimpleTerrain::ST_Task);

#endif
