#include "General.H"

#include "Global.H"
#include "GameStateModule.H"
#include "GameChanges.H"
#include "Game.H"
#include "BaseCommander.H"
#include "BuildCommander.H"
#include "History.H"
#include "ScoutCommander.H"
#include "ModuleSet.H"
#include "TerrainBase.H"
#include "Finance.H"
#include "BaseClassifier.H"
#include "DefenseCommander.H"
#include "Delegate.H"

using TerrainAnalysis::BaseInfo;

General::General(ModuleSet & ms) : 
    Commander(ms, 25),
    receivedFirstView(false),
    gsm(ms.get_game_state()),
    mc(ms.get_movement(), "Default", "Default"),
    fm(ms.get_finance()),
    bc(ms),
    expansion(NULL, TerrainBase::Loc(0,0), NULL, NULL),
    scoutCommander(NULL),
    assaultCommander(NULL),
    hasFactory(false)
{
    gsm.add_handler(this);

    mc.setStatusProcedure( make_delegate(this, &General::onMovement) );

    ms.get_history().trackBlueprint("mineral");
    ms.get_history().trackBlueprint("controlCenter");
    ms.get_history().trackBlueprint("barracks");
    ms.get_history().trackBlueprint("factory");

    scoutCommander = new ScoutCommander(this);
}

General::~General()
{
    gsm.remove_handler(this);
}

static sint4 locDistSq(const TerrainBase::Loc & a,
                    const TerrainBase::Loc & b)
{
    sint4 dx = a.x - b.x;
    sint4 dy = a.y - b.y;
    return dx*dx + dy*dy;
}

static real8 locDistance(const TerrainBase::Loc & a,
                         const TerrainBase::Loc & b)
{
    return sqrt( (real8)locDistSq(a, b) );
}

const BaseInfo * 
General::getNearestAlliedBase(const TerrainBase::Loc & loc) const
{
    const BaseInfo * best = NULL;
    sint4 bestDistance = 0;

    const Vector<const BaseInfo*> & bases = bc.get_bases();
    FORALL(bases, b)
    {
        const BaseInfo * base = (*b);
        if(base->getOwner() == BaseInfo::ALLIED ||
           base->getOwner() == BaseInfo::CONTESTED)
        {
            TerrainBase::Loc baseLoc = base->getCentralLocation();
            sint4 distance = locDistSq(baseLoc, loc);
            if(best == NULL || distance < bestDistance)
            {
                best = base;
                bestDistance = distance;
            }
        } 
    }

    return best;
}

const BaseInfo * 
General::getNearestEnemyBase(const TerrainBase::Loc & loc) const
{
    const BaseInfo * best = NULL;
    sint4 bestDistance = 0;

    const Vector<const BaseInfo*> & bases = bc.get_bases();
    FORALL(bases, b)
    {
        const BaseInfo * base = (*b);
        if(base->getOwner() == BaseInfo::ENEMY ||
           base->getOwner() == BaseInfo::CONTESTED)
        {
            TerrainBase::Loc baseLoc = base->getCentralLocation();
            sint4 distance = locDistSq(baseLoc, loc);
            if(best == NULL || distance < bestDistance)
            {
                best = base;
                bestDistance = distance;
            }
        } 
    }

    return best;
}

real8 General::getDistanceFromBase(const TerrainBase::Loc & loc,
                                   const BaseInfo * base)
{
    TerrainBase::Loc baseLoc = base->getCentralLocation();
    return locDistance(loc, baseLoc);
}

sint4 General::getExpansionScore(const BaseInfo * base) const
{
    static const real8 allyDistancePenalty = 1.0;
    static const real8 enemyDistanceBonus  = 1.0;

    TerrainBase::Loc loc = base->getResourceLocation();
    const BaseInfo * ally = getNearestAlliedBase(loc);
    const BaseInfo * enemy = getNearestEnemyBase(loc);

    sint4 score = base->resourceWorth();
    if(ally != NULL)
    {
        real8 distance = getDistanceFromBase(loc, ally);
        if(distance < 320)
        {
            score = 0;
        }
        else
        {
            score -= (sint4)(distance * allyDistancePenalty);
        }
    }
    if(enemy != NULL)
    {
        score += (sint4)(getDistanceFromBase(loc, enemy) * enemyDistanceBonus);
    }
    return score;
}

sint4 General::getTargetScore(const BaseInfo * base) const
{
    static const real8 allyDistancePenalty = 1.0;

    TerrainBase::Loc loc = base->getResourceLocation();
    const BaseInfo * ally = getNearestAlliedBase(loc);

    // TODO: Increase score for a weaker base
    sint4 score = 0;
    if(ally != NULL)
    {
        score -= (sint4)(getDistanceFromBase(loc, ally) * allyDistancePenalty);
    }
    return score;
}

const BaseInfo * General::getBestExpansion() const
{
    const BaseInfo * best = NULL;
    sint4 bestScore = 0;

    const Vector<const BaseInfo*> & bases = bc.get_bases();
    FORALL(bases, b)
    {    
        if((*b)->getOwner() == BaseInfo::UNCLAIMED)
        {
            sint4 score = getExpansionScore( (*b) );
            if(best == NULL || score > bestScore)
            {
                best = (*b);
                score = bestScore;
            }
        }
    }

    return best;
}

const BaseInfo * General::getBestTarget() const
{
    const BaseInfo * best = NULL;
    sint4 bestScore = 0;

    const Vector<const BaseInfo*> & bases = bc.get_bases();
    FORALL(bases, b)
    {    
        if((*b)->getOwner() == BaseInfo::ENEMY ||
           (*b)->getOwner() == BaseInfo::CONTESTED)
        {
            sint4 score = getTargetScore( (*b) );
            if(best == NULL || score > bestScore)
            {
                best = (*b);
                score = bestScore;
            }
        }
    }

    return best;    
}

BaseCommander * General::getBaseWithMostRequestable(const std::string & bp_name)
{
    BaseCommander * best = NULL;
    size_t bestAmount = 0;

    FORALL(bases, b)
    {
        size_t amount = (*b)->countRequestable(bp_name);
        if( best == NULL || amount > bestAmount )
        {
            best = (*b);
            bestAmount = amount;
        }
    }
        
    return best;
}

BaseCommander * General::getBaseWithMostOwned(const std::string & bp_name)
{
    BaseCommander * best = NULL;
    size_t bestAmount = 0;

    FORALL(bases, b)
    {
        size_t amount = (*b)->countOwned(bp_name);
        if( best == NULL || amount > bestAmount )
        {
            best = (*b);
            bestAmount = amount;
        }
    }
        
    return best;
}

BaseCommander * General::getBaseWithLeastOwned(const std::string & bp_name)
{
    BaseCommander * best = NULL;
    size_t bestAmount = 0;

    FORALL(bases, b)
    {
        size_t amount = (*b)->countOwned(bp_name);
        if( best == NULL || amount < bestAmount )
        {
            best = (*b);
            bestAmount = amount;
        }
    }
        
    return best;
}

size_t General::acquireUnits(const std::string & bp_name, 
                             size_t amount, bool priority)
{
    size_t count = 0; // Start with amount we actually have
    while(count < amount)
    {
        BaseCommander * base = getBaseWithMostRequestable(bp_name);
        if( base == NULL )
        {
            break;
        }
        count += base->requestUnits(this, bp_name, 1);
    }
    if(priority)
    {
        while(count < amount)
        {
            BaseCommander * base = getBaseWithMostOwned(bp_name);
            if( base == NULL )
            {
                break;
            }
            count += base->demandUnits(this, bp_name, 1);                
        }
    }
    return count;
}

size_t General::acquireAllUnits(const std::string & bp_name, bool priority)
{
    size_t ALL = 10000; // Some suitably big number

    size_t count = 0; // Start with amount we actually have
    FORALL(bases, b)
    {
        count += (*b)->requestUnits(this, bp_name, ALL);
    }
    if(priority)
    {
        FORALL(bases, b)
        {
            count += (*b)->demandUnits(this, bp_name, ALL);
        }
    }
    return count;
}

size_t General::countAllUnits(const std::string & bp_name, bool priority)
{
    size_t count = 0;
    if(assaultCommander != NULL)
    {
        count += assaultCommander->countOwned(bp_name);
    }
    if(priority)
    {
        FORALL(bases, b)
        {
            count += (*b)->countOwned(bp_name);
        }
    }
    else
    {
        FORALL(bases, b)
        {
            count += (*b)->countRequestable(bp_name);
        }
    }
    return count;
}

void General::pickExpansionTarget(TerrainBase::Loc & target)
{
    static Random r; // Hack to fix compilation break when calling rand()

    target = expansion.location;
    switch(r.rand_uint4() % 4)
    {
    case 0:
        target.x += 48;
        break;
    case 1:
        target.y += 48;
        break;
    case 2:
        target.x -= 48;
        break;
    case 3:
        target.y -= 48;
    }
}

bool General::pickExpansionLocation()
{
    const BaseInfo * best = getBestExpansion();

    if(best == NULL)
    {
        return false;
    } 

    // TODO: Evaluate whether or not it is worth expanding to
    // the base we have chosen

    std::cerr << "GENERAL considering expansion to unclaimed ";
    std::cerr << "base worth " << best->resourceWorth();
    std::cerr << std::endl;

    expansion.location = best->getResourceLocation();
    expansion.hold  = fm.hold( fm.get_cost("controlCenter") );
    expansion.state = ExpandTask::PICKED;

    return true;
}

void General::startExpansion()
{
    expansion.funds = expansion.hold->ticket();
    expansion.hold = NULL;
   
    if( acquireUnits("worker", 1, false) == 1 )
    {
        GameObj * w = *(workers.begin());
        this->removeUnit( w );
        expansion.worker = w;
        expansion.state = ExpandTask::MOVING;
    }

    if(expansion.state == ExpandTask::MOVING)
    {
        std::cerr << "GENERAL expanding to (";
        std::cerr << expansion.location.x << ", ";
        std::cerr << expansion.location.y << ")";
                
        TerrainBase::Loc target;
        pickExpansionTarget(target);

	mc.moveUnit(expansion.worker, Movement::TouchPoint(target));

        size_t count = 0;
        FORALL( bases, b )
        {
            count += (*b)->requestUnits(this, "marine", 2 - count);
        }
        if(assaultCommander != NULL)
        {
            count += assaultCommander->demandUnits(this, "marine", 2 - count);
        }

        if(expansion.cover != NULL) delete expansion.cover;
        expansion.cover = new DefenseCommander(this, target, 160);
        std::set<GameObj*> coverForce = army;
        this->removeUnits(coverForce);
        expansion.cover->addUnits(coverForce);
    }
}

bool General::pickAttackTarget()
{
    const BaseInfo * best = getBestTarget();

    if(best != NULL)
    {
        if(assaultCommander)
        {
            delete assaultCommander;
            assaultCommander = NULL;
        }
            
        acquireAllUnits("marine", false);
        acquireAllUnits("tank", false);
            
        assaultCommander = new DefenseCommander(
            this,
            best->getCentralLocation(),
            256);
        
        std::set<GameObj*> assaultForce = army;
        this->removeUnits(assaultForce);
        assaultCommander->addUnits(assaultForce);

        std::cerr << std::string(40, '!') << std::endl;
        std::cerr << "MASS ASSAULT HAS BEGUN" << std::endl;
        std::cerr << std::string(40, '!') << std::endl;
        
        return true;
    }
    else
    {
        return false;
    }
}

void General::cleanupFinishedBuilders()
{
    while(!finishedBuilders.empty())
    {
        BuildCommander * finished = *(finishedBuilders.begin());
        finishedBuilders.erase(finished);
        delete finished;
    }
}

void General::handleNewBases()
{
    // If we have spare control centers, start a base
    while(!controlCenters.empty())
    {           
        Vector<GameObj*> give;

        // Give one control center
        GameObj * cc = *(controlCenters.begin());
        this->removeUnit(cc);
        give.push_back(cc);

        // Give up to six workers
        // TODO: Request workers from other bases
        // Requires completion of TODO on acquireUnits()
        for(size_t i = 0; i < 6; i++)
        {
            if(!workers.empty())
            {
                GameObj * w = *(workers.begin());
                this->removeUnit(w);
                give.push_back(w);
            }
            else
            {
                break;
            }
        }

        // Delegate to a base commander
        TerrainBase::Loc loc; cc->get_center(loc.x, loc.y);
        BaseCommander * base = new BaseCommander(this, give, loc);

        if(expansion.cover != NULL)
        {
            delete expansion.cover;
            expansion.cover = NULL;
            std::set<GameObj*> defForce = army;
            this->removeUnits(defForce);
            base->addUnits(defForce);
        }

        bases.insert(base);
    }
}

void General::decideIfAttacking()
{
    //uint4 myWorth = fm.getAlliedWorth();
    //uint4 enemyWorth = fm.getAlliedWorth();

    // If we have an army and some targets, start fighting
    if( assaultCommander == NULL ) 
    {
        size_t marineCount = countAllUnits("marine", false);
        size_t tankCount = countAllUnits("tank", false);
        size_t forceSize = marineCount + tankCount * 3;

        // Under what conditions should we begin our assault?
        // Currently, if we have a small force to spare, attack
        if( forceSize > 4 )
        {
            pickAttackTarget();
        }
    }

    // If we are attacking
    if( assaultCommander != NULL )
    {
        // Reinforce with all available forces
        acquireAllUnits("marine", false);
        acquireAllUnits("tank", false);
        std::set<GameObj*> reinforcements = army;
        this->removeUnits(reinforcements);
        assaultCommander->addUnits(reinforcements);

        // If we have beaten the enemy in one location
        if(assaultCommander->isUnderControl())
        {
            // Assault another
            pickAttackTarget();
        }
    }
}


void General::decideIfExpanding()
{
    // If we have the opportunity to expand, try to
    if( expansion.state == ExpandTask::UNPICKED )
    {      
        // Only expand if we have enough forces thus
        // far.

        size_t baseCount = bases.size();
        size_t forceReq = baseCount * (baseCount + 5);
        size_t forceSize = countOwned("marine") + 
                           countOwned("tank") * 3;

        if( forceSize < forceReq )
        {
            return;
        }

        pickExpansionLocation();
    }
        
    // Once we have the funds, start moving out
    if( expansion.state == ExpandTask::PICKED && 
        expansion.hold->ready() &&
        countAllUnits("marine", false) >= 2 )
    {
        startExpansion();
    }

    // Create a build commander if we are expanding
    if( expansion.state == ExpandTask::READY )
    {
		// TODO: Pick a better location from which to search
		TerrainBase::Loc workerLoc;
		expansion.worker->get_center(workerLoc.x, workerLoc.y);
		
        BuildCommander * builder = new BuildCommander(
            this,
            expansion.worker,
            "controlCenter",
			workerLoc,
            expansion.funds,
            8);

        expansion.state = ExpandTask::EXPANDING;
        expansion.worker = NULL;
        expansion.funds = NULL;

        activeBuilders.insert(builder);
    }
}

void General::exec() 
{
    if (!receivedFirstView)
    {
        return;
    }

    bc.clear();
    bc.update(); // Update the base classifier

    std::cerr << "=== Recon Summary ====" << std::endl;
    std::cerr << "Allied net worth: " << fm.getAlliedWorth() << std::endl;
    std::cerr << "Enemy net worth:  " << fm.getEnemyWorth() << std::endl;

    const Vector<const BaseInfo*> & bi = bc.get_bases();
    FORALL(bi, b)
    {
        switch((*b)->getOwner())
        {
        case BaseInfo::UNCLAIMED: std::cerr << "Unclaimed"; break;
        case BaseInfo::ALLIED:    std::cerr << "Allied"; break;
        case BaseInfo::ENEMY:     std::cerr << "Enemy"; break;
        case BaseInfo::CONTESTED: std::cerr << "Contested"; break;
        }
        std::cerr << " base worth " << (*b)->resourceWorth();
        std::cerr << " minerals" << std::endl;
    }

    std::cerr << "======================" << std::endl;

    cleanupFinishedBuilders();

    handleNewBases();

    decideIfAttacking();

    decideIfExpanding();           

    // Base management

    size_t count = 0;
    FORALL (bases, b) 
    {
        BaseCommander * base = (*b);

        uint4 workerCost   = fm.get_cost("worker");
        uint4 marineCost   = fm.get_cost("marine");
        uint4 tankCost     = fm.get_cost("tank");
        uint4 barracksCost = fm.get_cost("barracks");
        uint4 factoryCost  = fm.get_cost("factory");            
        
        size_t workersOwned  = base->countOwned("worker");
        size_t workersQueued = base->countQueued("worker");
        size_t workerCount   = workersOwned + workersQueued;
        size_t workersNeeded = base->countWorkersNeeded();
        if(workersNeeded > 12) 
        {
            // Can thrash dangerously with too many units
            // gathering from one area. For now, this is
            // the only practical solution.
            workersNeeded = 12;
        }
        
        size_t marinesOwned  = base->countOwned("marine");
        size_t marinesQueued = base->countQueued("marine");
        size_t marineCount   = marinesOwned + marinesQueued;
        
        size_t tanksOwned  = base->countOwned("tank");
        size_t tanksQueued = base->countQueued("tank");
        size_t tankCount   = tanksOwned + tanksQueued;
        
        size_t barracksOwned  = base->countOwned("barracks");
        size_t barracksQueued = base->countQueued("barracks");
        size_t barracksCount  = barracksOwned + barracksQueued;

        size_t factoryOwned  = base->countOwned("factory");
        size_t factoryQueued = base->countQueued("factory");
        size_t factoryCount  = factoryOwned + factoryQueued;

        size_t scoutCount    = 0;
        if(scoutCommander != NULL)
        {
            scoutCommander->countObjects("worker", scoutCount);
            scoutCommander->countObjects("marine", scoutCount);
        }
        
        std::cerr << "=== Base " << ++count << " Summary ===" << std::endl;
        std::cerr << "Workers:  " << workersOwned;
        std::cerr << "/" << workerCount << std::endl;
        if(marineCount > 0)
        {
            std::cerr << "Marines:  " << marinesOwned;
            std::cerr << "/" << marineCount << std::endl;
        }
        if(tankCount > 0)
        {
            std::cerr << "Tanks:  " << tanksOwned;
            std::cerr << "/" << tankCount << std::endl;
        }
        if(barracksCount > 0)
        {
            std::cerr << "Barracks: " << barracksOwned;
            std::cerr << "/" << barracksCount << std::endl;
        }
        if(factoryCount > 0)
        {
            std::cerr << "Factory: " << factoryOwned;
            std::cerr << "/" << factoryCount << std::endl;
        }
        std::cerr << "======================" << std::endl;
        
        // If we have at least two barracks, and we have the funds,
        // construct a factory
        if ( factoryCount < 1 &&
             base->canBuild("factory") && 
             countAllUnits("barracks", true) > 1 &&
             fm.available() > factoryCost ) 
        {
            base->buildBuilding("factory", fm.reserve(factoryCost));
            continue;
        }

        // Any time we do not have a barracks built or queued,
        // but we can afford one, build it
        if ( barracksCount < 1 &&
             base->canBuild("barracks") && 
             fm.available() > barracksCost ) 
        {
            base->buildBuilding("barracks", fm.reserve(barracksCost));
            continue;
        }
     
        // Any time we can afford to do so, keep a tank queued
        if ( tanksQueued < 1 &&
             base->canBuild("tank") && 
             fm.available() > tankCost ) 
        {
            base->buildUnit("tank", fm.reserve(tankCost));
            continue;
        }
   
        // Any time we can afford to do so, keep a marine queued
        if ( marinesQueued < 1 &&
             base->canBuild("marine") && 
             fm.available() > marineCost ) 
        {
            base->buildUnit("marine", fm.reserve(marineCost));
            continue;
        }

        // Build up to six workers before constructing a barracks
        if ( workerCount < 6 &&
             workersQueued < 1 &&
             base->canBuild("worker") &&
             fm.available() > workerCost ) 
        {
            base->buildUnit("worker", fm.reserve(workerCost));
            continue;
        }

        // Build as many workers as possible once we have a barracks
        if ( workerCount < workersNeeded &&
             barracksCount > 0 &&
             workersQueued < 1 &&
             base->canBuild("worker") &&
             fm.available() > workerCost ) 
        {
            base->buildUnit("worker", fm.reserve(workerCost));
            continue;
        }
        
        // If we have at least six workers
        // we can let one worker scout
        if( (workersOwned > 5 && scoutCount < 1) ||
            (workersOwned > 7 && scoutCount < 2) ||
            (workersOwned > 9 && scoutCount < 3) )
        {
            base->requestUnits(this, "worker", 1);
            
            if(workers.empty() == false)
            {
                GameObj * w = *(workers.begin());
                this->removeUnit( w );
                scoutCommander->addUnit( w );
            }
            
            continue;
        }
    }
}

void General::onReceiveUnit(GameObj * unit, Watcher::WatchID)
{
    if(unit->bp_name() == "worker")
    {
        workers.insert(unit);
    }
    else if(unit->bp_name() == "marine" ||
            unit->bp_name() == "tank")
    {
        army.insert(unit);
    }
    else if(unit->bp_name() == "controlCenter")
    {
        expansion.state = ExpandTask::UNPICKED;
        controlCenters.insert(unit);
    }
    else
    {
        otherBuildings.insert(unit);
    }
}

void General::onLoseUnit(GameObj * unit)
{
    controlCenters.erase(unit);
    otherBuildings.erase(unit);
    workers.erase(unit);
    army.erase(unit);

    if(expansion.worker == unit)
    {
        expansion.funds->release();
        expansion.state = ExpandTask::UNPICKED;
    }
}

void General::onMovement(const Movement::Task &)
{
    if(expansion.state == ExpandTask::MOVING)
    {
	if(Distance::distance(*expansion.worker,
			      expansion.location.x,
			      expansion.location.y) < 64)
	{
	    expansion.state = ExpandTask::READY;
	}
	else
	{
	    TerrainBase::Loc target;
	    pickExpansionTarget(target);
	    mc.moveUnit(expansion.worker, Movement::TouchPoint(target));
	}
    }
}

bool General::onEvent(const Event & event) 
{
    if (event.get_what() == GameStateModule::VIEW_MSG) 
    {
        if (!receivedFirstView) 
        {
            insertUnits(gsm.get_changes().new_objs);
        }
        this->update();
        return true;
    }

    if (event.get_who() == BuildCommander::FROM)
    {
        if (event.get_what() == BuildCommander::FINISHED)
        {
            const BuildFinishedEvent & e = 
                reinterpret_cast<const BuildFinishedEvent &>(event);
            BuildCommander * builder = e.getSource();
            activeBuilders.erase(builder);
            finishedBuilders.insert(builder);
            return true;
        }
        if (event.get_what() == BuildCommander::FAILED)
        {
            const BuildFailedEvent & e = 
                reinterpret_cast<const BuildFailedEvent &>(event);
            BuildCommander * builder = e.getSource();
            activeBuilders.erase(builder);
            finishedBuilders.insert(builder);
            expansion.state = ExpandTask::UNPICKED;
            return true;
        }
    }
    return false;
}

void General::insertUnits(const Vector<ScriptObj::WPtr>& objects)
{
    const sint4 me = gsm.get_game().get_client_player();
        
    FORALL_DYN(objects, it) 
    {
        GameObj* gob = (*it)->get_GameObj();
        if( !gob ) continue;
        if( *(gob->sod.owner) == me ) 
        {
            addUnit(gob);
        }          
    }

    receivedFirstView = true;
}
