// $Id: TM_Score.C 7311 2008-07-29 17:18:36Z mburo $

// (c) M.Buro, licensed under the GPL 3

// AIIDE 2006/2007 tournament scoring functions

#include "TM_Score.H"
#include <string>
#include <sstream>
#include <iostream>
#include <cassert>
#include <cstdlib>

using namespace std;

#define SERR(s) {\
  std::stringstream ss; ss << __FILE__ << " " << __FUNCTION__ \
  << "() (line " << __LINE__ << "): " << s; \
  std::cerr << "GAME RESULT PARSE ERROR: " << ss.str() << std::endl;	\
  throw 0; \
}


// return true iff error in result string


// 2 two-player game results -> scores for player 1 (in first game) and player 2

// game result line:
//
// @ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 |
//              p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |
// or
//
// @ TIME 5000  p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 |
//              p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |

// game 4: same format
// @ EARLY 484 p0 m 0 s -2500 d 2250 ub 0 ul 50 bb 0 bl 0 t 0 |
//             p1 m 0 s -2250 d 2500 ub 0 ul 45 bb 0 bl 0 t 0 |

struct PlayerResult
{
  string player;
  int mineral_count;
  int standing_cost;
  int destroyed_cost;
  int units_built;
  int units_lost;
  int buildings_built;
  int buildings_lost;
  int time_my_last_building_was_destroyed;

  void parse(istream &is)
  {
    string s;
    
    is >> player; 

    is >> s; if (s != "m") SERR("m expected, but read " << s);
    is >> mineral_count; if (!is) SERR("integer expected");
    
    is >> s; if (s != "s") SERR("s expected, but read " << s);
    is >> standing_cost; if (!is) SERR("integer expected");

    is >> s; if (s != "d") SERR("d expected, but read " << s);
    is >> destroyed_cost; if (!is) SERR("integer expected");

    is >> s; if (s != "ub") SERR("ub expected, but read " << s);
    is >> units_built; if (!is) SERR("integer expected");
    
    is >> s; if (s != "ul") SERR("ul expected, but read " << s);
    is >> units_lost; if (!is) SERR("integer expected");

    is >> s; if (s != "bb") SERR("bb expected, but read " << s);
    is >> buildings_built; if (!is) SERR("integer expected");
    
    is >> s; if (s != "bl") SERR("bl expected, but read " << s);
    is >> buildings_lost; if (!is) SERR("integer expected");

    is >> s; if (s != "t") SERR("t expected, but read " << s);
    is >> time_my_last_building_was_destroyed; if (!is) SERR("integer expected");
  }
};


// one-player game result

struct GameResult1P
{
  PlayerResult p1;
  int time;
  bool timeout;
  
  void parse(const string &str)
  {
    istringstream is(str);
    char c;
    string s;

    is >> c; if (c != '@') SERR("@ expected, but read " << c << " in " << str);
    is >> s;
    timeout = s == "TIME";
    if (s != "TIME" && s != "EARLY")
      SERR("TIME or EARLY expected, but read " << s << " in " << str);

    is >> time; if (!is) SERR("integer expected after TIME/EARLY in " << str);

    p1.parse(is);
  }
};


// two-player game result

struct GameResult2P
{
  PlayerResult p1, p2;
  int time;
  bool timeout;
  
  void parse(const string &str)
  {
    istringstream is(str);
    char c;
    string s;

    is >> c; if (c != '@') SERR("@ expected, but read " << c << " in " << str);
    is >> s;
    timeout = s == "TIME";
    if (s != "TIME" && s != "EARLY")
      SERR("TIME or EARLY expected, but read " << s << " in " << str);

    is >> time; if (!is) SERR("integer expected after TIME/EARLY in " << str);

    p1.parse(is);
    is >> s; if (s != "|") SERR("| expected, but read " << s << " in " << str);

    p2.parse(is);
  }
};


//  Game 1: return mineral count

bool score1(const std::string &res1, const std::string &dummy, float &sc1, float &/*sc2*/)
{
  if (!dummy.empty()) { cerr << "dummy not empty in score1" << endl; exit(20); }
  
  GameResult1P g1;
  sc1 = -1;
  
  try {
    g1.parse(res1);
  }
  catch (...) {
    return true;
  }

  sc1 = g1.p1.mineral_count;
  return false;
}


/*
  Game 2:
  
  When all buildings of a player are destroyed the game ends instantly.

  s = # of destroyed opponent buildings in game 1 and game 2
  r = s(player 1) - s(player 2)

  If r > 0, player 1 wins.
  If r < 0, player 2 wins.
  If r = 0:

    (tiebreaker) 

    s = - elapsed time when last opponent building is destroyed in game 1
        - elapsed time when last opponent building is destroyed in game 2

    r = s(player 1) - s(player 2) 

    If r > 0, player 1 wins.
    If r < 0, player 2 wins.
    If r = 0, game is a tie
*/

bool score2(const std::string &res1, const std::string &res2, float &sc1, float &sc2)
{
  if (res2.empty()) { cerr << "res2 empty in score2" << endl; exit(20); }
  
  GameResult2P g1, g2;
  sc1 = sc2 = -1;
  
  try {
    g1.parse(res1);
    g2.parse(res2);
  }
  catch (...) {
    return true;
  }

  int s_1 = g1.p2.buildings_lost + g2.p1.buildings_lost;
  int s_2 = g1.p1.buildings_lost + g2.p2.buildings_lost;

  int r = s_1 - s_2;
  if      (r > 0) sc1 = 1.0;
  else if (r < 0) sc1 = 0.0;
  else {

    // tiebreaker : sum of destruction times

    s_1 = g1.p1.time_my_last_building_was_destroyed + g2.p2.time_my_last_building_was_destroyed;
    s_2 = g1.p2.time_my_last_building_was_destroyed + g2.p1.time_my_last_building_was_destroyed;    

    r = s_1 - s_2;
    if      (r > 0) sc1 = 1.0;
    else if (r < 0) sc1 = 0.0;
    else            sc1 = 0.5; // tie
  }
  
  sc2 = 1.0 - sc1;
  return false;
}


// 2 two-player game results -> scores for player 1 (in first game) and player 2

/*
  If player A achieves a wipe-out in one game and player B does not,
  player A wins

  If both players achieve a wipe-out and one player was faster, this player wins.

  Otherwise:

    s =   current mineral count / 2
        + construction cost of all standing buildings and units
        + construction cost of all opponent buildings and units that have
          been destroyed

    r = s(player 1) - s(player 2)

    If r > 0, player 1 wins.
    If r < 0, player 2 wins.
    If r = 0: game is a tie
*/

#define TEST3 0
#if TEST3
#define TEST(x) cout << x << endl;
#else
#define TEST(x)
#endif

bool score3(const std::string &res1, const std::string &res2, float &sc1, float &sc2)
{
  if (res2.empty()) { cerr << "dummy empty in score3" << endl; exit(20); }
  
  GameResult2P g1, g2;
  sc1 = sc2 = -1;

#if !TEST3
  try {
    g1.parse(res1);
    g2.parse(res2);
  }
  catch (...) {
    return true;
  }
#endif
  
  // wipeout: # lost buildings > 0 &&
  //          # lost buildings = 1 + #built buildings

  // # times player i was wiped out (player 1 vs player 2 match)

  int g1wo1 = g1.p1.buildings_lost > 0 && g1.p1.buildings_lost==1+g1.p1.buildings_built;
  int g2wo1 = g2.p2.buildings_lost > 0 && g2.p2.buildings_lost==1+g2.p2.buildings_built;
  int g1wo2 = g1.p2.buildings_lost > 0 && g1.p2.buildings_lost==1+g1.p2.buildings_built;
  int g2wo2 = g2.p1.buildings_lost > 0 && g2.p1.buildings_lost==1+g2.p1.buildings_built;

#if TEST3
  // read test data from stdin
  cin >> g1.time >> g2.time;
  cin >> g1wo1 >> g2wo1 >> g1wo2 >> g2wo2;
  cin >> g1.p1.mineral_count >> g1.p1.standing_cost >> g1.p1.destroyed_cost
      >> g1.p2.mineral_count >> g1.p2.standing_cost >> g1.p2.destroyed_cost
      >> g2.p1.mineral_count >> g2.p1.standing_cost >> g2.p1.destroyed_cost
      >> g2.p2.mineral_count >> g2.p2.standing_cost >> g2.p2.destroyed_cost;
#endif
  
  if (g1wo1+g2wo1 != g1wo2+g2wo2) {

    // one player wiped out less often -> wins
    
    if (g1wo1+g2wo1 < g1wo2+g2wo2) { sc1 = 1.0; TEST("A"); }
    else                           { sc1 = 0.0; TEST("B"); }

  } else if (g1wo1+g2wo1 == 2) {

    // 2+2 wipeouts: draw
    sc1 = 0.5; TEST("C"); 

  } else {

    if (g1wo1+g2wo1 == 1) {

      // one wipeout each

      // if in different games, time is tiebreaker

      if (g1wo1 && g2wo2) {

	int d = g1.time - g2.time;
	if      (d > 0) sc1 = 1.0; // player 1 wiped out later
	else if (d < 0) sc1 = 0.0;
	else            sc1 = 0.5;
	TEST("D"); 	
	sc2 = 1.0 - sc1;
	return false;
	
      } else if (g1wo2 && g2wo1) {

	int d = g2.time - g1.time;
	if      (d > 0) sc1 = 1.0; // player 1 wiped out later
	else if (d < 0) sc1 = 0.0;
	else            sc1 = 0.5;
	TEST("E");	
	sc2 = 1.0 - sc1;
	return false;
	
      } else {

	// simultaneous wipeout in one game
	// -> clear values for that game and apply tiebreaker

	if (g1wo1) {

	  assert(g1wo2);
	  g1.p1.mineral_count = g1.p1.standing_cost = g1.p1.destroyed_cost = 0;
	  g1.p2.mineral_count = g1.p2.standing_cost = g1.p2.destroyed_cost = 0;	  
	  TEST("F");
	
	} else if (g2wo1) {

	  assert(g2wo2);
	  g2.p1.mineral_count = g2.p1.standing_cost = g2.p1.destroyed_cost = 0;
	  g2.p2.mineral_count = g2.p2.standing_cost = g2.p2.destroyed_cost = 0;
	  TEST("G"); 		  
	}
      }
    }
    
    // tiebreaker
    // double scores to avoid fractions
    
    int s_1 =
      g1.p1.mineral_count + g2.p2.mineral_count +
      2*(g1.p1.standing_cost  + g2.p2.standing_cost +
	 g1.p1.destroyed_cost + g2.p2.destroyed_cost);
    
    int s_2 = 
      g1.p2.mineral_count + g2.p1.mineral_count +
      2*(g1.p2.standing_cost   + g2.p1.standing_cost +
	 g1.p2.destroyed_cost  + g2.p1.destroyed_cost);

    int r = s_1 - s_2;
    if      (r > 0) sc1 = 1.0;
    else if (r < 0) sc1 = 0.0;
    else            sc1 = 0.5; // tie
    TEST("H");    
  }

  sc2 = 1.0 - sc1;
  return false;
}


extern bool score4(const std::string &res1, const std::string &dummy, float &sc1, float &sc2)
{
  if (!dummy.empty()) { cerr << "dummy not empty in score4" << endl; exit(20); }
  
  GameResult2P g1;
  sc1 = sc2 = -1;

  try { g1.parse(res1); }
  catch (...) { return true; }

  int d12 = g1.p2.units_lost - g1.p1.units_lost;
  
  if      (d12 >  0) sc1 = 1.0;
  else if (d12 == 0) sc1 = 0.5;
  else               sc1 = 0.0;

  sc2 = 1.0 - sc1;
  return false;
}


void score1_test()
{
  cout << "score1 tests" << endl;  
  float dummy;
  
  {
    string r = "@ TIME 480 p0 m 560 s 0 d 0 ub 0 ul 0 bb 0 bl 0 t 0 | ";
    float sc;
    int ret = score1(r, "", sc, dummy);
    cout << "ret= " << ret << " s " << sc << " ";
    cout << "expected 560" << endl;
  }

  {
    string r = "@ TIME 480 p0 m 9 s 0 d 0 ub 0 ul 0 bb 0 bl 0 t 0 | ";
    float sc;
    int ret = score1(r, "", sc, dummy);
    cout << "ret= " << ret << " s " << sc << " ";
    cout << "expected 9" << endl;
  }

  {
    // syntax error
    string r = "@ TIME 480 p0 m 9 s 0 d 0uc= 0 ul 0 bb 0 bl 0 t 0 | ";
    float sc;
    int ret = score1(r, "", sc, dummy);
    cout << "ret= " << ret << " ";
    cout << "expected syntax error" << endl;
  }

};


void score2_test()
{
  cout << "score2 tests" << endl;

  {
    // equal strings
    string r1 = "@ EARLY 5660 p0 m 0 s -10005 d 9001 ub 0 ul 50 bb 0 bl 5 t 5660 | p1 m 0 s -9001 d 10005 ub 0 ul 45 bb 0 bl 1 t 260 |";
    string r2 = "@ EARLY 1501 p0 m 0 s -10005 d 6401 ub 0 ul 50 bb 0 bl 5 t 1500 | p1 m 0 s -6401 d 10005 ub 0 ul 32 bb 0 bl 1 t 337 |";
    float sc1, sc2;
    
    int ret = score2(r1, r2, sc1, sc2);
    cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << " ";
    cout << "expected 1.0" << endl;
  }
  
  {
    // equal strings
    string r1 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |";
    string r2 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |";  
    float sc1, sc2;
    
    int ret = score2(r1, r2, sc1, sc2);
    cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << " ";
    cout << "expected 0.5" << endl;
  }

  {
    // equal strings but player 1 faster -> 1.0
    string r1 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2643 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |";
    string r2 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |";  
    float sc1, sc2;
    
    int ret = score2(r1, r2, sc1, sc2);
    cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << " expected 1.0" << endl;    
  }

  {
    // equal strings but player 2 faster -> 0.0
    string r1 = "@ TIME 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 100 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 1000 |";
    string r2 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 100 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2000 |";  
    float sc1, sc2;
    
    int ret = score2(r1, r2, sc1, sc2);
    cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << " expected 0.0" << endl;    
  }

  {
    // p1 more buildings
    string r1 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 5 t 2533 |";
    string r2 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |";  
    float sc1, sc2;
    
    int ret = score2(r1, r2, sc1, sc2);
    cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << " ";
    cout << "expected 1.0" << endl;
  }

  {
    // p2 more buildings
    string r1 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 3 t 2533 |";
    string r2 = "@ EARLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |";  
    float sc1, sc2;
    
    int ret = score2(r1, r2, sc1, sc2);
    cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << " ";
    cout << "expected 0.0" << endl;
  }

  {
    // syntax errors
    string r1 = "@ RLY 2843 p0 m 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 3 t 2533 |";
    string r2 = "@ EARLY 2843 p0 c= 0 s -8405 d 6404 ub 0 ul 42 bb 0 bl 5 t 2843 | p1 m 0 s -6404 d 8405 ub 0 ul 32 bb 0 bl 4 t 2533 |";  
    float sc1, sc2;
    
    int ret = score2(r1, r2, sc1, sc2);
    cout << "ret= " << ret << " ";
    cout << "expected syntax error" << endl;
  }

}


void score3_test()
{
  // use test3 data (copy and paste);  TEST3 must be 1 for this to work.

#if TEST3

  cout << "score3 tests" << endl;
  while (1) {
    string r1, r2;
    float sc1, sc2;

    int ret = score3(r1, r2, sc1, sc2);
    cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << endl;    
  }

#endif  
}


void score4_test()
{
  cout << "score4 tests" << endl;
  float sc1, sc2;
  string res;
  int ret;
  
  res =
  "@ TIME 484 p0 m 0 s -2500 d 2250 ub 0 ul 50 bb 0 bl 0 t 0 | p1 m 0 s -2250 d 2500 ub 0 ul 45 bb 0 bl 0 t 0 |";
  ret = score4(res, "", sc1, sc2);
  cout << "res= " << res << endl;
  cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << endl;
  if (sc1 != 0.0 || sc2 != 1.0 - sc1) cerr << "error!" << endl;
  
  res =
  "@ EARLY 484 p0 m 0 s -2500 d 2250 ub 0 ul 23 bb 0 bl 0 t 0 | p1 m 0 s -2250 d 2500 ub 0 ul 50 bb 0 bl 0 t 0 |";
  ret = score4(res, "", sc1, sc2);
  cout << "res= " << res << endl;
  cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << endl;
  if (sc1 != 1.0 || sc2 != 1.0 - sc1) cerr << "error!" << endl;

  res =
  "@ EARLY 484 p0 m 0 s -2500 d 2250 ub 0 ul 49 bb 0 bl 0 t 0 | p1 m 0 s -2250 d 2500 ub 0 ul 49 bb 0 bl 0 t 0 |";
  ret = score4(res, "", sc1, sc2);
  cout << "res= " << res << endl;
  cout << "ret= " << ret << " sc1= " << sc1 << " sc2= " << sc2 << endl;
  if (sc1 != 0.5 || sc2 != 1.0 - sc1) cerr << "error!" << endl;  
}
