// $Id: TRTerrain.C 5727 2007-08-21 20:56:43Z ckai $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

#include "Base.h"
#include "TRTerrain.H"
#include "Global.H"
#include "Options.H"

using namespace std;

//DJD: new defines and declarations of static variables in TraFace {

//the epsilon to use for the DCDT
#define EPSILON 0.00001f
//the number of sections of a circle for the arcs around vertices
#define SECTIONS 20
//a float definition for pi
#define PI 3.1415926535897932384626433832795f
//a difference of less than this much influence doesn't matter
#define NO_INFLUENCE 0.01f
//number of sides of a polygon to put in for a circular object
#define SIDES 8
//fraction of a circle to move when trying a spot adjacent to an obstacle
#define TRIES 10

//Kai: remove the static members (processing and current) due to multiple definitions in the new pathfinding interface.
//SrArray<TraFace *> TraFace::processing;
//int TraFace::current = 0;

//new defines and declarations of static variables in TraFace }

namespace TRTerrain {

//DJD: new constructor logic {

//constructor - creates the DCDT but doesn't initialize it
TR_Terrain::TR_Terrain()
{
	pfDcdt = new TraDcdt;
}

//new constructor logic }

  //-----------------------------------------------------------------------------
  void TR_Terrain::add_options()
  {
    ::Options o("TRTerrain");
//    o.put("-gran",  sint4(4),  "granularity of the fine grid default 4x4");
//    o.put("-costonly", "represent boundaries only by higher edge cost");
  }

  //-----------------------------------------------------------------------------
  void TR_Terrain::init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_)
  {
    TerrainBasicImp<TR_Task>::init(tiles_x_, tiles_y_, tile_points_, client_pID_, neutral_pID_);
//    sint4 gran;
//    Options::get("-gran", gran);
//    pfEngine = boost::shared_ptr<PFEngine>(new PFEngine(tiles_x_, tiles_y_, tile_points_, gran));

//DJD: initialize triangulator with the boundary of the environment {
	float width = (float)(tiles_x_ * tile_points_);
	float height = (float)(tiles_y_ * tile_points_);
	SrPolygon boundary;
	boundary.push().set(0.0f, 0.0f);
	boundary.push().set(width, 0.0f);
	boundary.push().set(width, height);
	boundary.push().set(0.0f, height);
	pfDcdt->init(boundary, EPSILON, 1);
//initialize triangulator with the boundary of the environment }
  }

//DJD: function definitions for updating the terrain for obstacles {

//adds an obstacle to the environment representation
void TR_Terrain::AddObstacle(GameObj *gob)
{
	//create a closed polygon for the obstacle
	SrPolygon poly;
	poly.open(false);
	//check if the obstacle is rectangular
	if (gob->get_int("shape") == Object::RECTANGLE)
	{
		//get its corners
		sint4 x1, y1, x2, y2;
		gob->get_p1(x1, y1);
		gob->get_p2(x2, y2);
		//create a polygon using them
		poly.push().set((float)x1, (float)y1);
		poly.push().set((float)x2, (float)y1);
		poly.push().set((float)x2, (float)y2);
		poly.push().set((float)x1, (float)y2);
	}
	//check if the obstacle is circular
	else if (gob->get_int("shape") == Object::CIRCLE)
	{
		//get its center and radius
		sint4 cx, cy;
		gob->get_center(cx, cy);
		sint4 r = gob->get_radius();
		float x, y;
		float angle = 0.0f;
		float inc = 2.0f * PI / (float)SIDES;
		//create a certain number of vertices to approximate the circle
		for (int i = 0; i < SIDES; i++)
		{
			//calculate the position of the vertex
			x = (float)cx + cos(angle) * (float)r;
			y = (float)cy + sin(angle) * (float)r;
			//add it to the polygon
			poly.push().set(x, y);
			//and increment the angle to the next vertex
			angle += inc;
		}
	}
	else
	{
		//we only deal with circular and rectangular obstacles
		return;
	}
	//get the obstacle's ID within ORTS
	ObjId ortsId = GameObj::get_obj_id(gob);
	//insert the obstacle's polygon into the triangulation and record its ID there
	int dcdtId = pfDcdt->insert_polygon(poly);
	//add this to our list of obstacles
	obstacles.push_back(TR_Obstacle(ortsId, dcdtId));
}

//removes an obstacle from the environment representation
void TR_Terrain::RemoveObstacle(GameObj *gob)
{
	//get the ID of the obstacle within ORTS
	ObjId ortsId = GameObj::get_obj_id(gob);
	//go through our list of obstacles
	for (std::vector<TR_Obstacle>::iterator i = obstacles.begin(); i != obstacles.end(); i++)
	{
		//when we find one with a matching ID,
		if (ortsId == i->ortsId)
		{
			//if the object was successfully added to the triangulation,
			if (i->dcdtId != -1)
			{
				//remove it
				pfDcdt->remove_polygon(i->dcdtId);
			}
			//remove it from the list of obstacles
			obstacles.erase(i);
			break;
		}
	}
}

//updates the abstraction based on the changed triangulation
void TR_Terrain::UpdateAbstraction()
{
	//delete the abstraction
	pfDcdt->DeleteAbstraction();
	//rebuild it
	pfDcdt->Abstract();
	//repair the sectors that have been invalidated
	pfDcdt->InitializeSectors();
	//and record that the new faces have been processed
	TraFace::Clear();
}

//function definitions for updating the terrain for obstacles }

  //-----------------------------------------------------------------------------
  void TR_Terrain::add_obj(const Object *obj)
  {
    TerrainBasicImp<TR_Task>::add_obj(obj);
//    assert(pfEngine!=NULL);
//    pfEngine->insert_object(obj);

//DJD: add logic for keeping track of obstacles in the triangulation {
	GameObj *gob = (GameObj *)obj;
	//if the object is an obstacle,
	if (TR_Obstacle::IsObstacle(gob))
	{
		//add it to the representation
		AddObstacle(gob);
		UpdateAbstraction();
		//and don't consider it for influence
		return;
	}
//add logic for keeping track of obstacles in the triangulation }

//DJD: add logic for keeping track of enemy unit influence in the triangulation {
	//only processes enemy units
	const sintptr *owner = gob->get_int_ptr("owner");
	if ((*owner == client_pID) || (*owner == neutral_pID))
	{
		return;
	}
	//locates the triangle in which the object's center lies
	TraBase *s = NULL;
	sint4 x, y;
	obj->get_center(x, y);
	//sets the object's triangle to null if not found
	if (pfDcdt->LocatePoint(x, y, s) == SeTriangulator::NotFound)
	{
		objects.push_back(TR_Object(gob));
	}
	else
	{
		//otherwise sets it to the triangle face found
		TraFace *face = (TraFace *)s->fac();
		TR_Object trobj(gob, face);
		objects.push_back(trobj);
		//and updates the unit's influence in the triangulation
		pfDcdt->UpdateInfluence(face, trobj.influence);
	}
//add logic for keeping track of enemy unit influence in the triangulation }

  }

  //-----------------------------------------------------------------------------
  void TR_Terrain::update_obj(const Object *obj)
  {
    TerrainBasicImp<TR_Task>::update_obj(obj);
//    assert(pfEngine!=NULL);
//    pfEngine->update_object(obj);

//DJD: add logic for updating positions of obstacles in the triangulation {
	GameObj *gob = (GameObj *)obj;
	//if the object is an obstacle,
	if (TR_Obstacle::IsObstacle(gob))
	{
		//update it within the representation
		RemoveObstacle(gob);
		AddObstacle(gob);
		UpdateAbstraction();
		//and don't consider it for influence
		return;
	}	
//add logic for updating positions of obstacles in the triangulation }

//DJD: add logic for updating enemy position or influence in the triangulation {
	//only processes enemy units
	const sintptr *owner = gob->get_int_ptr("owner");
	if ((*owner == client_pID) || (*owner == neutral_pID))
	{
		return;
	}
	ObjId id = GameObj::get_obj_id(gob);
	//goes through the objects
	for (uint4 i = 0; i < objects.size(); i++)
	{
		//until it finds the desired object
		if (objects[i].id == id)
		{
			//record the object's old triangle and influence
			TraFace *oldFace = objects[i].face;
			float oldInfluence = objects[i].influence;
			//calculate the object's new influence
			objects[i].CalculateInfluence((GameObj *)obj);
			//remove the object's influence from it last triangle
			if (objects[i].face != NULL)
			{
				pfDcdt->UpdateInfluence(objects[i].face, -(objects[i].influence));
			}
			TraBase *s = NULL;
			SeTriangulator::LocateResult result;
			sint4 x, y;
			obj->get_center(x, y);
			//if possible, uses the last triangle as the starting point
			//otherwise, uses the triangle around the nearest sector midpoint
			//then finds the triangle in which the object is currently centered
			if (objects[i].face == NULL)
			{
				result = pfDcdt->LocatePoint(x, y, s);
			}
			else
			{
				result = pfDcdt->LocatePoint(objects[i].face, x, y, s);
			}
			//sets the object's face to NULL if it wasn't found
			if (result == SeTriangulator::NotFound)
			{
				objects[i].face = NULL;
			}
			else
			{
				//otherwise, sets it appropriately
				objects[i].face = (TraFace *)s->fac();
			}
			//checks if the object has moved triangles or if it has changed influence
			if ((oldFace != objects[i].face) || (fabs(oldInfluence - objects[i].influence) > NO_INFLUENCE))
			{
				//if it has and the old triangle was valid, subtract its influence from the old triangle
				if (oldFace != NULL)
				{
					pfDcdt->UpdateInfluence(oldFace, -oldInfluence);
				}
				//if the new triangle is valid, add the object's influence to the new triangle
				if (objects[i].face != NULL)
				{
					//and updates the object's new influence on the map
					pfDcdt->UpdateInfluence(objects[i].face, objects[i].influence);
				}
			}
			break;
		}
	}
//add logic for updating enemy position or influence in the triangulation }

  }

  //-----------------------------------------------------------------------------
  void TR_Terrain::remove_obj(const Object *obj)
  {
    TerrainBasicImp<TR_Task>::remove_obj(obj);
//    assert(pfEngine!=NULL);
//    pfEngine->remove_object(obj);
    objsToStop.erase(obj);

//DJD: add logic for removing dead or vanished obstacles from the triangulation {
	GameObj *gob = (GameObj *)obj;
	//if the object is an obstacle,
	if (TR_Obstacle::IsObstacle(gob))
	{
		//remove it from the representation
		RemoveObstacle(gob);
		UpdateAbstraction();
		//and don't consider it for influence
		return;
	}	
//add logic for removing dead or vanished obstacles from the triangulation }

//DJD: add logic for removing influence of vanished or dead enemy units {
	//only processes enemy units
	const sintptr *owner = gob->get_int_ptr("owner");
	if ((*owner == client_pID) || (*owner == neutral_pID))
	{
		return;
	}
	ObjId id = GameObj::get_obj_id(gob);
	//goes through the list of objects
	for (std::vector<TR_Object>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		//upon finding the matching object
		if (i->id == id)
		{
			pfDcdt->UpdateInfluence(i->face, -(i->influence));
			//removes it from the list
			objects.erase(i);
			break;
		}
	}
//add logic for removing influence of vanished or dead enemy units }

  }

  //-----------------------------------------------------------------------------
  void TR_Terrain::add_segments(const Vector<Segment> &segs)
  {
    TerrainBasicImp<TR_Task>::add_segments(segs);

//DJD: combine new segments with existing ones and update the triangulation {
	sm.clear_delta();
	// send all segments to merge class
	FORALL (segs, s)
	{
		DCDT::Segment s1(s->l1.x, s->l1.y, s->l2.x, s->l2.y);
		sm.new_segment(s1);
	}
	// process all changes
	bool changed = false;
	sint4 add_n = 0, del_n = 0;
	FORS (i, 4)
	{
		del_n += sm.del[i].size();
		FORALL (sm.del[i], it)
		{
			pfDcdt->remove_polygon(*it);
			changed = true;
		}
	}
	sint4 total = 0;
	FORS (i, 4)
	{
		total += sm.all[i].size();
		add_n += sm.add[i].size();
		FORALL (sm.add[i], it)
		{
			const SortSegment &ss = **it;
			SrPolygon pol;
			pol.push().set(ss.s.x1, ss.s.y1);
			pol.push().set(ss.s.x2, ss.s.y2);
			ss.label = pfDcdt->insert_polygon(pol);
			changed = true;
		}
	}
	//if the triangulation was changed, update the abstraction
	if (changed)
	{
		UpdateAbstraction();
	}
//combine new segments with existing ones and update the triangulation }

    invalidate_paths(segs);
  }

  //----------------------------------------------------------------------------------
  void TR_Terrain::cancel_task(const Object *obj)
  {
    TerrainBasicImp<TR_Task>::cancel_task(obj);
    objsToStop.insert(obj);
  }

  //----------------------------------------------------------------------------------
  void TR_Terrain::remove_task(TaskId tid)
  {
    FIND (taskId2task, it, tid);
    if( it != taskId2task.end() ) {
      Task& task = it->second->task;
      objsToStop.insert(task.objs.begin(), task.objs.end());
    }
    TerrainBasicImp<TR_Task>::remove_task(tid);
  }

  //----------------------------------------------------------------------------------
  real8 TR_Terrain::find_path(const Object *obj, const Goal &goal, Path *path, ConsiderObjects consider)
  {
//DJD: find a path from an object to some goal {
	//gets the current time
	Uint32 startTime = SDL_GetTicks();
	//gets the coordinates of the unit
	Loc start;
	obj->get_center(start.x, start.y);
	//if the pathfinding target is an area,
	if (goal.target == Goal::LOCATION)
	{
		//find a path there
		return find_path(start, goal.loc, obj->get_radius(), path, consider);
	}
	else
	{
		//if it is an object, determine its position
		Loc dest;
		GameObj *gob = (GameObj *)goal.obj;
		//if the object is not an obstacle, we can go to its center
		if (!TR_Obstacle::IsObstacle(gob))
		{
			if (gob->get_int("shape") == Object::RECTANGLE)
			{
				//if the destination object is rectangular,
				sint4 x1, y1, x2, y2;
				gob->get_p1(x1, y1);
				gob->get_p2(x2, y2);
				//calculate it center and go there
				dest.x = x1 + x2 / 2;
				dest.y = y1 + y2 / 2;
			}
			else if (gob->get_int("shape") == Object::CIRCLE)
			{
				//if the destination object is circular, go to its center
				gob->get_center(dest.x, dest.y);
			}
			else
			{
				//we're only dealing with rectangles and circles for now
				return -1;
			}
			//find a path there
			return find_path(start, dest, obj->get_radius(), path, consider);
		}
		//if it is an obstacle, we must set the destination to be outside it
		else
		{
			//calculate the center and "radius" of the obstacle
			sint4 x, y, r;
			if (gob->get_int("shape") == Object::RECTANGLE)
			{
				sint4 x1, y1, x2, y2;
				gob->get_p1(x1, y1);
				gob->get_p2(x2, y2);
				x = x2 - x1;
				y = y2 - y1;
				r = (sint4)(sqrt((float)(x * x + y * y)) / 2.0f) + 1 + obj->get_radius();
				x = x1 + x2 / 2;
				y = y1 + y2 / 2;
			}
			else if (gob->get_int("shape") == Object::CIRCLE)
			{
				gob->get_center(x, y);
				r = gob->get_radius() + obj->get_radius();
			}
			else
			{
				//we're only dealing with rectangles and circles for now
				return -1;
			}
			float base = atan2((float)start.y - (float)y, (float)start.x - (float)x);
			float diff = PI / (float)TRIES;
			for (int i = 0; i < TRIES; i++)
			{
				float theta = base + diff * (float)i;
				diff = -diff;
				dest.x = x + (sint4)(cos(theta) * (float)r);
				dest.y = y + (sint4)(sin(theta) * (float)r);
				float result = find_path(start, dest, obj->get_radius(), path, consider);
				if ((result != -1) || (SDL_GetTicks() >= startTime + 10))

				{
					return result;
				}
			}
			return -1;
		}
	}
//find a path from an object to some goal }
  }

  //----------------------------------------------------------------------------------
  real8 TR_Terrain::find_path(const Loc &l1, const Loc &l2, sint4 radius, Path *path, ConsiderObjects consider)
  {
//DJD: find a path from one location to another with given radius {
	//get the current time
	Uint32 time = SDL_GetTicks();
	//find the path with soft timeout of 2 milliseconds and hard timeout of 5 milliseconds
	if (!pfDcdt->SearchPathFast((float)l1.x + 0.5f, (float)l1.y + 0.5f, (float)l2.x + 0.5f, (float)l2.y + 0.5f,
		(float)radius, time + 200000, time + 500000))
	{
		return -1;
	}
	//get the path and record the first waypoint
	SrPolygon CurPath = pfDcdt->GetPath();
	path->locs.push_back(Loc((sint4)CurPath[0].x, (sint4)CurPath[0].y));
	//go through the straight segments of the path
	for (int i = 0; i < CurPath.size() - 1; i += 2)
	{
		//if another follows it, construct the curved segment in between
		if (i < CurPath.size() - 3)
		{
			//determine the angles of both adjoining segments
			float angle1 = atan2(CurPath[i + 1].y - CurPath[i].y, CurPath[i + 1].x - CurPath[i].x);
			float angle2 = atan2(CurPath[i + 3].y - CurPath[i + 2].y, CurPath[i + 3].x - CurPath[i + 2].x);
			//calculate the difference between the two
			float diff = angle1 - angle2;
			diff += (diff <= -PI) ? 2.0f * PI : (diff > PI) ? -2.0f * PI : 0.0f;
			//calculate the angle to the center of curvature
			float start = angle1 + ((diff < 0.0f) ? PI / -2.0f : PI / 2.0f);
			//followed by its coordinates
			float cornerX = CurPath[i + 1].x - cos(start) * radius;
			float cornerY = CurPath[i + 1].y - sin(start) * radius;
			//calculate the number of segments to draw based on the angle
			int number = (int)(fabs(diff) / (2.0f * PI) * (float)SECTIONS);
			float theta = -diff / (float)number;
			float r = radius / fabs(cos(theta * 0.5f));
			start += theta * 0.5f;
			//and go through the segments
			for (int j = 0; j < number; j++)
			{
				//calculate the next vertex in the line strip
				Loc l((sint4)(cornerX + cos(start) * r), (sint4)(cornerY + sin(start) * r));
				//only insert each vertex if it's different than the last
				if ((l.x != path->locs.back().x) && (l.y != path->locs.back().y))
				{
					path->locs.push_back(l);
				}
				//incrementing the angle each time
				start += theta;
			}
		}
	}
	//calculate and place the last vertex on the curve
	Loc l((sint4)CurPath[CurPath.size() - 1].x, (sint4)CurPath[CurPath.size() - 1].y);
	if ((l.x != path->locs.back().x) && (l.y != path->locs.back().y))
	{
		path->locs.push_back(l);
	}

 	//reverse order
  	for (uint4 i = 0; i <= (path->locs.size()-1)/2; i++)
	{
   		 Loc temp;
   		 temp = path->locs[i];
   		 path->locs[i] = path->locs[path->locs.size()-1-i];
    		path->locs[path->locs.size()-1-i] = temp;
  	}


	//return the length of the path found
	return pfDcdt->GetPathLength();
//find a path from one location to another with given radius }
  }

  //----------------------------------------------------------------------------------

  // fixme: we only check the path of the *center* of the unit vs the new boundaries,
  // and we should account for its size (radius).
  void TR_Terrain::invalidate_paths(const Vector<Segment> &segments)
  {
    FORALL(tasks, ti) {
      if( !ti->isActive() ) continue;
      // fixme: better handling of zcat != ON_LAND ?
      if( ti->path->obj->get_zcat() != Object::ON_LAND ) continue;

      FORALL(segments, si) {
        if( ti->path->is_intersect(*si) ) {
          // this path is invalid, mark as pending for (re)planning
          pendingTasks.push_back(ti);
          ti->isPending = true;
          cout << "Marked an invalid path" << endl;
          break;
        }
      }
    }

#if 0
    Vector<Segment> l;
    
    FORALL (tasks, i) {
      if( !i->isActive() ) continue;
      l = i->path->intersect(s);
      if (l.size() > 0) {
        UnitPath up = *i->path;
        FIND(objIt, objId2ObjMap, up.obj_id);
        if( objIt == objId2ObjMap.end() ) continue;

        // find-path
        Path newpath;
        pfEngine->find_path(objIt->second, i->path->goal, newpath);
        if( newpath.locs.size() > 0 ) {
          i->path->path.clear();
          i->path->path.insert(i->path->path.begin(),newpath.locs.begin(),newpath.locs.end());
          i->path->current_target = (sint4)newpath.locs.size()-1;
          i->path->tries = -1;
        } 
      }
    }
#endif
  }

  //----------------------------------------------------------------------------------
  void TR_Terrain::execute_tasks(Vector<TerrainBase::MoveCmd> &cmds, Vector<TerrainBase::StatusMsg> &msgs)
  {
    // remove dead paths
    // fixme: this should be removed after cleaning UnitPath
    for (typeof(tasks.begin()) i = tasks.begin(); i != tasks.end(); ) {
      if (i->path && i->path->dead) {
        i = tasks.erase(i);
      } else {
        ++i;
      }
    }

    // Execute all active tasks
    for (Tasks::iterator j = tasks.begin(); j != tasks.end();) 
    {
      Tasks::iterator i = j++;

      if( !i->isActive() ) continue;

      UnitPath &cur_path = *(i->path);

      if (cur_path.sleeping()) continue;

      const Object *obj = cur_path.obj;
      if (obj->get_speed() > 0 || obj->get_moving() != 0) continue;

      Loc p; obj->get_center(p.x, p.y);
      Loc cur_goal = cur_path.path.at(cur_path.current_target);

      if (p == cur_goal || cur_path.tries == -1) {

        // reached way-point || just getting started

        //      cout << "active = " << cur_path.active << endl;
        cur_path.tries = 0;

        if (cur_path.current_target == 0) { // path is in reverse order (goal first)
          // reached the goal

          TerrainBase::StatusMsg s;
          // fixme: this is not robust to future changes in the meaning of TaskId.
          s.task_id = (TerrainBase::TaskId)&(*i);
          s.obj = obj; 
          s.type = TerrainBase::StatusMsg::ARRIVED;
          msgs.push_back(s);
        } else {
          // move to next waypoint
          if (p == cur_goal)
            cur_path.current_target--;

          TerrainBase::MoveCmd m;
          m.obj = obj;
          m.next_loc = cur_path.path.at(cur_path.current_target);
          m.speed = obj->get_max_speed();
          cmds.push_back(m);
        }
        continue;
      }

      // unit is not moving
      // not at waypoint/goal

      if (cur_path.spot != p) {
        //      cur_path.tries = 0;
        cur_path.mark_spot(p);
      } else {
        cur_path.tries++;
      }

      if (cur_path.tries > 1) {
        // unexpected collition... try to repair

        cout << "collision " << cur_path.tries << endl;

        if (cur_path.tries <  3) {
          int delay = 500 + (p.x-cur_goal.x)+ (p.y-cur_goal.y) % 500;
          cur_path.sleep(delay);
          cur_path.tries++;
          continue;
        }

        cout << "failure" << endl;
/*
        SActParam param;
        param.obj.push_back(cur_path.unit);

        const char *c = "WOW!";
        param.val.push_back(stype(c));

        ObjAction *d = cur_path.unit->action("say");
        // if (d) d->Act(param);
        param.val.clear();
*/
        if (cur_path.tries > 5) {
          cout << "giving up" << endl;
//          c = "Giving up!";
          TerrainBase::StatusMsg s;
          // fixme: this is not robust to future changes in the meaning of TaskId.
          s.task_id = (TerrainBase::TaskId)&(*i);
          s.obj = obj;
          s.type = TerrainBase::StatusMsg::MOVEMENT_FAILURE;
          msgs.push_back(s);

          cur_path.kill();
          continue;
        }

        if (cur_path.tries >= 3 && cur_path.tries <= 5) {
          i->isFailed = true;
          //c = "Repathing...";
        }
/*
        param.val.push_back(stype(c));
        d = cur_path.unit->action("say");
        // if (d) d->Act(param);
*/
      }  
    }

    // Stop objects whose task was cancelled, and that did not yet get a new
    // move command
    FORALL(cmds, c) {
      objsToStop.erase(c->obj);
    }
    FORALL(objsToStop, oi) {
      TerrainBase::MoveCmd m;
      m.obj = *oi;
      (*oi)->get_center(m.next_loc.x, m.next_loc.y); // doesn't really matter
      m.speed = 0; // 0 -> stop
      cmds.push_back(m);
    }

    // remove finished tasks
    FORALL(msgs, t) {
      clean_task(t->task_id);
    }
  }

  //----------------------------------------------------------------------------------
bool TR_Terrain::plan_tasks(uint4 /*max_time*/)
  {
    // plan a failed path
    bool did_something = plan_failed_task();

    // plan a pending (new) task
    if( !did_something )
      did_something = plan_pending_task();

    return did_something;
  }
/*
  //----------------------------------------------------------------------------------
  std::deque<TIter>::iterator TR_Terrain::get_best_task(void)
  {    
    // Check if the oldest task must be planned now
    std::deque<TIter>::iterator oldest = max_element(pendingTasks.begin(), 
                                                     pendingTasks.end(),
                                                     is_younger);
    if( SDL_GetTicks() - (*oldest)->start_time >= MAXIMUM_TASK_AGE )
      return oldest;

    // else, return the 'shortest' (straight-line) path
    return min_element(pendingTasks.begin(), pendingTasks.end(), is_shorter);
  }
*/
  //----------------------------------------------------------------------------------
  bool TR_Terrain::plan_pending_task(void)
  {
/*
    if (pendingTasks.empty()) return false;

    // TIter tit = pendingTasks.front(); pendingTasks.pop_front();

    std::deque<TIter>::iterator temp = get_best_task();
    TIter tit = *temp;
    pendingTasks.erase(temp);

    // mark the task as handled
    tit->isPending = false;

    // if the object is moving, stop it first and come back to it later
    assert(tit->task.objs.size() > 0);
    const Object *obj = *(tit->task.objs.begin());

    if (obj->get_speed() > 0 || obj->get_moving()) {
      objsToStop.insert(obj);
      pendingTasks.push_back(tit);
      return false;
      // fixme: might be better to try another pending task already in this tick...
    }

    // plan
    TEST(cout << obj << " planning pending path" << endl);
    plan_task(tit);
    return true;
*/
    if( pendingTasks.empty() )
      return false;

    TIter tit = pendingTasks.front(); pendingTasks.pop_front();
    // mark the task as handled
    tit->isPending = false;

    // object to move
    assert(tit->task.objs.size() > 0);
    const Object *obj = *(tit->task.objs.begin());

    // if the object is moving, stop it first and come back to it later
    if( obj->get_speed() > 0 || obj->get_moving() ) {
      objsToStop.insert(obj);
      pendingTasks.push_back(tit);
      return false;
      // fixme: might be better to try another pending task already in this tick...
    }

    // target location
    Loc goal;
    if( tit->task.goal.target == Goal::LOCATION )
      goal = tit->task.goal.loc;
    else {
      tit->task.goal.obj->get_center(goal.x,goal.y);
    }

    // plan
    Path newpath;
//    pfEngine->find_path(obj, goal, newpath);
	Loc start;
	obj->get_center(start.x, start.y);
	find_path(start, goal, obj->get_radius(), &newpath, TerrainBase::IGNORE_ALL_MOBILE_OBJS);
    
    // insert results into data structures
    if (newpath.locs.size() > 0) {
      tit->init_path(obj, newpath.locs);
    }
    else {
      // fixme: maybe add to "failed" ?

//      char* c = "cannot find a path!";
//      SActParam p;
//      p.obj.push_back(obj);
//      p.val.push_back((stype)c);
//      ObjAction *d = obj->action("say");
//      if (d) d->Act(p);
    }
    return true;
  }

  bool TR_Terrain::plan_failed_task(void)
  {
//    assert(pfEngine!=NULL);

    // find a failed path
    FORALL_B(tasks,tit) {
      if( tit->isFailed )
        break;
    }
    // if none exists
    if( tit == tasks.end() )
      return false;

    // object to move
    assert(tit->task.objs.size() > 0);
    const Object *obj = *(tit->task.objs.begin());

    // if the object is moving, stop it first and come back to it later
    if( obj->get_speed() > 0 || obj->get_moving() ) {
      objsToStop.insert(obj);
      return false;
      // fixme: might be better to try another failed task already in this tick...
    }

    // find-path
    Path newpath;
    cout << "retry " << tit->path->tries << endl;
//    pfEngine->find_path(obj, tit->path->goal, newpath);
	Loc start;
	obj->get_center(start.x, start.y);
	find_path(start, tit->path->goal, obj->get_radius(), &newpath, TerrainBase::IGNORE_ALL_MOBILE_OBJS);
    if( newpath.locs.size() > 0 ) {
      tit->path->path.clear();
      tit->path->path.insert(tit->path->path.begin(),newpath.locs.begin(),newpath.locs.end());
      tit->path->current_target = (sint4)newpath.locs.size()-1;
      tit->path->tries = -1;
      tit->isFailed = false;
    } /*else {
      SActParam p;
      p.obj.push_back(i->unit);
      char* c = "repair failed :(";
      p.val.push_back((stype)c);
      ObjAction *d = i->unit->action("say");
      if (d) d->Act(p);
    }*/
    return true;
  }


//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

#define EPS 1e-9f


  UnitPath::UnitPath(const Object *obj_, const Vector<Loc> &path_)
    : obj(obj_), path(path_)
  {
    tries = -1;
    current_target = path.size()-1;
    dead = false;
    if (path.size() > 0) {
      goal = path[0];
    }
    spot = Loc(-1, -1);
    wakeup = 0;
  }


  bool UnitPath::is_intersect(const Segment &segment) const
  {
    typeof(path.begin()) pj = path.begin();
    typeof(path.begin()) pi = pj++;
    for( ; pj != path.end() ; pi++, pj++ ) {
      if( Segment(*pi, *pj).touches(segment) )
        return true;
    }
    return false;
  }

#if 0
  Vector<Segment> UnitPath::intersect(const Segment &segment)
  {
    Vector<Segment> lines;
    const Loc& a = segment.l1;
    const Loc& b = segment.l2;
  
    for (uint4 i = 1; i < path.size(); i++) {

      Loc c = path[i-1];
      Loc d = path[i];
    
      real8 m1 = (real8)(c.y-d.y) / (real8)(c.x-d.x + EPS);
      real8 m2 = (real8)(a.y-b.y) / (real8)(a.x-b.x + EPS);
    
      if ((int)m1 == (int)m2) continue; // fixme: equality for reals doesn't work
    
      real8 b1 = c.y - m1 * c.x;
      real8 b2 = a.y - m2 * a.x;        
    
      real8 X = (b2-b1) / (m1-m2);
      real8 Y = m1*X + b1;
    
      real8 t1x = (real8)(X - a.x) / ((real8)(b.x - a.x + EPS));
      real8 t1y = (real8)(Y - a.y) / ((real8)(b.y - a.y + EPS));
      real8 t2x = (real8)(X - c.x) / ((real8)(d.x - c.x + EPS));
      real8 t2y = (real8)(Y - c.y) / ((real8)(d.y - c.y + EPS));
    
      bool overlap1 = (t1x < 0 || t1x > 1);
      if (!overlap1) continue;
    
      bool overlap2 = (t1y < 0 || t1y > 1);
      if (!overlap2) continue;    
    
      bool overlap3 = (t2x < 0 || t2x > 1);
      if (!overlap3) continue;        

      bool overlap4 = (t2y < 0 || t2y > 1);
      if (!overlap4) continue;    

      lines.push_back(Segment(path[i-1], path[i]));
    }
  
    return lines;
  }
#endif

}   // End of namespace SimpleTerrain
