#ifndef BASEDEFENSE_H
#define BASEDEFENSE_H

// $Id: BaseDefense.H 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file (c) Michael Buro, licensed under the GPL
// Adapted from GatherAI class

#include "Global.H"
#include "EventHandler.H"
#include "ClientModule.H"
#include "GameObj.H"

class GameStateModule;

//===================================================================



//-------------------------------------------------------------------

/**
 * This AI module causes all units to defend the nearest base. If a base is destroyed, it's
 * units will defend the next nearest base. Defending units attack any units within a certain 
 * radius of the base, defined by a command line argument.
 */
class BaseDefense : public EventHandler, public ClientModule
{
protected:

  /** Class that holds data that each defender needs to know about
   */
  class DefenderData {
  public:
    GameObj *base;
    enum State {MOVE_TO, WAIT, DEFEND};
    State state;
  };
  
  /** Class that holds data that each base needs to know about */
  class BaseData{
  public:
    std::set<GameObj*> defenders;
    GameObj *threat; ///< Unit that is within the radius of the base
  };

  std::map<GameObj*, BaseData> base2data;
  std::map<GameObj*, DefenderData> defender2data; 
  std::set<GameObj*> enemies;
  GameStateModule &gsm;
  Vector<GameObj*> need_act; ///< GameObjects needing an action in the next tick
  sint4 clean_zone;
  
  /** Updates the threats for each base. If the current threat is no longer a threat
   * (out of threat radius or dead), then tries to find a new threat */ 
  void update_targets();

  /** Returns the nearest enemy object to the given base, if that enemy is within the
   * threat radius */
  GameObj* find_threat(GameObj *base);

  /**
   * Finds the nearest base to travel to, based on geometric distance. baseToAvoid is legacy
   * from SuicideAttackAI.
   */
  GameObj* find_base(GameObj *targeter, GameObj *baseToAvoid = 0);
public:

  static const sint4 FROM;
  
  /** Adds the threat radius information to the command line options*/
  static void add_options(void);

  BaseDefense(GameStateModule &gsm);
  ~BaseDefense();

  /**
   * Causes the units in the need_act vector to act according to
   * their current state.
   */
  void act(GameStateModule &gsm);
  
  /**
   * Adds an friendly base to the list (actually a map) of friendly bases.
   */
  void add_base(GameObj *gob);

  /**
   * Deletes a base from the list (actually a map) of friendly bases, and sets the units that
   * were defending this base to move to another base.
   */
  void del_base(GameObj *gob);
  
  /**
   * Deletes a unit from the list (actually a map) of defenders, and removes it from the set
   * of defenders of whatever base it was defending. 
   */
  void del_unit(GameObj *gob);

  /**
   * Deletes an enemy unit from the set of enemy units, and, if it was the threat for any
   * building, reset that threat.
   */
  void del_enemy_unit(GameObj *gob);
  
  /**
   * Handles events sent to SuicideAttackAI from other modules (eg. pathfinding events, game
   * state module view updates).
   */
  bool handle_event(const Event &e);
};

//===================================================================

#endif
