#ifndef SI_VECTOR3D_H
#define SI_VECTOR3D_H

// $Id: SI_Vector3D.h 4182 2007-04-05 21:13:47Z kalf $

#include <iostream>
#include <sstream>
#include <cmath>

namespace Simul{

  class Vector3D{

  public:
    Vector3D(double x = 0.0f, double y = 0.0f, double z = 0.0f);
    Vector3D(double v[3]);
    Vector3D(int v[3]);
    Vector3D(const Vector3D& v);

    bool operator == (const Vector3D& v) const;
    bool operator != (const Vector3D& v) const;

    bool isZero() const;

    Vector3D operator - (const Vector3D& v) const;
    Vector3D& operator -= (const Vector3D& v);
    Vector3D operator + (const Vector3D& v) const;
    Vector3D& operator += (const Vector3D& v);
    Vector3D operator * (double f) const;
    Vector3D& operator *= (double f);
    Vector3D operator / (double f) const;
    Vector3D& operator /= (double f);

    Vector3D operator | (double length) const;
    Vector3D& operator |= (double length);

    double dot(const Vector3D& v) const;
    Vector3D cross(const Vector3D& v) const;
    Vector3D normalized() const;
    double length() const;
    double length2() const;
    Vector3D operator - () const;
    double& operator [] (int i);
    const double& operator[] (int i) const;

    friend Vector3D operator*(double f, const Vector3D& v);

    std::string toString() const;

  public:
    double x,y,z;
  };

  inline std::istream& operator >> (std::istream& s, Vector3D& v)
  {
    return s >> v.x >> v.y >> v.z;
  }

  inline Vector3D::Vector3D(double x, double y, double z) : x(x), y(y), z(z)
  {
  }

  inline Vector3D::Vector3D(double v[3]) : x(v[0]), y(v[1]), z(v[2])
  {
  }

  inline Vector3D::Vector3D(int v[3]) : x(v[0]), y(v[1]), z(v[2])
  {
  }

  inline Vector3D::Vector3D(const Vector3D& v) : x(v.x), y(v.y), z(v.z)
  {
  }

  inline bool Vector3D::operator == (const Vector3D& v) const
  {
    return x == v.x && y == v.y && z == v.z;
  }

  inline bool Vector3D::operator != (const Vector3D& v) const
  {
    return !((*this) == v);
  }

  inline bool Vector3D::isZero() const
  {
    return ( x == 0 && y == 0 && z == 0 );
  }

  inline Vector3D Vector3D::operator - (const Vector3D& v) const
  {
    return Vector3D(x-v.x,y-v.y,z-v.z);
  }

  inline Vector3D& Vector3D::operator -= (const Vector3D& v)
  {
    x -= v.x;
    y -= v.y;
    z -= v.z;
    return *this;
  }

  inline Vector3D Vector3D::operator + (const Vector3D& v) const
  {
    return Vector3D(x+v.x,y+v.y,z+v.z);
  }

  inline Vector3D& Vector3D::operator += (const Vector3D& v)
  {
    x += v.x;
    y += v.y;
    z += v.z;
    return *this;
  }

  inline Vector3D Vector3D::operator * (double f) const
  {
    return Vector3D(x*f,y*f,z*f);
  }

  inline Vector3D& Vector3D::operator *= (double f)
  {
    x *= f;
    y *= f;
    z *= f;
    return *this;
  }

  inline Vector3D operator* (double f ,const Vector3D& v)
  {
    return v*f;
  }

  inline double& Vector3D::operator[]( int i )
  {
    return (&x)[i];
  }

  inline const double& Vector3D::operator[]( int i ) const
  {
    return (&x)[i];
  }

  inline Vector3D Vector3D::operator-() const
  {
    return Vector3D(-x,-y,-z);
  }

  inline double Vector3D::dot(const Vector3D& v) const
  {
    return v.x * x + v.y * y + v.z * z;
  }

  inline Vector3D Vector3D::cross(const Vector3D& v) const
  {
    return Vector3D(y*v.z-z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);
  }

  inline Vector3D Vector3D::normalized() const
  {
    double l = sqrt(x*x + y*y + z*z);
    if( fabs(l) < 1e-10 ) return Vector3D();
    return Vector3D(x/l,y/l,z/l);
  }

  inline double Vector3D::length() const
  {
    return sqrt(x*x+y*y+z*z);
  }

  inline double Vector3D::length2() const
  {
    return (x*x + y*y + z*z);
  }

  inline Vector3D Vector3D::operator / (double f) const
  {
    return Vector3D(x/f,y/f,z/f);
  }

  inline Vector3D& Vector3D::operator /= (double f){
    x /= f;
    y /= f;
    z /= f;
    return *this;
  }

  inline bool overlaps(const Vector3D& minA,
		      const Vector3D& maxA,
		      const Vector3D& minB,
		      const Vector3D& maxB )
  {
    return maxA.x >= minB.x && maxB.x >= minA.x &&
      maxA.y >= minB.y && maxB.y >= minA.y &&
      maxA.z >= minB.z && maxB.z >= minA.z;
  }

  // return vector with specified length
  inline Vector3D Vector3D::operator | (double length) const{
    return *this * (length / (double)sqrt(x*x + y*y + z*z));
  }

  // set length of vector equal to length
  inline Vector3D & Vector3D::operator |= (double length){
    return *this = *this | length;
  }

  inline std::string Vector3D::toString() const{
    std::ostringstream os;
    os << "(" << x << ", " << y << ", " << z << ")";
    return os.str();
  }

  inline void toArray(const Vector3D& v, double v2[3]){
    v2[0] = v.x; v2[1] = v.y; v2[2] = v.z;
  }
}

#endif
