#include <string>

#include "BaseCommander.H"

#include "GameStateModule.H"
#include "GatherCommander.H"
#include "Game.H"
#include "BuildCommander.H"
#include "DefenseCommander.H"

#include "TechTree.H"

BaseCommander::Building::Building(ModuleSet & ms,
                                  BaseCommander * base,
                                  GameObj * building) :
    ms(ms), base(base), building(building),
    training(), funds(NULL),  watch(Watcher::NONE),
    confirmed(false), complete(true)
{
    int x0, y0, x1, y1;
    building->get_p1(x0, y0);
    building->get_p2(x1, y1);
    x0 -= 32; y0 -= 32;
    x1 += 32; y1 += 32;
    sint4 me = ms.get_game_state().get_game().get_client_player();
    watch = ms.get_watcher().notifyOnCreation(
        base,
        Watcher::And(Watcher::OwnerIs(me),
                     Watcher::InRectangle(x0, y0, x1, y1)),
        100,
        false);
}

BaseCommander::Building::~Building()
{
    if(funds != NULL)
    {
        funds->release();
    }

    if(watch != Watcher::NONE)
    {
        ms.get_watcher().cancelNotify(watch);
    }
}

GameObj * BaseCommander::Building::getBuilding()
{
    return building;
}

bool BaseCommander::Building::isBusy() const
{
    return (!complete);
}

const std::string & BaseCommander::Building::getTrainingBP() const
{
    return training;
}

bool BaseCommander::Building::canTrain(std::string bp_name) const
{
    return TechTree::GetTrainer(bp_name) == building->bp_name();
}

void BaseCommander::Building::train(std::string bp_name, Finance::Ticket * funds)
{
    assert( !isBusy() );
    assert( canTrain(bp_name) );
    assert( funds->available() >= ms.get_finance().get_cost(bp_name) );

    training = bp_name;
    action = TechTree::GetTrainAction(bp_name);
    this->funds = funds;
    confirmed = false;
    complete = false;

    building->set_action(action, Vector<sint4>());
}

Watcher::WatchID BaseCommander::Building::getWatch() const
{
    return watch;
}

void BaseCommander::Building::update()
{
    const stype * active_ptr = building->get_val_ptr("active");
    assert(active_ptr != NULL);
    bool active = ( active_ptr->to_int() == 1 );

    if(!confirmed)
    {
        if(active)
        {
            confirmed = true;
            funds->spend();
            funds = NULL;

            std::cerr << "CONFIRMED: " << building->bp_name();
            std::cerr << " is training " << training << std::endl;
        }
        else
        {
            building->set_action(action, Vector<sint4>());
        }
    }
    else if(!complete)
    {
        if(!active)
        {
            complete = true;

            std::cerr << "COMPLETE: " << building->bp_name();
            std::cerr << " has trained " << training << std::endl;
        }
    }
}

BaseCommander::BaseCommander(Commander* const mySuperior,
                             Vector<GameObj*> units,
			     TerrainBase::Loc centre,
                             sint4 updateTime_)
    : Commander(mySuperior, updateTime_),
      centreLoc(centre),
      builder(NULL),
      askingForBuilder(false)
{
    unitsQueued["worker"] = 0;
    unitsQueued["marine"] = 0;
    unitsQueued["tank"] = 0;
    unitsQueued["controlCenter"] = 0;
    unitsQueued["barracks"] = 0;
    unitsQueued["factory"] = 0;

    init(units);
}

BaseCommander::~BaseCommander()
{
}

/////////////////////////////////private////////////////////////////////

REGISTER_TYPEOF2(732101, std::map<GameObj*, BaseCommander::Building*>::iterator);
REGISTER_TYPEOF2(732102, std::map<GameObj*, BaseCommander::Building*>::const_iterator);

void BaseCommander::onReceiveUnit( GameObj * unit, Watcher::WatchID watchID )
{
    FORALL( obj2building, o2b )
    {
        Building * building = o2b->second;
        if( building->getWatch() == watchID )
        {
            unitsQueued[unit->bp_name()]--;
        }
    }

    if (unit->bp_name() == "worker")
    {
        if (askingForBuilder)
        {
            builder = unit;
            askingForBuilder = false;
        }
        else if (gatherer->needsUnit(unit->bp_name()))
        {
            removeUnit( unit );
            gatherer->addUnit( unit );
        }
        else
        {
            removeUnit( unit );
            defender->addUnit( unit );
        }
    }
    else if (unit->bp_name() == "controlCenter" ||
             unit->bp_name() == "barracks" ||
             unit->bp_name() == "factory")
    {
        Building * building = new Building(moduleSet, this, unit);
        obj2building.insert(std::make_pair(unit, building));
    }
    else if (unit->bp_name() == "marine" ||
             unit->bp_name() == "tank")
    {
        removeUnit( unit );
        defender->addUnit( unit );
    }
    else // Should never happen
    {
        assert( false );
    }
}

void BaseCommander::onLoseUnit( GameObj * unit )
{
    std::map<GameObj*, Building*>::iterator it;
    if( (it = obj2building.find(unit)) != obj2building.end() )
    {
        // If we had a unit training at this building
        // Be sure to record that it is no longer training
        Building * building = it->second;
        building->update();
        if(building->isBusy())
        {
            unitsQueued[building->getTrainingBP()]--;
        }
        delete building;

        obj2building.erase(unit);
    }

    if( unit == builder )
    {
        builder = NULL;
    }
}

void BaseCommander::buildBuildingsInQueue()
{
    while ( !buildingsToBuild.empty() )
    {
        // What are we going to build?
        BuildTask task = buildingsToBuild.front();

        // Try to acquire a worker
        askingForBuilder = true;
        size_t count = 0;
        count += defender->requestUnits(this, "worker", 1 - count);
        count += gatherer->requestUnits(this, "worker", 1 - count);
        askingForBuilder = false;

        if ( builder != NULL ) // If we have a worker
        {
            std::cerr << "Found one!" << std::endl;

            // Transfer him over to a new build commander
            GameObj * worker = builder;
            removeUnit(worker);
            
            TerrainBase::Loc workerLoc;
            worker->get_center(workerLoc.x, workerLoc.y);
            
            BuildCommander * buildCommander
                = new BuildCommander(this,
                                     worker,
                                     task.bp_name,
                                     workerLoc,
                                     task.funds,
                                     5);
            activeBuilders.insert(buildCommander);

            // And we are done for this building
            buildingsToBuild.pop();
        }
        else
        {
            std::cerr << "Cannot build, too busy..." << std::endl;
            return; // Can't build right now, return early
        }
    }
}

void BaseCommander::buildUnitsInQueue()
{
    // TODO: Optimize so we don't wait on a free control center
    // simply because we have a marine queued and the barracks is busy
    while( !unitsToBuild.empty() )
    {
        BuildTask task = unitsToBuild.front();

        Building * building = NULL;
        FORALL(obj2building, b)
        {
            Building * current = b->second;
            if( current->isBusy() == false &&
                current->canTrain(task.bp_name) )
            {
                building = current;
                break;
            }
        }

        if( building != NULL ) // If we have a building
        {
            building->train(task.bp_name, task.funds);

            unitsToBuild.pop();
        }
        else // Otherwise
        {
            return; // Cannot train any more just yet
        }
    }
}

void BaseCommander::exec()
{
    FORALL(finishedBuilders, b)
    {
        BuildCommander * builder = (*b);
        unitsQueued[ builder->getBuildingName() ]--;
        delete builder;
    }
    finishedBuilders.clear();

    FORALL(obj2building, o2b)
    {
        o2b->second->update();
    }

    if (!unitsToBuild.empty())
    {
        buildUnitsInQueue();
    }

    if (!buildingsToBuild.empty())
    {
        buildBuildingsInQueue();
    }
}

void BaseCommander::init(const Vector<GameObj*> & initialUnits)
{
    std::set<GameObj*> unitsToAssign(initialUnits.begin(), initialUnits.end());

    std::cout << "Base Commander init!" << std::endl;
    
    // Grab any gathering related units for the gatherer
    Vector<GameObj*> objectsForGather;
    std::set<GameObj*>::iterator it, it2;
    for(it = unitsToAssign.begin(); it != unitsToAssign.end(); )
    {
        it2 = it++;
        if ((*it2)->bp_name() == "worker")
        {
            objectsForGather.push_back( (*it2) );
            unitsToAssign.erase( it2 );
        }
        else if ((*it2)->bp_name() == "controlCenter")
        {
            objectsForGather.push_back( (*it2) );
            // NOTE: By not removing this unit from
            // the list of unitsToAssign, we are in
            // fact allowing both the GatherCommander
            // and ourselves to call addUnit and assume
            // ownership of the controlCenter. This is
            // a bit of a hack to enable us to build
            // workers in the BaseCommander. It would
            // probably be a better idea to allow the
            // GatherCommand to produce workers, and
            // we can request it to do so.
        }
    }
    
    std::cout << "Creating GatherCommander" << std::endl;
    gatherer = new GatherCommander(this,
                                   objectsForGather,
                                   centreLoc,
                                   160,
                                   1);
    
    defender = new DefenseCommander(this,
                                    centreLoc,
                                    256);

    // Distribute any remaining units as we see fit
    addUnits( unitsToAssign );
}

//////////////////////////public////////////////////////////

void BaseCommander::buildUnit(const std::string & unitBPname,
                              Finance::Ticket * funds)
{
    // Reserve enough funds to build the building
    Finance::Module & fm = moduleSet.get_finance();
    uint4 unitCost = fm.get_cost(unitBPname);
    assert( funds->available() >= unitCost );
    funds->trim(unitCost);

    unitsToBuild.push( BuildTask(unitBPname, funds) );
    unitsQueued[unitBPname]++;
}

void BaseCommander::buildBuilding(const std::string & buildingBPname,
                                  Finance::Ticket * funds)
{
    // Reserve enough funds to build the building
    Finance::Module & fm = moduleSet.get_finance();
    uint4 buildingCost = fm.get_cost(buildingBPname);
    assert( funds->available() >= buildingCost );
    funds->trim(buildingCost);

    buildingsToBuild.push( BuildTask(buildingBPname, funds) );
    unitsQueued[buildingBPname]++;
}

bool BaseCommander::canBuild(const std::string & objectBPname)
{
    // If a building, determine its builder
    const std::string & breq = TechTree::GetBuilder(objectBPname);
    if( breq != "" )
    {
        return (this->countOwned(breq) > 0);
    }

    const std::string & treq = TechTree::GetTrainer(objectBPname);
    if( treq != "" )
    {
        return (this->countOwned(treq) > 0);
    }

    return false;
}

bool BaseCommander::onEvent(const Event &event)
{
    if (event.get_who() == BuildCommander::FROM)
    {
        if (event.get_what() == BuildCommander::FINISHED)
        {
            std::cout << "Received Build Finished Event!" << std::endl;
            const BuildFinishedEvent & e =
                dynamic_cast<const BuildFinishedEvent &>(event);
            BuildCommander * commander = e.getSource();
            
            activeBuilders.erase(commander);
            finishedBuilders.insert(commander);
        }
        
        else if (event.get_what() == BuildCommander::FAILED)
        {
            std::cout << "Received Build Failed Event!" << std::endl;
            const BuildFailedEvent & e =
                dynamic_cast<const BuildFailedEvent &>(event);
            BuildCommander * commander = e.getSource();
            
            activeBuilders.erase(commander);
            finishedBuilders.insert(commander);
        }
    }
    return false;
}

bool BaseCommander::needsUnit(const std::string & unitBPname)
{
    if (unitBPname == "worker")
    {
      return gatherer->needsUnit("worker");
    }
    return false;
}

size_t BaseCommander::countQueued(const std::string & bp_name) const
{
    std::map<std::string, size_t>::const_iterator it;
    it = unitsQueued.find(bp_name);
    if(it != unitsQueued.end())
    {
        return it->second;
    }
    else
    {
        return 0;
    }
}

size_t BaseCommander::countWorkersNeeded() const
{
    return gatherer->countWorkersNeeded();
}

size_t BaseCommander::countRequestable(const std::string & bp_name) const
{
    size_t count = 0;

    count += defender->countRequestable(bp_name);
    count += gatherer->countRequestable(bp_name);

    return count;
}

size_t BaseCommander::requestUnits(Commander * receiver,
                                   const std::string & bp_name,
                                   size_t amount)
{
    size_t count = 0;

    // In the event that a worker is asked for, the DefenseCommander
    // may be using some idle workers as melee units. Snag one of them
    // rather than disturbing the gatherer if at all possible.
    count += defender->requestUnits(receiver, bp_name, amount - count);
    count += gatherer->requestUnits(receiver, bp_name, amount - count);

    return count;
}
