/** @file */
/*
 *  Commander.h
 *  by: Aaron Young, Ryan Neighbour
 *  Subclassed from EventHandler as an abstract class.
 */
#ifndef COMMANDER_H
#define COMMANDER_H

#include "EventHandler.H"
#include "GameObj.H"
#include "WatcherAux.H"
#include "ModuleSet.H"
#include "Random.H"

class Commander: public EventHandler
{
private:
    bool constructed;           ///< True if constructed
    std::set<GameObj*> units;   ///< Set of owned units
    std::set<Commander*> comms; ///< Set of sub-commanders

protected:
    /**
     * Pointer to immediate superior (0 for root).
     * @see Commander()
     * @see ~Commander()
     */
    Commander * myCommander;

    /**
     * Easy access point to numerous helper modules.
     */
    ModuleSet & moduleSet;

    /**
     * Count of how many updates until exec() is called.
     * Derived by rand(X)%maxUpdateCount + 1.
     * @see update()
     * @see exec()
     */
    int updateCounter;
    /**
     * Commander's id within ORTS.
     */
    int commanderID;
    /**
     * Maximum amount of update calls before exec() will be called.
     * Actual number set randomly.
     * @see update()
     * @see exec()
     */
    int maxUpdateCount;

protected:

    /**
     * The main function (or brain) of the commander.
     * Must be implemented in derived Commanders.
     * Will be called only once every rand() % maxUpdateCount + 1 updates.
     * Runs when updateCounter == 0.
     * @see update()
     * @see maxUpdateCount
     * @see updateCounter
     */
    virtual void exec() = 0;
    virtual void onReceiveUnit(GameObj *, Watcher::WatchID) = 0;
    virtual void onLoseUnit(GameObj *) = 0;
    virtual bool onEvent(const Event &) = 0;
    
    /**
     * Recursively calls update on all Commanders.
     * Decrements updateCounter by 1.
     * When updateCounter == 0 updateCounter is set to
     * a new random number (maxUpdateCount = maximum) and exec() is called.
     * @see updateCounter
     * @see maxUpdateCount
     * @see exec()
     */
    void update();

    Random rand;
    
public:

    Commander(ModuleSet& moduleSet, int maxUpdate=25);
    Commander(Commander* superior, int maxUpdate=25);
    virtual ~Commander();
    
    bool handle_event(const Event &event);

    sint4 getID() const { return commanderID; }

    virtual void countObjects(const std::string & bp_name,
                              size_t & accumulator) const;

    void addUnit(GameObj * unit);
    void removeUnit(GameObj * unit);

    template <typename CONTAINER>
    void addUnits(const CONTAINER & cont)
    {
        typename CONTAINER::const_iterator it;
        for(it = cont.begin(); it != cont.end(); it++)
        {
            addUnit( (*it) );
        }
    }

    template <typename CONTAINER>
    void removeUnits(const CONTAINER & cont)
    {
        typename CONTAINER::const_iterator it;
        for(it = cont.begin(); it != cont.end(); it++)
        {
            removeUnit( (*it) );
        }
    }
    
    /// Should give as many units of the specified type as
    /// possible to receiver, up to amount, without adversely
    /// affecting the ability of this commander to perform its
    /// designated task.
    /// @param receiver the commander who will receive the units
    /// @param bp_name the type of unit that will be given
    /// @param amount the maximum number of units to give
    /// @return the number of units given to receiver
    /// @see countRequestable
    virtual size_t requestUnits(Commander * receiver,
                                const std::string & bp_name,
                                size_t amount);

    /// Should return the maximum number of units that will
    /// be provided in response to a call to requestUnits()
    /// @param bp_name the type of unit to be counted
    /// @return the maximum number of units requestUnits() will provide
    /// @see requestUnits
    virtual size_t countRequestable(const std::string & bp_name) const;

    /// Will give as many units of the specified type as
    /// possible to receiver, up to amount. Will first attempt
    /// to request units, and then will take whatever it needs
    /// past that.
    /// @param receiver the commander who will receive the units
    /// @param bp_name the type of unit that will be given
    /// @param amount the maximum number of units to give
    /// @return the number of units given to receiver
    /// @see countOwned
    size_t demandUnits(Commander * receiver,
                       const std::string & bp_name,
                       size_t amount);

    /// Returns the maximum number of units that will
    /// be provided in response to a call to demandUnits()
    /// @param bp_name the type of unit to be counted
    /// @return the maximum number of units demandUnits() will provide
    /// @see demandUnits
    size_t countOwned(const std::string & bp_name) const;
};

#endif //COMMANDER_H
