#include "DefenseCommander.H"
#include "GameStateModule.H"
#include "SquadCombatAI.H"
#include "WatcherAux.H"
#include "Delegate.H"

REGISTER_TYPEOF2(466001, std::map<GameObj*, DefenseCommander::UNITSTATE>::iterator);
REGISTER_TYPEOF2(466002, std::map<GameObj*, DefenseCommander::UNITSTATE>::const_iterator);

DefenseCommander::DefenseCommander(Commander * mySuperior,
				   const TerrainBase::Loc & center,
				   sint4 radius) :
    Commander(mySuperior, 8),
    center(center),
    radius(radius),
    me(moduleSet.get_game_state().get_game().get_client_player()),
    underControl(false),
    mc(moduleSet.get_movement(), "Default", "Default"),
    ga(moduleSet, center.x, center.y, radius)
{
    moduleSet.get_combat().add_handler(this);

    mc.setStatusProcedure( make_delegate(this, &DefenseCommander::onMovement) );
    
    ga.set_passable_height( ga.sample_height(center) );
    ga.set_object_border( "mineral", 24 );
    ga.set_object_border( "controlCenter", 32 );
}

DefenseCommander::~DefenseCommander()
{
    moduleSet.get_combat().remove_handler(this);
}

GameObj * 
DefenseCommander::pickClosest(const Watcher::UnitPredPtr & pred,
                              const std::set<GameObj*> & candidates) const
{
    GameObj * target = NULL;
    real8 targetDistance = 0.0;

    // Determine the closest enemy
    FORALL( candidates, e )
    {
        GameObj * enemy = (*e);
        if(pred->evaluate(enemy) == false)
        {
            continue;
        }

        TerrainBase::Loc enemyLoc;
        enemy->get_center(enemyLoc.x, enemyLoc.y);
        real8 enemyDistance = center.distance(enemyLoc);

        if(target == NULL || enemyDistance < targetDistance)
        {
            target = enemy;
            targetDistance = enemyDistance;
        }
    }

    return target;
}

GameObj * 
DefenseCommander::pickBestTarget(const std::set<GameObj*> & candidates) const
{
    GameObj * target = pickClosest(Watcher::IsCombat(), candidates);
    if(target) return target;
    target = pickClosest(Watcher::IsWorker(), candidates);
    if(target) return target;
    return pickClosest(Watcher::IsBuilding(), candidates);
}

void DefenseCommander::exec()
{
    if( unit2state.empty() )
    {
        // Can't do anything without units
        // TODO: Appropriate message to commander
        underControl = false;
        return;
    }

    std::set<GameObj*> enemies;
    moduleSet.get_watcher().queryUnits(And(InCircle(center.x, center.y, radius),
							 Not(OwnerIs(me)), Not(OwnerIs(2))),
			 enemies);

    Vector<GameObj::WPtr> activeUnits; // Units which are attacking
    Vector<GameObj::WPtr> passiveUnits; // Units which are moving/idle

    SquadCombatAI & scai = moduleSet.get_combat();

    // If all enemies have been defeated
    if( enemies.empty() )
    {
        underControl = true;

        // Iterate through all of our units
		FORALL(unit2state, u2s)
		{
			GameObj * unit = u2s->first;
			UNITSTATE & state = u2s->second;

			scai.disable_unit_motion(unit);

			if( state == CHASING || state == ATTACKING )
			{
				state = STOPPED;
			}

			if( state == RETURNING )
			{
				underControl = false;
			}

			if( state == STOPPED )
			{
				sint4 boundary = (radius * 3) / 4;

				// If the unit is outside of his zone
				TerrainBase::Loc objLoc;
				unit->get_center(objLoc.x, objLoc.y);
				sint4 distance = static_cast<sint4>(center.distance(objLoc));
				if (distance > boundary)
				{
					// If we can find a free spot, move to it
					TerrainBase::Loc target;
					if( ga.find_free_spot(center, boundary, target) )
					{
						mc.moveUnit(unit, Movement::TouchPoint(target));

						state = RETURNING;
						underControl = false;
					}
				}
				else
				{
					// Otherwise we couldn't find a free spot
					// Just wait around, we'll find one eventually
				}
			}
		}
    }
    else // There are enemies
    {
        GameObj * target = pickBestTarget(enemies);
		assert(target != NULL); // There should be at least one enemy

        Vector<GameObj::WPtr> wvec;




        FORALL( unit2state, u2s )
        {
	    GameObj * unit = u2s->first;
            UNITSTATE & state = u2s->second;

            real8 distance = unit->distance_to(*target);
            
            if( state == STOPPED || state == ATTACKING )
            {
                if( distance > 120.0f )
                {
		    mc.moveUnit(unit, Movement::AttackObject(target));

                    state = CHASING;
                }
                else
                {
                    state = ATTACKING;
                }
            }
            else // We are chasing or moving
            {
                if( distance < 80.0f )
                {
                    state = ATTACKING;
                }
            }

            if( state == ATTACKING )
            {
                scai.enable_unit_motion( unit );
                wvec.push_back( unit );
            }
            else
            {
                scai.disable_unit_motion( unit );
            }
        }

        scai.handle_event(SquadCombatEvent(commanderID,
                                           SquadCombatAI::ATTACK_TARGET,
                                           wvec, target));

        underControl = false;
    }
}

void DefenseCommander::onReceiveUnit ( GameObj * unit, Watcher::WatchID )
{
    unit2state.insert( std::make_pair(unit, STOPPED) );

    SquadCombatAI & scai = moduleSet.get_combat();
    scai.disable_unit_motion(unit);
}

void DefenseCommander::onLoseUnit ( GameObj * unit )
{
    unit2state.erase( unit );
}

bool DefenseCommander::onEvent ( const Event & )
{
    return false;
}

void DefenseCommander::onMovement ( const Movement::Task & task )
{
    FORALL( task.getUnits(), u )
    {
        std::map<GameObj*, UNITSTATE>::iterator it;
        it = unit2state.find(const_cast<GameObj*>(*u)); // TODO: Fix
        if(it != unit2state.end())
        {
	    std::cerr << "DefenseCommander got path message for ";
	    std::cerr << (*u)->bp_name() << std::endl;

	    it->second = STOPPED;
	}
    }
}

bool DefenseCommander::isUnderControl ( ) const
{
    return underControl;
}

size_t DefenseCommander::requestUnits(Commander * receiver,
                                      const std::string & bp_name,
                                      size_t amount)
{
    size_t numToGiveUp = countRequestable(bp_name);
    if(amount > numToGiveUp) amount = numToGiveUp;

    Vector<GameObj*> toGive;
    FORALL( unit2state, u2s )
    {
        if(toGive.size() == amount)
        {
            break;
        }

        GameObj * unit = u2s->first;
        if( unit->bp_name() == bp_name )
        {
            toGive.push_back( unit );
        }
    }
    this->removeUnits(toGive);
    receiver->addUnits(toGive);

    return toGive.size();
}

size_t DefenseCommander::countRequestable(const std::string & bp_name) const
{
    size_t count = 0;
    FORALL( unit2state, u2s )
    {
        GameObj * unit = u2s->first;
        if(unit->bp_name() == bp_name)
        {
            count++;
        }
    }

    size_t numToKeep = 0;
    if(bp_name == "marine")
    {
        numToKeep = 4;
    }
    if(underControl && count > numToKeep )
    {
        return count - numToKeep;
    }
    else
    {
        return 0;
    }
}
