#include "PlayerCommander.H"
#include "BuildCommander.H"
#include "GameChanges.H"
#include "SquadCombatAI.H"
#include "ScoutCommander.H"
#include "GatherCommander.H"

PlayerCommander::PlayerCommander(ModuleSet & ms) :
  Commander(ms),
  gsm(ms.get_game_state()),
  mc(ms.get_movement(), "Default", "Default")
{
  ms.get_history().trackBlueprint("mineral");
  
  scoutComm = new ScoutCommander(this);
  
  gsm.add_handler(this);
}

PlayerCommander::~PlayerCommander()
{
  gsm.remove_handler(this);
}

bool PlayerCommander::unitOwnedByPlayer(GameObj *unit) const
{
  return *unit->sod.owner == client;
}

GatherCommander *PlayerCommander::pickBestGC(GameObj *resource) const
{
  real8 bestDistance(0);
  GatherCommander *bestGC(0);
  
  TerrainBase::Loc resLoc;
  resource->get_center(resLoc.x, resLoc.y);
  
  std::map<GameObj::Ptr, GatherCommander*>::const_iterator it;
  for (it = cc2comm.begin(); it != cc2comm.end(); ++it)
  {
    GameObj *cc = it->first;
    TerrainBase::Loc ccLoc;
    cc->get_center(ccLoc.x, ccLoc.y);
    
    real8 distance = ccLoc.distance(resLoc);
    if (!bestGC || distance < bestDistance) {
      bestDistance = distance;
      bestGC = it->second;
    }
  }
  
  return bestGC;
}

void PlayerCommander::stop(GameObj *unit)
{
  if (!unitOwnedByPlayer(unit)) return;
  
  if (moduleSet.has_combat()) {
    moduleSet.get_combat().disable_unit_motion(unit);
  }
  
  std::map<GameObj*, Commander*>::iterator u2c;
  u2c = unit2commander.find(unit);
  if (u2c != unit2commander.end()) {
    u2c->second->removeUnit(unit);
    unit2commander.erase(u2c);
  }
}

void PlayerCommander::move(GameObj *unit, sint4 x, sint4 y)
{
  if (!unitOwnedByPlayer(unit)) return;
  
  stop(unit);
  mc.moveUnit(unit, Movement::TouchPoint(Movement::Vec2D(x,y)));
}

void PlayerCommander::gather(GameObj *unit, GameObj *target)
{
  if (!unitOwnedByPlayer(unit)) return;
  
  stop(unit);
  GatherCommander *gc = pickBestGC(target);
  if (gc) {
    std::cout << "Worker given to gatherer" << std::endl;
    gc->addUnit(unit);
    unit2commander.insert( std::make_pair(unit, gc) );
  } else {
    std::cout << "No suitable gatherer out of ";
    std::cout << cc2comm.size() << std::endl;
  }
}

void PlayerCommander::build(GameObj *unit, const std::string & bp_name,
                            sint4 x, sint4 y)
{
  if (!unitOwnedByPlayer(unit)) return;
  
  stop(unit);
  Commander *comm = new BuildCommander(this, unit, bp_name, TerrainBase::Loc(x,y));
  commanders.insert(comm);
}

void PlayerCommander::scout(GameObj *unit)
{
  if (!unitOwnedByPlayer(unit)) return;
  
  stop(unit);
  scoutComm->addUnit(unit);
  unit2commander.insert( std::make_pair(unit, scoutComm) );
}

void PlayerCommander::autoAttack(GameObj *unit)
{
  if (!unitOwnedByPlayer(unit)) return;
  
  stop(unit);
  if (moduleSet.has_combat()) {
    moduleSet.get_combat().enable_unit_motion(unit);
  }
}

bool PlayerCommander::onEvent(const Event & e)
{
  if (e.get_what() == GameStateModule::VIEW_MSG) {
    client = gsm.get_game().get_client_player();
    
    const GameChanges & gc(gsm.get_changes());
    
    // New player units should start with squad motion
    // disabled. We can enable it later with a command

    SquadCombatAI *scai = 0;
    if (moduleSet.has_combat()) scai = &moduleSet.get_combat();

    FORALL (gc.new_objs, u) {
      GameObj * unit((*u)->get_GameObj());
      if (*unit->sod.owner == client) {
        if (scai) scai->disable_unit_motion( unit );
        
        if (unit->bp_name() == "controlCenter")/* ||
            unit->bp_name() == "pantheon_acropolis" ||
            unit->bp_name() == "mound" ||
            unit->bp_name() == "tor" ||
            unit->bp_name() == "citadel")*/
          {
            GameObj::Ptr ccPtr(unit);
            Vector<GameObj*> empty;
            TerrainBase::Loc center;
            unit->get_center(center.x, center.y);
            GatherCommander * gc =
              new GatherCommander(this, empty, center, 256, 8);
            cc2comm.insert(std::make_pair(ccPtr, gc));
            gc->addUnit(unit);
          }
      }
    }
    
    // Remove any dead objects from the unit
    // to commander mapping. We don't need to
    // check if they belong to the player because
    // if they don't they will not be in the
    // mapping anyway.
    FORALL (gc.dead_objs, u) {
      GameObj *unit((*u)->get_GameObj());
      unit2commander.erase(unit);
    }
    
    std::set<Commander*>::iterator c, ct;
    for (c = commanders.begin(); c != commanders.end(); )
    {
      ct = c++; // My favorite language
      Commander *comm(*ct);
      if (comm->countOwned("") == 0) {
        delete comm;
        commanders.erase(ct);
      }
    }
    
    // Drive the update tree
    this->update();
    
    return true;
  }
  
  return false;
}
