#ifndef GatherCommander_H
#define GatherCommander_H

/* 
 * Gather Commander
 * Anthony Marcar
*/

#include "Commander.H"

#include "Global.H"
#include "ClientModule.H"

#include "TerrainBase.H"
#include "History.H"

#include "Worker.H"
#include "MineralCluster.H"

#include "Movement.H"

class RouteAllocator;
class GameStateModule;
class TerrainModule;

typedef boost::shared_ptr<RouteAllocator> RouteAllocatorPtr;
typedef boost::shared_ptr<Route> RoutePtr;
typedef boost::shared_ptr<MineralCluster> MineralClusterPtr;

/**
 * Stores the access points and centre of a mineral.  Used in 
 * computeAccessPoints()
 */
class Mineral {
public:
    Vector<TerrainBase::Loc> points;
    TerrainBase::Loc centre;
    Mineral(Vector<TerrainBase::Loc> points_, TerrainBase::Loc centre_)
        : points(points_), centre(centre_)
    {}
    ~Mineral(){}
};

typedef boost::shared_ptr<Mineral> MineralPtr;

/**
 * The purpose of the gather commander is to efficiently mine any minerals around
 * a base as efficiently as possible given a set of workers.  It operates in fog of
 * war and in perfect information.  
 */
class GatherCommander : public Commander, ClientModule 
{
    /// Identifier of this class when sending events
    static const sint4 FROM; 

    /// The minimum number of possible clusters
    static const sint4 MIN_NUM_CLUSTERS = 1;
    /// The maximum number of possible clusters
    static const sint4 MAX_NUM_CLUSTERS = 4;
    /// Maximum cluster/control center variance for a close cluster scheme
    static const sint4 RANGE_FOR_CLOSE_SCHEME = 25;
    /// Maximum cluster/control center distance for a close cluster scheme
    static const sint4 DIST_FOR_CLOSE_SCHEME = 100;

    /// The maximum amount of minerals a worker can carry 
    static const sint4 WORKER_MINERAL_LOAD_MAX = 10;
    /// Maximum worker/mineral distance which allows mining
    static const sint4 WORKER_MINERAL_DISTANCE_THRESHOLD = 2;
    /// Maximum worker/control-center distance which allows returning minerals
    static const sint4 WORKER_CC_DISTANCE_THRESHOLD = 3;

    /// worker's blueprint name
    static const std::string workerBPname;          
    /// minerals' blueprint name
    static const std::string mineralBPname;         
    /// control center's blueprint name
    static const std::string controlCenterBPname;   

    /// The centre of the sphere of influence.  Only minerals within this area
    /// will be mined.
    TerrainBase::Loc centreLoc;
    /// The distance from the centre that minerals can be used within.
    sint4 radius;

    /// Reference to the History Module.  Used for mineral locations.
    History::HistoryModule &history;
    /// Reference to the Game State Module.
    GameStateModule& gsm;
    /// Context for pathfinding and execution calls
    Movement::Context mc;

    /// a class that sorts and allocates the routes
    boost::shared_ptr<RouteAllocator> routeAllocator;
 
    /// a vector of all the workers owned by this commander
    Vector<Worker*> workers;
    /// a vector of control centers this commander can use
    Vector<GameObj*> controlCenters;
    /// a vector of mineral clusters containing organised Minerals
    Vector<boost::shared_ptr<MineralCluster> > mineralClusters;

    /// Contains the History records.  All calls to minerals are done through these
    /// records (for safety in fog of war).
    std::set<const History::Record*> mineralRecs;

    /// Contains any minerals that came into view (or out of fog of war) since the 
    /// last frame.
    std::set<GameObj*> newMinerals;

    //sint4 workerMaxMinerals;

    /// flag used for initialisation
    bool receivedFirstView;
    /// If a call to a history isActive() returns false, this flag is set to true.
    bool mineralDied;
    /// The ID of the watcher query, so that it can be cancelled.
    uint4 newMineralsWatchID;
    /// A set of any queries currently being calculated by the watcher.
    std::set<uint4> watcherIDs;

    ///////////////////////METHODS////////////////////////////

    /**
     * If there has been a change to the known minerals, this method will assess
     * whether the changes warrant recomputing of routes.  Ie, if a new mineral is 
     * found or a mineral is used up.
     */
    void assessChangesToMinerals();

    /** 
     * sorts the minerals into clusters.  These clusters are stored in the container
     * mineralClusters.
    */
    void computeMineralClusters();

    /**
     * Using the clusters, this method calculates Routes for each access point that
     * passes a series of tests.  After calculation, the only valid access points
     * should be on the outside of any cluster.
     */
    void computeRoutes();

    /**
     * called whenever a status update is received
     * @param task the task object whose status has changed
     */
    void handleMovementStatus(const Movement::Task & task);
    
    /**
     * called whenever a unit cannot move to its next point
     * @param worker the worker that cannot move
     */
    void handleMoveFailure(Worker *worker);
   
    /**
     * called whenever a unit cannot find a path to its destination
     * @param worker the worker that cannot find a path
     */
    void handlePlanFailure(Worker *worker);

    /**
     * called whenever a unit reaches its destination.
     * @param worker the worker that reached its destination
     */
    void handleSuccess(Worker *worker);

    /**
     * perform any needed initialisation.
     */
    void init();

    /**
     * scans the area nearby for any mineral patches and places them in NewMinerals
     */
    void locateNearbyMinerals();

    /**
     * looks at the state of each worker and decided its next action.
     */
    void performWorkerActions();

public:

    /**
     * creates a gather commander
     * @param mySuperior the superior commander (the commander who created this).
     * @param units the vector of game objects that this commander has control over
     * @param centre the centre of the commander's sphere of influence
     * @param radius the radius of the commander's sphere of influence
     * @param updateTime how often this commander should perform actions
     */
    GatherCommander(Commander* const mySuperior, 
                    const Vector<GameObj*> & units, 
                    TerrainBase::Loc centre,
                    sint4 radius,
                    sint4 updateTime);

    ~GatherCommander();

    /**
     * returns the number of units that this commander can part with
     * @param bp_name the blueprint name of the unit to request
     * @return the number of units this Commander can part with
     */
    size_t countRequestable(const std::string & bp_name);

    /**
     * Try to determine how many workers are needed by this gatherer.
     * @return the number of workers needed by this gatherer
     */
    size_t countWorkersNeeded() const;

    /**
     * this is where everything happens! (or should happen)
     */
    void exec();

    /**
     * returns the pathfinding distance between a game object and a location.
     * uses the radius of the game object.
     * @param obj the object from which to path from (uses radius of this object).
     * @param loc the location to reach.
     */
    real8 getDistance(GameObj *obj, TerrainBase::Loc loc);

    /**
     * returns the pathfinding distance between two locations.
     */
    //real8 getDistance(TerrainBase::Loc loc1, TerrainBase::Loc loc2, 
    //                  sint4 radius); 

    /**
     * returns the number of units owned by this commander
     * @param unitBPname the blueprint of the unit to test
     * @return the number of units owned by this commander
     */
    sint4 getNumUnits(const std::string & unitBPname);

    /**
     * Moves the unit to a random location in its immidiate vicinity
     * @param unit the object to move
     */
    void moveRandomly(GameObj *unit);

    /**
     * sends a unit to a game object using the pathfinder
     * @param unit the unit that should move
     * @param gob the Game Object to move to
     */
    void moveTo(GameObj *unit, GameObj* gob);

    /**
     * sends a unit to a location using the pathfinder
     * @param unit the unit that should move
     * @param location the location to move to
     */
    void moveTo(GameObj *unit, const TerrainBase::Loc &location);

    /**
     * returns whether this commander has capacity for more of the specified unit.
     */
    bool needsUnit(const std::string & unitBPname);

    /**
     * reacts to most events
     * @param event the event to process
     */
    bool onEvent(const Event& event);

    /**
     * reacts to any losses of resources.
     * @param object the object that was lost
     */
    void onLoseUnit(GameObj * object);

    /**
     * reacts to new resources
     * @param object the new object
     * @param watch the watch ID that triggered the new object
     */
    void onReceiveUnit(GameObj * object, Watcher::WatchID watch);

    /**
     * gives a number of units to another commander
     * @param receiver the commander who will receive the units
     * @param bp_name the blueprint of the unit that should be given
     * @param amount the number of units to give
     * @return the number of units that were actually given
     */
    size_t requestUnits(Commander * receiver,
                        const std::string & bp_name,
                        size_t amount);

    /**
     * gives a single worker to the requesting commander
     * @param requestingCommander the commander who will receive the worker
     * @return true if the transfer was succesful.  False if there are no workers.
     */
    bool requestWorker(Commander * requestingCommander);

    /**
     * Stops a worker from moving.
     * @param worker the unit that should stop moving
     */
    void stopWorker(GameObj *worker);
};

REGISTER_TYPEOF(110103, Vector<boost::shared_ptr<MineralCluster> >::iterator);
REGISTER_TYPEOF(110104, Vector<boost::shared_ptr<MineralCluster> >::const_iterator);
REGISTER_TYPEOF(110105, Vector<MineralPtr>::iterator);
REGISTER_TYPEOF(110106, Vector<MineralPtr>::const_iterator);
REGISTER_TYPEOF(110107, Vector<Worker*>::iterator);
REGISTER_TYPEOF(110108, Vector<Worker*>::const_iterator);

#endif
