#ifndef RouteAllocator_H
#define RouteAllocator_H

#include "TerrainBase.H"
#include "boost/shared_ptr.hpp"
#include "Route.H"

typedef TerrainBase::Loc Loc;
typedef boost::shared_ptr<Route> RoutePtr;

/** A comparison functor used for sorting mineral patches 
    according to descending distance from a location */
struct CompareRoutesByDistanceToLocation {
    TerrainBase::Loc cc;
    CompareRoutesByDistanceToLocation(TerrainBase::Loc cc_) : cc(cc_) {}
    bool operator() (const RoutePtr route1, const RoutePtr route2) const 
    {
        real8 route1Dist = route1->getRouteLength();
        real8 route2Dist = route2->getRouteLength();
        return (route1Dist < route2Dist);
    }
};

/**
 * The route allocator is responsible for assessing which route is the best to
 * assign to a given unit.  Units request a route using the getBestRoute() method.
 * The route allocator can sort routes according to certain criteria
 */
class RouteAllocator {
private:
    /// The maximum number of workers that can mine a mineral at one time.
    static const sint4 MAX_WORKERS_AT_ONE_TIME = 4;
    /// a vector of routes.
    Vector<RoutePtr> routes;
    /// the number of workers that could use the route allocator
    sint4 workers;
    /// the route position to use.
    sint4 pos;
    /// the number of the last ID given to a route.
    sint4 idPos;

    /**
     * create a new unique ID to assign to a new route
     * @return the ID number
     */
    sint4 newID();

public:
    /**
     * creates a new Route Allocator
     */
    RouteAllocator();

    /**
     * creates a new Route Allocator
     * @param the initial number of workers who will use this allocator
     */
    RouteAllocator(sint4 workers);

    ~RouteAllocator();

    /**
     * adds a route to this allocator.
     * @param route the pointer to a route
     */
    void addRoute(RoutePtr route);

    /**
     * increments the number of workers who could use the allocator.
     */
    void addWorker();

    /**
     * returns the best available route, bearing in mind the location of the unit
     */
    RoutePtr getBestRoute(TerrainBase::Loc locOfUnit);

    /**
     * returns the total number of routes
     */
    sint4 getNumRoutes();

    /**
     * returns the number of available routes
     */
    uint4 getNumAvailableRoutes();

    void markRouteAsBad(RoutePtr route);

    /**
     * returns the worst route currently being used (for taking out a unit)
     * @return a pointer to the route.
     */
    RoutePtr getWorstUsedRoute();

    /**
     * searches through the routes and erases any invalid route lenghts (ie -1, 0)
     */
    void removeInaccessibleRoutes();

    /**
     * removes a route from this allocator
     * @param route the route to remove
     */
    void removeRoute(RoutePtr route);

    /**
     * decrements the number of workers using the allocator
     */
    void removeWorker();

    /**
     * sets the number of workers using this allocator
     * @param numWorkers the number of workers
     */
    void setWorkers(sint4 numWorkers);

    /**
     * sorts the routes in ascending order of distance from a location.
     */
    void sortRoutesAccordingToLocation(TerrainBase::Loc location);

    /**
     * Determine the number of workers needed (in total) to use every
     * available route.
     * @return the number of workers needed
     */
    size_t countWorkersNeeded() const;
};

REGISTER_TYPEOF(110001, Vector<RoutePtr>::iterator);
REGISTER_TYPEOF(110002, Vector<RoutePtr>::const_iterator);

#endif
