#ifndef Worker_H
#define Worker_H

#include "TerrainBase.H"
#include "History.H"
#include "boost/shared_ptr.hpp"

class GatherCommander;
class Route;
class GameStateModule;
class GameObj;

typedef boost::shared_ptr<Route> RoutePtr;

class Worker {
protected:
 
    /// reference to the game state module
    GameStateModule *gsm;
    ///the commander who owns this worker
    GatherCommander *commander;
    /// the game object that is this worker
    GameObj *gob;
    /// the route that this worker is using
    RoutePtr route;
    /// the current state: MOVING, RETURING, IDLE, MINING
    sint4 state;
    /// whether this worker is ready to start on its route
    bool knowsRoute;

public:
    //states
    static const sint4 IDLE = 0;
    static const sint4 MOVING = 1;
    static const sint4 MINING = 2;
    static const sint4 RETURNING = 3;
    /// the maximum distance a worker has to be from a mineral in order to mine
    static const sint4 WORKER_MINERAL_DISTANCE_THRESHOLD = 2;
    /// the maximum distance a worker has to be from a control center to return
    static const sint4 WORKER_CC_DISTANCE_THRESHOLD = 3;


    /////////////////////// Constructors /////////////////////////
    /**
     * creates a new worker
     * @param gob the game object that is the worker
     * @param commander the commander that created this worker
     * @param gsm the game state module
     */
    Worker(GameObj* gob, 
           GatherCommander *commander, 
           GameStateModule *gsm);
    
    ~Worker();
   
    /**
     * sets knowsRoute to false
     */
    void clearRoute();

    /**
     * whether the worker is close enough to drop off its minerals at the control
     * centre.
     * @return whether it's close enough
     */
    bool closeEnoughToDropOff() const;

    /**
     * determines if the worker is close enough to mine and moves if necessary
     * @return whether it's close enough
     */
    bool closeEnoughToMine() const;

    /**
     * sends the action for a worker to drop its minerals at the control centre
     */
    void dropOffMinerals();

    /**
     * the control center that this worker returns to
     * @return a pointer to the control centre game object
     */
    GameObj* getControlCenter() const;

    /**
     * the amount of minerals this worker is carrying
     * @return amount of minerals
     */
    sint4 getCurrentMinerals() const;

    /**
     * the game object of this worker
     * @return a pointer to the game object
     */
    GameObj* get_gob() const;

    /**
     * the CURRENT location of this worker
     * @return the location
     */
    TerrainBase::Loc getLocation() const;

    /**
     * the mineral that this worker mines from
     * @return the history record of the mineral
     */
    const History::Record* getMineralRec() const;
    
    /**
     * The route that this worker is using
     * @return a pointer to the route
     */
    RoutePtr getRoute() const;

    /**
     * the current state of this worker
     * @return IDLE, MOVING, RETURNING, MINING
     */
    sint4 getState() const;

    /**
     * whether this worker has an allocated route
     */
    bool hasRoute() const;

    /**
     * whether this worker is currently mining
     */
    bool isMining() const;

    /**
     * sends the action to mine from the mineral
     */
    void mine();

    /**
     * moves this worker to its mineral
     */
    void moveToMineral();

    /**
     * moves this worker to its control center
     */
    void moveToCC();

    /**
     * sets the route that this worker should use
     */
    void setRoute(RoutePtr route_);

    /**
     * set the state of this worker.
     * @param state can be IDLE, MOVING, MINING or RETURNING
     */
    void setState(sint4 state_);
};

#endif
