#include "History.H"

#include "GameStateModule.H"
#include "GameChanges.H"
#include "Game.H"
#include "PlayerInfo.H"

#include <iostream>

namespace History
{
    Record::Record(GameObj * object) : 
        visible(true), object(GameObj::Ptr(object))
    {
        bp_name = object->bp_name();
        object->get_center(position.x, position.y);
    }

    Record::~Record()
    {
        
    }

    bool Record::isActive() const
    {
        return object.good();
    }

    bool Record::isVisible() const
    {
        return visible;
    }

    const std::string & Record::getBlueprint() const
    {
        return bp_name;
    }

    const TerrainBase::Loc & Record::getPosition() const
    {
        return position;
    }

    GameObj * Record::getObject() const
    {
        return (GameObj*)object;
    }

    void Record::setVisible(bool visible)
    {
        this->visible = visible;
    }

    void Record::changeObject(GameObj * newObject)
    {
        object = GameObj::Ptr(newObject);
    }

    HistoryModule::HistoryModule(GameStateModule& gsm) : gsm(gsm)
    {
        gsm.add_handler(this);
    }

    HistoryModule::~HistoryModule()
    {
        
    }

    void HistoryModule::handle_new(GameObj * object)
    {
        for(size_t i=0; i<records.size(); i++)
        {
            if(!records[i]->isActive()) continue;
            if(records[i]->isVisible()) continue;

            const TerrainBase::Loc & recPos = records[i]->getPosition();
            TerrainBase::Loc objPos; object->get_center(objPos.x, objPos.y);
            if(recPos.x != objPos.x) continue;
            if(recPos.y != objPos.y) continue;
            if(records[i]->getBlueprint() != object->bp_name()) continue;

            obj2rec.erase(records[i]->getObject());
            records[i]->setVisible(true);
            records[i]->changeObject(object);
            obj2rec.insert(std::make_pair(object, records[i]));
            return;
        }

        records.push_back(new Record(object));
        obj2rec.insert(std::make_pair(object, records.back()));
    }

    void HistoryModule::check_vis()
    {
        return;

        const Game & game = gsm.get_game();
        const PlayerInfo & player = game.get_cplayer_info();
        for(size_t i=0; i<records.size(); i++)
        {  
            if(records[i]->isActive() &&
               !records[i]->isVisible())
            {
                const TerrainBase::Loc & pos = records[i]->getPosition();
                int ppt = game.get_tile_points(); // Points per tile
                int half = ppt / 2;
                int x0 = (pos.x - half) / ppt; 
                int y0 = (pos.y - half) / ppt;
                int x1 = x0 + 1;
                int y1 = y0 + 1;
                int count = 0;
                if( player.visible_W.is_set(x0, y0) ) count++;
                if( player.visible_W.is_set(x1, y0) ) count++;
                if( player.visible_W.is_set(x0, y1) ) count++;
                if( player.visible_W.is_set(x1, y1) ) count++;
                if( player.visible_E.is_set(x0, y0) ) count++;
                if( player.visible_E.is_set(x1, y0) ) count++;
                if( player.visible_E.is_set(x0, y1) ) count++;
                if( player.visible_E.is_set(x1, y1) ) count++;
                if( count == 8 )
                {
                    // Mark object as dead
                    records[i]->changeObject(NULL);
                }
            }
        }
    }

    bool HistoryModule::handle_event(const Event &event)
    {
        if( event.get_who() != GameStateModule::FROM )
        {
            return false;
        }

        if( event.get_what() != GameStateModule::VIEW_MSG )
        {
            return false;
        }

        const GameChanges& gc = gsm.get_changes();
        Vector<ScriptObj::WPtr>::const_iterator it;
        for(it = gc.new_objs.begin(); it != gc.new_objs.end(); it++)
        {
            ScriptObj * sobj = (*it);
            GameObj * object = dynamic_cast<GameObj*>(sobj);
            if( object != NULL )
            {
                if(isTracking(object->bp_name()))
                {
                    handle_new(object);
                }
            }
        }

        for(it = gc.vanished_objs.begin(); it != gc.vanished_objs.end(); it++)
        {
            ScriptObj * sobj = (*it);
            GameObj * object = dynamic_cast<GameObj*>( sobj );
            if( object != NULL )
            {
                std::map<GameObj*, Record*>::iterator it;
                it = obj2rec.find(object);
                if(it != obj2rec.end())
                {
                    Record * record = it->second;
                    record->setVisible(false);
                }
            }            
        }

        for(it = gc.dead_objs.begin(); it != gc.dead_objs.end(); it++)
        {
            ScriptObj * sobj = (*it);
            GameObj * object = dynamic_cast<GameObj*>( sobj );
            if( object != NULL )
            {
                std::map<GameObj*, Record*>::iterator it;
                it = obj2rec.find(object);
                if(it != obj2rec.end())
                {
                    Record * record = it->second;
                    record->setVisible(false);
                    record->changeObject(NULL);
                }
            }
        }

        check_vis();

        return true;
    }

    void HistoryModule::trackBlueprint(const std::string & bp_name)
    {
        tracking.insert(bp_name);
    }

    bool HistoryModule::isTracking(const std::string & bp_name) const
    {
        return(tracking.find(bp_name) != tracking.end());
    }

    const Record * HistoryModule::getRecord(GameObj * object) const
    {
        std::map<GameObj*, Record*>::const_iterator it;
        it = obj2rec.find(object);
        if(it != obj2rec.end())
        {
            return it->second;
        }
        else
        {
            return NULL;
        }
    }

    void HistoryModule::getActiveRecords(std::vector<const Record*>& rec) const
    {
        for(size_t i=0; i<records.size(); i++)
        {
            if(records[i]->isActive())
            {
                rec.push_back(records[i]);
            }
        }
    }

    void HistoryModule::queryActiveRecords(
        const Watcher::UnitPredPtr & query,
        std::vector<const Record*> & result) const
    {
        for(size_t i=0; i<records.size(); i++)
        {
            if(records[i]->isActive())
            {
                if( query->evaluate( records[i]->getObject() ) )
                {
                    result.push_back(records[i]);
                }
                else
                {
                }
            }
        }
    }
}
