#ifndef __HISTORY_H__
#define __HISTORY_H__

#include "GameObj.H"      // For GameObj
#include "EventHandler.H" // For EventHandler, Event
#include "TerrainBase.H"  // For Loc (*sigh*)
#include "Watcher.H"      // For UnitPred (TODO: Factor out)

class GameStateModule;

namespace History
{
    /// Record objects exist as long as the HistoryModule is in scope. One
    /// Record object is created for each new instance of a tracked blueprint.
    /// Should the object it refers to vanish, the Record will keep the
    /// corresponding GameObj alive as a "memory", storing the last known
    /// attributes of the object, such as health or minerals remaining.
    /// The Record also provides easy access to the location of the object,
    /// whether or not it is currently visible, and whether or not it is
    /// active. An object is considered "active" as long as we do not yet
    /// know for sure that it is dead. If we see an object die, its Record
    /// will be set to inactive. Also, if we arrive at a location where an
    /// object used to exist and it is gone, its Record will be set to
    /// inactive. Note that even after a Record ceases to be active, the
    /// Record object will continue to exist for the remainder of the 
    /// simulation. Any time that the isActive() call returns true, we can
    /// call getObject() to retrieve the object containing the most recent
    /// information we know about this object.
    class Record
    {
        friend class HistoryModule;

        bool             visible;
        std::string      bp_name;
        TerrainBase::Loc position;
        GameObj::Ptr     object;

        Record(GameObj * object);
        ~Record();

        void setVisible(bool visible);
        void changeObject(GameObj * newObject);
    public:
        /** \brief determine whether or not this is a record of an active unit
         *  \return false if the unit is positively known to be dead
         *  \sa getObject
         */
        bool isActive() const;

        /** \brief determine if this object's information is up to date
         *  \return true if the object is visible by one of our units
         */
        bool isVisible() const;

        /** \brief retrieve the name of the blueprint used by this object
         *  \return the blueprint name as a string
         */
        const std::string & getBlueprint() const;

        /** \brief retrieve the location that this object is situated at
         *  \return the location of the object
         */
        const TerrainBase::Loc & getPosition() const;

        /** \brief retrieve the most recent information about this object
         *  \return a GameObj if this Record is active, NULL otherwise
         *  \sa isActive
         */
        GameObj * getObject() const;
    };

    /// The purpose of the HistoryModule is to provide a primitive but robust
    /// way of tracking certain types of static objects regardless of whether
    /// or not they are currently visible. When the module is first created, it
    /// needs a reference to the GameStateModule. You should also call
    /// trackBlueprint() with the blueprint names of each of the static object
    /// types you want the module to listen for. The only real requirement is
    /// that these objects should not move. Minerals and buildings are good
    /// candidates. There are two main ways to get information from the module.
    /// If you have an object, such as a mineral, that you wish to track for
    /// perpetuity, call getRecord() to obtain a Record pointer for the object.
    /// If you merely want a list of all known tracked objects, call 
    /// getActiveRecords() and provide it with a vector to fill out. The Record     /// pointers you obtain from either of these methods will remain valid for
    /// the entirety of the simulation.
    class HistoryModule : public EventHandler
    {
        GameStateModule&            gsm;
        std::set<std::string>       tracking;

        std::vector<Record*>        records;
        std::map<GameObj*, Record*> obj2rec;

        void handle_new(GameObj * object);
        void check_vis();
    public:
        HistoryModule(GameStateModule & gsm);
        ~HistoryModule();

        /** \brief to satisfy EventHandler requirements
         */
        bool handle_event(const Event &event);

        /** \brief create Record objects for instances of a specific blueprint
         *  \param the name of the blueprint to create Record objects for
         *  \sa isTracking
         */
        void trackBlueprint(const std::string & bp_name);

        /** \brief determine if a particular blueprint is being tracked
         *  \param the name of the blueprint to check whether we are tracking
         *  \return true if the module creates Records for this object type
         *  \sa trackBlueprint
         */
        bool isTracking(const std::string & bp_name) const;

        /** \brief retrieve the Record that tracks a particular object
         *  \param the object whose Record we wish to obtain
         *  \return the particular Record which tracks this object
         */
        const Record * getRecord(GameObj * object) const;

        /** \brief retrieve all Records for whom isActive() returns true
         *  \param a vector in which to insert the active Records
         *  \sa queryActiveRecords
         */
        void getActiveRecords(std::vector<const Record*> & rec) const;

        /** \brief query the set of Records for whom isActive() returns true
         *  \param query the test an object must passed to be in the result set
         *  \param result a vector in which to insert the results of the query
         *  \sa getActiveRecords
         */
        void queryActiveRecords(const Watcher::UnitPredPtr & query,
                                std::vector<const Record*> & result) const;
    };
}

REGISTER_TYPEOF(891001, std::vector<const History::Record*>::iterator);
REGISTER_TYPEOF(891002, std::vector<const History::Record*>::const_iterator);
REGISTER_TYPEOF(891003, std::set<const History::Record *>::iterator);
REGISTER_TYPEOF(891004, std::set<const History::Record *>::const_iterator);

#endif
