// $Id: InfluenceMap.C 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file (c) Tim Furtak licensed under the GPL

#include "Global.H"
#include "InfluenceMap.H"
#include "DrawOnTerrain.H"
#include "Game.H"
#include "Profiler.H"

using namespace std;

//===================================================================

InfluenceMap::InfluenceMap()
{
  points_x = points_y = 0;
  cell_size = 0;
  cells_x = cells_y = cell_num = 0;
}

//-------------------------------------------------------------------

InfluenceMap::InfluenceMap(sint4 px, sint4 py, sint4 cell_size)
{
  init(px, py, cell_size);
}

//-------------------------------------------------------------------

InfluenceMap::~InfluenceMap()
{
}

//===================================================================

void InfluenceMap::init(sint4 px, sint4 py, sint4 cs)
{
  assert(px >= 0 && py >= 0);
  assert(cs >= 1);

  points_x = px;
  points_y = py;
  cell_size = cs;

  cells_x = (px + cell_size-1) / cell_size;
  cells_y = (py + cell_size-1) / cell_size;
  cell_num = cells_x * cells_y;
  
  data.resize(cell_num);
  clear();
}

//===================================================================

void InfluenceMap::clear()
{
  FORU (i, cell_num) data[i] = 0.0;
}

//===================================================================

/** Drop a Gaussian centered on (cx, cy) with decay s, and max radius r. */

void InfluenceMap::add_gaussian(real4 cx, real4 cy, real4 s, real4 r, real4 val)
{
  sint4 min_x = ::max(sint4(floor(cx - r)), 0);
  sint4 max_x = ::min(sint4( ceil(cx + r)), points_x-1);
  sint4 min_y = ::max(sint4(floor(cy - r)), 0);
  sint4 max_y = ::min(sint4( ceil(cy + r)), points_y-1);

  min_x /= cell_size;
  max_x /= cell_size;
  min_y /= cell_size;
  max_y /= cell_size;

  cx /= cell_size;
  cy /= cell_size;
  r /= cell_size;

  real4 s2 = s*s;
  const real4 r2pi = 2.506628274;

  for (sint4 y = min_y; y <= max_y; ++y) {
    real4 ty = y+0.5;
    for (sint4 x = min_x; x <= max_x; ++x) {
      real4 tx = x+0.5;
      real4 d2 = square(cx - tx) + square(cy - ty);
      real4 d = sqrt(d2);

      real4 v = val / (s*r2pi) * expf(-d2 / (2*s2));

      d -= r;
      if (d <= 0.0) {
        get_cell(x,y) += v;
      } else if (d <= 1.0) {
        get_cell(x,y) += v * (1.0-d);
      }
    }
  }
}

//-------------------------------------------------------------------

void InfluenceMap::add_circle(real4 cx, real4 cy, real4 r, real4 val)
{
  sint4 min_x = ::max(sint4(floor(cx - r)), 0);
  sint4 max_x = ::min(sint4( ceil(cx + r)), points_x-1);
  sint4 min_y = ::max(sint4(floor(cy - r)), 0);
  sint4 max_y = ::min(sint4( ceil(cy + r)), points_y-1);

  min_x /= cell_size;
  max_x /= cell_size;
  min_y /= cell_size;
  max_y /= cell_size;

  cx /= cell_size;
  cy /= cell_size;
  r /= cell_size;

  real4 r2 = r*r;

  for (sint4 y = min_y; y <= max_y; ++y) {
    for (sint4 x = min_x; x <= max_x; ++x) {
      real4 d2 = square(cx - x) + square(cy - y);
      if (d2 <= r2) get_cell(x,y) += val; //data[xy2ind(x, y)] += val;
    }
  }
}

//-------------------------------------------------------------------

/** Not intended to be accurately anti-aliased. */

void InfluenceMap::add_aacircle(real4 cx, real4 cy, real4 r, real4 val)
{
  sint4 min_x = ::max(sint4(floor(cx - r)), 0);
  sint4 max_x = ::min(sint4( ceil(cx + r)), points_x-1);
  sint4 min_y = ::max(sint4(floor(cy - r)), 0);
  sint4 max_y = ::min(sint4( ceil(cy + r)), points_y-1);

  min_x /= cell_size;
  max_x /= cell_size;
  min_y /= cell_size;
  max_y /= cell_size;

  cx /= cell_size;
  cy /= cell_size;
  r /= cell_size;

  for (sint4 y = min_y; y <= max_y; ++y) {
    real4 ty = (y+0.5);
    for (sint4 x = min_x; x <= max_x; ++x) {
      real4 tx = (x+0.5);
      real4 d = sqrt(square(cx - tx) + square(cy - ty));
      d -= r;
      //      real4 d = d2 - r2;
      if (d <= 0.0) {
        get_cell(x,y) += val;
      } else if (d <= 1.0) {
        get_cell(x,y) += val * (1.0-d);
      }
    }
  }
}

//-------------------------------------------------------------------

void InfluenceMap::invert()
{
  FORU (i, cell_num) if (data[i] != 0.0) data[i] = 1.0 / data[i];
}

//===================================================================

void InfluenceMap::draw(DrawOnTerrain &dot, const Game &game, const Vec3<real4> &color) const
{
  sint4 tile_points = game.get_tile_points();

  // scaling factors to convert between cells and original points
  real4 dx = 1.0 / tile_points * cell_size;
  real4 dy = 1.0 / tile_points * cell_size;

  {
    PROFILE("draw cells");

    dot.begin_squares();

    FORS (yi, cells_y) {
      FORS (xi, cells_x) {
        real4 v = get_cell(xi, yi);

        if (v > 0.0) {

          if (v > 1.0) v = 1.0;
          Vec3<real4> c = color * v;

          real4 x1 = (xi) * dx;
          real4 x2 = (xi+1.0) * dx;
          real4 y1 = (yi) * dy;
          real4 y2 = (yi+1.0) * dy;
          
          dot.draw_square(x1, y1, x2, y2, c);
        }
      }
    }
    dot.end_squares();
  }
}

//===================================================================
