#ifndef VectorField_H
#define VectorField_H

// $Id: VectorField.H 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "LinAlg.H"

class DrawOnTerrain;
class Game;

//===================================================================

class VectorField
{
public:

  //==============================

  class Emitter {
  protected:
    real4 x, y;
  public:
    Emitter(real4 x_, real4 y_) : x(x_), y(y_) {}
    virtual ~Emitter() {}
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const =0;
  };

  //==============================

  class PointEmitter : public Emitter {
  protected:
    real4 radius, radius_2;
    real4 mag;
  public:
    PointEmitter(real4 x, real4 y, real4 radius, real4 mag);
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //------------------------------

  /** A truncated cone with radial direction. */

  class PointEmitter2 : public Emitter {
  protected:
    real4 r1, r2;
    real4 mag;
  public:
    PointEmitter2(real4 x, real4 y, real4 r1, real4 r2, real4 mag);
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //------------------------------

  //  r2    _______
  //     _--       --_
  //  __/   r1        \_
  //  |                 |

  /** Gaussian decay starting from an inner radius and cut off at an outer limit. */

  class GaussianEmitter : public PointEmitter2 {
  protected:
    real4 decay;
  public:
    GaussianEmitter(real4 x, real4 y, real4 r1, real4 r2, real4 mag, real4 decay);
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //==============================

  class OrthoEmitter : public Emitter {
  protected:
    real4 r1, r2;
    real4 mag;
  public:
    OrthoEmitter(real4 x, real4 y, real4 r1, real4 r2, real4 mag);
    Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //==============================
  // Modifiers
  //==============================

  /** Convolves another emitter by a truncated cone of height 1. */

  class RadialDecay : public Emitter {
  protected:
    const Emitter *em;
    real4 r1; ///< 1 until this radius;
    real4 r2; ///< Linear decay from 1 to 0 from r1 to r2;
  public:
    RadialDecay(real4 x, real4 y, real4 dx, real4 dy, const Emitter *em);
    ~RadialDecay();
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //------------------------------

  /** Restricts the field to one direction. */

  class DirMultiplier : public Emitter {
  protected:
    const Emitter *em;
    Vec2<real4> dir;
  public:
    DirMultiplier(real4 x, real4 y, real4 dx, real4 dy, const Emitter *em);
    ~DirMultiplier();
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //==============================

protected:
  Vector<Emitter*> emitters;

public:
  VectorField();
  ~VectorField();
  
  Vec2<real4> field(real4 x, real4 y, real4 dx=0, real4 dy=0) const;
  
  void add_emitter(Emitter *em);
  void clear();

  void draw(DrawOnTerrain &dot, const Game &game, const Vec3<real4> &color) const;
};

REGISTER_TYPEOF(2801, Vector<VectorField::Emitter*>::iterator);
REGISTER_TYPEOF(2802, Vector<VectorField::Emitter*>::const_iterator);

//===================================================================

#endif
