#include "Finance.H"

#include "GameStateModule.H"
#include "GameChanges.H"
#include "GameObjFactory.H"
#include "Blueprints.H"
#include "Game.H"
#include "History.H"
#include "WatcherAux.H"

REGISTER_TYPEOF(2842001, std::set<ScriptObj*>::iterator);
REGISTER_TYPEOF(2842002, std::set<ScriptObj*>::const_iterator);

namespace Finance
{
    Ticket::Ticket(Module & fm, uint4 value) :
        fm(fm), value(value)
    {

    }

    Ticket::~Ticket()
    {

    }

    uint4 Ticket::available() const
    {
        // Return the amount currently available
        return value;
    }

    Ticket * Ticket::reserve(uint4 amount)
    {
        // If we do not have this amount available
        if(amount > value)
        {
            // Return nothing
            return NULL;
        }
        else
        {
            // Reserve the amount and return a new ticket
            value -= amount;

            return new Ticket(fm, amount);
        }
    }

    void Ticket::trim(uint4 amount)
    {
        if(amount > value)
        {
            return;
        }
        else
        {
            uint4 released = value - amount;
            value = amount;
            fm.totalReserved -= released;
        }
    }

    void Ticket::spend()
    {
        // Spend the amount stored in this ticket
        fm.totalReserved -= value;
        fm.totalSpent    += value;
        value = 0;

        // Delete this ticket
        delete this;
    }

    void Ticket::release()
    {
        // Release the amount stored in this ticket
        fm.totalReserved -= value;
        value = 0;

        // Delete this ticket
        delete this;
    }

    Hold::Hold(Module & fm, uint4 value) : fm(fm), value(value)
    {
        
    }
    
    Hold::~Hold()
    {

    }

    bool Hold::ready() const
    {
        // Determine the amount of unreserved resources
        uint4 currentCosts = fm.totalReserved + fm.totalSpent;
        if(currentCosts > fm.totalResources)
        {
            return false;
        }
        uint4 currentResources = fm.totalResources - currentCosts;

        // We will allocate resources to holds in a first
        // come first served fashion
        std::list<Hold*>::iterator it;
        for(it = fm.holds.begin(); it != fm.holds.end(); it++)
        {
            // Determine the amount of the next hold
            uint4 amount = (*it)->value;
            
            // If we cannot satisfy it, fail
            if( amount > currentResources )
            {
                std::cerr << "Hold unsatisified: " << currentResources;
                std::cerr << " / " << amount << std::endl;
                return false;
            }
            currentResources -= amount;

            // If we have reached and our own hold, succeed
            if( (*it) == this )
            {
                return true;
            }
        }

        assert( false && "Hold not in queue" );
        return false;
    }

    void Hold::cancel()
    {
        fm.holds.erase( std::find(fm.holds.begin(), fm.holds.end(), this) );
        fm.totalHeld -= value;
        delete this;
    }

    Ticket * Hold::ticket()
    {
        assert( ready() && "Hold not yet satisfiable" );

        fm.totalReserved += value;
        Ticket * ticket = new Ticket(fm, value);

        fm.holds.erase( std::find(fm.holds.begin(), fm.holds.end(), this) );
        fm.totalHeld -= value;
        delete this;

        return ticket;
    }



    Module::Module(GameStateModule & gsm, History::HistoryModule & hm) : 
        gsm(gsm),
        hm(hm),
        totalResources(0),
        totalHeld(0),
        totalReserved(0),
        totalSpent(0),
        updateMapCounter(0),
        maxMapPatches(0),
        maxMapMinerals(0),
        alliedLiveUnitWorth(0),
        alliedDeadUnitWorth(0),
        enemyDeadUnitWorth(0)
    {
        gsm.add_handler(this);
    }

    Module::~Module()
    {
        gsm.remove_handler(this);
    }

    uint4 Module::available() const
    {
        // Calculate the amount of resources available
        uint4 totalCosts = totalReserved + totalSpent + totalHeld;
        if(totalCosts > totalResources)
        {
            return 0;
        }
        else
        {
            return (totalResources - totalCosts);
        }
    }
    
    Ticket * Module::reserve(uint4 amount)
    {
        // If we do not have this amount available
        if( amount > available() )
        {
            // Return nothing
            return NULL;
        }
        else
        {
            // Reserve the requested amount and return a new ticket
            totalReserved += amount;

            return new Ticket(*this, amount);
        }
    }

    Hold * Module::hold(uint4 amount)
    {
        holds.push_back(new Hold(*this, amount));
        totalHeld += amount;
        return holds.back();
    }

    bool Module::handle_event(const Event & event)
    {
        if( event.get_what() == GameStateModule::VIEW_MSG )
        {
            // Obtain the new amount for total resources
            PlayerInfo& pi = gsm.get_game().get_cplayer_info();
	    ScriptObj * so = pi.global_obj("player");
   	    totalResources = so->get_int("minerals");

            // Amount spent should have been taken off by now
            totalSpent     = 0;

            sint4 me = gsm.get_game().get_client_player();

            const GameChanges & gc = gsm.get_changes();
            FORALL(gc.new_objs, o)
            {
                ScriptObj * obj = (*o);
                if(obj->has_attr("cost_mineral"))
                {
                    if(obj->get_int("owner") == me)
                    {
                        sint4 cost = obj->get_int("cost_mineral");
                        alliedLiveUnitWorth += cost;

                        if(obj->bp_name() == "worker")
                        {
                            alliedWorkers.insert(obj);
                        }
                    }
                }
                //                std::cerr << std::endl;
            }

            FORALL(gc.dead_objs, o)
            {
                ScriptObj * obj = (*o);
                if(obj->has_attr("cost_mineral"))
                {
                    sint4 cost = obj->get_int("cost_mineral");
                    sint4 owner = obj->get_int("owner");
                    if(owner == me)
                    {
                        alliedLiveUnitWorth -= cost;
                        alliedDeadUnitWorth += cost;

                        if(obj->bp_name() == "worker")
                        {
                            alliedWorkers.erase(obj);
                        }
                    }
                    else if(owner != 2)
                    {
                        enemyDeadUnitWorth += cost;
                    }
                }
            }

            updateMapCounter++;
            if(updateMapCounter >= 32)
            {
                updateMapCounter = 0;

                // Determine the maximum and current amount of minerals
                // that we know about.
                std::vector<const History::Record*> results;
                hm.queryActiveRecords(Watcher::TypeIs("mineral"), results);
                if(results.size() > maxMapPatches)
                {
                    maxMapPatches = results.size();
                    maxMapMinerals = maxMapPatches * 1000; // Not generic?
                }

                curMapMinerals = 0;
                FORALL(results, r)
                {
                    GameObj * mineral = (*r)->getObject();
                    curMapMinerals += mineral->get_int("hp");
                }

                alliedCurrentMining = 0;
                FORALL(alliedWorkers, w)
                {
                    ScriptObj * worker = (*w);
                    alliedCurrentMining += worker->get_int("minerals");
                }
            }

            return true;
        }

        return false;
    }

    uint4 Module::get_cost(const std::string & bp_name) const
    {
        const Game & game          = gsm.get_game();
        const PlayerInfo & player  = game.get_cplayer_info();
        GameObjFactory & gof       = const_cast<GameObjFactory&>(player.gof);
        GameObjBlueprint * bp      = gof.get_blueprint(bp_name);
        if(bp == NULL) { return 0; }
        
        sint4 cost_id = bp->attr_index("cost_mineral");
        if(cost_id == -1) { return 0; }

        return bp->attr_vals[cost_id].to_int();
    }

    uint4 Module::getTotalMined() const
    {
        return maxMapMinerals - curMapMinerals;
    }

    uint4 Module::getTotalWorth() const
    {
        uint4 startingWorth = 3000; // 600 in cc, 300 in workers, 600 res (x2)
        uint4 worthToDate = getTotalMined() + startingWorth;
        uint4 totalDead = alliedDeadUnitWorth + enemyDeadUnitWorth;
        
        if(totalDead > worthToDate)
        {
            return 0;
        }
        else
        {
            return worthToDate - totalDead;
        }
    }

    uint4 Module::getAlliedWorth() const
    {
        return alliedLiveUnitWorth + totalResources + alliedCurrentMining;
    }

    uint4 Module::getEnemyWorth() const
    {
        if(getAlliedWorth() > getTotalWorth())
        {
            return 0;
        }
        else
        {
            return getTotalWorth() - getAlliedWorth();
        }
    }
}
