// $Id: PursueAttack.C 5279 2007-06-23 02:49:08Z mburo $
// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "PursueAttack.H"
#include "GameStateModule.H"
#include "Game.H"
#include "Distance.H"
#include "TerrainModule.H"
#include "TerrainBase.H"

using namespace std;

const sint4 PursueAttack::FROM = EventFactory::new_who();
const sint4 PursueAttack::PURSUE_START = EventFactory::new_what();
const sint4 PursueAttack::PURSUE_STOP = EventFactory::new_what();
const string PursueAttack::bp_name = "pursue_attack";
const sint4 PursueAttack::FAILURE_CUTOFF = 5;
const sint4 PursueAttack::ATTACK_PRIORITY = 3;

//===================================================================
PursueAttack::PursueAttack(GameStateModule &gsm_) : gsm(gsm_)
{
}
//===================================================================
PursueAttack::~PursueAttack()
{
}
//===================================================================
void PursueAttack::act(GameStateModule& /*gsm*/)
{
  FORALL (obj2dat, it){
    PursueData &pd = it->second;
    //const GameObj* targ = pd.targ;

    GameObj::ZCat zcat = pd.targ->get_zcat(0);
    sint4 range = get_range(it->first, zcat);
    real8 rdist = Distance::distance(*(it->first), *(pd.targ));
    sint4 sdist = (sint4)floor(rdist);
    if (sdist <= range && pd.state == PursueData::MOVE_TO){
      pd.state = PursueData::ATTACK;
      need_act.push_back(it->first);
    }
    else if (sdist > range && pd.state == PursueData::ATTACK){
      pd.state = PursueData::MOVE_TO;
      need_act.push_back(it->first);
    }
  }
  FORALL (need_act, i){
    FIND (obj2dat, it, *i);
    if (it == obj2dat.end()) continue;

    PursueData &pd = it->second;
    switch(pd.state) {
    case PursueData::MOVE_TO:
      {
	TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::VICINITY);
	task.objs.insert(*i);
	sint4 range = get_range(*i, pd.targ->get_zcat(0));
	task.goal.obj = pd.targ;
	if (range <= 12){
	  task.goal.mode = TerrainBase::Goal::TOUCH;
	}
	else{
	  task.goal.distance = range - 4; //move in slightly closer than the range
	}
	
	call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, task));
	continue;
      }
    case PursueData::ATTACK:
      {
	ScriptObj *weapon = (*i)->component("weapon");
	if (weapon){
	  const sintptr *shooting = weapon->get_int_ptr("shooting");
	  if (shooting && *shooting) continue; //assume that it is shooting at the right target
	  Vector<sint4> args;
	  args.push_back(gsm.get_game().get_cplayer_info().get_id(pd.targ));
	  weapon->set_action("attack", args, ATTACK_PRIORITY);
	}
      }
    }
  }
  need_act.clear();
}
//===================================================================
sint4 PursueAttack::get_range(GameObj *obj, GameObj::ZCat zcat){
  ScriptObj *weapon = obj->component("weapon");
  sint4 range;
  if (weapon){
    if (zcat == Object::ON_LAND){
      range = weapon->get_val_ptr("max_ground_range")->to_int();
	return range;
    }
    else if(zcat == Object::IN_AIR){
      range = weapon->get_val_ptr("max_air_range")->to_int();
	  return range;
    }
  }
  else{
    std::cerr << "Warning! Unit has no weapon, so it will not attack\n";
  }
  return 0;

}
//===================================================================
bool PursueAttack::handle_event(const Event &e){
  if (e.get_what() == TerrainModule::PATH_SUCCESS) {
    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
   
    assert(p && "SUCCESS: not a PathStatusEvent");
    FIND (obj2dat, it, p->get_obj());
    if (it == obj2dat.end()) return false;

    PursueData &pd = it->second;
    pd.move_failure_count = 0;
    if (pd.state == PursueData::MOVE_TO) //The object has successfully travelled the path but is not yet in range
      need_act.push_back(p->get_obj()); //Move again
    return true;
  }
  if (e.get_what() == TerrainModule::PATH_FAILURE) {
   
    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "FAILURE: not a PathStatusEvent");
    GameObj* obj = p->get_obj();
    FIND(obj2dat, it, obj);
    if (it == obj2dat.end()) return false;

    if (p->get_details() == TerrainModule::PATH_PLAN_FAILURE){
      del(obj); //give up control of this unit;
      cout << "PursueAttack got path plan failure\n";
    }
    if (p->get_details() == TerrainModule::PATH_MOVE_FAILURE){
      cout << "PursueAttack got path move failure\n";
      if (it->second.move_failure_count >= FAILURE_CUTOFF){
	del(obj);
      }
      else{
	if (it->second.state == PursueData::MOVE_TO){
	  need_act.push_back(obj);
	  it->second.move_failure_count++;
	}
	//if not in MOVE_TO mode, then the failure message does not matter
      }
    }
    return true;
  }

  if (e.get_what() == GameStateModule::VIEW_MSG) {
    update(gsm.get_changes());
    act(gsm);
    return true;
  }
  if (e.get_what() == PursueAttack::PURSUE_START) {
    REM("PURSUE EVENT");
    const PursueEvent &pe = static_cast<const PursueEvent&>(e);
    if (pe.pursuer && pe.targ) {
      add(pe.pursuer, pe.targ);
    }
    return true;
  }

  if (e.get_what() == PursueAttack::PURSUE_STOP){
    REM("STOP PURSUE EVENT");
    const PursueEvent &pe = static_cast<const PursueEvent&>(e);
    del(pe.pursuer);
  }
  return false;
} 

//===================================================================
void PursueAttack::update(const GameChanges &changes){

  // remove dead objects
  FORALL (changes.dead_objs, i) {
    GameObj *gob = (*i)->get_GameObj();
    del(gob);
    FORALL(obj2dat, it){
      if (it->second.targ == gob)
	del(it->first);
    }
  }

  // remove vanished objects
  FORALL (changes.vanished_objs, i) {
    GameObj *gob = (*i)->get_GameObj();
    del(gob);
    FORALL(obj2dat, it){
      if (it->second.targ == gob)
	del(it->first);
    }
  }

}

//===================================================================
void PursueAttack::add(GameObj* pursuer, GameObj* targ){
  assert(*(pursuer->sod.owner) == gsm.get_game().get_client_player());
  assert(pursuer->get_max_speed() > 0);
  PursueData pd;
  pd.targ = targ;
  pd.state = PursueData::MOVE_TO;
  obj2dat[pursuer] = pd;
  need_act.push_back(pursuer);
}
//===================================================================
void PursueAttack::del(GameObj* pursuer){
  obj2dat.erase(pursuer);
}

//===================================================================
