
#include "Global.H"
#include "FactorPath.H"
#include "ST_PFEngine.H"
#include "Options.H"
#include "Game.H"
#include "GameObj.H"
#include "GameStateModule.H"
#include "GameChanges.H"
#include "LinAlg.H"

#include <utility>
#include <iterator>
#include <math.h>

REGISTER_TYPEOF(102601, Movement::FactorPath::Obj2LocMap::iterator);
REGISTER_TYPEOF(102602, Movement::FactorPath::Obj2LocMap::const_iterator);
REGISTER_TYPEOF(102603, Movement::FactorPath::SegmentEndsMap::iterator);
REGISTER_TYPEOF(102604, Movement::FactorPath::SegmentEndsMap::const_iterator);
REGISTER_TYPEOF(102605, Movement::FactorPath::TaskSet::iterator);
REGISTER_TYPEOF(102606, Movement::FactorPath::TaskSet::const_iterator);
REGISTER_TYPEOF(102609, Movement::FactorPath::TaskQueue::iterator);
REGISTER_TYPEOF(102610, Movement::FactorPath::TaskQueue::const_iterator);
REGISTER_TYPEOF(102615, Vector<Movement::FactorTask::ptr>::iterator);
REGISTER_TYPEOF(102616, Vector<Movement::FactorTask::ptr>::const_iterator);
REGISTER_TYPEOF(102617, std::set<Movement::Pathfinder::ptr>::iterator);
REGISTER_TYPEOF(102618, std::set<Movement::Pathfinder::ptr>::const_iterator);
REGISTER_TYPEOF(102619, Vector<Movement::MoveCommand>::iterator);
REGISTER_TYPEOF(102620, Vector<Movement::MoveCommand>::const_iterator);

using namespace std;

#define TEST(x)
//#define TEST(x) x

namespace Movement 
{
	Module::ptr MakeModule(GameStateModule & gsm, sint4 max_ms_per_tick)
	{
		FactorPath::ptr factorPath( new FactorPath(gsm, max_ms_per_tick) );
		factorPath->setSelfPointer( FactorPath::wptr( factorPath ) );
		return Module::ptr(factorPath);
	}

	void FactorPath::add_options()
	{
		::Options o("SimpleTerrain - ForceField");
		o.put("-costonly", "represent boundaries only by higher edge cost");
		o.put("-highres", "use a high resolution world map (slower but more accurate)");
		o.put("-nopush", "turns off pushing");
		o.put("-wpattract", "units are attracted to wps, and not forced to stay on the path");
		o.put("-nostoptoplan", "units do not stop to plan tasks");
	}
   
	FactorPath::FactorPath(GameStateModule & gsm, sint4 max_ms_per_tick) : gsm(gsm), max_ms_per_tick(max_ms_per_tick)
	{
		gsm.add_handler(this);

		sint4 seed = 0; 
		::Options::get("-seed", seed); 
		rand.srand(seed);
 
		::Options::get("-nopush", pushing_on);
		pushing_on = !pushing_on;

		::Options::get("-wpattract", wp_attract);
		::Options::get("-nostoptoplan", stop_to_plan);
		stop_to_plan = !stop_to_plan;
	}

	FactorPath::~FactorPath()
	{
		gsm.remove_handler(this);
	}

	bool FactorPath::handle_event(const Event &e)
	{
		// fixme: add parameters to terrainmodule functions dealing with lag
		Game &game = gsm.get_game();
		sint4 vf = game.get_view_frame();
		sint4 ai = game.get_action_frame();
		bool behind = abs(vf-ai) > 10 || ai < 0;

		if (e.get_who() == GameStateModule::FROM && 
			e.get_what() == GameStateModule::VIEW_MSG) 
		{
			static sint4 first = 1; // TODO: Fix me

			// If this is the first view, initialize the TerrainBase implementation
			if (first)
			{
				first = 0;
				if( max_ms_per_tick == 0 )
				{    
					max_ms_per_tick = (uint4)(1000/gsm.get_game().get_freq() - 20);  // milliseconds
					cout << "MAXTIME " << max_ms_per_tick << endl;
				}
				init(game.get_map().get_width(),
					 game.get_map().get_height(),
					 game.get_tile_points(),
					 game.get_client_player(),
					 game.get_player_num()); //The total number of players is also the id of the neutral player.
			}

			update(gsm.get_changes());
			if( !behind ) 
			{
				plan_tasks(max_ms_per_tick);
				act();
			}
			return true;
		}
		return false;
	}

	void FactorPath::update(const GameChanges &changes)
	{
		// boundaries
		Vector<TerrainBase::Segment> segments;      
		FORALL (changes.new_boundaries, it) 
		{
			GameObj *gob = (*it)->get_GameObj();
			if (gob == 0) continue;
			segments.push_back(TerrainBase::Segment(
			TerrainBase::Loc(*gob->sod.x1, *gob->sod.y1),
			TerrainBase::Loc(*gob->sod.x2, *gob->sod.y2) ));
		} 

		if (!segments.empty()) add_segments(segments);

		// vanished objects
		FORALL (changes.vanished_objs, i) 
		{
			GameObj* p = (*i)->get_GameObj();
			if (p && p->has_attr("x") && p->has_attr("y") && p->has_attr("shape"))
			{
				// TODO: what about immobile objects getting out of sight?
				// remove or not?
				remove_obj(p);
			}
		}

		// dead objects
		FORALL (changes.dead_objs, i) 
		{
			GameObj* p = (*i)->get_GameObj();
			if (p && p->has_attr("x") && p->has_attr("y") && p->has_attr("shape"))
			{
				remove_obj(p);
			}
		}

		// new objects
		FORALL (changes.new_objs, i) 
		{
			GameObj *p = (*i)->get_GameObj();
			if (p && p->has_attr("x") && p->has_attr("y") && p->has_attr("shape"))
			{
				add_obj(p);
			}
		}

		// changed objects
		FORALL (changes.changed_objs, i) 
		{
			GameObj* p = (*i)->get_GameObj();
			if (!p) continue;
			if (p->attr_changed("x") || p->attr_changed("y") ||
				p->attr_changed("is_moving") || p->attr_changed("zcat") ||
				p->attr_changed("speed")) 
			{
				update_obj(p);
			}
		}

		FORALL( pathfinders, pf )
		{
			Pathfinder & pathfinder(**pf);
			pathfinder.updateData();
		}
	}

	void FactorPath::act(void)
	{
		// get the actions from the terrain implementation
		Vector<MoveCommand> moves;
		execute_tasks(moves);

		// execute the moves
		for(size_t i = 0; i < moves.size(); i++) 
		{
                    if (*moves[i].unit->sod.is_moving == 0) {
                        sint4 x = *moves[i].unit->sod.x;
                        sint4 y = *moves[i].unit->sod.y;
                        if (x == moves[i].next_loc.x &&
                            y == moves[i].next_loc.y) continue;
                    }

			Vector<sint4> params;
			params.push_back(moves[i].next_loc.x);
			params.push_back(moves[i].next_loc.y);
			params.push_back(max((sint4)0,moves[i].speed));
			moves[i].unit->set_action("move", params, 1); // TODO: Replace 1 with a MOVE_PRIORITY constant
                        /*
                        cout << "trying to move "
                             << " " << (*moves[i].unit->sod.x)
                             << " " << (*moves[i].unit->sod.y)
                             << " : " << params[0]
                             << " " << params[1] << endl;
                        */
		}
	}

	void FactorPath::addPathfinder(const std::string & name, const Pathfinder::ptr & pathfinder)
	{
		pathfinders.insert( pathfinder );
		name2pathfinder.insert( std::make_pair( name, pathfinder ) );
	}

	void FactorPath::addPathExecutor(const std::string & name, const PathExecutor::ptr & pathExecutor)
	{
		name2pathExec.insert( std::make_pair( name, pathExecutor ) );
	}

	Movement::Path::ptr FactorPath::findPath(const std::string & pathfinder, const Vec2D & start, sint4 radius, Movement::Goal::const_ptr goal, uint4 stopTime) const
	{
		std::map<std::string, Pathfinder::ptr>::const_iterator it;
		it = name2pathfinder.find(pathfinder);
		assert(it != name2pathfinder.end());
		return (it->second)->findPath(start, radius, goal, stopTime);
	}

	Movement::Task::ptr
	FactorPath::moveUnits(const std::string & pathfinder,
						  const std::string & pathExec,
						  const std::set<GameObj*> & units,
						  Movement::Goal::const_ptr goal,
						  const Movement::Task::StatusProc & proc)
	{
		FactorTask::ptr newTask(new FactorTask(self, units, goal,
											   name2pathfinder[pathfinder],
											   name2pathExec[pathExec], proc)); 
		newTask->setSelfPtr( boost::weak_ptr<FactorTask>(newTask) );
		
		tasks.insert(newTask);
		pendingTasks.push_back(newTask);

		FORALL(units, objIt) 
		{
			GameObj * obj(*objIt);
			clean_task_for_obj(obj);
			obj2task.insert( std::make_pair(obj, newTask) );
		}
		
		return Movement::Task::ptr(newTask);
	}

	Movement::Task::ptr
	FactorPath::moveUnitsViaPath(const std::string & pathExecutor,
								 const std::set<GameObj*> & units, 
								 Movement::Path::ptr path,
								 const Movement::Task::StatusProc & proc)
	{
		FactorTask::ptr newTask(new FactorTask(self, units, path,
											   name2pathExec[pathExecutor],
											   proc));
		newTask->setSelfPtr( boost::weak_ptr<FactorTask>(newTask) );

		tasks.insert(newTask);

		FORALL(units, objIt) 
		{
			GameObj * obj(*objIt);
			clean_task_for_obj(obj);
			obj2task.insert( std::make_pair(obj, newTask) );
		}
  
		return Movement::Task::ptr(newTask);
	}

	//-----------------------------------------------------------------------------
	void FactorPath::init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_)
	{
		client_pID = client_pID_;

		//Add the boundaries of the map to the list of segments for the forcefields  
		Loc topLeft(0,0);
		Loc bottomRight(tiles_x_ * tile_points_, tiles_y_ * tile_points_);
		Loc bottomLeft(topLeft.x, bottomRight.y);
		Loc topRight(bottomRight.x, topLeft.y);

		segments.push_back(Segment(topLeft, topRight));
		segments.push_back(Segment(topLeft, bottomLeft));
		segments.push_back(Segment(bottomLeft, bottomRight));
		segments.push_back(Segment(topRight, bottomRight));
	    
		segment_ends[topLeft] = true;
		segment_ends[bottomRight] = true;
		segment_ends[topRight] = true;
		segment_ends[bottomLeft] = true;

		FORALL( pathfinders, pf )
		{
			Pathfinder & pathfinder(**pf);
			pathfinder.initialize(tiles_x_, tiles_y_, tile_points_,
					      client_pID_, neutral_pID_);
		}
	}

	//-----------------------------------------------------------------------------
	void FactorPath::add_obj(GameObj * obj)
	{
		objects[obj] = Loc(-1,-1);

		FORALL( pathfinders, pf )
		{
			Pathfinder & pathfinder(**pf);
			pathfinder.addObject(obj);
		}
	}

	//-----------------------------------------------------------------------------
	void FactorPath::update_obj(GameObj * obj)
	{
		FORALL( pathfinders, pf )
		{
			Pathfinder & pathfinder(**pf);
			pathfinder.updateObject(obj);
		}
	}

	//-----------------------------------------------------------------------------
	void FactorPath::remove_obj(GameObj * obj)
	{
		FORALL( pathfinders, pf )
		{
			Pathfinder & pathfinder(**pf);
			pathfinder.removeObject(obj);
		}

		objsToStop.erase(obj);   // If we intended to stop it, we don't need to anymore
		clean_task_for_obj(obj); // remove a task (if any) associated with this object
		objects.erase(obj);
	}


	//-----------------------------------------------------------------------------
	void FactorPath::add_segments(const Vector<Segment> &segs)
	{
		FORALL(segs,i)
		{
			const Segment & seg(*i);
			segments.push_back(seg);
			segment_ends[seg.l1] = true;
			segment_ends[seg.l2] = true;

			FORALL( pathfinders, pf )
			{
				Pathfinder & pathfinder(**pf);
				pathfinder.addBoundary(Vec2D(seg.l1.x, seg.l1.y), Vec2D(seg.l2.x, seg.l2.y));
			}
		}
		invalidate_paths(segs);
	}

	//----------------------------------------------------------------------------------
	void FactorPath::cancel_task(GameObj *obj)
	{
		objsToStop.insert(obj);
		clean_task_for_obj(obj);
	}

	void FactorPath::remove_task(const FactorTask::ptr & task)
	{
		objsToStop.insert(task->getUnits().begin(), task->getUnits().end());
		clean_task(task);
	}
 
	//----------------------------------------------------------------------------------

	// fixme: we only check the path of the *center* of the unit vs the new boundaries,
	// and we should account for its size (radius).
	void FactorPath::invalidate_paths(const Vector<Segment> &segments)
	{
		FORALL(tasks, t)
		{
			FactorTask::ptr taskPtr(*t);
			FactorTask & task(*taskPtr);

			if( task.isValid() )
			{
				FORALL(segments, si)
				{
					if( task.isInvalidatedBySegment( (*si) ) )
					{
						// this path is invalid, mark as pending for (re)planning
						pendingTasks.push_back(taskPtr);
						task.setPending(true);
						cout << "Marked an invalid path" << endl;
						break;
					}
				}
			}
		}
	}

	void FactorPath::execute_tasks(Vector<MoveCommand> & cmds)
	{
		Vector<FactorTask::ptr> finishedTasks;

	    //clear the reserved locations in the object map
		FORALL(objects, it)
		{
			it->second = Loc(-1, -1);
		}
	
		// Execute all active tasks
        FORALL( tasks, t ) 
		{
			FactorTask::ptr taskPtr(*t);
			FactorTask & task(*taskPtr);

			// Report if task's goal is no longer valid
			if ( !task.getGoal()->isValid() )
			{
				task.setStatus(Movement::Task::PLAN_FAILURE);
				finishedTasks.push_back(taskPtr);
				continue;
			}

			// Report failure of planning (no path at all)
			if ( task.isNoPathFound() ) 
			{
				task.setStatus(Movement::Task::PLAN_FAILURE);
				finishedTasks.push_back(taskPtr);
				continue;
			}

			// Skip invalid tasks.
			if ( !task.isValid() ) 
			{
				continue;
			}

			Guide & guide = task.getGuide();
			if(guide.getUnits().empty())
			{
				continue;
			}

			if( !guide.isFinished() )
			{
				guide.updateTargets();
				FORALL( task.getUnits(), u )
				{
					GameObj * unit = (*u);
					bool unitAtGoal = guide.isUnitFinished(unit);
					if ( !unitAtGoal ) 
					{
						Vec2D target_vec = guide.getUnitTarget(unit);
						Loc target, attractor(target_vec.x, target_vec.y);
						get_forcefield_target(unit, attractor, target);
						add_move_command(unit, target, cmds);
					}
				}

				if(!guide.isFinished())
				{
					if(guide.isProgressing())
					{
						guide.updateState();
						task.resetCollisionCount();
					}
					else
					{
						if(task.shouldReplan())
						{
							task.setFailed(true);
							failedTasks.push_back(taskPtr);
						}
						else if(task.shouldFail())
						{
							task.setStatus(Movement::Task::MOVE_FAILURE);
							finishedTasks.push_back(taskPtr);
						}
						task.reportCollision();
					}
				}
			}
			else
			{
				task.setStatus(Movement::Task::SUCCESS);
				finishedTasks.push_back(taskPtr);

				const Movement::Goal & goal(*task.getGoal());

				// If we are moving to a location
				if(goal.getObject() == 0)
				{
					// Try to get as close as possible to the precise location
					Vec2D point = goal.getCenter();
					TerrainBase::Loc target(point.x, point.y);
					FORALL( task.getUnits(), u )
					{
						add_move_command( (*u), target, cmds );
					}
				}
			}
        }

		// Handle pushing
		if (pushing_on)
		{
			handle_pushing(cmds);
		}

		// Priority of commands:
		// 1. stops to avoid a collision.
		// 2. moves.
		// 3. other stops (resulting from a cancelled task, etc.).

		// 2 above 3:
		FORALL(cmds, c) 
		{
			objsToStop.erase(c->unit);
		}

		// 1 above 2:
		for(typeof(cmds.begin()) c = cmds.begin() ; c != cmds.end() ; ) 
		{
			FIND(objsToStop, s, c->unit);
			if( s != objsToStop.end() )
				c = cmds.erase(c);
			else
				c++;
		}

		// Return also the stop commands
		FORALL(objsToStop, o) 
		{
			GameObj * obj(*o);

			MoveCommand m;
			m.unit = obj;
			obj->get_center(m.next_loc.x, m.next_loc.y); // doesn't really matter
			m.speed = 0; // 0 -> stop
			cmds.push_back(m);
		}
		objsToStop.clear();

		// remove finished tasks
		FORALL (finishedTasks, t) 
		{
			remove_task((*t));
		}

		// call status procedures
		FORALL (finishedTasks, t)
		{
			FactorTask & task = (**t);
			task.callStatusProc();
		}
	}

	void FactorPath::handle_pushing(Vector<MoveCommand> &cmds)
	{
		FORALL(objects, objIt)
		{
			GameObj * gob(objIt->first);
			if (objIt->second == Loc(-1,-1) && *(gob->sod.owner) == client_pID && objIt->first->get_max_speed() > 0)
			{
				Loc target;
				get_push_target(gob, target);
				add_move_command(gob, target, cmds);
			}
		}
	}

	void FactorPath::add_move_command(GameObj * obj, Loc target, Vector<MoveCommand> &cmds)
	{
        if (abs(target.x) >= 500000) target.x = 499999;
        if( abs(target.y) >= 500000) target.y = 499999;

		Loc curr_loc; obj->get_center(curr_loc.x, curr_loc.y);
		Vec2D vec(target.x - curr_loc.x, target.y - curr_loc.y);
		real8 length = vec.length();
		sint4 speed = std::min(obj->get_max_speed(), static_cast<sint4>(length));

		MoveCommand m;
		m.unit = obj;
		m.next_loc = target;
		m.speed = speed;
		cmds.push_back(m);

		FIND (objects, it, obj);
		assert( it != objects.end() && "Object not in object map" );

		real8 k = static_cast<real8>(speed) / length;
	    vec = vec.multiply(k);
	    it->second.x = curr_loc.x + vec.x;
		it->second.y = curr_loc.y + vec.y;
	}

	sint4 FactorPath::what_team(const Object *obj)
	{
		const GameObj *p = dynamic_cast<const GameObj*>(obj);
		return (p) ? *(p->sod.owner) : -1;
	}

	void FactorPath::get_push_target(GameObj * obj, Loc & output)
	{
		if (obj->get_shape() != Object::CIRCLE)
		{
			obj->get_center(output.x, output.y);
			return; // only circle movement implemented -- this also functions as a crude check for buildings
		}

	    Vec2D force_vec(0,0);
		FORALL(objects, o)
		{
			GameObj * gob(o->first);
			if( obj->get_zcat(NULL) != gob->get_zcat(NULL) ) continue;
			
			if ( *gob->sod.owner != client_pID ) continue; // Only our units can push us
			const TerrainBase::Loc & reserved_pos(o->second);

			force_vec += ForceField::getObjectRepulsion(obj, gob, reserved_pos);
		}

		if(obj->get_zcat(NULL) != Object::IN_AIR) // For land and water units
		{
			Vec2D wall_repulsion;
			FORALL(segment_ends, ends_it)
			{
				ends_it->second = true;
			}
			FORALL(segments, seg_it)
			{
				Vec2D vec = ForceField::getLineRepulsion(obj, *seg_it);
				if (vec != Vec2D(0,0))
				{
					FIND(segment_ends, ends_it, seg_it->l1);
					if (ends_it != segment_ends.end())
						ends_it->second = false;
					FIND(segment_ends, ends_it2, seg_it->l2);
					if (ends_it != segment_ends.end())
						ends_it->second = false;
				}
				wall_repulsion += ForceField::getLineRepulsion(obj, *seg_it);
			}
			FORALL(segment_ends, endit)
			{
				if (endit->second)
				{
					wall_repulsion += ForceField::getPointRepulsion(obj, endit->first);
				}
			}
			force_vec += wall_repulsion;
		}

		if (force_vec != Vec2D(0,0))
		{
			Vec2D rand_vec;
			Vec2D half_force = Vec2D(force_vec.x/2, force_vec.y/2);

			uint4 rand_angle_int = rand.rand_uint4() % 62;
			real8 rand_angle = ((real8)rand_angle_int)/ 10;
			rand_vec.x = Vec2D::round(half_force.length() * cos(rand_angle));
			rand_vec.y = Vec2D::round(half_force.length() * sin(rand_angle));

			force_vec += rand_vec;
		}

		Loc object_center; obj->get_center(object_center.x, object_center.y);
		output.x = object_center.x + force_vec.x;
		output.y = object_center.y + force_vec.y;
	}

	void FactorPath::get_forcefield_target(GameObj * moving_obj, const Loc &attractor, Loc &output)
	{
		// TODO: Retrofit system for mobile buildings
		if(moving_obj->get_shape() != Object::CIRCLE)
		{
			output = attractor;
			return;
		}

	    Vec2D force_vec(0,0);
	    FORALL(objects, o)
		{
			const Object * static_obj(o->first);
			if(moving_obj->get_zcat(NULL) != static_obj->get_zcat()) continue;
			
			const TerrainBase::Loc & reserved_pos(o->second);
			force_vec += ForceField::getObjectRepulsion(moving_obj, static_obj, reserved_pos);
		}

		// For land and water objects
		if(moving_obj->get_zcat(NULL) != Object::IN_AIR) 
		{
			Vec2D wall_repulsion;
			FORALL(segment_ends, ends_it)
			{
				ends_it->second = true;
			}
			FORALL(segments, seg_it)
			{
				Vec2D vec;
				vec = ForceField::getLineRepulsion(moving_obj, *seg_it);
				if (vec != Vec2D(0,0))
				{
					FIND(segment_ends, ends_it, seg_it->l1);
					if (ends_it != segment_ends.end())
						ends_it->second = false;
					FIND(segment_ends, ends_it2, seg_it->l2);
					if (ends_it != segment_ends.end())
						ends_it->second = false;
				}
				wall_repulsion = wall_repulsion + ForceField::getLineRepulsion(moving_obj, *seg_it);
				// force_vec = force_vec + vec;
			}
			FORALL(segment_ends, endit)
			{
				if (endit->second)
				{
					wall_repulsion = wall_repulsion + ForceField::getPointRepulsion(moving_obj, endit->first);
				}
			}
			force_vec = force_vec + wall_repulsion;
		}

	    Vec2D attract_vec, scaled_attract_vec;
	    Loc obj_center;
	    moving_obj->get_center(obj_center.x, obj_center.y);
	    attract_vec = Vec2D(obj_center, attractor);

		real8 k;
		if (attract_vec.length() < 1)
		{
			k = ForceField::ATTRACTOR_FORCE;
	    }
		else
		{
			k = ForceField::ATTRACTOR_FORCE / attract_vec.length();
		}
		scaled_attract_vec = attract_vec.multiply(k);

		force_vec = force_vec + scaled_attract_vec;

		restrict_vector(attract_vec, force_vec);

		if (force_vec == Vec2D(0,0))
		{
			force_vec = Vec2D(attract_vec.y, -attract_vec.x);
		}

		Loc object_center;
		moving_obj->get_center(object_center.x, object_center.y);
	    output.x = object_center.x + force_vec.x;
		output.y = object_center.y + force_vec.y;
	}

	Loc find_point_on_line(const Loc & loc, const Loc & line_start, const Loc & line_end)
	{
		sint4 line_x = line_end.x - line_start.x;
		sint4 line_y = line_end.y - line_start.y;

		if (line_x == 0 && line_y == 0) return loc;

		sint4 v_x = loc.x - line_start.x;
		sint4 v_y = loc.y - line_start.y;

		sint4 a = line_x * v_x + line_y * v_y;
		sint4 b = line_x * line_x + line_y * line_y;

		return Loc(line_start.x + line_x * a / b, line_start.y + line_y * a / b);
	}



  void FactorPath::restrict_vector(const Vec2D &target_vec, Vec2D &vec){

	  real8 dot_prod = (real8)Vec2D::dot(vec, target_vec);
    if (dot_prod < 0){
      Vec2D perpendicular(target_vec.y, -target_vec.x);
      vec = vec.project_onto(perpendicular);
    }


  }

  


  

  //-------------------------------------------------------------------------

  bool firstPlan;

  bool FactorPath::plan_tasks(uint4 max_time)
  {
    sint4 stop_time = 0;
    if(max_time > 0){
      stop_time = (sint4)SDL_GetTicks() + max_time;
    }

    // cout << "Chance To Plan Task\n" ;

	firstPlan = true;
	
    // try to plan one failed and one pending tasks
    bool did_something = plan_queued_task(failedTasks, 0);
    bool did_some_more = plan_queued_task(pendingTasks, 0);
    sint4 counter = 0;
    //Continue planning until we have reached the max_time
    while( (sint4)SDL_GetTicks() < stop_time && (!pendingTasks.empty() || !failedTasks.empty() )) {
      if(counter++ % 2 == 0)
	did_something = plan_queued_task(failedTasks, stop_time);
      else
	did_some_more = plan_queued_task(pendingTasks, stop_time);


      did_some_more = did_something || did_some_more; // careful not to short-circuit!
    }



    did_something = did_some_more || did_something;
    return did_something;


  }	

  //-------------------------------------------------------------------
  struct FindFailed {
	  bool operator()(const FactorTask::ptr & task) { return task->isFailed() && !task->isNoPathFound(); };
  };

	//----------------------------------------------------------------------------------
	bool FactorPath::plan_queued_task(TaskQueue & queue, uint4 stop_time)
	{
		if (queue.empty()) return false;

		FactorTask::ptr taskPtr = queue.front(); queue.pop_front();
		assert( tasks.find(taskPtr) != tasks.end() && "Queued task not valid!" );

		FactorTask & task(*taskPtr);
		assert( !task.getUnits().empty() && "Invalid task: No objects in task!" );

		if (stop_to_plan)
		{
            bool moving = false;
            FORALL(task.getUnits(), it)
			{
				if ( (*it)->get_speed(NULL) > 0 || (*it)->get_moving(NULL) )
				{
					moving = true;
					break;
				}
			}
            if (moving) 
			{
				objsToStop.insert(task.getUnits().begin(), task.getUnits().end());
				queue.push_back(taskPtr);
				return false; // fixme: might be better to try another queud task already in this tick...
			}
		}

		TEST(cout << obj << " planning queued path" << endl);
		task.planTask(stop_time);

		/*if(firstPlan)
		{
			task.setFailed(false);
			task.setPending(true);
			pendingTasks.push_back(taskPtr);
			firstPlan = false;
		}*/

		return true;
	}

	bool Loc_less::operator() (Loc a, Loc b) const
	{
		if (a.x < b.x) return true;
        else return (a.x == b.x && a.y < b.y);
	}

	#define EPS 1e-9f

	void NullTaskStatusCallback(const Movement::Task &) {}

	void FactorPath::clean_task_for_obj(GameObj * obj)
	{
		Obj2Task::iterator o2t = obj2task.find(obj);
		if( o2t == obj2task.end() ) return;

		FactorTask::ptr taskPtr(o2t->second);
		FactorTask & task(*taskPtr);

		// if this is the only object of the task, erase the entire task
		if( task.getUnits().size() <= 1 ) 
		{
			clean_task(taskPtr);
		}
		// else, remove the object from the task's list of objects
		else 
		{
			task.removeUnit( obj );
			obj2task.erase(o2t);
		}
	}

	void FactorPath::clean_task(const FactorTask::ptr & task)
	{
		// erase the object -> task associations
		FORALL(task->getUnits(), u)
		{
			obj2task.erase( (*u) );
		}

		// remove the task from the pending queue, if it's there
		eraseFromQueue(pendingTasks, task);
 		eraseFromQueue(failedTasks, task);

		// remove the task
		tasks.erase(task);
	}

	void FactorPath::eraseFromQueue(TaskQueue & queue, const FactorTask::ptr & taskPtr)
	{
		TaskQueue::iterator it = std::find(queue.begin(), queue.end(), taskPtr);
		if( it != queue.end() ) queue.erase(it);
	}
} //end of namespace

