#ifndef FACTORPATH_H
#define FACTORPATH_H

// Based on the ST_ForceField class, the FactorPath class represents a core
// movement system based on task scheduling, forcefields for local collision
// avoidance, and pushing of friendly units. Details related to pathfinding
// and path execution have been abstracted.

#include "Movement.H"

#include "TerrainBase.H"
#include "GameObj.H"

#include "Random.H"
#include "Coord.H"
#include "SimpleTerrain.H"
#include "EventHandler.H"

#include "boost/weak_ptr.hpp"

REGISTER_TYPEOF(136003, std::set<const GameObj *>::iterator);
REGISTER_TYPEOF(136004, std::set<const GameObj *>::const_iterator);

class GameChanges;

namespace Movement
{
	typedef TerrainBase::Loc      Loc;
	typedef TerrainBase::Segment  Segment;

	/** Various utility methods for determining the strength of
		forcefields in various scenarios. */
	namespace ForceField
	{
		Vec2D getPointRepulsion(const Object * moving_obj,
								const TerrainBase::Loc & point,
								sint4 radius = 0);
		Vec2D getLineRepulsion(const Object * moving_obj,
							   const TerrainBase::Segment & seg);
		Vec2D getRectRepulsion(const Object * moving_obj,
							   const TerrainBase::Loc & topLeft,
							   const TerrainBase::Loc & bottomRight);
		Vec2D getObjectRepulsion(const Object * moving_obj,
								 const Object * static_obj,
								 const TerrainBase::Loc & reserved_pos);

		/** This multiplied by the max speed of an object gives the
			range at which objects and walls will exert a force on
			the object */
		extern const sint4 FIELD_RANGE_MULTIPLIER;
		
		/** The strength of attractors (does not vary with distance) */
		extern const sint4 ATTRACTOR_FORCE;

		/** This divided by the distance squared to the moving object
			gives the strength of the repulsors */
		extern const sint4 REPULSOR_BASELINE;
	}

	/// Utility Functions
	//@{
    /** Finds the location of point on a line that is closest to the
		given location */
    Loc find_point_on_line(const Loc & loc, const Loc & line_start, const Loc & line_end);
    /** Test if the object is at the given location,
		accounting for 1-tick lookahead */
    bool is_at_location(const Object* obj, const Loc& target);
    /** Test if the distance from the object to the target
		is within the given range */
    bool is_near_location(const Object* obj, const Loc& target, sint4 range);
    /** Test if the object is at a particular goal */
    bool is_at_goal(const Object* obj, const Goal & goal);
	/** Test if an object is in range */
    bool is_in_range(const Object * moving_obj, const Object * goal_obj);
    bool is_in_range(const Object * moving_obj, const Loc goal_loc);
//@}

	/// A low-level move command to be executed
	struct MoveCommand
	{
		GameObj * unit;		///< Unit which will move
		sint4     speed;    ///< Speed at which unit will move
		Loc       next_loc; ///< Destination for movement

		MoveCommand() : unit(0), speed(0) { }
	};

	/// Function object imposes weak ordering on locations
	class Loc_less : public std::binary_function<Loc, Loc, bool>{
	public:
		bool operator() (Loc a, Loc b) const;
	};

	class FactorPath;

	/// The internal data complementing Task data
	/// WARNING: This class requires EXTENSIVE refactoring, as it
	/// was initially a struct whose elements were manipulated by
	/// many different pieces of code. Most code related to
	/// FactorTask is still a little messy, and there are too
	/// many getter/setter methods.
    class FactorTask : public Movement::Task
	{
		boost::weak_ptr<FactorPath> manager; ///< Movement module
		boost::weak_ptr<FactorTask> self;    ///< Weak point to this

		std::set<GameObj *>		units;  ///< The units involved in this task
		Goal::const_ptr			goal;   ///< The goal for the units to reach
		
		Pathfinder::const_ptr	pf;		///< Algorithm for finding a path
		Path::ptr				path;   ///< A path for the units to follow

		PathExecutor::ptr		exec;	///< Algorithm for executing the path
		Guide::ptr				guide;  ///< Tracks completion of path

		StatusProc				proc;   ///< Procedure to call for status updates
		Status					status; ///< Status of task
		bool                    airOnly; ///< True if only air units in task

		bool pending;
		bool failed;
		bool noPathFound;
		
		uint4 collisionCount;
		uint4 planCount;
	public:
		FactorTask(const boost::weak_ptr<FactorPath> & manager, 
				   const std::set<GameObj *> & units, 
				   const Goal::const_ptr & goal,
				   const Pathfinder::ptr & pf,
				   const PathExecutor::ptr & exec,
				   const StatusProc & proc) :
			manager(manager), units(units), goal(goal),
			pf(pf), path(), exec(exec), guide(), proc(proc),
			airOnly(true), pending(true), failed(false),
			noPathFound(false), collisionCount(0), planCount(0)
		{
			FORALL(units, u)
			{
				if( (*u)->get_zcat(NULL) != Object::IN_AIR )
				{
					airOnly = false;
					break;
				}
			}
		}

		FactorTask(const boost::weak_ptr<FactorPath> & manager,
				   const std::set<GameObj *> & units, 
				   const Path::ptr & path,
				   const PathExecutor::ptr & exec,
				   const StatusProc & proc) :
			manager(manager), units(units), goal(path->getGoal()),
			pf(), path(path), exec(exec), guide(), proc(proc), 
			pending(false), failed(false), noPathFound(false),
			collisionCount(0), planCount(0)
		{
			guide = exec->getGuide(units, path);
		}

		~FactorTask() {}

		///{ Implementations for Movement::Task
		bool isPlanning() const { return pending && (!guide || !guide->isFinished()); }
		bool isRunning() const { return !pending && (!guide || !guide->isFinished()); }
		bool isFinished() const { return (guide && guide->isFinished()); }
		const std::set<GameObj *> & getUnits() const { return units; }
		Status getStatus() const { return status; }
		void cancel();
		///} End implementations

		void setSelfPtr(const boost::weak_ptr<FactorTask> & ptr)
		{
			self = ptr;
		}
		
		void removeUnit(GameObj * unit)
		{
			units.erase(unit);
			if(guide)
			{
				guide->removeUnit(unit);
			}
		}

		void planTask(uint4 stopTime);

		Guide & getGuide() { return *guide; }
		const Goal::const_ptr & getGoal() const { return goal; }

		void resetCollisionCount() { collisionCount = 0; }
		void reportCollision() { ++collisionCount; }
		bool shouldReplan() const { return collisionCount > REPLAN_TRIGGER && planCount < PLAN_HARDCAP; }
		bool shouldFail() const { return collisionCount > OCCILATION_TRIGGER; }

		void setPending(bool newPending) { pending = newPending; }
		void setFailed(bool newFailed) { failed = newFailed; }

		bool isPending() const { return pending; }
		bool isFailed() const { return failed; }
		bool isNoPathFound() const { return noPathFound; }
		bool isValid() const { return !pending && !failed && !noPathFound && path && guide; };

		bool isInvalidatedBySegment(const TerrainBase::Segment & segment) const;

		void setStatus(Status newStatus) { status = newStatus; }
		void callStatusProc() { proc( *this ); }

		typedef boost::shared_ptr<FactorTask> ptr;
		typedef boost::shared_ptr<const FactorTask> const_ptr;

		static const uint4 PLAN_HARDCAP = 3;         ///< Max number of plans for a task
		static const uint4 REPLAN_TRIGGER = 5;       ///< Number of occilation steps before a replan will be triggered
		static const uint4 OCCILATION_TRIGGER = 15;  ///< Number of occilation steps before a path movement failure will be passed
	};



	/// FactorPath, an implementation of Movement::Module
	/// This class handles scheduling of tasks, dispatching them through
	/// pluggable pathfinders/path execution systems, and low level
	/// collision avoidance via forcefields and pushing.
	class FactorPath : public Movement::Module, public EventHandler
	{
	public:
		typedef boost::shared_ptr<FactorPath> ptr;
		typedef boost::shared_ptr<FactorPath const> const_ptr;
		typedef boost::weak_ptr<FactorPath> wptr;
		typedef boost::weak_ptr<FactorPath const> const_wptr;

		typedef std::set<FactorTask::ptr>				TaskSet;
		typedef std::deque<FactorTask::ptr>             TaskQueue;
		typedef std::map<GameObj *, FactorTask::ptr>    Obj2Task;
		typedef std::map<GameObj *, Loc>				Obj2LocMap;
		typedef std::map<Loc, bool, Loc_less>			SegmentEndsMap;
	private:
		GameStateModule & gsm;
		FactorPath::wptr  self;
		sint4             max_ms_per_tick;

		TaskSet           tasks;        ///< All active tasks
		TaskQueue         pendingTasks; ///< New unplanned tasks
		TaskQueue         failedTasks;  ///< Failed tasks to replan
		Obj2Task          obj2task;     ///< Which objects are in which tasks
	
		/// All pathfinders used by the module
		std::set<Pathfinder::ptr> pathfinders;
		/// Binding of name to pathfinder
		std::map<std::string, Pathfinder::ptr> name2pathfinder;
		/// Binding of name to path execution system
		std::map<std::string, PathExecutor::ptr> name2pathExec;

		Obj2LocMap      objects;      ///< Used to calculate object repulsion
		Vector<Segment> segments;     ///< Used to calculate wall repulsion
		SegmentEndsMap  segment_ends; ///< Used to calculate repulsion from the ends of walls
		
		std::set<GameObj *> objsToStop; ///< Objects for which we should set a stop-moving action
		Random rand;                    ///< Pseudorandom number generator shared by all tasks
		bool pushing_on;                ///< True if allied units can be pushed out of the way
		bool wp_attract;                ///< True if units seek the waypoints of a path
		bool stop_to_plan;               
		sint4 client_pID;               ///< ID of the client player

		
		/// Handle events such as GameStateModule::VIEW_MSG
		bool handle_event(const Event & event);
		/// Update all abstractions (notably pathfinders) with game changes
		void update(const GameChanges & changes);
		/// Plan and executes tasks
		void act();
		/// Remove any associations to a particular object
		void clean_task_for_obj(GameObj * obj);
		/// Remove any associations to a particular task
		void clean_task(const FactorTask::ptr & taskPtr);
		/// Erase a task from a specific queue
		void eraseFromQueue(TaskQueue & queue, const FactorTask::ptr & taskPtr);
		/// Plan a task from a specific queue
		bool plan_queued_task(TaskQueue & queue, uint4 stop_time);
		/// Invalidate existing paths that intersect a given new boundary.
		void invalidate_paths(const Vector<Segment> &segments);
		/** Add a move command to the Vector of commands, telling the object to
			move to the target location at maximum speed. Also, updates the
			location associated with the object for forcefield purposes */
		void add_move_command(GameObj * obj, Loc target,
							  Vector<MoveCommand> &cmds);
		/// Issue additional move commands to accomodate pushing
		void handle_pushing(Vector<MoveCommand> &cmds);
		/// Determines what team an object is on. -1 if unkown
		sint4 what_team(const Object *obj);

		/// Produces an actual goal based on an object and a theoretical goal
		void get_forcefield_target(GameObj * obj, const Loc &attractor, Loc &output);
		/// Produces a pushing goal based on an object
		void get_push_target(GameObj * obj, Loc & output);
		
		/** Adjusts the vector so that it is not directed more than 90 degrees on
			either side of the target vector */
		void restrict_vector(const Vec2D & target_vec, Vec2D & vec);
		
	public:

		FactorPath(GameStateModule & gsm, sint4 max_ms_per_tick);

		~FactorPath();

		/// Set weak pointer to self
		void setSelfPointer(const FactorPath::wptr & self) { this->self = self; }
		///{ Implementations for Movement::Module
		void	  addPathfinder(const std::string & name,
								const Pathfinder::ptr & pathfinder);
		void	  addPathExecutor(const std::string & name,
								  const PathExecutor::ptr & pathExecutor);
		Path::ptr findPath(const std::string & pathfinder,
						   const Vec2D & start, sint4 radius,
						   Goal::const_ptr goal, uint4 stopTime = 0) const;
		Task::ptr moveUnits(const std::string & pathfinder,
							const std::string & pathExecutor,
							const std::set<GameObj*> & units,
							Goal::const_ptr goal,
							const Task::StatusProc & proc);
		Task::ptr moveUnitsViaPath(const std::string & pathExecutor,
								   const std::set<GameObj*> & units,
								   Path::ptr path,
								   const Task::StatusProc & proc);
		///} End implementations

		/// add command-line options specific to this implementation, using static methods of class Options
		static void add_options(void);

		/// Initialise abstractions (such as pathfinders)
		virtual void init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID);

		/// Add a new object for the module to consider
		virtual void add_obj(GameObj * obj);
		/// Notify the module that an object's state has changed
		virtual void update_obj(GameObj * obj);
		/// Remove an object from the module's consideration
		virtual void remove_obj(GameObj * obj);
		/// Insert boundaries into the module's representation
		virtual void add_segments(const Vector<TerrainBase::Segment> &s);

		/// Remove a unit from its current task
		virtual void cancel_task(GameObj *obj);

		/// Remove a task altogether
		virtual void remove_task(const FactorTask::ptr & task);

		/// Execute tasks by providing commands
		virtual void execute_tasks(Vector<MoveCommand> &cmds);

		/// Spend some time planning tasks
		virtual bool plan_tasks(uint4 max_time=0);
	};
}

#endif
