#include "FactorPath.H"

namespace Movement
{
	namespace ForceField
	{
		const sint4 FIELD_RANGE_MULTIPLIER	= 2;
		const sint4 ATTRACTOR_FORCE			= 26;
		const sint4 REPULSOR_BASELINE		= 200;

		Vec2D getPointRepulsion(const Object *moving_obj, const TerrainBase::Loc &point, sint4 radius)
		{
			Vec2D vec;
			TerrainBase::Loc center;
			moving_obj->get_center(center.x, center.y);
			real8 d = center.distance(point) - moving_obj->get_radius() - radius +1;

			if (d < moving_obj->get_max_speed() * FIELD_RANGE_MULTIPLIER)
			{
				vec.x = center.x - point.x;
				vec.y = center.y - point.y;
				real8 k = REPULSOR_BASELINE / (d * d * vec.length());
				vec = vec.multiply(k);
			}

			return vec;
		}

		Vec2D getLineRepulsion(const Object *moving_obj, const Segment &seg)
		{
			TerrainBase::Loc loc;
			moving_obj->get_center(loc.x, loc.y);
			const TerrainBase::Loc &line_start = seg.l1;
			const TerrainBase::Loc &line_end = seg.l2;

			sint4 line_x = line_end.x - line_start.x;
			sint4 line_y = line_end.y - line_start.y;

			sint4 v_x = loc.x - line_start.x;
			sint4 v_y = loc.y - line_start.y;

			sint4 a = line_x * v_x + line_y * v_y;
			sint4 b = line_x * line_x + line_y * line_y;

			TerrainBase::Loc proj(line_start.x + line_x * a / b, line_start.y + line_y * a / b);

			Vec2D line_to_obj(loc.x - proj.x, loc.y - proj.y);
			Vec2D to_l1, to_l2;

			to_l1 = Vec2D(seg.l1.x - proj.x, seg.l1.y - proj.y);
			to_l2 = Vec2D(seg.l2.x - proj.x, seg.l2.y - proj.y);

			if (Vec2D::dot(to_l1, to_l2) <= 0)
			{
				real8 d = loc.distance(proj) - moving_obj->get_radius() +1;
				if (d < moving_obj->get_max_speed() * FIELD_RANGE_MULTIPLIER /2)
				{
					real8 k = REPULSOR_BASELINE /  (d * d * line_to_obj.length());
					return line_to_obj.multiply(k);
				}
			}
			return Vec2D(0,0);
		}

		Vec2D getRectRepulsion(const Object * moving_obj, const TerrainBase::Loc & tl, const TerrainBase::Loc & br)
		{
			TerrainBase::Loc tr(br.x, tl.y);
			TerrainBase::Loc bl(tl.x, br.y);

			Vec2D vec(ForceField::getLineRepulsion(moving_obj, Segment(tl, tr)));
			vec += ForceField::getLineRepulsion(moving_obj, Segment(tr, br));
			vec += ForceField::getLineRepulsion(moving_obj, Segment(br, bl));
			vec += ForceField::getLineRepulsion(moving_obj, Segment(bl, tl));
			if (vec == Vec2D(0,0))
			{
				vec += ForceField::getPointRepulsion(moving_obj, tl);
				vec += ForceField::getPointRepulsion(moving_obj, tr);
				vec += ForceField::getPointRepulsion(moving_obj, bl);
				vec += ForceField::getPointRepulsion(moving_obj, br);
			}
			return vec;
		}

		Vec2D getObjectRepulsion(const Object * moving_obj, const Object * static_obj, const TerrainBase::Loc & reserved_pos)
		{
			if (moving_obj == static_obj)
			{
				return Vec2D(0,0);
			}
			if (static_obj->get_shape() == Object::CIRCLE)
			{
				TerrainBase::Loc obstacle_center;
				if (reserved_pos == TerrainBase::Loc(-1, -1))
				{
					static_obj->get_center(obstacle_center.x, obstacle_center.y);
				}
				else
				{
					obstacle_center = reserved_pos;
				}
				return ForceField::getPointRepulsion(moving_obj, obstacle_center, static_obj->get_radius());
			}
			else if (static_obj->get_shape() == Object::RECTANGLE)
			{
				TerrainBase::Loc u_l, l_r; //upper-left, lower-right
				static_obj->get_p1(u_l.x, u_l.y);
				static_obj->get_p2(l_r.x, l_r.y);
				return ForceField::getRectRepulsion(moving_obj, u_l, l_r);
			}
			else
			{
				return Vec2D(0,0);
			}
		}
	}
}
