#include "Movement.H"

#include "ScriptObj.H"
#include "GameObj.H"
#include "FactorPath.H"

#define UNUSED(x)
#define PI        3.141592654

namespace Movement
{
	enum GoalCondition { CONDITION_EXACT, CONDITION_TOUCH,
						 CONDITION_VICINITY, CONDITION_ATTACK_RANGE };

	bool isInRange( const GameObj * shooter, const Vec2D & target )
	{
		sint4 range = 1;
		ScriptObj * weapon = shooter->component("weapon");
		if( weapon )
		{
			const sintptr * r = weapon->get_int_ptr("max_ground_range");
			if( r )
			{
				range = *r;
			}
		}
		Vec2D position; shooter->get_center(position.x, position.y);
		return (target - position).lengthSq() <= (range * range);
	}

	bool isInRange( const GameObj * shooter, const Object * target )
	{
		sint4 range = 1;
		ScriptObj * weapon = shooter->component("weapon");
		if( weapon )
		{
			const sintptr * r = weapon->get_int_ptr("max_ground_range");
			if( r )
			{
				range = *r;
			}
		}
		return shooter->distance_to(*target) <= range;
	}

	class LocationGoal : public Goal
	{
	public:
		typedef boost::shared_ptr<LocationGoal> ptr;
	private:
		Vec2D            location;
		GoalCondition    condition;
		sint4            distance; // For vicinity condition
	public:
		LocationGoal(const Vec2D & location, GoalCondition condition,
					 sint4 distance = 0) :
			location(location), condition(condition), distance(distance) {}
		~LocationGoal() {}

		bool isValid() const
		{
			return true;
		}

		bool isObstacle() const
		{
			assert(isValid());
			return false;
		}

		void getBoundaries(std::vector<Vec2D> & bounds,
						   sint4 UNUSED(radius)) const
		{
			assert(isValid());
			bounds.push_back(location);
		}
		
		Vec2D getCenter() const
		{
			assert(isValid());
			return location;
		}
	    
		Object * getObject() const
		{
			assert(isValid());
			return 0;
		}

		real8 getApproxDistance(const Vec2D & point) const
		{
			assert(isValid());
			// TODO: Return octile approximation
			return getExactDistance(point);
		}

		real8 getExactDistance(const Vec2D & point) const
		{
			assert(isValid());
			return (point - location).length();
		}

		bool isAtGoal(const GameObj * unit) const
		{
			assert(isValid());
			
			Vec2D position; unit->get_center( position.x, position.y );
			switch( condition )
			{
			case CONDITION_EXACT:
				return position == location;
			case CONDITION_TOUCH:
				return (location - position).length() <= (unit->get_radius() + 1);
			case CONDITION_VICINITY:
				return (location - position).lengthSq() <= (distance * distance);
			case CONDITION_ATTACK_RANGE:
				return isInRange(unit, location);
			}
			return false;
		}
	};

	class ObjectGoal : public Goal
	{
	public:
		typedef boost::shared_ptr<ObjectGoal> ptr;
	private:
		mutable GameObj::Ptr       object;
		GoalCondition              condition;
		sint4                      distance; // For vicinity condition
	public:
		ObjectGoal(GameObj * object, GoalCondition condition, sint4 distance = 0) :
			object(object), condition(condition), distance(distance) {}
		~ObjectGoal() {}

		bool isValid() const
		{
		    if(!object.good()) return false;
			if(object->is_dead()) object = GameObj::Ptr();
			return object.good();
		}

		bool isObstacle() const
		{
			assert(isValid());
			return ((object->has_attr("is_mobile")) &&
					(object->get_int("is_mobile") == 0) &&
					(object->get_zcat(NULL) == Object::ON_LAND));
		}

		void getBoundaries(std::vector<Vec2D> & bounds, sint4 radius) const
		{
			if (object->get_shape() == Object::RECTANGLE)
			{
				sint4 x1, y1, x2, y2;
				object->get_p1(x1, y1);
				object->get_p2(x2, y2);
				for(sint4 x = x1; x <= x2; x += radius)
				{
					bounds.push_back(Vec2D(x, y1 - radius));
					bounds.push_back(Vec2D(x, y2 + radius));
				}
				for(sint4 y = y1; y <= y2; y += radius)
				{
					bounds.push_back(Vec2D(x1 - radius, y));
					bounds.push_back(Vec2D(x2 + radius, y));
				}
			}
			else if (object->get_shape() == Object::CIRCLE)
			{
				sint4 cx, cy; object->get_center(cx, cy);
				double dradius = object->get_radius() + radius;
				for(sint4 i=0; i<10; i++)
				{
					double angle = i * ((2.0 * PI) / 10.0);
					bounds.push_back(Vec2D(cx + (sint4)(cos(angle) * dradius),
										   cy + (sint4)(sin(angle) * dradius)));
				}
			}
		}
		
		Vec2D getCenter() const
		{
		    assert(isValid());
		    
			Vec2D location; object->get_center(location.x, location.y);
			return location;
		}
	    
		Object * getObject() const
		{
		    assert(isValid());
		    
			return object;
		}

		real8 getApproxDistance(const Vec2D & point) const
		{
			assert(isValid());
			
			// TODO: Return octile approximation
			return getExactDistance(point);
		}

		real8 getExactDistance(const Vec2D & point) const
		{
		    assert(isValid());
		    
			Vec2D location; object->get_center(location.x, location.y);
			return (point - location).length();
		}

		bool isAtGoal(const GameObj * unit) const
		{
	        assert(isValid());

			Vec2D position; unit->get_center( position.x, position.y );
			Vec2D location; object->get_center(location.x, location.y);
			real8 dist = unit->distance_to(*object);
			switch( condition )
			{
			case CONDITION_EXACT:
				return position == location;
			case CONDITION_TOUCH:
				return dist <= 2;
			case CONDITION_VICINITY:
				return dist <= distance;
			case CONDITION_ATTACK_RANGE:
				return isInRange(unit, object);
			}
			return false;
		}
	};

	Goal::const_ptr TouchPoint(const Vec2D & point)
	{
		LocationGoal::ptr goal(new LocationGoal(point, CONDITION_TOUCH));
		return Goal::const_ptr(goal);
	}

	Goal::const_ptr ReachPointVicinity(const Vec2D & point, sint4 distance)
	{
		LocationGoal::ptr goal(new LocationGoal(point, CONDITION_VICINITY, distance));
		return Goal::const_ptr(goal);
	}

	Goal::const_ptr AttackPoint(const Vec2D & point)
	{
		LocationGoal::ptr goal(new LocationGoal(point, CONDITION_ATTACK_RANGE));
		return Goal::const_ptr(goal);
	}

	Goal::const_ptr TouchObject(const GameObj * object)
	{
            ObjectGoal::ptr goal(new ObjectGoal(const_cast<GameObj*>(object), CONDITION_TOUCH));
		return Goal::const_ptr(goal);
	}

	Goal::const_ptr ReachObjectVicinity(const GameObj * object, sint4 distance)
	{
            ObjectGoal::ptr goal(new ObjectGoal(const_cast<GameObj*>(object), CONDITION_VICINITY, distance));
		return Goal::const_ptr(goal);
	}

	Goal::const_ptr AttackObject(const GameObj * object)
	{
            ObjectGoal::ptr goal(new ObjectGoal(const_cast<GameObj*>(object), CONDITION_ATTACK_RANGE));
		return Goal::const_ptr(goal);
	}
}
