/// This header file describes the primary interface of the movement module.
/// The module is designed so that almost all objects have virtual interfaces.
/// This allows future revisions to add various types of functionality with
/// minimal hassle. The core of this system is the Movement::Module interface,
/// whose implementations are responsible for coordinating the acts of finding
/// and executing paths for a variety of units with a variety of goals. There
/// is also an auxilliary class, Movement::Context, which provides helper and
/// convenience methods for using the Movement::Module. The recommended way
/// to use the movement subsystem is as follows:
///
/// 1. Spawn the module with Movement::MakeModule();
/// 2. Add pathfinders with Movement::Module::addPathfinder();
/// 3. Add path execution systems with Movement::Module::addPathExecutor();
/// 4. For every system that needs access to the module, instantiate a
///    Movement::Context object with the appropriate settings.
///
/// For instance, code for gathering resources could have a context set to
/// use a special pathfinding system, which makes use of oneway lane constructs
/// to minimise conflicts. Code for moving troops could have a special path
/// execution system designed to move squads in formation. To avoid having to
/// pass a large number of parameters each time a movement call is made, the
/// Movement::Context stores various "state" parameters such as which pathfinder
/// and path execution system to use.

/// TODO List
/// 1. Remove the dependency of SimpleTerrainPathfinder on knowing whether
///    the target is a location or an object, and remove the deprecated
///    Goal::getObject() method.
/// 2. Fix the const-correctness related to how the Path object is used.
///    When passing paths to the Module, Path::const_ptr should be sufficient.
/// 3. Implement the GroupFollowPath path execution system, which causes
///    a group of units to move along a path in compact clusters.

#ifndef MOVEMENT_H
#define MOVEMENT_H

#include "boost/function.hpp"
#include "boost/shared_ptr.hpp"
#include "boost/weak_ptr.hpp"
#include "boost/noncopyable.hpp"

#include "Global.H"
#include "Vec2D.H"

//#define USE_DCDTLITE


class GameStateModule;
class GameObj;

#ifdef USE_DCDTLITE
class DCDT;
#else
class TraDcdt;
#endif

namespace Movement
{
  /// Represents an arbitrary goal for pathfinding operations or for
  /// the movement of units. User-defined pathfinders should try not
  /// to call any deprecated methods on Goal objects. However, user-
  /// defined goals MAY wish to support deprecated methods for the
  /// sake of compatability with some older pathfinders.
  class Goal : public boost::noncopyable
  {
  public:
    typedef boost::shared_ptr<Goal> ptr;
    typedef boost::shared_ptr<const Goal> const_ptr;
    typedef boost::weak_ptr<Goal> wptr;
    typedef boost::weak_ptr<const Goal> const_wptr;
    
    Goal() {}
    virtual ~Goal() {}
    
    /// Returns false if the goal has somehow ceased to exist, in which
    /// case pathfinding operations should fail gracefully.
    virtual bool     isValid() const = 0;
    
    /// Returns true if the goal occupies a specific region in which it
    /// can block movement.
    virtual bool     isObstacle() const = 0;
    
    /// Returns the exact center of the goal. Pathing directly to this
    /// location should be acceptable if isObstacle() is false.
    virtual Vec2D    getCenter() const = 0;
    
    /// Fills the vector with a set of boundary points. These points should
    /// be outside of the target if isObstacle() is true.
    /// @param bounds Vector to be filled with boundary points
    /// @param radius Minimum distance from points to object boundary
    virtual void     getBoundaries(std::vector<Vec2D> & bounds,
                                   sint4 radius) const = 0;
    
    /// Returns distance to goal, fast, suitable as a heuristic.
    /// @param point Point from which to estimate distance
    virtual real8    getApproxDistance(const Vec2D & point) const = 0;
    
    /// Returns exact straightline distance to the goal.
    /// @param point Point from which to calculate distance
    virtual real8    getExactDistance(const Vec2D & point) const = 0;
    
    /// Returns true if an object is at a position which satisfies
    /// the goal's criteria.
    /// @param unit The unit to check
    virtual bool     isAtGoal(const GameObj * unit) const = 0;
    
    /// DEPRECATED: Returns an object if target is object, null otherwise.
    virtual Object * getObject() const = 0;
  };
  
  
  
  /// Interface for tracking the progress made by a particular
  /// entity as it travels along a path. Used to determine where
  /// an entity should travel to, based on its current location.
  /// These objects are primarily used to abstract away the internal
  /// implementation of any Path object. In essence, a Progress
  /// object tells you how to follow any given path.
  class Progress : public boost::noncopyable
  {
  public:
    typedef boost::shared_ptr<Progress> ptr;
    typedef boost::shared_ptr<const Progress> const_ptr;
    
    Progress() {}
    virtual ~Progress() {}
    
    /// Give the Progress object a chance to update its internal records
    /// with the most recent position of the "follower" entity.
    virtual void    updateState(const Vec2D & position) = 0;
    
    /// Give the Progress object a chance to pick a new target position
    /// based on the most recent position of the "follower" entity.
    virtual void    updateTarget(const Vec2D & position) = 0;
    
    /// Returns true if the "follower" entity has made some progress
    /// since the last time updateState() was called.
    virtual bool    isProgressing(const Vec2D & position) const = 0;
    
    /// Returns the next target for the "follower" entity to move to
    virtual Vec2D   getTarget() const = 0;
  };
  
  
  
  /// Interface representing an abstract path. Can be cached or
  /// manipulated in certain basic ways, and can be passed to the
  /// movement module to specify how units should move. User-defined
  /// Path implementations should also implement a corresponding
  /// Progress implementation. Paths with curved segments or annotated
  /// with metadata should easily be possible in this manner.
  class Path : public boost::noncopyable
  {
  public:
    typedef boost::shared_ptr<Path> ptr;
    typedef boost::shared_ptr<const Path> const_ptr;
    
    Path() {}
    virtual ~Path() {}
    
    /// Returns the approximate starting location
    virtual Vec2D getStart() const = 0;
    
    /// Returns the goal at the end of the path
    virtual Goal::const_ptr getGoal() const = 0;
    
    /// Returns the exact length of the path
    virtual real8 getLength() const = 0;
    
    /// Returns a new path which is reversed
    virtual Path::ptr getReversed() const = 0;
    
    /// Used internally to determine when to replan a
    /// previously planned path
    virtual bool intersectsSegment(const Vec2D & a,
                                   const Vec2D & b) const = 0;
    
    /// Used internally to determine how an entity should
    /// follow a given path.
    /// @param startPos The original location of the "following" entity
    /// @param lookahead The distance to generate targets at
    virtual Progress::ptr getProgress(const Vec2D & startPos,
                                      real8 lookahead) const = 0;
  };
  
  
  
  /// Interface representing a pathfinding algorithm. Allows
  /// multiple pathfinders to coexist simultaneously within
  /// an ORTS client. User-defined pathfinders should take
  /// care not to call deprecated methods on Goal objects. It
  /// is recommended that user-defined pathfinders do not
  /// listen directly to GameStateModule updates, but instead
  /// do their updates in the addObject, updateObject,
  /// removeObject, and updateBoundary methods. This ensures
  /// that each Pathfinder remains in synch with the Module
  /// during updates.
  class Pathfinder : public boost::noncopyable
  {
  public:
    typedef boost::shared_ptr<Pathfinder> ptr;
    typedef boost::shared_ptr<const Pathfinder> const_ptr;
    
    Pathfinder() {}
    virtual ~Pathfinder() {}
    
    /// Called once at the start of execution
    virtual void initialize(sint4 mapWidth, sint4 mapHeight,
                            sint4 tilePoints, sint4 clientPID,
                            sint4 neutralPID) = 0;
    
    /// Called every time a new object is discovered
    virtual void addObject(const Object * object) = 0;
    
    /// Called every time an object changes position
    virtual void updateObject(const Object * object) = 0;
    
    /// Called every time an object dies or disappears
    virtual void removeObject(const Object * object) = 0;
    
    /// Called every time a new line segment is encountered
    virtual void addBoundary(const Vec2D & a, const Vec2D & b) = 0;
    
    /// Called at the end of every update cycle, before any calls
    /// to findPath will be made. Useful if the pathfinder wishes
    /// to perform any expensive abstractions prior to pathfinding.
    virtual void updateData() = 0;
    
    /// Returns a path if one was found within the allotted time,
    /// or an empty pointer if none was found
    virtual Path::ptr findPath(const Vec2D & start,
                               sint4 radius,
                               Goal::const_ptr goal,
                               uint4 stopTime = 0) const = 0;
  };
  
  
  
  /// Interface for coordinating a group of units along a
  /// particular path via a particular algorithm. Guide
  /// objects will be spawned by a PathExecutor, similar
  /// to how Progress objects are spawned by a Path.
  class Guide : public boost::noncopyable
  {
  public:
    typedef boost::shared_ptr<Guide> ptr;
    typedef boost::shared_ptr<const Guide> const_ptr;
    
    Guide() {}
    virtual ~Guide() {}
    
    /// Inform the algorithm that a particular unit is gone
    virtual void removeUnit(GameObj * unit) = 0;
    
    /// Generate a new set of targets for the involved units
    virtual void updateTargets() = 0;
    
    /// Update any information regarding prior positions
    virtual void updateState() = 0;
    
    /// Determine if the units as a whole are making progress
    virtual bool isProgressing() const = 0;
    
    /// Determine if the units have reached their final destination
    virtual bool isFinished() const = 0;
    
    /// Retrieve the set of units
    virtual const std::set<GameObj *> & getUnits() const = 0;
    
    /// Determine if a particular unit has reached its destination
    virtual bool isUnitFinished(GameObj * unit) const = 0;
    
    /// Determine where a particular unit should travel
    virtual Vec2D getUnitTarget(GameObj * unit) const = 0;
  };
  
  
  
  /// Interface for encapsulating a particular path execution
  /// algorithm. Its sole purpose is to spawn "guide" objects
  /// when given a set of units and a path to execute. It also
  /// has methods to provide hints to the path planner about what
  /// sort of path the executor can best use.
  class PathExecutor : public boost::noncopyable
  {
  public:
    typedef boost::shared_ptr<PathExecutor> ptr;
    typedef boost::shared_ptr<const PathExecutor> const_ptr;

    PathExecutor() {}
    virtual ~PathExecutor() {}

    /// Create a guide object for a set of units and a given path
    virtual Guide::ptr getGuide(const std::set<GameObj *> & units,
                                const Path::ptr & path) = 0;

    /// Determine the best starting point for path planning with units
    virtual Vec2D getStartingPoint(const std::set<GameObj *> & units,
                                   const Goal::const_ptr & goal) = 0;

    /// Determine the radius of the path to plan for a set of units
    virtual real8 getPathRadius(const std::set<GameObj *> & units) = 0;
  };



  /// Interface representing a particular job given to the
  /// movement module. It can contain both path planning and
  /// execution phases, or simply an execution phase for a
  /// previously planned path.
  class Task : public boost::noncopyable
  {
  public:
    typedef boost::shared_ptr<Task> ptr;
    typedef boost::shared_ptr<const Task> const_ptr;
    typedef boost::function<void(const Task &)> StatusProc; 

    enum Status { PLANNING, RUNNING, PLAN_FAILURE, MOVE_FAILURE, SUCCESS };

    Task() {}
    virtual ~Task() {}

    /// Retrieves the set of units involved in the task
    virtual const std::set<GameObj *> & getUnits() const = 0;

    /// Returns the current status of the task
    virtual Status getStatus() const = 0;

    /// Returns true if the task is currently planning a path
    virtual bool isPlanning() const = 0;

    /// Returns true if the task is currently executing a path
    virtual bool isRunning() const = 0;

    /// Returns true if the set of units have reached their goal
    virtual bool isFinished() const = 0;

    /// Stops any planning/execution of a task immediately
    virtual void cancel() = 0;
  };



  /// Interface for a primary module which coordinates tasks with
  /// multiple independent pathfinders and path execution systems.
  /// It is not recommended that user-defined implementations be
  /// used for this interface. The existing implementation is
  /// designed to work with generic goals, pathfinders, paths,
  /// and path execution systems. It also automatically handles
  /// forcefields and pushing for local collision avoidance.
  class Module : public boost::noncopyable
  {
  public:
    typedef boost::shared_ptr<Module> ptr;
    typedef boost::shared_ptr<const Module> const_ptr;

    Module() {}
    virtual ~Module() {}

    /// Add a new named pathfinder to the module
    virtual void addPathfinder(const std::string & name,
                               const Pathfinder::ptr & pathfinder) = 0;

    /// Add a new named path execution algorithm to the module
    virtual void addPathExecutor(const std::string & name,
                                 const PathExecutor::ptr & pathExec) = 0;

    /// Find a path using a given named pathfinder
    virtual Path::ptr findPath(const std::string & pathfinder,
                               const Vec2D & start,
                               sint4 radius,
                               Goal::const_ptr goal,
                               uint4 stopTime = 0) const = 0;

    /// Move a set of units using the given parameters
    virtual Task::ptr moveUnits(const std::string & pathfinder,
                                const std::string & pathExecutor,
                                const std::set<GameObj*> & units,
                                Goal::const_ptr goal,
                                const Task::StatusProc & proc) = 0;

    /// Execute a known path with a given path execution system
    virtual Task::ptr moveUnitsViaPath(const std::string & pathExecutor,
                                       const std::set<GameObj*> & units,
                                       Path::ptr path,
                                       const Task::StatusProc & proc) = 0;
  };



  /// Class used to facilitate easy access to the movement module. It
  /// allows the user to specify a pathfinder and a path execution
  /// system to use, as well as a few common properties. It then provides
  /// helper methods so that access to the movement module is as convenient
  /// as possible. When the context passes out of scope, all tasks
  /// instantiated through it are canceled.
  class Context : public boost::noncopyable
  {
    Movement::Module &mm;
    uint4 stopTime;
    std::string pathfinder;
    std::string pathExec;
    Task::StatusProc statProc;
    std::set<Task::ptr> activeTasks;

    void handleStatusUpdate(const Task & task);
  public:
    Context(Movement::Module & mm,
            const std::string & pathfinder,
            const std::string & pathExec);
    ~Context();

    Path::ptr   findPath(GameObj * unit, Goal::const_ptr goal) const;
    Path::ptr   findPath(const Vec2D & start, sint4 radius,
                         Goal::const_ptr goal) const;

    Task::ptr   moveUnit(GameObj * unit, Goal::const_ptr goal);
    Task::ptr   moveUnits(const std::set<GameObj*> & units,
                          Goal::const_ptr goal);

    Task::ptr   moveUnitViaPath(GameObj * unit, Path::ptr path);
    Task::ptr   moveUnitsViaPath(const std::set<GameObj*> & units,
                                 Path::ptr path);

    void setPathfinderStopTime(uint4 milliseconds);
    void setPathfinderName(const std::string & pathfinder);
    void setPathExecName(const std::string & pathExec);
    void setStatusProcedure(const Task::StatusProc & statProc);
  };

  // Stock goals
  Goal::const_ptr TouchPoint(const Vec2D & point);
  Goal::const_ptr ReachPointVicinity(const Vec2D & point, sint4 distance);
  Goal::const_ptr AttackPoint(const Vec2D & point);

  Goal::const_ptr TouchObject(const GameObj * object);
  Goal::const_ptr ReachObjectVicinity(const GameObj * object, sint4 distance);
  Goal::const_ptr AttackObject(const GameObj * object);

  // Stock path implementation
  Path::ptr MakeWaypointPath(const Vector<Vec2D> & waypoints,
                             const Goal::const_ptr & goal);

  // Stock pathfinders
  Pathfinder::ptr MakeSimpleTerrainPathfinder();
   
#ifdef USE_DCDTLITE
  Pathfinder::ptr MakeTriangulationPathfinder(DCDT ** pPF = 0);
#else
  Pathfinder::ptr MakeTriangulationPathfinder(TraDcdt ** pPF = 0);
  
#endif

  // Stock path execution systems
  PathExecutor::ptr MakeMultiFollowExecutor();

  // Primary implementation of the module
  // Complete with forcefields and pushing
  Module::ptr MakeModule(GameStateModule & gsm, sint4 max_ms_per_tick);
}

#endif
