#include "FactorPath.H"

namespace Movement
{
	class MultiFollowPath : public Guide
	{
		std::set<GameObj *> units; /// Set of units following path
		std::map<GameObj *, Progress::ptr> unit2data; /// Data for individual units
		Path::ptr path; /// Path for units to follow
	public:
		MultiFollowPath(const std::set<GameObj *> & units, const Path::ptr & path);
		~MultiFollowPath() {};

		void removeUnit(GameObj * unit);
		void updateTargets();
		void updateState();

		bool isProgressing() const;
		bool isFinished() const;

		const std::set<GameObj *> & getUnits() const;
		bool isUnitFinished(GameObj * unit) const;
		Vec2D getUnitTarget(GameObj * unit) const;

		typedef boost::shared_ptr<MultiFollowPath> ptr;
		typedef boost::shared_ptr<const MultiFollowPath> const_ptr;

		/// this multiplied by the object max speed gives the distance
		/// ahead of the object on the path that the attractor will be set to
		static const sint4 PATH_LOOK_AHEAD = 3;
	};

	MultiFollowPath::MultiFollowPath(const std::set<GameObj *> & units, const Path::ptr & path) :
		units(units), unit2data(), path(path)
	{
		for(std::set<GameObj*>::const_iterator u = units.begin(); u != units.end(); ++u)
		{
			GameObj * unit = (*u);
			Vec2D position; unit->get_center(position.x, position.y);
			real8 lookahead = unit->get_max_speed() * PATH_LOOK_AHEAD;
			Progress::ptr progress = path->getProgress(position, lookahead);
			unit2data.insert( std::make_pair( unit, progress ) );
		}
		updateState();
	}

	void MultiFollowPath::removeUnit(GameObj * unit)
	{
		units.erase( unit );
		unit2data.erase( unit );
	}

	void MultiFollowPath::updateTargets()
	{
		std::map<GameObj *, Progress::ptr>::iterator u2d;
		for(u2d = unit2data.begin(); u2d != unit2data.end(); ++u2d)
		{
			const GameObj * unit = u2d->first;
			Progress::ptr & progress = u2d->second;
			Vec2D position; unit->get_center(position.x, position.y);
			progress->updateTarget(position);
		}
	}

	void MultiFollowPath::updateState()
	{
		std::map<GameObj *, Progress::ptr>::iterator u2d;
		for(u2d = unit2data.begin(); u2d != unit2data.end(); ++u2d)
		{
			GameObj * unit = u2d->first;
			Progress::ptr & progress = u2d->second;
			Vec2D position; unit->get_center(position.x, position.y);
			progress->updateState(position);
		}
	}

	bool MultiFollowPath::isFinished() const
	{
	    if( !path->getGoal()->isValid() )
		{
		    return true;
		}
	    
		for(std::set<GameObj *>::const_iterator it = units.begin(); it != units.end(); ++it)
		{
			if( isUnitFinished( (*it) ) == false )
			{
				return false;
			}
		}
		return true; // Otherwise all units are finished
	}

	bool MultiFollowPath::isProgressing() const
	{
		assert( !isFinished() );

		std::map<GameObj *, Progress::ptr>::const_iterator u2d;
		for(u2d = unit2data.begin(); u2d != unit2data.end(); ++u2d)
		{
			const GameObj * unit = u2d->first;
			const Progress::ptr & progress = u2d->second;
			Vec2D position; unit->get_center(position.x, position.y);
			if( !progress->isProgressing(position) )
			{
				return false;
			}
		}

		return true;
	}

	const std::set<GameObj *> & MultiFollowPath::getUnits() const
	{
		return units;
	}

	bool MultiFollowPath::isUnitFinished(GameObj * unit) const
	{
		assert( unit != 0 );
		assert( path->getGoal()->isValid() );
		
		return path->getGoal()->isAtGoal( unit );
	}

	Vec2D MultiFollowPath::getUnitTarget(GameObj * unit) const
	{   
		std::map<GameObj *, Progress::ptr>::const_iterator u2d;
		u2d = unit2data.find( unit );
		if( u2d != unit2data.end() )
		{
			return u2d->second->getTarget();
		}
		assert( false && "Invalid unit" );
		return Vec2D();
	}

	class MultiFollowExec : public PathExecutor
	{
	public:
		MultiFollowExec() {}
		~MultiFollowExec() {}

		Guide::ptr getGuide(const std::set<GameObj *> & units, const Path::ptr & path)
		{
			return Guide::ptr(new MultiFollowPath(units, path));
		}

		Vec2D getStartingPoint(const std::set<GameObj *> & units,
							   const Goal::const_ptr & goal)
		{
			Vec2D farthestPoint;
			real8 farthestDistance(-1.0);
			FORALL(units, u)
			{
				GameObj * unit(*u);
				Vec2D position; unit->get_center(position.x, position.y);
				real8 distance = goal->getExactDistance(position);
				if(farthestDistance < 0.0 || distance < farthestDistance)
				{
					farthestPoint = position;
					farthestDistance = distance;
				}
			}
			return farthestPoint;
		}
		
		real8 getPathRadius(const std::set<GameObj *> & units)
		{
			sint4 max_radius = 0;
			FORALL(units, u)
			{
				sint4 radius = (*u)->get_radius();
				if(radius > max_radius)
				{
					max_radius = radius;
				}
			}
			return max_radius;
		}
	};

	PathExecutor::ptr MakeMultiFollowExecutor()
	{
		return PathExecutor::ptr(new MultiFollowExec());
	}
}
