#include "Movement.H"
#include "GameStateModule.H"
#include "GameChanges.H"
#include "Game.H"
#include "ST_PFEngine.H"
#include "Options.H"

namespace Movement
{
    class FakeObject : public Object
    {
		Vec2D position;
		sint4 radius;
    public:
		FakeObject(const Vec2D & position, sint4 radius) : position(position), radius(radius) {}
		~FakeObject() {}
	
		Shape get_shape() const { return CIRCLE; }
		sint4 get_max_speed() const { return 0; }
		sint4 get_radius() const { return radius; }
	
		ZCat get_zcat(ZCat *prev = 0) const
		{
			if(prev) *prev = ON_LAND; return ON_LAND;
		}
		
		sint4 get_speed(sint4 *prev=0) const
		{
			if(prev) *prev = 0; return 0;
		}
		
		bool get_moving(bool *prev=0) const
		{
			if(prev) *prev = false; return false;
		}
		
		void get_center(sint4 &x, sint4 &y, sint4 *px=0, sint4 *py=0) const
		{
			x = position.x;
			y = position.y;
			if(px) *px = position.x;
			if(py) *py = position.y;
		}
	
		void get_p1(sint4 &x, sint4 &y, sint4 *px=0, sint4 *py=0) const
		{
			get_center(x, y, px, py);
		}
		
		void get_p2(sint4 &x, sint4 &y, sint4 *px=0, sint4 *py=0) const
		{
			get_center(x, y, px, py);
		}
	
		bool is_pending_action(void) const { return false; }
	
		real8 distance_to(const Object &other) const
		{
			return Game::distance(*this, other);
		}
    };
	
    /// Pathfinder wrapper which used ST_PFEngine, a grid-based A* pathfinder
    class STPathfinder : public Pathfinder
    {
		typedef SimpleTerrain::ST_Terrain::PFEngine PFEngine;
	
		boost::shared_ptr<PFEngine> pfEngine;
    public:
		STPathfinder() : Pathfinder()
		{
			
		}

		~STPathfinder()
		{
			
		}
	
		void initialize(sint4 mapWidth, sint4 mapHeight, sint4 tilePoints,
						sint4, sint4)
		{
			// initialize the internal PFEngine
			bool highResOnly;
			Options::get("-highres", highResOnly);
			Vector<sint4> lods;
			if( !highResOnly ) 
			{
				lods.push_back(4);
				// default - use a resolution reduced by a factor of 4.
			}
			lods.push_back(1);
			// also (or exclusively) maintain a full resolution view.
			
			pfEngine = boost::shared_ptr<PFEngine>(
				new PFEngine(mapWidth, mapHeight, tilePoints, lods));
		}

		bool isObjectStatic(const Object * object)
		{
			return object->get_max_speed() == 0;
		}
		
		void addObject(const Object * object)
		{
			if(isObjectStatic(object))
			{
				pfEngine->insert_object(object);
			}
		}
    
        void updateObject(const Object * object)
		{
			if(isObjectStatic(object))
			{
				pfEngine->update_object(object);
			}
		}

        void removeObject(const Object * object)
		{
			if(isObjectStatic(object))
			{
				pfEngine->remove_object(object);
			}
		}
		
        void addBoundary(const Vec2D & a, const Vec2D & b)
		{
			pfEngine->insert_boundary( TerrainBase::Segment( a.x, a.y,
															 b.x, b.y ) );
		}

		void updateData() {}

        Path::ptr findPath(const Vec2D & start, sint4 radius,
						   Goal::const_ptr goal, uint4 stopTime) const
		{
			if(!goal->isValid())
			{
				return Path::ptr();
			}
			
			Object * target_object = goal->getObject();
			if(target_object)
			{
				FakeObject fake_obj(start, radius);
				TerrainBase::Path newpath;
				if( pfEngine->find_path(&fake_obj, target_object, newpath,
										0, TerrainBase::IGNORE_ALL_MOBILE_OBJS,
										stopTime) )
				{
					Vector<Vec2D> wps;
					for(Vector<TerrainBase::Loc>::const_iterator it =
						newpath.locs.begin(); it != newpath.locs.end(); ++it)
					{
						wps.push_back( Vec2D( it->x, it->y ) );
					}
					return MakeWaypointPath(wps, goal);
				}
			}
			else
			{
				Vec2D target_loc = goal->getCenter();
				TerrainBase::Path newpath;
				TerrainBase::Loc start_loc(start.x, start.y);
				TerrainBase::Loc goal_loc(target_loc.x, target_loc.y);
				if( pfEngine->find_path(start_loc, goal_loc, radius, newpath,
										0, TerrainBase::IGNORE_ALL_MOBILE_OBJS,
										stopTime) )
				{
					newpath.locs[0] = goal_loc;
					Vector<Vec2D> wps;
					for(Vector<TerrainBase::Loc>::const_iterator it =
						newpath.locs.begin(); it != newpath.locs.end(); ++it)
					{
						wps.push_back( Vec2D( it->x, it->y ) );
					}
					return MakeWaypointPath(wps, goal);
				}
			}
			return Path::ptr();
		}
    };

    Pathfinder::ptr MakeSimpleTerrainPathfinder()
    {
		return Pathfinder::ptr(new STPathfinder());
    }
}
