

#if 1

#include "Base.h"
#include "GameObj.H"
#include "SegmentMerge.H"
#include "TerrainBasicImp.H"

struct XTR_Object
{
	//the ID of the object being monitored
	ObjId id;
	//the triangle in which the object lies
	TraFace *face;
	//the influence of the object
	float influence;

	//constructor - initializes member variables
	XTR_Object(ScriptObj *_obj, TraFace *_face = NULL) : id(GameObj::get_obj_id(_obj)), face (_face)
	{
		CalculateInfluence(_obj);
	}

	//calculates the object's influence based on its attributes
	void CalculateInfluence(ScriptObj *_obj)
	{
		//get the game object associated with the object
		GameObj *gob = _obj->get_GameObj();
		//get its hit points and weapon
		const sintptr *hp = gob->get_int_ptr("hp");
		ScriptObj *weapon = gob->component("weapon");
		//if it has no weapon, it has no influence
		if (weapon == NULL)
		{
			influence = 0.0f;
			return;
		}
		//get the minimum and maximum damage and cooldown for the weapon
		const sintptr *min = weapon->get_int_ptr("min_damage");
		const sintptr *max = weapon->get_int_ptr("max_damage");
		const sintptr *cooldown = weapon->get_int_ptr("cooldown");
		//if any value failed, it has no influence
		if ((hp == NULL) || (min == NULL) || (max == NULL) || (cooldown == NULL))
		{
			influence = 0.0f;
		}
		else
		{
			//otherwise the influence is calculated as the average damage of its weapon, divided by
			//the number of turns between its attacks, times its hit points
			influence = ((float)(*min) + (float)(*max)) / 2.0f * (float)(*hp) / (float)((*cooldown) + 1);
		}
	}
};

struct XTR_Obstacle
{
	//the ID of the obstacle within ORTS
	ObjId ortsId;
	//the ID of the obstacle within the triangulation
	int dcdtId;

	//constructor - initializes the member variables
	XTR_Obstacle(ObjId _ortsId, int _dcdtId) : ortsId (_ortsId), dcdtId (_dcdtId) {}

	//returns whether or not the given object is an obstacle
	static bool IsObstacle(GameObj *gob)
	{
		// Harvester units sometimes become immobile while harvesting.
		// Do NOT consider them as obstacles or they will be trapped.
		if(gob->bp_name() == "worker") return false;
		if(gob->bp_name() == "dactyl") return false;
		if(gob->bp_name() == "vector") return false;

		// Otherwise consider immobile land objects to be obstacles
		return ((gob->has_attr("is_mobile")) && (gob->get_int("is_mobile") == 0) &&
			    (gob->get_zcat(NULL) == Object::ON_LAND));
	}
};



#include "Global.H"
#include "Options.H"

using namespace std;

#define EPSILON 0.00001f
#define NO_INFLUENCE 0.01f
#define SIDES 8
#define TRIES 10

SrArray<TraFace *> TraFace::processing;
int TraFace::current = 0;
 
#include "Movement.H"

static sint4 myRound(real8 f)
{
	return static_cast<sint4>(f+0.5f);
}

namespace Movement
{
	class TRPathfinder : public Pathfinder
	{
		std::auto_ptr<TraDcdt> pfDcdt;
		SegmentMerge sm;
		Vector<XTR_Object> objects;
		Vector<XTR_Obstacle> obstacles;
		sint4 client, neutral;
	public:
		TRPathfinder(std::auto_ptr<TraDcdt> & pfDcdt);
		virtual ~TRPathfinder();
		void initialize(sint4 mapWidth, sint4 mapHeight,
								sint4 tilePoints, sint4 clientPID,
								sint4 neutralPID);
		void addObject(const Object * object);
		void updateObject(const Object * object);
		void removeObject(const Object * object);
		void addBoundary(const Vec2D & a, const Vec2D & b);
		void updateData() {}


		Path::ptr findPath(const Vec2D & start,
								   sint4 radius,
								   Goal::const_ptr goal,
								   uint4 stopTime = 0) const;

		void AddObstacle(GameObj *gob);
		void RemoveObstacle(GameObj *gob);
		void UpdateAbstraction();
		Path::ptr FindPathToPoint(const Vec2D & start, const Vec2D & point, sint4 radius, const Goal::const_ptr & goal) const;
	};

	TRPathfinder::TRPathfinder(std::auto_ptr<TraDcdt> & pfDcdt) : pfDcdt(pfDcdt)
	{

	}

	TRPathfinder::~TRPathfinder()
	{

	}

 
	void TRPathfinder::initialize(sint4 mapWidth, sint4 mapHeight,
								  sint4 tilePoints, sint4 clientPID,
								  sint4 neutralPID)
	{
		client = clientPID;
		neutral = neutralPID;

		float width = (float)(mapWidth * tilePoints);
		float height = (float)(mapHeight * tilePoints);
	
		SrPolygon boundary;
		boundary.push().set(0.0f, 0.0f);
		boundary.push().set(width, 0.0f);
		boundary.push().set(width, height);
		boundary.push().set(0.0f, height);
		pfDcdt->init(boundary, EPSILON, 1);
	}

 
	void TRPathfinder::addObject(const Object * obj)
    {
		GameObj *gob = (GameObj *)obj;
		if (XTR_Obstacle::IsObstacle(gob))
		{
			AddObstacle(gob);
			UpdateAbstraction();
			return;
		}
		
		// Otherwise, if unit is an enemy
		const sintptr *owner = gob->get_int_ptr("owner");
		if ((*owner == client) || (*owner == neutral))
		{
			return;
		}
		//locates the triangle in which the object's center lies
		TraBase *s = NULL;
		sint4 x, y;
		obj->get_center(x, y);
		//sets the object's triangle to null if not found
		if (pfDcdt->LocatePoint(x, y, s) == SeTriangulator::NotFound)
		{
			objects.push_back(XTR_Object(gob));
		}
		else
		{
			//otherwise sets it to the triangle face found
			TraFace *face = (TraFace *)s->fac();
			XTR_Object trobj(gob, face);
			objects.push_back(trobj);
			//and updates the unit's influence in the triangulation
			pfDcdt->UpdateInfluence(face, trobj.influence);
		}
	}


	void TRPathfinder::updateObject(const Object *obj)
	{
		GameObj *gob = (GameObj *)obj;
		if (XTR_Obstacle::IsObstacle(gob))
		{
			RemoveObstacle(gob);
			AddObstacle(gob);
			UpdateAbstraction();
			return;
		}	
	
		//only processes enemy units
		const sintptr *owner = gob->get_int_ptr("owner");
		if ((*owner == client) || (*owner == neutral))
		{
			return;
		}
		ObjId id = GameObj::get_obj_id(gob);
		//goes through the objects
		for (uint4 i = 0; i < objects.size(); i++)
		{
			//until it finds the desired object
			if (objects[i].id == id)
			{
				//record the object's old triangle and influence
				TraFace *oldFace = objects[i].face;
				float oldInfluence = objects[i].influence;
				//calculate the object's new influence
				objects[i].CalculateInfluence((GameObj *)obj);
				//remove the object's influence from it last triangle
				if (objects[i].face != NULL)
				{
					pfDcdt->UpdateInfluence(objects[i].face, -(objects[i].influence));
				}
				TraBase *s = NULL;
				SeTriangulator::LocateResult result;
				sint4 x, y;
				obj->get_center(x, y);
				//if possible, uses the last triangle as the starting point
				//otherwise, uses the triangle around the nearest sector midpoint
				//then finds the triangle in which the object is currently centered
				if (objects[i].face == NULL)
				{
					result = pfDcdt->LocatePoint(x, y, s);
				}
				else
				{
					result = pfDcdt->LocatePoint(objects[i].face, x, y, s);
				}
				//sets the object's face to NULL if it wasn't found
				if (result == SeTriangulator::NotFound)
				{
					objects[i].face = NULL;
				}
				else
				{
					//otherwise, sets it appropriately
					objects[i].face = (TraFace *)s->fac();
				}
				//checks if the object has moved triangles or if it has changed influence
				if ((oldFace != objects[i].face) || (fabs(oldInfluence - objects[i].influence) > NO_INFLUENCE))
				{
					//if it has and the old triangle was valid, subtract its influence from the old triangle
					if(oldFace != NULL)
					{	
						pfDcdt->UpdateInfluence(oldFace, -oldInfluence);
					}
					//if the new triangle is valid, add the object's influence to the new triangle
					if (objects[i].face != NULL)
					{
						//and updates the object's new influence on the map
						pfDcdt->UpdateInfluence(objects[i].face, objects[i].influence);
					}
				}
				break;
			}
		}
	}

	void TRPathfinder::removeObject(const Object *obj)
	{	
		GameObj *gob = (GameObj *)obj;
		if (XTR_Obstacle::IsObstacle(gob))
		{
			RemoveObstacle(gob);
			UpdateAbstraction();
			return;
		}	

		//only processes enemy units
		const sintptr *owner = gob->get_int_ptr("owner");
		if ((*owner == client) || (*owner == neutral))
		{
			return;
		}
		ObjId id = GameObj::get_obj_id(gob);
		//goes through the list of objects
		for (std::vector<XTR_Object>::iterator i = objects.begin(); i != objects.end(); i++)
		{
			//upon finding the matching object
			if (i->id == id)
			{
				pfDcdt->UpdateInfluence(i->face, -(i->influence));
				//removes it from the list
				objects.erase(i);
				break;
			}
		}
	}

	void TRPathfinder::addBoundary(const Vec2D & a, const Vec2D & b)
	{
		sm.clear_delta();
		DCDT::Segment s1(a.x, a.y, b.x, b.y);
		sm.new_segment(s1);

		// process all changes
		bool changed = false;
		sint4 add_n = 0, del_n = 0;
		FORS (i, 4)
		{
			del_n += sm.del[i].size();
			FORALL (sm.del[i], it)
			{
				pfDcdt->remove_polygon(*it);
				changed = true;
			}
		}
		sint4 total = 0;
		FORS (i, 4)
		{
			total += sm.all[i].size();
			add_n += sm.add[i].size();
			FORALL (sm.add[i], it)
			{
				const SortSegment &ss = **it;
				SrPolygon pol;
				pol.push().set(ss.s.x1, ss.s.y1);
				pol.push().set(ss.s.x2, ss.s.y2);
				ss.label = pfDcdt->insert_polygon(pol);
				changed = true;
			}
		}
		//if the triangulation was changed, update the abstraction
		if (changed)
		{
			UpdateAbstraction();
		}
	}
 
	struct CloserTo
	{
		Vec2D point;
		CloserTo(const Vec2D & point) : point(point) {}
		bool operator()(const Vec2D & a, const Vec2D & b)
		{
			return (point-a).lengthSq() < (point-b).lengthSq();
		}
	};

	Path::ptr TRPathfinder::findPath(const Vec2D & start, sint4 radius, Goal::const_ptr goal, uint4 stopTime) const
	{
		const Goal & g(*goal);
		if(!g.isObstacle())
		{
			return FindPathToPoint(start, g.getCenter(), radius, goal);
		}
		else
		{
			std::vector<Vec2D> bounds;
			g.getBoundaries(bounds, radius + 1);
			std::sort(bounds.begin(), bounds.end(), CloserTo(start));
			for(size_t i=0; i<bounds.size(); i++)
			{
				Path::ptr path = FindPathToPoint(start, bounds[i], radius, goal);
				if(path || SDL_GetTicks() > stopTime)
				{
					return path;
				}
			}
			return Path::ptr();
		}
	}
 
	void TRPathfinder::AddObstacle(GameObj *gob)
	{
		//create a closed polygon for the obstacle
		SrPolygon poly;
		poly.open(false);
		//check if the obstacle is rectangular
		if (gob->get_int("shape") == Object::RECTANGLE)
		{
			//get its corners
			sint4 x1, y1, x2, y2;
			gob->get_p1(x1, y1);
			gob->get_p2(x2, y2);
			//create a polygon using them
			poly.push().set((float)x1, (float)y1);
			poly.push().set((float)x2, (float)y1);
			poly.push().set((float)x2, (float)y2);
			poly.push().set((float)x1, (float)y2);
		}
		//check if the obstacle is circular
		else if (gob->get_int("shape") == Object::CIRCLE)
		{
			//get its center and radius
			sint4 cx, cy;
			gob->get_center(cx, cy);
			float r = (float)gob->get_radius() - 0.5f;
			float x, y;
			float angle = 0.0f;
			float inc = 2.0f * PI / (float)SIDES;
			//create a certain number of vertices to approximate the circle
			for (int i = 0; i < SIDES; i++)
			{
				//calculate the position of the vertex
				x = (float)cx + cos(angle) * r;
				y = (float)cy + sin(angle) * r;
				//add it to the polygon
				poly.push().set(x, y);
				//and increment the angle to the next vertex
				angle += inc;
			}
		}
		else
		{
			//we only deal with circular and rectangular obstacles
			return;
		}
		//get the obstacle's ID within ORTS
		ObjId ortsId = GameObj::get_obj_id(gob);
		//insert the obstacle's polygon into the triangulation and record its ID there
		int dcdtId = pfDcdt->insert_polygon(poly);
		//add this to our list of obstacles
		obstacles.push_back(XTR_Obstacle(ortsId, dcdtId));
	}

	void TRPathfinder::RemoveObstacle(GameObj *gob)
	{
		//get the ID of the obstacle within ORTS
		ObjId ortsId = GameObj::get_obj_id(gob);
		//go through our list of obstacles
		for (std::vector<XTR_Obstacle>::iterator i = obstacles.begin(); i != obstacles.end(); i++)
		{
			//when we find one with a matching ID,
			if (ortsId == i->ortsId)
			{
				//if the object was successfully added to the triangulation,
				if (i->dcdtId != -1)
				{
					//remove it
					pfDcdt->remove_polygon(i->dcdtId);
				}
				//remove it from the list of obstacles
				obstacles.erase(i);
				break;
			}
		}
	}

	void TRPathfinder::UpdateAbstraction()
	{
		//delete the abstraction
		pfDcdt->DeleteAbstraction();
		//rebuild it
		pfDcdt->Abstract();
		//repair the sectors that have been invalidated
		pfDcdt->InitializeSectors();
		//and record that the new faces have been processed
		TraFace::Clear();
	}


	Path::ptr TRPathfinder::FindPathToPoint(const Vec2D & start, const Vec2D & point, sint4 radius, const Goal::const_ptr & goal) const
	{
		Uint32 time = SDL_GetTicks();
		if (!pfDcdt->SearchPathFast((float)start.x, (float)start.y, (float)point.x, (float)point.y,
									(float)radius, time + 200000, time + 500000))
		{
			return Path::ptr();
		}

		Vector<Vec2D> waypoints;
		//get the path and record the first waypoint
		SrPolygon CurPath = pfDcdt->GetPath();
		waypoints.push_back( Vec2D(myRound(CurPath[0].x), myRound(CurPath[0].y)) );
		//go through the straight segments of the path
		for (int i = 0; i < CurPath.size() - 1; i += 2)
		{
			//if another follows it, construct the curved segment in between
			if (i < CurPath.size() - 3)
			{
				float p0x(CurPath[i+0].x), p0y(CurPath[i+0].y);
				float p1x(CurPath[i+1].x), p1y(CurPath[i+1].y);
				float p2x(CurPath[i+2].x), p2y(CurPath[i+2].y);
				float p3x(CurPath[i+3].x), p3y(CurPath[i+3].y);

				float nx(p2x - p3x), ny(p3y - p2x);
				float vx(p1x - p0x), vy(p1y - p0y);

				real8 denom = vx*nx + vy*ny; 
				if(denom != 0.0f)
				{
					real8 dx(p2x - p0x), dy(p2y - p0y);
					real8 t = (dx*nx + dy*ny) / denom;
					vx *= t; vy *= t; vx += p0x; vy += p0y;
					waypoints.push_back(Vec2D(myRound(vx), myRound(vy)));
				}
			}
		}

		//calculate and place the last vertex on the curve
		Vec2D l(myRound(CurPath[CurPath.size() - 1].x), myRound(CurPath[CurPath.size() - 1].y));
		if ((l.x != waypoints.back().x) && (l.y != waypoints.back().y))
		{
			waypoints.push_back(l);
		}

 		//reverse order
  		for (uint4 i = 0; i <= (waypoints.size()-1)/2; i++)
		{
			std::swap(waypoints[i], waypoints[waypoints.size()-1-i]);
  		}

		return Movement::MakeWaypointPath(waypoints, goal);
	}
 
	Pathfinder::ptr MakeTriangulationPathfinder(TraDcdt ** pPF)
    {
		std::auto_ptr<TraDcdt> pfDcdt(new TraDcdt());
		if(pPF) (*pPF) = pfDcdt.get();
		return Pathfinder::ptr(new TRPathfinder(pfDcdt));
    }
}

#endif
