#include "Movement.H"

namespace Movement
{
	Vec2D findPointOnLine(const Vec2D & loc, const Vec2D & line_start, const Vec2D & line_end)
	{
		sint4 line_x = line_end.x - line_start.x;
		sint4 line_y = line_end.y - line_start.y;

		if (line_x == 0 && line_y == 0) return loc;

		sint4 v_x = loc.x - line_start.x;
		sint4 v_y = loc.y - line_start.y;

		sint4 a = line_x * v_x + line_y * v_y;
		sint4 b = line_x * line_x + line_y * line_y;

		return Vec2D(line_start.x + line_x * a / b, line_start.y + line_y * a / b);
	}

	class WaypointPath : public Path
	{
	public:
		typedef boost::shared_ptr<WaypointPath> ptr;
		typedef boost::shared_ptr<const WaypointPath> const_ptr;
		typedef boost::weak_ptr<WaypointPath> wptr;
		typedef boost::weak_ptr<const WaypointPath> const_wptr;
	private:
		const_wptr self;
		Vector<Vec2D> waypoints;
		Goal::const_ptr goal;
		mutable real8 length;
	public:
		WaypointPath(const Vector<Vec2D> & waypoints, const Goal::const_ptr & goal) : 
			waypoints(waypoints), goal(goal), length(0.0) {}
		~WaypointPath() {}

		void setSelfPointer(const const_wptr & self) { this->self = self; }

		sint4 getWaypointCount() const { return waypoints.size(); }
		const Vec2D & getWaypoint(sint4 index) const { return waypoints[index]; }

		Vec2D getStart() const { return waypoints.back(); }
		Goal::const_ptr getGoal() const { return goal; }
		real8 getLength() const 
		{ 
			// If length not yet calculated
			if(length == 0.0)
			{
				// Calculate the length of the path
				for(size_t i=1; i<waypoints.size(); i++)
				{
					length += (waypoints[i] - waypoints[i-1]).length();
				}
			}
			// Return the length of the path
			return length;
		}

		bool intersectsSegment(const Vec2D & a, const Vec2D & b) const
		{
			TerrainBase::Segment s( TerrainBase::Loc(a.x, a.y), TerrainBase::Loc(b.x, b.y) );
			sint4 length = waypoints.size();
            for(sint4 i = 1; i < length; ++i)
			{
				TerrainBase::Segment t( TerrainBase::Loc( waypoints[i-1].x, waypoints[i-1].y ),
									    TerrainBase::Loc( waypoints[i].x,   waypoints[i].y   ) );
				if( s.touches(t) )
				{
					return true;
				}
			}
			return false;
		}

		Path::ptr getReversed() const 
		{ 
			Vector<Vec2D> revWaypoints;
			std::copy(waypoints.rbegin(), waypoints.rend(), std::back_inserter(revWaypoints));
			return Path::ptr(new WaypointPath(revWaypoints, TouchPoint(waypoints.back())));
		}

		Progress::ptr getProgress(const Vec2D & startPos, real8 lookahead) const;
	};

	class WaypointProgress : public Progress
	{
		WaypointPath::const_ptr path;
		sint4 nextWaypoint;
		real8 lookahead;
		Vec2D nextTarget;
		Vec2D lastSpot;
	public:
		WaypointProgress(WaypointPath::const_ptr path, sint4 nextWaypoint, real8 lookahead) :
			path(path), nextWaypoint(nextWaypoint), lookahead(lookahead) {}
		virtual ~WaypointProgress() {}

		void updateState(const Vec2D & position)
		{
			lastSpot = position;
		}

		void updateTarget(const Vec2D & position)
		{
			Vec2D attractor(position);
			if(nextWaypoint < path->getWaypointCount() - 1)
			{
				attractor = findPointOnLine(position, 
											path->getWaypoint(nextWaypoint), 
											path->getWaypoint(nextWaypoint+1));
			}
			Vec2D target = path->getWaypoint(nextWaypoint);
			Vec2D to_wp = target - attractor;
			Vec2D scaled_to_wp;

			real8 offset = lookahead;
			if (to_wp.length() == 0)
			{
				scaled_to_wp = Vec2D(1,1);
			}
			else
			{
				scaled_to_wp = to_wp.multiply(offset / to_wp.length());
			}

			bool wp_attract = false;
			//If wp_attract is set, set the attractor to the wp unless you're near it.
			if (wp_attract && scaled_to_wp.length() <= to_wp.length())
			{
				nextTarget = target;
			}

			while (scaled_to_wp.length() > to_wp.length())
			{
				offset = offset - to_wp.length();
				attractor = path->getWaypoint(nextWaypoint);

				if ( nextWaypoint == 0 )
				{
					scaled_to_wp = Vec2D(0,0);
					break;
				}
				--nextWaypoint;

				target = path->getWaypoint(nextWaypoint);
				to_wp = Vec2D(target.x - attractor.x, target.y - attractor.y);
				scaled_to_wp = to_wp.multiply(offset / to_wp.length());
			}

			attractor.x = attractor.x + scaled_to_wp.x;
			attractor.y = attractor.y + scaled_to_wp.y;
			nextTarget = attractor;
		}

		bool isProgressing(const Vec2D & position) const
		{
			if( nextWaypoint < 0 ) return true;
			Vec2D myTarget = path->getWaypoint(nextWaypoint);
			sint4 distanceSq = (myTarget - position).lengthSq();
			sint4 spotDistanceSq = (myTarget - lastSpot).lengthSq();
			return distanceSq < spotDistanceSq;
		}

		Vec2D getTarget() const
		{
			return nextTarget;
		}
	};

	Progress::ptr WaypointPath::getProgress(const Vec2D &, real8 lookahead) const
	{
		return Progress::ptr(new WaypointProgress(self.lock(), waypoints.size() - 1, lookahead));
	}

	Path::ptr MakeWaypointPath(const Vector<Vec2D> & waypoints, const Goal::const_ptr & goal)
	{
		WaypointPath::ptr path(new WaypointPath(waypoints, goal));
		path->setSelfPointer( WaypointPath::const_wptr( path ) );
		return Path::ptr(path);
	}
}
