#ifndef ST_PFEngine_H
#define ST_PFEngine_H

// $Id: ST_PFEngine.H 4209 2007-04-10 17:55:37Z atelliot $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

#include "ST_SimpleMap.H"
#include "ST_Bucket.H"
#include "SimpleTerrain.H"

#include "boost/smart_ptr.hpp"

namespace SimpleTerrain {

  // ------------------------------------------------------------------------------
  /** Adapt SimpleMap to use a given Resolution Factor. 
      The map will represent each RF X RF integer coordinate points of 
      the world as a single integer coordinate.
      It is assumed that w_ and h_ are divisible by RF.
      All of SimpleMap's cell access functions assume input coords are given
      in the *inner* scale, and not in the world scale.

      The main point for this intermediate class is code clarity and a joint
      interface for AStar maps and SizeCell maps.
  */
  template <typename T>
  struct ArbitraryResolutionSimpleMap : public SimpleMap<T>
  {
    /** 
    Construct an empty map according to given size specifications.
    All T cells in the map are constructed using their default ctor.
    @param w_                 Width (in *world* coordinates) of the map.
    @param h_                 Height (in *world* coordinates) of the map.
    @param resolution_factor_ Factor of decreased resolution; 1 => no decrease.
    */
    ArbitraryResolutionSimpleMap(sint4 w_, sint4 h_, sint4 resolution_factor_)
        // Keep a down-sized internal map, (+1 for what? find out/ask Sami)
      : SimpleMap<T>(w_/resolution_factor_ + 1, h_/resolution_factor_ + 1),
        resolution_factor(resolution_factor_), disp(resolution_factor_/2)
    {
      assert( (w_ % resolution_factor)==0 && "Width must be divisible by resolution_factor");
      assert( (h_ % resolution_factor)==0 && "Height must be divisible by resolution_factor");
    };

    //---------------------------------------------------------------------
    /// \name Coordinate translation: world <-> inner representation
    //---------------------------------------------------------------------
    //@{
    sint4 world2x(sint4 worldx) const { return (worldx+disp) / resolution_factor; };
    sint4 world2y(sint4 worldy) const { return (worldy+disp) / resolution_factor; };
    sint4 x2world(sint4 x) const { return x * resolution_factor; };
    sint4 y2world(sint4 y) const { return y * resolution_factor; };
    sint4 world2r(sint4 worldr) const {
      sint4 r = worldr / resolution_factor;
      return (worldr%resolution_factor >= (resolution_factor+1)/2)? r+1 : r;
    };
    /// This is max[ abs( c - x2world(world2x(c)) )]
    sint4 max_error_from_conversion() const { return resolution_factor/2; };
    //@}
    //---------------------------------------------------------------------

  protected:
    // For coordinate translations
    sint4 resolution_factor;      ///< how many world points per one point in our representation
    sint4 disp;                   ///< objects are displaced half an (internal) cell so boundaries fall in the middle of tiles
  };


// ------------------------------------------------------------------------------
// ------------------------------------------------------------------------------
// ------------------------------------------------------------------------------

  class ST_Terrain::PFEngine {

  // ------------------------------------------------------------------------------
    /** Extend the SimpleMap with set/access methods to allow multiple levels-of-detail.
        This factors out much of the old functionality of PFEngine, and the 
        implementation is tightly coupled to PFEngine, so it's an internal class.
    */
    class Map : public ArbitraryResolutionSimpleMap<SizeCell> {

    public:
      Map(sint4 w_, sint4 h_, sint4 resolution_factor_, sint4 client_pID_, bool include_vision_=false) 
        : ArbitraryResolutionSimpleMap<SizeCell>(w_, h_, resolution_factor_), client_pID(client_pID_), include_vision(include_vision_) {};

      //---------------------------------------------------------------------
      /// \name Changes to the world.
      //---------------------------------------------------------------------
      //@{
      void insert_object(const Object *obj);
      void remove_object(const Object *obj);
      void update_object(const Object *obj);
      void insert_boundary(const Segment& line);

      // insert/remove the given path (on the ground) as an obstacle to avoid.
      void ir_path(const Object *obj, const Path &path, bool insert);
      //@}
      //---------------------------------------------------------------------

    private:
      sint4 client_pID;
      bool include_vision;

      //---------------------------------------------------------------------
      /// \name Internal machinery for updating the map
      //---------------------------------------------------------------------
      //@{
      void ir_object(const Object *obj, bool insert,
        bool use_prev_attributes=false);                              ///< Expects world coordinates
      void insert_rectangle(const Loc &topleft,const Loc &bottomright, 
                            sint4 cost);                              ///< Expects world coordinates

      void insert_bucket(const Bucket &b, sint4 delta_maxsize, 
                         sint4 cost);                                 ///< Expects inner coordinates
      void insert_line(sint4 x1, sint4 y1, sint4 x2, sint4 y2,
                       sint4 delta_maxsize, sint4 cost);              ///< Expects inner coordinates
      void set_cell(sint4 x, sint4 y,
                    sint4 delta_maxsize, sint4 cost);                 ///< Expects inner coordinates
      //@}
      //---------------------------------------------------------------------

      // prevent copies
      Map(const Map &);
      Map& operator=(const Map &);
    };

    friend class Map;

// ------------------------------------------------------------------------------
// ------------------------------------------------------------------------------

  public:
    typedef uint4 LevelOfDetail;  ///< index into the LOD array passed to the ctor

    /**
    @param tiles_x_       Width of world in tiles.
    @param tiles_y_       Height of world in tiles.
    @param tile_points_   How many points per tile are in the full world representation.
    @param lods_needed    Vector indicating levels of detail of inner world representation needed.
                          1 - full details, resolution as in the world rep;
                          2 - half resolution; 4 - quarter resultion; etc.
                          (tiles_x_*tile_points_) and (tiles_y_*tile_points_) must be divisible by each LOD.
    */
    PFEngine(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, const Vector<sint4>& lods_needed, sint4 client_pID_);

    //---------------------------------------------------------------------
    /// \name Changes to the world.
    //---------------------------------------------------------------------
    //@{
    /** Insert a given object to the internal world representation.
        Pointers must remain valid until a respective remove_object() call. */
    void insert_object(const Object *obj);
    void remove_object(const Object *obj);
    void update_object(const Object *obj);
    void insert_boundary(const Segment& line);

    // insert/remove the given path (on the ground) as an obstacle to avoid.
    void ir_path(const Object *obj, const Path &path, bool insert);
    //@}
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    /// \name Perform path finding 
    //---------------------------------------------------------------------
    //@{
#if 0
    /** Flags that specify what objects to consider as obstacles. 
        Currently the range of options is limited, but this will be 
        expanded in the future. */
    enum ConsiderObjectsFlags { 
      CONSIDER_ALL      = 0,
      /** Specify IGNORE_MOVED_OBJ to disregard the moving object itself 
          as an obstacle (in it's current==start position).
          This is the default behaviour, but when PFing for a fake object
          (like planning a future move, planning a wide lane etc.) this flag
          must NOT be specified (or else the world view will be corrupted). */
      IGNORE_MOVED_OBJ            = 0x0001,
      /// Ignore all friendly (owned) mobile objects (subsumes IGNORE_MOVED_OBJ). Unimplemented yet.
      IGNORE_MOBILE_FRIEND_OBJS   = 0x0010 | IGNORE_MOVED_OBJ, 
      /// Ignore all enemy mobile objects. Unimplemented yet.
      IGNORE_MOBILE_FOE_OBJS      = 0x0100,
      /// Ignore all neutral ("sheep") mobile objects. Unimplemented yet.
      IGNORE_MOBILE_NEUTRAL_OBJS  = 0x1000,
      /** Ignore all mobile objects, considering only boundaries and non-mobile objects.
          Note that with the current implementation this amounts to removing all objects
          and then reinserting them, which might be a little costly. */
      IGNORE_ALL_MOBILE_OBJS      = IGNORE_MOBILE_FRIEND_OBJS | IGNORE_MOBILE_FOE_OBJS | IGNORE_MOBILE_NEUTRAL_OBJS,
    };
#endif
    /** Perform pathfinding to a location, return true iff path found, and the path itself in 'output'.
        If found then output is guaranteed to include at least one waypoint for the goal. 
        Waypoints are listed in reverse order, goal is in locs[0]. */
    bool find_path(const Object *obj, const Loc &goal, Path& output, LevelOfDetail lod = 0, ConsiderObjects consider = CONSIDER_ALL, uint4 stop_time = 0);
    /** Perform pathfinding to a target object, return true iff path found, and the path itself in 'output'
        If found then output is guaranteed to include at least one waypoint for the goal. 
        Waypoints are listed in reverse order, goal is in locs[0]. */
    bool find_path(const Object *obj, const Object *goal, Path& output, LevelOfDetail lod = 0, ConsiderObjects consider = CONSIDER_ALL, uint4 stop_time = 0);
    /** Perform pathfinding to a target object, return true iff path found, and the path itself in 'output'
        If found then output is guaranteed to include at least one waypoint for the goal. 
        Waypoints are listed in reverse order, goal is in locs[0]. */
    bool sub_path(const Object *obj, const Loc &topLeft, const Loc &bottomRight, const Object *goal_obj, Path &output, LevelOfDetail lod = 0, ConsiderObjects consider = CONSIDER_ALL, uint4 stop_time = 0);
    /** Perform pathfinding on a small submap, to a location, return true iff path found, and the path 
        itself in 'output'. If found then output is guaranteed to include at least one waypoint for 
        the goal. Waypoints are listed in reverse order, goal is in locs[0]. */
    bool sub_path(const Object *obj, const Loc &topLeft, const Loc &bottomRight, const Loc &goal_loc, Path &output, LevelOfDetail lod = 0, ConsiderObjects consider = CONSIDER_ALL, uint4 stop_time = 0);
    //@}
    //---------------------------------------------------------------------

  private:

    typedef ArbitraryResolutionSimpleMap<AStarCell> AStarMap;
    typedef Vector< boost::shared_ptr<Map> >        MapsVector;

    Vector<sint4> lods;                       ///< Levels Of Detail available.
    LevelOfDetail max_details_LOD;            ///< The most detailed level.
    MapsVector    ground_maps;
    MapsVector    vision_maps;
    MapsVector    air_maps;
    Vector< boost::shared_ptr<AStarMap> >  astar_maps;
    std::set<const Object*> mobile_objects;  ///< Set of all mobile objects, allows ignoring them during PFing
    sint4 client_pID;


    /// Initialize a given map to be empty with a blocked border.
    static void init_sizes(SimpleMap<SizeCell> &smap);

    //---------------------------------------------------------------------
    /// \name Actual computation of paths
    //---------------------------------------------------------------------
    //@{
    /** This implementation can handle multiple target locations and can 
        handle a target object. To this end it accepts a predicate for 
        testing whether a specific map location (inner coords + map pointer) 
        is an allowable goal location.
        It currently doesn't support a pluggable distance huristic, but 
        instead accepts 'the goal' location (probably the center of the 
        target) and uses the usual octile distance huristic. 
        This combination results in an inadmissible huristic, but should 
        give reasonable results without resorting to the slower huristic
        h = min distance to all goal locations. */
    template <class GoalTester>
    bool find_path(const Object *obj,             ///< the object to move
                   const Loc &start_position,     ///< Move it from where (not necessarily its current position)
                   const Object *goal_obj,               ///< the goal location used for distance estimation
                   GoalTester goal_tester,        ///< a predicate for an AStarCell which is true iff it is a goal location
                   const Loc &topleft,            ///< topleft corner of the submap to explore for this path; use Loc(0,0) to specify the entire map
                   const Loc &bottomright,        ///< bottomright corner of the submap to explore for this path; use Loc(-1,-1) or smaller to specify the entire map
                   LevelOfDetail lod,             ///< the LOD to use
                   ConsiderObjects consider,      ///< what objects to consider/ignore as obstacles
                   Path& output,                  ///< output path is returned here
                   uint4 stop_time,              ///< stop and report failure when SDL_GetTicks() >= stop_time; 0 => no time limit
                   bool use_vision=true);

    //---------------------------------------------------------------------
    /// \name Actual computation of paths
    //---------------------------------------------------------------------
    //@{
    /** This implementation can handle multiple target locations and can 
        handle a target object. To this end it accepts a predicate for 
        testing whether a specific map location (inner coords + map pointer) 
        is an allowable goal location.
        It currently doesn't support a pluggable distance huristic, but 
        instead accepts 'the goal' location (probably the center of the 
        target) and uses the usual octile distance huristic. 
        This combination results in an inadmissible huristic, but should 
        give reasonable results without resorting to the slower huristic
        h = min distance to all goal locations. */
    template <class GoalTester>
    bool find_path(const Object *obj,             ///< the object to move
                   const Loc &start_position,     ///< Move it from where (not necessarily its current position)
                   const Loc &goal,               ///< the goal location used for distance estimation
                   GoalTester goal_tester,        ///< a predicate for an AStarCell which is true iff it is a goal location
                   const Loc &topleft,            ///< topleft corner of the submap to explore for this path; use Loc(0,0) to specify the entire map
                   const Loc &bottomright,        ///< bottomright corner of the submap to explore for this path; use Loc(-1,-1) or smaller to specify the entire map
                   LevelOfDetail lod,             ///< the LOD to use
                   ConsiderObjects consider,      ///< what objects to consider/ignore as obstacles
                   Path& output,                  ///< output path is returned here
                   uint4 stop_time,              ///< stop and report failure when SDL_GetTicks() >= stop_time; 0 => no time limit
                   bool use_vision=true);

    //---------------------------------------------------------------------
    /// \name Actual computation of paths. Loc to Loc 
    //---------------------------------------------------------------------
    //@{
    /** This implementation can handle multiple target locations and can 
        handle a target location. To this end it accepts a predicate for 
        testing whether a specific map location (inner coords + map pointer) 
        is an allowable goal location.
        It currently doesn't support a pluggable distance huristic, but 
        instead accepts 'the goal' location (probably the center of the 
        target) and uses the usual octile distance huristic. 
        This combination results in an inadmissible huristic, but should 
        give reasonable results without resorting to the slower huristic
        h = min distance to all goal locations. */
    template <class GoalTester>
    bool find_path(const sint4 radius,             ///< the radius to calculate
                   const Loc &start_position,     ///< Move it from where (not necessarily its current position)
                   const Loc &goal,               ///< the goal location used for distance estimation
                   GoalTester goal_tester,        ///< a predicate for an AStarCell which is true iff it is a goal location
                   const Loc &topleft,            ///< topleft corner of the submap to explore for this path; use Loc(0,0) to specify the entire map
                   const Loc &bottomright,        ///< bottomright corner of the submap to explore for this path; use Loc(-1,-1) or smaller to specify the entire map
                   LevelOfDetail lod,             ///< the LOD to use
                   ConsiderObjects consider,      ///< what objects to consider/ignore as obstacles
                   Path& output,                  ///< output path is returned here
                   uint4 stop_time,               ///< stop and report failure when SDL_GetTicks() >= stop_time; 0 => no time limit
                   bool use_vision=true);


    /** \brief The actual A* implementation.
        Return true iff a path was found, and the goal location (cell index in astar_map)
        that was flagged by 'goal_tester' in 'end_of_path'. 
        Only exapnd nodes in the submap defined by 'topleft' and 'bottomright'. */
    template<class GoalTester>
    bool simple_astar(const Map &world_view, 
                      AStarMap &astar_data,
                      Bucket to_move, 
                      sint4 x2_1, sint4 y2_1,
                      sint4 x2_2, sint4 y2_2,
                      GoalTester goal_tester,
                      const Loc &topleft,
                      const Loc &bottomright,
                      uint4 &end_of_path,
                      uint4 stop_time);

    /// Remove objects we wish to ignore for PFing from the world map
    void ignore_objects(ConsiderObjects consider, boost::shared_ptr<Map> map, const Object *moved_obj);
    /// Return the objects we ignored for the last PFing back to the world map
    void reconsider_objects(ConsiderObjects consider, boost::shared_ptr<Map> map, const Object *moved_obj);

    /// Remove objects we wish to ignore for PFing from the world map
    void ignore_objects(ConsiderObjects consider, boost::shared_ptr<Map> map);
    /// Return the objects we ignored for the last PFing back to the world map
    void reconsider_objects(ConsiderObjects consider, boost::shared_ptr<Map> map);

    /// Remove redundant waypoints (that fall on a straight line)
    void smoothen_path(Vector<Loc> &path);
    /// Compute the octile distance between two points
    uint4 estimate(sint4 x1, sint4 y1, sint4 x2, sint4 y2);
    /// Check whether the given three points are on a straight line
    bool on_line(const Loc &a, const Loc &b, const Loc &c);
    //@}
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    /// \name Predicates for testing arrival to different types of goals
    //---------------------------------------------------------------------
    //@{
    /** Helper predicate for testing whether moving an object to a given location
        will place it adjacent to a given target, considering the shapes and sizes 
        of the object and target. */
    class IsTouchingTarget; friend class IsTouchingTarget;

    /** Helper predicate for testing whether a tested location is equal to the 
        target location. */
    class IsAtLocation; friend class IsAtLocation;

    /** Helper predicate for going to the closest clear location, i.e. one
        with no other overlapping objects (might happen when using a low resolution). */
    class IsAtClearLocation; friend class IsAtClearLocation;
    //@}
    //---------------------------------------------------------------------

  public:
    //---------------------------------------------------------------------
    /// \name Cost constants
    //---------------------------------------------------------------------
    //@{
    static const uint4 FACTOR = 128; ///< distance multiplier (real->int)
    static const uint4 DIAG   = sint4(FACTOR*1.4142);
    /// cost of passing though a square occupied by another unit
    static const sint4 UNIT_COST = 1000000;   // signed to allow decreasing the cost of a cell
    /// cost of passing though a square occupied by and orts boundery
    static const sint4 BOUND_COST = 5000000;  // signed to allow decreasing the cost of a cell
    /** cost of passing though a square within the direction of
    * another units x-second distance
    * x-second distance: if a unit continues to move in a straight
    * line it will be there in x-seconds */
    //static const uint4 CROSS_COST = 500;
    //@}
    //---------------------------------------------------------------------

  private:
    /// Serving the graphical widget that is used for developement.
    boost::shared_ptr<const Map> get_ground_map() const { return ground_maps[0]; };
    boost::shared_ptr<const Map> get_air_map() const { return air_maps[0]; };
    friend class SimpleTerrainWidget;
    friend class SimpleTerrainOverlay;
  };

} // end of namespace SimpleTerrain

#endif
