#include <assert.h>
#include "GridAbstraction.H"
#include "History.H"
#include "WatcherAux.H"
#include "GameObj.H"
#include "GameStateModule.H"
#include "GameObjFactory.H"
#include "Blueprints.H"
#include "ModuleSet.H"

using History::HistoryModule;
using History::Record;

namespace TerrainAnalysis
{
    GridAbstraction::GridAbstraction(ModuleSet & ms,
                                     const Rect & _area) : 
        ms(ms),
        gsm(ms.get_game_state()),
        game(gsm.get_game()),
        ppt(game.get_tile_points()),
        height(-1),
        valid(false)
    {
        const Map<GameTile> & map   = game.get_map();

        area = _area;

        if( area.x1 < area.x0 ) std::swap( area.x0, area.x1 );
        if( area.y1 < area.y0 ) std::swap( area.y0, area.y1 );

        if( area.x0 < 0 ) area.x0 = 0;
        if( area.y0 < 0 ) area.y0 = 0;

        if( area.x1 > map.get_width() ) area.x1 = map.get_width();
        if( area.y1 > map.get_height() ) area.y1 = map.get_height();

        dim = area.size();
        obs = new uint1[dim.x * dim.y];
    }

    GridAbstraction::GridAbstraction(ModuleSet & ms, 
                                     sint4 center_x, 
                                     sint4 center_y, 
                                     sint4 radius) :
        ms(ms),
        gsm(ms.get_game_state()),
        game(gsm.get_game()),
        ppt(game.get_tile_points()),
        height(0),
        valid(false)
    {
        const Map<GameTile> & map   = game.get_map();
        
        area.x0 = (center_x - radius) / ppt;
        area.y0 = (center_y - radius) / ppt;
        area.x1 = (center_x + radius + (ppt-1)) / ppt;
        area.y1 = (center_y + radius + (ppt-1)) / ppt;

        if( area.x1 < area.x0 ) std::swap( area.x0, area.x1 );
        if( area.y1 < area.y0 ) std::swap( area.y0, area.y1 );

        if( area.x0 < 0 ) area.x0 = 0;
        if( area.y0 < 0 ) area.y0 = 0;

        if( area.x1 > map.get_width() ) area.x1 = map.get_width();
        if( area.y1 > map.get_height() ) area.y1 = map.get_height();

        dim = area.size();
        obs = new uint1[dim.x * dim.y];
    }

    GridAbstraction::~GridAbstraction()
    {
        if(obs != NULL)
        {
            delete[] obs;
        }
    }

    bool GridAbstraction::get_building_size(const std::string &bp_name, 
                                            Size & result) const
    {
        const PlayerInfo & player  = game.get_cplayer_info();
        GameObjFactory & gof       = const_cast<GameObjFactory&>(player.gof);
        GameObjBlueprint * bp      = gof.get_blueprint(bp_name);
        if(bp == NULL) { return false; }
        
        sint4 width_id = bp->attr_index("width");
        if(width_id == -1) { return false; }

        sint4 height_id = bp->attr_index("height");
        if(height_id == -1) { return false; }

        result.x = bp->attr_vals[width_id].to_int();
        result.y = bp->attr_vals[height_id].to_int();
        return true;
    }

    void GridAbstraction::block_spot(sint4 x, sint4 y) const
    {
        if(x < area.x0) return;
        if(y < area.y0) return;
        if(x >= area.x1) return;
        if(y >= area.y1) return;

        obs[(y - area.y0) * dim.x + (x - area.x0)] = 1;
    }

    void GridAbstraction::block_terrain() const
    {
        const Game & game         = gsm.get_game();
        const Map<GameTile> & map = game.get_map();

        uint1 * data = obs;
        for(sint4 y = area.y0; y < area.y1; y++)
        {
            for(sint4 x = area.x0; x < area.x1; x++)
            {
                const GameTile & tile = map.get_tile(x, y);
                Tile::Type tw, te; tile.get_type(tw, te);
                bool blocked = false;
                blocked |= (height != tile.get_min_h());
                blocked |= (height != tile.get_max_h());
                blocked |= (tw == Tile::UNKNOWN);
                blocked |= (te == Tile::UNKNOWN);
                *data++ = (blocked ? 1 : 0);
            }
        }
    }

    void GridAbstraction::block_obj(GameObj * obj) const
    {
        sint4 border = get_object_border( obj->bp_name() );
        sint4 x, y, x0, x1, y0, y1;
        if(obj->get_shape() == Object::CIRCLE)
        {
            obj->get_center(x, y);
            border += obj->get_radius();
            x0 = (x - border) / ppt;
            y0 = (y - border) / ppt;
            x1 = (x + border) / ppt;
            y1 = (y + border) / ppt;
        }
        else if(obj->get_shape() == Object::RECTANGLE)
        {
            obj->get_p1(x0,y0);
            obj->get_p2(x1,y1);
            x0 = (x0 - border) / ppt;
            y0 = (y0 - border) / ppt;
            x1 = (x1 + border) / ppt;
            y1 = (y1 + border) / ppt;
        }
        else
        {
            return;
        }

        // TODO: Optimize by clipping rects
        for(y = y0; y <= y1; y++)
        {
            for(x = x0; x <= x1; x++)
            {
                block_spot(x, y);
            }
        }
    }

    void GridAbstraction::update() const
    {
        block_terrain();

        const HistoryModule & hist = ms.get_history();

        records.clear();
        hist.getActiveRecords(records);
        FORALL(records, r)
        {
            const Record * record = (*r);
  	    assert( record->isActive() );
            block_obj( record->getObject() );
        }
        
        uint1 * data = obs;
        for(sint4 y = 0; y < dim.y; y++)
        {
            for(sint4 x = 0; x < dim.x; x++)
            {
                std::cerr << ((*data++) ? '#' : ' ');
            }
            std::cerr << std::endl;
        }

        valid = true;
    }

    bool GridAbstraction::is_blocked(sint4 x, sint4 y) const
    {
        if( !area.contains(x, y) ) return true;

        if( !valid ) update();

        return (obs[(y - area.y0) * dim.x + (x - area.x0)] != 0);
    }

    bool GridAbstraction::is_blocked(const Rect & rect) const
    {
        assert( rect.x0 <= rect.x1 );
        assert( rect.y0 <= rect.y1 );

        if( rect.x0 < area.x0 ) return true;
        if( rect.y0 < area.y0 ) return true;
        if( rect.x1 > area.x1 ) return true;
        if( rect.y1 > area.y1 ) return true;

        if( !valid ) update();

        for(sint4 y = rect.y0; y <= rect.y1; y++)
        {
            uint1 * data = obs + (y - area.y0) * dim.x + (rect.x0 - area.x0);
            for(sint4 x = rect.x0; x <= rect.x1; x++)
            {
                if( *data++ != 0 )
                {
                    return true;
                }
            }
        }

        return false;
    }
    
    void GridAbstraction::set_object_border(const std::string & bp_name,
                                            sint4 border)
    {
        borders[bp_name] = border;
    }

    sint4 GridAbstraction::get_object_border(const std::string & bp_name) const
    {
        std::map<std::string, sint4>::const_iterator it;
        it = borders.find(bp_name);
        return ( it != borders.end() ) ? it->second : 0;
    }

    void GridAbstraction::set_passable_height(sint4 /*passHeight*/)
    {
        // Something more robust will be needed in the long term
        height = 0;
        valid = false;
    }

    sint4 GridAbstraction::sample_height(const TerrainBase::Loc & point) const
    {
        const Map<GameTile> & map   = game.get_map();

        sint4 px = point.x / ppt;
        sint4 py = point.y / ppt;
     
        if(!Rect(0, 0, map.get_width(), map.get_height()).contains(px, py))
        {
            return -1;
        }

        const GameTile & nearTile = map.get_tile(px, py);
        if(nearTile.get_min_h() != nearTile.get_max_h())
        {
            return -1;
        }

        return nearTile.get_min_h();
    }

    sint4 GridAbstraction::eval_nearest(const TerrainBase::Loc & candidate,
                                          const TerrainBase::Loc & near) const
    {
        sint4 dx = near.x - candidate.x;
        sint4 dy = near.y - candidate.y;
        return (dx*dx + dy*dy);
    }

    sint4 GridAbstraction::eval_prox_min(const TerrainBase::Loc & candidate) const
    {
        sint4 accum = 0;
        FORALL(records, r)
        {
            const Record * record = (*r);
  	    assert( record->isActive() );
            const GameObj * obj = record->getObject();
            if(obj->bp_name() == "mineral")
            {
                sint4 x, y; obj->get_center(x, y);
                sint4 dx = x - candidate.x;
                sint4 dy = y - candidate.y;
                accum -= 1024 / (dx*dx + dy*dy + 1);
            }
        }        
        return accum;
    }

    bool GridAbstraction::find_free_spot(const TerrainBase::Loc & near,
                                         real8 maxDistance,
                                         TerrainBase::Loc & result) const
    {
        sint4 limit = static_cast<sint4>(maxDistance)+1;
        Rect consider( near.x - limit, 
                       near.y - limit,
                       near.x + limit,
                       near.y + limit );
        Size considerDim = consider.size();

        real8 maxDistSq = maxDistance * maxDistance;

        for(size_t i=0; i<100; i++)
        {
            sint4 x = consider.x0 + rand() % considerDim.x;
            sint4 y = consider.y0 + rand() % considerDim.y;
            sint4 nx = x / ppt;
            sint4 ny = y / ppt;
            if( !is_blocked(nx, ny) )
            {
                sint4 dx = near.x - x;
                sint4 dy = near.y - y;
                if( dx*dx + dy*dy < maxDistSq )
                {
                    result.x = x;
                    result.y = y;

                    std::cerr << "Free spot at " << x;
                    std::cerr << ", " << y << std::endl;

                    return true;
                }
            }
        }

        return false;
    }

    size_t GridAbstraction::place_building(const std::string & building,
                                           const TerrainBase::Loc & near,
                                           Vector<TerrainBase::Loc> & results) const
    {
        std::multimap<sint4, TerrainBase::Loc> buildSites;

        Size buildSize;
        if(!get_building_size(building, buildSize))
        {
            return false;
        }
        sint4 border = get_object_border(building);

        buildSize.x = (buildSize.x + (2*border) + (ppt-1)) / ppt;
        buildSize.y = (buildSize.y + (2*border) + (ppt-1)) / ppt;

        sint4 dx = dim.x - buildSize.x;
        sint4 dy = dim.y - buildSize.y;

        size_t max = (size_t)((dim.x / buildSize.x) * (dim.y / buildSize.y))*8;

        for(size_t i=0; i<max; i++)
        {
            sint4 bx0 = area.x0 + rand() % dx;
            sint4 by0 = area.y0 + rand() % dy;

            sint4 bx1 = bx0 + buildSize.x;
            sint4 by1 = by0 + buildSize.y;

            if( !is_blocked( Rect( bx0, by0, bx1, by1 ) ) )
            {
                TerrainBase::Loc site;
                site.x = (bx0 + bx1) * ppt / 2;
                site.y = (by0 + by1) * ppt / 2;

                // More hacks, let's make this more general in the future
                // Possibly a set of attractors/repulsors or something
                sint4 value = eval_nearest(site, near);

                buildSites.insert( std::make_pair(value, site) );
            }
        }

        std::multimap<sint4, TerrainBase::Loc>::iterator it;
        for(it = buildSites.begin(); it != buildSites.end(); it++)
        {
			results.push_back( it->second );
        }
        return buildSites.size();
    }
}
