#ifndef __WATCHER_H__
#define __WATCHER_H__

#include <vector>
#include <set>
#include <map>

#include "boost/shared_ptr.hpp"

#include "EventHandler.H"

class GameObj;
class GameStateModule;

namespace Watcher
{
    typedef uint4 WatchID;
    const WatchID NONE = 0;


    // Base class for unit predicates
    class UnitPred
    {
    public:
        UnitPred() {}
        virtual ~UnitPred() {}

        virtual bool evaluate(const GameObj * unit) const = 0;
    };
    typedef boost::shared_ptr<UnitPred> UnitPredPtr;



    // Base class for set predicates
    class SetPred
    {
    public:
        SetPred() {}
        virtual ~SetPred() {}

        virtual bool evaluate(const std::set<GameObj*> & set) const = 0;
    };
    typedef boost::shared_ptr<SetPred> SetPredPtr;



    // Event fired due to trigger
    class ConditionEvent : public Event
    {
        WatchID id;
        std::set<GameObj*> units;
    public:
        ConditionEvent(WatchID id);
        ~ConditionEvent();

        WatchID getID() const;
        std::set<GameObj*> & getUnits();
        const std::set<GameObj*> & getUnits() const;
    };



    // Event fired due to object creation
    class CreationEvent : public Event
    {
        WatchID id;
        std::set<GameObj*> units;
    public:
        CreationEvent();
        ~CreationEvent();

        void setID(WatchID id);
        WatchID getID() const;
        std::set<GameObj*> & getUnits();
        const std::set<GameObj*> & getUnits() const;
    };



    // Event fired due to object death
    class DeathEvent : public Event
    {
        std::set<GameObj*> units;
    public:
        DeathEvent();
        ~DeathEvent();

        std::set<GameObj*> & getUnits();
        const std::set<GameObj*> & getUnits() const;
    };



    // Main module class
    class WatcherModule : public EventHandler
    {
        struct ConditionWatch 
        { 
            UnitPredPtr    query;      // Query for selecting units
            SetPredPtr     condition;  // Conditionto be passed by set
            EventHandler * handler;    // Handler to send event to
            uint4          ticks;      // Ticks since last check
            uint4          period;     // How many ticks before each check
            bool           autoDelete; // Automatically delete on success?

            ConditionWatch();
            ConditionWatch(const UnitPredPtr & query, 
                  const SetPredPtr & condition, 
                  EventHandler * handler,
                  uint4 period,
                  bool autoDelete);
            ConditionWatch & operator = (const ConditionWatch & o);
        };

        struct CreationWatch
        {
            UnitPredPtr    query;      // Query for selecting units
            EventHandler * handler;    // Handler to send event to
            uint4          priority;   // Priority for selection
            bool           autoDelete; // Automatically delete on success?

            CreationWatch();
            CreationWatch(const UnitPredPtr & query, 
                        EventHandler * handler,
                        uint4 priority,
                        bool autoDelete);
            CreationWatch & operator = (const CreationWatch & o);
        };

        GameStateModule&                        gsm;
        std::set<GameObj*>                      objects;
        std::multimap<GameObj*, EventHandler*>  mapDeathHandlers;
        std::map<WatchID, ConditionWatch>       mapWatches;
        std::map<WatchID, CreationWatch>        mapCreationWatches;
        WatchID                                 nextID;

        void handleNew();
        void handleDeaths();
        void checkWatches();
    public:
        WatcherModule(GameStateModule& gsm);
        ~WatcherModule();

        /** \brief to satisfy EventHandler requirements
         */
        bool handle_event(const Event &event);

        /** \brief query for the set of units satisfying a given query
         *  \param query the object containing the tests which a unit must pass
         *  \param result a set to be filled with the results of the query
         *  \return the number of units added to the set due to the query
         */
        size_t queryUnits(const UnitPredPtr & query, 
                          std::set<GameObj*> & result) const;

        /** \brief instruct the Watcher to notify a handler when a condition is true
         *  \param query the object containing the tests which a unit must pass
         *  \param condition a predicate that the set of units must pass to fire the event
         *  \param handler the handler to send the event to
         *  \param period the number of cycles which pass between updates
         *  \param autoDelete whether or not to delete the Watch on success
         *  \return a unique identifier which can be used to cancel the trigger
         *  \sa notifyOnCreation, notifyOnDeath, cancelNotify
         */ 
        WatchID notifyOnCondition(EventHandler * handler,
                                  const UnitPredPtr & query,
                                  const SetPredPtr & condition,
                                  uint4 period = 1,
                                  bool autoDelete = true);

        /** \brief instruct the Watcher to notify when unit created that passes query
         *  \param query object containing tests which unit must pass
         *  \param handler handler to be notified
         *  \param priority determines who takes precedence in accepting units
         *  \param autoDelete whether or not to delete the Watch on success
         *  \return unique identifier which can be used to cancel trigger
         *  \sa notifyOnCondition, notifyOnDeath, cancelNotify
         */
        WatchID notifyOnCreation(EventHandler * handler,
                                 const UnitPredPtr & query,
                                 uint4 priority,
                                 bool autoDelete = false);

        /** \brief instruct the Watcher to notify a handler when a unit dies
         *  \param unit the unit whose death will trigger notification
         *  \param handler the handler to be informed
         *  \sa notifyOnCondition, notifyOnCreation, cancelNotifyOnDeath
         */
        void notifyOnDeath(EventHandler * handler, GameObj * unit);

        /** \brief remove a particular event trigger from the list of active triggers
         *  \param id the unique identifier returned by notifyOnTrue()
         *  \sa notifyOnCondition, notifyOnCreation cancelNotifyOnDeath
         */
        void cancelNotify(WatchID id);

        /** \brief instruct the Watcher that it no longer needs to monitor unit
         *  \param handler the handler which no longer wants notification
         *  \param unit the unit to stop monitoring
         *  \sa notifyOnDeath, cancelNotify
         */
        void cancelNotifyOnDeath(EventHandler * handler, GameObj * unit);

        static const sint4 FROM; ///< returned by Event::get_who()

        static const sint4 CONDITION; ///< event ID for condition notifications
        static const sint4 CREATION; ///< event ID for creation notifications
        static const sint4 DEATH; ///< event ID for death notifications
    };
}

#endif // __WATCHER_H__
