

#include "BTask.h"

#include "BuildTasksAI.h"

#include "BuildTasksState.h"

namespace BSimul {

  Task::Task()
  {
    parent = NULL;
    state = NULL;
    playerID = 0;
    lineNum = 0;
    function = NULL;
    extraData = NULL;

    startTime = getCurrentTime();
    timeout = 0.0;
  }

  // this is the task constructor for when we have a parent
/*
  Task::Task(Task* parent, GroupVector& origGroups,
	     TaskCreateFunctionPtr function, std::string name,
	     bool baseAttackPreference, MoveStruct* extraData)
  */
  Task::Task(Task* parent, ObjectVector& origObjects,
             TaskCreateFunctionPtr function, std::string name,real8 timeout,
             MoveStruct* extraData)
  {
    if(parent == NULL) {
      ERR("we should always have a parent task in this Task constructor!!!");
    }
    this->parent = parent;
    playerID = parent->playerID;
    state = parent->state;

    if(function == NULL) {
      ERR("we should never have a NULL function in task constructor!");
    }
    this->function = function;
    this->name = name;
    this->timeout = timeout;
    this->extraData = extraData;

    lineNum = 0;

    if(origObjects.empty()) {
      ERR("warning,  trying to create a task with no objects, shouldn't be doing this!");
    }

    this->objects = origObjects;

    // add ourselves as a child of our parent
    // now we remove all the objects this child gets from the parent
    FORALL(objects,it) {

      bool isFound = false;
      FORALL(parent->objects,it2) {
	if(*it == *it2) {
	  parent->objects.erase(it2);
	  isFound = true;
	  break;
	}
      }
      if(!isFound) {
	ERR("BAD, group in parent should always exist when giving it to a child");
      }
    }
    parent->children.push_back(this);
    startTime = getCurrentTime();
  }

  // when we don't have a parent, this is the task constructor
/*
  Task::Task(uint4 playerID,GameState* state,GroupVector& origGroups,TaskCreateFunctionPtr function,
	     std::string name,bool baseAttackPreference,MoveStruct* extraData) {
*/
    Task::Task(uint4 playerID,BuildTasksState* state,ObjectVector& origObjects,TaskCreateFunctionPtr function,
               std::string name,real8 timeout,MoveStruct* extraData) {

    this->playerID = playerID;

    if(state == NULL) {
      ERR("warning: trying to set a null state for task, this should not happen");
    }
    this->state = state;
    this->parent = NULL;
    children.clear();

    if(origObjects.empty()) {
      ERR("warning,  trying to create a task with no groups, shouldn't be doing this!");
    }

    this->objects = origObjects; // we can just do a straight copy of this vector, since all it contains is ints.

    if(function == NULL) {
      ERR("we should never have a NULL function in parentless task constructor!");
    }
    this->function = function;
    this->name = name;

    this->timeout = timeout;
    this->extraData = extraData;
    lineNum = 0;

    startTime = getCurrentTime();
  }

  Task::Task(const Task& t) {
    objects = t.objects;
    playerID = t.playerID;
    name = t.name;
    startTime = t.startTime;
    timeout = t.timeout;
    function = t.function;
    lineNum = t.lineNum;
    if(t.extraData != NULL) {
      extraData = t.extraData->clone();
    }else{
      extraData = NULL;
    }
  }

  Task::~Task() {
    FORALL(children, it) {
      delete *it;
    }
    delete extraData;
    children.clear();
  }

  // get rid of unit with  from task and all subtasks
  void Task::removeObject(GameObj* obj) {
    FORALL(objects,it) {
      if(*it == obj) {
	objects.erase(it);
	break;
      }
    }
    // TODO: not sure if we even need this, it should be taken care of automatically
    FORALL(children,it) {
      (*it)->removeObject(obj);
    }
  }

  void Task::childCompleted(Task* child) {
    assert(child != NULL);

    // add our child's objects back to the parent once the child is done
    FORALL(child->objects,it) {
      objects.push_back(*it);
    }

  }

  bool Task::update() {
    TaskVector::iterator it= children.begin();
    while(it != children.end()) {
      if((*it)->update()) {
	childCompleted(*it);
	delete *it;
	it = children.erase(it);
      }else{
	++it;
      }
    }

    // if we have no children, and no groups, we are definitely done!
    if(objects.empty() && children.empty()) return true;

    // if we're out of time, then everyone is done, and mark them so...(not sure if marking is needed, but better
    // to be safe than sorry
    if(timeout > 0.0 && (getCurrentTime() - startTime) > timeout){
      FORALL(objects,it) {
        FIND(state->getFriendlyUnits(),it2,*it);
        if (it2 != state->getFriendlyUnits().end()){
          (it2->second).isDone = true;
        }
      }
      return true;
    }

    //if we have children, but no groups, then we are not done
    if(objects.empty()) return false;

    // if any of the groups of this task are not done their orders, then
    // this task is not done
    FORALL(objects,it) {

      // all objects MUST be done in order for this task to be complete
      FIND(state->getFriendlyUnits(),it2,*it);
      if (it2 == state->getFriendlyUnits().end()){
	FIND(state->getFriendlyBases(), it2, *it);
	if(it2 == state->getFriendlyBases().end())
	  ERR("a problem in task->update(), seems like task objects are out of sync with state objects... SHOULD NOT HAPPEN!");
      }

      if(!it2->second.isDone){
        return false;
      }
    }
    // finally, execute the next line of the function, if it is done, inform parent

    return exec();
  }

}
