#ifndef GfxEmitter_H
#define GfxEmitter_H

// $Id: GfxEmitter.H 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file
// (c) Keith Yerex
// (c) Tim Furtak
// (c) Erik Benner
// licensed under the GPL

#include "GfxGlobal.H"
#include "GfxObject.H"

//===================================================================

struct Particle
{
  Vec3<real4> pos;
  Vec3<real4> vel;
  Vec3<real4> acc;
  real4 fric;
  real4 life;
  real4 age;
  
  void update(real4 dt);
};

// For simple particles with reduced memory usage
struct ParticleLite
{
  Vec3<real4> pos;
  Vec3<real4> vel;
  real4 life;
  real4 age;
  
  bool update(real4 dt, const Vec3<real4> &acc);
};

//-------------------------------------------------------------------

class GfxEmitter : public GfxObject
{
protected:
  bool enabled; ///< Spawning particles.

public:
  GfxEmitter() {}
  virtual ~GfxEmitter() {}

  static void beginDraw();
  static void endDraw();

  void disable() { enabled = false; }
  void enable() { enabled = true; }

//  virtual void kill() { dying=true; enabled=false; };
  virtual void update(real4 /*dt*/) {}
  virtual void draw() {}
  virtual void pick() {}
  virtual void draw_emitters() = 0;
  DrawOrder order() const { return O_PARTICLES; }

  enum BlendType { STANDARD, ADDITIVE };
  BlendType blend;
};

//===================================================================

/** Emitter that maintains a circular buffer of particles. */

class GfxPListEmitter : public GfxEmitter
{
protected:

  bool dying;

  real4 create;   ///< Number of pending particles to create.
  real4 rate;     ///< Number of particles created during one timestep.
  real4 life;     ///< Length of time before each particle is destroyed.
  real4 limit;    ///< Length of time to emit particles.
  bool unlimited; ///< True iff emit particles forever.
  
  Vector<Particle> list;
  int first;      ///< Index of first active particle
  int last;       ///< Index of last active particle
  int max_particles;

  Particle& newparticle();
  void reorder();
  
  real4 w;
  real4 h;
  GLuint billboard_tex;
  GLuint color_tex;

  virtual void draw_as_points();
  virtual void draw_as_streaks();

  GfxPListEmitter();

public:
  enum DRAW_TYPE { POINTS=0, STREAKS };
  DRAW_TYPE draw_type;

  virtual void draw_emitters();
  void set_rate(real4 rate);
};

//===================================================================

/** Create particles within a box volume. */

class GfxBoxEmitter : public GfxPListEmitter
{
public:

  GfxBoxEmitter(GfxModule &gfxm, real4 rate, real4 life, real4 time_limit,
                Vec3<real4> pos, Vec3<real4> pos_rand, Vec3<real4> vel,
                Vec3<real4> vel_rand, Vec3<real4> accel,
                real4 fric, real4 fric_rand, real4 p_width, real4 p_height,
                GLuint texid, GLuint colid);
  
  virtual void update(real4 dt);

protected:

  Vec3<real4> pos;       ///< Spawn point.
  Vec3<real4> p_rand;    ///< Initial position randomization.
  Vec3<real4> vel;       ///< Initial particle velocity.
  Vec3<real4> v_rand;    ///< Initial velocity randomization.
  Vec3<real4> acc;       ///< Global accelerating force.
  real4 fric;     ///< Constant negative drag coefficient.
  real4 f_rand;   ///< Randomness for each particle's drag coeff (affects initial state).
  
  real4 cos_ang;
  real4 sin_ang;

  virtual void spawn_particle();
};

//-------------------------------------------------------------------

/** Spray particles from a common center within a limited cone. */

class GfxSprayEmitter : public GfxPListEmitter
{
public:
  GfxSprayEmitter(GfxModule &gfxm, real4 rate, real4 life, real4 time_limit,
                  const Vec3<real4> &pos, real4 min_lat, real4 max_lat,
                  real4 offset, real4 o_rand, real4 speed, real4 s_rand,
                  const Vec3<real4> &accel, real4 fric, real4 f_rand,
                  real4 ww, real4 hh, GLuint texid, GLuint colid);

  virtual void update(real4 dt);

protected:

  Vec3<real4> pos;       ///< Origin of the spray.
  real4 offset;          ///< Distance from the origin to spawn.
  real4 o_rand;          ///< Offset distance randomization.
  real4 min_ang;         ///< Latitude to start emitting
  real4 max_ang;         ///< Latitude to stop emitting
  real4 speed;           ///< Initial particle speed.
  real4 s_rand;          ///< Initial speed randomization.
  Vec3<real4> acc;       ///< Global accelerating force.
  real4 fric;     ///< Constant negative drag coefficient.
  real4 f_rand;   ///< Randomness for each particle's drag coeff (affects initial state).
  
  virtual void spawn_particle();
};
//-------------------------------------------------------------------

/** Spray particles from a common center within a limited cone. Moves in a circular path. */

class GfxCircleEmitter : public GfxPListEmitter
{
public:
  GfxCircleEmitter(GfxModule &gfxm, real4 rate, real4 life, real4 time_limit,
                  const Vec3<real4> &pos, real4 min_lat, real4 max_lat,
                  real4 offset, real4 o_rand, real4 speed, real4 s_rand,
                  const Vec3<real4> &accel, real4 fric, real4 f_rand,
                  real4 ww, real4 hh, GLuint texid, GLuint colid);

  virtual void update(real4 dt);

protected:

  Vec3<real4> pos;       ///< Origin of the spray.
  real4 offset;          ///< Distance from the origin to spawn.
  real4 o_rand;          ///< Offset distance randomization.
  real4 min_ang;         ///< Latitude to start emitting
  real4 max_ang;         ///< Latitude to stop emitting
  real4 speed;           ///< Initial particle speed.
  real4 s_rand;          ///< Initial speed randomization.
  Vec3<real4> acc;       ///< Global accelerating force.
  real4 fric;     ///< Constant negative drag coefficient.
  real4 f_rand;   ///< Randomness for each particle's drag coeff (affects initial state).
  real4 rot_amt;  ///<Rotational value in pixels
  
  virtual void spawn_particle();
  virtual void draw_emitters();
};

//===================================================================

/** 
* A light flare/corona emitter.  Note: More of a billboard then a particle 
* emitter since there is only ever a single particle.  
*/

class GfxLightCorona : public GfxEmitter
{
public:
  GfxLightCorona(GfxModule &gfxm, real4 size, real4 speed, GLuint texid, GLuint colid);
  
  virtual void draw_emitters()  { draw(); }
  virtual void draw();
  virtual void update(real4 dt);
  
protected:
  
  GLuint billboard_tex;	//!< Corona texture
  GLuint color_tex;	    //!< Corona colour texture	
  real4  size;
  real4  age;
  real4  fade;
};

//===================================================================

/** 
* A general fire and smoke emitter
*/

class GfxFireEmitter : public GfxEmitter
{
public:
  GfxFireEmitter(GfxModule &gfxm, real4 size, real4 speed);
  
  virtual void draw_emitters()  { draw(); }
  virtual void draw();
  virtual void update(real4 dt);
  
  void set_size(real4 amount) { size = amount; } 
  void set_speed(real4 amount) { speed = amount; } 
  
  //protected:
  
  uint4  max_spark, max_smoke;
  Vector<ParticleLite> spark;
  Vector<ParticleLite> smoke;
  
  static GLuint flame_tex;  //!< Flame texture
  static GLuint spark_tex;  //!< Spark texture
  static GLuint fire_tex;   //!< Fire colour texture
  static GLuint cloud_tex;  //!< Smoke texture
  static GLuint smoke_tex;  //!< Smoke colour texture
  
  Vec3<real4> acc;          //!< Global accelerating force
  real4  spread;            //!< Spray cone angle in degrees
  real4  theta;
  real4  phi;
  real4  size;
  real4  speed;
  real4  target;            //!< Flame pulse size target
  
  uint2 num_ptc;
  real4 new_ptc;
  
  void spawn(ParticleLite &particle);
};

//===================================================================

/** 
* A gun shell emitter
*/

class GfxShellEmitter : public GfxEmitter
{
public:
  GfxShellEmitter(GfxModule &gfxm, real4 size, real4 speed);
  
  virtual void draw();
  virtual void draw_emitters()  { draw(); }
  virtual void draw_at(const Mat4<real4> &mat);
  virtual void draw_emitters_at(const Mat4<real4> &mat)  { draw_at(mat); }
  virtual void update(real4 dt);
  
  void set_size(real4 amount) { size = amount; } 
  void set_speed(real4 amount) { speed = amount; } 
  
protected:
  
  static GLuint id;         //!< Display list id
  
  uint4  max_shell;
  Vector<ParticleLite> shell;
  
  Vec3<real4> acc;          //!< Global accelerating force
  real4  spread;            //!< Spray cone angle in degrees
  real4  theta;
  real4  phi;
  real4  size;
  real4  speed;
  
  uint2 num_ptc;
  real4 new_ptc;
  
  void spawn(ParticleLite &particle);
};

//===================================================================

class GfxLightningEmitter : public GfxEmitter
{
public:
  GfxLightningEmitter(GfxModule &gfxm, Vec3<real4> &targ, real4 size, real4 freq, sint4 num_arcs);

  virtual void draw();
  virtual void draw_emitters() { draw(); }
  virtual void update(real4 dt);

  struct Node {
    Vec3<real4> pos;
    Vec4<real4> color;
    real4 size;
  };

  struct Tree {
    Vector<sint4> parent;
    Vector<Node> nodes1;
    Vector<Node> nodes2;
    real4 blend;
  };

  struct Line {
    Vector<Node> nodes1;
    Vector<Node> nodes2;
    real4 blend;
  };

  Vector<Line> arc;

  void make_line(Vector<Node> &nodes);
  void draw_line(const Line &line);

  Tree make_tree();


  void draw_node(const Node &node);
  void draw_arc();

  void set_targ(const Vec3<real4> &targ_) { targ = targ_; }

protected:

  Vec3<real4> targ;
  real4 size;
  real4 freq;
  sint4 num_arcs;
};

//===================================================================

REGISTER_TYPEOF(1400, std::vector<ParticleLite>::iterator);

#endif
