// $Id: GfxTerrainTile.C 5474 2007-07-25 22:56:52Z sorsten $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxTerrainTile.H"
#include "GfxTerrain.H"
#include "Game.H"
#include "GameTile.H"
#include "Image.H"
#include "GfxModule.H"
#include "WaterTile.H"
#include "Random.H"

using namespace std;

//===================================================================

Vec3<real4> GfxTerrainTile::noise_scale1;
Vec3<real4> GfxTerrainTile::noise_scale2;
Image GfxTerrainTile::noise;
real4 GfxTerrainTile::sz = 1.0/8.0;
real4 GfxTerrainTile::eps = 1.0/512.0;
bool GfxTerrainTile::outline = false;

Vector<GfxTerrainTile::GeometryType> GfxTerrainTile::types;
ShaderProgram GfxTerrainTile::tile_shader;

//===================================================================

GfxTerrainTile::GfxTerrainTile()
{
  initialized = false;
  flattened[0] = 0;
  flattened[1] = 0;
  flattened[2] = 0;
  flattened[3] = 0;
  tx = 0;
  ty = 0;
  typeW = typeE = Tile::UNKNOWN;

  for(size_t i=0; i<4; i++)
  {
	  for(size_t j=0; j<4; j++)
	  {
		  vert[i][j] = Vec3<real4>(0,0,0);
	  }
  }
}

//-------------------------------------------------------------------

void GfxTerrainTile::add_type(sint4 id, GfxTerrainTile::GeometryType t)
{
  if (id >= (sint4)types.size()) {
    types.resize(id+1);
  }
  types[id] = t;
}

//-------------------------------------------------------------------

bool GfxTerrainTile::load_noise(const string &filename, const Vec3<real4> &scale1, const Vec3<real4> &scale2)
{
  if (!noise.loadTGA(filename.c_str())) {
    cout << "Unable to load terrain noise texture " << filename << endl;
    return 0;
  }

  if (noise.bpp() != 1) {
    cout << "Terrain noise texture is not 1bpp grayscale " << filename << endl;
    return 0;
  }
  
  noise_scale1 = scale1;
  noise_scale2 = scale2;

  return 1;
}

//-------------------------------------------------------------------

real4 GfxTerrainTile::get_noise_at(real4 x, real4 y)
{
  if (noise.width() == 0 || noise.height() == 0) return 0.0;

  real4 px = noise_scale1.x * x;
  real4 py = noise_scale1.z * y;
  
  sint4 nx = (sint4)(floor(px)) % noise.width(); if (nx < 0) nx += noise.width();
  sint4 ny = (sint4)(floor(py)) % noise.height(); if (ny < 0) ny += noise.height();
  real4 fx = px-floor(px);
  real4 fy = py-floor(py);

#if 1
  real4 n1 = noise_scale1.y * (real4)(noise[noise.width()*ny+nx]-128);
  real4 n2 = noise_scale1.y * (real4)(noise[noise.width()*ny+(nx+1) % noise.width()]-128);
  real4 n3 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + nx]-128);
  real4 n4 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + (nx+1) % noise.width()]-128);
#else
  real4 n1 = noise_scale1.y * (real4)(noise[noise.width()*ny+nx]);
  real4 n2 = noise_scale1.y * (real4)(noise[noise.width()*ny+(nx+1) % noise.width()]);
  real4 n3 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + nx]);
  real4 n4 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + (nx+1) % noise.width()]);
#endif

  real4 n = n1*(1-fx)*(1-fy) + n2*fx*(1-fy) + n3*(1-fx)*fy + n4*fx*fy;
  //  n *= 0.5;

/*  ny = (sint4)(floor(noise_scale2.z*y)) % noise_h;
  nx = (sint4)(floor(noise_scale2.x*x)) % noise_w;
  n += noise_scale2.y*(real4)(noise[noise_w*ny+nx]-128); */
  return n;
}

//-------------------------------------------------------------------

void GfxTerrainTile::compute_tc()
{
  int u=0, v=2;
  if (trans) { u=2; v=0;}

  FORS (i, 4) {
    FORS (j, 4) {
      texcoord[i][j][0] = hf*(vert[i][j][u]-vert[0][0][u])*(sz-eps)+tt1[0] + eps*hf;
      texcoord[i][j][1] = vf*(vert[i][j][v]-vert[0][0][v])*(sz-eps)+tt1[1] + eps*vf;
    }
  }
}

//-------------------------------------------------------------------

/*     N

   y  0-1             0-1
   |  |\|  TB_SPLIT   |/|  BT_SPLIT
   V  2-3             2-3
      x ->
*/

/**
   Smooth interpolation between tile corner heights.
   Does not necessarily corespond to tile geometry.
*/

#if GCC
#else
double cbrt(double v)
{
	if(v > 0.0)
	{
		return pow(v, 1.0/3.0);
	}
	else
	{
		return -pow(-v, 1.0/3.0);
	}
}
#endif

real4 GfxTerrainTile::height_at(real4 x, real4 y) const
{
  assert(0 <= x && x <= 1.0);
  assert(0 <= y && y <= 1.0);  
 
  real4 i = x * 2.0 - 1.0;
  real4 j = y * 2.0 - 1.0;

  i = (cbrt(i) + 1) / 2.0;
  j = (cbrt(j) + 1) / 2.0;

  real4 h1 = h[3] * i + h[2] * (1.0 - i);
  real4 h2 = h[1] * i + h[0] * (1.0 - i);

  real4 h = h1 * j + h2 * (1.0 - j);

  return h;
}

//-------------------------------------------------------------------

void GfxTerrainTile::set_edge(sint4 a, sint4 b, real4 width)
{
  if (h[a] == h[b]) width = 0.0;
  set_edge(a, b, width, 0.0);
}

//-------------------------------------------------------------------

/**
   Sets tile interpolation points along one edge defined by indices
   a and b.  Each point is distance "width" from the middle of the edge,
   and is lowered by "depth" from the default height given by height_at().
*/

void GfxTerrainTile::set_edge(sint4 a, sint4 b, real4 width, real4 depth)
{
  Vec3<real4> &va = vert[a][a];
  Vec3<real4> &vb = vert[b][b];

  Vec3<real4> &vab = vert[a][b];
  Vec3<real4> &vba = vert[b][a];

#if 0
  // bilinear interpolation
  vert[a][b] = va * (0.5 - width) + vb * (0.5 + width);
  vert[b][a] = va * (0.5 + width) + vb * (0.5 - width);
#else

  real4 i = 0.5 + width;
  vab.x = va.x * i + vb.x * (1-i);
  vab.z = va.z * i + vb.z * (1-i);

  vba.x = va.x * (1-i) + vb.x * i;
  vba.z = va.z * (1-i) + vb.z * i;

  vab.y = height_at(vab.x, vab.z);
  vba.y = height_at(vba.x, vba.z);
#endif

  vab.y -= depth;
  vba.y -= depth;
}

//-------------------------------------------------------------------

void w_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  if (j & 1) {
    --j;
  } else {
    ++j;
    if (i & 1) {
      --i;
    } else {
      ++i;
      --x;
      w_point(i,j,x,y);
    }
  }
}

void e_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  if (j & 1) {
    --j;
    if (i & 1) {
      --i;
      ++x;
      e_point(i,j,x,y);
    } else {
      ++i;
    }
  } else {
    ++j;
  }
}

void n_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  if (j & 2) {
    j -= 2;
  } else {
    j += 2;
    if (i & 2) {
      i -= 2;
    } else {
      i += 2;
      --y;
      n_point(i,j,x,y);
    }
  }
}

void s_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  if (j & 2) {
    j -= 2;
    if (i & 2) {
      i -= 2;
      ++y;
      s_point(i,j,x,y);
    } else {
      i += 2;
    }
  } else {
    j += 2;
  }
}

void sw_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  s_point(i,j,x,y);
  w_point(i,j,x,y);
}

void nw_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  n_point(i,j,x,y);
  w_point(i,j,x,y);
}

void se_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  s_point(i,j,x,y);
  e_point(i,j,x,y);
}

void ne_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  n_point(i,j,x,y);
  e_point(i,j,x,y);
}

//-------------------------------------------------------------------

void GfxTerrainTile::initverts()
{
  GeometryType tW = types[typeW];
  GeometryType tE = types[typeE];

  real4 miny = numeric_limits<real4>::max();
  real4 maxy = numeric_limits<real4>::min();
  //  real4 x = aabb_o.x;
  //  real4 y = aabb_o.z;
  //  const real4 a = 0.4;
  //  const real4 b = 0.3;

  // corner positions

  vert[0][0].x = 0.0;
  vert[0][0].y = h[0];
  vert[0][0].z = 0.0;

  vert[1][1].x = 1.0;
  vert[1][1].y = h[1];
  vert[1][1].z = 0.0;

  vert[2][2].x = 0.0;
  vert[2][2].y = h[2];
  vert[2][2].z = 1.0;

  vert[3][3].x = 1.0;
  vert[3][3].y = h[3];
  vert[3][3].z = 1.0;

  // adjust interpolation vertices for different tile types ---------

  // corner positions
  {
    const real4 depth = 0.6;
    GfxTerrainTile *a, *b, *c;
    bool w;

    if (tW == WATER) {

      if (split != Tile::TB_SPLIT || tE == WATER) {
        a = ter->tile(tx-1, ty);
        b = ter->tile(tx-1, ty-1);
        c = ter->tile(tx,   ty-1);
        w = true;
        if (!a || a->get_typeE() != WATER) w = false;
        if (!c || c->get_typeS() != WATER) w = false;
        if (b) {
          if (b->split == Tile::TB_SPLIT) {
            if (types[b->typeW] != WATER || types[b->typeE] != WATER) w = false;
          } else {
            if (types[b->typeE] != WATER) w = false;
          }
        } else w = false;
        if (w) {
          vert[0][0].y -= depth;
        }
      }
      
      //------------------------------
      
      if (split != Tile::BT_SPLIT || tE == WATER) {
        a = ter->tile(tx-1, ty);
        b = ter->tile(tx-1, ty+1);
        c = ter->tile(tx,   ty+1);
        w = true;
        if (!a || a->get_typeE() != WATER) w = false;
        if (!c || c->get_typeN() != WATER) w = false;
        if (b) {
          if (b->split == Tile::BT_SPLIT) {
            if (types[b->typeW] != WATER || types[b->typeE] != WATER) w = false;
          } else {
            if (types[b->typeE] != WATER) w = false;
          }
        } else w = false;
        if (w) {
          vert[2][2].y -= depth;
        }
      }
    }

    //------------------------------

    if (tE == WATER) {
      
      if (split != Tile::BT_SPLIT || tW == WATER) {
        a = ter->tile(tx+1, ty);
        b = ter->tile(tx+1, ty-1);
        c = ter->tile(tx,   ty-1);
        w = true;
        if (!a || a->get_typeW() != WATER) w = false;
        if (!c || c->get_typeS() != WATER) w = false;
        if (b) {
          if (b->split == Tile::BT_SPLIT) {
            if (types[b->typeW] != WATER || types[b->typeE] != WATER) w = false;
          } else {
            if (types[b->typeW] != WATER) w = false;
          }
        } else w = false;
        if (w) {
          vert[1][1].y -= depth;
        }
      }
      
      //------------------------------

      if (split != Tile::TB_SPLIT || tW == WATER) {
        a = ter->tile(tx+1, ty);
        b = ter->tile(tx+1, ty+1);
        c = ter->tile(tx,   ty+1);
        w = true;
        if (!a || a->get_typeW() != WATER) w = false;
        if (!c || c->get_typeN() != WATER) w = false;
        if (b) {
          if (b->split == Tile::TB_SPLIT) {
            if (types[b->typeW] != WATER || types[b->typeE] != WATER) w = false;
          } else {
            if (types[b->typeW] != WATER) w = false;
          }
        } else w = false;
        if (w) {
          vert[3][3].y -= depth;
        }
      }
    }
  }

  // edges between tiles (adds a trench for adjacent water tiles)

  {
    const real4 depth = 0.6;
    const real4 width = 0.3;  // for water - distance from the midline (0.5)
    const real4 cliff = 0.3; // if the heights are different, the distance the cliff extends from the midline

    GfxTerrainTile *t;
#if 1
    if (get_typeW() == WATER && (t = ter->tile(tx-1, ty)) && t->get_typeE() == WATER) set_edge(0, 2, width, depth);
    else                                                                              set_edge(0, 2, cliff);

    if (get_typeN() == WATER && (t = ter->tile(tx, ty-1)) && t->get_typeS() == WATER) set_edge(0, 1, width, depth);
    else                                                                              set_edge(0, 1, cliff);

    if (get_typeE() == WATER && (t = ter->tile(tx+1, ty)) && t->get_typeW() == WATER) set_edge(1, 3, width, depth);
    else                                                                              set_edge(1, 3, cliff);

    if (get_typeS() == WATER && (t = ter->tile(tx, ty+1)) && t->get_typeN() == WATER) set_edge(2, 3, width, depth);
    else                                                                              set_edge(2, 3, cliff);
#else
    set_edge(0, 2, cliff);
    set_edge(2, 3, cliff);
    set_edge(1, 3, cliff);
    set_edge(0, 1, cliff);
#endif
  }

  // interior points
  {
    Vec3<real4> center(0.5, (h[0] + h[1] + h[2] + h[3]) * 0.25, 0.5);
    
    const real4 width = 0.2; // width of basins for WATER tiles
    const real4 depth = 0.6; // water depth
    const real4 cliff = 0.35; // distance from the center to place cliff interpolation points

    vert[1][2] = vert[2][1] = vert[3][0] = vert[0][3] = center;

    //    vert[3][0].x = 1.0;
    //    vert[3][0].y = 1.0;
    //    vert[3][0].z = 1.0;

    if (h[0] != h[1] || h[2] != h[3] || split != Tile::NO_SPLIT) {
      vert[1][2].x += cliff;
      vert[3][0].x += cliff;
      vert[0][3].x -= cliff;
      vert[2][1].x -= cliff;
      has_cliffs = true;
    }

    if (h[0] != h[2] || h[1] != h[3] || split != Tile::NO_SPLIT) {
      vert[1][2].z -= cliff;
      vert[0][3].z -= cliff;
      vert[3][0].z += cliff;
      vert[2][1].z += cliff;
      has_cliffs = true;
    }

    if (split != Tile::TB_SPLIT) {
      if (h[0] == h[1] && h[0] == h[2]) vert[0][3] = vert[1][2] = vert[2][1] = center;
      if (h[3] == h[1] && h[3] == h[2]) vert[3][0] = vert[1][2] = vert[2][1] = center;
    }

    if (split != Tile::BT_SPLIT) {
      if (h[2] == h[0] && h[2] == h[3]) vert[2][1] = vert[3][0] = vert[0][3] = center;
      if (h[1] == h[0] && h[1] == h[3]) vert[1][2] = vert[3][0] = vert[0][3] = center;
    }

    if (tW == WATER || tE == WATER) has_cliffs = true;

    if (tW == WATER && tE == WATER) {
      vert[2][1].x = vert[0][3].x = 0.5 - width;
      vert[3][0].x = vert[1][2].x = 0.5 + width;
      vert[2][1].z = vert[3][0].z = 0.5 + width;
      vert[0][3].z = vert[1][2].z = 0.5 - width;
      sint4 a[] = {1,2,3,0};
      sint4 b[] = {2,1,0,3};
      FORU (i, 4) {
        Vec3<real4> &v = vert[a[i]][b[i]];
        v.y = height_at(v.x, v.z) - depth;
      }
    } else {
      if (tW == WATER) {
        if (split == Tile::BT_SPLIT) {
          Vec3<real4> &v = vert[0][3];
          v.x = 0.5 - width;
          v.z = 0.5 - width;
          v.y = height_at(v.x, v.z) - depth;
        } else {
          Vec3<real4> &v = vert[2][1];
          v.x = 0.5 - width;
          v.z = 0.5 + width;
          v.y = height_at(v.x, v.z) - depth;
        }
      }
      else if (tE == WATER) {
        if (split == Tile::BT_SPLIT) {
          Vec3<real4> &v = vert[3][0];
          v.x = 0.5 + width;
          v.z = 0.5 + width;
          v.y = height_at(v.x, v.z) - depth;
        } else {
          Vec3<real4> &v = vert[1][2];
          v.x = 0.5 + width;
          v.z = 0.5 - width;
          v.y = height_at(v.x, v.z) - depth;
        }
      } else {
        sint4 a[] = {1,2,3,0};
        sint4 b[] = {2,1,0,3};
        FORU (i, 4) {
          Vec3<real4> &v = vert[a[i]][b[i]];
          v.y = height_at(v.x, v.z);
        }
      }
    }
  }

  bool all_flat = (flattened[0] && flattened[1] && flattened[2] && flattened[3]);
 
  // offset for tile position
  FORS (i, 4) {
    FORS (j, 4) {
      Vec3<real4> &v = vert[i][j];
      v.x += tx;
      v.z += ty;

      if (!all_flat) {
        bool flat = false;
        if (flattened[i] && flattened[j]) {
          if (i == 0 && j != 3) flat = true;
          if (i == 1 && j != 2) flat = true;
          if (i == 2 && j != 1) flat = true;
          if (i == 3 && j != 0) flat = true;
        }
        if (!flat) v.y += get_noise_at(v.x, v.z);
      }
    }
  }
 
   FORS (i, 4) {
    FORS (j, 4) {
      Vec3<real4> &v = vert[i][j];

      // calculate normal (mostly accurate)

      typedef void (*move_func)(sint4&, sint4&, sint4&, sint4&);
      move_func func[4] = {w_point, e_point, n_point, s_point};
      move_func pfunc[4][6] = {{ne_point, nw_point, se_point, sw_point, n_point, s_point},
                               {ne_point, nw_point, se_point, sw_point, n_point, s_point},
                               {ne_point, nw_point, se_point, sw_point, w_point, e_point},
                               {ne_point, nw_point, se_point, sw_point, w_point, e_point}};

      Vec3<real4> pt[4]; // position of vertices one move away in each compass direction
      sint4 i2, j2, x2, y2;
      
      FORU (q, 4) {
        Vec3<real4> &u = pt[q];
        {
          i2 = i;
          j2 = j;
          x2 = tx;
          y2 = ty;
          do {
            (*func[q])(i2, j2, x2, y2);
            GfxTerrainTile *t = ter->tile(x2, y2);
            if (t == 0) {
              u = v;
              break;
            } else {
              u = t->vert[i2][j2];
            }
          } while (u.x == v.x && u.z == v.z);
        }
        {
          Vec3<real4> s;
          bool second = false;
          FORU (r, 6) {
            i2 = i;
            j2 = j;
            x2 = tx;
            y2 = ty;
            (*pfunc[q][r])(i2, j2, x2, y2);
            GfxTerrainTile *t = ter->tile(x2, y2);
            if (t) {
              Vec3<real4> &w = t->vert[i2][j2];
              if (w.x == v.x && w.z == v.z) { // perpendicular point that is coincident
                FOREVER {
                  (*func[q])(i2, j2, x2, y2);
                  t = ter->tile(x2, y2);
                  if (!t) break;
                  s = t->vert[i2][j2];
                  if (s.x != v.x || s.z != v.z) {
                    second = true;
                    break;
                  }
                }
                if (second) break;
              }
            }
          }
          if (second) { // take the average
            u += s;
            u *= 0.5;
          }
        }
      }

      FORU (q, 4) { pt[q].y += get_noise_at(pt[q].x, pt[q].z); }

      norm[i][j] = (pt[2]-pt[3]).cross(pt[0]-pt[1]);
      norm[i][j].normalize();

      maxy = max(maxy, v.y);
      miny = min(miny, v.y);
    }
  }

  aabb_o.y = miny;
  aabb_d.y = maxy-miny;

  compute_tc();
}

//-------------------------------------------------------------------

void GfxTerrainTile::initblack(sint4 index, GfxTerrain *t)
{
  if (initialized) return;
  ter = t;
  this->index = index;

  sint4 x = index % t->getWidth();
  sint4 y = index / t->getWidth();

  tx = x;
  ty = y;

  GfxTerrainTile *gtt;

  gtt = ter->tile(x-1, y);   if (gtt && gtt->initialized) { h[0] = gtt->h[1]; h[2] = gtt->h[3]; }
  gtt = ter->tile(x+1, y);   if (gtt && gtt->initialized) { h[1] = gtt->h[0]; h[3] = gtt->h[2]; }
  gtt = ter->tile(x, y-1);   if (gtt && gtt->initialized) { h[0] = gtt->h[2]; h[1] = gtt->h[3]; }
  gtt = ter->tile(x, y+1);   if (gtt && gtt->initialized) { h[2] = gtt->h[0]; h[3] = gtt->h[1]; }
  gtt = ter->tile(x-1, y-1); if (gtt && gtt->initialized) { h[0] = gtt->h[3]; }
  gtt = ter->tile(x+1, y+1); if (gtt && gtt->initialized) { h[3] = gtt->h[0]; }
  gtt = ter->tile(x+1, y-1); if (gtt && gtt->initialized) { h[1] = gtt->h[2]; }
  gtt = ter->tile(x-1, y+1); if (gtt && gtt->initialized) { h[2] = gtt->h[1]; }

  has_cliffs = false;

  hf = 1;
  vf = 1;
  trans = 0;
  tt1[0] = 0;
  tt1[1] = sz;

  aabb_o.x = x;
  aabb_o.z = y;
  aabb_d.x = 1;
  aabb_d.z = 1;

  initverts();

//  initialized = true;
}

//-------------------------------------------------------------------

WaterTile* GfxTerrainTile::new_WaterTile()
{
  static real4 depth = 0.25;
  Vec3<real4> watervert[4];
  
  watervert[0] = Vec3<real4>(tx+1, h[1]-depth, ty);
  watervert[1] = Vec3<real4>(tx,   h[0]-depth, ty);
  watervert[2] = Vec3<real4>(tx+1, h[3]-depth, ty+1);
  watervert[3] = Vec3<real4>(tx,   h[2]-depth, ty+1);

  //    FORU(i, 4) watervert[i].y += get_noise_at(watervert[i].x, watervert[i].z);

  return new WaterTile(watervert);
}

//-------------------------------------------------------------------

void GfxTerrainTile::init(sint4 index, Game *game, GfxTerrain *t)
{
  set_gfxm(t->gfxm);
  ter = t;
  this->index = index;

  const real4 inv_h = 1.0/GameConst::HEIGHT_MULT;
  sint4 x = game->get_map().ind2x(index);
  sint4 y = game->get_map().ind2y(index);

  tx = x;
  ty = y;

  const GameTile &tile = game->get_map()(index);

  Random &rng = gfxm->rand;

  sint4 htemp[4]; // get_topo takes integer parameters
  tile.get_topo(htemp[0], htemp[1], htemp[3], htemp[2], split);
  FORS (i, 4) h[i] = inv_h * (real8)htemp[i];

  tile.get_type(typeW, typeE);
  GeometryType tW = types[typeW];
  GeometryType tE = types[typeE];

  has_cliffs = (tW == CLIFF || tE == CLIFF);


  // pick texture tiles
  hf = 1;
  vf = 1;
  trans = 0;
  if (tW == WATER && tE != WATER) {

    tt1[0] = sz*2;
    tt1[1] = sz*5;
    if (split == Tile::BT_SPLIT) vf = -1;
    
  } else if (tE == WATER && tW != WATER) {
    tt1[0] = sz*2;
    tt1[1] = sz*5;
    hf = -1;
    if (split==Tile::TB_SPLIT) vf = -1;
    
  } else if (has_water()) {

    tt1[0] = sz * (3+(rng.rand_uint4() % 5));
    tt1[1] = sz*5;
    
  } else {

    tt1[0] = sz * (rng.rand_uint4() % 8);
    tt1[1] = sz * min(int(rng.rand_uint4() % 5), 1);
    hf = (2*(float)(rng.rand_uint4() % 2))-1;
    vf = (2*(float)(rng.rand_uint4() % 2))-1;
    trans=(rng.rand_uint4() % 2);
  }
  
  tt1[0] += sz*(hf<0);
  tt1[1] += sz*(vf<0);

  tt2[0] = sz * (rng.rand_uint4() % 8);
  tt2[1] = 6*sz;


  aabb_o.x = x;
  aabb_o.z = y;
  aabb_d.x = 1;
  aabb_d.z = 1;

  initverts();

  initialized = true;
}

//===================================================================

void GfxTerrainTile::flatten(sint4 f)
{
  flattened[0] += f;
  flattened[1] += f;
  flattened[2] += f;
  flattened[3] += f;
  
  //  if (initialized)
  {
    initverts();

    // align surrounding tiles
    int i = tx;
    int j = ty;

    GfxTerrainTile *t;

    if (i > 0) {
      t = ter->tile(i-1, j); // W tile
      if (t) {
        t->flattened[1] += f;
        t->flattened[3] += f;
        t->initverts();
      }

      t = ter->tile(i-1, j-1); // NW tile
      if (t) {
        t->flattened[3] += f;
        t->initverts();
      }

      t = ter->tile(i-1, j+1); // SW tile
      if (t) {
        t->flattened[1] += f;
        t->initverts();
      }
    }

    t = ter->tile(i, j-1); // N tile
    if (t) {
      t->flattened[2] += f;
      t->flattened[3] += f;
      t->initverts();
    }

    t = ter->tile(i, j+1); // S tile
    if (t) {
      t->flattened[0] += f;
      t->flattened[1] += f;
      t->initverts();
    }

    if (i+1 < ter->getWidth()) {
      t = ter->tile(i+1, j); // E tile
      if (t) {
        t->flattened[0] += f;
        t->flattened[2] += f;
        t->initverts();
      }

      t = ter->tile(i+1, j-1); // NE tile
      if (t) {
        t->flattened[2] += f;
        t->initverts();
      }

      t = ter->tile(i+1, j+1); // SE tile
      if (t) {
        t->flattened[0] += f;
        t->initverts();
      }
    }
  }
}

//===================================================================

void GfxTerrainTile::setTexture(real4 row, real4 col, sint4 ncol,
                                float hflip, float vflip, bool transpose)
{
  tt1[0] = sz * (col+(gfxm->rand.rand_uint4() % ncol));
  tt1[1] = row*sz;
  trans = transpose;
  hf = hflip;
  vf = vflip;
  tt1[0] += sz*(hf<0);
  tt1[1] += sz*(vf<0);
}

//===================================================================

void GfxTerrainTile::NW_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx-1, ty-1);
  if (!t || !t->initialized || !initialized) return;
  h[0] = t->h[3];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::N_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx, ty-1);
  if (!t || !t->initialized || !initialized) return;
  h[0] = t->h[2];
  h[1] = t->h[3];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::NE_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx+1, ty-1);
  if (!t || !t->initialized || !initialized) return;
  h[1] = t->h[2];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::E_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx+1, ty);
  if (!t || !t->initialized || !initialized) return;
  h[1] = t->h[0];
  h[3] = t->h[2];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::SE_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx+1, ty+1);
  if (!t || !t->initialized || !initialized) return;
  h[3] = t->h[0];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::S_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx, ty+1);
  if (!t || !t->initialized || !initialized) return;
  h[2] = t->h[0];
  h[3] = t->h[1];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::SW_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx-1, ty+1);
  if (!t || !t->initialized || !initialized) return;
  h[2] = t->h[1];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::W_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx-1, ty);
  if (!t || !t->initialized || !initialized) return;
  h[0] = t->h[1];
  h[2] = t->h[3];
  initverts();
}

//===================================================================

#define CNV(i,j)\
glColor3f(vert[i][j].y/8+0.35, vert[i][j].y/8+0.35, vert[i][j].y/8+0.35);\
glNormal3fv(norm[i][j].ptr());\
glVertex3fv(vert[i][j].ptr());

void GfxTerrainTile::draw()
{
  //  draw_shader(); return;

  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);
    glTexCoord2fv(texcoord[0][2]); CNV(0,2);
    glTexCoord2fv(texcoord[0][1]); CNV(0,1);
    glTexCoord2fv(texcoord[0][3]); CNV(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[1][0]); CNV(1,0);
    glTexCoord2fv(texcoord[1][2]); CNV(1,2);
    glTexCoord2fv(texcoord[1][1]); CNV(1,1);
    glTexCoord2fv(texcoord[1][3]); CNV(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[2][0]); CNV(2,0);
    glTexCoord2fv(texcoord[2][2]); CNV(2,2);
    glTexCoord2fv(texcoord[2][1]); CNV(2,1);
    glTexCoord2fv(texcoord[2][3]); CNV(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[3][0]); CNV(3,0);
    glTexCoord2fv(texcoord[3][2]); CNV(3,2);
    glTexCoord2fv(texcoord[3][1]); CNV(3,1);
    glTexCoord2fv(texcoord[3][3]); CNV(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[0][1].y, tt2[1]+sz-eps); CNV(0,1);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+sz/2);   CNV(0,3);
    glTexCoord2f(tt2[0]+sz*vert[1][0].y, tt2[1]+sz-eps); CNV(1,0);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+sz/2);   CNV(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/2); CNV(2,1);
    glTexCoord2f(tt2[0]+sz*vert[2][3].y, tt2[1]+eps);  CNV(2,3);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/2); CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[3][2].y, tt2[1]+eps);  CNV(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[2][0].y, tt2[1]+eps);  CNV(2,0);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/2); CNV(2,1);
    glTexCoord2f(tt2[0]+sz*vert[0][2].y, tt2[1]+eps);  CNV(0,2);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+sz/2); CNV(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/2);   CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[3][1].y, tt2[1]+sz-eps); CNV(3,1);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+sz/2);   CNV(1,2);
    glTexCoord2f(tt2[0]+sz*vert[1][3].y, tt2[1]+sz-eps); CNV(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+eps);  CNV(1,2);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+eps);  CNV(0,3);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/4); CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/4); CNV(2,1);
    glEnd();

  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    glTexCoord2fv(texcoord[0][3]); CNV(0,3);
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);
    glTexCoord2fv(texcoord[0][2]); CNV(0,2);
    glTexCoord2fv(texcoord[2][0]); CNV(2,0);
    glTexCoord2fv(texcoord[2][2]); CNV(2,2);
    glTexCoord2fv(texcoord[2][3]); CNV(2,3);
    glTexCoord2fv(texcoord[3][2]); CNV(3,2);
    glTexCoord2fv(texcoord[3][3]); CNV(3,3);
    glTexCoord2fv(texcoord[3][1]); CNV(3,1);
    glTexCoord2fv(texcoord[1][3]); CNV(1,3);
    glTexCoord2fv(texcoord[1][1]); CNV(1,1);
    glTexCoord2fv(texcoord[1][0]); CNV(1,0);
    glTexCoord2fv(texcoord[0][1]); CNV(0,1);
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);
    glEnd();
  }

  if (outline)draw_outline();
}

//===================================================================

#define STCNV(i,j)                                              \
  c = vert[i][j].y/8+0.35;                                      \
  glColor3f(c,c,c);                                             \
  glNormal3fv(norm[i][j].ptr());                                \
  glVertex3fv(vert[i][j].ptr());

/*
  if (GfxTerrainTile::tile_shader.begin()) {
    GLint shadow_tex = glGetUniformLocation(GfxTerrainTile::tile_shader.id(), "shadow");
    glUniform1i(shadow_tex, 1);
  }
*/

void GfxTerrainTile::draw_shader()
{
  real4 c;

  if (has_cliffs) {

    // non cliff pieces

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(0,1);
    STCNV(0,0);
    STCNV(0,3);
    STCNV(0,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(1,0);
    STCNV(1,2);
    STCNV(1,1);
    STCNV(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(2,0);
    STCNV(2,2);
    STCNV(2,1);
    STCNV(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(3,1);
    STCNV(3,0);
    STCNV(3,3);
    STCNV(3,2);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    STCNV(0,1);
    STCNV(0,3);
    STCNV(1,0);
    STCNV(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(2,1);
    STCNV(2,3);
    STCNV(3,0);
    STCNV(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(2,0);
    STCNV(2,1);
    STCNV(0,2);
    STCNV(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(3,0);
    STCNV(3,1);
    STCNV(1,2);
    STCNV(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(1,2);
    STCNV(0,3);
    STCNV(3,0);
    STCNV(2,1);
    glEnd();

  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);

    STCNV(0,3);
    STCNV(0,0);
    STCNV(0,2);
    STCNV(2,0);
    STCNV(2,2);
    STCNV(2,3);
    STCNV(3,2);
    STCNV(3,3);
    STCNV(3,1);
    STCNV(1,3);
    STCNV(1,1);
    STCNV(1,0);
    STCNV(0,1);
    STCNV(0,0);
    glEnd();

  }

  if (outline) draw_outline();
}

//-------------------------------------------------------------------

#define CNVf(i,j)\
glNormal3fv(norm[i][j].ptr());\
glVertex3fv(vert[i][j].ptr());

/** Draw the tile while ignoring texture.
    Combine with blending to highlight certain tiles.
*/

void GfxTerrainTile::draw_flat()
{
  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVf(0,0);
    CNVf(0,2);
    CNVf(0,1);
    CNVf(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(1,0);
    CNVf(1,2);
    CNVf(1,1);
    CNVf(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,0);
    CNVf(2,2);
    CNVf(2,1);
    CNVf(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(3,0);
    CNVf(3,2);
    CNVf(3,1);
    CNVf(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVf(0,1);
    CNVf(0,3);
    CNVf(1,0);
    CNVf(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,1);
    CNVf(2,3);
    CNVf(3,0);
    CNVf(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,0);
    CNVf(2,1);
    CNVf(0,2);
    CNVf(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(3,0);
    CNVf(3,1);
    CNVf(1,2);
    CNVf(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(1,2);
    CNVf(0,3);
    CNVf(3,0);
    CNVf(2,1);
    glEnd();
  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    CNVf(0,3);
    CNVf(0,0);
    CNVf(0,2);
    CNVf(2,0);
    CNVf(2,2);
    CNVf(2,3);
    CNVf(3,2);
    CNVf(3,3);
    CNVf(3,1);
    CNVf(1,3);
    CNVf(1,1);
    CNVf(1,0);
    CNVf(0,1);
    CNVf(0,0);
    glEnd();
  }
  if (outline) draw_outline();
}

//-------------------------------------------------------------------

#define CNVp(i,j)\
glColor3fv(((vert[i][j]-aabb_o)/aabb_d).ptr());\
glVertex3fv(vert[i][j].ptr());

void GfxTerrainTile::pick()
{
  glLoadName(my_id << GfxObject::SUB_ID_LOG);

  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVp(0,0);
    CNVp(0,2);
    CNVp(0,1);
    CNVp(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(1,0);
    CNVp(1,2);
    CNVp(1,1);
    CNVp(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,0);
    CNVp(2,2);
    CNVp(2,1);
    CNVp(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(3,0);
    CNVp(3,2);
    CNVp(3,1);
    CNVp(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVp(0,1);
    CNVp(0,3);
    CNVp(1,0);
    CNVp(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,1);
    CNVp(2,3);
    CNVp(3,0);
    CNVp(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,0);
    CNVp(2,1);
    CNVp(0,2);
    CNVp(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(3,0);
    CNVp(3,1);
    CNVp(1,2);
    CNVp(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(1,2);
    CNVp(0,3);
    CNVp(3,0);
    CNVp(2,1);
    glEnd();
  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    CNVp(0,3);
    CNVp(0,0);
    CNVp(0,2);
    CNVp(2,0);
    CNVp(2,2);
    CNVp(2,3);
    CNVp(3,2);
    CNVp(3,3);
    CNVp(3,1);
    CNVp(1,3);
    CNVp(1,1);
    CNVp(1,0);
    CNVp(0,1);
    CNVp(0,0);
    glEnd();
  }
}

//-------------------------------------------------------------------

/** Draws an outline along the South and West sides. */

void GfxTerrainTile::draw_outline()
{
  glLoadName(my_id << GfxObject::SUB_ID_LOG);

  real4 oset = 0.01;
  
  glColor3f(0.1, 0.1, 0.1);
  glBegin(GL_LINE_STRIP);
  glVertex3f(vert[0][0].x, vert[0][0].y+oset, vert[0][0].z);
  glVertex3f(vert[0][2].x, vert[0][2].y+oset, vert[0][2].z);
  glVertex3f(vert[2][0].x, vert[2][0].y+oset, vert[2][0].z);
  glVertex3f(vert[2][2].x, vert[2][2].y+oset, vert[2][2].z);
  glVertex3f(vert[2][3].x, vert[2][3].y+oset, vert[2][3].z);
  glVertex3f(vert[3][2].x, vert[3][2].y+oset, vert[3][2].z);
  glVertex3f(vert[3][3].x, vert[3][3].y+oset, vert[3][3].z);
  glEnd();
}

//===================================================================
