// $Id: MD2.C 6725 2008-01-28 09:12:35Z furtak $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxGlobal.H"
#include "GfxAnimesh.H"
#include "MD2.H"
#include "Options.H"

#ifndef GCC
static sint4 round(real8 x) { return (sint4)floor(x+0.5); }
#endif

using namespace std;

//===================================================================

#define MD2_ID (((int)'I')|(((int)'D')<<8)|(((int)'P')<<16)|(((int)'2')<<24))


MD2_Model::MD2_Model()
{
  // init. pointers

  vert = 0;
  tvert = 0;
  vnorm = 0;
  poly = 0;
  tpoly = 0;
  command = 0;
  vertex_array = 0;
  normal_array = 0;
  texcoord_array = 0;
  index_array = 0;
  loaded = false;
}


MD2_Model::~MD2_Model()
{
  free();
}

void MD2_Model::free()
{
  delete [] vert; vert = 0;
  delete [] tvert; tvert = 0;
  delete [] vnorm; vnorm = 0;
  delete [] poly; poly = 0;
  delete [] command; command = 0;
  delete [] vertex_array; vertex_array = 0;
  delete [] index_array; index_array = 0;
}

struct MD2Header
{
  sint4 id;
  sint4 version;
  sint4 skinWidth;
  sint4 skinHeight;
  sint4 frameSize;
  sint4 numSkins;
  sint4 numVertices;
  sint4 numTexCoords;
  sint4 numTriangles;
  sint4 numGlCommands;
  sint4 numFrames;
  sint4 offsetSkins;
  sint4 offsetTexCoords;
  sint4 offsetTriangles;
  sint4 offsetFrames;
  sint4 offsetGlCommands;
  sint4 offsetEnd;

  // convert to machine endianess
  
  void little2machine()
  {
    ::little2machine(id);
    ::little2machine(version);
    ::little2machine(skinWidth);
    ::little2machine(skinHeight);
    ::little2machine(frameSize);
    ::little2machine(numSkins);
    ::little2machine(numVertices);
    ::little2machine(numTexCoords);
    ::little2machine(numTriangles);
    ::little2machine(numGlCommands);
    ::little2machine(numFrames);
    ::little2machine(offsetSkins);
    ::little2machine(offsetTexCoords);
    ::little2machine(offsetTriangles);
    ::little2machine(offsetFrames);
    ::little2machine(offsetGlCommands);
    ::little2machine(offsetEnd);
  }
};

//-------------------------------------------------------------------

struct MD2TexCoord
{
  sint2 s;
  sint2 t;

  void little2machine()
  {
    ::little2machine(s);
    ::little2machine(t);    
  }
};

//-------------------------------------------------------------------

struct MD2Vertex
{
  uint1 pos[3];
  uint1 normalIndex;
};

//-------------------------------------------------------------------

struct MD2Frame
{
  real4     scale[3];
  real4     translate[3];  
  char      name[16];      
  MD2Vertex vert[1];
};

//===================================================================

void MD2_Model_DL::init()
{
  // build display lists
  display_lists = glGenLists(nframe);
  FORU (i, nframe) {
    glNewList(display_lists+i, GL_COMPILE);
    MD2_Model::draw(i);
    glEndList();
  }

  select_list = glGenLists(1);
  glNewList(select_list, GL_COMPILE);
  MD2_Model::select(0);
  glEndList();
}

//===================================================================

void MD2_Model_VB::init()
{
  glGenBuffersARB(1, &vertex_buffer);
  glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffer);
  glBufferDataARB(GL_ARRAY_BUFFER_ARB,
                  nframe*array_size*sizeof(Vec3<real4>)*2 + array_size*2*sizeof(real4),
                  vertex_array,
                  GL_STATIC_DRAW_ARB);

  glGenBuffersARB(1,&index_buffer);
  glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, index_buffer);
  glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB,
                  index_array_size*sizeof(sint2),
                  index_array,
                  GL_STATIC_DRAW_ARB);

  //  free(); // delete data
}

//===================================================================

sint4 MD2_Model::load(const string &modelfile, const string &texfile,
                      real4 scale, real4 dy, real4 gamma)
{
  loaded = false;

  scale *= 0.01; 

  if (verbose_load) cout << "reading MD2: " << modelfile;
  
  FILE* f = fopen(modelfile.c_str(), "rb");
  if (!f) {
    REM2("couldn't open file", modelfile.c_str());
    return 0;
  }

  MD2Header header;

  if (fread(&header, sizeof(header), 1, f) != 1) {
    fclose(f); 
    REM2("failed reading md2 file header", modelfile.c_str());
    return 0;
  }

  header.little2machine();
  
  if (header.id != MD2_ID) {
    fclose(f);    
    REM2("error reading md2 file (bad magic number)", modelfile.c_str());
    return 0;
  }

  nvert    = header.numVertices;
  npoly    = header.numTriangles;
  ntvert   = header.numTexCoords;
  nframe   = header.numFrames;
  ncommand = header.numGlCommands;

  tex = NULL;
  bbox.resize(nframe);
  vert    = new Vec3<real4>[nvert*nframe];
  vnorm   = new Vec3<real4>[nvert*nframe];
  tvert   = new real4[ntvert*2];
  poly    = new Poly[npoly];
  tpoly   = new Poly[npoly];
  command = new sint4[ncommand];

  if (verbose_load) {
    sint4 l2 = 1;
    sint4 np = npoly;
    while (np /= 10) ++l2;
    sint4 max = 62 - modelfile.length() - l2;
    if (max < 0) max = 0;
    FORU (i, max+1) cout << " ";
    cout << npoly << " polygons" << endl;
  }

  //  read texture coords
  {
    // fixme: check for error?
    fseek(f, header.offsetTexCoords, SEEK_SET);

    MD2TexCoord *tc = new MD2TexCoord[ntvert];

    if (fread(tc, sizeof(MD2TexCoord), ntvert, f) != ntvert) {
      fclose(f);
      REM2("failed reading texture coordinates", modelfile.c_str());
      delete [] tc;
      return 0;
    }
  
    FORU (i, ntvert) {
      tc[i].little2machine();
      tvert[2*i]   = ((real4)tc[i].s) / ((real4)header.skinWidth);
      tvert[2*i+1] = ((real4)tc[i].t) / ((real4)header.skinHeight);
    }
  
    delete [] tc;
  }
  
  //  read vertices
  {
    uint1 *buffer = new uint1[header.frameSize*nframe];
  
    fseek(f, header.offsetFrames, SEEK_SET); 
    uint4 bytecount = fread(buffer, 1, header.frameSize*nframe, f); 

    if (header.frameSize*nframe != bytecount) {
      cout << "error reading md2 file. " << bytecount << " != " << header.frameSize*nframe << endl;
      fclose(f); return 0;
    }

    const uint4 MAX_LEN = 16;
    char name[MAX_LEN+1];
    name[0] = 0;
    sint4 starting = 0;

    FORU (j, nframe) {

      bbox[j].min = bbox[j].max = Vec3<real4>(0, 0, 0);
      MD2Frame *frame = reinterpret_cast<MD2Frame *>(buffer+header.frameSize*j);
    
      FORU (k, 3) {
	little2machine(frame->scale[k]);
	little2machine(frame->translate[k]);
      }

      sint4 l = strlen(frame->name);
      if (frame->name[l-1] >= '0' && frame->name[l-1] <= '9') {
	frame->name[l-1] = 0;
	if (frame->name[l-2] >= '0' && frame->name[l-2] <= '9') frame->name[l-2] = 0;
      }
    
      if (strcmp(frame->name, name)) {
	if (j > 0) {
          name2frames[name] = pair<sint2,sint2>(starting, j-1);
          if (verbose_load) cout << "\t" << name << " : " << starting << " " << j-1 << endl;
        }
	assert(strlen(frame->name) <= MAX_LEN);
	strcpy(name, frame->name);
	starting = j;
      }
    
      FORU (i, nvert) {

	vert[nvert*j+i][2] = (frame->vert[i].pos[0]*frame->scale[0]) + frame->translate[0];
	vert[nvert*j+i][0] = (frame->vert[i].pos[1]*frame->scale[1]) + frame->translate[1];
	vert[nvert*j+i][1] = (frame->vert[i].pos[2]*frame->scale[2]) + frame->translate[2];

	vert[i+nvert*j][0] *= scale;
	vert[i+nvert*j][1] *= scale; vert[i+nvert*j][1] += dy;
	vert[i+nvert*j][2] *= scale;
	vert[i+nvert*j]    =  vert[i+nvert*j].rotate(Vec3<real4>(0, 1, 0), -M_PI/2);

	FORS (k, 3) {
          bbox[j].min.A[k] = min(bbox[j].min.A[k], vert[nvert*j+i][k]);
          bbox[j].max.A[k] = max(bbox[j].max.A[k], vert[nvert*j+i][k]);
	}
      
	vnorm[i+nvert*j] = normal_lookup[frame->vert[i].normalIndex] * -1;
      }
    }

    {
      centroid.x = centroid.y = centroid.z = 0;
      FORU (i, nvert) centroid += vert[i];
      centroid /= nvert;

      radius = 0;
      FORU (i, nvert) {
	Vec3<real4> d = centroid - vert[i];
	real4 l = d.length();
	radius = max(radius, l);
      }
    }
  
    name2frames[name] = pair<sint2,sint2>(starting,nframe-1);
    if (verbose_load) cout << "\t" << name << " : " << starting << " " << nframe-1 << endl;
  
    delete [] buffer;
  }

  // read triangles
  {
    MD2Triangle *tbuffer = new MD2Triangle[npoly];

    fseek(f, header.offsetTriangles, SEEK_SET);

    if (fread(tbuffer, sizeof(MD2Triangle), npoly, f) != npoly) {
      REM2("failed reading md2 triangles", modelfile.c_str());
      delete [] tbuffer;
      fclose(f); return 0;
    }

    FORU (i, npoly) {
      tbuffer[i].little2machine();
      FORS (j, 3) {
	poly[i].vert[j]  = tbuffer[i].vert[j];
	tpoly[i].vert[j] = tbuffer[i].tvert[j];
      }
    }

    delete [] tbuffer;
  }
  
  // read commands

  // fixme: check for error
  
  fseek(f, header.offsetGlCommands, SEEK_SET);

  if (fread(command, sizeof(*command), ncommand, f) != ncommand) {
    REM2("failed reading md2 command list", modelfile.c_str());
    fclose(f); return 0;
  }

  FORU (i, ncommand) little2machine(command[i]);
  fclose(f);

  // parse gl commands
  
  sint4 *p = command;
  array_size=index_array_size=0;

  while (sint4 i=*(p++)) {
    if (i < 0) i = -i;
    array_size += i;
    index_array_size += 3*(i-2);
    p += i*3;
  }

  vertex_array   = reinterpret_cast<Vec3<real4> *>(new real4[nframe*array_size*6 + array_size*2]);
  normal_array   = &vertex_array[nframe*array_size];
  texcoord_array = reinterpret_cast<real4*>(&vertex_array[2*nframe*array_size]);
  index_array    = new sint2[index_array_size];

  // build index array
  
  p = command;
  sint4 ind  = 0;
  sint4 ind2 = 0;

  while (sint4 i = *(p++)) {

    if (i < 0) {
      i = -i;
      for (sint4 j=1; j < i-1; j++) {
        index_array[ind2++] = ind;
        index_array[ind2++] = ind+j;
        index_array[ind2++] = ind+j+1;
      }

    } else {
      
      for (sint4 j=1; j < i-1; j++) {
        if (j & 1) {
          index_array[ind2++] = ind+j-1;
          index_array[ind2++] = ind+j;
          index_array[ind2++] = ind+j+1;
        } else {
          index_array[ind2++] = ind+j;
          index_array[ind2++] = ind+j-1;
          index_array[ind2++] = ind+j+1;
        }
      }
    }
    p += 3*i;
    ind += i;
  }

  // build texcoord array
  
  p = command;
  ind = 0;

  while (sint4 i = *(p++)) {
    if (i < 0) i = -i;
    for (; i > 0; i--, p += 3) {
      texcoord_array[ind*2]   = (reinterpret_cast<real4*>(p))[0];
      texcoord_array[ind*2+1] = (reinterpret_cast<real4*>(p))[1];
      ind++;
    }
  }
  
  // build vertex and normal arrays

  sint4 aj = 0, vj = 0;
  
  FORU (j, nframe) {

    p = command;
    ind = 0;

    while (sint4 i = *(p++)) {

      if (i < 0) i = -i;
      for(; i > 0; i--, p += 3) {
        vertex_array[aj+ind] = vert[vj+p[2]];
        normal_array[aj+ind] = vnorm[vj+p[2]];
        ind++;
      }
    }

    aj += array_size;
    vj += nvert;
  }

  if (load_textures(texfile.substr(0, texfile.length()-4), gamma) == 0) {
    REM2("failed loading textures!", texfile.substr(0, texfile.length()-4).c_str());
    return 0;
  }

  loaded = true; // before init()
  init();
  return 1;
}

//===================================================================

// no vertex interpolation - display lists
void MD2_Model_DL::draw(sint4 frame)
{
  assert(loaded);
  assert(frame >= 0 && frame < (sint4)nframe);
  glCallList(display_lists + frame);
}

void MD2_Model_DL::draw_interp(sint4 f1, sint4, real4, sint4 team)
{
  assert(loaded);
  // can't do interpolation with display lists
  glColor4f(1,1,1,1);
  glBindTexture(GL_TEXTURE_2D, tex->base_tex);
  draw(f1);
  glDepthFunc(GL_LEQUAL);
  glBindTexture(GL_TEXTURE_2D, tex->team_tex);
  glColor4fv(colors[team].ptr());
  draw(f1);
  glColor4f(1,1,1,1);
}

void MD2_Model_DL::select(sint4)
{
  glCallList(select_list);
}

//===================================================================

void MD2_Model_VB::draw_interp(sint4 f1, sint4, real4, sint4 team)
{
  assert(loaded);
  // don't interpolate for now
  glColor4f(1,1,1,1);
  glBindTexture(GL_TEXTURE_2D, tex->base_tex);
  draw(f1);
  glDepthFunc(GL_LEQUAL);
  glBindTexture(GL_TEXTURE_2D, tex->team_tex);
  glColor4fv(colors[team].ptr());
  draw(f1);
  glColor4f(1,1,1,1);
}

//===================================================================

// no vertex interpolation - vertex buffers
void MD2_Model_VB::draw(sint4 frame)
{
  assert(loaded);

  glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffer);
  glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, index_buffer);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);

  glVertexPointer(3, GL_FLOAT, 0, (void *)(frame*array_size*sizeof(Vec3<real4>)));
  glNormalPointer(GL_FLOAT, 0, (void *)((nframe+frame)*array_size*sizeof(Vec3<real4>)));
  glTexCoordPointer(2, GL_FLOAT, 0, (void *)(nframe*array_size*2*sizeof(Vec3<real4>)));

  if (!glDrawRangeElements) ERR("foo"); // fixme
  
  glDrawRangeElements(GL_TRIANGLES, 0, array_size, index_array_size, GL_UNSIGNED_SHORT, 0);

  glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
  glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

//===================================================================

// no vertex interpolation not using gl-commands
void MD2_Model::draw_t(sint4 frame, sint4 team)
{
  assert(loaded);
  
  glColor4f(1,1,1,1);
  glBindTexture(GL_TEXTURE_2D, tex->base_tex);

  glBegin(GL_TRIANGLES);
  FORU (i, npoly) {
    glTexCoord2fv(&tvert[2*(tpoly[i].vert[0])]);
    glNormal3fv(vnorm[nvert*frame+poly[i].vert[0]].ptr());
    glVertex3fv(vert[nvert*frame+poly[i].vert[0]].ptr());
    glTexCoord2fv(&tvert[2*(tpoly[i].vert[1])]);
    glNormal3fv(vnorm[nvert*frame+poly[i].vert[1]].ptr());
    glVertex3fv(vert[nvert*frame+poly[i].vert[1]].ptr());
    glTexCoord2fv(&tvert[2*(tpoly[i].vert[2])]);
    glNormal3fv(vnorm[nvert*frame+poly[i].vert[2]].ptr());
    glVertex3fv(vert[nvert*frame+poly[i].vert[2]].ptr());
  }    
  glEnd();


  glDepthFunc(GL_LEQUAL);
  glBindTexture(GL_TEXTURE_2D, tex->team_tex);
  glColor4fv(colors[team].ptr());

  glBegin(GL_TRIANGLES);
  FORU (i, npoly) {
    glTexCoord2fv(&tvert[2*(tpoly[i].vert[0])]);
    glNormal3fv(vnorm[nvert*frame+poly[i].vert[0]].ptr());
    glVertex3fv(vert[nvert*frame+poly[i].vert[0]].ptr());
    glTexCoord2fv(&tvert[2*(tpoly[i].vert[1])]);
    glNormal3fv(vnorm[nvert*frame+poly[i].vert[1]].ptr());
    glVertex3fv(vert[nvert*frame+poly[i].vert[1]].ptr());
    glTexCoord2fv(&tvert[2*(tpoly[i].vert[2])]);
    glNormal3fv(vnorm[nvert*frame+poly[i].vert[2]].ptr());
    glVertex3fv(vert[nvert*frame+poly[i].vert[2]].ptr());
  }    
  glEnd();

  glColor4f(1,1,1,1);
}

//===================================================================

void MD2_Model::select(sint4 frame)
{
  assert(loaded);
  assert(frame >= 0 && frame < (sint4)nframe);

  sint4 *p = command;

  while (sint4 i = *(p++)) {
    if (i < 0) { glBegin( GL_TRIANGLE_FAN ); i = -i; }
    else { glBegin( GL_TRIANGLE_STRIP ); }

    for(; i>0; i--, p += 3) {
      Vec3<real4> col = (vert[nvert*frame+p[2]] - bbox[frame].min) / 
                        (bbox[frame].max - bbox[frame].min);
      glColor3fv(col.ptr());
      glVertex3fv(vert[nvert*frame+p[2]].ptr());
    }

    glEnd();
  }
}

//===================================================================

// no vertex interpolation
void MD2_Model::draw(sint4 frame, sint4 team)
{
  assert(loaded);

  glColor4f(1,1,1,1);
  glBindTexture(GL_TEXTURE_2D, tex->base_tex);
  draw(frame);
  glDepthFunc(GL_LEQUAL);
  glBindTexture(GL_TEXTURE_2D, tex->team_tex);
  glColor4fv(colors[team].ptr());
  draw(frame);
  glColor4f(1,1,1,1);
}

#if 0

void MD2_Model::draw(sint4 frame)
{
  glBindTexture(GL_TEXTURE_2D, tex);

  sint4 *p = command;
  
  while(sint4 i = *(p++)) {
    if (i < 0) { glBegin(GL_TRIANGLE_FAN); i = -i; }
    else glBegin(GL_TRIANGLE_STRIP);

    for(; i > 0; i--, p += 3) {
      glTexCoord2fv(((real4 *)p));
      glNormal3fv(normal_lookup[vnorm[nvert*frame+p[2]]]* -1);
      glVertex3fv(vert[nvert*frame+p[2]]);
    }

    glEnd();
  }
}

#endif

//-------------------------------------------------------------------

void MD2_Model::draw(sint4 frame)
{
  assert(loaded);

  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  
  glVertexPointer(3, GL_FLOAT,0, vertex_array[frame*array_size].ptr());
  glNormalPointer(GL_FLOAT, 0, normal_array[frame*array_size].ptr());
  glTexCoordPointer(2, GL_FLOAT, 0, texcoord_array);

  if (!glDrawRangeElements) ERR("foo2"); // fixme
  
  glDrawRangeElements(GL_TRIANGLES, 0, array_size, index_array_size,
                      GL_UNSIGNED_SHORT, index_array);
  
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

//===================================================================

// software vertex interpolation
void MD2_Model::draw_interp(sint4 f1, sint4 f2, real4 blend, sint4 team)
{
  assert(loaded);

  glColor4f(1,1,1,1);
  glBindTexture(GL_TEXTURE_2D, tex->base_tex);

  sint4 *p = command;
  while (sint4 i = *(p++)) {
    if (i < 0) { glBegin( GL_TRIANGLE_FAN ); i = -i; }
    else { glBegin( GL_TRIANGLE_STRIP ); }

    for(; i > 0; i--, p += 3) {
      glTexCoord2fv(reinterpret_cast<real4*>(p));
      glNormal3fv((vnorm[nvert*f1 + p[2]]*(1.0-blend) + vnorm[nvert*f2 + p[2]]*blend).ptr());
      glVertex3fv((vert [nvert*f1 + p[2]]*(1.0-blend) + vert [nvert*f2 + p[2]]*blend).ptr());
    }
    glEnd();
  }

  glDepthFunc(GL_LEQUAL);
  glBindTexture(GL_TEXTURE_2D, tex->team_tex);
  glColor4fv(colors[team].ptr());

  p = command;
  while (sint4 i = *(p++)) {
    if (i < 0) { glBegin( GL_TRIANGLE_FAN ); i = -i; }
    else { glBegin( GL_TRIANGLE_STRIP ); }

    for(; i > 0; i--, p += 3) {
      glTexCoord2fv(reinterpret_cast<real4*>(p));
      glNormal3fv((vnorm[nvert*f1 + p[2]]*(1.0-blend) + vnorm[nvert*f2 + p[2]]*blend).ptr());
      glVertex3fv((vert [nvert*f1 + p[2]]*(1.0-blend) + vert [nvert*f2 + p[2]]*blend).ptr());
    }
    glEnd();
  }

  glColor4f(1,1,1,1);
}

//===================================================================

void MD2_Model::draw_bbox(sint4 frame)
{
  const Vec3<real4> &b0 = bbox[frame].min;
  const Vec3<real4> &b1 = bbox[frame].max;
  
  glBegin(GL_QUAD_STRIP);
  glVertex3f(b0.x, b1.y, b0.z);
  glVertex3f(b0.x, b0.y, b0.z);
  glVertex3f(b1.x, b1.y, b0.z);
  glVertex3f(b1.x, b0.y, b0.z);
  glVertex3f(b1.x, b1.y, b1.z);
  glVertex3f(b1.x, b0.y, b1.z);
  glVertex3f(b0.x, b1.y, b1.z);
  glVertex3f(b0.x, b0.y, b1.z);
  glVertex3f(b0.x, b1.y, b0.z);
  glVertex3f(b0.x, b0.y, b0.z);
  glEnd();

  glBegin(GL_QUADS);
  glVertex3f(b0.x, b1.y, b0.z);
  glVertex3f(b1.x, b1.y, b0.z);
  glVertex3f(b1.x, b1.y, b1.z);
  glVertex3f(b0.x, b1.y, b1.z);

  glVertex3f(b0.x, b0.y, b1.z);
  glVertex3f(b1.x, b0.y, b1.z);
  glVertex3f(b1.x, b0.y, b0.z);
  glVertex3f(b0.x, b0.y, b0.z);
  glEnd();
}

//===================================================================

void MD2_Model::draw_shadow(sint4 frame, real4 shadow_sc)
{
  assert(frame >= 0 && frame < (sint4)nframe);

  Vec3<real4> diff = bbox[frame].max - bbox[frame].min;
  Vec3<real4> f0 =   bbox[frame].min - diff*shadow_sc;
  Vec3<real4> f1 =   bbox[frame].max + diff*shadow_sc;

  glBegin(GL_QUADS);
  glTexCoord2f(0, 0);  glVertex3f(f0.x, f0.y, f0.z);
  glTexCoord2f(1, 0);  glVertex3f(f1.x, f0.y, f0.z);
  glTexCoord2f(1, 1);  glVertex3f(f1.x, f0.y, f1.z);
  glTexCoord2f(0, 1);  glVertex3f(f0.x, f0.y, f1.z);
  glEnd();
}

//===================================================================

// normal lookup table from Q2 source

typedef Vec3<real4> V3R;

static V3R foo[] = {
  V3R(-0.525731, 0.000000, 0.850651), 
  V3R(-0.442863, 0.238856, 0.864188), 
  V3R(-0.295242, 0.000000, 0.955423), 
  V3R(-0.309017, 0.500000, 0.809017), 
  V3R(-0.162460, 0.262866, 0.951056), 
  V3R(0.000000, 0.000000, 1.000000), 
  V3R(0.000000, 0.850651, 0.525731), 
  V3R(-0.147621, 0.716567, 0.681718), 
  V3R(0.147621, 0.716567, 0.681718), 
  V3R(0.000000, 0.525731, 0.850651), 
  V3R(0.309017, 0.500000, 0.809017), 
  V3R(0.525731, 0.000000, 0.850651), 
  V3R(0.295242, 0.000000, 0.955423), 
  V3R(0.442863, 0.238856, 0.864188), 
  V3R(0.162460, 0.262866, 0.951056), 
  V3R(-0.681718, 0.147621, 0.716567), 
  V3R(-0.809017, 0.309017, 0.500000), 
  V3R(-0.587785, 0.425325, 0.688191), 
  V3R(-0.850651, 0.525731, 0.000000), 
  V3R(-0.864188, 0.442863, 0.238856), 
  V3R(-0.716567, 0.681718, 0.147621), 
  V3R(-0.688191, 0.587785, 0.425325), 
  V3R(-0.500000, 0.809017, 0.309017), 
  V3R(-0.238856, 0.864188, 0.442863), 
  V3R(-0.425325, 0.688191, 0.587785), 
  V3R(-0.716567, 0.681718, -0.147621), 
  V3R(-0.500000, 0.809017, -0.309017), 
  V3R(-0.525731, 0.850651, 0.000000), 
  V3R(0.000000, 0.850651, -0.525731), 
  V3R(-0.238856, 0.864188, -0.442863), 
  V3R(0.000000, 0.955423, -0.295242), 
  V3R(-0.262866, 0.951056, -0.162460), 
  V3R(0.000000, 1.000000, 0.000000), 
  V3R(0.000000, 0.955423, 0.295242), 
  V3R(-0.262866, 0.951056, 0.162460), 
  V3R(0.238856, 0.864188, 0.442863), 
  V3R(0.262866, 0.951056, 0.162460), 
  V3R(0.500000, 0.809017, 0.309017), 
  V3R(0.238856, 0.864188, -0.442863), 
  V3R(0.262866, 0.951056, -0.162460), 
  V3R(0.500000, 0.809017, -0.309017), 
  V3R(0.850651, 0.525731, 0.000000), 
  V3R(0.716567, 0.681718, 0.147621), 
  V3R(0.716567, 0.681718, -0.147621), 
  V3R(0.525731, 0.850651, 0.000000), 
  V3R(0.425325, 0.688191, 0.587785), 
  V3R(0.864188, 0.442863, 0.238856), 
  V3R(0.688191, 0.587785, 0.425325), 
  V3R(0.809017, 0.309017, 0.500000), 
  V3R(0.681718, 0.147621, 0.716567), 
  V3R(0.587785, 0.425325, 0.688191), 
  V3R(0.955423, 0.295242, 0.000000), 
  V3R(1.000000, 0.000000, 0.000000), 
  V3R(0.951056, 0.162460, 0.262866), 
  V3R(0.850651, -0.525731, 0.000000), 
  V3R(0.955423, -0.295242, 0.000000), 
  V3R(0.864188, -0.442863, 0.238856), 
  V3R(0.951056, -0.162460, 0.262866), 
  V3R(0.809017, -0.309017, 0.500000), 
  V3R(0.681718, -0.147621, 0.716567), 
  V3R(0.850651, 0.000000, 0.525731), 
  V3R(0.864188, 0.442863, -0.238856), 
  V3R(0.809017, 0.309017, -0.500000), 
  V3R(0.951056, 0.162460, -0.262866), 
  V3R(0.525731, 0.000000, -0.850651), 
  V3R(0.681718, 0.147621, -0.716567), 
  V3R(0.681718, -0.147621, -0.716567), 
  V3R(0.850651, 0.000000, -0.525731), 
  V3R(0.809017, -0.309017, -0.500000), 
  V3R(0.864188, -0.442863, -0.238856), 
  V3R(0.951056, -0.162460, -0.262866), 
  V3R(0.147621, 0.716567, -0.681718), 
  V3R(0.309017, 0.500000, -0.809017), 
  V3R(0.425325, 0.688191, -0.587785), 
  V3R(0.442863, 0.238856, -0.864188), 
  V3R(0.587785, 0.425325, -0.688191), 
  V3R(0.688191, 0.587785, -0.425325), 
  V3R(-0.147621, 0.716567, -0.681718), 
  V3R(-0.309017, 0.500000, -0.809017), 
  V3R(0.000000, 0.525731, -0.850651), 
  V3R(-0.525731, 0.000000, -0.850651), 
  V3R(-0.442863, 0.238856, -0.864188), 
  V3R(-0.295242, 0.000000, -0.955423), 
  V3R(-0.162460, 0.262866, -0.951056), 
  V3R(0.000000, 0.000000, -1.000000), 
  V3R(0.295242, 0.000000, -0.955423), 
  V3R(0.162460, 0.262866, -0.951056), 
  V3R(-0.442863, -0.238856, -0.864188), 
  V3R(-0.309017, -0.500000, -0.809017), 
  V3R(-0.162460, -0.262866, -0.951056), 
  V3R(0.000000, -0.850651, -0.525731), 
  V3R(-0.147621, -0.716567, -0.681718), 
  V3R(0.147621, -0.716567, -0.681718), 
  V3R(0.000000, -0.525731, -0.850651), 
  V3R(0.309017, -0.500000, -0.809017), 
  V3R(0.442863, -0.238856, -0.864188), 
  V3R(0.162460, -0.262866, -0.951056), 
  V3R(0.238856, -0.864188, -0.442863), 
  V3R(0.500000, -0.809017, -0.309017), 
  V3R(0.425325, -0.688191, -0.587785), 
  V3R(0.716567, -0.681718, -0.147621), 
  V3R(0.688191, -0.587785, -0.425325), 
  V3R(0.587785, -0.425325, -0.688191), 
  V3R(0.000000, -0.955423, -0.295242), 
  V3R(0.000000, -1.000000, 0.000000), 
  V3R(0.262866, -0.951056, -0.162460), 
  V3R(0.000000, -0.850651, 0.525731), 
  V3R(0.000000, -0.955423, 0.295242), 
  V3R(0.238856, -0.864188, 0.442863), 
  V3R(0.262866, -0.951056, 0.162460), 
  V3R(0.500000, -0.809017, 0.309017), 
  V3R(0.716567, -0.681718, 0.147621), 
  V3R(0.525731, -0.850651, 0.000000), 
  V3R(-0.238856, -0.864188, -0.442863), 
  V3R(-0.500000, -0.809017, -0.309017), 
  V3R(-0.262866, -0.951056, -0.162460), 
  V3R(-0.850651, -0.525731, 0.000000), 
  V3R(-0.716567, -0.681718, -0.147621), 
  V3R(-0.716567, -0.681718, 0.147621), 
  V3R(-0.525731, -0.850651, 0.000000), 
  V3R(-0.500000, -0.809017, 0.309017), 
  V3R(-0.238856, -0.864188, 0.442863), 
  V3R(-0.262866, -0.951056, 0.162460), 
  V3R(-0.864188, -0.442863, 0.238856), 
  V3R(-0.809017, -0.309017, 0.500000), 
  V3R(-0.688191, -0.587785, 0.425325), 
  V3R(-0.681718, -0.147621, 0.716567), 
  V3R(-0.442863, -0.238856, 0.864188), 
  V3R(-0.587785, -0.425325, 0.688191), 
  V3R(-0.309017, -0.500000, 0.809017), 
  V3R(-0.147621, -0.716567, 0.681718), 
  V3R(-0.425325, -0.688191, 0.587785), 
  V3R(-0.162460, -0.262866, 0.951056), 
  V3R(0.442863, -0.238856, 0.864188), 
  V3R(0.162460, -0.262866, 0.951056), 
  V3R(0.309017, -0.500000, 0.809017), 
  V3R(0.147621, -0.716567, 0.681718), 
  V3R(0.000000, -0.525731, 0.850651), 
  V3R(0.425325, -0.688191, 0.587785), 
  V3R(0.587785, -0.425325, 0.688191), 
  V3R(0.688191, -0.587785, 0.425325), 
  V3R(-0.955423, 0.295242, 0.000000), 
  V3R(-0.951056, 0.162460, 0.262866), 
  V3R(-1.000000, 0.000000, 0.000000), 
  V3R(-0.850651, 0.000000, 0.525731), 
  V3R(-0.955423, -0.295242, 0.000000), 
  V3R(-0.951056, -0.162460, 0.262866), 
  V3R(-0.864188, 0.442863, -0.238856), 
  V3R(-0.951056, 0.162460, -0.262866), 
  V3R(-0.809017, 0.309017, -0.500000), 
  V3R(-0.864188, -0.442863, -0.238856), 
  V3R(-0.951056, -0.162460, -0.262866), 
  V3R(-0.809017, -0.309017, -0.500000), 
  V3R(-0.681718, 0.147621, -0.716567), 
  V3R(-0.681718, -0.147621, -0.716567), 
  V3R(-0.850651, 0.000000, -0.525731), 
  V3R(-0.688191, 0.587785, -0.425325), 
  V3R(-0.587785, 0.425325, -0.688191), 
  V3R(-0.425325, 0.688191, -0.587785), 
  V3R(-0.425325, -0.688191, -0.587785), 
  V3R(-0.587785, -0.425325, -0.688191), 
  V3R(-0.688191, -0.587785, -0.425325)
};

#if 0
static real4 normal_lookup_f[162][3] = {
  {-0.525731, 0.000000, 0.850651}, 
  {-0.442863, 0.238856, 0.864188}, 
  {-0.295242, 0.000000, 0.955423}, 
  {-0.309017, 0.500000, 0.809017}, 
  {-0.162460, 0.262866, 0.951056}, 
  {0.000000, 0.000000, 1.000000}, 
  {0.000000, 0.850651, 0.525731}, 
  {-0.147621, 0.716567, 0.681718}, 
  {0.147621, 0.716567, 0.681718}, 
  {0.000000, 0.525731, 0.850651}, 
  {0.309017, 0.500000, 0.809017}, 
  {0.525731, 0.000000, 0.850651}, 
  {0.295242, 0.000000, 0.955423}, 
  {0.442863, 0.238856, 0.864188}, 
  {0.162460, 0.262866, 0.951056}, 
  {-0.681718, 0.147621, 0.716567}, 
  {-0.809017, 0.309017, 0.500000}, 
  {-0.587785, 0.425325, 0.688191}, 
  {-0.850651, 0.525731, 0.000000}, 
  {-0.864188, 0.442863, 0.238856}, 
  {-0.716567, 0.681718, 0.147621}, 
  {-0.688191, 0.587785, 0.425325}, 
  {-0.500000, 0.809017, 0.309017}, 
  {-0.238856, 0.864188, 0.442863}, 
  {-0.425325, 0.688191, 0.587785}, 
  {-0.716567, 0.681718, -0.147621}, 
  {-0.500000, 0.809017, -0.309017}, 
  {-0.525731, 0.850651, 0.000000}, 
  {0.000000, 0.850651, -0.525731}, 
  {-0.238856, 0.864188, -0.442863}, 
  {0.000000, 0.955423, -0.295242}, 
  {-0.262866, 0.951056, -0.162460}, 
  {0.000000, 1.000000, 0.000000}, 
  {0.000000, 0.955423, 0.295242}, 
  {-0.262866, 0.951056, 0.162460}, 
  {0.238856, 0.864188, 0.442863}, 
  {0.262866, 0.951056, 0.162460}, 
  {0.500000, 0.809017, 0.309017}, 
  {0.238856, 0.864188, -0.442863}, 
  {0.262866, 0.951056, -0.162460}, 
  {0.500000, 0.809017, -0.309017}, 
  {0.850651, 0.525731, 0.000000}, 
  {0.716567, 0.681718, 0.147621}, 
  {0.716567, 0.681718, -0.147621}, 
  {0.525731, 0.850651, 0.000000}, 
  {0.425325, 0.688191, 0.587785}, 
  {0.864188, 0.442863, 0.238856}, 
  {0.688191, 0.587785, 0.425325}, 
  {0.809017, 0.309017, 0.500000}, 
  {0.681718, 0.147621, 0.716567}, 
  {0.587785, 0.425325, 0.688191}, 
  {0.955423, 0.295242, 0.000000}, 
  {1.000000, 0.000000, 0.000000}, 
  {0.951056, 0.162460, 0.262866}, 
  {0.850651, -0.525731, 0.000000}, 
  {0.955423, -0.295242, 0.000000}, 
  {0.864188, -0.442863, 0.238856}, 
  {0.951056, -0.162460, 0.262866}, 
  {0.809017, -0.309017, 0.500000}, 
  {0.681718, -0.147621, 0.716567}, 
  {0.850651, 0.000000, 0.525731}, 
  {0.864188, 0.442863, -0.238856}, 
  {0.809017, 0.309017, -0.500000}, 
  {0.951056, 0.162460, -0.262866}, 
  {0.525731, 0.000000, -0.850651}, 
  {0.681718, 0.147621, -0.716567}, 
  {0.681718, -0.147621, -0.716567}, 
  {0.850651, 0.000000, -0.525731}, 
  {0.809017, -0.309017, -0.500000}, 
  {0.864188, -0.442863, -0.238856}, 
  {0.951056, -0.162460, -0.262866}, 
  {0.147621, 0.716567, -0.681718}, 
  {0.309017, 0.500000, -0.809017}, 
  {0.425325, 0.688191, -0.587785}, 
  {0.442863, 0.238856, -0.864188}, 
  {0.587785, 0.425325, -0.688191}, 
  {0.688191, 0.587785, -0.425325}, 
  {-0.147621, 0.716567, -0.681718}, 
  {-0.309017, 0.500000, -0.809017}, 
  {0.000000, 0.525731, -0.850651}, 
  {-0.525731, 0.000000, -0.850651}, 
  {-0.442863, 0.238856, -0.864188}, 
  {-0.295242, 0.000000, -0.955423}, 
  {-0.162460, 0.262866, -0.951056}, 
  {0.000000, 0.000000, -1.000000}, 
  {0.295242, 0.000000, -0.955423}, 
  {0.162460, 0.262866, -0.951056}, 
  {-0.442863, -0.238856, -0.864188}, 
  {-0.309017, -0.500000, -0.809017}, 
  {-0.162460, -0.262866, -0.951056}, 
  {0.000000, -0.850651, -0.525731}, 
  {-0.147621, -0.716567, -0.681718}, 
  {0.147621, -0.716567, -0.681718}, 
  {0.000000, -0.525731, -0.850651}, 
  {0.309017, -0.500000, -0.809017}, 
  {0.442863, -0.238856, -0.864188}, 
  {0.162460, -0.262866, -0.951056}, 
  {0.238856, -0.864188, -0.442863}, 
  {0.500000, -0.809017, -0.309017}, 
  {0.425325, -0.688191, -0.587785}, 
  {0.716567, -0.681718, -0.147621}, 
  {0.688191, -0.587785, -0.425325}, 
  {0.587785, -0.425325, -0.688191}, 
  {0.000000, -0.955423, -0.295242}, 
  {0.000000, -1.000000, 0.000000}, 
  {0.262866, -0.951056, -0.162460}, 
  {0.000000, -0.850651, 0.525731}, 
  {0.000000, -0.955423, 0.295242}, 
  {0.238856, -0.864188, 0.442863}, 
  {0.262866, -0.951056, 0.162460}, 
  {0.500000, -0.809017, 0.309017}, 
  {0.716567, -0.681718, 0.147621}, 
  {0.525731, -0.850651, 0.000000}, 
  {-0.238856, -0.864188, -0.442863}, 
  {-0.500000, -0.809017, -0.309017}, 
  {-0.262866, -0.951056, -0.162460}, 
  {-0.850651, -0.525731, 0.000000}, 
  {-0.716567, -0.681718, -0.147621}, 
  {-0.716567, -0.681718, 0.147621}, 
  {-0.525731, -0.850651, 0.000000}, 
  {-0.500000, -0.809017, 0.309017}, 
  {-0.238856, -0.864188, 0.442863}, 
  {-0.262866, -0.951056, 0.162460}, 
  {-0.864188, -0.442863, 0.238856}, 
  {-0.809017, -0.309017, 0.500000}, 
  {-0.688191, -0.587785, 0.425325}, 
  {-0.681718, -0.147621, 0.716567}, 
  {-0.442863, -0.238856, 0.864188}, 
  {-0.587785, -0.425325, 0.688191}, 
  {-0.309017, -0.500000, 0.809017}, 
  {-0.147621, -0.716567, 0.681718}, 
  {-0.425325, -0.688191, 0.587785}, 
  {-0.162460, -0.262866, 0.951056}, 
  {0.442863, -0.238856, 0.864188}, 
  {0.162460, -0.262866, 0.951056}, 
  {0.309017, -0.500000, 0.809017}, 
  {0.147621, -0.716567, 0.681718}, 
  {0.000000, -0.525731, 0.850651}, 
  {0.425325, -0.688191, 0.587785}, 
  {0.587785, -0.425325, 0.688191}, 
  {0.688191, -0.587785, 0.425325}, 
  {-0.955423, 0.295242, 0.000000}, 
  {-0.951056, 0.162460, 0.262866}, 
  {-1.000000, 0.000000, 0.000000}, 
  {-0.850651, 0.000000, 0.525731}, 
  {-0.955423, -0.295242, 0.000000}, 
  {-0.951056, -0.162460, 0.262866}, 
  {-0.864188, 0.442863, -0.238856}, 
  {-0.951056, 0.162460, -0.262866}, 
  {-0.809017, 0.309017, -0.500000}, 
  {-0.864188, -0.442863, -0.238856}, 
  {-0.951056, -0.162460, -0.262866}, 
  {-0.809017, -0.309017, -0.500000}, 
  {-0.681718, 0.147621, -0.716567}, 
  {-0.681718, -0.147621, -0.716567}, 
  {-0.850651, 0.000000, -0.525731}, 
  {-0.688191, 0.587785, -0.425325}, 
  {-0.587785, 0.425325, -0.688191}, 
  {-0.425325, 0.688191, -0.587785}, 
  {-0.425325, -0.688191, -0.587785}, 
  {-0.587785, -0.425325, -0.688191}, 
  {-0.688191, -0.587785, -0.425325}
}; 
#endif

Vec3<real4> *MD2_Model::normal_lookup = foo; // (Vec3<real4> *)(normal_lookup_f);

