// $Id: MD3.C 6725 2008-01-28 09:12:35Z furtak $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxGlobal.H"
#include "GfxAnimesh.H"
#include "MD3.H"
#include "Profiler.H"
#include "Options.H"

#ifndef GCC
static sint4 round(real8 x) { return (sint4)floor(x+0.5); }
#endif

using namespace std;

//===================================================================

#define MD3_ID (((int)'I')|(((int)'D')<<8)|(((int)'P')<<16)|(((int)'3')<<24))

//===================================================================

// convert to machine endianess
void MD3Header::little2machine()
{
  ::little2machine(id);
  ::little2machine(version);
  ::little2machine(flags);
  ::little2machine(numFrames);
  ::little2machine(numTags);
  ::little2machine(numSurfaces);
  ::little2machine(numSkins);
  ::little2machine(surfaceSize);
  ::little2machine(offsetTags);
  ::little2machine(offsetSurfaces);
  ::little2machine(offsetEnd);
}

void MD3Header::print() const
{
  cout << "-- MD3 header --"<< endl
       << "id="             << id << endl
       << "version="        << version << endl
       << "name="           << name << endl
       << "flags="          << flags << endl
       << "numFrames="      << numFrames << endl
       << "numTags="        << numTags << endl
       << "numSurfaces="    << numSurfaces << endl
       << "numSkins="       << numSkins << endl
       << "surfaceSize="    << surfaceSize << endl
       << "offsetTags="     << offsetTags << endl
       << "offsetSurfaces=" << offsetSurfaces << endl
       << "offsetEnd="      << offsetEnd << endl;
}

//===================================================================

// convert to machine endianess
void MD3Surface::little2machine()
{
  ::little2machine(id);
  ::little2machine(flags);
  ::little2machine(numFrames);
  ::little2machine(numShaders);
  ::little2machine(numVerts);
  ::little2machine(numTriangles);
  ::little2machine(offsetTriangles);
  ::little2machine(offsetShaders);
  ::little2machine(offsetST);
  ::little2machine(offsetVerts);
  ::little2machine(offsetEnd);
}

void MD3Surface::print() const
{
  cout << "-- MD3 surface --" << endl
       << "id="               << id << endl
       << "name="             << name << endl
       << "flags="            << flags << endl
       << "numFrames="        << numFrames << endl
       << "numShaders="       << numShaders << endl
       << "numVerts="         << numVerts << endl
       << "numTriangles="     << numTriangles << endl
       << "offsetTriangles="  << offsetTriangles << endl
       << "offsetShaders="    << offsetShaders << endl
       << "offsetST="         << offsetST << endl
       << "offsetVerts="      << offsetVerts << endl
       << "offsetEnd="        << offsetEnd << endl;
}

//===================================================================

void MD3Vertex::little2machine()
{
  ::little2machine(x);
  ::little2machine(y);
  ::little2machine(z);
  ::little2machine(normal);
}

//===================================================================

void MD3Tag::little2machine()
{
  FORU (i, 3) ::little2machine(origin[i]);
  FORU (i, 3)
    FORU (j, 3) ::little2machine(axis[i][j]);
}

//===================================================================

MD3_Model::MD3_Model()
{
  vert = 0;
  tvert = 0;
  vnorm = 0;
  poly = 0;
  loaded = false;
}

MD3_Model::~MD3_Model()
{
  delete [] vert;
  delete [] tvert;
  delete [] vnorm;
  delete [] poly;
}

MD3_Model_DL::~MD3_Model_DL()
{
  // Free display lists
  if (glIsList(display_lists)) glDeleteLists(display_lists, nframe);
  if (glIsList(select_list)) glDeleteLists(select_list, 1);
}

//===================================================================

void MD3_Model::init()
{ /* Nothing to init */ }

void MD3_Model_DL::init()
{
  // Build a display list for each frame of the animation
  display_lists = glGenLists(nframe);
  FORU (i, nframe) {
    glNewList(display_lists+i, GL_COMPILE);
    MD3_Model::draw(i);
    glEndList();
  }

  // Only build one diplay list for select
  select_list = glGenLists(1);
  glNewList(select_list, GL_COMPILE);
  MD3_Model::select(0);
  glEndList();
}

//-------------------------------------------------------------------

bool MD3_Model::load_header(FILE* &f, MD3Header &header)
{
  if (fread(&header, sizeof(header), 1, f) != 1) {
    fclose(f); 
    cerr << "failed reading md3 file header." << endl;
    return false;
  }

  header.little2machine();
  //  header.print();
  
  if (header.id != MD3_ID) {
    cerr << "error reading md3 file. bad magic number." << endl;
    fclose(f);    
    return false;
  }

  return true;
}

//-------------------------------------------------------------------

bool MD3_Model::load_tags(FILE* &f, const MD3Header &header)
{
  ntag = header.numTags;

  if (ntag) {
    MD3Tag *tags = new MD3Tag[ntag * nframe];
    fseek(f, header.offsetTags, SEEK_SET);
    if (fread(tags, sizeof(MD3Tag) * ntag * nframe, 1, f) != 1) {
      delete [] tags;
      return false;
    }
    FORS (i, ntag * nframe) tags[i].little2machine();

    all_tags = new ModelTag[nframe*ntag];
    tag.resize(ntag);
    tag_name.resize(ntag);
    FORS (i, ntag) {
      tag[i] = &all_tags[nframe*i];
      tag_name[i] = reinterpret_cast<char*>(tags[i].name);
      if (tag_name[i].size() > 50) tag_name[i] = tag_name[i].substr(0,50);
    }

    FORS (i, nframe) {
      FORS (j, ntag) {

        sint4 ind = j + i*ntag;

        if (i == 0 && verbose_load) cout << "\ttag: " << tags[ind].name << "\n";

        real4 (*a)[3] = tags[ind].axis;

        // adjust the axis for tags to match the default model transformation
        Vec3<real4> ax(a[1][0], a[1][1], a[1][2]);
        Vec3<real4> ay(a[2][0], a[2][1], a[2][2]);
        Vec3<real4> az(a[0][0], a[0][1], a[0][2]);

        Vec3<real4> of(tags[ind].origin[0],
                       tags[ind].origin[1],
                       tags[ind].origin[2]);

        // transform md3 coordinates into our system
        // tag coordinates are 1/64 of the vertex scale
        // 1.0/15.625 = magic constant based on 1000 scaling factor below
        // (1000 / 64 = 15.624)

        ax = Vec3<real4>(ax.x, ax.z, -ax.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2);
        ay = Vec3<real4>(ay.x, ay.z, -ay.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2);
        az = Vec3<real4>(az.x, az.z, -az.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2);
        of = Vec3<real4>(of.x, of.z, -of.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2) * (1.0/15.625);

        Vec4<real4> axis_x(ax.x, ax.y, ax.z, 0);
        Vec4<real4> axis_y(ay.x, ay.y, ay.z, 0);
        Vec4<real4> axis_z(az.x, az.y, az.z, 0);
        Vec4<real4> foo   (of.x, of.y, of.z, 1); // position offset

        //        axis_x.normalize();
        //        axis_y.normalize();
        //        axis_z.normalize();

        tag[j][i].axis = Mat4<real4>(axis_x, axis_y, axis_z, foo);				
      }
    }
    delete [] tags;
  }

  return true;
}

//-------------------------------------------------------------------

void MD3_Model::load_cfg(const string &fname)
{
  ifstream file;
  file.open(fname.c_str());
  if (!file.is_open()) return;
  
  while (file.good()) {
    char line[256];
    file.getline(line, 256);
    string ln = line;
    FORU (i, ln.size())
      if (ln[i] == '\r' || ln[i] == '\n') {
        ln = ln.substr(0, i);
        break;
      }
    if (ln == "" || ln.substr(0, 2) == "//") continue;
    istringstream ss(ln);
    
    sint4 first, num, looping, fps;
    ss >> first;
    ss >> num;
    ss >> looping;
    ss >> fps;
    
    //      cout << first << " " << num << " " << looping << " " << fps << endl;
    
    string temp;
    ss >> temp;
    if (temp != "//") REM("expecting animation name");
    ss >> temp;
    name2frames[temp] = pair<sint2, sint2>(first, first+num-1);
  }

  file.close();
}

//-------------------------------------------------------------------

void MD3_Model::compute_centroid()
{
  centroid[0] = centroid[1] = centroid[2] = 0.0;
  FORU (i, nvert) centroid += vert[i];
  centroid /= nvert;
  
  radius = 0;
  FORU (i, nvert) {
    Vec3<real4> d = centroid - vert[i];
    real4 l = d.length();
    radius = max(radius, l);
  }
}

//===================================================================

bool MD3_Model::fast_load(const string &filename)
{
  ifstream file(filename.c_str(), ios::in | ios::binary);
  if (!file) return false;

  file.read(reinterpret_cast<char*>(&nframe),   sizeof(nframe));
  file.read(reinterpret_cast<char*>(&ntag),     sizeof(ntag));
  file.read(reinterpret_cast<char*>(&centroid), sizeof(centroid));
  file.read(reinterpret_cast<char*>(&radius),   sizeof(radius));

  bbox.resize(nframe);
  if (nframe > 0) {
    file.read(reinterpret_cast<char*>(&bbox[0]), sizeof(bbox[0])*nframe);
  }

  if (ntag) {
    sint1 *s = new sint1[ntag];
    file.read(reinterpret_cast<char*>(s), sizeof(*s)*ntag);
    sint4 len = 0;
    FORU (i, ntag) len += s[i];
    char *str = new char[len];
    file.read(str, len);
    
    len = 0;
    FORU (i, ntag) {
      string foo(&str[len], s[i]);
      tag_name.push_back(foo);
      len += s[i];
    }
    delete [] s;
    delete [] str;
  }

  all_tags = new ModelTag[nframe*ntag];
  tag.resize(ntag);
  if (nframe > 0) {
    FORU (i, ntag) tag[i] = &all_tags[nframe*i];
    file.read(reinterpret_cast<char*>(all_tags), sizeof(*all_tags)*nframe*ntag);
  }

  file.read((char*)&nvert, sizeof(nvert));
  file.read((char*)&npoly, sizeof(npoly));

  vert = new Vec3<real4>[nvert*nframe];
  vnorm = new Vec3<real4>[nvert*nframe];
  tvert = new real4[nvert*2];
  poly = new Poly[npoly];

  file.read((char*)vert, sizeof(vert[0])*nvert*nframe);
  file.read((char*)tvert, sizeof(tvert[0])*nvert*2);
  file.read((char*)vnorm, sizeof(vnorm[0])*nvert*nframe);
  file.read((char*)poly, sizeof(poly[0])*npoly);

  return true;
}

//-------------------------------------------------------------------

bool MD3_Model::fast_save(const string &filename)
{
  ofstream file;
  file.open(filename.c_str(), ios::in | ios::binary);
  if (!file) return false;

  file.write((char*)&nframe,   sizeof(nframe));
  file.write((char*)&ntag,     sizeof(ntag));
  file.write((char*)&centroid, sizeof(centroid));
  file.write((char*)&radius,   sizeof(radius));

  if (nframe > 0) {
    file.write((char*)&bbox[0], sizeof(bbox[0])*nframe);
  }

  FORU (i, ntag) {
    sint1 s = tag_name[i].size();
    file.write((char*)&s, sizeof(s));
  }
  FORU (i, ntag) {
    file.write(tag_name[i].c_str(), tag_name[i].size());
  }

  if (nframe > 0) {
    file.write((char*)all_tags, sizeof(*all_tags)*nframe*ntag);
  }

  file.write((char*)&nvert, sizeof(nvert));
  file.write((char*)&npoly, sizeof(npoly));

  file.write((char*)vert,  sizeof(*vert)*nvert*nframe);
  file.write((char*)tvert, sizeof(*tvert)*nvert*2);
  file.write((char*)vnorm, sizeof(*vnorm)*nvert*nframe);
  file.write((char*)poly,  sizeof(*poly)*npoly);

  return true;
}

//===================================================================

void MD3_Model::strip(const string &modelfile)
{
  FILE *f = fopen(modelfile.c_str(), "r+b");
  if (!f) {
    REM2("couldn't open file", modelfile.c_str());
    return;
  }

  fseek(f, 8, SEEK_SET);
  uint1 name[64];
  memset(name, 0, 64);
  fwrite(name, 1, 64, f);

  fseek(f, 0, SEEK_SET);

  MD3Header header;
  if (!load_header(f, header)) return;

  MD3Surface surface;
  int surface_offset = header.offsetSurfaces;

  FORS (i, header.numSurfaces) {

    fseek(f, surface_offset, SEEK_SET);
    if (fread(&surface, sizeof(MD3Surface), 1, f) != 1) {
      fclose(f); 
      cerr << "failed reading md3 surface. " << modelfile << endl;
      return;
    }

    fseek(f, surface_offset+4, SEEK_SET); // +4 -> 'name' position
    fwrite(name, 1, 64, f);

    surface_offset += surface.offsetEnd;
  }

  fclose(f);
}

//-------------------------------------------------------------------

sint4 MD3_Model::load(const string &modelfile, const string &texfile,
                      real4 /*scale*/, real4 /*dy*/, real4 gamma)
{
  PROFILE("MD3 load");

  loaded = false;

  //  strip(modelfile);

  // check for fast model
  if (!clear_fast && fast_load(modelfile + ".fast")) {
    if (load_textures(texfile.substr(0, texfile.length()-4), gamma) == 0) {
      REM2("failed loading textures!", texfile.substr(0, texfile.length()-4).c_str());
      return 0;
    }

    // try to find a cfg file with animation names
    string fname = modelfile.substr(0, modelfile.length()-4) + ".cfg";
    load_cfg(fname);

    init();
    return 1;
  }

  // standard md3 file

  FILE* f = fopen(modelfile.c_str(), "rb");
  if (!f) {
    REM2("couldn't open file", modelfile.c_str());
    return 0;
  }

  MD3Header header;
  if (verbose_load) cout << "reading MD3: " << modelfile;
  if (!load_header(f, header)) return 0;


  nvert = 0;
  npoly = 0;
  nframe = header.numFrames;

  MD3Surface *surfaces = new MD3Surface[header.numSurfaces];

  int surface_offset = header.offsetSurfaces;

  FORS (i, header.numSurfaces) {

    fseek(f, surface_offset, SEEK_SET);

    if (fread(&surfaces[i], sizeof(MD3Surface), 1, f) != 1) {
      fclose(f); 
      cerr << "failed reading md3 surface. " << modelfile << endl;
      delete [] surfaces;
      return 0;
    }

    surfaces[i].little2machine();
    //    cout << "\t\tsurf: " << surfaces[i].name << endl;
    //    surfaces[i].print();

    if (surfaces[i].id != MD3_ID) {
      fclose(f); 
      cerr << "failed reading md3 surface. " << modelfile << endl;
      delete [] surfaces;
      return 0;
    }

    nvert += surfaces[i].numVerts;
    npoly += surfaces[i].numTriangles;
    surfaces[i].offsetVerts += surface_offset;
    surfaces[i].offsetST += surface_offset;
    surfaces[i].offsetTriangles += surface_offset;
    surface_offset += surfaces[i].offsetEnd;
  }

  if (verbose_load) {
    sint4 l2 = 1;
    sint4 np = npoly;
    while (np /= 10) ++l2;
    sint4 max = 62 - modelfile.length() - l2;
    if (max < 0) max = 0;
    FORU (i, max+1) cout << " ";
    cout << npoly << " polygons\n";
    cout << "\tframes: " << nframe << "\n";
  }

  if (!load_tags(f, header)) {
    fclose(f);
    cerr << "failed reading md3 tag. " << modelfile << endl;
    return 0;
  }

  tex = 0;
  bbox.resize(nframe);
  vert    = new Vec3<real4>[nvert*nframe];
  vnorm   = new Vec3<real4>[nvert*nframe];
  tvert   = new real4[nvert*2];
  memset(tvert, 0, sizeof(tvert[0])*nvert*2);

  poly = new Poly[npoly];

  int vert_offset = 0;
  int poly_offset = 0;

  sint4 *buffer = new sint4[3*npoly];

  FORS (s, header.numSurfaces) {
    
    // read texture coords
    fseek(f, surfaces[s].offsetST, SEEK_SET);  // fixme: check if OK
    
    if (fread(tvert + 2*vert_offset, sizeof(real4)*2*surfaces[s].numVerts, 1, f) != 1) {
      cerr << "failed reading texture coordinates. " << modelfile << endl;
      delete [] buffer;      
      delete [] surfaces;
      return 0;
    }
    
    // read vertices
    
    MD3Vertex tmp;

    fseek(f, surfaces[s].offsetVerts, SEEK_SET); // fixme: check if OK
    
    FORU (j, nframe) {
      FORS (i, surfaces[s].numVerts) {
        
        if (fread(&tmp, sizeof(tmp), 1, f) != 1) {
          cerr << "error reading md3 file" << endl;
	  delete [] buffer;	  
          delete [] surfaces;
          fclose(f); return 0;
        }
	tmp.little2machine();

        sint4 ind = nvert*j + i + vert_offset;
	assert(ind < (sint4)(nvert*nframe));
        vert[ind] =
	  Vec3<real4>(tmp.x, tmp.z, -tmp.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2)*(1.0/1000.0);
        
        FORS (k, 3) {
          bbox[j].min.A[k] = min(bbox[j].min.A[k], vert[ind][k]);
          bbox[j].max.A[k] = max(bbox[j].max.A[k], vert[ind][k]);
        }
        
        real4 lat = (tmp.normal & 255) * (2 * M_PI)/255.0;
        real4 lng = ((tmp.normal >> 8) & 255) * (2 * M_PI)/255.0;
        vnorm[ind].x = cos(lat) * sin(lng);
        vnorm[ind].y = sin(lat) * sin(lng);
        vnorm[ind].z = cos(lng);
      }
    }
    
    // read triangles
    fseek(f, surfaces[s].offsetTriangles, SEEK_SET);

    if (fread(buffer, sizeof(buffer[0])*3*surfaces[s].numTriangles, 1, f) != 1) {
      cerr << "failed reading md3 triangles. " << modelfile << endl;
      delete [] buffer;
      delete [] surfaces;
      fclose(f); return 0;
    }

    FORS (i, surfaces[s].numTriangles) {
      FORS (j, 3) {
        little2machine(buffer[i*3+j]);
        poly[i + poly_offset].vert[j] = buffer[i*3+j] + vert_offset; 
      }
    }

    poly_offset += surfaces[s].numTriangles;
    vert_offset += surfaces[s].numVerts;
  }

  fclose(f);
  delete [] buffer; buffer = 0;
  delete [] surfaces; surfaces = 0;

  FORU (i, nvert*2) little2machine(tvert[i]);
  
  if (load_textures(texfile.substr(0, texfile.length()-4), gamma) == 0) {
    REM2("failed loading textures!", texfile.substr(0, texfile.length()-4).c_str());
    return 0;
  }


  // try to find a cfg file with animation names
  string fname = modelfile.substr(0, modelfile.length()-4) + ".cfg";
  load_cfg(fname);

  compute_centroid();

  //  fast_save(modelfile + ".fast");  // save fast model

  loaded = true; // before init()
  init();
  return 1;
}

//-------------------------------------------------------------------

void MD3_Model::draw(sint4 frame)
{
  assert(loaded);
  assert(frame >= 0 && frame < (sint4)nframe);

  glBegin(GL_TRIANGLES);

  const sint4 start = nvert*frame;
  
  FORU (i, npoly) {

    // a bit faster
    const sint2 v0 = poly[i].vert[0];
    const sint2 v1 = poly[i].vert[1];
    const sint2 v2 = poly[i].vert[2];    
    
    glTexCoord2fv(&tvert[2*v0]);
    glNormal3fv(vnorm[start+v0].ptr());
    glVertex3fv(vert[start+v0].ptr());
    
    glTexCoord2fv(&tvert[2*v1]);
    glNormal3fv(vnorm[start+v1].ptr());
    glVertex3fv(vert[start+v1].ptr());
    
    glTexCoord2fv(&tvert[2*v2]);
    glNormal3fv(vnorm[start+v2].ptr());
    glVertex3fv(vert[start+v2].ptr());

  }    
  glEnd();
}

//-------------------------------------------------------------------

void MD3_Model::draw(sint4 frame, sint4 team)
{
  //  assert(loaded); // checked in draw(frame)

  glColor4f(1,1,1,1);
  glBindTexture(GL_TEXTURE_2D, tex->base_tex);
  draw(frame);
  glDepthFunc(GL_LEQUAL);
  glBindTexture(GL_TEXTURE_2D, tex->team_tex);
  glColor4fv(colors[team].ptr());
  draw(frame);
  glColor4f(1,1,1,1);
}

//-------------------------------------------------------------------

void MD3_Model::draw_interp(sint4 f1, sint4, real4, sint4 team)
{
  draw(f1, team);
}

//-------------------------------------------------------------------

void MD3_Model::draw_shadow(sint4 frame, real4 shadow_sc)
{
  assert(loaded);
  assert(frame >= 0 && frame < (sint4)nframe);

  glBegin(GL_QUADS);

  Vec3<real4> diff = bbox[frame].max - bbox[frame].min;
  Vec3<real4> f0   = bbox[frame].min - diff*shadow_sc;
  Vec3<real4> f1   = bbox[frame].max + diff*shadow_sc;

  glTexCoord2f(0, 0);  glVertex3f(f0.x, f0.y, f0.z);
  glTexCoord2f(1, 0);  glVertex3f(f1.x, f0.y, f0.z);
  glTexCoord2f(1, 1);  glVertex3f(f1.x, f0.y, f1.z);
  glTexCoord2f(0, 1);  glVertex3f(f0.x, f0.y, f1.z);
  glEnd();
}

//-------------------------------------------------------------------

void MD3_Model::select(sint4 frame)
{
  assert(loaded);
  assert(frame >= 0 && frame < (sint4)nframe);

  glColor3f(0.5, 0.5, 0.5);
  glBegin(GL_TRIANGLES);
  FORU (i, npoly) {
    FORS (j, 3) {
      glVertex3fv(vert[poly[i].vert[j] + nvert*frame].ptr());
    }
  }
  glEnd();
}

//===================================================================

void MD3_Model_DL::draw(sint4 frame)
{
  assert(loaded);
  assert(frame >= 0 && frame < (sint4)nframe);
  glCallList(display_lists + frame);
}

void MD3_Model_DL::draw_interp(sint4 f1, sint4, real4, sint4 team)
{
  // Can't do interpolation with display lists
  glColor4f(1,1,1,1);
  glBindTexture(GL_TEXTURE_2D, tex->base_tex);
  draw(f1);
  glDepthFunc(GL_LEQUAL);
  glBindTexture(GL_TEXTURE_2D, tex->team_tex);
  glColor4fv(colors[team].ptr());
  draw(f1);
  glColor4f(1,1,1,1);
}

void MD3_Model_DL::select(sint4)
{
  glCallList(select_list);
}

//===================================================================
