// $Id: Minimap.C 7355 2008-08-06 11:58:11Z furtak $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "GfxGlobal.H"
#include "Global.H"
#include "Minimap.H"
#include "GameObj.H"
#include "Game.H"
#include "Options.H"

using namespace std;

#define MAX_RES_BITS 8 // 256x256

//===================================================================

Minimap::Minimap(Game *g) : game(g)
{
  map_width  = game->get_map().get_width();
  map_height = game->get_map().get_height();
  map_tiles  = map_width * map_height;

  tile_points = game->get_tile_points();

  // find the closest available texture resolution that fits the map size

  texture_width = 1;
  while (texture_width < map_width) texture_width <<= 1;
  scale_w = real4(map_width) / real4(texture_width);
  texture_height = 1;
  while (texture_height < map_height) texture_height <<= 1;
  scale_h = real4(map_height) / real4(texture_height);

  texture_pixels = texture_width*texture_height;

  changed = false;
  initialize_colors();

  assert(texture_pixels != 0);
  
  color_texture = new GLubyte[texture_pixels*3];
  visible_texture = new GLubyte[texture_pixels];
  explored = new uint1[texture_pixels];

  assert(map_tiles != 0);
  
  memset(color_texture,  0, texture_pixels * 3);
  memset(visible_texture, 0xFF, texture_pixels);
  memset(explored, 0, texture_pixels);

  //----------texture----------

  vtexture_mode = GL_ALPHA; // visibility texture

  glGenTextures(1, &color_texture_id);
  glBindTexture(GL_TEXTURE_2D, color_texture_id);
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGB, texture_width, texture_height, 0,
               GL_RGB, GL_UNSIGNED_BYTE, color_texture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

  glGenTextures(1, &visible_texture_id);
  glBindTexture(GL_TEXTURE_2D, visible_texture_id);
  glTexImage2D(GL_TEXTURE_2D,0,vtexture_mode,  texture_width, texture_height, 0,
               vtexture_mode, GL_UNSIGNED_BYTE, visible_texture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

  cout << "map size = " << map_width <<"X"<< map_height << endl;
  cout << "texture size = " << texture_width<<"X"<<texture_height << endl;
}

//===================================================================

Minimap::~Minimap()
{
  delete [] color_texture;
  delete [] visible_texture;
  delete [] explored;
}

//===================================================================

void Minimap::new_tile(uint4 x, uint4 y, uint1 height, uint1 ground)
{
  if (height >= height_colors.size()) height = height_colors.size()-1;
  if (ground >= ground_colors.size()) ground = ground_colors.size()-1;

  sint4 o1 = y*texture_width+x;
  sint4 offset = 3*o1;
  const Col3 &gc = ground_colors[ground] * 255.0;
  const Col3 &hc = height_colors[height];
    
  color_texture[offset]   = static_cast<GLubyte>(hc.r * gc.r + 0.5);
  color_texture[offset+1] = static_cast<GLubyte>(hc.g * gc.g + 0.5);
  color_texture[offset+2] = static_cast<GLubyte>(hc.b * gc.b + 0.5);

  explored[o1] = 1;
  changed = true;
}

//===================================================================

bool Minimap::update_texture()
{
  Vector<MapView*> mv;
  sint4 watch;
  Options::get("-watch", watch);
  if (watch > 0) {
    FORU (i, game->get_player_num()) {
      if (watch & (1 << i)) {
        MapView &m = game->get_player_info(i).visible_W;
        mv.push_back(&m);
      }
    }
  } else {
    MapView &m = game->get_cplayer_info().visible_W;
    mv.push_back(&m);
  }

  // these values are shadow intensities out of 255
  const sint4 known_color = 160; // known terrain, but not visible


  // only one player is looking
  if (mv.size() <= 1) {

    FORS (y, map_height) {
      sint4 i0 = y*texture_width;
      FORS (x, map_width) {
        sint4 i = i0+x;
        
        if (explored[i]) {
          sint4 mult = mv[0]->is_set(x, y) ? 0 : known_color;
          // blend towards the new visibility state, avoids sharp transitions
          visible_texture[i] = (visible_texture[i] + (uint1)mult) / 2;
        }
      }
    }

  } else {

    // try to hint at multiple players' visibility maps for the watcher

    FORS (y, map_height) {
      sint4 i0 = y*texture_width;
      FORS (x, map_width) {
        sint4 i = i0+x;
        
        if (explored[i]) {
          sint4 mult = 0;  // number of players seeing the square
          FORU (q, mv.size()) {
            if (mv[q]->is_set(x, y)) mult += 1;
          }
          mult = known_color - mult * 100; // each player contributes 100
          if (mult < 0) mult = 0;          // can't go lighter than this
          visible_texture[i] = (visible_texture[i] + (uint1)mult) / 2;
        }
      }
    }
  }

  // We are drawing the SHADOW here, so the lower visible_texture[i], the brighter
  // the terrain is going to be.

  glBindTexture(GL_TEXTURE_2D, visible_texture_id);
  glTexSubImage2D(GL_TEXTURE_2D, 0,0,0, texture_width, map_height,
		  vtexture_mode, GL_UNSIGNED_BYTE, visible_texture);

  if (changed) {
    glBindTexture(GL_TEXTURE_2D, color_texture_id);
    glTexSubImage2D(GL_TEXTURE_2D, 0,0,0, texture_width, map_height,
		    GL_RGB, GL_UNSIGNED_BYTE, color_texture);
    changed = false;
  }

  return true;
}

//===================================================================

void Minimap::initialize_colors()
{
  height_colors.clear();
  ground_colors.clear();
  visible_colors.clear();
  
  height_colors.push_back(Col3(0.4, 0.4, 0.4));
  height_colors.push_back(Col3(0.6, 0.6, 0.6));
  height_colors.push_back(Col3(0.8, 0.8, 0.8));  

  ground_colors.push_back(Col3(1, 1, 1));          // unknown
  ground_colors.push_back(Col3(0.38, 0.38, 0.62)); // water
  ground_colors.push_back(Col3(0.4, 0.9, 0.3));    // grass

  visible_colors.push_back(0x80);
  visible_colors.push_back(0x00);
}

//===================================================================

bool Minimap::has_changed()
{
  bool rvalue = changed;
  changed = false;
  return rvalue;
}

//===================================================================



#if 0
uint4 Minimap::col2uint4(const Col3 &color)
{
  uint4 r,g,b;
  r = (uint4) color.r * 255;
  g = (uint4) color.g * 255;
  b = (uint4) color.b * 255;
  
  return (0xFF << 24) + (b << 16) + (g << 8) + r;
}
#endif
