
// $Id: Shader.C 7357 2008-08-07 02:38:57Z furtak $

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "Shader.H"
#include "Global.H"
#include "TextureStore.H"

using namespace std;

//===================================================================

set<ShaderProgram*> ShaderProgram::shaders;
set<CustomShaderProgram*> CustomShaderProgram::shaders;

//===================================================================

void Shader::clear()
{
  if (shader) {
    if (glDeleteShader) glDeleteShader(shader);
    shader = 0;
  }
}

//-------------------------------------------------------------------

void Shader::compile()
{
  if (!glCompileShader || !glGetObjectParameterivARB) return;

  glCompileShader(shader);
 
  GLint compiled = 0;
  glGetObjectParameterivARB(shader, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);
  if (compiled) {
    // OK
  } else {
    if (shader == 0) return;

    GLsizei len = 0;
    GLsizei log_len = 0;
    
    glGetObjectParameterivARB(shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, &len);
    
    char *log = new char[len];

    glGetInfoLogARB(shader, len, &log_len, log);
    cout << "COMPILE ERROR: " << log << endl;

    delete [] log;
    //    ERR("");
  }
}

//-------------------------------------------------------------------

void ShaderProgram::link()
{
  if (!glLinkProgram || !glGetProgramiv) return;

  glLinkProgram(prog); 

  GLint linked = 0;
  glGetProgramiv(prog, GL_LINK_STATUS, &linked);
  
  if (linked) {
    // OK
  } else {
    if (prog == 0) return;

#if 1

    ERR2("link error! ", prog);

#else
    // fixme: not supported by some 
    GLsizei len = 0;
    GLsizei log_len = 0;
    
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &len);
    char *log = new char[len];

    glGetProgramInfoLog(prog, len, &log_len, log);
    cout << "LINK ERROR: " << log << endl;

    delete [] log;
    //    ERR("");
#endif    
  }
}

//-------------------------------------------------------------------

char* Shader::read_source(const string &filename)
{
  ifstream file;
  file.open(filename.c_str(), ios::binary);
  if (!file.is_open()) {
    return 0;
    ERR2("can't open file:", filename.c_str());
  }

  file.seekg(0, ios::end);
  sint4 length = file.tellg();
  file.seekg(0, ios::beg);

  char *buffer = new char[length+1];
  file.read(buffer, length);
  file.close();
  buffer[length] = 0;

  return buffer;
}

//===================================================================

bool VertexShader::load(const string &filename)
{
  if (!glCreateShader || !glDeleteShader || !glShaderSource) return false;

  if (shader) {
    glDeleteShader(shader);
    shader = 0;
  }
  const char *s = read_source(filename);
  if (s == 0) return false;
  shader = glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(shader, 1, &s, NULL);
  compile();
  delete [] s;
  return true;
}

//===================================================================

bool FragmentShader::load(const string &filename)
{
  if (!glCreateShader || !glDeleteShader || !glShaderSource) return false;

  if (shader) {
    glDeleteShader(shader);
    shader = 0;
  }
  const char *s = read_source(filename);
  if (s == 0) return false;
  shader = glCreateShader(GL_FRAGMENT_SHADER_ARB);
  glShaderSource(shader, 1, &s, NULL);
  compile();
  delete [] s;
  return true;
}

//===================================================================

ShaderProgram::ShaderProgram()
{
  fs = 0;
  vs = 0;
  prog = 0;
}

//-------------------------------------------------------------------

ShaderProgram::~ShaderProgram()
{
  shaders.erase(this);
}

//-------------------------------------------------------------------

void ShaderProgram::cleanup()
{
  FORALL (shaders, it) {
    delete *it;
  }
  shaders.clear();
}

//-------------------------------------------------------------------

void ShaderProgram::fragment_shader(const string &filename)
{
  if (!glCreateProgram || !glAttachShader || !glDetachShader) return;
  shaders.insert(this);

  if (fs) {
    glDetachShader(prog, fs->shader);
    delete fs;
    fs = 0;
  }
  if (filename != "") {
    fs = new FragmentShader;
    if (fs->load(filename)) {
      cout << "frag load " << filename << " = true" << endl;
      if (prog == 0) prog = glCreateProgram();
      glAttachShader(prog, fs->shader);
      link();
    } else {
      delete fs;
      fs = 0;
    }
  }
  fname = filename;
}

//-------------------------------------------------------------------

void ShaderProgram::vertex_shader(const string &filename)
{
  if (!glCreateProgram || !glAttachShader || !glDetachShader) return;
  shaders.insert(this);

  if (vs) {
    glDetachShader(prog, vs->shader);
    delete vs;
    vs = 0;
  }
  if (filename != "") {
    vs = new VertexShader;
    if (vs->load(filename)) {
      cout << "vert load " << filename << " = true" << endl;
      if (prog == 0) prog = glCreateProgram();
      glAttachShader(prog, vs->shader);
      link();
    } else {
      delete vs;
      vs = 0;
    }
  }
  vname = filename;
}

//-------------------------------------------------------------------

void ShaderProgram::reload()
{
  vertex_shader(vname);
  fragment_shader(fname);
}

//-------------------------------------------------------------------

void ShaderProgram::reload_all()
{
  FORALL (shaders, it) (*it)->reload();
}

//===================================================================

CustomShaderProgram::CustomShaderProgram()
{
  shaders.insert(this);
}

//-------------------------------------------------------------------

CustomShaderProgram::~CustomShaderProgram()
{
  set<CustomShaderProgram*>::iterator it = shaders.find(this);
  if (it != shaders.end()) shaders.erase(it);
}

//-------------------------------------------------------------------

CustomShaderProgram::Pass::Pass()
{
  prog = 0;
}

//-------------------------------------------------------------------

CustomShaderProgram::Pass::~Pass()
{
  delete prog;
}

//-------------------------------------------------------------------

void CustomShaderProgram::cleanup()
{
  FORALL (shaders, it) {
    delete *it;
  }
  shaders.clear();
}

//-------------------------------------------------------------------

bool CustomShaderProgram::load(ShaderProgram *sp)
{
  passes.clear();
  if (sp->id() == 0) return false;

  shaders.insert(this);
  passes.resize(1);
  Pass &p = passes[0];
  p.prog = sp;
  return true;
}

//-------------------------------------------------------------------

/**
   Format: vertex_shader0 frag_shader0
           varname0 texname0
           varname1 texname1
           ...
*/

bool CustomShaderProgram::load(const string &filename)
{
  fname = filename;
  
  ifstream file;
  file.open(filename.c_str());
  if (!file.is_open()) {
    //    REM2("can't open file:", filename.c_str());
    return false;
  }

  cout << "custom shader load " << filename << endl;

  string fbase;
  {
    size_t i = filename.find_last_of(PATH_SEP);
    if (i == string::npos) i = filename.size()+1;
    fbase = filename.substr(0, i);
  }

  passes.clear();

  FOREVER {
    string vfile, ffile;
    string t;
    getline(file, t);
    stringstream ss(t);

    ss >> vfile >> ffile;
    if (vfile == "" || ffile == "") break;
    vfile = fbase + PATH_SEP + vfile;
    ffile = fbase + PATH_SEP + ffile;

    passes.resize(passes.size()+1);
    Pass &p = passes.back();

    p.prog = new ShaderProgram;
    p.prog->vertex_shader(vfile);
    p.prog->fragment_shader(ffile);

    FOREVER {
      string t;
      getline(file, t);
      stringstream ss(t);
      string vname, tex;
      ss >> vname >> tex;
      if (vname == "" || tex == "") break;

      // see comments in setup() if things seriously break here
      GLint ploc = glGetUniformLocation(p.prog->id(), vname.c_str());
      if (ploc < 0) {
        REM2("invalid/unused shader variable:", vname.c_str());
      }

      string tfile = fbase + PATH_SEP + tex;

      GLuint id = TextureStore::load(tfile, Image::BILINEAR);
      if (id == 0) {
        REM2("invalid shader texture:", tfile.c_str());
      }

      p.var_loc.push_back(ploc);
      p.var_name.push_back(vname);
      p.var_tex.push_back(id);
    }
  }
  return true;
}

//-------------------------------------------------------------------

void CustomShaderProgram::reload()
{
  load(fname);
}

//-------------------------------------------------------------------

void CustomShaderProgram::reload_all()
{
  FORALL (shaders, it) (*it)->reload();
}

//-------------------------------------------------------------------

void CustomShaderProgram::Pass::setup() const
{
#if 1
  FORU (i, var_loc.size()) {
    if (var_loc[i] < 0) continue;
    glActiveTextureARB(GL_TEXTURE0 + 4 + i);
    glBindTexture(GL_TEXTURE_2D, var_tex[i]);
    glUniform1i(var_loc[i], 4+i);
  }
#else
  // Possible but unlikely that the shader needs to be active before
  // the location can be queried -- so the cached value is wrong => requery.
  FORU (i, var_name.size()) {
    GLint loc = glGetUniformLocation(prog->id(), var_name[i].c_str());
    if (loc < 0) continue;
    glActiveTextureARB(GL_TEXTURE0 + 4 + i);
    glBindTexture(GL_TEXTURE_2D, var_tex[i]);
    glUniform1i(loc, 4+i);
  }
#endif
}

//===================================================================

