#ifndef Blueprints_H
#define Blueprints_H

/** @file Blueprints.H
    GameObj blueprints.

    $Id: Blueprints.H 5279 2007-06-23 02:49:08Z mburo $
    $Source$
*/

// This is an ORTS file (c) Tim Furtak, Michael Buro, licensed under the GPL

#include "Global.H"
#include "Script.H"
#include "Action.H"
#include "GameObjFactory.H"
#include "Attr.H"

class Tokenizer;
class PlayerInfo;
class ScriptNode;

//===================================================================

/** Describes GameObj structure and initial starting values/states. */
/** \nosubgrouping */

class GameObjBlueprint {

public:
// This must be public to enable the "REGISTER_TYPEOF" definitions:
  typedef EXT_NAMESPACE::hash_map<std::string, sint4, StringHash> AttrMap;

protected:

  static Vector<ObjActionDefinition> ext_action_defs; ///< Actions not local to the blueprint.
  static Vector<ObjAction*>          ext_actions;

  uint4 action_count;      ///< # of actions in the subtree rooted here (inclusive)
  uint4 root_action_count; ///< # of actions at (only) this level

  Vector<ObjActionDefinition> action_defs;
  Vector<ObjActionDefinition> constructor_defs;
  Vector<ObjActionDefinition> destructor_defs;
  Vector<ObjAction*> constructor; ///< Automatically called when the object is created.
  Vector<ObjAction*> destructor;  ///< Automatically called when the object is killed (not deleted).

  Vector<ObjAction*> hooks;       ///< Actions to call when an attribute value has changed.

  bool locked;
  static bool all_locked; ///< Set to true after the game is loaded.  Extensions don't need to be locked.

  Vector<ScriptNode*> init_vals; ///< These need to be kept around in the case of string pointers.

  GameObjBlueprint& operator= (const GameObjBlueprint &x);
  void sort_attributes();

  AttrMap attributes;

  bool can_instance; ///< If the blueprint can exist as a stand-alone object.

  void parse_action(Tokenizer &tokenizer, std::string act_type,
                    Game &game, PlayerInfo &pi, bool read_constructors);
  void parse_attr(Tokenizer &tokenizer, std::string cmd);
  void parse_inherit(Tokenizer &tokenizer, GameObjFactory::Name2BlueprintMap &classes);
  void parse_comp(Tokenizer &tokenizer,
                  GameObjFactory::Name2BlueprintMap &classes,
                  GameObjFactory::Name2ConstructorMap &classes2);

  std::set<std::string> base_classes;

public:

  std::string name;   ///< Set by register_blueprint. Needed by gfx.
  bool enabled;  ///< Whether (sub)objects created from this blueprint are initially enabled.
  sint4 enable_count; ///< 1 - # of disabled parent objects.

  sint4 code;

  //===================================================================
  /// @name Attributes
  //===================================================================
  ///@{
  sint4 attr_index(const std::string &a);

  Vector<std::string> attr_names;
  Vector<stype>       attr_vals;  ///< Default values assigned to each GameObj.
  Vector<AttrProp>    attr_props;
  sint4 num_const, num_stable, num_variable;

  void add_attr(const std::string &name, const AttrProp &af, stype val);
  void change_val(const std::string &name, stype val, bool force=false);
  void change_val(const std::string &name, const AttrProp &af, stype val, bool force=false);
  ///@}

  //===================================================================
  /// @name Components
  //===================================================================
  ///@{
  typedef std::map<std::string, sint4> Name2Comp;
  struct BP_Const {
    GameObjBlueprint *bp;
    GameObjFactory::BPConstructorData c;
  };

  sint4 num_comp;      ///< Number of components in the game version (extended may be more).
  Name2Comp name2comp; // bound after parsing
  Vector<BP_Const> component;
  Vector<CompProp> comp_props;

  void attach_component(const std::string &var_name, const GameObjBlueprint *comp, bool enabled);
  ///@}

  //===================================================================
  /// @name Actions
  //===================================================================
  ///@{

  // all the actions at this level, regardless of type or location
  Vector<std::string>               action_names;
  std::map<std::string, ObjAction*> action_map;
  std::map<std::string, sint4>      action_name2id; ///< Actions that the client can send to the server.
  Vector<ObjAction*>                actions; // also includes all component actions

  static void add_ext_action(const ObjActionDefinition &def);
  void attach_action(const ObjActionDefinition &def);
  ///@}
  //===================================================================

  GameObjBlueprint(const std::string &name, bool can_instance=true);
  ~GameObjBlueprint();
  GameObjBlueprint(const GameObjBlueprint &x);

  void parse_class_def(Tokenizer &tokenizer,
                       GameObjFactory::Name2BlueprintMap &classes,
                       GameObjFactory::Name2ConstructorMap &classes2,
                       Game &game, PlayerInfo &pi, bool read_constructors=true);


  void lock(sint4 id_start = 0);
  void bind(PlayerInfo &pi, Game &game, const std::string &bp_name);
  void clear_action_map();

  friend class GameObjFactory;
  friend class GameObj;
  friend class Game;
};

//===================================================================

REGISTER_TYPEOF(90, Vector<ObjActionDefinition>::iterator);
REGISTER_TYPEOF(91, Vector<ObjActionDefinition>::const_iterator);

//REGISTER_TYPEOF(92, GameObjBlueprint::Name2Comp::iterator);
//REGISTER_TYPEOF(93, GameObjBlueprint::Name2Comp::const_iterator);

REGISTER_TYPEOF(94, Vector<GameObjBlueprint::BP_Const>::iterator);
REGISTER_TYPEOF(95, Vector<GameObjBlueprint::BP_Const>::const_iterator);

REGISTER_TYPEOF(96, GameObjBlueprint::AttrMap::iterator);
REGISTER_TYPEOF(97, GameObjBlueprint::AttrMap::const_iterator);

#endif
