/** @file Collision.C

    @todo Speed improvement (pythagoras) possible! (No one knows what this refers to.)

    $Id: Collision.C 5279 2007-06-23 02:49:08Z mburo $
    $Source$
*/

// This is an ORTS file
// (c) Michael Buro  
// (c) Tim Furtak    
// licensed under the GPL

#include "Game.H"
#include "GameObj.H"
#include "Collision2D.H"

using namespace std;
using namespace collision2d;

#define COLL_PAR 0   // print collision parameters

#define COUNT_PATHS 0
#define PCN 28
sint4 pc[PCN];
sint4 pcn;

#if COUNT_PATHS
#define INCPC(i)   { assert(i>=0 && i < PCN); ++pc[i]; }
#define PRINTPC    print_path_counts()
#else
#define INCPC(i)
#define PRINTPC
#endif

#define TEST_OUT 0

template <class T> T sgn(const T& x)
{
  if (x >= 0) return 1;
  return -1;
}

//===================================================================

/** Compute collision time between two objects relative to some starting time.
    @return >= 2 if no such collision exists; -1 if too close.
*/

real8 Game::collision_time(GameObj &e1, GameObj &e2, real8 ct, sint4 &coll_type) const
{
  real8 t;

  bool can_collide = (*e1.sod.collides & *e2.sod.collides);
  if (can_collide) can_collide = !(*e1.sod.ncollides & *e2.sod.ncollides);

  if (!can_collide) return 2.0;

  PRINTPC;

  Object::Shape s1 = e1.get_shape();
  Object::Shape s2 = e2.get_shape();

  // circle - circle -------------
  if (s1 == Object::CIRCLE &&
      s2 == Object::CIRCLE)
    t = collision_time_circle_circle(e1, e2, ct, coll_type);

  // circle - [hv]line -------------
  else if (s1 == Object::CIRCLE &&
           s2 == Object::LINE)
    t = collision_time_circle_line(e1, e2, ct, coll_type);

  else if (s1 == Object::LINE &&
           s2 == Object::CIRCLE)
    t = collision_time_circle_line(e2, e1, ct, coll_type);

  // circle - rectangle ----------
  else if (s1 == Object::CIRCLE &&
           s2 == Object::RECTANGLE)
    t = collision_time_circle_rect(e1, e2, ct, coll_type);

  else if (s1 == Object::RECTANGLE &&
           s2 == Object::CIRCLE)
    t = collision_time_circle_rect(e2, e1, ct, coll_type);

  // [hv]line - [hv]line ----------
  else if (s1 == Object::LINE &&
           s2 == Object::LINE)
    t = collision_time_line_line(e2, e1, ct, coll_type);

  // [hv]line - rectangle ---------
  else if (s1 == Object::LINE &&
           s2 == Object::RECTANGLE)
    t = collision_time_line_rect(e1, e2, ct, coll_type);

  else if (s2 == Object::LINE &&
           s1 == Object::RECTANGLE)
    t = collision_time_line_rect(e2, e1, ct, coll_type);

  // rectangle - rectangle --------
  else if (s1 == Object::RECTANGLE &&
           s2 == Object::RECTANGLE)
    t = collision_time_rect_rect(e2, e1, ct, coll_type);

  // ------------------------------
  else {
    cout << s1 << " " << s2 << endl;
    ERR("undefined shape pair");
    return 2;
  }

#ifndef COLL_PAR
  cout << "colltime " << &e1 << " " << &e2 << " " << ct << " -> " << t << endl;
#endif

  return t;
}

//===================================================================

/** Computes the collision normal between two objects at time ct to determine
    which objects should be stopped based on their headings.
    Computed separately to avoid unnecessary work in collision_time().
    collision_type is set by collision_time(), and is defined by the
    types of objects colliding.

    Note: collision_time() aggressively stops movement if the objects intersect,
          this check resolves cases where objects are moving away from each other.
          If only one object is moving you should not assume that the object
          should be stopped.

    @return 1 if the first object is stopped, 2 if the second, 3 if both.
*/
sint4 Game::resolve_collision(const GameObj &e1, const GameObj &e2, real8 ct, sint4 collision_type) const
{
  sint4 stopped = 3;

  sint4 s1 = *e1.sod.shape;
  sint4 s2 = *e2.sod.shape;

  //  cout << "shapes = " << s1 << " " << s2 << endl;

  //------------------------------
  if (s1 == Object::CIRCLE && s2 == Object::CIRCLE)
    stopped = resolve_collision_circle_circle(e1, e2, ct, collision_type);
  //------------------------------
  else if (s1 == Object::CIRCLE && s2 == Object::LINE)
    stopped = resolve_collision_circle_line(e1, e2, ct, collision_type);

  else if (s2 == Object::CIRCLE && s1 == Object::LINE) {
    stopped = resolve_collision_circle_line(e2, e1, ct, collision_type);
    if (stopped == 1) stopped = 2;
    else if (stopped == 2) stopped = 1;
  }
  //------------------------------
  else if (s1 == Object::CIRCLE && s2 == Object::RECTANGLE)
    stopped = resolve_collision_circle_rect(e1, e2, ct, collision_type);

  else if (s1 == Object::RECTANGLE && s2 == Object::CIRCLE) {
    stopped = resolve_collision_circle_rect(e2, e1, ct, collision_type);
    if (stopped == 1) stopped = 2;
    else if (stopped == 2) stopped = 1;
  }
  //------------------------------
  /*
  else if (s1 == Object::HVLINE &&
           s2 == Object::HVLINE)
    stopped = resolve_collision_hvline_hvline(e2, e1, ct, collision_type);
  else {
    cout << s1 << " " << s2 << endl;
    ERR("undefined shape pair");
    stopped = 3;
  }
  */
  return stopped;
}

//===================================================================

#define CIRCLE_VARS(a, c, h, r)\
  ScalarPoint c;                     \
  ScalarPoint h;                     \
  real8 r;                     \
  if (*a.sod.speed) {          \
    c = a.smd.fine_pos;        \
    h = a.smd.fine_heading;    \
  } else {                     \
    c.x = fine_grid(*a.sod.x); \
    c.y = fine_grid(*a.sod.y); \
    h.x = h.y = 0;             \
  }                            \
  r = fine_grid(*a.sod.radius);

//===================================================================

sint4 Game::resolve_collision_circle_point(const GameObj &e1,
                                           const ScalarPoint &c2, const ScalarPoint &h2,
                                           real8 ct) const
{
  CIRCLE_VARS(e1, c1, h1, r1);
  real8 rr1 = r1*r1;

  return collision2d::resolve_collision_circle_point(c1, rr1, h1, c2, h2, ct);
}

//===================================================================

sint4 Game::resolve_collision_circle_circle(const GameObj &e1, const GameObj &e2,
                                            real8 ct, sint4 /*collision_type*/) const
{
  CIRCLE_VARS(e1, c1, h1, r1);
  CIRCLE_VARS(e2, c2, h2, r2);

  real8 rr1 = r1*r1;
  real8 rr2 = r2*r2;

  return collision2d::resolve_collision_circle_circle(c1, rr1, h1, c2, rr2, h2, ct);
}

//===================================================================

sint4 Game::resolve_collision_circle_line(const GameObj &e1, const GameObj &e2,
                                          real8 ct, sint4 collision_type) const
{
  CIRCLE_VARS(e1, c1, h1, r1);
  
  ScalarPoint a1, a2;
  a1.x = fine_grid(*e2.sod.x1);
  a1.y = fine_grid(*e2.sod.y1);
  a2.x = fine_grid(*e2.sod.x2);
  a2.y = fine_grid(*e2.sod.y2);
  ScalarPoint h2(0,0);
  if (*e2.sod.speed) h2 = e2.smd.fine_heading;

  return collision2d::resolve_collision_circle_line(c1, r1, h1, a1, a2, h2, collision_type, ct);
}

//===================================================================

sint4 Game::resolve_collision_circle_rect(const GameObj &e1, const GameObj &e2,
                                          real8 ct, sint4 collision_type) const
{
  //------------------------------

  CIRCLE_VARS(e1, c, h1, r);

  //------------------------------

  ScalarPoint a1, a2;
  ScalarPoint h2(0,0);
  if (*e2.sod.speed) {
    sint4 w = e2.get_int("width");
    sint4 h = e2.get_int("height");
    a1.x = e2.smd.fine_pos.x - fine_grid(w/2);
    a1.y = e2.smd.fine_pos.y - fine_grid(h/2);
    a2.x = a1.x + fine_grid(w);
    a2.y = a1.y + fine_grid(h);
    h2 = e2.smd.fine_heading;
  } else {
    sint4 w = e2.get_int("width");
    sint4 h = e2.get_int("height");
    a1.x = fine_grid(*e2.sod.x) - fine_grid(w/2);
    a1.y = fine_grid(*e2.sod.y) - fine_grid(h/2);
    a2.x = a1.x + fine_grid(w);
    a2.y = a1.y + fine_grid(h);
  }

  //------------------------------

  return collision2d::resolve_collision_circle_rect(c, r, h1, a1, a2, h2, collision_type, ct);
}

//===================================================================
//===================================================================

/** Compute collision time of a moving circle and a still line.
    Moves circle to time ct first and scales heading difference accordingly.
    Ignore collisions before time ct.
*/

real8 Game::collision_time_circle_line(const GameObj &e1,
                                       const GameObj &e2, real8 ct, sint4 &coll_type) const
{
  assert(*e1.sod.shape == Object::CIRCLE &&
	 *e2.sod.shape == Object::LINE);

  //------------------------------

  CIRCLE_VARS(e1, c, h1, r);

  //------------------------------

  ScalarPoint a1, a2;
  a1.x = fine_grid(*e2.sod.x1);
  a1.y = fine_grid(*e2.sod.y1);
  a2.x = fine_grid(*e2.sod.x2);
  a2.y = fine_grid(*e2.sod.y2);
  ScalarPoint h2(0,0);
  if (*e2.sod.speed) h2 = e2.smd.fine_heading;

  //------------------------------

  return collision2d::collision_time_circle_line(c, r, h1, a1, a2, h2, coll_type, ct);
}

//===================================================================

/** Compute collision time of a moving circle and a still rectangle.
    Moves circle to time ct first and scales heading difference accordingly.
    @return Collision time relative to ct.
*/

real8 Game::collision_time_circle_rect(const GameObj& e1,
                                       const GameObj& e2, real8 ct, sint4 &coll_type) const
{
  assert(*e1.sod.shape == Object::CIRCLE &&
         *e2.sod.shape == Object::RECTANGLE);

  //------------------------------

  CIRCLE_VARS(e1, c, h1, r);

  //------------------------------

  ScalarPoint a1, a2;
  ScalarPoint h2(0,0);
  if (*e2.sod.speed) {
    sint4 w = e2.get_int("width");
    sint4 h = e2.get_int("height");
    a1.x = e2.smd.fine_pos.x - fine_grid(w/2);
    a1.y = e2.smd.fine_pos.y - fine_grid(h/2);
    a2.x = a1.x + fine_grid(w);
    a2.y = a1.y + fine_grid(h);
    h2 = e2.smd.fine_heading;
  } else {
    sint4 w = e2.get_int("width");
    sint4 h = e2.get_int("height");
    a1.x = fine_grid(*e2.sod.x) - fine_grid(w/2);
    a1.y = fine_grid(*e2.sod.y) - fine_grid(h/2);
    a2.x = a1.x + fine_grid(w);
    a2.y = a1.y + fine_grid(h);
  }

  //------------------------------

  return collision2d::collision_time_circle_rect(c, r, h1, a1, a2, h2, coll_type, ct);
}

//===================================================================

/** Compute collision time of two circles.
    Moves circles to time ct first and scales heading difference accordingly.
    stopped is set to reflect which of the objects actually ran into something
    (1 = e1, 2 = e2, 3 = both)
    @return Collision time relative to ct.
*/

real8 Game::collision_time_circle_circle(const GameObj &e1, const GameObj &e2,
                                         real8 ct, sint4 &/*coll_type*/) const
{
  assert(*e1.sod.shape == Object::CIRCLE &&
         *e2.sod.shape == Object::CIRCLE);
  assert (ct >= 0.0L && ct <= 1.0L);

  // 2D for now

  if ((e1.smd.fine_heading.x == e2.smd.fine_heading.x) &&
      (e1.smd.fine_heading.y == e2.smd.fine_heading.y)) return 2.0;

  //------------------------------

  CIRCLE_VARS(e1, c, h1, r);

  //------------------------------

  ScalarPoint p;       // center
  ScalarPoint h2(0,0); // heading
  if (*e2.sod.speed) {
    p = e2.smd.fine_pos;
    h2 = e2.smd.fine_heading;
  } else {
    p.x = fine_grid(*e2.sod.x);
    p.y = fine_grid(*e2.sod.y);
  }
  r += fine_grid(*e2.sod.radius); // radius
  real8 rr = r*r;

  //------------------------------

  return collision2d::collision_time_circle_point(c, rr, h1, p, h2, ct);
}

//===================================================================

/** Compute collision time of two lines.
*/

real8 Game::collision_time_line_line(const GameObj &e1, const GameObj &e2,
                                     real8 ct, sint4 &coll_type) const
{
  ScalarPoint a1(fine_grid(*e1.sod.x1), fine_grid(*e1.sod.y1));
  ScalarPoint a2(fine_grid(*e1.sod.x2), fine_grid(*e1.sod.y2));
  ScalarPoint b1(fine_grid(*e2.sod.x1), fine_grid(*e2.sod.y1));
  ScalarPoint b2(fine_grid(*e2.sod.x2), fine_grid(*e2.sod.y2));

  ScalarPoint h1(0,0);
  ScalarPoint h2(0,0);
  if (*e1.sod.speed) h1 = e1.smd.fine_heading;
  if (*e2.sod.speed) h2 = e2.smd.fine_heading;

  return collision2d::collision_time_line_line(a1, a2, h1, b1, b2, h2, coll_type, ct);
}

//===================================================================

/** Compute collision time of a line and a rectangle.
*/

real8 Game::collision_time_line_rect(const GameObj &e1, const GameObj &e2,
                                     real8 ct, sint4 &coll_type) const
{
  assert(*e1.sod.shape == Object::LINE);
  assert(*e2.sod.shape == Object::RECTANGLE);

  //------------------------------

  ScalarPoint a1, a2;
  a1.x = fine_grid(*e1.sod.x1);
  a1.y = fine_grid(*e1.sod.y1);
  a2.x = fine_grid(*e1.sod.x2);
  a2.y = fine_grid(*e1.sod.y2);
  ScalarPoint h1(0,0);
  if (*e1.sod.speed) h1 = e1.smd.fine_heading;

  //------------------------------

  ScalarPoint b1, b2;
  ScalarPoint h2(0,0);
  if (*e2.sod.speed) {
    sint4 w = e2.get_int("width");
    sint4 h = e2.get_int("height");
    b1.x = e2.smd.fine_pos.x - fine_grid(w/2);
    b1.y = e2.smd.fine_pos.y - fine_grid(h/2);
    b2.x = b1.x + fine_grid(w);
    b2.y = b1.y + fine_grid(h);
    h2 = e2.smd.fine_heading;
  } else {
    sint4 w = e2.get_int("width");
    sint4 h = e2.get_int("height");
    b1.x = fine_grid(*e2.sod.x) - fine_grid(w/2);
    b1.y = fine_grid(*e2.sod.y) - fine_grid(h/2);
    b2.x = b1.x + fine_grid(w);
    b2.y = b1.y + fine_grid(h);
  }

  //------------------------------

  return collision2d::collision_time_line_rect(a1, a2, h1, b1, b2, h2, coll_type, ct);
}

//===================================================================

/** Compute collision time of two rectangles.
    @return Collision time relative to ct.
*/

real8 Game::collision_time_rect_rect(const GameObj &e1, const GameObj &e2,
                                     real8 ct, sint4 &coll_type) const
{
  assert(*e1.sod.shape == Object::RECTANGLE);
  assert(*e2.sod.shape == Object::RECTANGLE);

  //------------------------------

  ScalarPoint a1, a2;
  ScalarPoint h1(0,0);
  if (*e1.sod.speed) {
    sint4 w = e1.get_int("width");
    sint4 h = e1.get_int("height");
    a1.x = e1.smd.fine_pos.x - fine_grid(w/2);
    a1.y = e1.smd.fine_pos.y - fine_grid(h/2);
    a2.x = a1.x + fine_grid(w);
    a2.y = a1.y + fine_grid(h);
    h1 = e1.smd.fine_heading;
  } else {
    sint4 w = e1.get_int("width");
    sint4 h = e1.get_int("height");
    a1.x = fine_grid(*e1.sod.x) - fine_grid(w/2);
    a1.y = fine_grid(*e1.sod.y) - fine_grid(h/2);
    a2.x = a1.x + fine_grid(w);
    a2.y = a1.y + fine_grid(h);
  }

  //------------------------------

  ScalarPoint b1, b2;
  ScalarPoint h2(0,0);
  if (*e2.sod.speed) {
    sint4 w = e2.get_int("width");
    sint4 h = e2.get_int("height");
    b1.x = e2.smd.fine_pos.x - fine_grid(w/2);
    b1.y = e2.smd.fine_pos.y - fine_grid(h/2);
    b2.x = b1.x + fine_grid(w);
    b2.y = b1.y + fine_grid(h);
    h2 = e2.smd.fine_heading;
  } else {
    sint4 w = e2.get_int("width");
    sint4 h = e2.get_int("height");
    b1.x = fine_grid(*e2.sod.x) - fine_grid(w/2);
    b1.y = fine_grid(*e2.sod.y) - fine_grid(h/2);
    b2.x = b1.x + fine_grid(w);
    b2.y = b1.y + fine_grid(h);
  }

  //------------------------------

  return collision2d::collision_time_rect_rect(a1, a2, h1, b1, b2, h2, coll_type, ct);
}

//===================================================================

#if 0
static void print_path_counts()
{
  ++pcn;
  if ((pcn & 127) == 0) {
    cout << "PATH COUNTS:" << endl;
    FORS (i, PCN)
      cout << i << ": " << pc[i] << " " << (!pc[i] ? " !!!" : "") << endl;
  }
}
#endif

//===================================================================
