#ifndef Coord_H
#define Coord_H

// $Id: Coord.H 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"

#define USE_LRINT          0
#define USE_ROUNDING_MACRO 1


typedef sint4 Scalar;  // fine grid

class FixedPoint {

public:

  inline static Scalar add(const Scalar &x, const Scalar &y) { return x+y; }
  inline static Scalar sub(const Scalar &x, const Scalar &y) { return x-y; }
  inline static Scalar abs(const Scalar &x) { return ::abs(x); }
  inline static sint8  to_sint8(const Scalar &x) { return sint8(x); }
  inline static sint8  square(const Scalar &x) { return sint8(x)*sint8(x); }
  
  static sint8 slow_from_real8(const real8 &f) {

    Scalar v;
    if      (f > 0.0) v = sint4(floor(f+0.5));
    else if (f < 0.0) v = sint4(ceil(f-0.5));
    else v = 0;
    return v;

  }

  static Scalar from_real8(const real8 &f) {

    assert(fabs(f) < 2e9);

    Scalar v;
    ROUND_REAL8_TO_SINT4(f, v);
    return v;
  }

  static Scalar scale(const Scalar &x, const real8 &f) {
    return from_real8(x*f);
  }
  
  static Scalar round(const Scalar &x, sint4 bits) {

    // untested
    
    Scalar y = x;
    sint4 mask = 1 << bits;
    if      (y > 0) y += mask >> 1;
    else if (y < 0) y -= mask >> 1;
    y &= ~(mask-1);
    return y;
  }

};


class ScalarPoint {

public:
  
  Scalar x, y;

  ScalarPoint() : x(0), y(0) {}

  ScalarPoint(Scalar x_, Scalar y_) : x(x_), y(y_) {}

  bool bad() const {

    return (x == min_sint4 || abs(x) >= 500000000 ||
	    y == min_sint4 || abs(y) >= 500000000);
  }
  
  Scalar l2_norm() const {

    if (bad()) {
      write(std::cerr);
      ERR("scalar out of range");
    }

    return FixedPoint::from_real8(sqrt(real8(FixedPoint::square(x) +
					   FixedPoint::square(y))));
  }

  sint8 l2_norm_sq() const {

    if (bad()) {
      write(std::cerr);
      ERR("scalar out of range");
    }

    return FixedPoint::square(x) + FixedPoint::square(y);
  }

  Scalar max_norm() const {
    return std::max(FixedPoint::abs(x), FixedPoint::abs(y));
  }

  Scalar l2_dist(const ScalarPoint &y_) const {
    ScalarPoint d = *this;
    d.sub(y_);
    return d.l2_norm();
  }

  sint8 l2_dist_sq(const ScalarPoint &y_) const {
    ScalarPoint d = *this;
    d.sub(y_);
    return d.l2_norm_sq();
  }

  Scalar max_dist(const ScalarPoint &y_) const {
    ScalarPoint d = *this;
    d.sub(y_);
    return d.max_norm();
  }


  //  sint8 dot(const ScalarPoint &p) {
  sint8 operator*(const ScalarPoint &p) const {
    return sint8(x)*p.x + sint8(y)*p.y;
  }

  sint8 cross(const ScalarPoint &p) const {
    return sint8(x)*p.y - sint8(y)*p.x;
  }

  void add(const ScalarPoint &p) {
    x = FixedPoint::add(x, p.x);
    y = FixedPoint::add(y, p.y);
  }

  void sub(const ScalarPoint &p) {
    x = FixedPoint::sub(x, p.x);
    y = FixedPoint::sub(y, p.y);
  }  

  void scale(real8 f) {
    x = FixedPoint::scale(x, f);
    y = FixedPoint::scale(y, f);
  }

  void round(sint4 bits) {
    x = FixedPoint::round(x, bits);
    y = FixedPoint::round(y, bits);
  }

  void write(std::ostream &os = std::cout) const {
    os << "(" << x << "," << y << ")";
  }
  
};

inline bool operator==(const ScalarPoint &c1, const ScalarPoint &c2) {
  return c1.x == c2.x && c1.y == c2.y;
}
/*
inline bool operator==(const ScalarPoint c1, const ScalarPoint c2) {
  return c1.x == c2.x && c1.y == c2.y;
}
*/
inline bool operator!=(const ScalarPoint &c1, const ScalarPoint &c2) {
  return c1.x != c2.x || c1.y != c2.y;
}

REGISTER_TYPEOF(301,  Vector<ScalarPoint>::iterator);
REGISTER_TYPEOF(302,  Vector<ScalarPoint>::const_iterator);
#endif
