#ifndef LinAlg_H
#define LinAlg_H

// $Id: LinAlg.H 12785 2012-11-06 04:50:57Z mburo $

// This is an ORTS file
// (c) Keith Yerex
// (c) Tim Furtak
// licensed under the GPL

#include "Random.H"
#include <algorithm>

//===================================================================

#define UNWIELDY_MACRO(Vec, N)                                          \
  Vec(const T *B) { FORU (i, N) A[i] = B[i]; }                          \
  Vec(T n) { FORU (i, N) A[i] = n; }                                    \
                                                                        \
  const T* ptr() const { return A; }                                    \
  T* ptr() { return A; }                                                \
                                                                        \
  const T & operator [](int i) const { return A[i]; }                   \
  T & operator [](int i) { return A[i]; }                               \
                                                                        \
  inline Vec operator +(const Vec &v) const { Vec V = (*this); FORU (i, N) V[i] += v[i]; return V; } \
  inline Vec operator -(const Vec &v) const { Vec V = (*this); FORU (i, N) V[i] -= v[i]; return V; } \
  inline Vec operator *(const Vec &v) const { Vec V = (*this); FORU (i, N) V[i] *= v[i]; return V; } \
  inline Vec operator /(const Vec &v) const { Vec V = (*this); FORU (i, N) V[i] /= v[i]; return V; } \
                                                                        \
  inline Vec operator +(const T &n) const { Vec V = (*this); FORU (i, N) V[i] += n; return V; } \
  inline Vec operator -(const T &n) const { Vec V = (*this); FORU (i, N) V[i] -= n; return V; } \
  inline Vec operator *(const T &n) const { Vec V = (*this); FORU (i, N) V[i] *= n; return V; } \
  inline Vec operator /(const T &n) const { Vec V = (*this); T m = (1.0/n); FORU (i, N) V[i] *= m; return V; } \
                                                                        \
  inline void operator +=(const Vec &v) { FORU (i, N) A[i] += v[i]; }   \
  inline void operator -=(const Vec &v) { FORU (i, N) A[i] -= v[i]; }   \
  inline void operator *=(const Vec &v) { FORU (i, N) A[i] *= v[i]; }   \
  inline void operator /=(const Vec &v) { FORU (i, N) A[i] /= v[i]; }   \
                                                                        \
  inline void operator *=(const T &n) { FORU (i, N) A[i] *= n; }        \
  inline void operator /=(const T &n) { T m = (1.0/n); FORU (i, N) A[i] *= m; } \
                                                                        \
  Vec& operator=(const Vec &v) {                                        \
    FORU (i, N) A[i] = v[i];                                            \
    return *this;                                                       \
  }                                                                     \
                                                                        \
  inline T length() const {                                             \
    T sum = A[0] * A[0];                                                \
    for (int i = 1; i < N; ++i) sum += A[i]*A[i];                       \
    return sqrt(sum);                                                   \
  }                                                                     \
                                                                        \
  inline void normalize() { if (length() == 0.0) return; (*this) /= length(); } \
                                                                        \
  inline T dot(const Vec &v) const {                                    \
    T sum = A[0] * v[0];                                                \
    for (int i = 1; i < N; ++i) sum += A[i]*v[i];                       \
    return sum;                                                         \
  }                                                                     \
                                                                        \
  void print(std::ostream &os=std::cout) const {                        \
    os << "[" << A[0];                                                  \
    for (int i = 1; i < N; ++i) os << ", " << A[i];                     \
    os << "]" << std::endl;                                             \
  }                                                                     \
                                                                        \
  static Vec vrand(Random &r) {                                         \
    Vec V;                                                              \
    FORU (i, N) V[i] = 2.0*r.rand_real8()-1.0;                          \
    return V;                                                           \
  }                                                                     \
                                                                        \
  static Vec min(const Vec &a, const Vec &b) {                          \
    Vec V;                                                              \
    FORU (i, N) V[i] = std::min(a[i], b[i]);                            \
    return V;                                                           \
  }                                                                     \
                                                                        \
  static Vec max(const Vec &a, const Vec &b) {                          \
    Vec V;                                                              \
    FORU (i, N) V[i] = std::max(a[i], b[i]);                            \
    return V;                                                           \
  }

//===================================================================

template <class T>
class Vec2
{
public:

  Vec2() : x(0), y(0) {}
  Vec2(T X, T Y) : x(X), y(Y) {}

  union {
    T A[2];
    struct {
      T x, y;
    };
  };

  UNWIELDY_MACRO(Vec2, 2);

  Vec2 operator -() const { return Vec2(-x,-y); }

  Vec2 project(const Vec2 &v) {
    if (v.x == (T)0 && v.y == (T)0) return v;
    real8 d = (real8)dot(v) / v.dot(v);
    return v * d;
  }

  Vec2 rotate(T a) {
    T c = cos(a), s = sin(a);
    return Vec2(x*c - y*s, x*s + y*c);
  }
};

//===================================================================

template <class T>
class Vec3
{
public:

  Vec3() : x(0), y(0), z(0) {}
  Vec3(T X, T Y, T Z) : x(X), y(Y), z(Z) {}

  union {
    T A[3];
    struct {
      T x, y, z;
    };
    struct {
      T r, g, b;
    };
  };

  UNWIELDY_MACRO(Vec3, 3);

  Vec3 operator -() const { return Vec3(-x,-y,-z); }

  Vec3 cross(const Vec3 &v) const {
    return Vec3(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);
  }

  Vec3 rotate(const Vec3 &axis, T a) {
    T c = cos(a), s = sin(a);
    return (*this)*c + cross(axis)*s + axis*dot(axis)*(1-c);
  }

  // false iff error
  bool write(FILE *f) const { return fwrite(A, sizeof(*this), 1, f) == 1; }
  bool read (FILE *f)       { return fread (A, sizeof(*this), 1, f) == 1; }
};

//===================================================================

template <class T>
class Vec4
{
public:

  Vec4() : x(0), y(0), z(0), w(1) {}
  Vec4(T X, T Y, T Z, T W) : x(X), y(Y), z(Z), w(W) {}

  operator Vec3<T>() const { return Vec3<T>(x,y,z); }

  union {
    T A[4];
    struct {
      T x, y, z, w;
    };
    struct {
      T r, g, b, a;
    };
  };
  
  UNWIELDY_MACRO(Vec4, 4);

  Vec4 project() const {
    return *this / w;
  }

  Vec4 cross(const Vec4 &v) const {
    return Vec4(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x, 0);
  }

  // false iff error
  bool write(FILE *f) const { return fwrite(A, sizeof(*this), 1, f) == 1; }
  bool read (FILE *f)       { return fread (A, sizeof(*this), 1, f) == 1; }
};

//===================================================================

#undef UNWIELDY_MACRO

//===================================================================

template <class T> class Mat4
{
public:

  Mat4() {}

  Mat4(Vec4<T> &r0, Vec4<T> &r1, Vec4<T> &r2, Vec4<T> &r3) {
    A[0] = r0; A[1] = r1; A[2] = r2; A[3] = r3;
  }
  
  const T* ptr() const { return &A[0][0]; }
  T* ptr() { return &A[0][0]; }
  // operator const T *() const { return &(A[0][0]); }
  // operator T *() { return &(A[0][0]); }

  Vec4<T> & operator [](int i) { return A[i]; }
  // const Vec4<T> & operator [](int i) const { return A[i]; }

  Vec4<T> col(sint4 i) const { return Vec4<T>(A[0][i], A[1][i], A[2][i], A[3][i]); }

  Mat4 operator *(const Mat4 &c) const;
  Vec4<T> operator *(const Vec4<T> &v) const;
  Vec3<T> operator *(const Vec3<T> &v) const;
  Mat4 inv_quick() const; // inverse of rotation/translation matrix only
  void invert();
  Mat4 inverse() const;
  Mat4 trans() const;
  void identity();

  void print(std::ostream &os=std::cout) const { FORS (i, 4) A[i].print(os); }

  // false iff error
  bool write(FILE *f) const { return fwrite(A, sizeof(A), 1, f) == 1; }
  bool read(FILE *f) { return fread(A, sizeof(A), 1, f) == 1; }

private:

  Vec4<T> A[4];
};


// implementation

template <typename T> Mat4<T> Mat4<T>::operator *(const Mat4<T> &c) const
{
  Mat4<T> r;

  FORS (i, 4)
    FORS (j, 4) {
      r[i][j] = 0;
      FORS (k, 4) r[i][j] += A[i][k] * c.A[k][j];
    }
  return r;
};

template <typename T> Vec4<T> Mat4<T>::operator *(const Vec4<T> &v) const
{
  Vec4<T> r;

  FORS (i, 4) {
    r[i] = 0;
    FORS (j, 4) r[i] += A[i][j]*v[j];
  }
  return r;
};

template <typename T> Vec3<T> Mat4<T>::operator *(const Vec3<T> &v) const
{
  Vec3<T> r;
  FORS (i, 3) {
    r[i] = 0;
    FORS (j, 3) r[i] += A[i][j]*v[j];
    r[i] += A[i][3];
  }
  return r;
};

template <typename T> Mat4<T> Mat4<T>::trans() const
{
  Mat4 r;
  r.A[0][0] = A[0][0];
  r.A[0][1] = A[1][0];
  r.A[0][2] = A[2][0];
  r.A[0][3] = A[3][0];

  r.A[1][0] = A[0][1];
  r.A[1][1] = A[1][1];
  r.A[1][2] = A[2][1];
  r.A[1][3] = A[3][1];

  r.A[2][0] = A[0][2];
  r.A[2][1] = A[1][2];
  r.A[2][2] = A[2][2];
  r.A[2][3] = A[3][2];

  r.A[3][0] = A[0][3];
  r.A[3][1] = A[1][3];
  r.A[3][2] = A[2][3];
  r.A[3][3] = A[3][3];

  return r;
};

template <typename T> void Mat4<T>::invert()
{
  T data[16];
  T inv[16];
  T smallest = std::numeric_limits<T>::max();

  FORU (i, 4) data[ 0+i] = A[0][i];
  FORU (i, 4) data[ 4+i] = A[1][i];
  FORU (i, 4) data[ 8+i] = A[2][i];
  FORU (i, 4) data[12+i] = A[3][i];

  memset(inv, 0, sizeof(inv));

  inv[0] = 1;
  inv[5] = 1;
  inv[10] = 1;
  inv[15] = 1;

  FORS (j, 4) {
    T largest = 0;
    int largei = 0;
    for (int i = j; i < 4; i++){
      if (fabs(data[i+4*j]) > largest){
	largei = i;
	largest = fabs(data[i+4*j]);
      }
    }
    // pivot
    FORS (k, 4) {

      std::swap(data[j+4*k], data[largei+4*k]);
      std::swap(inv[j+4*k], inv[largei+4*k]);

#if 0      
      T temp = data[j+4*k];
      data[j+4*k] = data[largei+4*k];
      data[largei+4*k] = temp;

      temp = inv[j+4*k];
      inv[j+4*k] = inv[largei+4*k];
      inv[largei+4*k] = temp;
#endif
      
    }
    if (fabs(data[j*4+j]) < smallest)
      smallest = fabs(data[j*4+j]);
    
    T ival = 1.0/data[j+4*j];
    FORS (k, 4) {
      data[j+4*k] *= ival;
      inv[j+4*k] *= ival;
    }
    FORS (i, 4) {
      if (i != j){
	T scale = data[i+4*j];
	FORS (k, 4) {
	  data[i+4*k] = data[i+4*k] - scale*data[j+4*k];
	  inv[i+4*k] = inv[i+4*k] - scale*inv[j+4*k];
	}
      }
    }
  }

  FORU (i, 4) A[0][i] = inv[ 0+i];
  FORU (i, 4) A[1][i] = inv[ 4+i];
  FORU (i, 4) A[2][i] = inv[ 8+i];
  FORU (i, 4) A[3][i] = inv[12+i];
}

template <typename T> Mat4<T> Mat4<T>::inverse() const
{
  Mat4<T> r(*this);
  r.invert();
  return r;
}


template <typename T> Mat4<T> Mat4<T>::inv_quick() const
{
  Mat4 r;
  r.A[0][0] = A[0][0];
  r.A[0][1] = A[1][0];
  r.A[0][2] = A[2][0];
  r.A[0][3] = -A[0][3];

  r.A[1][0] = A[0][1];
  r.A[1][1] = A[1][1];
  r.A[1][2] = A[2][1];
  r.A[1][3] = -A[1][3];

  r.A[2][0] = A[0][2];
  r.A[2][1] = A[1][2];
  r.A[2][2] = A[2][2];
  r.A[2][3] = -A[2][3];

  r.A[3][0] = 0;
  r.A[3][1] = 0;
  r.A[3][2] = 0;
  r.A[3][3] = 1;

  return r;
};

template <typename T> void Mat4<T>::identity()
{
  A[0][0] = 1; A[0][1] = 0; A[0][2] = 0; A[0][3] = 0;
  A[1][0] = 0; A[1][1] = 1; A[1][2] = 0; A[1][3] = 0;
  A[2][0] = 0; A[2][1] = 0; A[2][2] = 1; A[2][3] = 0;
  A[3][0] = 0; A[3][1] = 0; A[3][2] = 0; A[3][3] = 1;
};

#endif
