/** @file Script.C
    @see Script.H.

    $Id: Script.C 6867 2008-04-05 18:59:22Z furtak $
    $Source$
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Script.H"
#include "ScriptGame.H"
#include "Action.H"
#include "GameObj.H"
#include "GameObjFactory.H"
#include "PlayerInfo.H"
#include "Game.H"
#include "Random.H"
#include "Blueprints.H"
#include "ScriptParser.H"

using namespace std;

REGISTER_TYPEOF(44, list<uint4*>::iterator);
REGISTER_TYPEOF(45, list<uint4*>::const_iterator);

Game *FunctionNode::game = 0;

#ifdef SCRIPT_DEBUG
stack<ScriptParserContext> StmtList::context;
#endif

//=============================================================================

stype ActionScript::Act(const SActParam &p)
{
  if (!root) return stype(); // the action is blank -- can probably ignore

  if (call_depth > 0) {
    Vector<stype> cur_obj_stack;
    Vector<stype> cur_param_stack;
    FORS(i, vlist.num_objs)  { cur_obj_stack.push_back(vlist.objs[i]); }
    FORS(i, vlist.num_params){ cur_param_stack.push_back(vlist.param[i]); }
    FORS(i, vlist.num_local) { cur_param_stack.push_back(vlist.local[i]); }
    obj_stack.push(cur_obj_stack);
    param_stack.push(cur_param_stack);
  }

  if (vlist.num_pobjs+1 > (sint4)(p.obj.size()))
    SCRIPT_ERR2("not enough parameters1", name.c_str());
  if (vlist.num_params > (sint4)(p.val.size()))
    SCRIPT_ERR2("not enough parameters2", name.c_str());

  FORS(i, vlist.num_pobjs+1) vlist.objs[i] = stype(p.obj[i]);
  FORS(i, vlist.num_params) vlist.param[i] = p.val[i];

  ++call_depth;
  stype r = root->eval();
  --call_depth;

  if (call_depth > 0) {
    Vector<stype> old_obj_stack = obj_stack.top();
    Vector<stype> old_param_stack = param_stack.top();
    FORS(i, vlist.num_objs)  { vlist.objs[i] = old_obj_stack[i]; }
    FORS(i, vlist.num_params){ vlist.param[i] = old_param_stack[i]; }
    FORS(i, vlist.num_local) { vlist.local[i] = old_param_stack[i+vlist.num_params]; }
    obj_stack.pop();
    param_stack.pop();
  }

  return r;
}

//-----------------------------------------------------------------------------

ActionScript::~ActionScript()
{
  if (root) delete root;
}

//=============================================================================

stype StmtList::eval()
{
  stype RVal;
  uint4 size = stmts.size();

  FOR (PC, size) {

#ifdef SCRIPT_DEBUG
    context.push(stmt_context[PC]);
#endif

#if 0
    RVal = stmts[PC]->eval();
#else
    try {
      RVal = stmts[PC]->eval();
    }
    catch (std::runtime_error &e) {
      string msg = e.what();
      msg = "\n" + msg; // new context
      for (sint4 i=PC; i >= 0; --i) {
	ScriptParserContext &context = stmt_context[i];
	msg =
	  context.file + "::" + (boost::format("%d") % context.line_num).str()
	  + " " + context.line + "\n" + msg;
      }
      throw ScriptExecException(msg);
    }
#endif
    
#ifdef SCRIPT_DEBUG
    context.pop();
#endif
  }
  return RVal;
}

//-----------------------------------------------------------------------------

#ifdef SCRIPT_DEBUG
void StmtList::dump_context()
{
  stack<ScriptParserContext> temp;
  stack<ScriptParserContext> foo = context;
  while (!context.empty()) {
    temp.push(context.top());
    context.pop();
  }
  while (!temp.empty()) {
    errstr << endl << temp.top().file << "::" << temp.top().line_num << temp.top().line;
    temp.pop();
  }
  context = foo;
  errstr << endl;
}
#endif

//=============================================================================

AssignStmt::AssignStmt(VariableNode *v, ScriptNode *e) : var(v), expression(e)
{
  SCRIPT_ASSERT(var && expression && "ptr 0");
}

//-----------------------------------------------------------------------------

stype AssignStmt::eval()
{
  stype v = expression->eval();
  stype *a = var->addr();
  SCRIPT_ASSERT(a && "var ptr 0");
  // cout << "t=" << v.type << " v=" << v.val.i << endl;
  if (a->type == stype::OBJ) a->val.o = v.to_obj();
  else *a = v;
  return v;
}

//=============================================================================

stype GlobalObj::eval()
{
  ScriptObj *s = pi->global_obj(name);
  ScriptObj::check_ptr(s);
  if (s) {
    obj = stype(s);
  } else {
    SCRIPT_ERR2("Could not find global object ", name.c_str());
    obj = stype(0);
  }
  return stype();
}

//=============================================================================

stype SharedObj::eval()
{
  ScriptObj *s = pi->shared_obj(name);
  ScriptObj::check_ptr(s);
  if (s) {
    obj = stype(s);
  } else {
    SCRIPT_ERR2("Could not find shared object ", name.c_str());
    obj = stype(0);
  }
  return stype();
}

//=============================================================================

GOB_Component::GOB_Component(stype *g, Vector<string> &comp)
{
  obj = stype(0);
  gob = g;
  comp_names = comp;
  first = 0;
}

stype GOB_Component::eval()
{
  if (first) first->eval();
  ScriptObj *sob = gob->to_obj();

  ScriptObj::check_ptr(sob);
  FORALL (comp_names, i) {
    if (*i == "parent") {
      sob = sob->parent;
      continue;
    }
    ScriptObj *comp = sob->component(*i);
    if (comp == 0) {
      cerr << "object type = " << sob->bp_name() << endl;
      cerr << "parent type = " << sob->parent->bp_name() << endl;
      SCRIPT_ERR2("Object does not have component:", (*i).c_str());
    }
    ScriptObj::check_ptr(comp);
    sob = comp;
  }
  obj = stype(sob);
  return stype();
}

//=============================================================================

stype GOB_AttrNode::eval()
{
  if (par) par->eval(); // dereference any components
  ScriptObj *obj = gob->to_obj();
  if (!obj) SCRIPT_ERR2("couldn't find component(s) - attr =", attr.c_str());
  ScriptObj::check_ptr(obj);

  stype *v = obj->get_val_ptr(attr);
  if (v) return *v;

  if (attr == "parent") return stype(obj->parent);

  ScriptObj *comp = obj->component(attr);
  if (comp != 0) {
    ScriptObj::check_ptr(comp);
    return stype(comp); // a pointer to the component
  }

  cout << "BP = " << obj->bp_name() << endl;
  cout << "this = " << reinterpret_cast<void*>(obj) << endl;
  SCRIPT_ERR2("no such attribute:", attr.c_str());
}

stype* GOB_AttrNode::addr()
{
  if (par) par->eval();
  stype *ptr = gob->to_obj()->get_val_ptr(attr);
  SCRIPT_ASSERT2(ptr && "Attribute ptr 0", attr);
  return ptr;
}

//=============================================================================

stype GOB_AddrNode::eval()
{
  if (par) par->eval();
  //  cout << "gob_addrnode eval = " << (void*)*gob << endl;
  ScriptObj::check_ptr(gob->to_obj()); 
  return stype(*gob);
}

//=============================================================================

BP_AttrNode::BP_AttrNode(const Vector<std::string> &a, GameObjFactory *f)
{
  SCRIPT_ASSERT(a.size() >= 2 && "not enough parameters");
  attrs = a;
  gof = f;
  ptr = 0;
}

void BP_AttrNode::get_ptr()
{
  string bp_name = attrs[0];
  GameObjBlueprint *bp = gof->get_blueprint(bp_name);
  if (!bp) SCRIPT_ERR2("searching for BP value ... not registered:", bp_name.c_str());
  
  for (uint4 i = 1; i < attrs.size()-1; ++i) {
    const string &comp = attrs[i];
    FIND (bp->name2comp, it2, comp);
    if (it2 == bp->name2comp.end()) SCRIPT_ERR2("unknown component:", comp.c_str());
    bp = bp->component[it2->second].bp;
    if (bp == 0) SCRIPT_ERR2("non-blueprint component:", comp.c_str());
  }
  
  string attr = attrs[attrs.size()-1];

  sint4 i = bp->attr_index(attr);
  if (i < 0) SCRIPT_ERR2("unknown attribute", attr.c_str());
  ptr = &bp->attr_vals[i];
}

stype BP_AttrNode::eval()
{
  if (!ptr) get_ptr();
  return *ptr;
}

stype* BP_AttrNode::addr()
{
  if (!ptr) get_ptr();
  return ptr;
}

//=============================================================================

StringNode::StringNode(const string &s)
{
  string temp = s.substr(1, s.length()-2);
  char *t = new char[temp.length()+1];
  strcpy(t, temp.c_str());
  addr = t;
}

//=============================================================================

stype ArrayNode::eval()
{
  sint4 i = index->eval().to_int();
  ScriptObj *o = ptr->eval().to_obj();
  ScriptObj::check_ptr(o);
  ObjAction *a = o->action("op");
  SCRIPT_ASSERT2(a && "undefined action op in ", o->bp_name());
  SActParam p;
  p.obj.push_back(o);
  p.val.push_back(stype(i));
  return a->Act(p); // the result of the object's array operator
}
 
// fixme: have an assignment operator instead of using the address

stype* ArrayNode::addr()
{
  sint4 i = index->eval().to_int();
  ScriptObj *o = ptr->eval().to_obj();
  ScriptObj::check_ptr(o);
  ObjAction *a = o->action("op_addr");
  SCRIPT_ASSERT2(a && "undefined action op_addr in ", o->bp_name());
  SActParam p;
  p.obj.push_back(o);
  p.val.push_back(stype(i));
  return a->Act(p).to_addr(); // address of element i
}

//=============================================================================
//=== functions ===============================================================
//=============================================================================

/** Parse the parameters for a function call (objects, integers). */

ScriptNode* FunctionCall::parse_core(ScriptFunctionDef &def, ScriptParser &p)
{
  ScriptNode *f = this;
  ScriptNode *n;

  if (p.getToken() != "(") p.parseError(": " + def.name + " expecting '('");

  if (def.num_obj_ids > 0) {

    FORS (i, def.num_obj_ids) {
      VariableNode *temp = p.getVariable(true); // try to read as object
      objs.push_back(temp);

      if (i < def.num_obj_ids -1) {
        if (p.getToken() != ",") p.parseError(": " + def.name + " expecting ','");
      } else if (def.num_params != 0) {
        if (p.getToken() != ";") p.parseError(": " + def.name + " expecting ';'");
      }
    }
  } else if (def.num_obj_ids < 0) { // variable number of objects
    FOREVER {
      string t = p.getToken();
      if (t == ";") break;
      p.putbackToken(t);

      VariableNode *temp = p.getVariable();
      objs.push_back(temp);

      t = p.getToken();
      if (t == ",") continue;
      if (t == ";" && def.num_params != 0) break;
      p.parseError(": " + def.name + " expecting ',' or ';'");
    }
  }

  if (def.num_params == 0) {
    if (p.getToken() != ")") p.parseError(": " + def.name + " expecting ')'");
  } else if (def.num_params > 0) {
    FORS(i, def.num_params) {
      if (i < def.num_params -1) {
        n = p.parseExp(",");
      } else {
        n = p.parseExp(")");
      }
      params.push_back(n);
    }
  } else { // variable number of arguments
    set<string> end_markers;
    string end_marker;
    end_markers.insert(",");
    end_markers.insert(")");
    FOREVER {
      n = p.parseExp(end_markers, end_marker);
      params.push_back(n);
      if (end_marker == ")") break;
    }
  }

  return f;
}

//===================================================================

stype StaticFunctionCall::eval()
{
  SActParam p;

  FORALL (objs, i) p.obj.push_back((*i)->eval().to_obj());
  FORALL (params, i) p.val.push_back((*i)->eval());

  return (*func)(p);
}

//=============================================================================

stype LostSight::eval()
{
  sint4 p = player->eval().to_int();
  GameObj *g = dynamic_cast<GameObj*>(gob->eval().to_obj());
  SCRIPT_ASSERT(g && "GameObj ptr 0");
  SCRIPT_ASSERT(p >= 0 && p < g->sod.last_lost.size() && "index out of range");
  return stype(g->sod.last_lost[p]);
}

ScriptNode* LostSight::parse(ScriptParser &p)
{
  if (p.getToken() != "(") p.parseError("expecting '('");
  gob = p.getVariable();
  if (p.getToken() != ",") p.parseError("expecting ','");
  player = p.parseExp(")");

  return this;
}

//=============================================================================

stype RandomNode::eval()
{
  sint4 a = min->eval().to_int();
  sint4 b = max->eval().to_int();

  if (a == b) return stype(a);
  if (a < b)  return stype((game->rand.rand_uint4() % (b-a+1)) + a);
  return stype((game->rand.rand_uint4() % (a-b+1)) + b);
}

ScriptNode* RandomNode::parse(ScriptParser &p)
{
  if (p.getToken() != "(") p.parseError("expecting '('");
  min = p.parseExp(",");
  max = p.parseExp(")");
  if (!min || !max) p.parseError("expecting ranges");
  return this;
}

//=============================================================================

stype NumIntersect::eval()
{
  GameObj *g = dynamic_cast<GameObj*>(gob->eval().to_obj()); // fixme: GameObj->ScriptObj
  SCRIPT_ASSERT(g && "no gameobj");
  return stype(game->num_intersect(*g));
}

ScriptNode* NumIntersect::parse(ScriptParser &p)
{
  if (p.getToken() != "(") p.parseError("expecting '('");
  gob = p.getVariable();
  if (p.getToken() != ")") p.parseError("expecting ')'");

  return this;
}

//=============================================================================

stype BP_Code::eval()
{
  /// @bug Won't work if players are created differently.
  return stype(game->get_player_info(0).gof.bp_code(bp_name));
}

ScriptNode* BP_Code::parse(ScriptParser &p) {
  ScriptNode *f = this;
  if (p.getToken() != "(") p.parseError("expecting '('");
  bp_name = p.getToken();
  if (p.getToken() != ")") p.parseError("expecting ')'");

  return f;
}

//=============================================================================
//-----------------------------------------------------------------------------
//=============================================================================

ForAllObj::ForAllObj(stype *g, Vector<ScriptObj::Ptr> *_objs)
{
  gob = g;
  objs = _objs;
  expression = 0;
  ScriptObj::check_ptr(gob->to_obj());
  PC = 0;
}

//-----------------------------------------------------------------------------

stype ForAllObj::eval()
{
  stype res;
  PC = 0;
  if (expression) {
    FORALL(*objs, i) {
      if (PC) break;
      *gob = stype(&*(*i)); // fixme: is that OK?
      res = expression->eval();
    }
  }
  return res;
}

//=============================================================================

ForAllType::ForAllType(stype *g, Vector<ScriptObj::Ptr> *_objs, set<string> &_bp_types)
{
  gob = g;
  objs = _objs;
  expression = 0;
  bp_types = _bp_types;
  ScriptObj::check_ptr(gob->to_obj());
  PC = 0;
}

//-----------------------------------------------------------------------------

stype ForAllType::eval()
{
  PC = 0;
  stype res;
  if (expression) {

  //   FORALL(bp_types, it){
//      cout << "Name = " << *it << endl;
//     }

    FORALL(*objs, i) {
      if (PC) break;
      ScriptObj *t = *i;
      *gob = stype(t);
      // cout << "BP name = " << t->bp_name() << "\n" ;
      if (bp_types.find(t->bp_name()) != bp_types.end())
	res = expression->eval();
    }
  }
  return res;
}

//=============================================================================

stype ForAllIntersect::eval()
{
  if (!expression) return stype();

  GameObj *q = dynamic_cast<GameObj*>(area->eval().to_obj()); // fixme: GameObj->ScriptObj
  SCRIPT_ASSERT(q && "no gameobj");
  GameObj &a = *q;
  sint4 x0, x1, y0, y1;

  if (*a.sod.shape == Object::CIRCLE) {
    sint4 x = game->fine_grid(*a.sod.x);
    sint4 y = game->fine_grid(*a.sod.y);
    sint4 r = game->fine_grid(*a.sod.radius);
    x0 = x - r;
    x1 = x + r;
    y0 = y - r;
    y1 = y + r;
  } else if (*a.sod.shape == Object::RECTANGLE ||
             *a.sod.shape == Object::LINE) {
    a.update_points();
    x0 = game->fine_grid(*a.sod.x1);
    y0 = game->fine_grid(*a.sod.y1);
    x1 = game->fine_grid(*a.sod.x2);
    y1 = game->fine_grid(*a.sod.y2);
  } else {
    ERR("BAD SHAPE");
    return stype(-1);
  }

  sint4 sec_wh = game->get_sec_wh();
  sint4 sec_nx = game->get_sec_nx();
  sint4 sec_ny = game->get_sec_ny();

  x0 /= sec_wh;
  x1 /= sec_wh;
  y0 /= sec_wh;
  y1 /= sec_wh;

  if (x0 < 0) x0 = 0; else if (x1 >= sec_nx) x1 = sec_nx-1;
  if (y0 < 0) y0 = 0; else if (y1 >= sec_ny) y1 = sec_ny-1;

  set<GameObj*> objects;

  for(sint4 y = y0; y <= y1; ++y) {
    sint4 s = y*sec_nx + x0;
    for(sint4 x = x0; x <= x1; ++x) {
      const MotionSectors &ms = game->get_motion_secs();
      // the first 'size' objects are boundaries (the (mostly) static part of the array)
      //      for (sint4 i = ms->sizes[s]; i < ms->secs[s].size(); ++i) {

      const MotionSectorInfo &msi = ms.secs[s];
      FORT (i, msi.size()) objects.insert(msi[i]);
      ++s;
    }
  }

  PC = 0;
  stype res;
  FORALL (objects, it) {

    if (PC) break;

    if (*a.sod.zcat == Object::ZCAT_UNDEF || *(**it).sod.zcat == Object::ZCAT_UNDEF ||
        *a.sod.zcat == *(**it).sod.zcat) {
      if (*it == &a) continue;
      if (Game::distance(a, **it) <= 0) {
        *gob = stype(*it);
        res = expression->eval();
      }
    }
  }

  return res;
}

//=============================================================================
//=============================================================================

QueueAction::QueueAction(const string &_action, Vector<VariableNode*> &o, Vector<ScriptNode*> &p,
                         ScriptNode *_delay, Game *_game, list<uint4*> &_pcs)
{
  action_name = _action;
  objs = o;
  params = p;
  delay = _delay;
  game = _game;
  // was: if (game) ...
  is_server = (game->get_client_player() == -1);
  sa = &game->get_simple_actions();
  FORALL (_pcs, i) pcs.push_back(*i);
}

//-----------------------------------------------------------------------------

QueueAction::~QueueAction()
{
  FORALL (objs, it) delete *it;
  FORALL (params, it2) delete *it2;
  if (delay) { delete delay; delay = 0; }
}

//-----------------------------------------------------------------------------

stype QueueAction::eval()
{
  Vector<ScriptObj*> obj;

  //  cout << "QueueAction action name = " << action_name << endl;
  //  cout << "objs[0] VariableNode* = " << (void*)(objs[0]) << endl;

  FORALL (objs, it) {
    ScriptObj *temp = (*it)->eval().to_obj();
    //    cout << "temp = " << (void*) temp << endl;
    //    cout << "tp = " << (void*)temp->parent << endl;
    SCRIPT_ASSERT(temp && "ptr 0");
    ScriptObj::check_ptr(temp);
    obj.push_back(temp);

  }

  //  cout << "QueueAction obj[0] = " << (void*)obj[0] << endl;
  //  cout << "en = " << obj[0]->is_enabled() << endl;
  //  cout << "par = " << (void*)obj[0]->parent << endl;

  // no actions from dead objects (not marked "dead" until 1 tick after they die)

  sint4 wait = delay ? delay->eval().to_int() : -1;

  if (obj[0]->parent->is_dead() && wait >= 0) return stype();
  //  if (is_server && !obj[0]->is_enabled()) return stype();

  uint4 zero = game->tick() + wait;
  uint4 key = game->rand.rand_uint4();

  SimpleActions::saction *q = new SimpleActions::saction(zero, key);
  q->active = true;
  q->name = action_name;

  FORALL (params, it) q->p.val.push_back((*it)->eval());
  FORALL (obj, it)    q->p.obj.push_back(*it);

  ObjAction *a = obj[0]->action(action_name);

  if (a == 0) {
    //      cout << "ACTION NOT FOUND " << action_name << endl;
    delete q;
    return stype();
  }
  q->a = a;
  q->obj_act_id = obj[0]->action_id(action_name); // -1 if it's not found
  
  GameObj *gob = obj[0]->get_GameObj();
  if (gob != 0) {
    SimpleActions::saction *p = gob->last_command_map[action_name];
    if (p != 0) {
      if (!p->queue_ref) delete p;
      else               p->obj_ref = false;
      gob->last_command_map[action_name] = 0;
    }
  }

  q->queue_ref = true;

  // do the setup here so that the action can be placed as the obj's last command
  if (wait < 0) {
    {
      Vector<uint4> cur_pc_stack;
      FORALL(pcs, i) cur_pc_stack.push_back(**i);
      pc_stack.push(cur_pc_stack);
    }
    q->obj_ref = false;
    stype r = sa->spawn_now(q);
    {
      Vector<uint4> old_pc_stack = pc_stack.top();
      sint4 j = 0;
      FORALL(pcs, i) **i = old_pc_stack[j++];
      pc_stack.pop();
    }
    return stype(r);
  }

  if (gob != 0) {
    gob->last_command_map[action_name] = q;
    q->obj_ref = true;
  } else {
    q->obj_ref = false;
  }
  sa->spawn(q);
  return stype();
}

//=============================================================================

ReturnStmt::ReturnStmt(ScriptNode *v, const list<uint4*> &pcs_) : val(v)
{
  FORALL(pcs_, i) pcs.push_back(*i);
}

//-----------------------------------------------------------------------------

stype ReturnStmt::eval()
{
  FORALL (pcs, i) **i = uint4(-2);
  return val->eval();
}

//-----------------------------------------------------------------------------

stype BreakStmt::eval()
{
  *pc = uint4(-2); // fixme: this only works because pc is unsigned
  return stype();
}

//-----------------------------------------------------------------------------

stype WhileStmt::eval()
{
  stype res;
  PC = 0;
  if (right_child) {
    while (!!(res = left_child->eval()) && PC == 0) {
      res = right_child->eval();
      if (PC != 0) break;
    }
  } else {
    while (!!(res = left_child->eval()) && PC == 0) {}
  }
  return res;
}

//-----------------------------------------------------------------------------

stype ForLoop::eval()
{
  stype res;
  PC = 0;
  if (init) init->eval();

  if (cond) {
    while (!!cond->eval()) {
      if (loop)
        res = loop->eval();
      if (PC != 0) break;
      if (inc) inc->eval();
    }
  } else {
    FOREVER {
      if (loop)
        res = loop->eval();
      if (PC != 0) break;
      if (inc) inc->eval();
    }
  }
  return res;
}

//=============================================================================
