#ifndef ScriptObj_H
#define ScriptObj_H

/** @file ScriptObj.H
    The objects that populate the world.

    $Id: ScriptObj.H 6853 2008-03-26 08:19:42Z furtak $
    $Source$
*/

// This is an ORTS file
// (c) Tim Furtak
// (c) Michael Buro
// licensed under the GPL

#include "Global.H"
#include "GameConst.H"
#include "ServerObjData.H"
#include "IdFactory.H"
#include "Attr.H"
#include "SaveLoad.H"

class PlayerInfo;
class ScriptObj;
class ObjAction;
class GameObj;
class ByteStream;

typedef uint4 ObjId;

// keep outside of class to use with typeof (windows)
typedef std::map<ObjId,ObjId> IdMap;
typedef std::map<ObjId,ScriptObj*> IdObjMap;
typedef std::map<ScriptObj*,ObjId> ObjIdMap;
typedef std::map<std::string, ScriptObj*> NameObjMap;
typedef std::map<std::string, stype*> NameVarMap;

//===================================================================

/** A generic container class that defines script objects. */

class ScriptObj : public SaveLoadAbs
{
public:
  class BasePtr;
  class WPtr;
  class Ptr;

  ScriptObj *parent;   ///< The object that this is a node of (can be self referential).
  PlayerInfo *creator;

  ScriptObj();
  virtual ~ScriptObj();

  void save(std::ostream &os) const;
  void load(std::istream &is);
  
  static void get_ids(Vector<ScriptObj*> &objs, Vector<ObjId> &ids);
  static void set_ids(const Vector<ScriptObj*> &objs, const Vector<ObjId> &ids);

  static std::set<ScriptObj*> ptr_list;
  static void check_ptr(ScriptObj*);
  static void collect_dead(std::set<ScriptObj*> &dead);
  static void collect_destroyed();
  static ObjId get_obj_id(ScriptObj *s);

  void add_ref() { ++num_refs; /* SCRIPT_ASSERT(num_refs <= 10000); WAS REACHED IN ACTUAL GAME WHEN A HUGE ARMY WAS CREATED */ } // not a hard assert - only for bug detection
  void del_ref() { --num_refs; SCRIPT_ASSERT(num_refs >= 0); }
  virtual sint4 get_refs() const { return num_refs; }
  virtual void enable() { enabled = true; }
  virtual void disable() { enabled = false; }
  virtual bool is_enabled() { return enabled; }

  void kill();
  bool is_dead() const { return dead; }

  virtual std::string bp_name() const { return ""; }
  virtual uint4 code() const { return 0; }

  virtual void attr_names(Vector<std::string> &) const {}
  virtual bool has_attr    (const std::string &) const;
  virtual bool attr_changed(const std::string &) const { return false; }

  // get value or pointer to value by name
  virtual stype get_val     (const std::string &) const;
  virtual stype get_val_prev(const std::string &) const { return stype(); }
  virtual stype* get_val_ptr(const std::string &)       { return 0; }
  virtual const stype* get_val_ptr(const std::string &) const { return 0; }
  virtual void  set_val(const std::string &name, stype val, bool warn=true);

  // for integers
  virtual sint4 get_int     (const std::string &) const;
  virtual sint4 get_int_prev(const std::string &) const;
  virtual sintptr* get_int_ptr(const std::string &);
  virtual const sintptr* get_int_ptr(const std::string &) const;
  virtual void  set_int(const std::string &name, sint4 val, bool warn=true);

  virtual void gen_diff(bool, bool, std::string &, PlayerInfo *) const {}
  virtual void gen_comp_diff(bool first, bool allied, std::string &s, PlayerInfo *pi) const
  { gen_diff(first, allied, s, pi); }
  virtual bool apply_diff(ByteStream &, PlayerInfo *) { return false; }
  virtual bool apply_comp_diff(bool, ByteStream &s, bool /*first*/, PlayerInfo *pi)
  { return apply_diff(s, pi); }

  virtual void copy_to_prev() {}
  virtual void reset_frame_data() {}

  virtual void action_names(Vector<std::string> &) const {}

  virtual uint4 set_action(const std::string &, const Vector<sint4> &, sint4 =1) {
    ERR("set_action called for non-game object (ScriptObj)");
    return 0;
  }

  virtual uint4 set_action(sint4, const Vector<sint4> &, sint4 =1) {
    ERR("set_action called for non-game object (ScriptObj)");    
    return 0;
  }

  virtual void clear_action() {
    ERR("clear_action called for non-game object (ScriptObj)");
  }
  
  virtual std::string write_action(ObjId) const { return ""; }

  virtual ObjAction* action(const std::string &)   const { return 0; }
  virtual ObjAction* action(sint4)                 const { return 0; }
  virtual sint4 action_id(const std::string &)     const { return -1; }
  virtual bool action_enabled(sint4)               const { return false; }
  virtual bool action_enabled(const std::string &) const { return false; }
  virtual bool action_hidden(sint4)                const { return false; }
  virtual bool action_hidden(const std::string &)  const { return false; }

  virtual void component_names(Vector<std::string> &) const {}
  virtual ScriptObj* component(const std::string &) const { return 0; }
  virtual ScriptObj* action_component(sint4)        const { return 0; }
  
  virtual void construct() {}
  virtual void destruct() {}
  virtual GameObj* get_GameObj() { return 0; }

  friend class GameObjFactory;
  friend class ScriptParser;

  ObjId uid() const { return id; }

protected:

  bool dead;           ///< State of this object, regardless of any parent's state.
  bool enabled;

  sint4 num_refs;
  std::set<Ptr*> hard_refs;
  std::set<WPtr*> soft_refs;

private:

  ObjId id;
  static IdFactory idf;
  static IdObjMap id2obj_map;
  static ObjIdMap obj2id_map;

  static Vector<ScriptObj*> dead_objs;         ///< Dead (in the script sense) but not deleted.
                                               ///< Garbage code relies on Sequence container.
  static std::list<ScriptObj*> destroyed_objs; ///< Dead objects whose destructor has already
                                               ///< been called, now waiting to be deleted.
};

//-------------------------------------------------------------------

/** One of those horrible 'smart pointers'.
    A necessary evil to maintain reasonable saving ability.
    Lets ScriptObjs be referenced by id. */

class ScriptObj::BasePtr
{
public:
  inline ScriptObj* operator->() const { return p; }
  inline ScriptObj& operator*() const { if (!p) { std::cout << "DANGER" << std::endl; } return *p; }
  inline operator ScriptObj* () const { return p; }
  inline bool operator==(const ScriptObj::BasePtr &b) const { return id == b.id; }
  inline bool operator!=(const ScriptObj::BasePtr &b) const { return id != b.id; }
  inline bool operator< (const ScriptObj::BasePtr &b) const { return id <  b.id; }
  inline bool operator<=(const ScriptObj::BasePtr &b) const { return id <= b.id; }
  inline bool operator> (const ScriptObj::BasePtr &b) const { return id >  b.id; }
  inline bool operator>=(const ScriptObj::BasePtr &b) const { return id >= b.id; }
  /*
  inline bool operator==(const ScriptObj::BasePtr &b) const { return p == b.p; }
  inline bool operator!=(const ScriptObj::BasePtr &b) const { return p != b.p; }
  inline bool operator< (const ScriptObj::BasePtr &b) const { return p <  b.p; }
  inline bool operator<=(const ScriptObj::BasePtr &b) const { return p <= b.p; }
  inline bool operator> (const ScriptObj::BasePtr &b) const { return p >  b.p; }
  inline bool operator>=(const ScriptObj::BasePtr &b) const { return p >= b.p; }
  */
  inline ScriptObj* ptr() const { return p; }

  void reset();
  inline bool good() const { return p != 0; }
  inline ObjId get_id() const { return id; }

protected:
  ScriptObj *p;
  ObjId id;
  
  BasePtr() { p = 0; id = 0; }
  BasePtr(ScriptObj *s);
  BasePtr(const BasePtr &ptr);
};

/** A 'weak' pointer.
    Not strong enough to prevent its target from being deleted.
*/

class ScriptObj::WPtr : public ScriptObj::BasePtr
{
public:
  WPtr() {}
  WPtr(ScriptObj *s);
  WPtr(const WPtr &ptr);
  WPtr &operator=(const WPtr &ptr);
  ~WPtr();
  
  void save(std::ostream &os) const;
  void load(std::istream &is);
};

/** A 'strong' pointer.
    The object is maintained as long as at least one strong pointer
    is referencing it.
*/

class ScriptObj::Ptr : public ScriptObj::BasePtr
{
public:
  Ptr() {}
  Ptr(ScriptObj *s);
  Ptr(const Ptr &ptr);
  Ptr &operator=(const Ptr &ptr);
  ~Ptr();
  
  inline operator WPtr () const { return WPtr(p); }
  
  void save(std::ostream &os) const;
  void load(std::istream &is);
};

//===================================================================

REGISTER_TYPEOF(1100, Vector<ScriptObj*>::iterator);
REGISTER_TYPEOF(1101, Vector<ScriptObj*>::const_iterator);
REGISTER_TYPEOF(1102, Vector<ScriptObj::Ptr>::iterator);
REGISTER_TYPEOF(1103, Vector<ScriptObj::Ptr>::const_iterator);
REGISTER_TYPEOF(1104, IdMap::iterator);
REGISTER_TYPEOF(1105, IdMap::const_iterator);
REGISTER_TYPEOF(1106, IdObjMap::iterator);
REGISTER_TYPEOF(1107, IdObjMap::const_iterator);
REGISTER_TYPEOF(1108, ObjIdMap::iterator);
REGISTER_TYPEOF(1109, ObjIdMap::const_iterator);
REGISTER_TYPEOF(1110, NameObjMap::iterator);
REGISTER_TYPEOF(1111, NameObjMap::const_iterator);
REGISTER_TYPEOF(1112, NameVarMap::iterator);
REGISTER_TYPEOF(1113, NameVarMap::const_iterator);
REGISTER_TYPEOF(1114, Vector<ScriptObj::WPtr>::iterator);
REGISTER_TYPEOF(1115, Vector<ScriptObj::WPtr>::const_iterator);

REGISTER_TYPEOF(1116, std::set<ScriptObj::WPtr*>::iterator); 
//REGISTER_TYPEOF(1117, std::set<ScriptObj::WPtr*>::const_iterator); 
REGISTER_TYPEOF(1118, std::set<ScriptObj::Ptr*>::iterator); 
//REGISTER_TYPEOF(1119, std::set<ScriptObj::Ptr*>::const_iterator); 

REGISTER_TYPEOF(1120, std::set<ScriptObj::WPtr>::iterator); 
//REGISTER_TYPEOF(1121, std::set<ScriptObj::WPtr>::const_iterator); 
REGISTER_TYPEOF(1122, std::set<ScriptObj::Ptr>::iterator); 
//REGISTER_TYPEOF(1123, std::set<ScriptObj::Ptr>::const_iterator); 

#endif
