#ifndef ServerObjData_H
#define ServerObjData_H

// $Id: ServerObjData.H 6867 2008-04-05 18:59:22Z furtak $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "Coord.H"
#include "Array.H"
#include "PlayerSet.H"
#include "Object.H"
#include "SType.H"
#include "SaveLoad.H"

class GameObj;

typedef Array<GameObj*> Neighbors;

//===================================================================

/** Cached motion & sight properties.
    Pointers to oft-used variables so they don't have to be repeatedly located.
    Update GameObj::init_sod() if properties change.
*/

class ServerObjData : public SaveLoad {

public:

  sintptr *shape;
  sintptr *x, *y;
  sintptr *x1, *y1;  // rectangle upper left  / line endpoint
  sintptr *x2, *y2;  // rectangle lower right / line endpoint
  sintptr *radius;

  bool   is_boundary;

  sintptr *owner;

  sintptr *max_speed;
  sintptr *speed;     ///< It may be possible for an object to be moving with zero speed,
                      ///< if, say, it's caught in a spell, or it hops.
  sintptr *is_moving; ///< The object is at least trying to move and hasn't encountered an obstacle.
  sintptr *zcat;
  sintptr *collides;  ///< Collision mask determines which objects can collide.
  sintptr *ncollides; ///< Exceptions to the collision mask, objects with intersecting masks don't collide.

  sintptr *sight;     ///< Visibility radius. 360 degree vision for now.
  sintptr *cloaked;   ///< >= 1 if cannot be seen by non-detectors, enemies
  sintptr *cloaked2;  ///< >= 1 iff can only be seen by allies
  sintptr *detector;  ///< True iff the object can see cloaked units.

  //-------------------------------------

  struct PrevData {
    sint4 shape;
    sint4 x, y;
    sint4 x1, y1;
    sint4 x2, y2;
    sint4 radius;
    
    sint4 owner;
    
    sint4 max_speed;
    sint4 speed;
    sint4 is_moving;
    sint4 zcat;
    sint4 collides;
    sint4 ncollides;
    
    sint4 sight;
    sint4 cloaked;
    sint4 cloaked2;
    sint4 detector;

    PrevData();
  } prev;

  //-------------------------------------

  PlayerSet seen_by, prev_seen_by;
  Array<sint4> last_lost; // sight of

  bool in_game, is_global, is_shared;

  ServerObjData() {
    is_boundary = false;
    in_game = false;
    is_global = false;
    is_shared = false;

    shape = 0;
    x = y = 0;
    x1 = y1 = 0;
    x2 = y2 = 0;
    radius = 0;
    owner = 0;
    max_speed = speed = 0;
    is_moving = 0;
    zcat = 0;
    collides = 0;  
    ncollides = 0; 
    sight = 0;     
    cloaked = 0;   
    cloaked2 = 0;  
    detector = 0;  
  }

  void copy_to_prev();

  void save(std::ostream &os) const;
  void load(std::istream &is);
};

/** Data used by the server for object motion and collisions.
    Distinguished by being information not available to the scripts.
*/

class ServerMotionData : public SaveLoad {

public:

  ScalarPoint  start_f_pos;   ///< Position the current straight line movement started from (in fine coordinates).
  ScalarPoint  target_f_pos;  ///< Movement destination (in fine coordinates).
  ScalarPoint  f_delta;       ///< Vector from the start to the target in fine units.

  real8  inv_ttt;             ///< 1 / time to reach target

  ScalarPoint  fine_pos;      ///< Sub-grid pos (sv_pos is nearest grid point)
  ScalarPoint  fine_heading;  ///< Movement vector in fine coordinates.
  ScalarPoint  bc_fine_center;///< Motion bounding circle center.
  Scalar bc_fine_r;           ///< Motion bounding circle radius.
  sint4  ticks_so_far;        ///< Number of ticks the object has moved along the current path.
  sint4  ticks_total;         ///< Total number of ticks needed to move from the start to the destination.

  uint4  neighbor_magic;      ///< For Game::move_objects().
  uint4  visit_magic;
  uint4  unique_magic;        ///< For Game::handle_component().

  Neighbors neighbors;

  ServerMotionData() {
    inv_ttt = 0;
    fine_pos.x = fine_pos.y = -1;
    bc_fine_center.x = bc_fine_center.y = -1;
    bc_fine_r = -1;
    ticks_so_far = ticks_total = -1;
    visit_magic = neighbor_magic = unique_magic = 0;
  }

  void save(std::ostream &os) const;
  void load(std::istream &is);
};

//===================================================================

#endif
