#ifndef Terrain_H
#define Terrain_H

/** @file Terrain.H
    Generic Terrain description. The ORTS server will read this information,
    choose start locations at random and run the game. The game script will
    populate start and resource locations, generate trees, introduce NPCs, etc.

    $Id: Terrain.H 5279 2007-06-23 02:49:08Z mburo $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "Coord.H"
#include "SaveLoad.H"
#include "Tile.H"
#include "Map.H"

class Terrain : public SaveLoadAbs {

public:

  class TileOpt;
  class BuildLocation;
  
  sint4 max_players;   // maximal number of players
  
  Map<Tile>    tiles;         // height field + tile type information
  Map<TileOpt> tile_opts;     // tiles that can(/should?) be blocked by trees etc.
  Vector<BuildLocation> start_loc; // potential starting positions (# >> max_players)
  Vector<BuildLocation> resource_loc; // other resource locations

  Terrain() { max_players = 0; }
  virtual ~Terrain() { }

  void save(std::ostream &os) const {
    OS_WRITE(max_players);
    tiles.save(os);
    tile_opts.save(os);
    SaveLoad::save(start_loc, os);
    SaveLoad::save(resource_loc, os);    
  }
  
  void load(std::istream &is) {
    IS_READ(max_players);
    tiles.load(is);
    tile_opts.load(is);
    SaveLoad::load(start_loc, is);
    SaveLoad::load(resource_loc, is);    
  }

  
  // local classes

  // available resource at build location
  
  class Resource : public SaveLoad {

  public:

    sint4 type;         // 0,1,... (game dependent!)
    real4 amount_ratio; // >= 0.0 and <= 1.0 = max

    Resource() { type = 0; amount_ratio = 0.0; }
    virtual ~Resource() { }

    void save(std::ostream &os) const {
      OS_WRITE(type);
      OS_WRITE(amount_ratio);      
    }
    
    void load(std::istream &is) {
      IS_READ(type);
      IS_READ(amount_ratio);      
    }
  };


  // build location information
  
  class BuildLocation : public SaveLoad {

  public:

    ScalarPoint loc;   // center point
    sint4 orientation; // 0:N 1:NE ... 7:NW, allow NE,SE,SW,NW?
    Vector<Resource> resources; // what resources
    sint4 w, h;        // width and height (game dependent!)
  
    void save(std::ostream &os) const {
      OS_WRITE(loc);
      OS_WRITE(orientation);
      SaveLoad::save(resources, os);
      OS_WRITE(w);
      OS_WRITE(h);      
    }
    
    void load(std::istream &is) {
      IS_READ(loc);
      IS_READ(orientation);
      SaveLoad::load(resources, is);
      IS_READ(w);
      IS_READ(h);      
    }

    BuildLocation() {
      orientation = 0; w = h = 0;
    }

    virtual ~BuildLocation() { }
  };


  // additional tile options/hints
  
  class TileOpt : public SaveLoad {

  public:

    bool to_be_occupied;

    void save(std::ostream &os) const { OS_WRITE(to_be_occupied); }
    void load(std::istream &is)       { IS_READ(to_be_occupied);  }

    TileOpt() {
      to_be_occupied = false;
    }

    virtual ~TileOpt() { }
  };

};

REGISTER_TYPEOF(10001, Vector<Terrain::BuildLocation>::iterator );
REGISTER_TYPEOF(10002, Vector<Terrain::BuildLocation>::const_iterator );
REGISTER_TYPEOF(10003, Vector<Terrain::Resource>::iterator );
REGISTER_TYPEOF(10004, Vector<Terrain::Resource>::const_iterator );

#endif
