#ifndef Tile_H
#define Tile_H

/** @file Tile.H
    Terrain tile class.

    $Id: Tile.H 5279 2007-06-23 02:49:08Z mburo $
    $Source$
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "SaveLoad.H"

class ByteStream;

// square component of a height field
// quite general: can define endpoint heights of all four edges individually
// supports split tiles (two triangles): |/| or |\|

// special case:
//   only one height per grid intersection point
// => continuous height field
// all provided map generators now create continuous height fields only
// set/get_topo() and set/get_type() are sufficient to write/read all
// relevant tile data in this case

class Tile : public SaveLoad {

public:

  typedef uint4 Type;
  enum Types   { UNKNOWN = 0 }; // other types (>0) defined in game script
  enum Split   { NO_SPLIT=0, TB_SPLIT /* |\| */, BT_SPLIT /* |/| */ };
  enum Compass { N=1, NE, E, SE, S, SW, W, NW }; // don't change
  static const sint4 MAX_HEIGHT = 15;
  
  Tile() { clear(); }

  bool is_flat() const { return flat; }     // same types, same corner heights
  bool is_split() const { return split != NO_SPLIT; } // is tile split?

  /// flat tile, uniform type
  void set_flat(sint4 height);

  //  set/get:
  //   - four corner heights,
  //   - split: NO_SPLIT (corner heights must be equal), TB_SPLIT-|\| , BT_SPLIT-|/|
  void set_topo(sint4 hNW, sint4 hNE, sint4 hSE, sint4 hSW, Split split);
  void get_topo(sint4 &hNW, sint4 &hNE, sint4 &hSE, sint4 &hSW, Split &split) const;

  Split get_split() const { return static_cast<Split>(split); }

  /// Sets the tile type of the parts adjacent to the West and East sides of the tile.
  /// Uses the current split.
  void set_type(Type type) { typeW = typeE = type; }
  void set_type(Type tW, Type tE) { typeW = tW; typeE = tE; }
  void set_type(Type tW, Type tE, Split s);

  void get_type(Type &tW, Type &tE) const {
    tW = static_cast<Type>(typeW); tE = static_cast<Type>(typeE);
  }

  Type get_typeW() const { return static_cast<Type>(typeW); }
  Type get_typeE() const { return static_cast<Type>(typeE); }
  // typeW is the default for unsplit tiles, so make sure it's returned if split == NO_SPLIT
  Type get_typeN() const {
    return (static_cast<Split>(split) == TB_SPLIT) ? static_cast<Type>(typeE) : static_cast<Type>(typeW);
  }
  Type get_typeS() const {
    return (static_cast<Split>(split) == BT_SPLIT) ? static_cast<Type>(typeE) : static_cast<Type>(typeW);
  }

  // minimum/maximum corner height
  sint4 get_min_h() const { return min_h; }
  sint4 get_max_h() const { return max_h; }

  /// returns height of edge endpoints
  /// direction in (N,E,S,W); corners are ordered clockwise
  ///   ( (h1, h2) = (NE, SE), (SE, SW), ... )
  /// direction = diagonal (NW,NE,SE,SW) generates error message

  void get_edge_height(sint4 &h1, sint4 &h2, Compass dir) const;

  // return height at point (x,y) in [0,1]x[0,1]
  real4 height_at(real4 x, real4 y) const;

  /* c:
     
      N 

     0 1
    7   2
    6   3
     5 4
  */
  
  Type get_type_at_corner(sint4 c) const;

  // see h[] below
  sint4 get_height(sint4 i) const { return h[i]; }

  bool is_half() const { return half; }     // two plateaus at different heights
                                            // => discontinuous (not used anymore)
  // binary I/O
  void read_b(ByteStream &s);
  void write(std::string &s);

  void display() const { std::cout << h[0]; } // debug output
  
  void save(std::ostream &os) const;
  void load(std::istream &is);

  void clear(); // reset 

  /// Makes a ramp pointing in a certain direction (1-8) (N,NE,E,SE,S,SW,W,NW).
  /// The end of the ramp is the part of the tile closest to the direcion, the
  /// start is the part farthest away.
  /// For a diagonal ramp dir=(2,4,6,8), the height along the edge perpendicular to
  /// the direction is the average of the start and end heights.  If this edge
  /// is different from the current split, the tile is reset to a uniform tile type.

  // end_height is the height of the side pointed to by dir
  // start_height is the height of the side opposite
  // mid_height is the height of the crease for diagonal ramps

  void set_ramp(sint4 start_height, sint4 end_height, Compass dir);
  void set_ramp(sint4 start_height, sint4 mid_height, sint4 end_height, Compass dir);

  /// Splits the tile along the edge perpendicular to the direction dir (see above).
  void set_half_tile(sint4 start_height, sint4 end_height, Compass dir); // set types later

protected:
  
  boost::array<uint1,3> data;       ///< Bit-encoding of tile data.  Described in set_corners().
                       ///< This is the tile information sent to clients and read from the world description.

  // values extracted from data[3]

  uint1 split;         // the diagonal along which the tile is split / creased. see Split above
  Type  typeW, typeE;  // if the tile is split, the types of each portion
  boost::array<sint2,8> h;          ///< Corner heights of the tile.
                       ///< Clockwise ordering; 0,1 = North wall (NW NE); 2,3 = East wall (NE, SE); ...

  bool  flat;
  bool  half;          // if the tile is a half-tile
  sint2 min_h;         ///< Lowest height on the tile.
  sint2 max_h;         ///< Highest height on the tile.

  // internal helpers
  
  void set_corners(); // Convert the bit-encoded data to corner heights.
  void set_bits();    // Encodes the slope direction and height values as bit data.
};

//===================================================================


#endif
