/** @file Token.C
    @see Token.H.

    $Id: Token.C 12791 2012-11-07 18:19:57Z mburo $
    $Source$
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Token.H"
#include "Options.H"

using namespace std;

//===================================================================

Tokenizer::DefineTable Tokenizer::cmd_def_table;

//===================================================================

//#define VERBOSE

Tokenizer::Tokenizer()
{
  define_table = new DefineTable;
  my_def_table = true;
  my_data_buf = false;
  cleanup();
}

//-------------------------------------------------------------------

/* Points to the original definition table, updating it while parsing.  For recursive calls. */

Tokenizer::Tokenizer(Tokenizer *t)
{
  if (t) {
    define_table = t->define_table;
    my_def_table = false;
  } else {
    define_table = new DefineTable;
    my_def_table = true;
  }
  
  my_data_buf = false;
  cleanup();
}

//-------------------------------------------------------------------

Tokenizer::~Tokenizer()
{
  if (my_def_table) delete define_table;
  if (my_data_buf) delete [] data_start;
}

//-------------------------------------------------------------------

void Tokenizer::cleanup()
{
  if (my_data_buf) {
    delete [] data_start;
    my_data_buf = false;
  }
  data_start = data_ptr = data_end = 0;
}

//-------------------------------------------------------------------

void Tokenizer::init_cmd_def(Vector<string> &var, Vector<string> &def)
{
  assert(var.size() == def.size());
  FORU (i, var.size()) {
    cmd_def_table.define_var(var[i], def[i]);
  }
}

//-------------------------------------------------------------------

void Tokenizer::init_cmd_def()
{
  Vector<string> def;
  ::Options::get("-D", def);
  FORALL (def, i) {
    const string &s = *i;
    size_t p = s.find("=");
    if (p == string::npos) {
      cmd_def_table.define_var(s, "");
    } else {
      string var = s.substr(0,p);
      string def = s.substr(p+1, s.size()-1);
      cmd_def_table.define_var(var, def);
    }
  }
}

//-------------------------------------------------------------------

void Tokenizer::init(const string &str, string file, string path)
{
  cleanup();
  char *temp = new char[str.length()+1];
  memcpy(temp, str.c_str(), str.length()+1);

  data_ptr = temp;
  data_start = temp;
  data_end = data_start + str.length();
  type = UNKNOWN;
  cur_line_num = 1;
  fname = file;
  fpath = path;
  my_data_buf = true;
}

void Tokenizer::init(const char *ptr, sint4 length, string file, string path)
{
  cleanup();
  data_ptr = ptr;
  data_start = ptr;
  data_end = data_start + length;
  type = UNKNOWN;
  cur_line_num = 1;
  fname = file;
  fpath = path;
  my_data_buf = false;
}

bool Tokenizer::good() const
{
  return (!tokenStack.empty()) || (data_ptr && *data_ptr);
}

void Tokenizer::readto(string &s, istream &ss, char C)
{
  char c;
  s.clear();
  while (1) {
    ss.get(c);
    if (c != C) s += c;
    else return;
  }
}


char Tokenizer::readto(string &s, istream &ss, char c1, char c2)
{
  char c;
  s.clear();
  while (1) {
    ss.get(c);
    if (c != c1 && c != c2) s += c;
    else return c;
  }
}

/** Skip define table lookup iff redef. */

const string Tokenizer::next_raw_token(bool redef)
{
  char ch;
  string token;

  assert(data_ptr);

  if (!tokenStack.empty()) {
    token = tokenStack.top();
    tokenStack.pop();
    type = type_of(token[0]);
    return token;
  }

  if (*data_ptr == 0) return "";

  type = UNKNOWN;

  while ((ch = *(data_ptr++))) {

    if (ch == '\r') continue;

    if (ch == '\n') ++cur_line_num;

    if (type == UNKNOWN) {
      type = type_of(ch);
      assert(type);
      token += ch;

      if (ch==',') return token;
      if (ch==';' || ch=='(' || ch==')') return token;
      if (ch=='[' || ch==']')            return token;
      if (ch=='!' || ch == '-') {
        if (*data_ptr) {
          ch = *data_ptr;
          if (ch=='=') {
            token += ch;
            ++data_ptr;
          }
        }
        return token;
      }
      continue;
    }
    
    if (type == STRING) {
      if (ch == '\\') {
        if (!(ch = *(data_ptr++))) {
          --data_ptr;
          return token;
        }
        if (ch == '"') token += '"';
        if (ch == 'n') token += '\n';
        if (ch == 't') token += '\t';
        if (ch == '\\') token += '\\';
        continue;
      }
      token += ch;
      if (ch == '"') break; // end of string
      continue;
    }

    if ((type==NUMBER  && (isdigit(ch) || ch=='.')) ||
        (type==ID      && (isalnum(ch) || ch=='.' || ch=='_')) ||
        (type==SPACE   && isspace(ch)) ||
        (type==COMMENT && ch!='\n') ||
        (type==PUNCT   && ispunct(ch) && ch!=';' && ch!='(' && ch!=')' && ch!='[' && ch!=']' )) {
      token += ch;
      continue;
    }

    // otherwise, the token has ended
    --data_ptr;
    if (ch == '\n') --cur_line_num;
    break;
  }

  assert(type);
  if (ch == 0) {
    --data_ptr;
    assert(*data_ptr == 0);
  }

  // check if the token has been redefined
  
  Vector<string> def;

  if (redef && cmd_def_table.is_def(token)) {
    cmd_def_table.get_def(token, def);
  } else if (define_table->is_def(token)) {
    define_table->get_def(token, def);
  }

  if (!def.empty()) {
    string o = token;
    for (sint4 i = def.size()-1; i > 0; --i) {
      putback_token(def[i]);
    }
    token = def[0];
    type = type_of(token[0]);
  }

  return token;
}

/** Skip define table lookup iff redef. */

const string Tokenizer::next_token(bool redef)
{
  string token;
  while (good()) { // not end of file
    token = next_raw_token(redef);
    if ((token_type() != SPACE) && (token_type() != COMMENT)) break;
    token = "";
  }
#ifdef VERBOSE
  cout << "TOKEN = '" << token << "'" << endl;
#endif
  return token;
}

// if there is a tag next, return it; otherwise nothing

const string Tokenizer::next_tag()
{
  string tag;
  string a,b,c;
  
  push_bookmark();

  a = next_token();
  if (a != "<" && a != "</") {
    flip_bookmark();
    drop_bookmark();
    return "";
  }
  
  b = next_token();
  if (token_type() != ID) {
    flip_bookmark();
    drop_bookmark();
    return "";
  }

  c = next_token();
  if (c != ">") {
    flip_bookmark();
    drop_bookmark();
    return "";
  }

  tag = a + b + c;

  if (!tokenStack.empty()) {
    string token = tokenStack.top();
    if (token[0] != '\n') ERR2("expecting newline immediately after tag:", token.c_str());
    token = token.substr(1, token.size()-1);
    if (!token.empty()) tokenStack.push(token);
  } else {
    if (*data_ptr != '\n') ERR2("expecting newline immediately after tag:", tag.c_str());
    ++data_ptr;
  }
  ++cur_line_num;

  return tag;
}

/// The current line of text the cursor is on.

const string Tokenizer::scan_line() const
{
  const char *start = data_ptr;
  while ((start > data_start) && (*start != '\n' && *start != '\r')) --start;
  string line(1, *start++);
  while (*start != '\n' && *start != '\r' && *start) line += *start++;
  // strip surrounding white-space
  while (!line.empty() && (line[0] == '\n' || line[0] == '\r' || line[0] == ' ')) line.erase(0, 1);
  while (!line.empty() && (line[line.size()-1] == '\n' || line[line.size()-1] == '\r')) line.erase(line.size()-1);
  return line;
}

/// The rest of the line.

const string Tokenizer::read_line()
{
  string line;
  while (*data_ptr != '\n' && *data_ptr != '\r' && *data_ptr) {
    line += *data_ptr;
    ++data_ptr;
  }
  ++cur_line_num;
  return line;
}

/** Returns the type of token starting with the given character. */

Tokenizer::TokenType Tokenizer::type_of(const char ch)
{
  if (ch == '#')    return COMMENT;
  if (ch == '"')    return STRING;
  if (isspace(ch))  return SPACE;
  if (ispunct(ch))  return PUNCT;
  if (isdigit(ch))  return NUMBER;
  if (isalpha(ch))  return ID;

  return UNKNOWN;
}


void Tokenizer::putback(const string &data)
{
#ifdef VERBOSE
  cout << "PUTBACK '" << data << "'" << endl;
#endif
  if (data == "") return;

  static Vector<char> buf;
  buf.resize(data.length()+1);
  memcpy(&buf[0], data.c_str(), data.length()+1);

  Tokenizer temp(0); // mic: was (false)
  temp.init(&buf[0]);
  Vector<string> raw;
  while (temp.good()) { raw.push_back(temp.next_raw_token()); }
  for (sint4 i = raw.size()-1; i >= 0; --i)
    if (raw[i] != "") {
      #ifdef VERBOSE
      cout << "PUTBACK : TOKENSTACK '" << raw[i] << "'" << endl;
      #endif
      tokenStack.push(raw[i]);
    }

  if (buf.size() > 256) buf.resize(256);
}

void Tokenizer::putback_token(const string &data)
{
#ifdef VERBOSE
  cout << "PUTBACKT '" << data << "'" << endl;
#endif
  if (data == "") return;
  tokenStack.push(data);
}

void Tokenizer::putback_tag(const string &data)
{
  putback(data + "\n");
#ifdef VERBOSE
  cout << "PUTBACKG '" << data << "'" << endl;
#endif
}

void Tokenizer::push_bookmark()
{
  Bookmark b;
  b.ptr = data_ptr;
  b.line = cur_line_num;
  b.tStack = tokenStack;
  bookmark.push(b);
}

void Tokenizer::flip_bookmark()
{
  data_ptr = bookmark.top().ptr;
  cur_line_num = bookmark.top().line;
  tokenStack = bookmark.top().tStack;
}

//===================================================================

void Tokenizer::define_var(const string &var, const string &def)
{
#ifdef VERBOSE
  cout << "DEFINE '" << var << "' as '" << def << "'" << endl;
#endif
  define_table->define_var(var, def);
}

//===================================================================

void Tokenizer::DefineTable::define_var(const string &var, const string &def)
{
  Tokenizer temp(0); // !!! was (false)
  const char *cdef = def.c_str();
  temp.init(cdef);

  Vector<string> defn;
  string token;

  token = temp.next_raw_token(false);
  while (token != "") {
    defn.push_back(token);
    token = temp.next_raw_token(false);
  }
  assert(!defn.empty());
  def_table[var] = defn;
}

//-------------------------------------------------------------------

bool Tokenizer::DefineTable::is_def(const string &var)
{
  return def_table.find(var) != def_table.end();
}

//-------------------------------------------------------------------

void Tokenizer::DefineTable::get_def(const string &var, Vector<string> &def)
{
  if (!is_def(var)) {
    def.clear();
    return;
  }
  def = def_table[var];
}

//===================================================================
