/** @file View.C
    Computing what each player can see.

    @see Game.H

    $Id: View.C 12791 2012-11-07 18:19:57Z mburo $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/View.C,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "Game.H"
#include "GameObj.H"
#include "CycleCounter.H"
#include "ServerObjData.H"
#include "GameTile.H"
#include "PlayerInfo.H"
#include "Compress.H"
#include "Slope.H"
#include "BoolMatrix.H"
#include "Profiler.H"
 
using namespace std;

REGISTER_TYPEOF(120, set<Slope>::iterator);

//===================================================================

/** Allocate memory for the vision computations.
    Memory for storing the byte alignment of each tile in the map,
    as well as a permanent pool used by the visibility computations.
*/

void Game::allocate_vision_memory()
{
  const sint4 TS = sizeof(VisionT)*8;
  uint4 blocks_x = (pf.get_width() + TS - 1) / TS;
  uint4 n = pf.get_num();

  assert(n > 0);
  assert(blocks_x * pf.get_height() > 0);

  tind2bitmap   = new sint4[n];
  bitmap2tind   = new sint4[blocks_x * pf.get_height()];
  lshift_amount = new sint4[n]; // distance from the left edge of a tileview to a word boundary

  // cleared and reset for each player at each call to compute_view / client_compute_visible
  v1w  = new sint1[n];  // max land sight range (of the units centered on the tile)
  v1e  = new sint1[n];
  v2   = new sint1[n];
  dv1w = new sint1[n]; // detector ranges
  dv1e = new sint1[n];
  dv2  = new sint1[n];

  const uint4 wsq = GameConst::WIDTH_SQUARED;
  tile_known_range = new BitStrip[wsq];
  tile_fshadow_range = new BitStrip[wsq];
  tile_f_visible_range = new BitStrip[wsq];
  
  // see "Shadows and Visibility" in precompute_vision_data for usage
  FORU (i, 4) {
    edge_shadows[i] = new BitStrip*[wsq];
    tile_shadow_range[i] = new sint4*[wsq];
    tile_visible_range_W[i] = new sint4*[wsq];
    tile_visible_range_E[i] = new sint4*[wsq];
    FORS (j, wsq) {
      edge_shadows[i][j] = new BitStrip[6];
      tile_shadow_range[i][j] = new sint4[12]; // # of different shadow profiles
      tile_visible_range_W[i][j] = new sint4[12];
      tile_visible_range_E[i][j] = new sint4[12];
    }
  }
}

//===================================================================

void Game::deallocate_vision_memory()
{
  delete [] tind2bitmap;
  delete [] bitmap2tind;
  delete [] lshift_amount;

  delete [] v1w;
  delete [] v1e;
  delete [] v2;

  delete [] dv1w;
  delete [] dv1e;
  delete [] dv2;

  delete [] tile_known_range;
  delete [] tile_fshadow_range;
  delete [] tile_f_visible_range;

  FORU (i, 4) {
    FORS (j, GameConst::WIDTH_SQUARED) {
      delete [] edge_shadows[i][j];
      delete [] tile_shadow_range[i][j];
      delete [] tile_visible_range_W[i][j];
      delete [] tile_visible_range_E[i][j];
    }
    delete [] edge_shadows[i];
    delete [] tile_shadow_range[i];
    delete [] tile_visible_range_W[i];
    delete [] tile_visible_range_E[i];
  }
}

//===================================================================

sint4 slope_to_index(const Slope &s, Vector<Slope> &v)
{
  FORU (i, v.size()) {
    if (s == v[i]) return i;
  }
  return -1;
}

//-------------------------------------------------------------------

void slope_to_segment(const Slope &s, Vector<Slope> &v, sint4 &x, sint4 &y)
{
  x = -1;
  y = 0;
  FORU (i, v.size()-1) {
    if (s >= v[i]) {
      if (s > v[i]) ++x;
      ++y;
    }
  }
  if (x < 0) x = 0;
}

//===================================================================

/** Generate tile slope data for relative positions and storage offsets.
    Stores results such as how much each tileview must be shifted to be word-aligned.
    Calculates the range of slopes taken up by the different shadow types at all possible
    positions relative to the viewing tile.
*/

void Game::precompute_vision_data()
{
  const sint4 TS = sizeof(VisionT)*8;
  const uint4 blocks_x = (pf.get_width() + TS - 1) / TS;

  GameTile::edge_shadows = &edge_shadows;

  // ---- Tile offsets ----------------------
  
  FORU (i, blocks_x * pf.get_height()) {
    bitmap2tind[i] = ((i / blocks_x) * pf.get_width()) + (i % blocks_x) * TS;
  }

  FORS (x, pf.get_width()) {
    FORS (y, pf.get_height()) {
      uint4 tind = pf.xy2ind(x, y);
      tind2bitmap[tind] = (y-GameConst::VIEW_MAX)*blocks_x + (x-GameConst::VIEW_MAX+TS)/TS -1;
      // # of tiles between the left edge of this view and the nearest T word boundary
      lshift_amount[tind] = (x-GameConst::VIEW_MAX+TS) % TS;
    }
  }

  // ---- Slopes ----------------------------

  Vector<Slope> slopes;
  sint4 num_slopes;

  Slope m[(GameConst::VIEW_WIDTH+1)*(GameConst::VIEW_WIDTH+1)];

  {
    set<Slope> slope_set;

    FORS (i, GameConst::VIEW_WIDTH+1) {
      FORS (j, GameConst::VIEW_WIDTH+1) {
        sint4 k  = i + j * (GameConst::VIEW_WIDTH+1);
        sint4 ii = (i - GameConst::VIEW_WIDTH / 2) * 2 - 1;
        sint4 jj = (j - GameConst::VIEW_WIDTH / 2) * 2 - 1;
        if (abs(jj) > abs(ii)) swap(ii, jj);
        m[k].set(ii, jj);
        slope_set.insert(m[k]);
      }
    }

    FORALL (slope_set, it) { slopes.push_back(*it); }
    num_slopes = slopes.size();
  }

  qstrip.alloc(num_slopes-1);

  Slope min_bound(1,-1); // the quadrant dividing lines
  Slope max_bound(1,1);
  
  // ---- Full Visibility Ranges ------------

  {
    FORS (i, GameConst::VIEW_WIDTH) {
      FORS (j, GameConst::VIEW_WIDTH) {
        sint4 k = i + j * GameConst::VIEW_WIDTH;
        sint4 mk = i + j * (GameConst::VIEW_WIDTH+1);

        sint4 min, max, si;

        min = max = slope_to_index(m[mk], slopes);

        si = slope_to_index(m[mk+1], slopes);
        if (si < min) min = si; if (si > max) max = si;

        si = slope_to_index(m[mk+GameConst::VIEW_WIDTH+1], slopes);
        if (si < min) min = si; if (si > max) max = si;

        si = slope_to_index(m[mk+GameConst::VIEW_WIDTH+2], slopes);
        if (si < min) min = si; if (si > max) max = si;

        tile_known_range[k].alloc(num_slopes-1);
        tile_known_range[k].clear();
        for (sint4 q = min; q < max; q++) tile_known_range[k].set(q);

        tile_fshadow_range[k].alloc(num_slopes-1);
        tile_fshadow_range[k].fill();
        for (sint4 q = min; q < max; q++) tile_fshadow_range[k].unset(q);

        sint4 ii = -GameConst::VIEW_MAX + i;
        sint4 jj = -GameConst::VIEW_MAX + j;
        if (abs(jj) > abs(ii)) swap(ii, jj);

#if 1
        // line through the middle
        if (ii != 0) {
          Slope vis(ii, jj);
          slope_to_segment(vis, slopes, min, max);
        } else {
          min = max = 0;
        }
#else
        // instead of a point, use a centered box, 1/2 x 1/2
        Slope sl[4];
        sl[0].set(ii*4+1, jj*4+1);
        sl[1].set(ii*4+1, jj*4-1);
        sl[2].set(ii*4-1, jj*4+1);
        sl[3].set(ii*4-1, jj*4-1);
        
        min = num_slopes;
        max = -1;

        FORU (k, 4) {
          if (sl[k] < min_bound) sl[k] = min_bound;
          if (sl[k] > max_bound) sl[k] = max_bound;
          sint4 mn, mx;
          slope_to_segment(sl[k], slopes, mn, mx);
          if (mn < min) min = mn;
          if (mx > max) max = mx;
        }
#endif
        BitStrip &temp = tile_f_visible_range[k];
        temp.alloc(num_slopes-1);
        temp.clear();

        if (max > num_slopes-1) max = num_slopes-1;
        for (k = min; k < max; ++k)
          temp.set(k);
      }
    }

    sint4 d2 = ok_dist2(GameConst::VIEW_MAX);

    fv_yc[0] = 0;
    for (sint4 x = 1; x <= GameConst::VIEW_MAX; ++x) {
      fv_yc[x] = 0;
      for (sint4 y = 0; y <= GameConst::VIEW_MAX; ++y) {
        sint4 sq = square(x) + square(y);
        if (sq <= d2) fv_yc[x]++;
        if (y >= x) break;
        if (sq > d2) break;
      }
    }
  }

  // ---- Shadows and Visibility ------------

  {
    // 1 2
    // 0 3
    // which corners are connected by the walls given in 'profiles' below
    sint4 walls[6][2] = { {0, 1}, {1, 2}, {2, 3}, {3, 0}, {0, 2}, {1, 3} };
    sint4 offset[4][2] = { {-1,-1}, {-1,1}, {1,1}, {1,-1} };

    sint4 quad_transform[4][4] = { {1,0,0,1}, {0,1,1,0}, {-1,0,0,-1}, {0,-1,-1,0} };

    // calculate the shadows for each individual edge

    FORU (quad, 4) { // quadrant
      FORU (i, 6) {  // for each wall
        for (sint4 x = 1; x <= GameConst::VIEW_MAX; ++x) {
          for (sint4 y = 1-fv_yc[x]; y < fv_yc[x]; ++y) { // distance from the axis

            const sint4 r = GameConst::VIEW_MAX;

            sint4 xx = x * quad_transform[quad][0] + y * quad_transform[quad][1];
            sint4 yy = x * quad_transform[quad][2] + y * quad_transform[quad][3];
            sint4 k = (xx + r) + GameConst::VIEW_WIDTH * (yy + r);
            

            BitStrip &temp = edge_shadows[quad][k][i];
            temp.alloc(num_slopes-1);
            temp.fill();
            
            Slope c1(1,1);
            Slope c2(1,-1);

            sint4 min = walls[i][0];
            sint4 max = walls[i][1];

            sint4 run1 = xx*2 + offset[min][0];
            sint4 run2 = xx*2 + offset[max][0];

            sint4 rise1 = yy*2 + offset[min][1];
            sint4 rise2 = yy*2 + offset[max][1];


            if (quad == 1 || quad == 3) {
              swap(rise1, run1);
              swap(rise2, run2);
            }


            Slope s1( run1, rise1);
            Slope s2( run2, rise2);

            if (s1 > c1) s1 = c1;
            if (s1 < c2) s1 = c2;
            if (s2 > c1) s2 = c1;
            if (s2 < c2) s2 = c2;

            min = slope_to_index( s1, slopes );
            max = slope_to_index( s2, slopes );
            
            if (min > max) swap(min, max);
            if (min == max) { // this is going to occur when a diagonal edge wraps around a diagonal
              if (min < (num_slopes-2)-max)
                min = 0;
              else
                max = num_slopes-2;
            }

            for (sint4 q = min; q < max; ++q) temp.unset(q);
          }
        }
      }
    }

    // There are a total of 2^4 * 3 wall configurations: each of the 4 side walls can be on or off,
    // independent of the 3 split options for the diagonal wall.
    // In terms of the "shadow" cast by the tile, there are 12 options (ignoring the two extremes
    // of completely obscured and completely visible).
    // E.g. 3 side walls are equivalent to all 4, 2 corner files imply the associated diagonal wall, etc.

    sint4 profiles[12][6] = { {1, 0, 1, 0, 0, 0}, // parallel lines 0
                              {0, 1, 0, 1, 0, 0},
                              {1, 0, 0, 1, 0, 0}, // corner         2
                              {0, 0, 1, 1, 0, 0},
                              {0, 1, 1, 0, 0, 0},
                              {1, 1, 0, 0, 0, 0},
                              {0, 0, 0, 0, 1, 0}, // diagonal       6
                              {0, 0, 0, 0, 0, 1},
                              {1, 0, 0, 0, 0, 0}, // single edge    8
                              {0, 0, 0, 1, 0, 0},
                              {0, 0, 1, 0, 0, 0},
                              {0, 1, 0, 0, 0, 0} };

    // [Quadrant] [1st loop, 2nd loop] [wall]

    sint4 vis_blocking_W[4][2][6] = {{{1,0,0,1, 0,0}, {1,1,0,0, 0,0}},
                                     {{1,0,0,1, 1,0}, {0,0,1,1, 1,0}},
                                     {{0,1,1,0, 1,1}, {0,0,1,1, 1,1}},
                                     {{0,1,1,0, 0,1}, {1,1,0,0, 0,1}} };

    sint4 vis_blocking_E[4][2][6] = {{{1,0,0,1, 1,1}, {1,1,0,0, 1,1}},
                                     {{1,0,0,1, 0,1}, {0,0,1,1, 0,1}},
                                     {{0,1,1,0, 0,0}, {0,0,1,1, 0,0}},
                                     {{0,1,1,0, 1,0}, {1,1,0,0, 1,0}} };

    // go through the tiles in roughly the same order that the vision computation code would

    BitStrip temp, v_temp_W, v_temp_E;
    temp.alloc(num_slopes-1);
    v_temp_W.alloc(num_slopes-1);
    v_temp_E.alloc(num_slopes-1);

    FORU (quad, 4) {
      for (sint4 x = 1; x <= GameConst::VIEW_MAX; ++x) {
        for (sint4 y = 1-fv_yc[x]; y < fv_yc[x]; ++y) { // distance from the axis
          FORU (i, 12) {
            temp.fill();

            sint4 b = (y < 0) ? 1 : 0;

            const sint4 r = GameConst::VIEW_MAX;

            sint4 xx = x * quad_transform[quad][0] + y * quad_transform[quad][1];
            sint4 yy = x * quad_transform[quad][2] + y * quad_transform[quad][3];
            sint4 k = (xx + r) + GameConst::VIEW_WIDTH * (yy + r);

            v_temp_W = tile_f_visible_range[k];
            v_temp_E = tile_f_visible_range[k];

            FORU (j, 6) {
              if (profiles[i][j]) {
                temp.and_strip(edge_shadows[quad][k][j]);
                if (vis_blocking_W[quad][b][j])
                  v_temp_W.and_strip(edge_shadows[quad][k][j]);
                if (vis_blocking_E[quad][b][j])
                  v_temp_E.and_strip(edge_shadows[quad][k][j]);
              }
            }
            uint4 index;

            //---------------------------------------
            {
              FOR (index, tile_shadows.size()) {
                if (tile_shadows[index] == temp) break;
              }
              if (index == tile_shadows.size())
                tile_shadows.push_back(temp);
              tile_shadow_range[quad][k][i] = index;
            } 
            //---------------------------------------
            {
              FOR (index, tile_visibility.size()) {
                if (tile_visibility[index] == v_temp_W) break;
              }
              if (index == tile_visibility.size())
                tile_visibility.push_back(v_temp_W);
              tile_visible_range_W[quad][k][i] = index;
            }
            //---------------------------------------
            {
              FOR (index, tile_visibility.size()) {
                if (tile_visibility[index] == v_temp_E) break;
              }
              if (index == tile_visibility.size())
                tile_visibility.push_back(v_temp_E);
              tile_visible_range_E[quad][k][i] = index;
            }
            //---------------------------------------
          }
        }
      }
    }
    cout << tile_shadows.size() << endl;
  }

  
  // ----------------------------------------

  // compute full tile views
  FORS (i, GameConst::VIEW_MAX+1) {
    generate_full_tile_view(i, full_tile_views[i]);

    FORS (j, TS) {
      dfull_tile_views[i][j].alloc(full_tile_views[i], j);
    }
  }

  TileView tv;
  tv.alloc();

  edge_masks = new DualView[pf.get_width()];

  FORS (i, pf.get_width()) {
    tv.fill();
    const sint4 r = GameConst::VIEW_MAX;
    
    uint4 mask = 0xFFFFFFFF;
    if (i + TS-(r+1) >= pf.get_width()) { // if the edge of the bytes goes outside the map
      mask <<= (i + TS-(r+1) + 1) - pf.get_width();
      mask >>= (i + TS-(r+1) + 1) - pf.get_width();
    }
    if (i - r < 0) {
      mask >>= -(i-r);
      mask <<= -(i-r);
    }
    
    for (sint4 y =-r; y <= r; ++y) {
      tv.row_mask(mask, y+GameConst::VIEW_MAX);
    }

    edge_masks[i].alloc(tv, lshift_amount[i]);
  }

  // ----------------------------------------
  // compute vision for every tile - maximize memory allocation
  // ----------------------------------------
#if 0
  FORS (i, pf.get_num()) {
    GameTile &tile = pf(i);
    sint4 dv_offset = lshift_amount[i]; // amount this tile's bitmap is offset within a DualView

    generate_tile_view(i, tile, true); // compute West view
    tile.dtvW.alloc(tile.tvW, dv_offset);
    tile.dtvE.alloc(tile.tvE, dv_offset);
    tile.dknown_tv.alloc(tile.known_tv, dv_offset);

    generate_tile_view(i, tile, false); // compute East view
    tile.dtvW2.alloc(tile.tvW2, dv_offset);
    tile.dtvE2.alloc(tile.tvE2, dv_offset);
    tile.dknown_tv2.alloc(tile.known_tv2, dv_offset);
  }
#endif
}

//===================================================================

void Game::compute_footprint(const GameObj &gob, Vector<sint4> &obj_tiles) const
{
  obj_tiles.clear();

  if (*gob.sod.shape != Object::LINE) {

    if (*gob.sod.x < 0 || *gob.sod.y < 0 ||
        *gob.sod.x / tile_points >= pf.get_width() ||
        *gob.sod.y / tile_points >= pf.get_height()) return;
  }

  Vector< Vec2<sint4> > fp;

  if (*gob.sod.shape == Object::CIRCLE) {
    Object::circle_footprint(*gob.sod.x, *gob.sod.y, *gob.sod.radius, fp, tile_points);
  } else if (*gob.sod.shape == Object::RECTANGLE) {
    Object::rect_footprint(*gob.sod.x1, *gob.sod.y1, *gob.sod.x2, *gob.sod.y2, fp, tile_points);
  } else if (*gob.sod.shape == Object::LINE) {
    Object::line_footprint(*gob.sod.x1, *gob.sod.y1, *gob.sod.x2, *gob.sod.y2, fp, tile_points);
  }

  FORU (i, fp.size()) {
    sint4 x = fp[i].x;
    sint4 y = fp[i].y;
    if (x < 0 || x >= pf.get_width()) continue;
    if (y < 0 || y >= pf.get_height()) continue;
    obj_tiles.push_back(pf.xy2ind(x,y));
  }
}

//===================================================================

/** Generate a pair of vectors for the West and East sides of the
    tiles an object intersects.
*/

void Game::compute_footprint(const GameObj &gob,
                             Vector<sint4> &obj_tiles_W, Vector<sint4> &obj_tiles_E) const
{
  obj_tiles_W.clear();
  obj_tiles_E.clear();

  if (*gob.sod.x < 0 || *gob.sod.y < 0 ||
      *gob.sod.x / tile_points >= pf.get_width() ||
      *gob.sod.y / tile_points >= pf.get_height()) return;


  if (*gob.sod.shape == Object::CIRCLE) {

    sint4 rad = *gob.sod.radius;
    sint4 rr = square(rad);
    sint4 ox = *gob.sod.x;
    sint4 oy = *gob.sod.y;
    
    sint4 tx0 = (ox - rad)/tile_points;
    sint4 tx1 = (ox + rad)/tile_points;
    sint4 ty0 = (oy - rad)/tile_points;
    sint4 ty1 = (oy + rad)/tile_points;

    if (tx0 < 0) tx0 = 0;
    if (tx1 >= pf.get_width()) tx1 = pf.get_width()-1;
    if (ty0 < 0) ty0 = 0;
    if (ty1 >= pf.get_height()) ty1 = pf.get_height()-1;

    sint4 tx = ox/tile_points;
    sint4 ty = oy/tile_points;
    sint4 dx = ox % tile_points;
    sint4 dy = oy % tile_points;

    // The split names may be misleading; they are given as the splits would appear
    // when shown with (0,0) in the upper left (NW).  To use the more intuitive Cartesian
    // coordinates the splits should be reversed.
    // The following comments will assume use of the Cartesian plane.

    sint4 tb_index = dx + dy; // distance from the infinite / line
    sint4 bt_index = dx - dy; // distance from the infinite \ line

    sint4 ry = ty0 - ty;
    for (sint4 y = ty0; y <= ty1; ++y, ++ry) {
      sint4 rx = tx0 - tx;
      for (sint4 x = tx0; x <= tx1; ++x, ++rx) {
        const Tile &t = pf(x,y);
        bool west = false, east = false;
        if (t.get_split() == Tile::NO_SPLIT) {
          if (rx == 0 || ry == 0) west = true;
          else {
            sint4 mx = tile_points * rx - dx;
            sint4 my = tile_points * ry - dy;
            if (rx < 0) mx += tile_points;
            if (ry < 0) my += tile_points;
            sint4 dist = square(mx) + square(my);
            if (dist < rr) west = true;
          }

        } else  if (rx == 0 && ry == 0) {
          if (t.get_split() == Tile::TB_SPLIT) {
            sint4 dist2 = square(dx - dy) / 2;
            if (rr > dist2) {
              east = true;
              west = true;
            } else {
              if (dx > dy) east = true;
              else         west = true; // dx < dy
            }
          } else {
            sint4 dist2 = square(dx + dy - tile_points) / 2;
            if (rr > dist2) {
              east = true;
              west = true;
            } else {
              if (dx > tile_points - dy) east = true;
              else                       west = true; // dx > tile_points - dy
            }
          }
        } else {

          sint4 min_x = tile_points * rx - dx;
          sint4 min_y = tile_points * ry - dy;
          if (rx < 0) min_x += tile_points;
          if (ry < 0) min_y += tile_points;
          sint4 min_dist = square(min_x) + square(min_y);
          if (min_dist < rr) {
            if (t.get_split() == Tile::TB_SPLIT) {
              if (rx == ry) {
                obj_tiles_E.push_back(pf.xy2ind(x,y));
                obj_tiles_W.push_back(pf.xy2ind(x,y));
                continue;
              } else if (rx > ry) { // below '/'
                west = true;
              } else {              // above '/'
                east = true;
              }
              // Indices used as a measure of where along the infinite \ line the corners of the tile are.
              // This determines which corner of the line (in the tile) is closest.
              sint4 i1 = (rx - ry - 1) * tile_points;
              sint4 i2 = (rx - ry + 1) * tile_points;
              if (bt_index <= i1) {
                sint4 mx = rx * tile_points - dx;
                sint4 my = ry * tile_points - dy;
                sint4 dist = square(mx) + square(my);
                if (dist < rr) {
                  if (rx > ry) east = true;
                  else         west = true;
                }
              } else if (bt_index >= i2) {
                sint4 mx = (rx+1) * tile_points - dx;
                sint4 my = (ry+1) * tile_points - dy;
                sint4 dist = square(mx) + square(my);
                if (dist < rr) {
                  if (rx > ry) east = true;
                  else         west = true;
                }
              } else {
                // Closest point lies on the diagonal line
                sint4 dist2 = (rx - ry) * tile_points - dx + dy;
                dist2 = square(dist2) / 2;
                if (dist2 < rr) {
                  if (rx - ry > 0) east = true;
                  else             west = true;
                }
              }
            } else { // Tile::BT_SPLIT
              if (rx == -ry) {
                obj_tiles_E.push_back(pf.xy2ind(x,y));
                obj_tiles_W.push_back(pf.xy2ind(x,y));
                continue;
              } else if (rx > -ry) { // above '\'
                west = true;
              } else {               // below '\'
                east = true;
              }
              sint4 i1 = (rx + ry - 1) * tile_points;
              sint4 i2 = (rx + ry + 1) * tile_points;
              if (tb_index <= i1) {
                sint4 mx = (rx+1) * tile_points - dx;
                sint4 my = ry * tile_points - dy;
                sint4 dist = square(mx) + square(my);
                if (dist < rr) {
                  if (rx + ry > 0) east = true;
                  else             west = true;
                }
              } else if (tb_index >= i2) {
                sint4 mx = rx * tile_points - dx;
                sint4 my = (ry+1) * tile_points - dy;
                sint4 dist = square(mx) + square(my);
                if (dist < rr) {
                  if (rx + ry > 0) east = true;
                  else             west = true;
                }
              } else {
                sint4 dist2 = (rx + ry) * tile_points - dx + dy;
                dist2 = square(dist2) / 2;
                if (dist2 < rr) {
                  if (rx + ry > 0) east = true;
                  else             west = true;
                }
              }
            }
          }
        }
        if (east) obj_tiles_E.push_back(pf.xy2ind(x,y));
        if (west) obj_tiles_W.push_back(pf.xy2ind(x,y));
      }
    }
  } else if (*gob.sod.shape == Object::RECTANGLE) {
    sint4 tx0 = *gob.sod.x1 / tile_points;
    sint4 tx1 = *gob.sod.x2 / tile_points;
    sint4 ty0 = *gob.sod.y1 / tile_points;
    sint4 ty1 = *gob.sod.y2 / tile_points;

    if (tx0 < 0) tx0 = 0;
    if (tx1 >= pf.get_width()) tx1 = pf.get_width()-1;
    if (ty0 < 0) ty0 = 0;
    if (ty1 >= pf.get_height()) ty1 = pf.get_height()-1;

    for (sint4 ty = ty0; ty <= ty1; ++ty) {
      for (sint4 tx = tx0+1; tx < tx1; ++tx) {
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));
      }
    }
    for (sint4 ty = ty0+1; ty < ty1; ++ty) {
      {
        sint4 tx = tx0;
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));
      }
      {
        sint4 tx = tx1;
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));
      }
    }

    // now check the corners

    {
      sint4 tx = tx0;
      sint4 ty = ty1;
      const Tile &t = pf(tx,ty);
      if (t.get_split() == Tile::TB_SPLIT) {

        sint4 dx = *gob.sod.x1 % tile_points;
        sint4 dy = *gob.sod.y1 % tile_points;
        if (dy < tile_points - dx) obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));

      } else {
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));
      }
    }

    {
      sint4 tx = tx1;
      sint4 ty = ty1;
      const Tile &t = pf(tx,ty);
      if (t.get_split() == Tile::BT_SPLIT) {

        sint4 dx = *gob.sod.x2 % tile_points;
        sint4 dy = *gob.sod.y1 % tile_points;
        if (dy < dx) obj_tiles_E.push_back(pf.xy2ind(tx,ty));
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));

      } else {
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));
      }
    }

    {
      sint4 tx = tx0;
      sint4 ty = ty0;
      const Tile &t = pf(tx,ty);
      if (t.get_split() == Tile::BT_SPLIT) {

        sint4 dx = *gob.sod.x1 % tile_points;
        sint4 dy = *gob.sod.y2 % tile_points;
        if (dy > dx) obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));

      } else {
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));
      }
    }

    {
      sint4 tx = tx1;
      sint4 ty = ty0;
      const Tile &t = pf(tx,ty);
      if (t.get_split() == Tile::TB_SPLIT) {
        
        sint4 dx = *gob.sod.x2 % tile_points;
        sint4 dy = *gob.sod.y2 % tile_points;
        if (dy > tile_points - dx) obj_tiles_E.push_back(pf.xy2ind(tx,ty));
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));

      } else {
        obj_tiles_W.push_back(pf.xy2ind(tx,ty));
        obj_tiles_E.push_back(pf.xy2ind(tx,ty));
      }
    }

  }
}

//===================================================================

/** Compute the players that can see an object based on the visibility
    of the tiles the object intersects.
*/

void Game::compute_visibility(GameObj &gob) const
{
  gob.sod.seen_by.clear();

  if (gob.sod.is_shared) {
    FORS (i, player_n) {
      gob.sod.seen_by.set(i);
    }
    return;
  }

  if (*gob.sod.x < 0 || *gob.sod.y < 0 ||
      *gob.sod.x / tile_points >= pf.get_width() ||
      *gob.sod.y / tile_points >= pf.get_height()) return;

  if (*gob.sod.cloaked2 >= 1) return;

  MapView* *mapW = new MapView*[player_n];
  MapView* *mapE = new MapView*[player_n];

  if (*gob.sod.cloaked >= 1) {
    FORS(i, player_n) { mapW[i] = &players[i]->detected_W; mapE[i] = &players[i]->detected_E; }
  } else {
    FORS(i, player_n) { mapW[i] = &players[i]->visible_W;  mapE[i] = &players[i]->visible_E; }
  }

  Vector<sint4> obj_tiles_W;
  Vector<sint4> obj_tiles_E;

  compute_footprint(gob, obj_tiles_W, obj_tiles_E);

  FORU (i, obj_tiles_W.size()) {
    sint4 x = pf.ind2x(obj_tiles_W[i]);
    sint4 y = pf.ind2y(obj_tiles_W[i]);

    FORS (i, player_n) {
      if (mapW[i]->is_set(x, y))
        gob.sod.seen_by.set(i);
    }
  }
  FORU (i, obj_tiles_E.size()) {
    sint4 x = pf.ind2x(obj_tiles_E[i]);
    sint4 y = pf.ind2y(obj_tiles_E[i]);

    FORS(i, player_n) {
      if (mapE[i]->is_set(x, y))
        gob.sod.seen_by.set(i);
    }
  }

  delete [] mapW;
  delete [] mapE;
}

//===================================================================

/** Checks if one object can see another.
*/

bool Game::can_see_obj(const GameObj &a, const GameObj &b)
{
  sint4 x1, y1; // !!!unused, x2, y2;
  sint4 rx, ry;

  const sint4 ax = *a.sod.x;
  const sint4 ay = *a.sod.y;
  // const sint4 bx = *b.sod.x;
  // const sint4 by = *b.sod.y;
  const sint4 sight = *a.sod.sight;

  x1 = ax / tile_points;
  y1 = ay / tile_points;
  // x2 = bx / tile_points;
  // y2 = by / tile_points;

  Vector<sint4> obj_tiles_W;
  Vector<sint4> obj_tiles_E;
  compute_footprint(b, obj_tiles_W, obj_tiles_E);

  GameTile &t = pf(x1,y1);
  sint4 tind = pf.xy2ind(x1,y1);

  // set these values for a check within generate_tile_view
  v1w[tind] = sight;
  v1e[tind] = sight;

  generate_tile_view(tind, t, true);
  if (t.is_split()) generate_tile_view(tind, t, false);

  TileView viewW; // visible West sides of tiles
  TileView viewE;

  TileView *pvW = 0, *pvE = 0;

  if (*a.sod.zcat == Object::IN_AIR) {
    viewW.copy(full_tile_views[sight]);
    viewE.copy(full_tile_views[sight]);
  } else {
    if (t.get_split() != Tile::NO_SPLIT) {
      rx = ax % tile_points;
      ry = ay % tile_points;
      if (t.get_split() == Tile::TB_SPLIT) {
        if (ry < rx) {
          pvW = t.dtvW2.make_TileView();
          pvE = t.dtvE2.make_TileView();
        } else {
          pvW = t.dtvW.make_TileView();
          pvE = t.dtvE.make_TileView();
        }
      } else {
        if (ry > tile_points - rx) {
          pvW = t.dtvW2.make_TileView();
          pvE = t.dtvE2.make_TileView();
        } else {
          pvW = t.dtvW.make_TileView();
          pvE = t.dtvE.make_TileView();
        }
      }
    } else {
      pvW = t.dtvW.make_TileView();
      pvE = t.dtvE.make_TileView();
    }

    viewW.copy(*pvW);
    viewE.copy(*pvE);
    delete pvW;
    delete pvE;


    viewW.view_and(full_tile_views[sight]);
    viewE.view_and(full_tile_views[sight]);
  }

  // the West sides
  FORALL (obj_tiles_W, i) {
    sint4 targ_ind = *i;
    sint4 targ_x = pf.ind2x(targ_ind) - x1 + GameConst::VIEW_MAX;
    sint4 targ_y = pf.ind2y(targ_ind) - y1 + GameConst::VIEW_MAX;
    
    if (targ_x < 0) continue;
    if (targ_y < 0) continue;
    if (targ_x >= GameConst::VIEW_WIDTH) continue;
    if (targ_y >= GameConst::VIEW_WIDTH) continue;
    
    if (viewW.is_set(targ_x, targ_y)) return true;
  }
  
  // now the East sides
  FORALL (obj_tiles_E, i) {
    sint4 targ_ind = *i;
    sint4 targ_x = pf.ind2x(targ_ind) - x1 + GameConst::VIEW_MAX;
    sint4 targ_y = pf.ind2y(targ_ind) - y1 + GameConst::VIEW_MAX;
    
    if (targ_x < 0) continue;
    if (targ_y < 0) continue;
    if (targ_x >= GameConst::VIEW_WIDTH) continue;
    if (targ_y >= GameConst::VIEW_WIDTH) continue;
    
    if (viewW.is_set(targ_x, targ_y)) return true;
  }

  return false;
}

//===================================================================

/** Create a tileview mask for unobstructed visibility of a given radius (i.e. a filled circle).
    The resulting tileview is used to mask out tiles beyond a unit's sight range.
*/

void Game::generate_full_tile_view(sint4 r, TileView &v)
{
  if (r < 0 || r > GameConst::VIEW_MAX) ERR("r out of range");
  v.alloc();
  v.clear();
  
  sint4 d2 = ok_dist2(r);

  for (sint4 y=-r; y <= r; ++y) {
    for (sint4 x=-r; x <= r; ++x) {
      sint4 sq = square(x) + square(y);
      if (sq <= d2) v.set(x+GameConst::VIEW_MAX, y+GameConst::VIEW_MAX);
    }
  }
}

//===================================================================

/** Computes the join allied views of all players, or just one player.
    Shouldn't be called more than once per tick.
*/

void Game::compute_views(sint4 player)
{
  PROFILE("compute views");

  FORS (i, player_n) {
    if (player >= 0 && !x_allied_with_y->get(i, player)) continue;

    PlayerInfo &p = *players[i];

    // store the tiles which are currently known, used to compute the new tiles later
    // copy is actually much faster than swapping pointers (caching?)
    // (check this again)
    p.p_explored.copy(p.explored);

    compute_view(i);
  }

  FORS (i, player_n) {
    if (player >= 0 && i != player) continue;
    PlayerInfo &pi = *players[i];

    FORS (j, player_n) {
      PlayerInfo &pj = *players[j];
      if (i == j) continue;

      if (x_allied_with_y->get(i, j)) {
        pi.visible_W. map_or(pj.visible_W);
        pi.visible_E. map_or(pj.visible_E);
        pi.detected_W.map_or(pj.detected_W);
        pi.detected_W.map_or(pj.detected_E);
        pi.explored.  map_or(pj.explored);
      }
    }
  }
}

//-------------------------------------------------------------------

/** Generate the difference and store the indices in new_tiles.
    Explored is increasing, the difference will always be new tiles.
    This works even if a tile has been invalidated; information will
    only be sent if the player can see the tile.
*/

void Game::get_new_tiles(sint4 player, Vector<sint4> &new_tiles)
{
  PlayerInfo &p = *players[player];
  p.explored.map_whatsnew(p.p_explored, new_tiles);
}

//===================================================================

/** Update a player's mapviews to reflect what is now visible, known, and detected.
    Store the indices of newly explored tiles in a vector.

    Side effects:
    \li Clears and uses Game data structures (v1w,v1e,v2,dv1w,dv1e,dv2) which are
        only used by this function and in generate_tile_view.
    \li Sets the "previous view" mapviews to the current state and updates the new current state.
*/

void Game::compute_view(sint4 player)
{
  assert(player >= 0 && player < GameConst::PLAYER_MAX);
  PlayerInfo &p = *(players[player]);

  p.visible_W.clear();
  p.visible_E.clear();
  p.detected_W.clear();
  p.detected_E.clear();

  // reset vision counters and maximum sight for each tile

  const uint4 n = pf.get_num();
  
  memset(v1w, 0xFF, sizeof(*v1w) * n);
  memset(v1e, 0xFF, sizeof(*v1e) * n);
  memset(v2,  0xFF, sizeof(*v2)  * n);

  memset(dv1w, 0xFF, sizeof(*dv1w) * n); // detected vision ranges
  memset(dv1e, 0xFF, sizeof(*dv1e) * n);
  memset(dv2,  0xFF, sizeof(*dv2)  * n);

  Array<sint4, 3000> used_tiles_w;
  Array<sint4, 3000> used_tiles_e;

  // compute used tiles and maximum sight for each
  
  FORALL (objs[player], it) {

    GameObj *temp = (*it)->get_GameObj();
    if (temp == 0) continue;

    const GameObj &gob = *temp;
    if (!gob.has_attr("sight")) continue;

    const sint4 x = *gob.sod.x;
    const sint4 y = *gob.sod.y;

    const sint4 tx = x / tile_points;
    const sint4 ty = y / tile_points;

    // object isn't on the field
    if (x < 0 || tx >= gtiles_x) continue;
    if (y < 0 || ty >= gtiles_y) continue;

    sint4 tind = pf.xy2ind(tx, ty);

    sint1 *p_v1;
    sint1 *p_dv1;

    const sint2 zcat = *gob.sod.zcat;
    const sint2 sight = *gob.sod.sight;
    const bool detector = *gob.sod.detector;

    if (sight <= 0) continue;

    if (zcat == Object::IN_AIR) {
      if (sight > v2[tind])
        v2[tind] = sight;
      if (detector)
        if (sight > dv2[tind])
          dv2[tind] = sight;
      if (v1w[tind] < 0 && v1e[tind] < 0) {
        v1w[tind] = 0;
        v1e[tind] = 0;
        used_tiles_w.push_back(tind);
      }
    } else {

      if (pf(tind).get_split()) { // the tile is split, perform an unfortunate test

        sint4 rx = x % tile_points;
        sint4 ry = y % tile_points;
        
        if (pf(tind).get_split() == Tile::TB_SPLIT) {
          if (ry < rx) {
            p_v1 = v1e;
            p_dv1 = dv1e;
          } else {
            p_v1 = v1w;
            p_dv1 = dv1w;
          }

        } else {

          if (ry > tile_points - rx) {
            p_v1 = v1e;
            p_dv1 = dv1e;
          } else {
            p_v1 = v1w;
            p_dv1 = dv1w;
          }
        }

        if (p_v1 == v1w) {

          if (v1w[tind] < 0) {
            v1w[tind] = 0; // save tile index only once
            used_tiles_w.push_back(tind);
          }

        } else {

          if (v1e[tind] < 0) {
            v1e[tind] = 0; // save tile index only once
            used_tiles_e.push_back(tind);
          }
        }
        
      } else {
        p_v1 = v1w;
        p_dv1 = dv1w;
        
        if (v1w[tind] < 0) {
          v1w[tind] = 0; // save tile index only once
          used_tiles_w.push_back(tind);
        }
      }

      if (sight > p_v1[tind])
        p_v1[tind] = sight;
      if (detector)
        if (sight > p_dv1[tind])
          p_dv1[tind] = sight;
    }
  }

  /////////////////////

  // set flags for visible tiles (maximum sight)

  FORS (i, used_tiles_w.size()) {
    sint4 tind = used_tiles_w[i];
    sint4 tx = pf.ind2x(tind);
    sint4 dv_offset = lshift_amount[tind]; // amount this tile's bitmap is offset within a DualView
    GameTile &tile = pf(tind);

    generate_tile_view(tind, tile, true); // checks if necessary and updates tile info
    
    DualView dual_vW, dual_dW, dual_vE, dual_dE;
    DualView dual_k;
    const DualView &mask = edge_masks[tx]; // stop the view from wrapping around
    
    // fixme: switch dfull_tile_views indices for faster access?
    if (v1w[tind] > v2[tind]) {
      dual_vW.alloc(tile.dtvW);
      dual_vE.alloc(tile.dtvE);
      dual_k.alloc(tile.dknown_tv);
      dual_vW.view_and(dfull_tile_views[v1w[tind]][dv_offset]);
      dual_vE.view_and(dfull_tile_views[v1w[tind]][dv_offset]);
      dual_k.view_and(dfull_tile_views[v1w[tind]][dv_offset]);
      if (v2[tind] > 0) {
        dual_vW.view_or(dfull_tile_views[v2[tind]][dv_offset]);
        dual_vE.view_or(dfull_tile_views[v2[tind]][dv_offset]);
        dual_k.view_or(dfull_tile_views[v2[tind]][dv_offset]);
        dual_vW.view_and(mask);
        dual_vE.view_and(mask);
        dual_k.view_and(mask);
      }
    } else if (v2[tind] > 0) {
      dual_vW.alloc(dfull_tile_views[v2[tind]][dv_offset]);
      dual_vE.alloc(dfull_tile_views[v2[tind]][dv_offset]);
      dual_k.alloc(dfull_tile_views[v2[tind]][dv_offset]);
      dual_vW.view_and(mask);
      dual_vE.view_and(mask);
      dual_k.view_and(mask);
    }

    if (dv1w[tind] > dv2[tind]) {
      dual_dW.alloc(tile.dtvW);
      dual_dE.alloc(tile.dtvE);
      dual_dW.view_and(dfull_tile_views[dv1w[tind]][dv_offset]);
      dual_dE.view_and(dfull_tile_views[dv1w[tind]][dv_offset]);
      if (dv2[tind] > 0) {
        dual_dW.view_or(dfull_tile_views[dv2[tind]][dv_offset]);
        dual_dE.view_or(dfull_tile_views[dv2[tind]][dv_offset]);
        dual_dW.view_and(mask);
        dual_dE.view_and(mask);
      }
    } else if (dv2[tind] > 0) {
      dual_dW.alloc(dfull_tile_views[dv2[tind]][dv_offset]);
      dual_dE.alloc(dfull_tile_views[dv2[tind]][dv_offset]);
      dual_dW.view_and(mask);
      dual_dE.view_and(mask);
    }

    p.explored.view_or(dual_k, tind);

    p.visible_W.view_or(dual_vW, tind);
    p.visible_E.view_or(dual_vE, tind);

    if (dv1w[tind] > 0 || dv2[tind] > 0) {
      p.detected_W.view_or(dual_dW, tind);
      p.detected_E.view_or(dual_dE, tind);
    }
  }

  /////////////////////

  FORS (i, used_tiles_e.size()) {
    sint4 tind = used_tiles_e[i];
    sint4 tx = pf.ind2x(tind);
    sint4 dv_offset = lshift_amount[tind]; // amount this tile's bitmap is offset within a DualView
    GameTile &tile = pf(tind);

    generate_tile_view(tind, tile, false); // checks if necessary and updates tile info

    DualView dual_vW, dual_dW, dual_vE, dual_dE;
    DualView dual_k;
    const DualView &mask = edge_masks[tx]; // stop the view from wrapping around

    // fixme: switch dfull_tile_views indices for faster access?
    if (v1e[tind] > v2[tind]) {
      dual_vW.alloc(tile.dtvW2);
      dual_vE.alloc(tile.dtvE2);
      dual_k.alloc(tile.dknown_tv2);
      dual_vW.view_and(dfull_tile_views[v1e[tind]][dv_offset]);
      dual_vE.view_and(dfull_tile_views[v1e[tind]][dv_offset]);
      dual_k.view_and(dfull_tile_views[v1e[tind]][dv_offset]);
      if (v2[tind] > 0) {
        dual_vW.view_or(dfull_tile_views[v2[tind]][dv_offset]);
        dual_vE.view_or(dfull_tile_views[v2[tind]][dv_offset]);
        dual_k.view_or(dfull_tile_views[v2[tind]][dv_offset]);
        dual_vW.view_and(mask);
        dual_vE.view_and(mask);
        dual_k.view_and(mask);
      }
    } else if (v2[tind] > 0) {
      dual_vW.alloc(dfull_tile_views[v2[tind]][dv_offset]);
      dual_vE.alloc(dfull_tile_views[v2[tind]][dv_offset]);
      dual_k.alloc(dfull_tile_views[v2[tind]][dv_offset]);
      dual_vW.view_and(mask);
      dual_vE.view_and(mask);
      dual_k.view_and(mask);
    }

    if (dv1e[tind] > dv2[tind]) {
      dual_dW.alloc(tile.dtvW2);
      dual_dE.alloc(tile.dtvE2);
      dual_dW.view_and(dfull_tile_views[dv1e[tind]][dv_offset]);
      dual_dE.view_and(dfull_tile_views[dv1e[tind]][dv_offset]);
      if (dv2[tind] > 0) {
        dual_dW.view_or(dfull_tile_views[dv2[tind]][dv_offset]);
        dual_dE.view_or(dfull_tile_views[dv2[tind]][dv_offset]);
        dual_dW.view_and(mask);
        dual_dE.view_and(mask);
      }
    } else if (dv2[tind] > 0) {
      dual_dW.alloc(dfull_tile_views[dv2[tind]][dv_offset]);
      dual_dE.alloc(dfull_tile_views[dv2[tind]][dv_offset]);
      dual_dW.view_and(mask);
      dual_dE.view_and(mask);
    }

    p.explored.view_or(dual_k, tind);

    p.visible_W.view_or(dual_vW, tind);
    p.visible_E.view_or(dual_vE, tind);

    if (dv1e[tind] > 0 || dv2[tind] > 0) {
      p.detected_W.view_or(dual_dW, tind);
      p.detected_E.view_or(dual_dE, tind);
    }
  }

  if (opt.nfog) {
    p.visible_W.fill();
    p.visible_E.fill();
    p.explored.fill();
  }
}

//===================================================================

/*
\ 2 /
 \ /
3 X 1
 / \
/ 4 \


-1  0  1
  \ | /
   \|/
 0--X--0
   /|\
  / | \
 1  0 -1

To generate the tile view it makes sense to note the places where a transition
occurs between visible and occluded sections (i.e. ranges of angles or slopes
that can be seen at a certain radius from the tile's center).  Because tiles
are currently the only objects that can obscure vision, the only places where
such a transition can occur is at the corners of a tile.  The set of all
angles at which such a transition can occur is then the set of angles that
pass through all grid line intersections within the maximum visible radius.

The visual region can be considered to be the set of slices whose boundaries are
these angles.  Both the current visual region and the visual field taken up by
tiles are represented by an array of bits indicating the state of each slice.
This allows for fast bit operations to determine intersections and mask out
obscured regions.

Starting off with an unobscured field of view, tiles are processed in an
increasing radius from the tile whose view is being computed.  First, the range
of angles that the tile takes up within the visual field is checked against
the currently visible ranges
to determine whether the tile is "known".  Then, if the tile is an obscuring
tile, its precomputed range of angles is marked as obscured.  Finally, similar
to the first step, the ranges that mark the center of the tile are checked
to determine if the tile is "visible".

Instead of a strictly distance-ordered traversal of the surrounding tiles,
it is sufficient to proceed by working outward in a sequence of hollow squares,
providing that these squares are in turn traversed outward from the axis.
In this way, of two tiles having a common visible region, the closer tile will
be processed first.

 <-- ^ --> 
^    |    ^
|    |    |
<----o---->
|    |    |
v    |    v
 <-- v -->

Traversing the tiles in the manner leads to a natural division of the visible
region into quadrants divided along the diagonals.  Considering the slope
to be mirrored about the diagonals allows for the same field of view description
to be used to define tiles along the diagonal regardless of which quadrant is
currently being computed.

*/

/** Update which tiles surrounding it a tile can see, based on the height of
    the neighbouring terrain. */

void Game::generate_tile_view(sint4 tind, GameTile &tile, bool West)
{
  assert(tind >= 0 && tind < pf.get_num());
  assert(West || tile.is_split());

  if (West) {
    if (v1w[tind] <= tile.v1w && tile.dtvW.defined()) return;
  } else {
    if (v1e[tind] <= tile.v1e && tile.dtvE.defined()) return;
  }

  // which side of the tile the view is being computed for

  TileView vW;
  TileView vE;
  TileView kv;

  vW.alloc(); vW.clear();
  vE.alloc(); vE.clear();
  kv.alloc(); kv.clear();

  const sint4 r = GameConst::VIEW_MAX;

  // mark the current (center) square known
  kv.set(r,r);

  // splitting into specific quadrants allows for more aggressive termination when hitting map boundaries
  // fixme: 4 adjacent tiles checked, perform more general pruning for entire quadrant being occluded

  if (tile.is_split()) { // the tile is composed of two separate triangles (possibly connected)

    sint4 max_z;
    if (West) {
      max_z = tile.get_vis_h1() + z_sight;
      vW.set(r,r);
    } else {
      max_z = tile.get_vis_h2() + z_sight;
      vE.set(r,r);
    }

    if (tile.get_split() == Tile::TB_SPLIT) {     //  |\|
      if (West) {
        gen_tile_view_W(tind, max_z, vW, vE, kv);
        gen_tile_view_S(tind, max_z, vW, vE, kv);
        if (max_z >= min(tile.get_height(0), tile.get_height(3))) {
          vE.set(r,r);
          if (max_z >= min(tile.get_height(2), tile.get_height(3))) // can see past the diagonal wall
            gen_tile_view_E(tind, max_z, vW, vE, kv);
          if (max_z >= min(tile.get_height(0), tile.get_height(1)))
            gen_tile_view_N(tind, max_z, vW, vE, kv);
        }
      } else {
        gen_tile_view_E(tind, max_z, vW, vE, kv);
        gen_tile_view_N(tind, max_z, vW, vE, kv);
        if (max_z >= min(tile.get_height(7), tile.get_height(4))) { // can see past the diagonal wall
          vW.set(r,r);
          if (max_z >= min(tile.get_height(6), tile.get_height(7)))
            gen_tile_view_W(tind, max_z, vW, vE, kv);
          if (max_z >= min(tile.get_height(4), tile.get_height(5)))
            gen_tile_view_S(tind, max_z, vW, vE, kv);
        }
      }
    } else {                   //  |/|
      if (West) {
        gen_tile_view_W(tind, max_z, vW, vE, kv);
        gen_tile_view_N(tind, max_z, vW, vE, kv);
        if (max_z >= min(tile.get_height(2), tile.get_height(5))) {
          vE.set(r,r);
          if (max_z >= min(tile.get_height(2), tile.get_height(3)))
            gen_tile_view_E(tind, max_z, vW, vE, kv);
          if (max_z >= min(tile.get_height(4), tile.get_height(5)))
            gen_tile_view_S(tind, max_z, vW, vE, kv);
        }
      } else {
        gen_tile_view_E(tind, max_z, vW, vE, kv);
        gen_tile_view_S(tind, max_z, vW, vE, kv);
        if (max_z >= min(tile.get_height(1), tile.get_height(6))) {
          vW.set(r,r);
          if (max_z >= min(tile.get_height(6), tile.get_height(7)))
            gen_tile_view_W(tind, max_z, vW, vE, kv);
          if (max_z >= min(tile.get_height(0), tile.get_height(1)))
            gen_tile_view_N(tind, max_z, vW, vE, kv);
        }
      }
    }

  } else {
    vW.set(r,r); // the tile isn't split so only this value will be checked

    sint4 max_z = tile.get_max_h() + z_sight; // assume visiblity from the highest point

    gen_tile_view_E(tind, max_z, vW, vE, kv);
    gen_tile_view_S(tind, max_z, vW, vE, kv);
    gen_tile_view_W(tind, max_z, vW, vE, kv);
    gen_tile_view_N(tind, max_z, vW, vE, kv);
  }

  sint4 lshift = lshift_amount[tind];
  if (West) {
    tile.dtvW.alloc(vW, lshift);
    tile.dtvE.alloc(vE, lshift);
    tile.dknown_tv.alloc(kv, lshift);
    tile.v1w = v1w[tind];
  } else {
    tile.dtvW2.alloc(vW, lshift);
    tile.dtvE2.alloc(vE, lshift);
    tile.dknown_tv2.alloc(kv, lshift);
    tile.v1e = v1e[tind];
  }
}

/** Determines what parts of a tile located at some offset are visible from a certain z-height.
   @param tile  The tile whose visibility is being determined.
   @param vW    Visible Western/full halves.
   @param vE    Visible Eastern halves.
   @param kv    Known view.
   @param q     The quadrant that the tile is in, for the purposes of the vision computation.
   @param vx    x location of tile
   @param vy    y location of tile
   @param max_z The height from which tile is being viewed.
 */

inline void Game::check_tile(GameTile &tile, TileView &vW, TileView &vE, TileView &kv,
                             sint4 q, sint4 vx, sint4 vy, sint4 max_z)
{
  sint4 SP = tile.shadow_profile(max_z);
  
  sint4 t = vx + vy*(GameConst::VIEW_WIDTH);
  
  if (SP < 0) { // the entire tile is low enough to see over
    if (qstrip.intersects(tile_f_visible_range[t])) {
      vW.set(vx, vy);
      if (tile.is_split()) vE.set(vx, vy);
      kv.set(vx, vy);
    } else if (qstrip.intersects(tile_known_range[t])) {
      kv.set(vx, vy);
    }
  } else if (qstrip.intersects(tile_known_range[t])) { // some part of the tile is visible
    kv.set(vx, vy);
    
    if (SP < 12) { // the tile isn't too high to see over
      sint4 vis_index = tile_visible_range_W[q][t][SP];
      sint4 shadow_index = tile_shadow_range[q][t][SP];
      const BitStrip &vis = tile_visibility[vis_index];
      const BitStrip &shadow = tile_shadows[shadow_index];
      if (qstrip.intersects(vis)) vW.set(vx, vy);
      
      if (tile.is_split()) {
        sint4 vis_index_E = tile_visible_range_E[q][t][SP];
        const BitStrip &vis_E = tile_visibility[vis_index_E];
        if (qstrip.intersects(vis_E)) vE.set(vx, vy);
      }
      
      qstrip.and_strip(shadow);
    } else {
      qstrip.and_strip(tile_fshadow_range[t]);
    }
  }
}

inline void Game::gen_tile_view_E(sint4 tind, sint4 max_z,
                                  TileView &vW, TileView &vE, TileView &kv)
{
  sint4 tx = pf.ind2x(tind);
  sint4 ty = pf.ind2y(tind);

  if (tx == pf.get_width()-1) return;

  const sint4 r = GameConst::VIEW_MAX;
  sint4 r1 = ::min(r, pf.get_width()-1 - tx);

  sint4 X, Y;

  qstrip.fill();

  if (pf(tx+1,ty).get_min_h() > max_z) {
    kv.set(r+1,r);
    return;
  }

  for (sint4 x = 1; x <= r1; ++x) { // move the radius outward
    X = tx+x;
    for (sint4 y = 0; y < fv_yc[x]; ++y) { // distance from the axis
      Y = ty+y;
      if (Y == pf.get_height()) break;

      GameTile &tile = pf(X,Y);
      sint4 vx = x+r;
      sint4 vy = y+r;
      
      check_tile(tile, vW, vE, kv, 0, vx, vy, max_z);
    }

    for (sint4 y = -1; y > -fv_yc[x]; --y) { // distance from axis, moving in other direction
      Y = ty+y;
      if (Y < 0) break;

      GameTile &tile = pf(X,Y);
      sint4 vx = x+r;
      sint4 vy = y+r;
      
      check_tile(tile, vW, vE, kv, 0, vx, vy, max_z);
    }
  }
}

inline void Game::gen_tile_view_S(sint4 tind, sint4 max_z,
                                  TileView &vW, TileView &vE, TileView &kv)
{
  sint4 tx = pf.ind2x(tind);
  sint4 ty = pf.ind2y(tind);

  if (ty == pf.get_height()-1) return;

  const sint4 r = GameConst::VIEW_MAX;
  sint4 r2 = ::min(r, pf.get_height()-1 - ty);

  sint4 X, Y;

  qstrip.fill();

  if (pf(tx,ty+1).get_min_h() > max_z) {
    kv.set(r,r+1);
    return;
  }

  for (sint4 x = 1; x <= r2; ++x) { // move the radius outward
    Y = ty+x;
    for (sint4 y = 0; y < fv_yc[x]; ++y) { // distance from the axis
      X = tx+y;
      if (X == pf.get_width()) break;
      
      GameTile &tile = pf(X,Y);
      sint4 vx = y+r;
      sint4 vy = x+r;
      
      check_tile(tile, vW, vE, kv, 1, vx, vy, max_z);
    }

    for (sint4 y = -1; y > -fv_yc[x]; --y) {
      X = tx+y;
      if (X < 0) break;
      
      GameTile &tile = pf(X,Y);
      sint4 vx = y+r;
      sint4 vy = x+r;

      check_tile(tile, vW, vE, kv, 1, vx, vy, max_z);
    }
  }
}

inline void Game::gen_tile_view_W(sint4 tind, sint4 max_z,
                                  TileView &vW, TileView &vE, TileView &kv)
{
  sint4 tx = pf.ind2x(tind);
  sint4 ty = pf.ind2y(tind);

  if (tx == 0) return;

  const sint4 r = GameConst::VIEW_MAX;
  sint4 r3 = ::min(r, tx);

  sint4 X, Y;

  qstrip.fill();

  if (pf(tx-1,ty).get_min_h() > max_z) {
    kv.set(r-1,r);
    return;
  }

  for (sint4 x = 1; x <= r3; ++x) { // move the radius outward
    X = tx-x;
    for (sint4 y = 0; y < fv_yc[x]; ++y) { // distance from the axis
      Y = ty-y;
      if (Y < 0) break;
      
      GameTile &tile = pf(X,Y);
      sint4 vx = r-x;
      sint4 vy = r-y;
      
      check_tile(tile, vW, vE, kv, 2, vx, vy, max_z);
    }
    
    for (sint4 y = -1; y > -fv_yc[x]; --y) {
      Y = ty-y;
      if (Y == pf.get_height()) break;

      GameTile &tile = pf(X,Y);
      sint4 vx = r-x;
      sint4 vy = r-y;
      
      check_tile(tile, vW, vE, kv, 2, vx, vy, max_z);
    }
  }
}

inline void Game::gen_tile_view_N(sint4 tind, sint4 max_z,
                                  TileView &vW, TileView &vE, TileView &kv)
{
  sint4 tx = pf.ind2x(tind);
  sint4 ty = pf.ind2y(tind);

  if (ty == 0) return;

  const sint4 r = GameConst::VIEW_MAX;
  sint4 r4 = ::min(r, ty);

  sint4 X, Y;

  qstrip.fill();

  if (pf(tx,ty-1).get_min_h() > max_z) {
    kv.set(r,r-1);
    return;
  }
 
  for (sint4 x = 1; x <= r4; ++x) { // move the radius outward
    Y = ty-x;
    for (sint4 y = 0; y < fv_yc[x]; ++y) { // distance from the axis
      X = tx-y;
      if (X < 0) break;
 
      GameTile &tile = pf(X,Y);
      sint4 vx = r-y;
      sint4 vy = r-x; 
      
      check_tile(tile, vW, vE, kv, 3, vx, vy, max_z);
    }
    
    for (sint4 y = -1; y > -fv_yc[x]; --y) {
      X = tx-y;
      if (X == pf.get_width()) break;
      
      GameTile &tile = pf(X,Y);
      sint4 vx = r-y;
      sint4 vy = r-x;
      
      check_tile(tile, vW, vE, kv, 3, vx, vy, max_z);
    }
  }
} 

//==================================================================


