/** @file View2.C
    Creating and applying view differences.

    @see Game.H

    $Id: View2.C 8267 2009-04-28 02:17:38Z furtak $
    $Source$
*/

// This is an ORTS file (c) Michael Buro, Tim Furtak, licensed under the GPL

#include "Global.H"
#include "Game.H"
#include "GameObj.H"
#include "GameChanges.H"
#include "ServerObjData.H"
#include "GameTile.H"
#include "BoolMatrix.H"
#include "PlayerInfo.H"
#include "GameObj.H"
#include "GameObjFactory.H"
#include "ByteData.H"

using namespace std;

#define TEST 0

//===================================================================

/** Compares pointers to game objects by their blueprint code.
    Used for sorting prior to transmission.
*/

struct GameObjCodeCmp {

  bool operator()(GameObj * const &p1, GameObj * const &p2) {
    return p1->code() < p2->code();
  }

};

//-------------------------------------------------------------------

void Game::gen_view_diffs(Vector<string> &views)
{
  views.clear();
  sint4 num_views = player_n + watchers.size();
  FORS (i, num_views) views.push_back("");
  
  // copy previous visibility flags

  FORS (p, player_n+1) {
    FORALL (objs[p], i) {
      GameObj *gob = (*i)->get_GameObj();
      if (!gob) continue;
      gob->sod.prev_seen_by = gob->sod.seen_by;
    }
  }

  // currently visible tiles

  Vector<sint4> *new_tiles = new Vector<sint4>[num_views];

  compute_views();
  FORS (i, player_n) {
    get_new_tiles(i, new_tiles[i]);
  }

  FORS (i, watchers.size()) {
    PlayerInfo &p = *watchers[i];
    p.p_explored.copy(p.explored);
    uint4 vis = watcher_visibility[i];
    if (vis == 0) {
      p.explored.fill();
    } else {
      FORS (i, player_n) {
        if ((vis & (1 << i)) == 0) continue;
        p.explored.map_or(players[i]->explored);
      }
    }
    p.explored.map_whatsnew(p.p_explored, new_tiles[player_n+i]);
  }

  // current visibility flags

  FORS (p, player_n+1) {
    PlayerSet seen_by;
    seen_by.set(p);
    FORS (i, player_n) {
      // if (x_shares_view_with_y->get(p,i)) seen_by.set(i);
      if (x_allied_with_y->get(p,i)) seen_by.set(i);
    }
    
    ObjCont &obj = objs[p];
    FORALL (obj, i) {
      GameObj *gob = (*i)->get_GameObj();
      if (!gob) continue;
      compute_visibility(*gob);
      gob->sod.seen_by.unite(seen_by);
    }
  }
  /*
  FORALL (objs[player_n], i) { // visibility of NPCs
    GameObj *gob = (*i)->get_GameObj();
    if (!gob) continue;
    compute_visibility(*gob);
  }
  */

  // new visible objects

  ObjCont cobjs;   // pointers to objects known to client
  ObjCont del;     // pointers to objects no longer visible to client
  
  //  FORS (p, player_n) {
  FORS (p, num_views) {

    PlayerInfo *pi = p < player_n ? players[p] : watchers[p-player_n];

    // compress list of new tiles:
    // 1. sort tile indexes
    // 2. gap: gap-length-1
    // 3. run: <tile-description>+ 
    
    sort(new_tiles[p].begin(), new_tiles[p].end());
    sint4 next = 0;
    string &s = views[p];

    FORALL (new_tiles[p], it) {
      sint4 pos = *it;
      write_b_int4(s, next-pos); // offset from the expected tile position ( value is <= 0 )

      GameTile &t = pf(pos);
      t.write(s);

      next = pos+1;
    }
    s += char(0x40); // end of section
    
    // count new visible objects and create new cids for them after sorting

    Vector<GameObj*> new_objs;
    
    if (p < player_n) {
      FORS (q, player_n+1) {
        FORALL (objs[q], i) {
          GameObj *gob = (*i)->get_GameObj();
          if (!gob) continue;
          if (gob->sod.prev_seen_by.is_set(p) || !gob->sod.seen_by.is_set(p)) continue;
          if (gob->is_dead()) continue;
          if (gob->sod.is_global) continue;
          new_objs.push_back(gob);
        }
      }
    } else { // watcher
      PlayerInfo &w = *watchers[p-player_n];
      uint4 vis = watcher_visibility[p-player_n];
      bool first_view = new_watcher[p-player_n];

      FORS (q, player_n+1) {
        FORALL (objs[q], i) {
          GameObj *gob = (*i)->get_GameObj();
          if (!gob) continue;
          if (gob->is_dead()) continue;
          if (gob->sod.is_global) continue;
          if (w.get_id(gob) != 0) continue;
          bool prev = false;
          bool now = false;
          FORS (r, player_n+1) {
            if (vis && ((vis & (1 << r)) == 0)) continue;
            prev |= gob->sod.prev_seen_by.is_set(r);
            now |= gob->sod.seen_by.is_set(r);
          }
          prev &= !first_view;
          if (prev || !now) continue;
          new_objs.push_back(gob);
        }
      }
    }

    // sort new object points w.r.t. blueprint code
    sort(new_objs.begin(), new_objs.end(), GameObjCodeCmp());

    // assign client ids to new objects
    FORALL (new_objs, i) pi->new_id(*i);
    

    // create list of new objects in the form <blueprint code, owner+, -1>+
    // terminate list by -1
    
    if (new_objs.size() > 0) {
      uint4 last_code = new_objs[0]->code();
      write_b_int4(s, last_code);
      FORU (i, new_objs.size()) {
        GameObj *gob = new_objs[i];
        if (gob->code() != last_code) {
          write_b_int4(s, -1);
          last_code = gob->code();
          write_b_int4(s, last_code);
        }
        sint4 owner = *gob->sod.owner;
        if (owner < 0) owner = player_n;
        write_b_int4(s, owner);
      }
      write_b_int4(s, -1);
    }
    s += char(0x41); // -1

    // get sorted object pointers
    pi->objs(cobjs);
    del.clear();

    // send updates, collect objects that are no longer visible
    
    bool first_view = false;
    if (p >= player_n) {
      first_view = new_watcher[p-player_n];
      new_watcher[p-player_n] = 0;
    }

    FORALL (cobjs, i) {

      GameObj *gob2 = (*i)->get_GameObj();
      if (gob2 == 0) {
        cout << "non GameObj object" << endl;
        continue;
      }
      GameObj &gob = *gob2;

      bool prev = false, now = false, allied = false;

      if (p < player_n) {
        prev = gob.sod.prev_seen_by.is_set(p);
        now = gob.sod.seen_by.is_set(p);
        allied = x_allied_with_y->get(p, *gob.sod.owner);
        allied |= gob.sod.is_shared;
      } else { // watcher
        uint4 vis = watcher_visibility[p-player_n];
        allied = gob.sod.is_shared;
        FORS (r, player_n+1) {
          if (vis && ((vis & (1 << r)) == 0)) continue;
          prev |= gob.sod.prev_seen_by.is_set(r);
          now  |= gob.sod.seen_by.is_set(r);
          allied |= x_allied_with_y->get(r, *gob.sod.owner);
        }

        prev &= !first_view;
      }


      if (prev) {

        if (gob.is_dead()) {
          s += char(0x05); // 5 (dead)
          del.push_back(&gob);
        } else if (!now) {
          s += char(0x04); // 4 (no longer visible)
          del.push_back(&gob);  
        } else {
          string t;
          // update (line starts with "0")
          gob.gen_diff(false, allied, t, pi);
          s += t; 
        }

      } else if (now) {

        if (gob.is_dead()) {
          // because the object is dead it is no longer in objs[],
          // so seen_by was never copied to prev_seen_by --
          // the object WAS actually seen previously.

          s += char(0x05);
          del.push_back(&gob);

        } else {

          // new (line starts with "1")
          string t;
          gob.gen_diff(true, allied, t, pi);
          s += t;
        }
      }
    }

    // delete cids of invisible or dead objects
    FORALL (del, j) pi->delete_obj(*j, tick_count);
    s += char(0x07); // 7 (end of updates)
  }

  // copy object attributes

  FORS (p, player_n+1) {
    ObjCont &obj = objs[p];
    FORALL (obj, i) {
      (*i)->copy_to_prev();
      (*i)->reset_frame_data(); // single frame data such as attack direction
    }
  }

  delete [] new_tiles;
}

//===================================================================

/** Merge changes from the server with the current perceived state.
    Summarize the differences in changes.
    Previous attribute values are maintained by GameObj::apply_diff()
    @bug The call to apply_diff assumes that sod.owner is already set,
         new objects can break this.
*/

void Game::apply_view_diff(const string &view, GameChanges &changes, ActionChanges &act_changes)
{
  ++tick_count; // since there is no simulation_step in the client
  
  ByteStream bs(view);

  //  cout << "VIEW" << endl << view << endl;

  changes.clear();
  act_changes.clear();
  
  //------------------- frame info --------------------  

  view_frame      = bs.read_b_int4(); // current simulation frame
  action_id       = bs.read_b_int4(); // id of action executed (-1 if none)
  skipped_actions = bs.read_b_int4(); // # of skipped actions
  merged_actions  = bs.read_b_int4(); // # of merged actions
  save_and_quit   = bs.read_b_int4(); // save and quit flag
  freq            = bs.read_b_int4()/1024.0; // server simulation frequency

  if (save_and_quit == -1) {
    game_over = true;
  }
  
  if (save_and_quit == 1) {
    cout << "SAVE REQUEST" << endl;
    ostringstream oss;
    save(oss);
    string file = get_savefile();
    ofstream savefile(file.c_str(), ios::binary);
    savefile.write(oss.str().c_str(), oss.str().size());
    cout << "SAVING TO " << file << " " << oss.str().size() << endl;
    return;
  }

  //------------------- new tiles --------------------

  {
    sint4 pos = 0;
    
    while (bs.good()) {
      if (bs.peek_b_char() == 0x40) { bs.advance(); break; }
      
      sint4 index = bs.read_b_int4();
      pos -= index;
      
      pf(pos).read_b(bs);
      update_tile(pos, &changes.new_boundaries);
      
      cplayer->explored.set(pf.ind2x(pos), pf.ind2y(pos));
      
      changes.new_tile_indexes.push_back(pos);
      ++pos;
    }
  }

  // ------------------- new objects --------------------

  while (bs.good()) {

    sint4 c, p;  // code, owner
    
    c = bs.read_b_int4();
    if (c < 0) {
      if (c != -1) ERR("update message corrupt 1");
      break;
    }

    FOREVER {
      p = bs.read_b_int4();
      if (p < 0) {
        if (p != -1) ERR("update message corrupt 2");
        break;
      }

      ScriptObj *gob = players[p]->gof.create(c);
      changes.new_objs.push_back(gob);
      cplayer->new_id(gob);

      objs[p].push_back(gob);

      GameObj *g = gob->get_GameObj();
      if (g) {
        g->sod.in_game = true;
        g->set_int("owner", p, false);
      } else {
        ERR("non GameObj object");
      }
    }
  }

  // ------------------- update objects --------------------

  ObjCont all_objs, remove;

  cplayer->objs(all_objs);

  FORALL (all_objs, it) {

    char q;

    q = bs.read_b_char(); //read_b_int4(s);
    assert(abs(q) < 64);

    if (q == 5) { // object dead
      changes.dead_objs.push_back(*it);
      (*it)->kill();
      remove.push_back(*it);

    } else if (q == 4) { // object no longer visible
      changes.vanished_objs.push_back(*it);
      remove.push_back(*it);

    } else if (q >= 0 && q <= 3) {
      bs.advance(-1); // the first/allied flags should be passed to apply_diff()
      if ((*it)->get_int("owner") < 0) REM("owner < 0");

      if ((*it)->apply_diff(bs, cplayer))
        changes.changed_objs.push_back(*it);

    } else ERR("corrupt update line");
 
  }

  if (bs.read_b_char() != 7) { // EOM
    ERR("wrong # of update messages");
  }

  // -------------------------------------------------------

  // add new objects to their respective obj vector

  /*
  FORALL (changes.new_objs, it) {
    sint4 owner = (*it)->get_int("owner");
    if (owner < 0) owner = player_n;
    assert(owner <= player_n);
    objs[owner].push_back(*it);
    //    objs[owner].push_back(const_cast<GameObj*>(*it));
  }
  */
  
  // free ids of removed objects
  // delete removed objects from obj containers

  FORALL (remove, it) {

    // fixme: TIM
    //    if (!(*it)->get_int("static", false)) (*it)->destruct();

    cplayer->delete_obj(*it, tick_count);
    sint4 owner = (*it)->get_int("owner");
    if (owner < 0) owner = player_n;
    // fixme: finding objects should be faster
    FIND (objs[owner], j, *it);
    if (j == objs[owner].end()) ERR("obj not found");
    objs[owner].erase(j);
  }

  update_sectors();
  
  // new: update action queues

  if (get_action_frame() >= 0) {

    // remove all pending actions with frame_sent <= get_action_frame()

    // Example:
    //             | 3 skipped |   4 merged   |
    // msg num:  2   3   4   5   6   7   8   9
    //           ^           ^               ^
    //                                       E = id of last action 
    //                                           msg executed = 
    //                                           get_action_frame() = 9
    //                       S = E - get_merged_actions() = 9 - 4
    //           P = S - get_skipped_actions() = 5 - 3

    // E[xecuted] S[kipped] P[revious]
    const sint4 E = get_action_frame();
    const sint4 S = E - get_merged_actions();
    const sint4 P = S - get_skipped_actions();

#if TEST
    cout << " vf:" << get_view_frame()    
         << " af:" << get_action_frame()
         << " sa:" << get_skipped_actions()
         << " ma:" << get_merged_actions()
         << endl;
#endif
    
    if (opt.watch < 0) //opt.cpind < player_n)
    FORALL (objs[opt.cpind], i) {

      GameObj *p = dynamic_cast<GameObj*>(&(**i));
      if (!p) continue;
      
      typeof(p->pending_actions.begin())
        j   = p->pending_actions.begin(),
        end = p->pending_actions.end();

      // sanity check
      while(j != end && j->frame_sent <= P) {
        REM("should not happen! frame_sent <= P ? " << j->frame_sent << " " << P);
        j = p->pending_actions.erase(j);
      }

      // skipped actions
      while (j != end && j->frame_sent <= S) {
        act_changes.skipped.push_back(*j);
        j = p->pending_actions.erase(j);
      }

      // executed actions
      while (j != end && j->frame_sent <= E) {
        act_changes.executed.push_back(*j);
        j = p->pending_actions.erase(j);
      }
    }
  }

#if TEST
  cout << "S:" << act_changes.skipped.size() << " "
       << "E:" << act_changes.executed.size()
       << endl;
#endif
  
}

//===================================================================

void Game::add_watcher(uint4 visibility)
{
  PlayerInfo *w = new PlayerInfo(&shared_objs, pf.get_width(), pf.get_height(), opt.cpind < 0);

  // add ids for global/shared variables

  FORU (j, players[0]->global_shared.size()) {
    FORU (i, players.size()) {
      if (visibility == 0 || (visibility > 0 && (visibility & (1 << i)))) {
        ScriptObj *obj = players[i]->global_shared[j];
        w->new_id(obj);
      }
    }
  }

  w->my_objs = 0;
  w->gof.game = this;
  watchers.push_back(w);
  watcher_visibility.push_back(visibility);
  new_watcher.push_back(1);
}

//===================================================================
