#ifndef GfxGlobal_H
#define GfxGlobal_H

// $Id: GfxGlobal.H 5279 2007-06-23 02:49:08Z mburo $
 
// (c) Michael Buro, licensed under the GNU Public License, version 2

#include "Global.H"

#ifdef GCC
#  include "glew.h"
#  include <gl.h>
#  include <glu.h>
#  include <glut.h>
#endif

#include "GfxLinAlg.H"
#include "Random.H"

extern Random gfxRand;

struct Col3
{
  real4 r,g,b;
  Col3() {}
  Col3(real4 r_, real4 g_, real4 b_) : r(r_), g(g_), b(b_) {}
  Col3 operator+( Col3 &c ) { return Col3(r+c.r, g+c.g, b+c.b); }
  Col3 operator*( real4 m ) { return Col3(r*m, g*m, b*m); }

  bool operator==(const Col3 &c) { return r == c.r && g == c.g && b == c.b; }
  bool operator!=(const Col3 &c) { return !(*this == c); }
};

struct Coor3
{
  real4 x,y,z;
  Coor3() {}
  Coor3(real4 x_, real4 y_, real4 z_) : x(x_), y(y_), z(z_) {}
};

inline Coor3 operator+(const Coor3 &a, const Coor3 &b) {
  return Coor3(a.x + b.x, a.y + b.y, a.z + b.z);
}

inline Coor3 operator-(const Coor3 &a, const Coor3 &b) {
  return Coor3(a.x - b.x, a.y - b.y, a.z - b.z);
}

inline Coor3 operator*(real4 t, const Coor3 &a) {
  return Coor3(t*a.x, t*a.y, t*a.z);
}

inline Coor3 operator*(const Coor3 &a, real4 t) {
  return Coor3(t*a.x, t*a.y, t*a.z);
}

inline real4 dot(const Coor3 &a, const Coor3 &b) {
  return a.x*b.x + a.y*b.y + a.z*b.z;
}

// intersect line and plane
//
// input: line (l1,l2)
//        plane (point,normal)
//
// returns:
//   0 = intersection (result in intersection)
//   1 = no intersection
//   2 = segment in plane

template <typename T>
inline sint4 intersect_line_plane(const Vec3<T> &l1, const Vec3<T> &l2,
				  const Vec3<T> &point, const Vec3<T> &normal,
				  Vec3<T> &intersection, T &t)
{
  Vec3<T> u = l2 - l1;
  Vec3<T> w = l1 - point;
  
  T d = normal.dot(u);
  T n = -normal.dot(w);
  
  if (fabs(d) < 1e-6) { // segment parallel to plane
    if (n == 0) 
      return 2;         // segment in plane
    else
      return 1;         // no intersection
  }

  t = n/d;
  
  intersection = l1 + u * t;
  return 0;
}


// normalize
inline void compute_rect(sint4 x1, sint4 y1, sint4 x2, sint4 y2, SDL_Rect &r)
{
  r.x = std::min(x1,x2);
  r.y = std::min(y1,y2);  
  r.w = abs(x1-x2);
  r.h = abs(y1-y2);  
}

inline void rect_intersection(const SDL_Rect &r1, const SDL_Rect &r2, SDL_Rect &r)
{
  sint4 x1 = std::max(r1.x, r2.x);
  sint4 y1 = std::max(r1.y, r2.y);
  sint4 x2 = std::min(r1.x+r1.w, r2.x + r2.w);
  sint4 y2 = std::min(r1.y+r1.h, r2.y + r2.h);    
  
  r.x = x1;
  r.y = y1;
  r.w = x2-x1;
  r.h = y2-y1;
}

#endif
