// $Id: GfxTerrain.C 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxModule.H"
#include "GfxTerrain.H"
#include "WaterTile.H"
#include "Options.H"
#include "Image.H"
#include "Octree.H"

using namespace std;

//======================================================================

TerrainBlock::TerrainBlock()
  : empty((sector_size+2)*(sector_size+2)), dlist(0), olist(0)
{
  FORU (i, sector_size+2) {
    FORU (j, sector_size+2) {
      tiles[i][j] = 0;
    }
  }
}

//-------------------------------------------------------------------

TerrainBlock::~TerrainBlock() {
  FORS (i, sector_size) {
    FORS (j, sector_size) {
      if (tiles[i+1][j+1]) delete tiles[i+1][j+1];
    }
  }
}

//======================================================================

void GfxTerrain::beginDraw()
{
  glActiveTextureARB(GL_TEXTURE1);
  glEnable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE0);
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_LIGHTING);
  glBindTexture(GL_TEXTURE_2D, ground_tex);
  glColor4f(1, 1, 1, 1);

  glEnable(GL_BLEND);
  glDepthFunc(GL_LEQUAL);

  if (GfxTerrainTile::tile_shader.begin()) {
    GLint shadow = glGetUniformLocation(GfxTerrainTile::tile_shader.id(), "shadow");
    if (shadow >= 0) glUniform1i(shadow, 1);

    GLint ntex = glGetUniformLocation(GfxTerrainTile::tile_shader.id(), "ntex");
    if (ntex >= 0) glUniform1i(ntex, 2);

    GLint ntex3 = glGetUniformLocation(GfxTerrainTile::tile_shader.id(), "ntex3");
    if (ntex3 >= 0) glUniform1i(ntex3, 3);
  }
}

//----------------------------------------------------------------------

void GfxTerrain::endDraw()
{
  glDepthFunc(GL_LESS);
  //  glDisable(GL_BLEND);

  GfxTerrainTile::tile_shader.end();
}

//======================================================================

void GfxTerrain::uflatten(real4 x1r, real4 y1r, real4 x2r, real4 y2r, sint4 f)
{
  sint4 x1 = (sint4) floor(x1r);
  sint4 x2 = (sint4) floor(x2r);
  sint4 y1 = (sint4) floor(y1r);
  sint4 y2 = (sint4) floor(y2r);

  if (x1 < 0) x1 = 0;
  if (y1 < 0) y1 = 0;
  if (x2 >= w) x2 = w-1;
  if (y2 >= h) y2 = h-1;

  //  real4 y = tiles[w*((y1+y2)/2)+(x1+x2)/2].h[0]; // height at the centre

  for(int i = x1; i <= x2; i++)
    for(int j = y1; j <= y2; j++)
      tiles[w*j+i].flatten(f);

  sint4 ss = TerrainBlock::sector_size;
  sint4 sx1 = max((x1-1) / ss, (sint4)(0));
  sint4 sx2 = min((x2+1) / ss, (sint4)(sw-1));
  sint4 sy1 = max((y1-1) / ss, (sint4)(0));
  sint4 sy2 = min((y2+1) / ss, (sint4)(sh-1));

  for (int i = sx1; i <= sx2; i++) {
    for (int j = sy1; j <= sy2; j++) {
      if (sectors[j*sw+i].empty == 0)
	sectors[j*sw+i].build_list();
    }
  }
}

//----------------------------------------------------------------------

void GfxTerrain::flatten(real4 x1r, real4 y1r, real4 x2r, real4 y2r)
{
  uflatten(x1r, y1r, x2r, y2r, 1);
}

void GfxTerrain::unflatten(real4 x1r, real4 y1r, real4 x2r, real4 y2r)
{
  uflatten(x1r, y1r, x2r, y2r, -1);
}

//======================================================================

GfxTerrain::GfxTerrain(GfxModule &gfxm_, sint4 tw, sint4 th, GLuint /*tex*/, Octree *o) : gfxm(gfxm_)
{
  string f = sanitize_path(gfxm.opt.path + "/terrain.tga");
  
  Image imgTerrain;
  if (imgTerrain.loadTGA(f, gfxm.opt.terrain_gamma)) {
    ground_tex = imgTerrain.createTexture();
    if (!ground_tex) ERR2("failed to load terrain image", f.c_str());
  }	
  
  w = tw;
  h = th;

  sw = ((w-1) / TerrainBlock::sector_size)+1;
  sh = ((h-1) / TerrainBlock::sector_size)+1;

  tiles   = new GfxTerrainTile[w*h];
  sectors = new TerrainBlock[sw*sh];
  octree = o;  // new Octree(Coor3(0, 0, 0), Coor3(w, 10, h), 6);

  FORU (i, w*h) {
    tiles[i].initblack(i, this);
  }
  

  // adjust empty counts for sectors that are on the edge
  FORS (j, sh) {
    FORS (i, sw) {
      sint4 index = i + j * sw;
      sint4 empty = (TerrainBlock::sector_size+2) * (TerrainBlock::sector_size+2);
      if (i == 0)    empty -= TerrainBlock::sector_size + 2;
      if (j == 0)    empty -= TerrainBlock::sector_size + 2;
      if (i == sw-1) empty -= TerrainBlock::sector_size + 2;
      if (j == sh-1) empty -= TerrainBlock::sector_size + 2;

      if (i == 0    && j == 0)    ++empty;
      if (i == 0    && j == sh-1) ++empty;
      if (i == sw-1 && j == 0)    ++empty;
      if (i == sw-1 && j == sh-1) ++empty;

      sectors[index].empty = empty;
    }
  }

  f = sanitize_path(gfxm.opt.path + "/noise.tga");

  if (!GfxTerrainTile::load_noise(f,
				  Vec3<real4>(16, gfxm.opt.terrain_noise1, 16),
				  Vec3<real4>(4, gfxm.opt.terrain_noise2, 4)))
    ERR2("failed to load noise image", f.c_str());

  //load_shaders();
}

//----------------------------------------------------------------------

GfxTerrainTile* GfxTerrain::new_tile(sint4 index, Game *game)
{
  tiles[index].init(index, game, this);
  sint4 x = index % w;
  sint4 y = index / w;

  // Look at the new tile's neighbours; if they don't exist then add a black tile.
  // In either case, adjust corner heights and textures to match.

  GfxTerrainTile *t;
  
  t = tile(x-1, y);
  if (t) {
    if (!t->initialized) {
      t->initblack(index-1, this);
      octree->update(t);
    }
    t->E_neighbor();
  }

  t = tile(x-1, y-1);
  if (t) {
    if (!t->initialized) {
      t->initblack(index-1-w, this);
      octree->update(t);
    }
    t->SE_neighbor();
  }

  t = tile(x, y-1);
  if (t) {
    if (!t->initialized) {
      t->initblack(index-w, this);
      octree->update(t);
    }
    t->S_neighbor();
  }

  t = tile(x+1, y-1);
  if (t) {
    if (!t->initialized) {
      t->initblack(index+1-w, this);
      octree->update(t);
    }
    t->SW_neighbor();
  }

  t = tile(x+1, y);
  if (t) {
    if (!t->initialized) {
      t->initblack(index+1, this);
      octree->update(t);
    }
    t->W_neighbor();
  }

  t = tile(x+1, y+1);
  if (t) {
    if (!t->initialized) {
      t->initblack(index+1+w, this);
      octree->update(t);
    }
    t->NW_neighbor();
  }

  t = tile(x, y+1);
  if (t) {
    if (!t->initialized) {
      t->initblack(index+w, this);
      octree->update(t);
    }
    t->N_neighbor();
  }

  t = tile(x-1, y+1);
  if (t) {
    if (!t->initialized) {
      t->initblack(index-1+w, this);
      octree->update(t);
    }
    t->NE_neighbor();
  }


  tiles[index].compute_tc();


  // compute TerrainBlock sector indices
  const sint4 ss = TerrainBlock::sector_size;
  sint4 x2 = x % ss;
  sint4 y2 = y % ss;
  sint4 sx = x / ss;
  sint4 sy = y / ss;
  sint4 index2 = sx + sy * sw;


  sectors[index2].new_tile(x2, y2, &tiles[index], octree);
  if (sx > 0 && x2 == 0) {
    sectors[index2-1].new_tile(ss, y2, &tiles[index], octree);
    if (sy > 0 && y2 == 0)       sectors[index2-sw-1].new_tile(ss, ss, &tiles[index], octree);
    if (sy < sh-1 && y2 == ss-1) sectors[index2+sw-1].new_tile(ss, -1, &tiles[index], octree);
  }
  if (sx < sw-1 && x2 == ss-1) {
    sectors[index2+1].new_tile(-1, y2, &tiles[index], octree);
    if (sy > 0 && y2 == 0)       sectors[index2-sw+1].new_tile(-1, ss, &tiles[index], octree);
    if (sy < sh-1 && y2 == ss-1) sectors[index2+sw+1].new_tile(-1, -1, &tiles[index], octree);
  }
  if (sy > 0 && y2 == 0)       sectors[index2-sw].new_tile(x2, ss, &tiles[index], octree);
  if (sy < sh-1 && y2 == ss-1) sectors[index2+sw].new_tile(x2, -1, &tiles[index], octree);

  octree->remove(&tiles[index]);
  if (sectors[index2].empty) // still has empty tiles
  {
    octree->add(&tiles[index]);
  }
  
  if (tiles[index].has_water()) octree->add(tiles[index].new_WaterTile());
  return &(tiles[index]);
}

//======================================================================

/** Add a new tile t to the terrain block at local position (x,y).
    If this fills up the TerrainBlock then update the Octree.
*/

void TerrainBlock::new_tile(sint4 x, sint4 y, GfxTerrainTile *t, Octree *oct)
{
  x++; // offset for border tiles
  y++;

  // adjust the bounding box to fit the new tile
  if (aabb_d.x==0) { // TerrainBlock is empty (bbox width = 0)
    aabb_o = t->aabb_o;
    aabb_d = t->aabb_d;
  } else {
    aabb_o = Vec3<real4>::min(aabb_o, t->aabb_o);
    aabb_d = Vec3<real4>::max(aabb_o+aabb_d, t->aabb_d+t->aabb_o)-aabb_o;
  }

  if (tiles[x][y] == 0) --empty;
  tiles[x][y] = t;

  // The TerrainBlock is full, remove all of the individual tiles
  // from the octree and replace them with this TerrainBlock.
  if (empty == 0) {
    build_list();

    FORS (i, sector_size) {
      FORS (j, sector_size) {
	GfxTerrainTile *t = tiles[i+1][j+1]; // offset for border tiles
	assert(t);
	oct->remove(t);
      }
    }

    oct->add(this);
  }
}

//-------------------------------------------------------------------

void TerrainBlock::build_list()
{
  if (!dlist) {
    dlist = glGenLists(1);
    olist = glGenLists(1);
  }

  glNewList(dlist, GL_COMPILE);
  //  GfxTerrainTile::tile_shader.begin();
  bool o = GfxTerrainTile::outline; // turn off outlines while making the display list
  GfxTerrainTile::outline = false;
  FORS (i, sector_size) {
    FORS (j, sector_size) {
      if (tiles[i+1][j+1])
        tiles[i+1][j+1]->draw();
    }
  }
  //  GfxTerrainTile::tile_shader.end();
  glEndList();

  GfxTerrainTile::outline = o;

  glNewList(olist, GL_COMPILE);
  //  GfxTerrainTile::tile_shader.begin();
  FORS (i, sector_size) {
    FORS (j, sector_size) {
      if (tiles[i+1][j+1])
        tiles[i+1][j+1]->draw_outline();
    }
  }
  //  GfxTerrainTile::tile_shader.end();
  glEndList();

}

//-------------------------------------------------------------------

void TerrainBlock::pick()
{
  FORS (i, sector_size) {
    FORS (j, sector_size) {
      GfxTerrainTile *t = tiles[i+1][j+1];
      if (t) {
	t->pick();
      }
    }
  }
}

//-------------------------------------------------------------------

void TerrainBlock::draw()
{
  //  GfxTerrainTile::tile_shader.begin();
  // in the current implementation dlist should always be defined
  if (dlist) {
    glCallList(dlist);
    if (GfxTerrainTile::outline)
      glCallList(olist);
  } else {
    FORS (i, sector_size) {
      FORS (j, sector_size) {
	GfxTerrainTile *t = tiles[i+1][j+1];
	if (t) {
          t->draw();
          if (GfxTerrainTile::outline)
            t->draw_outline();
        }
      }
    }
  }
  //  GfxTerrainTile::tile_shader.end();
}

//===================================================================
/*
void GfxTerrain::load_shaders()
{
  if (gfxm.opt.terrain_vertex_shader != "") {
    string vshader = sanitize_path(gfxm.opt.path + "/" + gfxm.opt.terrain_vertex_shader);
    GfxTerrainTile::tile_shader.vertex_shader(vshader);
  }
  if (gfxm.opt.terrain_fragment_shader != "") {
    string fshader = sanitize_path(gfxm.opt.path + "/" + gfxm.opt.terrain_fragment_shader);
    GfxTerrainTile::tile_shader.fragment_shader(fshader);
  }

  if (gfxm.opt.water_vertex_shader != "") {
    string vshader = sanitize_path(gfxm.opt.path + "/" + gfxm.opt.water_vertex_shader);
    WaterTile::water_shader.vertex_shader(vshader);
  }
  if (gfxm.opt.water_fragment_shader != "") {
    string fshader = sanitize_path(gfxm.opt.path + "/" + gfxm.opt.water_fragment_shader);
    WaterTile::water_shader.fragment_shader(fshader);
  }
}
*/
//======================================================================

