// $Id: Image.C 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file
// (c) Tim Furtak
// (c) Erik Benner
// licensed under the GPL

#include "Image.H"
#include "FileArchive.H"

using namespace std;

//===================================================================

Image::Image()
{
  w = 0; h = 0;
  channels = 0;
  size = 0;
  data = 0;
}

//-------------------------------------------------------------------

Image::Image(const std::string &file, real4 gamma)
{
  w = 0; h = 0;
  channels = 0;
  size = 0;
  data = 0;
  loadTGA(file, gamma);
}

//-------------------------------------------------------------------

void Image::free()
{
  if (data) {
    delete [] data; 
    data = 0;
  }
}

//-------------------------------------------------------------------

Image::~Image()
{
  free();
}

//-------------------------------------------------------------------

void Image::empty(uint2 width, uint2 height, uint2 bpp)
{
  free();
  
  w = width;
  h = height;
  channels = bpp;
  
  size = w*h*channels;
  data = new uint1[size];
}

//-------------------------------------------------------------------

void Image::copy_rect(const Image &i, sint4 sx, sint4 sy, sint4 rw, sint4 rh, sint4 dx, sint4 dy)
{
  if (dx >= w) return;
  if (dx + rw >= w) rw = w - dx - 1;

  if (dy >= h) return;
  if (dy + rh >= h) rh = h - dy - 1;


  if (sx >= i.w) return;
  if (sx + rw >= i.w) rw = i.w - sx - 1;

  if (sy >= i.h) return;
  if (sy + rh >= i.h) rh = i.h - sy - 1;

  if (rw <= 0 || rh <= 0) return;

  FORU (y, rh) { // for each row to copy
    memcpy(&data  [( (y+dy)*w   + dx ) * channels],
           &i.data[( (y+sy)*i.w + sx ) * channels],
           rw*channels);
  }
}

//-------------------------------------------------------------------

bool Image::loadTGA(const std::string &filename, real4 gamma)
{
  // Load the bitmap; check for errors; if bitmap's not found, quit.
  string fdata;
  if (!FileArchive::read_file(filename, fdata)) return false;

  uint4 fsize = fdata.size();
  if (fsize < 18) REM2("Not a TGA file", filename.c_str());

  const char *ptr = fdata.c_str();

  free();
  
  // Supported Targa formats 
  const sint1 TGA_RGB     = 2;  // Uncompressed, RGB images
  const sint1 TGA_BW      = 3;  // Uncompressed, black and white images
  const sint1 TGA_RLE_RGB = 10; // Runlength encoded RGB images
  const sint1 TGA_RLE_BW  = 11; // Runlength encoded BW images

  sint1 img_header[6]; // width, height, bits per pixel
  sint1 comment;       // Length of comment string located after header
  sint1 colourmap;     // Colour map type (must equal 0)
  sint1 typecode;      // Type code (must be one of the supported types)
  bool compressed = false;
  
  // Read one byte for 'image id length'
  comment = ptr[0];
  // Read 'colour map type'
  colourmap = ptr[1];
  if (colourmap) {
    cout << "Unsupported TGA colour maps found in " << filename.c_str() << endl;
    return false;
  }
  
  // Read 'image type' field
  typecode = ptr[2];
  if (typecode == TGA_RLE_RGB || typecode == TGA_RLE_BW)
    compressed = true;
  else if (typecode != TGA_RGB && typecode != TGA_BW) {
    REM2("Unsupported TGA type found in", filename.c_str());
    return false;
  }
  
  // Ignore unsupported fields of the TGA header:
  // 2-'colour map origin', 2-'colour map length', 1-'colour map depth',
  // 2-'x origin', & 2-'y origin' = 9 bytes + first 3
  ptr += 12;

  // Read image header
  memcpy(&img_header, ptr, 6);
  ptr += 6;
  
  w = abs(img_header[0] + img_header[1] * 256);
  h = abs(img_header[2] + img_header[3] * 256);
  channels = img_header[4] / 8;

  // Allocate memory
  empty(w, h, channels);

  // Read uncompressed data
  if (!compressed) {
    // Read uncompressed data
    if (fsize < size + 18) {
      REM2("Problems loading uncompressed TGA:", filename.c_str());
      free();
      return false;
    }
    memcpy(data, ptr, size);

  } else {
    // Read RLE compressed data
    if (!readRLE(ptr, fsize - 18)) {
      REM2("Problems loading compressed TGA:", filename.c_str());
      return false;
    }
  }
  
  // Second pass: BGR-to-RGB and gamma correction
  if (channels >= 3) {

    //----------------------------------------
    if (gamma == 1.0) {
      Vector<uint1> g;
      createGammaTable(gamma, g);
      
      // flip red and blue for OpenGL usage   
      for (uint4 i = 0; i < size; i += channels) swap(data[i], data[i+2]); 
    } else { //------------------------------
      Vector<uint1> g;
      createGammaTable(gamma, g);
      
      for (uint4 i = 0; i < size; i += channels) {
        // flip red and blue for OpenGL usage   
        swap(data[i], data[i+2]); 
        
        data[i]   = g[data[i]];   // gamma correction
        data[i+1] = g[data[i+1]];
        data[i+2] = g[data[i+2]];
      }
    }
    //----------------------------------------
  } else {
    if (gamma != 1.0) {
      Vector<uint1> g;
      createGammaTable(gamma, g);
      for (uint4 i = 0; i < size; i += channels) data[i] = g[data[i]];   // gamma correction
    }
  }
  
  // Check flip bit
  if((img_header[5] & 0x20) == 0) flip();
  
  return true; 
}

//-------------------------------------------------------------------

/** Read in run-length encoded data from a data stream. */

bool Image::readRLE(const char *ptr, uint4 fsize)
{
  const char *end_ptr = ptr + fsize;

  uint1 chunk = 0; // RLE code
  uint4 pixel = 0; // Index of next pixel
  
  // Load in all the pixel data
  while (pixel < (uint4) size) {

    assert(ptr <= end_ptr);

    // Read next RLE code chunk, fixme: error check?
    chunk = *ptr++;
    
    // If we're reading in a string of unique pixels
    if (chunk < 128) {
      // Increase the count by 1
      ++chunk;
      
      // Read a block of unique colors
      memcpy(&data[pixel], ptr, channels*chunk);
      ptr += channels * chunk;
      pixel += channels * chunk;

    } else {  // Otherwise, we need to expand the RLE pixel
      // Determine number of pixels to colour
      chunk -= 127;

      //      memcpy(color, ptr, channels); // color to be duplicated

      FORU (i, chunk) {
        //        memcpy(&data[pixel], color, channels);
        memcpy(&data[pixel], ptr, channels);
        pixel += channels;
      }
      ptr += channels;
    }
  }
  
  // Ensure pixel data was read successfully
  if (ptr < end_ptr) return true;

  cout << "ptr = " << (void*)ptr << endl;
  cout << "end ptr = " << (void*)end_ptr << endl;
  cout << "NO GOOD" << endl;

  free();
  return false;
}

//-------------------------------------------------------------------

bool Image::saveTGA(const std::string &filename)
{
  // Load the bitmap; check for errors; if bitmap's not found, quit.
  ofstream file;
  file.open(filename.c_str(), ios::binary);
  if (!file.is_open()) { 
    cout << "Can't open file " << filename.c_str() << endl;
    return false;
  }
  
  // Supported Targa formats 
  const sint1 TGA_RGB     = 2;  // Uncompressed, RGB images
  //  const sint1 TGA_BW      = 3;  // Uncompressed, black and white images
  //  const sint1 TGA_RLE_RGB = 10; // Runlength encoded RGB images
  //  const sint1 TGA_RLE_BW  = 11; // Runlength encoded BW images

  sint1 img_header[6]; // width, height, bits per pixel
  sint1 comment = 0;   // Length of comment string located after header
  sint1 colourmap = 0; // Colour map type (must equal 0)
  sint1 typecode = TGA_RGB;
  //  bool compressed = false;
  
  file.put(comment);
  file.put(colourmap);
  file.put(typecode);

  // Ignore unsupported fields of the TGA header:
  // 2-'colour map origin', 2-'colour map length', 1-'colour map depth',
  // 2-'x origin', & 2-'y origin' = 9 bytes
  FORU (i, 9) file.put(0);

  img_header[0] = w % 256;
  img_header[1] = w / 256;
  img_header[2] = h % 256;
  img_header[3] = h / 256;
  img_header[4] = channels * 8;
  img_header[5] = 0;

  // Read image header
  FORU (i, sizeof(img_header)) file.put(img_header[i]);

  // Write uncompressed data
  FORU (i, size) file.put(data[i]);

  file.close();
  
  return true; 
}

//-------------------------------------------------------------------

GLuint Image::createTexture(FILTER filter, sint4 level) const
{
  GLuint id;
  GLuint type;

  switch (channels)  {
  case 1:  type=GL_LUMINANCE; break;
  case 2:  type=GL_LUMINANCE_ALPHA; break;
  case 3:  type=GL_RGB; break;
  default: type=GL_RGBA; break;
  }

  glGenTextures(1, &id);
  glBindTexture(GL_TEXTURE_2D, id);
  
  switch (filter) {
    // Nearest - no mipmap, nearest sample
  case NEAREST:
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    break;
    // Linear - no mipmap, linear sample
  case LINEAR:
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    break;
    // Trilinear - linear mipmap, linear sample
  case TRILINEAR:
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    break;
    
  case BILINEAR:
  default:
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    break;
  }
  
  // Always repeat textures (don't clamp)
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

  uint1 *ndata;
  sint4 b = 1 << level;
  if (level > 0) {
    ndata = new uint1[size / (b*b)];
    gluScaleImage(type, w, h, GL_UNSIGNED_BYTE, data, w/b, h/b, GL_UNSIGNED_BYTE, ndata);
  } else {
    ndata = data;
  }

  // Create mipmaps if needed
  if (filter == BILINEAR || filter == TRILINEAR)
    gluBuild2DMipmaps(GL_TEXTURE_2D, type, w/b, h/b, type, GL_UNSIGNED_BYTE, ndata);
  else
    glTexImage2D(GL_TEXTURE_2D, 0, type, w/b, h/b, 0, type, GL_UNSIGNED_BYTE, ndata);

  if (level > 0) delete[] ndata;

  return id;
}

//-------------------------------------------------------------------

bool Image::deleteTexture(GLuint id)
{
  if (glIsTexture(id)) {
    glDeleteTextures(1, &id);
    return true;
  }
  return false;
}

//-------------------------------------------------------------------

void Image::flip()
{
  uint1 *flipped = new uint1[size];
  
  // For each row in the image
  FORU (y, h) {
    
    // Copy each row to the new image
    memcpy(static_cast<void*>(&flipped[(h-y-1)*channels*w]),
	   static_cast<void*>(&data[y*channels*w]),
	   w*channels);
  }
  
  free();
  data = flipped;
}

//-------------------------------------------------------------------

void Image::createGammaTable(real4 gamma, Vector<uint1> &g)
{
  g.clear();
  
  gamma = 1.0/gamma;

  FORS (i, 256) {
    if (gamma == 1.0)
      g.push_back(i);
    else
      g.push_back((uint1)(std::pow(real8(i/255.0), real8(gamma))*255.0+0.5));
  }
}

//===================================================================

