// $Id: Lexer.H 5279 2007-06-23 02:49:08Z mburo $ 

// (c) Jonathan Yip, Michael Buro, licensed under the GPL

#ifndef Lexer_H
#define Lexer_H

#include "OSL_RecStream.H"

namespace OSL {
  
  enum TokenType {
    TK_ERROR,  // token value contains error message
    TK_EOF,
    TK_NIL,
    TK_ID,
    TK_IF,
    TK_ELSE,
    TK_WHILE,
    TK_FOR,
    TK_BREAK,
    TK_CONT,
    TK_SWITCH,
    TK_CASE,
    TK_DEFAULT,
    TK_CHARC,
    TK_STRINGC,
    TK_INTC,
    TK_UINTC,
    TK_LONGC,
    TK_ULONGC,
    TK_FLOATC,
    TK_DOUBLEC,
    TK_VOID,
    TK_BOOL,
    TK_BYTE,
    TK_UBYTE,
    TK_SHORT,
    TK_USHORT,
    TK_INT,
    TK_UINT,
    TK_LONG,
    TK_ULONG,
    TK_FLOAT,
    TK_DOUBLE,
    TK_STRING,
    TK_CLASS,
    TK_ASSIGN, 
    TK_AND,
    TK_ANDAND,
    TK_OR,
    TK_OROR,
    TK_XOR,
    TK_EQ,
    TK_NEQ,
    TK_LT,
    TK_LEQ,
    TK_GT,
    TK_GEQ, 
    TK_ADD,
    TK_SUB,
    TK_MUL,
    TK_DIV,
    TK_MOD,
    TK_COMPLEMENT,
    TK_INVERSE,
    TK_LSHIFT,
    TK_RSHIFT,
    TK_COLON,
    TK_SCOPE,
    TK_DOT,
    TK_LPAR,
    TK_RPAR,
    TK_LBRACE,
    TK_RBRACE,
    TK_LBRACKET,
    TK_RBRACKET,
    TK_COMMA,
    TK_SEMICOLON,
    //    TK_MAIN,
    TK_NEW,
    //    TK_DELETE,
    TK_RETURN, 
    TK_EXTERN,
    TK_STATIC,
    TK_PUBLIC,
    TK_PROTECTED,
    TK_PRIVATE
  };

  
  class Token {

  public:

    Token() { type = TK_ERROR; }
    Token(TokenType type_) : type(type_) {}
    Token(TokenType type_, const std::string &value_) : type(type_), value(value_) {}

    TokenType get_type() const { return type; }
    std::string get_value() const { return value; }

    void write(std::ostream &os = std::cout) const;
    
  private:
    
    TokenType type;
    std::string value;
  };


  class Lexer {

  public:

    Lexer(RecStream &rstream_);

    Token get_token();

    Token peek_token();

    void reset();
    
    uint4 cur_line_num() const { return rstream.get_line_num(); }

    std::string cur_file() const { return rstream.get_name(); }

    struct SyntaxErrorException : public std::exception {
      std::string msg;

      SyntaxErrorException(const std::string &m)  throw() : msg(m) {}
      virtual ~SyntaxErrorException() throw() { }
    };

    /** thrown exceptions:

    SyntaxErrorException: syntax error like illegal character

    RecStream::ReadErrorException: file read error
    
    */

    
  private:

    RecStream &rstream;    // input stream
    bool      peeked;
    Token     current_token;
    bool      error;       // true after error occurred

    /** file read error or syntax error */
    void throw_error(const std::string &msg);

    /** I/O in current read context */
    sint4 peek() { return rstream.peek(); }
    sint4 get_char();
    void put_char(sint4 c) { return rstream.put_char(c); }    

    /** generate error token and set error status */
    Token ErrToken(const std::string &msg);

    // preprocessor stuff

    std::set<std::string> defs;  // definitions
    
    void skip_to_eol()
    {
      FOREVER {
	sint4 c = get_char();
	if (c == '\n' || c < 0) break;
      }
    }

    static const uint4 MAX_SKIP_LEVEL = 30;
    uint4 next_skipping_level;
    uint4 skipping_flags;
    bool skip_override;

    bool skipping() const { return !skip_override && skipping_flags != 0; }

    sint4 do_preprocessor_command();
    void do_include();
    void do_undef(); 
    void do_define();
    void do_ifdef(bool negate);
    void do_endif(); 
    void do_else();      

    // disallow copy and assignment
    Lexer(const Lexer& rhs);
    Lexer& operator=(const Lexer& rhs);
  };

}
#endif
