// $Id: OSL_Program.C 5279 2007-06-23 02:49:08Z mburo $

// (c) Michael Buro, mburo@cs.ualberta.ca, licensed under the GPL

#include "OSL.H"

using namespace std;

namespace OSL {

  // must match Type enum
  
  const sint4 Type::sizes[TC_END] = {
    /* TC_unknown */ 0,  // can't occur in user data 
    /* TC_pc */      0,  // -"-
    /* TC_fp */      0,  // -"-
    /* TC_fu */      0,  // -"-
    /* TC_void */    0,
    /* TC_sint1 */   1,
    /* TC_uint1 */   1,
    /* TC_sint2 */   2,
    /* TC_uint2 */   2,
    /* TC_sint4 */   4,
    /* TC_uint4 */   4,
    /* TC_sint8 */   8,
    /* TC_uint8 */   8,
    /* TC_real4 */   4,
    /* TC_real8 */   8
  };

  
  void Type::write_cat(Category cat, ostream &os)
  {
    switch (cat) {

    case Type::TC_unknown: os << "unknown"; break;      
    case Type::TC_pc:      os << "pc"; break;
    case Type::TC_fp:      os << "fp"; break;
    case Type::TC_fu:      os << "fu"; break;
    case Type::TC_void:    os << "void"; break;
    case Type::TC_sint1:   os << "s1" ; break;
    case Type::TC_uint1:   os << "u1" ; break;
    case Type::TC_sint2:   os << "s2" ; break;
    case Type::TC_uint2:   os << "u2" ; break;
    case Type::TC_sint4:   os << "s4" ; break;
    case Type::TC_uint4:   os << "u4" ; break;
    case Type::TC_sint8:   os << "s8" ; break;
    case Type::TC_uint8:   os << "u8" ; break;
    case Type::TC_real4:   os << "r4" ; break;
    case Type::TC_real8:   os << "r8" ; break;
    case Type::TC_object:  os << "object"; break;      
    case Type::TC_class:   os << "class"; break;
    case Type::TC_array:   os << "array"; break;
    case Type::TC_string:  os << "string"; break;      

    default: throw Exception("unknown category" + to_string(cat)); break;
    }
  }


  Class::Class()
  {
    ref_tind = Type::TC_unknown;
    name = "?";
    super_class_index = -1;
    inst_data_size = 0;
  }

  
  Program::Program() {

    // init. types
    
    FORS (i, Type::TC_END) {
      Type t;
      t.cat = static_cast<Type::Category>(i);
      types.push_back(t);
    }

    ClassInstance *ci;
    class_data_offset = ((uint1*)&ci->data[0]) - ((uint1*)ci);

    ArrayInstance *ai;
    array_data_offset = ((uint1*)&ai->data[0]) - ((uint1*)ai);

    std::cout << "class_data_offset=" << class_data_offset
	      << " array_data_offset=" << array_data_offset
	      << std::endl;

    heap = new Heap(this);
  }

  Class* Program::add_class(const string& name) 
  {
    uint4 class_index = classes.size();
    Class c;
    c.defined = false;
    c.name = name;
    c.super_class_index = -1;
    c.ref_tind = types.size();
    classes.push_back(c);

    // Order important!
    class_ids[name] = class_index;

    // check
    Type t;
    t.cat = Type::TC_class;
    t.class_index = class_index;
    types.push_back(t);

    return &classes[class_index];
  }

  void Program::add_array(TypeIndex tind, sint4 dimensions)
  {
    Type t;
    t.cat = Type::TC_array;
    t.array_rank = dimensions;
    t.T_cat = (Type::Category)tind;
    types.push_back(t);
  }

  /*
  void Program::add_type(const Type &type)
  {
    types.push_back(type);
  }
  */

  void Program::add_ext(ExtFunc::C_Func c_func, const string& name, vector<TypeIndex> par_types)
  {
    ExtFunc fn;
    fn.name = name;
    fn.c_func = c_func;
    fn.par_types = par_types;
    fn.ret_tind = 0;
    extf_ids[name] = extfs.size();
    extfs.push_back(fn);
  }

  Value Program::run(const string &cl, const string &func, 
		     vector<Value> &params)
  {
    Context* cont = new Context(*this);
    contexts.push_back(cont);
    Class* c = &classes[class_ids[cl]];
    cont->init_run(&c->static_funcs[c->static_func_ids[func]], params);
    try {
      cont->run();
    }
    catch (Exception e) { cout << e.msg << endl; }
    Value ret = cont->result();
    delete contexts[contexts.size()-1];
    contexts.pop_back();
    return ret;
  }

  Value Program::run(const uint4 class_index, const uint4 func_index,
		     vector<Value> & params)
  {
    Context* cont = new Context(*this);
    contexts.push_back(cont);
    cont->init_run(&classes[class_index].static_funcs[func_index], params);
    try {
      cont->run();
    }
    catch (Exception e) { cout << e.msg << endl; }
    return cont->result();
  }
  
  void Program::write_type(const Type &type, ostream &os) const
  {
    Type::write_cat(type.cat, os);

    switch (type.cat) {

    case Type::TC_class:
      os << "[" << classes[type.class_index].name << "]";
      break;

    case Type::TC_array:
      os << "[" << type.array_rank << ":"; Type::write_cat(type.T_cat, os);
      if (type.T_cat == Type::TC_class)
	os << "[" << classes[type.class_index].name << "]";
      os << "]";
      break;
      
    default: break;
    }
  }

  
  void Program::write_value(const Value &v, ostream &os) 
  {
    Type::write_cat(v.get_cat(), os);
    os << ":";
    
    switch (v.get_cat()) {
    case Type::TC_pc:      os << v.v_pc;    break;
    case Type::TC_fp:      os << v.v_fp;    break;
    case Type::TC_fu:      os << v.v_fu;    break;
    case Type::TC_sint1:   os << v.v_sint4; break;
    case Type::TC_uint1:   os << v.v_sint4; break;
    case Type::TC_sint2:   os << v.v_sint4; break;
    case Type::TC_uint2:   os << v.v_sint4; break;
    case Type::TC_sint4:   os << v.v_sint4; break;
    case Type::TC_uint4:   os << v.v_uint4; break;
    case Type::TC_sint8:   os << v.v_sint8; break;
    case Type::TC_uint8:   os << v.v_sint8; break;
    case Type::TC_real4:   os << v.v_real4; break;
    case Type::TC_real8:   os << v.v_real8; break;
    case Type::TC_object:  os << v.v_object; break;
    default: ERR("unknown category"); break;
    }
  }


  void Program::write_code(const FuncMember &fm, ostream &os) const
  {
    uint4 pc = 0;
    uint4 size = fm.code.size();
    const vector<Code> &code = fm.code;
      
    while (pc < size) {

      sint4 f = 0;
      string n;
      Type::Category t[3] = { Type::TC_uint4, Type::TC_uint4, Type::TC_uint4 };
      
      switch (fm.code[pc]) {
	
      case Opcodes::UNKNOWN:     n="???????"; break;

      case Opcodes::NOP:         n="NOP"; break;

      case Opcodes::JUMP:        n="JUMP";        f=1; t[0]=Type::TC_pc; break;
      case Opcodes::JUMP_FALSE:  n="JUMP_FALSE";  f=1; t[0]=Type::TC_pc; break;
      case Opcodes::JUMP_TRUE:   n="JUMP_TRUE";   f=1; t[0]=Type::TC_pc; break; 
      case Opcodes::JUMP_PFALSE: n="JUMP_PFALSE"; f=1; t[0]=Type::TC_pc; break; 
      case Opcodes::JUMP_PTRUE:  n="JUMP_PTRUE";  f=1; t[0]=Type::TC_pc; break; 
	
      case Opcodes::CALL_VIRT:   n="CALL_VIRT"; f=1; break;
      case Opcodes::CALL_STAT:   n="CALL_STAT"; f=2; break;
      case Opcodes::CALL_EXT:    n="CALL_EXT";  f=1; break;
	
      case Opcodes::RETURN:      n="RETURN"; break;
      case Opcodes::RETURN_VAL:  n="RETURN_VAL"; break;
	
      case Opcodes::NEW_CLASS:   n="NEW_CLASS"; f=1; break;
      case Opcodes::NEW_ARRAY:   n="NEW_ARRAY"; f=2; break;

      case Opcodes::LOAD_LOCAL:  n="LOAD_LOCAL";  f=1; break;
      case Opcodes::LOAD_PARAM:  n="LOAD_PARAM";  f=1; break;
      case Opcodes::LOAD_STATIC: n="LOAD_STATIC"; f=2; break;
      case Opcodes::LOAD_INST:   n="LOAD_INST";   f=1; break;
      case Opcodes::LOAD_ARRAY:  n="LOAD_ARRAY";  f=0; break;

      case Opcodes::STORE_LS1: n="STORE_LS1"; f=1; break;
      case Opcodes::STORE_LU1: n="STORE_LU1"; f=1; break;
      case Opcodes::STORE_LS2: n="STORE_LS2"; f=1; break;
      case Opcodes::STORE_LU2: n="STORE_LU2"; f=1; break;
      case Opcodes::STORE_LS4: n="STORE_LS4"; f=1; break;
      case Opcodes::STORE_LU4: n="STORE_LU4"; f=1; break;
      case Opcodes::STORE_LS8: n="STORE_LS8"; f=1; break;
      case Opcodes::STORE_LU8: n="STORE_LU8"; f=1; break;
      case Opcodes::STORE_LR4: n="STORE_LR4"; f=1; break;
      case Opcodes::STORE_LR8: n="STORE_LR8"; f=1; break;
      case Opcodes::STORE_LREF: n="STORE_LREF"; f=1; break;

      case Opcodes::STORE_PS1: n="STORE_PS1"; f=1; break;
      case Opcodes::STORE_PU1: n="STORE_PU1"; f=1; break;
      case Opcodes::STORE_PS2: n="STORE_PS2"; f=1; break;
      case Opcodes::STORE_PU2: n="STORE_PU2"; f=1; break;
      case Opcodes::STORE_PS4: n="STORE_PS4"; f=1; break;
      case Opcodes::STORE_PU4: n="STORE_PU4"; f=1; break;
      case Opcodes::STORE_PS8: n="STORE_PS8"; f=1; break;
      case Opcodes::STORE_PU8: n="STORE_PU8"; f=1; break;
      case Opcodes::STORE_PR4: n="STORE_PR4"; f=1; break;
      case Opcodes::STORE_PR8: n="STORE_PR8"; f=1; break;
      case Opcodes::STORE_PREF: n="STORE_PREF"; f=1; break;
	
      case Opcodes::STORE_SS1: n="STORE_SS1"; f=2; break;
      case Opcodes::STORE_SU1: n="STORE_SU1"; f=2; break;
      case Opcodes::STORE_SS2: n="STORE_SS2"; f=2; break;
      case Opcodes::STORE_SU2: n="STORE_SU2"; f=2; break;
      case Opcodes::STORE_SS4: n="STORE_SS4"; f=2; break;
      case Opcodes::STORE_SU4: n="STORE_SU4"; f=2; break;
      case Opcodes::STORE_SS8: n="STORE_SS8"; f=2; break;
      case Opcodes::STORE_SU8: n="STORE_SU8"; f=2; break;
      case Opcodes::STORE_SR4: n="STORE_SR4"; f=2; break;
      case Opcodes::STORE_SR8: n="STORE_SR8"; f=2; break;
      case Opcodes::STORE_SREF: n="STORE_SREF"; f=2; break;
	
      case Opcodes::STORE_IS1: n="STORE_IS1"; f=1; break;
      case Opcodes::STORE_IU1: n="STORE_IU1"; f=1; break;
      case Opcodes::STORE_IS2: n="STORE_IS2"; f=1; break;
      case Opcodes::STORE_IU2: n="STORE_IU2"; f=1; break;
      case Opcodes::STORE_IS4: n="STORE_IS4"; f=1; break;
      case Opcodes::STORE_IU4: n="STORE_IU4"; f=1; break;
      case Opcodes::STORE_IS8: n="STORE_IS8"; f=1; break;
      case Opcodes::STORE_IU8: n="STORE_IU8"; f=1; break;
      case Opcodes::STORE_IR4: n="STORE_IR4"; f=1; break;
      case Opcodes::STORE_IR8: n="STORE_IR8"; f=1; break;
      case Opcodes::STORE_IREF: n="STORE_IREF"; f=1; break;
	
      case Opcodes::STORE_AS1: n="STORE_AS1"; f=0; break;
      case Opcodes::STORE_AU1: n="STORE_AU1"; f=0; break;
      case Opcodes::STORE_AS2: n="STORE_AS2"; f=0; break;
      case Opcodes::STORE_AU2: n="STORE_AU2"; f=0; break;
      case Opcodes::STORE_AS4: n="STORE_AS4"; f=0; break;
      case Opcodes::STORE_AU4: n="STORE_AU4"; f=0; break;
      case Opcodes::STORE_AS8: n="STORE_AS8"; f=0; break;
      case Opcodes::STORE_AU8: n="STORE_AU8"; f=0; break;
      case Opcodes::STORE_AR4: n="STORE_AR4"; f=0; break;
      case Opcodes::STORE_AR8: n="STORE_AR8"; f=0; break;
      case Opcodes::STORE_AREF: n="STORE_AREF"; f=0; break;		

      case Opcodes::DUP:      n="DUP";      break;
      case Opcodes::POP:      n="POP";      break;
	
      case Opcodes::LINE:     n="LINE";     f=1; break;
      case Opcodes::YIELD:    n="YIELD";    break;
	
	//      case Opcodes::ISA_REF:  n="ISA_REF";  f=1; break;
	//      case Opcodes::CONV_REF: n="CONV_REF"; f=1; break;
	
	// load constant (also for creating local variables)
	
      case Opcodes::PUSH_CS4: n="PUSH_CS4"; f=1; t[0]=Type::TC_sint4; break;
      case Opcodes::PUSH_CU4: n="PUSH_CU4"; f=1; t[0]=Type::TC_uint4; break;
      case Opcodes::PUSH_CS8: n="PUSH_CS8"; f=1; t[0]=Type::TC_sint8; break;
      case Opcodes::PUSH_CU8: n="PUSH_CU8"; f=1; t[0]=Type::TC_uint8; break;
      case Opcodes::PUSH_CR4: n="PUSH_CR4"; f=1; t[0]=Type::TC_real4; break;
      case Opcodes::PUSH_CR8: n="PUSH_CR8"; f=1; t[0]=Type::TC_real8; break;    

	// fixme: implement PUSH_CARR
	
	// case Opcodes::PUSH_CARR: n="PUSH_CARR"; f=1; t[0] = Type::; break;
	// fetch(string) push(reference to new char[] = string)
	// string format in bytecode: length(4): sequence of bytes (padded)
	
      case Opcodes::PUSH_NULL: n="PUSH_NULL"; break; 
      case Opcodes::PUSH_ZS4:  n="PUSH_ZS4";  break;
      case Opcodes::PUSH_ZU4:  n="PUSH_ZU4";  break; 
      case Opcodes::PUSH_ZS8:  n="PUSH_ZS8";  break;
      case Opcodes::PUSH_ZU8:  n="PUSH_ZU8";  break;
      case Opcodes::PUSH_ZR4:  n="PUSH_ZR4";  break;
      case Opcodes::PUSH_ZR8:  n="PUSH_ZR8";  break; 
	
      case Opcodes::ADD_R4:  n="ADD_R4"; break; 
      case Opcodes::ADD_R8:  n="ADD_R8"; break; 
      case Opcodes::ADD_SS4: n="ADD_SS4"; break; 
      case Opcodes::ADD_US4: n="ADD_US4"; break; 
      case Opcodes::ADD_SU4: n="ADD_SU4"; break; 
      case Opcodes::ADD_UU4: n="ADD_UU4"; break; 
	
	//ADD_SS4_CHK, ADD_US4_CHK, ADD_SU4_CHK, ADD_UU4_CHK,
	//ADD_SS8, ADD_US8, ADD_SU8, ADD_UU8,
	//ADD_SS8_CHK, ADD_US8_CHK, ADD_SU8_CHK, ADD_UU8_CHK,

      case Opcodes::SUB_R4:  n="SUB_R4"; break; 
      case Opcodes::SUB_R8:  n="SUB_R8"; break; 
      case Opcodes::SUB_SS4: n="SUB_SS4"; break; 
      case Opcodes::SUB_US4: n="SUB_US4"; break; 
      case Opcodes::SUB_SU4: n="SUB_SU4"; break; 
      case Opcodes::SUB_UU4: n="SUB_UU4"; break; 

	//SUB_SS4_CHK, SUB_US4_CHK, SUB_SU4_CHK, SUB_UU4_CHK,
	//SUB_SS8, SUB_US8, SUB_SU8, SUB_UU8,
	//SUB_SS8_CHK, SUB_US8_CHK, SUB_SU8_CHK, SUB_UU8_CHK,

      case Opcodes::MUL_R4:  n="MUL_R4"; break; 
      case Opcodes::MUL_R8:  n="MUL_R8"; break; 
      case Opcodes::MUL_SS4: n="MUL_SS4"; break; 
      case Opcodes::MUL_US4: n="MUL_US4"; break; 
      case Opcodes::MUL_SU4: n="MUL_SU4"; break; 
      case Opcodes::MUL_UU4: n="MUL_UU4"; break; 

	//MUL_SS4_CHK, MUL_US4_CHK, MUL_SU4_CHK, MUL_UU4_CHK,
	//MUL_SS8, MUL_US8, MUL_SU8, MUL_UU8,
	//MUL_SS8_CHK, MUL_US8_CHK, MUL_SU8_CHK, MUL_UU8_CHK,

      case Opcodes::DIV_R4:  n="DIV_R4"; break; 
      case Opcodes::DIV_R8:  n="DIV_R8"; break; 
      case Opcodes::DIV_SS4: n="DIV_SS4"; break; 
      case Opcodes::DIV_US4: n="DIV_US4"; break; 
      case Opcodes::DIV_SU4: n="DIV_SU4"; break; 
      case Opcodes::DIV_UU4: n="DIV_UU4"; break; 

	//DIV_SS4_CHK, DIV_US4_CHK, DIV_SU4_CHK, DIV_UU4_CHK,
	//DIV_SS8, DIV_US8, DIV_SU8, DIV_UU8,
	//DIV_SS8_CHK, DIV_US8_CHK, DIV_SU8_CHK, DIV_UU8_CHK,

	//case Opcodes::MOD_R4:  n="MOD_R4"; break; 
	//case Opcodes::MOD_R8:  n="MOD_R8"; break; 
      case Opcodes::MOD_SS4: n="MOD_SS4"; break; 
      case Opcodes::MOD_US4: n="MOD_US4"; break; 
      case Opcodes::MOD_SU4: n="MOD_SU4"; break; 
      case Opcodes::MOD_UU4: n="MOD_UU4"; break; 
	
	//MOD_SS4_CHK, MOD_US4_CHK, MOD_SU4_CHK, MOD_UU4_CHK,
	//MOD_SS8, MOD_US8, MOD_SU8, MOD_UU8,
	//MOD_SS8_CHK, MOD_US8_CHK, MOD_SU8_CHK, MOD_UU8_CHK,

      case Opcodes::AND_I4: n="AND_I4"; break;
	// case Opcodes::AND_I8: 

      case Opcodes::OR_I4:  n="OR_I4"; break;
	// case Opcodes::OR_I8:

      case Opcodes::XOR_I4: n="XOR_I4"; break;
	// case Opcodes::XOR_I8:

      case Opcodes::SHL_I4: n="SHL_I4"; break;
	// case Opcodes::SHL_I8:

      case Opcodes::SHR_I4: n="SHR_I4"; break;
	// case Opcodes::SHR_I8:

      case Opcodes::EQ_SS4: n="EQ_SS4"; break;
      case Opcodes::EQ_SU4: n="EQ_SU4"; break;
      case Opcodes::EQ_US4: n="EQ_US4"; break;
      case Opcodes::EQ_UU4: n="EQ_UU4"; break;
      case Opcodes::EQ_SS8: n="EQ_SS8"; break;
      case Opcodes::EQ_SU8: n="EQ_SU8"; break;
      case Opcodes::EQ_US8: n="EQ_US8"; break;
      case Opcodes::EQ_UU8: n="EQ_UU8"; break;

      case Opcodes::NEQ_SS4: n="NEQ_SS4"; break;
      case Opcodes::NEQ_SU4: n="NEQ_SU4"; break;
      case Opcodes::NEQ_US4: n="NEQ_US4"; break;
      case Opcodes::NEQ_UU4: n="NEQ_UU4"; break;
      case Opcodes::NEQ_SS8: n="NEQ_SS8"; break;
      case Opcodes::NEQ_SU8: n="NEQ_SU8"; break;
      case Opcodes::NEQ_US8: n="NEQ_US8"; break;
      case Opcodes::NEQ_UU8: n="NEQ_UU8"; break;

      case Opcodes::LT_SS4: n="LT_SS4"; break;
      case Opcodes::LT_SU4: n="LT_SU4"; break;
      case Opcodes::LT_US4: n="LT_US4"; break;
      case Opcodes::LT_UU4: n="LT_UU4"; break;
      case Opcodes::LT_SS8: n="LT_SS8"; break;
      case Opcodes::LT_SU8: n="LT_SU8"; break;
      case Opcodes::LT_US8: n="LT_US8"; break;
      case Opcodes::LT_UU8: n="LT_UU8"; break;

      case Opcodes::GT_SS4: n="GT_SS4"; break;
      case Opcodes::GT_SU4: n="GT_SU4"; break;
      case Opcodes::GT_US4: n="GT_US4"; break;
      case Opcodes::GT_UU4: n="GT_UU4"; break;
      case Opcodes::GT_SS8: n="GT_SS8"; break;
      case Opcodes::GT_SU8: n="GT_SU8"; break;
      case Opcodes::GT_US8: n="GT_US8"; break;
      case Opcodes::GT_UU8: n="GT_UU8"; break;

      case Opcodes::LEQ_SS4: n="LEQ_SS4"; break;
      case Opcodes::LEQ_SU4: n="LEQ_SU4"; break;
      case Opcodes::LEQ_US4: n="LEQ_US4"; break;
      case Opcodes::LEQ_UU4: n="LEQ_UU4"; break;
      case Opcodes::LEQ_SS8: n="LEQ_SS8"; break;
      case Opcodes::LEQ_SU8: n="LEQ_SU8"; break;
      case Opcodes::LEQ_US8: n="LEQ_US8"; break;
      case Opcodes::LEQ_UU8: n="LEQ_UU8"; break;

      case Opcodes::GEQ_SS4: n="GEQ_SS4"; break;
      case Opcodes::GEQ_SU4: n="GEQ_SU4"; break;
      case Opcodes::GEQ_US4: n="GEQ_US4"; break;
      case Opcodes::GEQ_UU4: n="GEQ_UU4"; break;
      case Opcodes::GEQ_SS8: n="GEQ_SS8"; break;
      case Opcodes::GEQ_SU8: n="GEQ_SU8"; break;
      case Opcodes::GEQ_US8: n="GEQ_US8"; break;
      case Opcodes::GEQ_UU8: n="GEQ_UU8"; break;
	
      case Opcodes::NEG_S4: n="NEG_S4"; break;
      case Opcodes::NEG_U4: n="NEG_U4"; break;
	
	// case Opcodes::NEG_R4, NEG_R8

      case Opcodes::COMPL_S4: n="COMPL_S4"; break;
      case Opcodes::COMPL_U4: n="COMPL_U4"; break;

      case Opcodes::INV_S4: n="INV_S4"; break;
      case Opcodes::INV_U4: n="INV_U4"; break;

      case Opcodes::PRINT: n="PRINT"; break; 
      case Opcodes::EXIT:  n="EXIT"; break;

      default: ERR("unknown opcode");
      }

      form(os, "%06d %04x ", pc, code[pc]);
      pc++;
    
      FORS (i, 2) {
	if (i < f) form(os, "%04x %04x ", code[pc+i] >> 16, code[pc+i] & 0xffff);
	else       os << "          ";
      }

      form(os, "%03d ", fm.code[pc-1]);
      form(os, "%-10s ", n.c_str());
    
      FORS (i, f) {

	switch (t[i]) {

	case Type::TC_pc:    os << (uint4)code[pc+i]; break;
	case Type::TC_uint1: os << (uint4)code[pc+i]; break;
	case Type::TC_sint2: os << (sint4)code[pc+i]; break;
	case Type::TC_uint2: os << (uint4)code[pc+i]; break;
	case Type::TC_sint4: os << (sint4)code[pc+i]; break;
	case Type::TC_uint4: os << (uint4)code[pc+i]; break;
	case Type::TC_sint8: os << *reinterpret_cast<const sint8*>(&code[pc+i]); break;
	case Type::TC_uint8: os << *reinterpret_cast<const uint8*>(&code[pc+i]); break;
	case Type::TC_real4: os << *reinterpret_cast<const real4*>(&code[pc+i]); break;
	case Type::TC_real8: os << *reinterpret_cast<const real8*>(&code[pc+i]); break;

	default: ERR2("unknown type", t[i]);
	}
	os << " ";
      }

      os << endl;
      pc += f;
    }
  }
}
