// $Id: Script.H 5279 2007-06-23 02:49:08Z mburo $ 

// (c) Jonathan Yip, licensed under the GPL


#ifndef Script_H
#define Script_H

#include "Global.H"
#include "Lexer.H"
#include "Mem.H"
#include "API.H"
#include "VarTable.H"


using namespace std;
using namespace OSL;

// fixme: check to ensure all memory is deallocated properly
// fixme: add in private copy constructors/assign operator where necessary




// Class ExpressionNode
// - ABC of all expressions
template <class T>
class ExpressionNode {
public:
  virtual ~ExpressionNode() {}
  virtual T eval(Mem& stack)=0;
};

union Expression {
  ExpressionNode<ScriptObject::Char>*  char_exp;
  ExpressionNode<sint4>*  sint4_exp;
  ExpressionNode<real4>*  float_exp;
  ExpressionNode<string>* string_exp;
  ExpressionNode<ScriptObject*>* obj_exp;
};



// Class ConstNode
// - a constant node with the value of the constant as a private data member
template <class T>
class ConstNode: public ExpressionNode<T> {
public:
  ConstNode(T value) { m_value = value; }
  T eval(Mem& /*stack*/) { return m_value; }
private:
  T m_value;
};


class CharVarNode: public ExpressionNode<ScriptObject::Char> {
public:
  CharVarNode(sint4 loc, bool func_var) : m_loc(loc), m_func_var(func_var) {}

  ScriptObject::Char eval(Mem& stack) {
    sint4 value;
    if (m_func_var)
      value = stack.get_stack_val_from_frame(m_loc);
    else
      value = stack.get_stack_val(m_loc);
    return *reinterpret_cast<uint1*>(&value);
  }
private:
  sint4 m_loc;
  bool m_func_var;
};


class IntVarNode: public ExpressionNode<sint4> {
public:
  IntVarNode(sint4 loc, bool func_var) : m_loc(loc), m_func_var(func_var) {}

  sint4 eval(Mem& stack) {
    sint4 value;
    if (m_func_var)
      value = stack.get_stack_val_from_frame(m_loc);
    else
      value = stack.get_stack_val(m_loc);
    return *reinterpret_cast<sint4*>(&value);
  }
private:
  sint4 m_loc;
  bool m_func_var;
};

class FloatVarNode: public ExpressionNode<real4> {
public:
  FloatVarNode(sint4 loc, bool func_var) : m_loc(loc), m_func_var(func_var) {}

  real4 eval(Mem& stack) {
    sint4 value;
    if (m_func_var) value = stack.get_stack_val_from_frame(m_loc);
    else  value = stack.get_stack_val(m_loc);
    return *reinterpret_cast<real4*>(&value);
  }
private:
  sint4 m_loc;
  bool m_func_var;
};

class StringVarNode: public ExpressionNode<string> {
public:
  StringVarNode(sint4 loc, bool func_var) : m_loc(loc), m_func_var(func_var) {}

  string eval(Mem& stack) {
    sint4 value;
    if (m_func_var) value = stack.get_stack_val_from_frame(m_loc);
    else value = stack.get_stack_val(m_loc);
    return **reinterpret_cast<string**>(&value);
  }
private:
  sint4 m_loc;
  bool m_func_var;
};

class ObjVarNode: public ExpressionNode<ScriptObject*> {
public:
  ObjVarNode(sint4 loc, bool func_var) : m_loc(loc), m_func_var(func_var) {}

  ScriptObject* eval(Mem& stack) {
    sint4 value;
    if (m_func_var) value = stack.get_stack_val_from_frame(m_loc);
    else value = stack.get_stack_val(m_loc);

    return *reinterpret_cast<ScriptObject**>(&value);
  }
private:
  sint4 m_loc;
  bool m_func_var;
};
  
/******************************************/

template <class T>
class NegNode : public ExpressionNode<T> {
public:
  NegNode(ExpressionNode<T> *exp) : m_exp(exp) {}
  ~NegNode() { delete m_exp; }

  T eval (Mem& stack) {
    return -m_exp->eval(stack);
  }

private:
  ExpressionNode<T> *m_exp;
};

template <class T>
class ComplementNode : public ExpressionNode<T> {
public:
  ComplementNode(ExpressionNode<T> *exp) : m_exp(exp) {}
  ~ComplementNode() { delete m_exp; }

  T eval (Mem& stack) {
    return ~m_exp->eval(stack);
  }

private:
  ExpressionNode<T> *m_exp;
};


template <class T>
class InverseNode : public ExpressionNode<T> {
public:
  InverseNode(ExpressionNode<T> *exp) : m_exp(exp) {}
  ~InverseNode() { delete m_exp; }

  T eval (Mem& stack) {
    return !m_exp->eval(stack);
  }

private:
  ExpressionNode<T> *m_exp;
};


template <class T, class V>
class BinaryNode : public ExpressionNode<T> {
public:
  BinaryNode() : m_lhs(0), m_rhs(0), _op("") {}
  ~BinaryNode() { delete m_lhs; delete m_rhs; }

  void init(ExpressionNode<V> *lhs, ExpressionNode<V> *rhs, string op) {
    m_lhs = lhs;
    m_rhs = rhs;
    _op = op;
  }

protected:
  ExpressionNode<V> *m_lhs;
  ExpressionNode<V> *m_rhs;
  string _op;
};

template <class T, class V>
class PlusNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) {
    return this->m_lhs->eval(stack) + this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class MinusNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) {
    return this->m_lhs->eval(stack) - this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class MultiplyNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) { 
    return this->m_lhs->eval(stack) * this->m_rhs->eval(stack); 
  }
};

template <class T, class V>
class DivideNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) { 
    return this->m_lhs->eval(stack) / this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class ModulusNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) {
    return this->m_lhs->eval(stack) % this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class ShiftLeftNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) {
    return this->m_lhs->eval(stack) << this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class ShiftRightNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) {
    return this->m_lhs->eval(stack) >> this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class EqualNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) { 
    return this->m_lhs->eval(stack) == this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class NotEqualNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) {
    return this->m_lhs->eval(stack) != this->m_rhs->eval(stack); 
  }
};

template <class T, class V>
class LessThanNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) {
    return this->m_lhs->eval(stack) < this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class LessEqualNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) { 
    return this->m_lhs->eval(stack) <= this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class GreaterThanNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) { 
    return this->m_lhs->eval(stack) > this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class GreaterEqualNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) { 
    return this->m_lhs->eval(stack) >= this->m_rhs->eval(stack);
  }
};

template <class T, class V>
class AndNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) { 
    return this->m_lhs->eval(stack) && this->m_rhs->eval(stack); 
  }
};

template <class T, class V>
class OrNode: public BinaryNode<T, V> {
public:
  T eval(Mem& stack) { 
    return this->m_lhs->eval(stack) || this->m_rhs->eval(stack); }
};


///////////////////////////////////////////////////////////////

class TypeNode {
public:
  virtual ~TypeNode() {}

  Mem::mem_types get_type() { return m_type; }
protected:
  Mem::mem_types m_type;
};

template<class T>
class KnownTypeNode : public TypeNode {
public:
  KnownTypeNode(ExpressionNode<T>* exp, Mem::mem_types type) : 
    m_exp(exp) {
    m_type = type;
  }

  ExpressionNode<T>* get_exp() { return m_exp; }
private:
  ExpressionNode<T>* m_exp;
};


////////////////////////////////////////////////////////////////





class Node {
public:
  virtual ~Node() {}
  virtual bool exec(Mem& stack)=0;
};

class OpenBlock: public Node {
public:
  bool exec(Mem& stack);
private:
};

class CloseBlock: public Node {
public:
  bool exec(Mem& stack);
private:
  OpenBlock* m_open;
};

class AllocCharNode : public Node {
public:
  AllocCharNode(sint4 loc, bool func_var, ExpressionNode<ScriptObject::Char>* init_value) :
    m_loc(loc), m_func_var(func_var), m_init_value(init_value) {}
  
  bool exec(Mem& stack) {
    if (m_func_var) stack.alloc_stack_from_frame(m_loc, false);
    else stack.alloc_stack(m_loc, false);
    
    sint4 value;
    if (m_init_value) 
      *reinterpret_cast<ScriptObject::Char*>(&value) = m_init_value->eval(stack);
    
    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);



    return true;
  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<ScriptObject::Char>* m_init_value;
};


class AllocIntNode : public Node {
public:
  AllocIntNode(sint4 loc, bool func_var, ExpressionNode<sint4>* init_value) :
    m_loc(loc), m_func_var(func_var), m_init_value(init_value) {}
  
  bool exec(Mem& stack) {
    if (m_func_var) stack.alloc_stack_from_frame(m_loc, false);
    else stack.alloc_stack(m_loc, false);
    
    sint4 value;
    if (m_init_value) 
      *reinterpret_cast<sint4*>(&value) = m_init_value->eval(stack);
    
    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);



    return true;
  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<sint4>* m_init_value;
};


class AllocFloatNode : public Node {
public:
  AllocFloatNode(sint4 loc, bool func_var, ExpressionNode<real4>* init_value) :
    m_loc(loc), m_func_var(func_var), m_init_value(init_value) {}
  
  bool exec(Mem& stack) {
    if (m_func_var) stack.alloc_stack_from_frame(m_loc, false);
    else stack.alloc_stack(m_loc, false);
    
    sint4 value;
    if (m_init_value)
      *reinterpret_cast<real4*>(&value) = m_init_value->eval(stack);

    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);
    return true;

  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<real4>* m_init_value;
};

class AllocStringNode : public Node {
public:
  AllocStringNode(sint4 loc, bool func_var, 
		  ExpressionNode<string>* init_value) :
    m_loc(loc), m_func_var(func_var), m_init_value(init_value) {}
  
  bool exec(Mem& stack) {
    if (m_func_var) stack.alloc_stack_from_frame(m_loc, false);
    else stack.alloc_stack(m_loc, false);

    sint4 value;
    string *s = new string;
    if (m_init_value) 
      *s = m_init_value->eval(stack) ;
    else
      *s = "";

    *reinterpret_cast<string**>(&value) = s;
    
    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);
    return true;

  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<string>* m_init_value;
};

class AllocObjectNode : public Node {
public:
  AllocObjectNode(sint4 loc, bool func_var,
		  ExpressionNode<ScriptObject*>* init_value) :
    m_loc(loc), m_func_var(func_var), m_init_value(init_value) {}
  
  bool exec(Mem& stack) {
    if (m_func_var) stack.alloc_stack_from_frame(m_loc, true);
    else stack.alloc_stack(m_loc, true);
    //    cout << "ALLOCATED object at " << m_loc << endl;
    // fixme: need to be able have an initial value for object either through
    // contstructor or assignment

    return true;

  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<ScriptObject*>* m_init_value;
};


class StmtList: public Node {
public:
  //  StmtList(vector<Node*> stmts) : m_stmts(stmts) {}
  ~StmtList() { FORU(i, m_stmts.size()) delete m_stmts[i]; }
  
  void add(Node* stmt) {
    m_stmts.push_back(stmt);
  }

  bool exec(Mem& stack) {
    FORU(i, m_stmts.size()) 
      if (!m_stmts[i]->exec(stack)) return false;
    return true;
  }
private:
  vector<Node*> m_stmts;
};



class IfStmt : public Node {
public:
  IfStmt() : _cond(0), _if_block(0), _else_block(0) {}
  IfStmt(ExpressionNode<bool> *cond, Node *if_block,
	   Node *else_block) : 
    _if_block(if_block), _else_block(else_block) { _cond = cond;}
  ~IfStmt() { delete _cond; delete _if_block; delete _else_block; }

  bool exec(Mem& stack) {
    if (_cond->eval(stack) != 0)
      return _if_block->exec(stack);
    else if (_else_block) return _else_block->exec(stack);
    return true;
  }

private:
  ExpressionNode<bool> *_cond;
  Node *_if_block;
  Node *_else_block;
};

class WhileStmt: public Node {
public:
  //  WhileStmt() : _cond(0), _block(0) {}
  WhileStmt(ExpressionNode<bool>* cond, Node* block) : 
    _cond(cond), _block(block) {}
  ~WhileStmt() { delete _cond; delete _block; }

  bool exec(Mem& stack) {
    while (_cond->eval(stack) != 0) {
      _block->exec(stack); }
    return true;
  }
     
protected:
  ExpressionNode<bool> *_cond;
  Node* _block;
};

template<class T>
class SwitchStmt: public Node {
public:
  SwitchStmt(ExpressionNode<T>* cond, map<T, Node*> case_blocks,
	       Node* default_block) : _cond(cond), _case_blocks(case_blocks),
					_default_block(default_block) {}
  ~SwitchStmt() {}
  
  bool exec(Mem& stack) {
    T cur_case = _cond->eval(stack);
    if (_case_blocks.find(cur_case) != _case_blocks.end())
      _case_blocks[cur_case]->exec(stack);
    else if (_default_block)
      _default_block->exec(stack);
    return true;
  }

private:
  ExpressionNode<T>* _cond;
  map<T, Node*> _case_blocks;
  Node* _default_block;
};

// fixme take out spaces and endl
template<class T>
class PrintStmt: public Node {
public:
  PrintStmt(ExpressionNode<T>* first, Node* rest) :
    _first(first), _rest(rest) {}
  ~PrintStmt() { delete _first; delete _rest; }

  bool exec(Mem& stack) {
    cout << _first->eval(stack);
    if (_rest) {
      cout << " ";
      _rest->exec(stack);
    }
    else cout << endl;
    return true;
  }
private:
  ExpressionNode<T>* _first;
  // _rest should be of class PrintStmt to work properly
  Node* _rest;
};

class CharAssignStmt: public Node {
public:
  CharAssignStmt(sint4 loc, bool func_var, ExpressionNode<ScriptObject::Char>* expression) :
    m_loc(loc), m_func_var(func_var), m_expression(expression) {}

  bool exec(Mem& stack) {
    sint4 value;
    *reinterpret_cast<ScriptObject::Char*>(&value) = m_expression->eval(stack);

    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);

    return true;
  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<ScriptObject::Char>* m_expression;
};

    

class IntAssignStmt: public Node {
public:
  IntAssignStmt(sint4 loc, bool func_var, ExpressionNode<sint4>* expression) :
    m_loc(loc), m_func_var(func_var), m_expression(expression) {}

  bool exec(Mem& stack) {
    sint4 value;
    *reinterpret_cast<sint4*>(&value) = m_expression->eval(stack);

    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);

    return true;
  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<sint4>* m_expression;
};


class FloatAssignStmt: public Node {
public:
  FloatAssignStmt(sint4 loc, bool func_var, 
		  ExpressionNode<real4>* expression) :
    m_loc(loc), m_func_var(func_var), m_expression(expression) {}

  bool exec(Mem& stack) {
    sint4 value;
    *reinterpret_cast<real4*>(&value) = m_expression->eval(stack);

    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);
    return true;

  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<real4>* m_expression;
};


class StringAssignStmt: public Node {
public:
  StringAssignStmt(sint4 loc, bool func_var,
		   ExpressionNode<string>* expression) :
    m_loc(loc), m_func_var(func_var), m_expression(expression) {}

  bool exec(Mem& stack) {
    sint4 value;
    string* s = new string;
    *s = m_expression->eval(stack);
    *reinterpret_cast<string**>(&value) = s;

    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);
    return true;
  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<string>* m_expression;
};

class ObjAssignStmt: public Node {
public:
  ObjAssignStmt(sint4 loc, bool func_var,
		ExpressionNode<ScriptObject*>* expression) :
    m_loc(loc), m_func_var(func_var), m_expression(expression) {}

  bool exec(Mem& stack) {
    sint4 value;
    ScriptObject* o = m_expression->eval(stack);
    *reinterpret_cast<ScriptObject**>(&value) = o;

    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);
    return true;
  }
private:
  sint4 m_loc;
  bool m_func_var;
  ExpressionNode<ScriptObject*>* m_expression;
};

class ForStmt: public WhileStmt {
public:
  ForStmt(ExpressionNode<bool> *cond, Node *block, vector<Node*> vars,
	    vector<Node*> iter) : WhileStmt(cond, block),
				    _vars(vars), _iter(iter) {
  }
  ~ForStmt() {}

  bool exec(Mem& stack) {
    FORU(i, _vars.size()) _vars[i]->exec(stack);
    while (_cond->eval(stack)) {
      _block->exec(stack);
      FORU(i, _iter.size()) _iter[i]->exec(stack);
    }
    return true;
  }

private:
  vector<Node*> _vars;
  vector<Node*> _iter;
};

template <class T>
class ScriptFunction {
public:
  // Constructor for prototype
  ScriptFunction(vector<Mem::mem_types> param_types) :
    m_defined(false), m_has_param_names(false), m_param_types(param_types) {}
  ~ScriptFunction() {}

  void add_param_names(vector<string> param_names) {
    m_has_param_names = true;
    m_param_names = param_names;
  }

  void init(Node* stmts, sint4 /* fixme: unused block */) { 
    m_defined = true; 
    m_stmts = stmts;
    //    m_block = block;
  }

  vector<Mem::mem_types> get_types() { return m_param_types; }

  bool is_defined() { return m_defined; }

  void set_params(vector<Expression> params, Mem& stack) {
    for (sint4 i = params.size()-1; i>=0; i--) {
      if (m_param_types[i] == Mem::INT) {
	sint4 value;

	*reinterpret_cast<sint4*>(&value) = params[i].sint4_exp->eval(stack);

	stack.set_next_stack_val(value, false);
      }
      else if (m_param_types[i] == Mem::FLOAT) {
	sint4 value;

	*reinterpret_cast<real4*>(&value) = params[i].float_exp->eval(stack);
	
	stack.set_next_stack_val(value, false);
      }
      else if (m_param_types[i] == Mem::STRING) {
	sint4 value;
	
	string* s = new string;
	*s = params[i].string_exp->eval(stack);
	
	*reinterpret_cast<string**>(&value) = s;
	
	stack.set_next_stack_val(value, false);
      }
      else if (m_param_types[i] == Mem::OBJECT) {
	sint4 value;

	ScriptObject* o = params[i].obj_exp->eval(stack);

	*reinterpret_cast<ScriptObject**>(&value) = o;

	stack.set_next_stack_val(value, true);	
      }
      else
	ERR("HUH");
    }
    stack.set_frame_index();


  }

  void set_return(T ret_value) {
    m_return_value = ret_value;
  }

  T eval(Mem& stack) {


    m_stmts->exec(stack);
    // ********** remove that local block of variables







    stack.reset_frame_index();

    return m_return_value;
  }

    
private:
  bool m_defined;
  bool m_has_param_names;

  Mem::mem_types m_type;
  T m_return_value;

  //  sint4 m_block;
  vector<string> m_param_names;
  vector<Mem::mem_types> m_param_types;

  map<uint4, string> m_param_obj_types;


  Node* m_stmts;
};

template <class T>
class FunctionExp: public ExpressionNode<T> {
public:
  FunctionExp(ScriptFunction<T>* function, vector<Expression> params) :
    m_function(function), m_params(params) {
  }
  ~FunctionExp() {}

  
  T eval(Mem& stack) {

    m_function->set_params(m_params, stack);
    return m_function->eval(stack);
  }
  
private:

  ScriptFunction<T>* m_function;
  vector<Expression> m_params;
};

template <class T>
class FunctionStmt: public Node {
public:
  FunctionStmt(ScriptFunction<T>* function, vector<Expression> params) :
    m_function(function), m_params(params) {
  }
  ~FunctionStmt() {}

  
  bool exec(Mem& stack) {
    m_function->set_params(m_params, stack);
    m_function->eval(stack);
    return true;
  }
  
private:

  ScriptFunction<T>* m_function;
  vector<Expression> m_params;

};


template <class T>
class ReturnStmt: public Node {
public:
  ReturnStmt(ScriptFunction<T>* function, ExpressionNode<T>* expression) :
    m_function(function) { m_expression = expression; }
  ~ReturnStmt() { delete m_expression; }

  bool exec(Mem& stack) {
    m_function->set_return(m_expression->eval(stack)); 
    return false;
  }

private:
  ScriptFunction<T>* m_function;
  ExpressionNode<T> *m_expression;
};




class HostFunctionStmt: public Node {
public:
  HostFunctionStmt(sint4 loc, bool func_var, Cfunction cfunc) :
    m_loc(loc), m_func_var(func_var), m_cfunc(cfunc) {}
  ~HostFunctionStmt() {}
  
  bool exec(Mem& stack) {
    sint4 value;
    if (m_func_var) value = stack.get_stack_val_from_frame(m_loc);
    else value = stack.get_stack_val(m_loc);

    ScriptObject* params = *reinterpret_cast<ScriptObject**>(&value);

    m_cfunc(params);
    return true;
  }
  
private:
  sint4 m_loc;
  bool m_func_var;

  Cfunction m_cfunc;
  //  ScriptObject* _params;
};




class HostFunctionExp: public ExpressionNode<sint4> {
public:
  HostFunctionExp(sint4 loc, bool func_var, Cfunction cfunc) :
    m_loc(loc), m_func_var(func_var), m_cfunc(cfunc) {}
  ~HostFunctionExp() {}
  
  sint4 eval(Mem& stack) {
    sint4 value;
    if (m_func_var) value = stack.get_stack_val_from_frame(m_loc);
    else value = stack.get_stack_val(m_loc);

    ScriptObject* params = *reinterpret_cast<ScriptObject**>(&value);

    return m_cfunc(params);
  }
  
private:
  sint4 m_loc;
  bool m_func_var;

  Cfunction m_cfunc;
  //  ScriptObject* _params;
};





template <class T>
class ObjMemberAssignStmt : public Node {
public:
  ObjMemberAssignStmt(sint4 loc, bool func_var, uint4 member_loc, 
		      ExpressionNode<T>* expression) :
    m_loc(loc), m_func_var(func_var), m_member_loc(member_loc), 
    m_expression(expression) {}


  bool exec(Mem& stack) {
    sint4 value;
    if (m_func_var) value = stack.get_stack_val_from_frame(m_loc);
    else value = stack.get_stack_val(m_loc);

    ScriptObject* obj = *reinterpret_cast<ScriptObject**>(&value);

    obj->set_value(m_member_loc, m_expression->eval(stack));

    return true;
  }
  
private:
  sint4 m_loc;
  bool m_func_var;
  uint4 m_member_loc;
  ExpressionNode<T>* m_expression;
};


//%%
#if 0
template <class T>
class ObjFunctionExp: public ExpressionNode<T> {
public:
  ObjFunctionExp(Mem::mem_types type, string name, sint4 block, ScriptFunction<T>* function, vector<Expression> params) :
    m_type(type), m_name(name), m_block(block), m_function(function), m_params(params) {}
  ~ObjFunctionExp() {}

  T eval(Mem& stack) {
    sint8 value = stack.get_value( Mem(m_name, m_type, m_block) );
    uint1 offset = stack.get_offset( Mem(m_name, m_type, m_block) );

    ScriptObject* obj = *(reinterpret_cast<ScriptObject**>(&value) +
			     offset);

    m_function->set_params(m_params, obj->get_stack());
    return m_function->eval(stack);
  }

private:
  Mem::mem_types m_type;
  string m_name;
  sint4 m_block;

  ScriptFunction<T>* m_function;
  vector<Expression> m_params;
  
};

template <class T>
class ObjFunctionStmt: public Node {
public:
  ObjFunctionStmt(Mem::mem_types type, string name, sint4 block, ScriptFunction<T>* function, vector<Expression> params) :
    m_type(type), m_name(name), m_block(block), m_function(function), m_params(params) {}
  ~ObjFunctionStmt() {}

  bool exec(Mem& stack) {
    sint8 value = stack.get_value( Mem(m_name, m_type, m_block) );
    uint1 offset = stack.get_offset
      ( Mem(m_name, m_type, m_block) );

    ScriptObject* obj = *(reinterpret_cast<ScriptObject**>(&value) +
			     offset);

    m_function->set_params(m_params, obj->get_stack());
    m_function->eval(stack);
    return true;
  }

private:
  Mem::mem_types m_type;
  string m_name;
  sint4 m_block;

  ScriptFunction<T>* m_function;
  vector<Expression> m_params;
  
};
#endif
//%%

class NewObjStmt : public Node {
public:
  NewObjStmt(sint4 loc, bool func_var, ScriptObjectType* definition) :
    m_loc(loc), m_func_var(func_var), m_definition(definition) {}

  bool exec(Mem& stack) {
    sint4 value;
    if (m_func_var) value = stack.get_stack_val_from_frame(m_loc);
    else value = stack.get_stack_val(m_loc);

    cout << "NEWING object " << m_loc << endl;

    /*
    ScriptObject *o = new ScriptObject(m_definition, stack);
    *reinterpret_cast<ScriptObject**>(&value) = o;
    */


    sint4* heap_ptr = stack.alloc_heap(m_definition->size());

    *reinterpret_cast<ScriptObject*>(heap_ptr) = ScriptObject(m_definition);
    *reinterpret_cast<sint4**>(&value) = heap_ptr;
    //    stack.alloc_heap(m_definition->size());

    if (m_func_var) stack.set_stack_val_from_frame(m_loc, value);
    else stack.set_stack_val(m_loc, value);
    return true;
  }
private:
  sint4 m_loc;
  bool m_func_var;
  ScriptObjectType* m_definition;

};

// fixme !!!!
class DelObjStmt: public Node {
public:
  DelObjStmt(Mem::mem_types type, string name, sint4 block,
	       ScriptObjectType* definition) :
    m_type(type), m_name(name), m_block(block), m_definition(definition) {}


  bool exec(Mem& stack) {
    //    stack.del_mem( Mem(m_name, m_type, 0, m_block), 4 );
    // fixme
    Mem s = stack;
    return true;
  }
private:
  Mem::mem_types m_type;
  string m_name;
  sint4 m_block;
  ScriptObjectType* m_definition;

};

template <class T>
class ObjMemberVarNode : public ExpressionNode<T> {
public:
  ObjMemberVarNode(sint4 loc, bool func_var, uint4 member_loc) :
    m_loc(loc), m_func_var(func_var), m_member_loc(member_loc) {}


  T eval(Mem& stack) { 
    T dummy;



    sint4 value;
    if (m_func_var) value = stack.get_stack_val_from_frame(m_loc);
    else value = stack.get_stack_val(m_loc);


    ScriptObject* obj = *reinterpret_cast<ScriptObject**>(&value);
    
    return obj->get_value(m_member_loc, dummy);
  }

private:
  sint4 m_loc;
  bool m_func_var;
  uint4 m_member_loc;
};



class ScriptInstance {
public:
  ScriptInstance(Node* root, Node* main, uint4 index) :
    _root(root), m_main(main), _index(index) {}    
  ~ScriptInstance() {}
  
  // check to see if index is valid perhaps?
  void eval() {
    cout << "IN AI SCRIPT " << " using index " << _index << endl;

    // took out:    if (m_init) m_init->exec(stack, _index, 0);



    if (_root)
      _root->exec(m_stack);


    if (m_main) m_main->exec(m_stack);

  }



private:
  // took out:  StmtList* m_init;
  Node* _root;
  Node* m_main;

  Mem m_stack;



  uint4 _index;
};






class Script {
public:
  Script() : m_index(0) {}
  ~Script() {}


  // add c functions
  void register_func(Cfunction c_func, string script_func_name);
		   
  sint4 load(string filename) {
    _lex.init(filename);
    return parse();
  }

  ScriptInstance* new_instance() {
    ScriptInstance* script = new ScriptInstance(m_root, m_main, 
						m_index++);
    return script;
  }






private:
  sint4 m_index;
  Lexer _lex;

  Node* m_root;
  // took out:  StmtList* m_init;
  Node* m_main;

  VarTable m_vars;
  FuncTable m_funcs;

  // ??? what is difference between _objs and m_objs_vars???

  // This is to store the different objects
  map<string, ScriptObjectType*> _objs;

  // This is to store the var names associated with specific object
  //  map<string, ScriptObjectType*> m_objs_vars;

  map<string, uint4> m_cfuncs_ids;
  vector<Cfunction>  m_cfuncs;

  map<string, ScriptFunction<sint4>*>  m_int_functions;
  map<string, ScriptFunction<real4>*>  m_float_functions;
  map<string, ScriptFunction<string>*> m_string_functions;
  map<string, ScriptFunction<bool>*>   m_void_functions;
  map<string, ScriptFunction<ScriptObject>*> m_obj_functions;

  
  // This heap is shared by all objs with same script
  // ??? should we just have 1 heap...or many heaps
  Mem m_mem;
  

  ///// other stuff






  string m_current_func_name;
  Mem::mem_types m_current_func_type;

  //  bool m_parsing_func;
  string _current_obj_name;



  sint4 parse();

  void parse_extern();

  //  void parse_vars();
  //  void parse_prototypes();

  void parse_var_declaration(StmtList* stmts);

  void parse_main();

  IfStmt* parse_if();
  WhileStmt* parse_while();
  ForStmt* parse_for();

  Node* parse_switch();

  Node* parse_return();
  

  //  Node* parse_class_init(const string class_name);
  
  Node* parse_assign(const string var_name);

  Node* parse_obj_variable(const string var_name, sint4 block, bool func_var);
  Node* parse_obj_assign(const string var_name, sint4 block, bool func_var);


  Node* parse_print(bool check);

  void parse_function_body();
  void parse_obj();

  Node* parse_stmt();
  StmtList* parse_stmt_list(const TokenType expected_end_marker);
  

  TypeNode* parse_exp(const TokenType expected_end_marker,
			const Mem::mem_types expected_type);



  TypeNode* parse_till_low_op(sint4 prec, const Mem::mem_types expected_type);


  TypeNode* get_term(const Mem::mem_types expected_type);


  TypeNode* parse_func_exp(const string func_name, Mem::mem_types type,
			   bool is_obj, sint4 block);

  Node* parse_func_stmt(const string func_name, Mem::mem_types type,
			bool is_obj, sint4 block);

  TypeNode* parse_host_func_exp(const string host_func_name);

  Node* parse_host_func_stmt(const string host_func_name);

  sint4 precedence_of(sint4 op);

};


#endif

