// DCDT template function implementations

#include "Global.h"
#include "DCDTLite.h"
#include "HalfEdge.h"
#include "Face.h"

using namespace std;

 
DCDT::DCDT(Mesh &m, int xSectors, int ySectors)
  : _mesh(m)
{
  _xSectors = xSectors;
  _ySectors = ySectors;

  
 //initialize sectors
 sectors.resize(extents[_xSectors][_ySectors]);
  for (int i = 0; i < ySectors; ++i) {
    for (int j = 0; j < xSectors; ++j) {
      sectors[i][j] = -1;
    }
  }
}

 
DCDT::~DCDT()
{
}

 
int DCDT::interHoriz(C& px,  C& py,
                     C& p1x, C& p1y,
                     C& p2x, C& p2y)
{
  if (p1y > p2y)  { 
    C tmp1 = p1x;
    p1x = p2x;
    p2x = tmp1;
        
    C tmp2 = p1y;
    p1y = p2y;
    p2y = tmp2;
  } // swap

  if (p1y >= py) return false; // not intercepting
  if (p2y <  py) return false; // not intercepting or = max 

  C numa  = p1x*(p2y-p1y) + (py-p1y) * (p2x-p1x) ;
  C denom = (p2y-p1y);

  if (denom >= Geometry::zero) {
    C lhs = px * denom;
    return lhs < numa;
  } else {
    C lhs = px * denom;
    return lhs > numa;
  }
}

 
int DCDT::insidePolygon(C& x, C& y)
{
  if (_polygons.size() <= 1) return -1; // if there is only the domain, return.

  std::list<Polygon<V*> >::iterator p = _polygons.begin();
  ++p; // skip boundary
     
  for (; p!=_polygons.end(); ++p) { 

    int cont = 0;
    Polygon<V*>& pol = *p;
    size_t size = pol.size();
      
    for (size_t j=0; j < size; j++) {
      V& v1 = *pol[j];
      V& v2 = *pol[(j+1) % size];
      cont ^= interHoriz(x, y, v1.getX(), v1.getY(), v2.getX(), v2.getY());
    }

    if (cont) { 
      return pol.getID();
    }
  }
   
  return -1;
}

 
bool DCDT::hasCoincidentSegments(V* v1, V* v2, H** coincidentEdge)
{
  C x1, y1, x2, y2, x3, y3, x4, y4;
  x1 = v1->getX();
  y1 = v1->getY();
  x2 = v2->getX();
  y2 = v2->getY();

  H* e1 = v1->getOutEdge();
  H* e2 = e1;
     
  do {
    V* v3 = e1->getDestVertex();
    //V* v3 = e1->leftEnd;
           
    //V* v4 = e1->rightEnd;

    x3 = v3->getX();
    y3 = v3->getY();
         
    if(Geometry::onSegment(x1,y1,x2,y2,x3,y3) || Geometry::onSegment(x1,y1,x3,y3,x2,y2)) {
      *coincidentEdge = e1;
      return true;
    }
               
    e1 = e1->getOppEdge()->getNext();
    
  } while (e1 != e2);

  return false;
}

 
DCDT::F* DCDT::locateStartTriangle(V* v1, V* v2)
{
  H* e1 = v1->getOutEdge()->getNext();
  H* e2 = e1;
     
  do {
    V* v3 = e1->getOrgVertex();
    V* v4 = e1->getDestVertex();
         
    if (Geometry::segments_intersect(v1->getX(),v1->getY(),v2->getX(),v2->getY(),
                                     v3->getX(),v3->getY(),
                                     v4->getX(),v4->getY())) {
      return e1->getFace();
    }
           
    e1 = e1->getNext()->getOppEdge()->getNext();

  } while(e1 != e2);

  //collinear case? should be handled by trivial case prior to common insert!

  H* e3 = v1->getOutEdge();
  //FIXME: not needed? H* e4 = e3;
     
  do {
    if (Geometry::onSegment(v1->getX(), v1->getY(),
                            v2->getX(), v2->getY(),
                            e3->getDestVertex()->getX(), e3->getDestVertex()->getY())) {
      return e3->getFace();
    }
           
    e3 = e3->getPrior()->getOppEdge();
    
  } while (e1 != e2);

  printf("Start face not found, Stop!");
  exit(-1);
}

 
void DCDT::removeVertexIfPossible(V* v, int /*id*/)  // fixme: id not used?
{
  std::vector<H*> incidentConstraints;
 
  H* e1 = v->getOutEdge();
  H* e2 = e1;
  int numberOfConstrainedEdges = 0;
     
  do {
    if (e1->isConstrained()) {
      numberOfConstrainedEdges++;
      incidentConstraints.push_back(e1);

    }
    e1 = e1->getPrior()->getOppEdge();
  } while(e1 != e2);

  //simple vertex removal
  if (numberOfConstrainedEdges == 0) {
    removeVertex(v);
    return;
  }

  if (numberOfConstrainedEdges != 2) {
    return;
  }

  //remove the vertex on a constrained edge
     
  V* v1 = incidentConstraints[0]->getDestVertex();
  V* v2 = incidentConstraints[1]->getDestVertex();
    
  //incidentConstraints[0] and incidentConstraints[1] are not collinear
  //v is an extremity of both constraints
  //(the V shape case)
  if (!Geometry::onSegment(v1->getX(), v1->getY(),
                           v2->getX(), v2->getY(),
                           v->getX(), v->getY())) {
    return;
  }

  if (incidentConstraints[1]->getLeftEnd().first && incidentConstraints[1]->getRightEnd().first) {
    v1 = incidentConstraints[1]->getLeftEnd().second;
    v2 = incidentConstraints[1]->getRightEnd().second;
  }

  std::list<int> oldids = incidentConstraints[1]->getConstraintIDs();
  removeVertex(v);

  insertLineConstraint(v1 , v2, oldids );
}

 
void DCDT::removeVertex(V* v)
{
  std::list<H*> elist;
  std::list<H*> outEdges;

  H* e1 = v->getOutEdge();
  H* e2 = e1;
     
  do {
    elist.push_front(e1->getNext()); 
    outEdges.push_front(e1);
    e1 = e1->getPrior()->getOppEdge();
  } while(e1 != e2);

  F* newFace;
  //process the edges
  for(std::list<H*>::iterator e = outEdges.begin(); e != outEdges.end(); ++e) {
         
    F* f1 = (*e)->getFace();
    F* f2 = (*e)->getOppEdge()->getFace();

    if (f1 == f2) {
      // _mesh.removeEdge((*e));
      newFace = _mesh.joinFaces(f1, f2, (*e) );
    } else {
      newFace = _mesh.joinFaces(f1, f2, (*e) );
    }
         
  }

  _mesh.removeVertex(v);
  triangulateFace(elist);
  //  TriangulatePseudoPolygon(vlist,baseEdge);
}


//collect polygon edges, vertices, intermediate steiner points


// fixme: does this also work if constraints have been intersected by other constraints?

void DCDT::findVerticesAndEdges(int polygonid, std::vector<V*>& vs, std::vector<H*>& es)
{
  vs.clear();
  es.clear();
  
  const Polygon<V*>* pol = getPolygon(polygonid);
  
  if (!pol || pol->empty()) return;
  
  V* v1 = (*pol)[0];
  std::list<V*> stack; 

  _mesh.beginMarking();

  //initialize the stack with a polygon vertex.
  stack.push_back(v1);
     
  while(stack.size() > 0) {
    V* v = stack.front();
    stack.pop_front();
    
    H* e1 = v->getOutEdge();
    H* e2 = e1;
    do {
      if (e2->isConstrained() && !_mesh.marked(e2) ) {
        if (e2->hasConstraintID( polygonid) ) {
          _mesh.mark(e2);
          _mesh.mark(e2->getOppEdge());
          es.push_back(e2);
          stack.push_back(e2->getDestVertex());
          vs.push_back(e2->getDestVertex());
          break;
        }
      }

      e2 = e2->getPrior()->getOppEdge();

    } while (e2 != e1);
  }
  _mesh.endMarking();
}

 
void DCDT::removePolygon(int polygonid)
{
  if (polygonid == 0) {
    printf("Domain cannot be removed by removePolygon()");
    exit(0);
  }

  std::vector<V*> vs;
  std::vector<H*> es;

  // find all vertices and edges:
  findVerticesAndEdges(polygonid,vs,es); 
   

  // Remove all ids which are equal to polygonid in the edge constraints
  for (std::vector<H*>::iterator e=es.begin(); e!=es.end(); ++e) { 
    _mesh.removeConstraint(*e, polygonid); // remove all occurences
  }

  for (std::vector<V*>::iterator v=vs.begin(); v!=vs.end(); ++v) { 
    removeVertexIfPossible(*v, polygonid); // remove all occurences
  }

  //remove the polygon
  for (std::list<Polygon<V*> >::iterator p = _polygons.begin(); p!= _polygons.end(); p++) {
    if (p->getID() == polygonid) {
      _polygons.erase(p);
      break;
    }
  }
}


void DCDT::init(const IPolygon& domain, int radius)
{
  // compute min,max rectangle from polygon

 int _xmin=domain[0].getX();
 int _ymin=domain[0].getY();
 int _xmax=domain[2].getX();
 int _ymax=domain[2].getY();


 // int marginX = (_xmax-_xmin)/5;
  //int marginY = (_ymax-_ymin)/5;
 
 int marginX = 0;
 int marginY = 0;

  if (_xmax == _xmin || _ymax == _ymin ||
      _xmin-marginX < 0 || _xmax+marginX >= INTERSECTOVERFLOWBOUND ||
      _ymin-marginY < 0 || _ymax+marginY >= INTERSECTOVERFLOWBOUND) {
    printf("coordinates out of bounds in init()\n");
    exit(-1);
  }

  if (marginX > 0 && marginY>0) { 
    _xmax += marginX;
    _xmin -= marginX;
    _ymax += marginY;
    _ymin -= marginY;
  }
   
  // creates the triangulated external rectangle
  _mesh.init(_xmax, _ymax);
  // Inserts the vertices of the domain according to radius:
  if ( radius > 0 )
  { 
	insertPolygon ( domain, 0 );
  }
   

}



 
DCDT::V* DCDT::findOrInsertVertex(C& x, C& y)
{
  F* resultFace;
  H* resultEdge;
  V* resultVertex;
  DCDT::LocateResult resType;

         
  //std::cout<<"inserting or find Vertex (" <<x <<" "<< y<<")" <<std::endl;
  resType = locatePoint(NULL, x, y, &resultFace,&resultEdge,&resultVertex);

  if (resType == DCDT::VertexFound) {
    //printf ( "Duplicated vertex, use existing vertex.\n" );
    return resultVertex;
  } else if (resType == DCDT::EdgeFound) {
    //printf ( "Inserted in an edge.\n" );
    return insertPointOnEdge ( resultEdge, x, y, resultEdge->getLeftEnd().second, resultEdge->getRightEnd().second );
  } else if( resType==DCDT::TriangleFound) {
    // a face was found 
    //printf ( "Inserted in a triangle.\n" );
    return insertPointInFace ( resultFace, x, y );
  } else {
    printf ( "Error: Not in domain, or faces not ccw!\n" );
    exit(-1);
  }
     
}

 
void DCDT::insertConstraint(C& x1, C& y1,
                            C& x2, C& y2,
                            int id)
{
  V* v1 = findOrInsertVertex(x1, y1);
  V* v2 = findOrInsertVertex(x2, y2);

  std::cout<<"inserting constraint "<<id <<" (" << x1 <<" "<< y1<< " "<< x2 <<" "<< y2<<")" <<std::endl;
  insertLineConstraint(v1, v2, id);
}

 
void DCDT::insertPolygon(const IPolygon& pol, int id)
{
  V* v;
  Polygon<V*> constrainedVertices;

  for (size_t i=0; i < pol.size(); i++) {

    const IPoint& p = pol[i];

    if (!inputOK(p)) { ERR("input out of bounds " << p.getX() << " " << p.getY()); }
    
    C px = p.getX();
    C py = p.getY();

    v = findOrInsertVertex(px, py);
    constrainedVertices.push_back(v);        
  }

  for (size_t i=0; i < constrainedVertices.size()-1; i++) {
    insertLineConstraint(constrainedVertices[i], constrainedVertices[i+1], id);
  }

  if (pol.isClosed()) {
    // close the polygon
    insertLineConstraint(constrainedVertices[constrainedVertices.size()-1],
                         constrainedVertices[0],
                         id);
  }

  constrainedVertices.setID(id);
  _polygons.push_back(constrainedVertices);
}


 
const Polygon<DCDT::V*>* DCDT::getPolygon(int id)
{
  FORALL (_polygons, it) {
    if ((*it).getID() == id) {
      return &*it;
    }
  }

  return 0;
}

 
void DCDT::findUnconstrainedCrossEdges(V* v1, V* v2,
                                       std::list<V*>& pu,
                                       std::list<V*>& pl,
                                       std::list<H*>& ul,
                                       V* /*left*/, V* right)
// fixme: left not used?
{
  F* t = locateStartTriangle(v1, right); 
  V* v = v1;
  H* shareEdge;
  _mesh.getOpposedFace(t, v,&shareEdge);

  if(!shareEdge->isConstrained()) {
    ul.push_back(shareEdge);
  }
    
  //initialize the vetex list pu and pl
  if(_mesh.vertexAboveSegment2(v1, v2,shareEdge->getOrgVertex())) {
    pu.push_back(shareEdge->getOrgVertex());
    pl.push_front(shareEdge->getDestVertex());
  } else {
    pu.push_back(shareEdge->getDestVertex());
    pl.push_front(shareEdge->getOrgVertex());
  }
    
  V* vSeg = shareEdge->getOppEdge()->getNext()->getDestVertex();

  while(vSeg != v2) {
    if(_mesh.vertexAboveSegment2(v1,v2,vSeg)) {
      pu.push_back(vSeg);
      shareEdge = shareEdge->getOppEdge()->getNext();
    } else {
      pl.push_front(vSeg);
      shareEdge = shareEdge->getOppEdge()->getPrior();
    }
        
    if(!shareEdge->isConstrained()) {
      ul.push_back(shareEdge);
    }
    vSeg = shareEdge->getOppEdge()->getNext()->getDestVertex();

  }//end while
}


 
void DCDT::findConstrainedCrossEdges(V* v1, V* v2,
                                     std::list<std::pair<int,H*> >& cl,
                                     std::vector<V*>& pm)
{
  int vertexIndex = 0;
  F* t = locateStartTriangle(v1, v2); 
  V* v = v1;
  H* shareEdge;
  _mesh.getOpposedFace(t, v,&shareEdge);

  if (shareEdge->isConstrained()) {
    cl.push_back(std::pair<int,H* >(vertexIndex,shareEdge));
    pm.push_back(NULL);
    vertexIndex++;
  }
   
  V* vSeg = shareEdge->getOppEdge()->getNext()->getDestVertex();

  while (vSeg != v2) {
       
    if (_mesh.vertexAboveSegment(v1,v2,vSeg)) {
           
      shareEdge = shareEdge->getOppEdge()->getNext();
      
    } else if(_mesh.vertexBelowSegment(v1,v2,vSeg)) {
           
      shareEdge = shareEdge->getOppEdge()->getNext()->getNext();

    } else {
      
      shareEdge = shareEdge->getOppEdge()->getNext()->getNext();
      pm.push_back(vSeg);
      vertexIndex++;
    }
     
    if (shareEdge->isConstrained()) {
      cl.push_back(std::pair<int,H* >(vertexIndex,shareEdge));
      pm.push_back(NULL);
      vertexIndex++;
    }
        
    vSeg = shareEdge->getOppEdge()->getNext()->getDestVertex();
  }//end while
}

#if 0 
void DCDT::findCrossVertices(V* v1, V* v2, std::list<V*>& vl)
{
  // vl.clear(); fixme: added this (vl could be non-empty, check whether users of this function actually
  // want points added to vl ...
  F* t = locateStartTriangle(v1, v2, vl);  // fixme: undefined
  V* v = v1;
  H* shareEdge;
  _mesh.getOpposedFace(t, v, shareEdge);
  
  V* vSeg = shareEdge->getOppEdge()->getNext()->getDestVertex();

  while (vSeg != v2) {
    
    if(_mesh.vertex_above_segment(v1,v2,vSeg)) {

      shareEdge = shareEdge->getOppEdge()->getNext();

    } else if (_mesh.vertex_below_segment(v1,v2,vSeg)) {
           
      shareEdge = shareEdge->getOppEdge()->getPrior();

    } else {

      vl.push_back(vSeg);
    }
                   
    vSeg = shareEdge->getOppEdge()->getNext()->getDestVertex();

  }//end while
}

#endif

 
void DCDT::insertLineConstraintSimple(V* v1, V* v2,int id,
                                                     V* left, V* right)
{
  //if the insertion is connected, just mark the unconstrained edge as constrained
  H* connectedEdge;
  if( _mesh.isConnected(v1, v2, &connectedEdge)) {
    _mesh.addConstraint(connectedEdge,id,left,right);
    return;
  }

  H* coincidentEdge;
  if(hasCoincidentSegments(v1,v2,&coincidentEdge)) {
         
    C x1, y1, x2, y2, x3, y3, x4, y4;
    x1 = v1->getX();
    y1 = v1->getY();
    x2 = v2->getX();
    y2 = v2->getY();

    x3 = coincidentEdge->getOrgVertex()->getX();
    y3 = coincidentEdge->getOrgVertex()->getY();
          
    x4 = coincidentEdge->getDestVertex()->getX();
    y4 = coincidentEdge->getDestVertex()->getY();
    if (Geometry::onSegment(x1,y1,x2,y2,x4,y4)) {
               
      _mesh.addConstraint(coincidentEdge,id,left,right);
              
      insertLineConstraint(coincidentEdge->getDestVertex(),v2,id);
      return;
    }

    if (Geometry::onSegment(x3,y3,x4,y4,x2,y2)) {
      V* target = coincidentEdge->getOrgVertex();

      if (!coincidentEdge->isConstrained()) {
        V* theIntersect = insertPointOnEdge(coincidentEdge,x2,y2,left,right);

        H* e1 = theIntersect->getOutEdge();
        H* e2 = e1;
                   
        do {                  
                         
          V* v4 = e1->getDestVertex();

          if(v4 == target) {
            _mesh.addConstraint(e1,id,left,right);
            break;

          }
                         
          e1 = e1->getOppEdge()->getNext();
        } while (e1 != e2);

      } else {
        /*FIXME: result not needed? V* theIntersect =*/ insertPointOnEdge(coincidentEdge,x2,y2,left,right);
      }
    }
  } else {

    //the common case: insertion starts in a triangle and crosses at least one unconstrained edge

    std::list<H*> ul;
       
    //upper vertices
    std::list<V*> pu;
    //lower vertices
    std::list<V*> pl;

    findUnconstrainedCrossEdges(v1, v2,pu,pl,ul,left,right);

    F* newFace = 0;
    //process the edges
    for (std::list<H*>::iterator e = ul.begin(); e != ul.end(); ++e) {
      F* f1 = (*e)->getFace();
      F* f2 = (*e)->getOppEdge()->getFace();
      newFace = _mesh.joinFaces(f1, f2, (*e) );
    }
          
    H* constrainedEdge = _mesh.splitFace(newFace,v1,v2);
         
    _mesh.addConstraint(constrainedEdge,id,left,right);

    triangulatePseudoPolygon(pu,constrainedEdge);
    triangulatePseudoPolygon(pl,constrainedEdge->getOppEdge());
          
  }//end else
}

 
void DCDT::insertLineConstraint(V* v1, V* v2, int id)
{   
  //if the insertion is connected, just mark the unconstrained edge as constrained

  H* connectedEdge;
  if (_mesh.isConnected(v1, v2, &connectedEdge)) {
    _mesh.addConstraint(connectedEdge,id,v1,v2);
    return;
  }
    
  //checks if the insertion constains an existing edge
  //if so mark the existing edge constrained
  //and insert the remaining part using insertLineConstraint 
  H* coincidentEdge;

  if (hasCoincidentSegments(v1,v2,&coincidentEdge)) {
         
    C x1, y1, x2, y2, x3, y3, x4, y4;
    x1 = v1->getX();
    y1 = v1->getY();
    x2 = v2->getX();
    y2 = v2->getY();

    x3 = coincidentEdge->getOrgVertex()->getX();
    y3 = coincidentEdge->getOrgVertex()->getY();
          
    x4 = coincidentEdge->getDestVertex()->getX();
    y4 = coincidentEdge->getDestVertex()->getY();
    if (Geometry::onSegment(x1,y1,x2,y2,x4,y4)) {
               
      _mesh.addConstraint(coincidentEdge,id,v1,v2);
      insertLineConstraint(coincidentEdge->getDestVertex(),v2,id);
      return;
    }

    if (Geometry::onSegment(x3,y3,x4,y4,x2,y2)) {
      V* theIntersect = insertPointOnEdge(coincidentEdge,x2,y2);
      H* e1 = theIntersect->getOutEdge();
      H* e2 = e1;
               
      do {                  
                     
        V* v4 = e1->getDestVertex();

        if (v4 == coincidentEdge->getOrgVertex()) {
          _mesh.addConstraint(e1,id,v1,v2);
          break;

        }
                     
        e1 = e1->getOppEdge()->getNext();
            
      } while (e1 != e2);
    }

  } else {

    //the common case: insertion starts in a triangle and crosses at least one edge
          
    std::list<H*> ul;
    std::list<std::pair<int,H*> > cl;

    std::vector<V*> pm;

    //upper vertices
    std::list<V*> pu;
    //lower vertices
    std::list<V*> pl;
        
    findConstrainedCrossEdges(v1, v2, cl, pm);

    int xmax = _mesh.getXMax();
    int ymax = _mesh.getYMax();

    FORALL (cl, it) {
      
      C vx, vy;
      //H* e = (*it).second;

      // fixme: always false
      if (xmax > V::INTERSECTION_LIMIT || ymax > V::INTERSECTION_LIMIT) {
        Geometry::exactFindIntersect(v1->getX(), v1->getY(), v2->getX(), v2->getY(),
                                     (*it).second->getLeftEnd().second->getX(),
                                     (*it).second->getLeftEnd().second->getY(),
                                     (*it).second->getRightEnd().second->getX(),
                                     (*it).second->getRightEnd().second->getY(),
                                     vx, vy);
      } else {
        
        Geometry::findIntersect(v1->getX(), v1->getY(), v2->getX(), v2->getY(),
                                (*it).second->getLeftEnd().second->getX(),
                                (*it).second->getLeftEnd().second->getY(),
                                (*it).second->getRightEnd().second->getX(),
                                (*it).second->getRightEnd().second->getY(),
                                vx,vy);
      }
      
      V* theIntersection = insertPointOnEdge((*it).second,
                                             vx, vy,
                                             (*it).second->getLeftEnd().second,
                                             (*it).second->getRightEnd().second);
         
      pm[(*it).first] = theIntersection;
    }
      
   
      V* vStart = v1;
      pm.push_back(v2);
        
      //insert constrained intervals

      FORALL (pm, it) {
        // for (std::vector<V*>::iterator pm = pm.begin(); pm != pm.end(); ++pm) {
        /* if(count ==2 && id == 1) 
           {
           printf("");
           }*/
        insertLineConstraintSimple(vStart, *it, id, v1, v2);
        vStart = *it; // was (*pm); fixme: what is vStart used for?
        //count--;
      }
    
    
  }//end else
}

 
void DCDT::insertLineConstraintSimple(V* v1, V* v2,std::list<int>& ids,
                                                     V* left, V* right)
{
  //if the insertion is connected, just mark the unconstrained edge as constrained
  H* connectedEdge;
  if (_mesh.isConnected(v1, v2, &connectedEdge)) {
    _mesh.addConstraint(connectedEdge,ids,v1,v2);
    return;
  }

  H* coincidentEdge;
  if (hasCoincidentSegments(v1, v2, &coincidentEdge)) {
         
    C x1, y1, x2, y2, x3, y3, x4, y4;
    x1 = v1->getX();
    y1 = v1->getY();
    x2 = v2->getX();
    y2 = v2->getY();

    x3 = coincidentEdge->getOrgVertex()->getX();
    y3 = coincidentEdge->getOrgVertex()->getY();
          
    x4 = coincidentEdge->getDestVertex()->getX();
    y4 = coincidentEdge->getDestVertex()->getY();
    
    if (Geometry::onSegment(x1,y1,x2,y2,x4,y4)) {
      _mesh.addConstraint(coincidentEdge,ids,v1,v2);
      insertLineConstraint(coincidentEdge->getDestVertex(),v2,ids);
      return;
    }

    if (Geometry::onSegment(x3,y3,x4,y4,x2,y2)) {
      V* theIntersect = insertPointOnEdge(coincidentEdge,x2,y2);
      H* e1 = theIntersect->getOutEdge();
      H* e2 = e1;
               
      do {                  
        V* v4 = e1->getDestVertex();

        if (v4 == coincidentEdge->getOrgVertex()) {
          _mesh.addConstraint(e1,ids,v1,v2);
          break;
        }
        e1 = e1->getOppEdge()->getNext();
      } while(e1 != e2);
    }

  } else {

    //the common case: insertion starts in a triangle and crosses at least one unconstrained edge

    std::list<H*> ul;
       
    //upper vertices
    std::list<V*> pu;
    //lower vertices
    std::list<V*> pl;

    findUnconstrainedCrossEdges(v1, v2, pu, pl, ul, v1, v2);

    F* newFace = 0;
    //process the edges

    FORALL (ul, it) {
      // for (std::list<H*>::iterator e = ul.begin(); e != ul.end(); ++e) {
      F* f1 = (*it)->getFace();
      F* f2 = (*it)->getOppEdge()->getFace();
      newFace = _mesh.joinFaces(f1, f2, *it);
    }
          
    H* constrainedEdge = _mesh.splitFace(newFace, v1, v2);
         
    _mesh.addConstraint(constrainedEdge, ids, left, right);
    triangulatePseudoPolygon(pu, constrainedEdge);
    triangulatePseudoPolygon(pl, constrainedEdge->getOppEdge());
  }//end else
}


  //insert with multiple ids. restore after vertex removal
 
void DCDT::insertLineConstraint(V* v1, V* v2,std::list<int>& ids)
{   //if the insertion is connected, just mark the unconstrained edge as constrained
  H* connectedEdge;
  if(_mesh.isConnected(v1, v2, &connectedEdge)) {
    _mesh.addConstraint(connectedEdge,ids,v1,v2);
    return;
  }

  //checks if the insertion constains an existing edge
  //if so mark the existing edge constrained
  //and insert the remaining part using insertLineConstraint 
  H* coincidentEdge;
  if (hasCoincidentSegments(v1, v2, &coincidentEdge)) {
         
    C x1, y1, x2, y2, x3, y3, x4, y4;
    x1 = v1->getX();
    y1 = v1->getY();
    x2 = v2->getX();
    y2 = v2->getY();

    x3 = coincidentEdge->getOrgVertex()->getX();
    y3 = coincidentEdge->getOrgVertex()->getY();
          
    x4 = coincidentEdge->getDestVertex()->getX();
    y4 = coincidentEdge->getDestVertex()->getY();

    if (Geometry::onSegment(x1,y1,x2,y2,x4,y4)) {
      _mesh.addConstraint(coincidentEdge,ids,v1,v2);
      insertLineConstraint(coincidentEdge->getDestVertex(),v2,ids,v1,v2);
      return;
    }

    if (Geometry::onSegment(x3,y3,x4,y4,x2,y2)) {
      V* theIntersect = insertPointOnEdge(coincidentEdge,x2,y2);
      H* e1 = theIntersect->getOutEdge();
      H* e2 = e1;
               
      do {                  
        V* v4 = e1->getDestVertex();

        if (v4 == coincidentEdge->getOrgVertex()) {
          _mesh.addConstraint(e1,ids,v1,v2);
          break;
        }
        e1 = e1->getOppEdge()->getNext();
      } while(e1 != e2);
    }

  } else {

    //the common case: insertion starts in a triangle and crosses at least one edge
          
    std::list<H*> ul;
    std::list<std::pair<int,H*> > cl;

    std::vector<V*> pm;
    //upper vertices
    std::list<V*> pu;
    //lower vertices
    std::list<V*> pl;
        
    findConstrainedCrossEdges(v1, v2,cl,pm);

    int xmax = _mesh.getXMax();
    int ymax = _mesh.getYMax();
    
    FORALL (cl, it) {
      
      C  vx, vy;
      //H* e = (*cl).second;


      // fixme: xmax/ymax never > INPUT_BOUND, so why do we check this here?
      
      if (xmax > V::INTERSECTION_LIMIT || ymax > V::INTERSECTION_LIMIT) {
        Geometry::exactFindIntersect(v1->getX(), v1->getY(), v2->getX(), v2->getY(),
                                     (*it).second->getLeftEnd().second->getX(),
                                     (*it).second->getLeftEnd().second->getY(),
                                     (*it).second->getRightEnd().second->getX(),
                                     (*it).second->getRightEnd().second->getY(),
                                     vx,vy);
      } else {
        Geometry::findIntersect(v1->getX(), v1->getY(), v2->getX(), v2->getY(),
                                (*it).second->getLeftEnd().second->getX(),
                                (*it).second->getLeftEnd().second->getY(),
                                (*it).second->getRightEnd().second->getX(),
                                (*it).second->getRightEnd().second->getY(),
                                vx,vy);
      }

      /* Geometry::exact_find_intersect(v1->getX(), v1->getY(), v2->getX(), v2->getY(),
         (*it).second->getOrgVertex()->getX(),
         (*it).second->getOrgVertex()->getY(),
         (*it).second->getDestVertex()->getX(),
         (*it).second->getDestVertex()->getY(),
         vx,vy);*/
          
  
      V* theIntersection = insertPointOnEdge((*it).second,vx,vy);
      pm[(*it).first] = theIntersection;
    }
   
    //boost::timer t;
    V* vStart = v1;
    pm.push_back(v2);
    //insert constrained intervals
    FORALL (pm, it) {
      insertLineConstraintSimple(vStart, *it, ids, v1, v2);
      vStart = *it; // fixme: what is vStart used for?
    }
    // if(t.elapsed()>0) std::cout << t.elapsed() << std::endl; 
  }//end else
}


  //insert with multiple ids. restore after vertex removal
 
void DCDT::insertLineConstraint(V* v1, V* v2,std::list<int>& ids,
                                V* left, V* right)
{   //if the insertion is connected, just mark the unconstrained edge as constrained
  H* connectedEdge;
  if(_mesh.isConnected(v1, v2, &connectedEdge)) {
    _mesh.addConstraint(connectedEdge,ids,left,right);
    return;
  }

  //checks if the insertion constains an existing edge
  //if so mark the existing edge constrained
  //and insert the remaining part using insertLineConstraint 
  H* coincidentEdge;

  if (hasCoincidentSegments(v1, v2, &coincidentEdge)) {
         
    C x1, y1, x2, y2, x3, y3, x4, y4;
    x1 = v1->getX();
    y1 = v1->getY();
    x2 = v2->getX();
    y2 = v2->getY();

    x3 = coincidentEdge->getOrgVertex()->getX();
    y3 = coincidentEdge->getOrgVertex()->getY();
          
    x4 = coincidentEdge->getDestVertex()->getX();
    y4 = coincidentEdge->getDestVertex()->getY();
    if (Geometry::onSegment(x1,y1,x2,y2,x4,y4)) {
      _mesh.addConstraint(coincidentEdge,ids,left,right);
      insertLineConstraint(coincidentEdge->getDestVertex(),v2,ids,left,right);
      return;
    }

    if (Geometry::onSegment(x3,y3,x4,y4,x2,y2)) {
      V* theIntersect = insertPointOnEdge(coincidentEdge,x2,y2);

      H* e1 = theIntersect->getOutEdge();
      H* e2 = e1;
               
      do {                  
        V* v4 = e1->getDestVertex();

        if (v4 == coincidentEdge->getOrgVertex()) {
          _mesh.addConstraint(e1,ids,left,right);
          break;
        }
                     
        e1 = e1->getOppEdge()->getNext();
      } while (e1 != e2);
    }
  } else {

    //the common case: insertion starts in a triangle and crosses at least one edge
         
    std::list<H*> ul;
    std::list<std::pair<int,H*> > cl;

    std::vector<V*> pm;
    //upper vertices
    std::list<V*> pu;
    //lower vertices
    std::list<V*> pl;
        
        
    findConstrainedCrossEdges(v1, v2,cl,pm);

    int xmax = _mesh.getXMax();
    int ymax = _mesh.getYMax();
    C vx, vy;
    
    FORALL (cl, it) {
      
      //H* e = (*it).second;

      // fixme: this test is always false
      if(xmax > V::INTERSECTION_LIMIT || ymax > V::INTERSECTION_LIMIT) {
        Geometry::exactFindIntersect(v1->getX(), v1->getY(), v2->getX(), v2->getY(),
                                     (*it).second->getLeftEnd().second->getX(),
                                     (*it).second->getLeftEnd().second->getY(),
                                     (*it).second->getRightEnd().second->getX(),
                                     (*it).second->getRightEnd().second->getY(),
                                     vx, vy);
      } else {
        Geometry::findIntersect(v1->getX(), v1->getY(), v2->getX(), v2->getY(),
                                (*it).second->getLeftEnd().second->getX(),
                                (*it).second->getLeftEnd().second->getY(),
                                (*it).second->getRightEnd().second->getX(),
                                (*it).second->getRightEnd().second->getY(),
                                vx,vy);
      }

      /* Geometry::exact_find_intersect(v1->getX(), v1->getY(), v2->getX(), v2->getY(),
         (*it).second->getOrgVertex()->getX(),
         (*it).second->getOrgVertex()->getY(),
         (*it).second->getDestVertex()->getX(),
         (*it).second->getDestVertex()->getY(),
         vx,vy);*/
          
      V* theIntersection = insertPointOnEdge((*it).second, vx, vy, left, right);
      pm[(*it).first] = theIntersection;
    }
   
    //boost::timer t;
    V* vStart = v1;
    pm.push_back(v2);
    //insert constrained intervals

    FORALL (pm, it) {
      insertLineConstraintSimple(vStart, *it, ids, left, right);
      vStart = *it; // fixme: what is vStart used for?
    }
    // if(t.elapsed()>0) std::cout << t.elapsed() << std::endl; 
  }//end else
}


 
DCDT::V* DCDT::insertPointInFace(F* face,C& x, C& y)
{
  //assert(!face->isBoundary());
  //int oldFaceCount = _mesh.getFaceCount();
  //int oldEdgeCount = _mesh.getHalfEdgeCount();
  //int oldVetexCount = _mesh.getVertexCount();


     
  H* e1 = face->getEdge(0);
  H* e2 = e1->getNext();
  H* e3 = e2->getNext();

  V* v1 = e1->getOrgVertex();
  V* v2 = e2->getOrgVertex();
  V* v3 = e3->getOrgVertex();
     
  V* v5 = _mesh.createVertex(x,y);

  //face 1
  H* e7 = _mesh.createEdge(v5, v1);
  H* e9 = _mesh.createEdge(v2, v5);
  _mesh.createFace(e7, e9, e1);

  //face 2
  H* e11 = e9->getOppEdge();
  H* e13 = _mesh.createEdge(v3, v5);

  _mesh.createFace(e11, e2,e13);

  //face 3
  _mesh.createFace(e13->getOppEdge(), e3, e7->getOppEdge());

  //remove face 
  _mesh.removeFace(face);

  //restore delaunay property

  legalizedEdge(e1);
  legalizedEdge(e2);
  legalizedEdge(e3);


  //assert(_mesh.getFaceCount() == oldFaceCount +2);
  //assert(_mesh.getHalfEdgeCount() == oldEdgeCount +6);
  //assert(_mesh.getVertexCount() == oldVetexCount +1);

  return v5;
}

 
void DCDT::splitAndLinkEdge(H* e1, V* v5)
{     
  H* e2 = e1->getOppEdge();
  H* e3 = e1->getNext();
  H* e4 = e3->getNext();

  H* e5 = e2->getNext();
  H* e6 = e5->getNext();

  V* v1 = e1->getOrgVertex();
  // FIXME: not used? V* v2 = e3->getOrgVertex();
  V* v3 = e4->getOrgVertex();
  V* v4 = e6->getOrgVertex();

  Face* f1 = e1->getFace();
  Face* f2 = e2->getFace();

  //process face 1 
  H* e7 = _mesh.createEdge(v3,v5);
  H* e8 = e7->getOppEdge();

  e1->setOrgVertex(v5);
    
  H* e9 = _mesh.createEdge(v1,v5);
  H* e10 = e9->getOppEdge();
  if (e1->isConstrained()) {
    e9->setConstrained(e1->getConstraintIDs());
    e10->setConstrained(e1->getConstraintIDs());
  }

  std::vector<H*> edges1;
  edges1.push_back(e1);
  edges1.push_back(e3);
  edges1.push_back(e7);    
  e1->setFace(f1);
  e3->setFace(f1);
  e7->setFace(f1);
  f1->setEdges(edges1);

  std::vector<H*> edges2;
  edges2.push_back(e9);
  edges2.push_back(e8);
  edges2.push_back(e4);
  _mesh.createFace(edges2);
   
  //process face 2
  H* e11 = _mesh.createEdge(v4,v5);
  H* e12 = e11->getOppEdge();
    
  e2->setDestVertex(v5);
  std::vector<H*> edges3;
  edges3.push_back(e2);
  edges3.push_back(e12);
  edges3.push_back(e6);
  e2->setFace(f2);
  e12->setFace(f2);
  e6->setFace(f2);
  f2->setEdges(edges3);

  std::vector<H*> edges4;
  edges4.push_back(e10);
  edges4.push_back(e5);
  edges4.push_back(e11);
  _mesh.createFace(edges4);
}



 
void DCDT::splitAndLinkEdge(H* e1, V* v5,V* left,V* right)
{
     
    
    
  H* e2 = e1->getOppEdge();
  H* e3 = e1->getNext();
  H* e4 = e3->getNext();

  H* e5 = e2->getNext();
  H* e6 = e5->getNext();

  V* v1 = e1->getOrgVertex();
  // FIXME: not needed? V* v2 = e3->getOrgVertex();
  V* v3 = e4->getOrgVertex();
  V* v4 = e6->getOrgVertex();

  F* f1 = e1->getFace();
  F* f2 = e2->getFace();

  //process face 1 
  H* e7 = _mesh.createEdge(v3,v5);
  H* e8 = e7->getOppEdge();

  e1->setOrgVertex(v5);
    
  H* e9 = _mesh.createEdge(v1,v5);
  H* e10 = e9->getOppEdge();
  if (e1->isConstrained()) {
    e9->setConstrained(e1->getConstraintIDs());
    e10->setConstrained(e1->getConstraintIDs());
    
    if (!e1->getLeftEnd().first)  {
      std::pair<bool,V*> lf(true,left);
      e1->setLeftEnd(lf);
                
    }
    
    if (!e1->getRightEnd().first) {
      std::pair<bool,V*> rt(true,right);
      e1->setRightEnd(rt);
    }

    if (!e2->getLeftEnd().first) {
      std::pair<bool,V*> rt(true,right);
      e2->setLeftEnd(rt);
                
    }
    if (!e2->getRightEnd().first) {
      std::pair<bool,V*> lf(true,left);
      e2->setRightEnd(lf);
    }
            
    if (!e9->getLeftEnd().first)  {
      std::pair<bool,V*> lf(true,left);
      e9->setLeftEnd(lf);
    }
    
    if (!e9->getRightEnd().first) {
      std::pair<bool,V*> rt(true,right);
      e9->setRightEnd(rt);
    }

    if (!e10->getLeftEnd().first) {
      std::pair<bool,V*> rt(true,right);
      e10->setLeftEnd(rt);
    }

    if (!e10->getRightEnd().first) {
      std::pair<bool,V*> lf(true,left);
      e10->setRightEnd(lf);
    }
  }

  std::vector<H*> edges1;
  edges1.push_back(e1);
  edges1.push_back(e3);
  edges1.push_back(e7);    
  e1->setFace(f1);
  e3->setFace(f1);
  e7->setFace(f1);
  f1->setEdges(edges1);

  std::vector<H*> edges2;
  edges2.push_back(e9);
  edges2.push_back(e8);
  edges2.push_back(e4);
  _mesh.createFace(edges2);
   
  //process face 2
  H* e11 = _mesh.createEdge(v4,v5);
  H* e12 = e11->getOppEdge();
    
  e2->setDestVertex(v5);
  std::vector<H*> edges3;
  edges3.push_back(e2);
  edges3.push_back(e12);
  edges3.push_back(e6);
  e2->setFace(f2);
  e12->setFace(f2);
  e6->setFace(f2);
  f2->setEdges(edges3);

  std::vector<H*> edges4;
  edges4.push_back(e10);
  edges4.push_back(e5);
  edges4.push_back(e11);
  _mesh.createFace(edges4);
}


 
DCDT::V* DCDT::insertPointOnEdge(H* e, C& x, C& y)
{
  //post condition
  //int oldFaceCount = _mesh.getFaceCount();
  //int oldEdgeCount = _mesh.getHalfEdgeCount();
  //int oldVetexCount = _mesh.getVertexCount();

  H* e1 = e;
  H* e2 = e1->getNext();
  H* e3 = e2->getNext();

  V* v1 = e1->getOrgVertex();
  V* v2 = e2->getOrgVertex();
     
  if (x == v1->getX() && y == v1->getY()) {
    return v1;
  }
  
  if (x == v2->getX() && y == v2->getY()) {
    return v2;
  }
  
  V* v5 = _mesh.createVertex(x,y);
    
  //C x1, y1, x2, y2, x3, y3;
  

  //assert(Geometry::in_segment(v1->getX(),v1->getY(),v2->getX(),v2->getY(),x,y) ) ;
 
  // boost::timer t;
  
  splitAndLinkEdge(e1 ,v5);
 
  H* e5 = e3->getNext()->getOppEdge()->getNext();
  H* e6 = e2->getPrior()->getOppEdge()->getPrior();
    
  legalizedEdge(e2);
  legalizedEdge(e3);
  legalizedEdge(e5);
  legalizedEdge(e6);

  //post condition
  // FIXME: why is this uncommented?
  //assert(_mesh.getFaceCount() == oldFaceCount +2);
  //assert(_mesh.getHalfEdgeCount() == oldEdgeCount +6);
  //assert(_mesh.getVertexCount() == oldVetexCount +1);

  return v5;
}

 
DCDT::V* DCDT::insertPointOnEdge(H* e, C& x, C& y,V* left, V* right)
{
  //post condition
  //int oldFaceCount = _mesh.getFaceCount();
  //int oldEdgeCount = _mesh.getHalfEdgeCount();
  //int oldVetexCount = _mesh.getVertexCount();
  

  H* e1 = e;
  H* e2 = e1->getNext();
  H* e3 = e2->getNext();

  V* v1 = e1->getOrgVertex();
  V* v2 = e2->getOrgVertex();
     
  if (x == v1->getX() && y == v1->getY()) {
    return v1;
  }

  if (x == v2->getX() && y == v2->getY()) {
    return v2;
  }

    
  V* v5 = _mesh.createVertex(x,y);
    
  //C x1, y1, x2, y2, x3, y3;
     
  //assert(Geometry::in_segment(v1->getX(),v1->getY(),v2->getX(),v2->getY(),x,y) ) ;
 
  // boost::timer t;
  
  splitAndLinkEdge(e1 ,v5,left,right);
 
  H* e5 = e3->getNext()->getOppEdge()->getNext();
  H* e6 = e2->getPrior()->getOppEdge()->getPrior();
   
  legalizedEdge(e2);
  legalizedEdge(e3);
  legalizedEdge(e5);
  legalizedEdge(e6);

    
  //if(t.elapsed()>0) std::cout << t.elapsed() << std::endl;    

  // FIXME: why is this uncommented?
  
  //post condition
  //assert(_mesh.getFaceCount() == oldFaceCount +2);
  //assert(_mesh.getHalfEdgeCount() == oldEdgeCount +6);
  //assert(_mesh.getVertexCount() == oldVetexCount +1);

  return v5;
}

  /*
     
    void DCDT::legalizedEdge(H* e)
    {
    

    std::list<H*> stack;
    stack.push_back(e);

    while(stack.size()>0)
    {
    H* e = *(stack.begin());
    stack.pop_front();
    if(!isLegalEdge(e) && !e->isConstrained() && !e->isBoundary())
    {
            
    stack.push_back(e->getOppEdge()->getNext());
    stack.push_back(e->getOppEdge()->getPrior());

                 
    flip(e);

    }
    }
     

    }*/
 
void DCDT::legalizedEdge(H* e)
{
  if (e->isConstrained() ||e->isBoundary()) {
    return;
  } else {    
         
    if (!isLegalEdge(e)) {
                         
      H* ei = e->getOppEdge()->getNext();
      H* ek = ei->getNext();
      
      flip(e);

      legalizedEdge(ei);
      legalizedEdge(ek);
    }
  }
}


  //new implementation of flip that reuses existing faces.
 
DCDT::H* DCDT::flip(H* e)
{
  //the ccw edges
  H* eOpp = e->getOppEdge();
  H* e2 = e->getNext();
  H* e3 = e2->getNext();
  H* e4 = eOpp->getNext();
  H* e5 = e4->getNext();

  //create the new edge
  V* v3 = e2->getDestVertex();
  V* v4 = e4->getDestVertex();
  H* e6 = _mesh.createEdge(v3,v4);
  H* e7 = e6->getOppEdge();
    

  //old faces
  F* f1 = e->getFace();
  F* f2 = eOpp->getFace();

  std::vector<H*> f1Edges;
  f1Edges.push_back(e6);
  f1Edges.push_back(e5);
  f1Edges.push_back(e2);

  f1->setEdges(f1Edges);

  e6->setFace(f1);
  e5->setFace(f1);
  e2->setFace(f1);

  std::vector<H*> f2Edges;
  f2Edges.push_back(e7);
  f2Edges.push_back(e3);
  f2Edges.push_back(e4);
  f2->setEdges(f2Edges);
    
  e7->setFace(f2);
  e3->setFace(f2);
  e4->setFace(f2);

  _mesh.removeEdge(e);

  return e6;
}

 
void DCDT::triangulateFace(std::list<H*>& elist)
{
  H* xn;
  H* xp;
  H* tmp; 
  //FIXME not used? V* v;
  //V* vp;
  //V* vn;
  
  std::list<H*> stack ;
  bool ok;

  H* x = *(elist.begin());
  H* s = x;

  _mesh.beginMarking();
  do {
    _mesh.mark(x); 
    x=x->getNext(); 
  } while (x != s);
    
  while (!_mesh.isTriangle(x)) { 
    ok = true;
    xn=x->getNext(); 
    xp=x->getPrior();
    tmp=xn->getNext();

    if (Geometry::ccw(xp->getOrgVertex()->getX(), xp->getOrgVertex()->getY(),
                      x->getOrgVertex()->getX(),  x->getOrgVertex()->getY(),
                      xn->getOrgVertex()->getX(), xn->getOrgVertex()->getY()) > 0) {
      do { 
        if ( _mesh.inTriangle(xp->getOrgVertex(), x->getOrgVertex(), xn->getOrgVertex(), tmp->getOrgVertex()) ) {
          // boundary is considered inside
          ok=false; 
          break; 
        }
        tmp = tmp->getNext();
      } while (tmp != xp);
    } else {
      ok=false;
    }

    if (ok) {
      s = _mesh.splitFace(xp->getFace(), xp->getOrgVertex(), xn->getOrgVertex() );
      if(s->getFace()->getEdgeCount()==3) {
        s= s->getOppEdge();
      }
      stack.push_back(s);
      s=xn;
    }
    x=xn;
    if (!ok && x == s) {
      printf("Loop occurs in triangulate face");
      _mesh.endMarking();
      return ;
    }
  }

  //FIXME: not used H* e1;
  //H* e2;
  H* e3;
  H* e4;

  while (stack.size()> 0) {
    x = stack.front();
    stack.pop_front();
    if (!x->isBoundary() && !isLegalEdge(x)) {
            
      e3=x->getOppEdge()->getNext();
      e4=e3->getNext();
            
      if (!_mesh.marked(e3)) {
        _mesh.mark(e3);
        stack.push_back(e3);
      }
      if (!_mesh.marked(e4))  {
        _mesh.mark(e4);
        stack.push_back(e4);
      }
      flip(x);
    }
  }

  _mesh.endMarking();
}

/*****************************************************************
Note: vertices in vlist should be arranged such that they can be
traversed from e->getOrgVertex to e->getDestVertex
******************************************************************/
 
void DCDT::triangulatePseudoPolygon(std::list<V*>& vlist, H* e)
{
  if (vlist.size() > 1) {
    V* a = e->getOrgVertex();
    V* b = e->getDestVertex();
    V* c = *(vlist.begin());
    std::list<V*>::iterator cit = vlist.begin();

    FORALL (vlist, it) {

      /*C x1, y1, x2, y2, x3, y3, x4, y4;
   
        x1 = a->getX();
        y1 = a->getY();
        x2 = b->getX();
        y2 = b->getY();
        x3 = c->getX();
        y3 = c->getY();
        x4 = (*v)->getX();
        y4 = (*v)->getY();*/
               
      if (Geometry::inCircle(a->getX(), a->getY(),
                             b->getX(), b->getY(),
                             c->getX(), c->getY(),
                             (*it)->getX(), (*it)->getY())) {
        c = *it;
        cit = it;
      }
    }
         
    std::list<V*> pe;
    pe.insert(pe.begin(), vlist.begin(), cit);
          
    std::list<V*> pd;
    pd.insert(pd.begin(), ++cit, vlist.end());

    H* peEdge = 0;
    H* pdEdge = 0;

    if (e->getNext()->getDestVertex() == c) {

      peEdge = _mesh.splitFace(e->getFace(), a, c);

      //switch to the outer SimplePolygon face edge
      if(peEdge->getFace() == e->getFace()) {
        peEdge = peEdge->getOppEdge();
      }

    } else if (e->getPrior()->getOrgVertex() == c) {

      pdEdge = _mesh.splitFace(e->getFace(), b, c);

      //switch to the outer SimplePolygon face edge
      if (pdEdge->getFace() == e->getFace()) {
        pdEdge = pdEdge->getOppEdge();
      } 

    } else {

      peEdge = _mesh.splitFace(e->getFace(), a, c);

      //switch to the outer SimplePolygon face edge
      if (peEdge->getFace() == e->getFace()) {
        peEdge = peEdge->getOppEdge();
      }

      pdEdge = _mesh.splitFace(e->getFace(), b, c);

      //switch to the outer SimplePolygon face edge
      if (pdEdge->getFace() == e->getFace()) {
        pdEdge = pdEdge->getOppEdge();
      }
    }

    triangulatePseudoPolygon(pe,peEdge);
    triangulatePseudoPolygon(pd,pdEdge);
  }
}

  //------------------------------------------------------------------------------------


#if 1

// need to inovoke it once to create sector array
// and compute sectorWidth and sectorHeight 
void DCDT::initializeSectors()
{
  C xmin, xmax, ymin, ymax; 
  
  float _xmin = 0;
  float _ymin = 0;
  float _xmax = _mesh.getXMax();
  float _ymax = _mesh.getYMax();

  //calculate the width and height of the triangulation
  float width = _xmax - _xmin;
  float height = _ymax - _ymin;
  //calculate the width and height of each sector
  _sectorWidth = width / (float)_xSectors;
  _sectorHeight = height / (float)_ySectors;
  
  #if 0
  float x, y;
  float x1, y1, x2, y2, x3, y3;
  
	  std::vector<F*>::iterator f;

	  //go through the unprocessed triangles
	  for (f = _mesh.getFaceBegin(); f != _mesh.getFaceEnd(); ++f) {
	    if( (*f)->isBoundary()) {
		 continue;
	    }

	    //retrieve the next valid triangle
	    F* currentFace = *f;
			
	    //get the minimum and maximum x and y values
	    TriangleVertices(currentFace->getFaceEdge(0), x1, y1, x2, y2, x3, y3);
	    float left = Min(x1, x2, x3);
	    float right = Max(x1, x2, x3);
	    float top = Min(y1, y2, y3);
	    float bottom = Max(y1, y2, y3);
	    //calculate the minimum and maximum x and y sector indices covered
	    int xIndexMin = (int)((left - _xmin + 0.5f * sectorWidth) / sectorWidth);
	    int xIndexMax = (int)((right - _xmin - 0.5f * sectorWidth) / sectorWidth);
	    int yIndexMin = (int)((top - _ymin + 0.5f * sectorHeight) / sectorHeight);
	    int yIndexMax = (int)((bottom - _ymin - 0.5f * sectorHeight) / sectorHeight);
	    //go through the covered sector midpoints
	    for (int i = yIndexMin; i <= yIndexMax; i++) {
		 for (int j = xIndexMin; j <= xIndexMax; j++) {
		   //calculate the location of the midpoint
		   x = ((float)j + 0.5f) * sectorWidth + _xmin;
		   y = ((float)i + 0.5f) * sectorHeight + _ymin;
		   //if the midpoint is in the current triangle,
		   if (InTriangle(currentFace, x, y)) {
		     //point the sector to that triangle
		     sectors[i][j] = currentFace->vecIndex;
		     std::pair<int,int> p(i,j);
		     currentFace->attachedSectors.push_back(p);
		   }
		 }
	    }
	  }
  #endif
}

#endif

//returns the maximum of the 3 values passed

 
DCDT::LocateResult
DCDT::locatePointRandom(C& x, C& y,
                           F** resultFace, 
                           H** resultEdge,
                           V** resultVertex)
{
  int res = NotFound;
  int n = _mesh.getVertexCount();

  int sampleSize =(int)ceil(pow(n, 1.0/3.0));
  double d = 10000;
  int resIndex = 1;

  for(int i= 0; i< sampleSize;i++) {
    int index = rand() % n;
    V* tmp = _mesh.getVertices()[index];
    double x2 = tmp->getX().getD();
    double y2 = tmp->getY().getD();

    double dx, dy;
    dx=x.getD()-x2;
    dy=y.getD()-y2;
    double distbetween = pow(dx*dx + dy*dy,1.0/2.0);
    if (distbetween < d) {
      d =distbetween;
      resIndex = tmp->getVecIndex();
    }
  }
  int visited =0;
  res = locatePoint(_mesh.getVertices()[resIndex]->getOutEdge()->getFace(),
                    NULL,
                    x, y,
                    resultFace, resultEdge, resultVertex,
                    visited);
  return (LocateResult)res;
}

#if 0

// no longer needed

 
DCDT::LocateResult
DCDT::locatePointThreshold(F* discard,
                              C x,
                              C y,
                              F** resultFace, 
                              H** resultEdge,
                              V** resultVertex)
{
  int res = NotFound;
  C xmin, xmax, ymin, ymax; 
  _mesh.getBounds( xmax, ymax, xmin, ymin );

  float _xmin = xmin.getD();
  float _ymin = ymin.getD();
  //float _xmax = xmax.getD();
  //float _ymax = ymax.getD();

  float _x = x.getD();
  float _y = y.getD();

  //calculates the indices of the sector in which the point lies
  int xIndex = (int)((_x - _xmin) / _sectorWidth);
  int yIndex = (int)((_y - _ymin) / _sectorHeight);
  //makes sure the indices are within the proper bounds
  xIndex = (xIndex < 0) ? 0 : (xIndex >= xSectors) ? xSectors - 1 : xIndex;
  yIndex = (yIndex < 0) ? 0 : (yIndex >= ySectors) ? ySectors - 1 : yIndex;
  //retrieve the triangle in which that sector midpoint is located

  int faceIndex = sectors[yIndex][xIndex];
  F* iniface;


  // check if the faceIndex is not a valid interior face index
  if(faceIndex > _mesh.getFaceCount() || faceIndex < 1)
    {
      iniface = *(_mesh.getFaceEnd()-1);

    }
  else 
    {
      iniface = _mesh.getFace(sectors[yIndex][xIndex]);
  
    }
    
  int visited = 0;
    
  res = locatePoint(iniface, NULL, x, y, resultFace, resultEdge,resultVertex,visited);
   
	
  if (visited > SECTORREINITIALIZELIMIT) {
    InitializeSectors();
  }
  //use this starting triangle to start point location

    
  return (LocateResult)res;

}

#endif

 
DCDT::LocateResult
DCDT::locatePoint(F* /*iniFace*/,
                     C& x,
                     C& y,
                     F** resultFace, 
                     H** resultEdge,
                     V** resultVertex)
{
    
  int res = NotFound;
  float _x = x.getD();
  float _y = y.getD();

  //calculates the indices of the sector in which the point lies
  int xIndex = (int)((_x - 0) / _sectorWidth); // min = 0
  int yIndex = (int)((_y - 0) / _sectorHeight);
  //makes sure the indices are within the proper bounds
  xIndex = (xIndex < 0) ? 0 : (xIndex >= _xSectors) ? _xSectors - 1 : xIndex;
  yIndex = (yIndex < 0) ? 0 : (yIndex >= _ySectors) ? _ySectors - 1 : yIndex;
  //retrieve the triangle in which that sector midpoint is located

  int faceIndex = sectors[yIndex][xIndex];
  F* iniface;
  int visited = 0;
  
  // check if the faceIndex is not a valid interior face index
  if (faceIndex > _mesh.getFaceCount()-1 || faceIndex < 1) {
    assert(!_mesh.getFaces().empty());
    iniface = _mesh.getFaces().back();
        
    res = locatePoint(iniface, NULL, x, y, resultFace, resultEdge,resultVertex,visited);

    if (res == TriangleFound) {
      sectors[yIndex][xIndex] = (*resultFace)->getVecIndex();
      //std::pair<int,int> p(yIndex,xIndex);
      //(*resultFace)->attachedSectors.push_back(p);
    }

  } else {
    
    iniface = _mesh.getFaces()[faceIndex];
    res = locatePoint(iniface, NULL, x, y, resultFace, resultEdge, resultVertex, visited);

    if (res == TriangleFound) {
      sectors[yIndex][xIndex] = (*resultFace)->getVecIndex();
      //std::pair<int,int> p(yIndex,xIndex);
      //(*resultFace)->attachedSectors.push_back(p);
    }
   
  }
    
  return (LocateResult)res;
}

 
DCDT::LocateResult DCDT::locatePoint(F* iniface,
                                     H* dummy,
                                     C& x, C& y,
                                     F** resultFace,
                                     H** resultEdge,
                                     V** resultVertex,
                                     int& visited)
{
  int res = NotFound;
  V* v1;
  V* v2;
  H* e1;

  //add randomness to avoid loops in non-Delaunay triangulations
  double randNum = double(rand()) / double(RAND_MAX);
  if (randNum > 0.5) {
    e1 = iniface->getEdge(0);
  } else {
    e1 = iniface->getEdge(1);
  }
    
  H* e2 = e1;

  C x1, y1, x2, y2;

  // loop clockwise over the edges
  do {
    if (e2 != dummy) {
      //skip one ccw test. The oppEdge of edge dummy is tested in the previous triangle.

      v1=e2->getOrgVertex();
      v2=e2->getDestVertex(); 

      x1 = v1->getX();
      y1 = v1->getY();
      x2 = v2->getX();
      y2 = v2->getY();

      if (Geometry::ccw(x2,y2,x1,y1,x,y) > 0) {
        //check for enclosure

        dummy = e2->getOppEdge();

        //search on adjacent triangle
        return locatePoint(dynamic_cast<F*>(dummy->getFace()), // fixme: not nice ...
                           dummy, x, y, resultFace, resultEdge, resultVertex, ++visited);
      }
    }

    e2 = e2->getPrior();

  } while (e2 != e1);
    

  //base case found: Triangle found with triangle vertices enclosing (x,y) 
  //TRAP! It could also be on the edge.
  res = TriangleFound; 
  *resultFace = iniface;

  //extra test for point on vertex/edge
  if(_mesh.onVertex(*resultFace , x, y, resultVertex)) {
    res = VertexFound;
  } else if(_mesh.onEdge(*resultFace , x, y, resultEdge)){
    res = EdgeFound;
  }
    
  //std::cout<<visited<<std::endl;
  return (LocateResult)res;
}

 
bool DCDT::isLegalEdge(H* e)
{
  if (e->isConstrained())
    return true;

  V* v1 = e->getOrgVertex();
  V* v2 = e->getDestVertex();

  V* v3 = e->getNext()->getDestVertex();
  V* v4 = e->getOppEdge()->getNext()->getDestVertex();

  return !Geometry::inCircle(v1->getX(), v1->getY(),
                             v2->getX(), v2->getY(),
                             v3->getX(), v3->getY(),
                             v4->getX(), v4->getY());
}
