#ifndef DCDT_H
#define DCDT_H

#include "Polygon.h"
#include "Point.h"
#include "Geometry.h"
#include "GMP_helper.h"

#include <list>
#include <vector>
#include <cmath>
#include <cstdio>
#include <iostream>
#include <queue>

#include <boost/timer.hpp>
#include <boost/array.hpp>
#include <boost/multi_array.hpp>

#include "Mesh.h"

class DCDT
{
  //static const int SECTORREINITIALIZELIMIT = 20;

  //static const int UNABSTRACTED = -1;
  static const long long INTERSECTOVERFLOWBOUND = 1000000;

public:

  typedef Vertex     V;
  typedef HalfEdge   H;
  typedef Face       F;
  typedef V::coord_t C;

  // constructor , destructor
  DCDT(Mesh& m, int xSectors, int ySectors);
  virtual ~DCDT();

  // enums for return values;
  enum LocateResult { NotFound, TriangleFound, EdgeFound, VertexFound };

public:

  // FIXME: briefly describe what these functions do
  
  //operators

  void init(const IPolygon& pol, int radius = 1);
  // tries to insert vertex (x,y). return existing vertex if insertion overlaps   
  V* findOrInsertVertex(C& x, C& y);
  // deprecated
  //void findCrossVertices(V* v1, V* v2,std::list<V*>& vl);

  //perform straight-line walk on triangls cut by line segment v1-v2
  //store intersecting constrained line segments in cl
  //store newly inserted intersection points in vl. (partitions the constraint
  //insertion to small constrained intervals between intersections.
  // vl is to be used later to insert each intervals) 
  void findConstrainedCrossEdges(V* v1, V* v2,
                                 std::list<std::pair<int,H*> >& cl,
                                 std::vector<V*>& vl);
  
  //perform straight-line walk on triangls cut by line segment v1-v2
  //store intersecting unconstrained line segments in ul
  void findUnconstrainedCrossEdges(V* v1, V* v2,
                                   std::list<V*>& pu,
                                   std::list<V*>& pl,
                                   std::list<H*>& ul,
                                   V* left, V* right);

  //insert point v1=(x1,y1) and v2 = (x2,y2) then invoke insertLineConstraint()
  void insertConstraint( C& x1, C& y1, C& x2, C& y2, int id );
  
  //insert constrained line segments with endpoints v1 and v2
  //it invokes insertLineConstraintSimple in the common case
  void insertLineConstraint(V* v1, V* v2,int id);
  void insertLineConstraintSimple(V* v1, V* v2,int id,
                                  V* left, V* right);

  //insert constrained line segment assuming no intersecting constraints 
  //leftmost and rightmost endpoint is used in intersection computation.
  void insertLineConstraintSimple(V* v1, V* v2,std::list<int>& ids,
                                  V* left, V* right);
                                  
  //constraint insertion.
  //it is used to re-insert the constraint after vertex(steiner) removal and restoring correct constraint ids
  void insertLineConstraint(V* v1, V* v2,std::list<int>& ids);
  
  //constraint insertion.
  //it is used to re-insert the constraint after vertex(steiner) removal and restoring correct constraint ids
  //left and right integer coordinate is maintained in this constraint
  void insertLineConstraint(V* v1, V* v2,std::list<int>& ids,
                            V* left, V* right);
  
  //constrained polygon insertion.   
  void insertPolygon(const IPolygon& pol, int id);
  
  //returns the constrained polygon with ID == id.
  const Polygon<V*>* getPolygon(int id);

  //constrained Delaunay point insertion within a face 
  V* insertPointInFace(F* f, C& x,C& y);
  
  //constrained Delaunay point insertion on an edge
  V* insertPointOnEdge(H* e, C& x,C& y);
  
  //constrained Delaunay point insertion on an edge.
  //left and right endpoint of the splitting constraint is reset.
  V* insertPointOnEdge(H* e, C& x,C& y,V* left,V* right);
  
  //restore constrained Delaunay property using edge flip procedure
  void legalizedEdge(H* e);
  
  //edge flip operation on quadrilateral with diagonal edge e
  //edge is flipped to the other diagonal
  H* flip(H* e);
  
  //traingulatio of simple polygon using ear clipping.
  //connect vertices by observing empty circumcircle property.  
  //invoke by vertex removal
  void triangulateFace(std::list<H*>& elist);
  
  //retriangulate the open region form by constraint insertion.
  void triangulatePseudoPolygon(std::list<V*>& vlist, H* e);
  
  //vertex removal by deleting it from vertex vector
  void removeVertex(V* v);
  
  //try to remove a vertex.
  //skip the removal if it is still an intersection vertex.
  //skip when the vertex is the endpoint.  
  void removeVertexIfPossible(V* v, int id);
  //void removeEdge(V* v);
 
   // collect all edges and vertices associated with a constrained polygon with ID == polygonid
  void findVerticesAndEdges(int polygonid, std::vector<V*>& vs, std::vector<H*>& es); 
  
  //remove the constrained polygon
  void removePolygon(int id);

private:

  // data
  
  // std::list<H*> tmpedge;
	
  Mesh& _mesh;
  int _xSectors, _ySectors;
  int _sectorWidth, _sectorHeight;//initialized through initializeSector()
  std::list<Polygon<V*> > _polygons;

public:

  //point location
  //sector-based speedup references Douglas Demyen's implementation

  //perform initialization for sector-based point location
  //setting sectorWidth and sectorHeight
  void initializeSectors();

  //use sector-based point location to find the face in which the given point resides
  LocateResult locatePoint(F* iniface,
                           C& x, C& y,
                           F** resultFace, 
                           H** resultEdge,
                           V** resultVertex);
  
  LocateResult locatePointThreshold(F* iniface,
                                    C& x, C& y,
                                    F** resultFace, 
                                    H** resultEdge,
                                    V** resultVertex);
  
//use random sampling point location to find the face in which the given point resides
  LocateResult locatePointRandom(C& x, C& y,
                                 F** resultFace, 
                                 H** resultEdge,
                                 V** resultVertex);
	
  //remembering stochastic walk point location
  LocateResult locatePoint(F* iniface,
                           H* dummy,
                           C& x, C& y,
                           F** resultFace, 
                           H** resultEdge,
                           V** resultVertex,
                           int& visited);

  Mesh& getMesh() { return _mesh; }
  
  //the array of sectors
  //std::pair<bool, F*> sectors[ySectors][xSectors];
  //std::vector<std::pair<int, int> > dirtySectors;
  //int **sectors;
  boost::multi_array<int, 2>::extent_gen extents;
  boost::multi_array<int, 2> sectors;
 
  
public:

  //triangulation:

  bool isLegalEdge(H* e);
  bool hasCoincidentSegments(V* v1, V* v2, H** coincidentEdge);
  F*   locateStartTriangle(V* v1, V* v2);
  int  interHoriz(C& px, C& py, C& p1x, C& p1y, C& p2x, C& p2y);

  // assumes polygon is closed
  int  insidePolygon(C& x, C& y);

private:

  bool inputOK(const IPoint& p) {
    return p.getX() >= 0 && p.getX() <= _mesh.getXMax()
      && p.getY() >= 0 && p.getY() <= _mesh.getYMax();
  }
  
  void splitAndLinkEdge(H* e1, V* v5, V* left, V* right);
  void splitAndLinkEdge(H* e1, V* v5);

};

#endif
