#include "GMP_helper.h"
#include <cassert>
#include <cstdlib>
#include <iostream>


using namespace std;

// copied from gmp source
#define MPZ_REALLOC(z,n) (UNLIKELY ((n) > ALLOC(z))     \
  ? (mp_ptr) _mpz_realloc(z,n)  \
 : PTR(z))

#define UNLIKELY(cond) __GMP_UNLIKELY(cond)
#define ALLOC(x) ((x)->_mp_alloc)
#define PTR(x) ((x)->_mp_d)

#if GMP_LIMB_BITS != 32 && GMP_LIMB_BITS != 64
#error "GMP_LIMB_BITS != 32,64 "
#endif

#if GMP_NAIL_BITS != 0
#error "gmp nail bits != 0"
#endif    



bool mpz_fits_uint8(mpz_srcptr x)
{
  mp_size_t n = x->_mp_size;
  if (n == 0) return true;

#if GMP_NUMB_BITS == 32
  if (n > 2) return false;
#else
  if (n > 1) return false;
#endif
  return true;
}

// assumes mpz_fits_uint8_p is true
uint8 mpz_get_uint8(mpz_srcptr x)
{
  mp_size_t n = x->_mp_size;
  if (n == 0) return 0;

#if GMP_NUMB_BITS == 32
  if (n > 2) return 0;
  return ((uint8)x->_mp_d[1]) << 32 | x->_mp_d[0];
#else
  if (n > 1) ERR("too big");
  return x->_mp_d[0];
#endif
}

bool mpz_fits_sint8_p(mpz_srcptr x)
{
  mp_size_t n = x->_mp_size;
  mp_limb_t limb = x->_mp_d[0];
  
  if (n == 0) return true;
  
#if GMP_NUMB_BITS == 32
  //cout << "xxx " << n << endl;  
  if (n > 2 || n < -2)  return false;
  if (n == 1|| n == -1) return true;
  uint8 ull = x->_mp_d[1];
  ull <<= 32;
  ull |= limb;
  if (n > 0)
    return ull <= 0x7fffffffffffffffULL;
  else 
    return ull <= 0x8000000000000000ULL;      
#else
  // cout << "yyy " << n << endl;  
  if (n > 1 || n < -1) return false;
  if (n > 0)
    return limb <= 0x7fffffffffffffffULL;
  else
    return limb <= 0x8000000000000000ULL;
#endif
}

// assumes mpz_fits_sint8_p is true  
sint8 mpz_get_sint8(mpz_srcptr x)
{
  mp_size_t n = x->_mp_size;
  mp_limb_t limb = x->_mp_d[0];

  if (n == 0) return 0;

#if GMP_NUMB_BITS == 32
  if (n > 2 || n < -2)  ERR("too big");

  if (n == 1) {
    return limb;
  }
  if (n == -1) {
    return -(sint8)limb;
  }

  // 2 or -2
  uint8 ull = x->_mp_d[1];
  ull <<= 32;
  ull |= limb;

  if (n > 0) {
    // +2
    if (ull > 0x7fffffffffffffffULL) ERR("too big");
    return ull;
  }

  // -2
  if (ull > 0x8000000000000000ULL) ERR("too big");
  return -ull;

#else

  if (n > 1 || n < -1) ERR("too big");

  if (n > 0) {
    if (limb > 0x7fffffffffffffffULL) ERR("too big");
    return limb;
  }

  if (limb > 0x8000000000000000ULL) ERR("too big");
  return -limb;

#endif
}

void mpz_set_uint8(mpz_ptr dest, const uint8& val)
{
  mp_size_t size;

  dest->_mp_d[0] = val & GMP_NUMB_MASK;
  size = val != 0;

#if 64 > GMP_NUMB_BITS  /* avoid warnings about shift amount */
  if (val > GMP_NUMB_MAX) {
    MPZ_REALLOC (dest, 2);
    dest->_mp_d[1] = val >> GMP_NUMB_BITS;
    size = 2;
  }
#endif

  dest->_mp_size = size;
}

void mpz_set_sint8(mpz_ptr dest, const sint8& val)
{
  mp_size_t size;
  uint8 vl;

  vl = (uint8) (val >= 0 ? val : -val);

  dest->_mp_d[0] = vl & GMP_NUMB_MASK;
  size = vl != 0;

#if 64 > GMP_NUMB_BITS
  if (vl > GMP_NUMB_MAX) {
    MPZ_REALLOC (dest, 2);
    dest->_mp_d[1] = vl >> GMP_NUMB_BITS;
    size = 2;
  }
#endif
    
  dest->_mp_size = val >= 0 ? size : -size;
}


// e1 * e2 without gcd
Exact mult(const Exact& e1, const Exact& e2)
{
  Exact res;
  mpq_srcptr p1 = e1.get_mpq_t();
  mpq_srcptr p2 = e2.get_mpq_t();
  // a/b * c/d
  // num = a * c
  mpz_mul(mpq_numref(res.get_mpq_t()), mpq_numref(p1), mpq_numref(p2));
  // den = b * d
  mpz_mul(mpq_denref(res.get_mpq_t()), mpq_denref(p1), mpq_denref(p2));
  return res;
}

// e1 + e2 without gcd
Exact add(const Exact& e1, const Exact& e2)
{
  Exact res;
  mpq_srcptr p1 = e1.get_mpq_t();
  mpq_srcptr p2 = e2.get_mpq_t();
  // a/b + c/d:
  // ad
  mpz_mul(mpq_numref(res.get_mpq_t()), mpq_numref(p1), mpq_denref(p2));  
  // bc
  mpz_mul(mpq_denref(res.get_mpq_t()), mpq_denref(p1), mpq_numref(p2));
  // num = ad + bc
  mpz_add(mpq_numref(res.get_mpq_t()), mpq_numref(res.get_mpq_t()), mpq_denref(res.get_mpq_t()));
  // den = bd
  mpz_mul(mpq_denref(res.get_mpq_t()), mpq_denref(p1), mpq_denref(p2));  
  return res;
}

// e1 - e2 without gcd
Exact sub(const Exact& e1, const Exact& e2)
{
  Exact res;
  mpq_srcptr p1 = e1.get_mpq_t();
  mpq_srcptr p2 = e2.get_mpq_t();
  // a/b + c/d:
  // ad
  mpz_mul(mpq_numref(res.get_mpq_t()), mpq_numref(p1), mpq_denref(p2));  
  // bc
  mpz_mul(mpq_denref(res.get_mpq_t()), mpq_denref(p1), mpq_numref(p2));
  // num = ad - bc
  mpz_sub(mpq_numref(res.get_mpq_t()), mpq_numref(res.get_mpq_t()), mpq_denref(res.get_mpq_t()));
  // den = bd
  mpz_mul(mpq_denref(res.get_mpq_t()), mpq_denref(p1), mpq_denref(p2));  
  return res;
}

/** convert LimRat to Exact.
    @return true iff OK
*/
bool convert(const LimRat& a, Exact& b)
{
  if (a.den == 0) return false;
  mpz_class& num = b.get_num();
  mpz_class& den = b.get_den();
  mpz_set_sint8(num.get_mpz_t(), a.num);
  mpz_set_sint8(den.get_mpz_t(), a.den);    
  b.canonicalize();
  return true;
}

/** convert Exact to LimRat.
    @return true iff OK
*/
bool convert(const Exact& a, LimRat& b)
{
  if (a == 0) {
    b.num = 0;
    b.den = 1;
    return true;
  }
  
  const mpz_class& num = a.get_num();
  const mpz_class& den = a.get_den();
  
  if (!mpz_fits_sint8_p(num.get_mpz_t()) ||
      !mpz_fits_sint8_p(den.get_mpz_t())) {
    return false;
  }
  
  b.num = mpz_get_sint8(num.get_mpz_t());
  b.den = mpz_get_sint8(den.get_mpz_t());
  return true;
}
  

static void test_convert(sint8 n, sint8 d)
{
  LimRat r, s, t;
  Exact e;
  r.num = n;
  r.den = d;
  
  if (!convert(r, e)) { cout << "error1 " << r.num << " " << r.den << endl; exit(-1); }
  if (!convert(e, s)) { cout << "error2 " << r.num << " " << r.den << endl; exit(-1); }
  if (!convert(s, e)) { cout << "error3 " << r.num << " " << r.den << endl; exit(-1); }
  if (!convert(e, t)) { cout << "error4 " << r.num << " " << r.den << endl; exit(-1); }
  if (s.num != t.num || s.den != t.den) {
    cerr << "oops " << s.num << " " << s.den << " " << t.num << " " << t.den << endl;
    exit(-1);
  }
}

void gmp_helper_test()
{
  cout << "limb: " << GMP_NUMB_BITS << " bits" << endl;

  if (sizeof(uint4) != 4) {
    cout << "uint4 != 4 bytes" << endl;
  }
  if (sizeof(uint8) != 8) {
    cout << "uint8 != 8 bytes" << endl;
  }

  // test highest bit

  for (int i=63; i >= 0; i--) {
    uint8 ull = 1ULL << i;
    if (highestBitSet(ull) != i) ERR("hb ull" << i);
  }

  for (int i=31; i >= 0; i--) {  
    uint4 ul = 1UL << i;
    if (highestBitSet(ul) != i) ERR("hb ul  " << i);    
    uint4 ui = 1U << i;
    if (highestBitSet(ui) != i) ERR("hb ui  " << i);    
  }

  if (highestBitSet(0ULL) >= 0) ERR("hb 0 ");

  // test conversions
  
  sint8 n = 0x7fffffffffffffffULL;
  sint8 d = 0x7fffffffffffffffULL;
  test_convert(n, d);
  test_convert(-n, d);
  test_convert(n, -d);
  test_convert(-n, -d);
  
  for (int i=0; i < 10000000; i++) {

    n = ((uint8)rand()) << 32 | (rand() << 1) | (rand() % 2);
    d = ((uint8)rand()) << 32 | (rand() << 1) | (rand() % 2);
    if (d == 0) continue;

    if (rand() & 256) n = -n;

    test_convert(n, d);
    if ((i & ((1<<16)-1)) == 0) cout << "." << flush;
  }
}
