#ifndef Geometry_H
#define Geometry_H


// exact in-circle test

// Uses Boost interval arithmetic to quickly decide whether the
// in-circle determinant is != 0. If that fails, falls back to exact
// rational computation (GMP)

// compile with g++ -O3 -Wall -Wuninitialized incircle.c gmp_helper.c -lgmpxx

// (c) 2008 Michael Buro, licensed under GPLv3
#include "GMP_helper.h"
#include <boost/lexical_cast.hpp>
#include <boost/numeric/interval.hpp>
#include <boost/array.hpp>
#include <boost/rational.hpp>
#include <iostream>
#include <cassert>
#include <boost/pool/object_pool.hpp>
//#include <boost/detail/singleton.hpp>
//#include <boost/detail/mutex.hpp>
//#include <boost/detail/guard.hpp>
#include <boost/timer.hpp>
#include "Rational.h"


#define TEST 0

class Geometry
{
  //    //copied from gmp source to solve the linking errors between C and C++ linkage
  //
  //#ifdef _WIN32 
  //static size_t geo_mpz_size(mpz_srcptr __gmp_z) __GMP_NOTHROW
  //{
  //     return __GMP_ABS (__gmp_z->_mp_size);
  //}
  //
  //static mp_limb_t geo_mpz_getlimbn (mpz_srcptr __gmp_z, mp_size_t __gmp_n) __GMP_NOTHROW
  //{
  //  mp_limb_t  __gmp_result = 0;
  //  if (__GMP_LIKELY (__gmp_n >= 0 && __gmp_n < __GMP_ABS (__gmp_z->_mp_size)))
  //    __gmp_result = __gmp_z->_mp_d[__gmp_n];
  //  return __gmp_result;
  //}
  //#endif


public:
  //static const int poolSize = 100; 
  //static Exact varPool[poolSize];
  //static int poolIndex;
  //static void setVarPool();
  //static Exact* getVar();

  static Exact p1xe;
  static Exact p1ye ;
  static Exact p2xe ;
  static Exact p2ye;
  static Exact p3xe ;
  static Exact p3ye ;
  static Exact pxe ;
  static Exact pye ;

public: 
  //typedef Exact coord_t;
  typedef Rational<> coord_t;
  typedef boost::rational<long long> coord_t2;

public:

  static const int COOR_MAX = 1321000; // -> intersection coordinate rationals fit in two sint8
  static int appCount;
  static int exactCount;
  
  struct LimRatPoint
  {
    LimRat x, y;
  };
  
  struct IntPoint
  {
    sint4 x, y;
  };

  struct ExactPoint
  {
    Exact x, y;
  };
  
  // ensure all coordinates are within range
  static bool withinRange(const Geometry::IntPoint& p1, const Geometry::IntPoint& p2,
                          const Geometry::IntPoint& p3, const Geometry::IntPoint& p4)
  {
    if (p1.x > Geometry::COOR_MAX || p1.x < -Geometry::COOR_MAX ||
        p1.y > Geometry::COOR_MAX || p1.y < -Geometry::COOR_MAX ||	
        p2.x > Geometry::COOR_MAX || p2.x < -Geometry::COOR_MAX ||
        p2.y > Geometry::COOR_MAX || p2.y < -Geometry::COOR_MAX ||	
        p3.x > Geometry::COOR_MAX || p3.x < -Geometry::COOR_MAX ||
        p3.y > Geometry::COOR_MAX || p3.y < -Geometry::COOR_MAX ||	
        p4.x > Geometry::COOR_MAX || p4.x < -Geometry::COOR_MAX ||
        p4.y > Geometry::COOR_MAX || p4.y < -Geometry::COOR_MAX) {
      ERR("intpoint out of range");
    }
    return true;
  }

  // returns 0 if all points lie on circle spanned by a b c, +1 if d is
  // inside, -1 if outside assuming ccw ordering of a b c
  static int exactInCircle( const Exact& p1x, const Exact& p1y, 
                            const Exact& p2x, const Exact& p2y,
                            const Exact& p3x, const Exact& p3y, 
                            const Exact& px, const Exact& py)
  {
    // doesn't compute gcd!
    const Exact adx = sub(p1x, px);
    const Exact bdx = sub(p2x, px);
    const Exact cdx = sub(p3x, px);
    const Exact ady = sub(p1y, py);
    const Exact bdy = sub(p2y, py);
    const Exact cdy = sub(p3y, py);
  
    Exact bdxcdy = mult(bdx, cdy);
    Exact cdxbdy = mult(cdx, bdy);
    Exact alift = add(mult(adx, adx), mult(ady, ady));

    Exact cdxady = mult(cdx, ady);
    Exact adxcdy = mult(adx, cdy);
    Exact blift = add(mult(bdx, bdx), mult(bdy, bdy));

    Exact adxbdy = mult(adx, bdy);
    Exact bdxady = mult(bdx, ady);
    Exact clift = add(mult(cdx, cdx), mult(cdy, cdy));

    Exact det =
      add(add(mult(alift, sub(bdxcdy, cdxbdy)),
              mult(blift, sub(cdxady, adxcdy))),
          mult(clift, sub(adxbdy, bdxady)));

    if (det > 0) return +1;
    if (det < 0) return -1;
    return 0;
  }

  // old version using GCD, 2+ times slower than without in this application!
  
  // returns 0 if all points lie on circle spanned by a b c, +1 if d is
  // inside, -1 if outside assuming ccw ordering of a b c
  static int exactInCircleOld(const Exact& p1x, const Exact& p1y, 
                              const Exact& p2x, const Exact& p2y,
                              const Exact& p3x, const Exact& p3y, 
                              const Exact& px, const Exact& py);
  
  // approximates a point using double intervals
  struct ApproxPoint
  {
    double xmin, xmax;
    double ymin, ymax;

    ApproxPoint(double x=0, double y=0)
    {
      xmin = xmax = x;
      ymin = ymax = y;    
    }

    ApproxPoint(double xmin, double xmax, double ymin, double ymax)
    {
      this->xmin = xmin;
      this->xmax = xmax;
      this->ymin = ymin;
      this->ymax = ymax;    
    }
  };

  /** returns +1 if d lies in circle spanned by a b c, -1 if it
      doesn't, 0 if we don't know; asssumes ccw ordering of a b c
  */
  static int approxInCircle(const ApproxPoint& pa, const ApproxPoint& pb,
                            const ApproxPoint& pc, const ApproxPoint& pd)
  {
    using namespace boost::numeric;
    using namespace interval_lib;
    typedef interval<double> I_aux;

    I_aux::traits_type::rounding rnd; // switches on default rounding mode
    typedef unprotect<I_aux>::type I;

    const I pax(pa.xmin, pa.xmax);
    const I pay(pa.ymin, pa.ymax);
    const I pbx(pb.xmin, pb.xmax);
    const I pby(pb.ymin, pb.ymax);
    const I pcx(pc.xmin, pc.xmax);
    const I pcy(pc.ymin, pc.ymax);
    const I pdx(pd.xmin, pd.xmax);
    const I pdy(pd.ymin, pd.ymax);
  
    const I adx = pax - pdx;
    const I bdx = pbx - pdx;
    const I cdx = pcx - pdx;
    const I ady = pay - pdy;
    const I bdy = pby - pdy;
    const I cdy = pcy - pdy;
  
    I bdxcdy = bdx * cdy;
    I cdxbdy = cdx * bdy;
    I alift = adx * adx + ady * ady;

    I cdxady = cdx * ady;
    I adxcdy = adx * cdy;
    I blift = bdx * bdx + bdy * bdy;

    I adxbdy = adx * bdy;
    I bdxady = bdx * ady;
    I clift = cdx * cdx + cdy * cdy;

    I det =
      alift * (bdxcdy - cdxbdy)
      + blift * (cdxady - adxcdy)
      + clift * (adxbdy - bdxady);

    using namespace compare::certain;

    // cout << "det= " << det.lower() << " " << det.upper() << endl;
  
    if (det > 0.) return +1;
    if (det < 0.) return -1;
    return 0;
  }



  ///** approximate rational number by double interval [xmin,xmax] using
  //    a small number of bits to represent the numerator and denominator.
  //*/
  //static void approximate(const Exact& x, int bits, double& xmin, double& xmax)
  //  {

  //   
  //    if (bits < 2 || bits > 30) { cout << "bits < 2 | > 30" << endl; exit(-1); }

  //    const mpz_srcptr n = x.get_num().get_mpz_t();
  //    const mpz_srcptr d = x.get_den().get_mpz_t();    
  //      
  //   
  //  #ifdef  _WIN32

  //    int sn = geo_mpz_size(n);
  //    int sd = geo_mpz_size(d);
  //  #else

  //    int sn = mpz_size(n);
  //    int sd = mpz_size(d);
  //  #endif
  //    // collect highest bits from both numbers

  //  #if GMP_LIMB_BITS == 64
  //    typedef uint8 Limb;
  //  #elif GMP_LIMB_BITS == 32
  //    typedef uint4 Limb;
  //  #endif

  //    // cout << x << " " << GMP_LIMB_BITS << endl;
  //    // cout << sn << " " << sd << endl;

  //  #ifdef  _WIN32
  //    // relevant bits in highest limb
  //    int hn = highestBitSet(geo_mpz_getlimbn(n, sn-1))+1;
  //    int hd = highestBitSet(geo_mpz_getlimbn(d, sd-1))+1;

  //  #else
  //    // relevant bits in highest limb
  //    int hn = highestBitSet(mpz_getlimbn(n, sn-1))+1;
  //    int hd = highestBitSet(mpz_getlimbn(d, sd-1))+1;
  //  #endif
  //   
  //    //the rational number is 0/den ;
  //    if (hn < 1 || hd < 1) 
  //    { 
  //         //cout << "0 limb" << endl; exit(-1);

  //         //0/den -> exact
  //         xmin = xmax = mpz_get_sint8(n);
  //         return;

  //    }

  //    // how many relevant bits in n and d
  //    int nbits = hn + (sn-1)*GMP_LIMB_BITS;
  //    int dbits = hd + (sd-1)*GMP_LIMB_BITS;

  //    if (dbits == 1 && nbits < 50) {
  //      // small integer -> exact
  //      xmin = xmax = mpz_get_sint8(n);

  //  #if TEST
  //      double v = x;
  //      if (v != xmin) {
  //        ERR("not exact! " <<  v << " " << xmin << " " << mpz_get_sint8(n) << " " << mpz_get_sint8(d));
  //      }
  //  #endif    
  //      return;
  //    }
  //    
  //    if (abs(nbits - dbits) >= 200) {
  //      ERR("huge difference " << nbits << " " << dbits);
  //    }

  //    // how many bits to extract from n and d
  //    int nebits = min(bits, nbits);
  //    int debits = min(bits, dbits);
  //    
  //    sint4 an, ad; // highest bits approximating n and d

  //    // assemble numerator bits
  //    if (nebits <= hn) {
  //         #ifdef  _WIN32
  //            // all bits in highest limb    
  //             an = geo_mpz_getlimbn(n, sn-1) >> (hn-nebits);

  //         #else
  //            // all bits in highest limb    
  //            an = mpz_getlimbn(n, sn-1) >> (hn-nebits);
  //         #endif
  //      
  //    } else {
  //      assert(sn >= 2);
  //      assert(nebits > hn);
  //      #ifdef  _WIN32
  //      // some bits in second limb    
  //      an =
  //        (geo_mpz_getlimbn(n, sn-1) << (nebits-hn)) |
  //        (geo_mpz_getlimbn(n, sn-2) >> (GMP_LIMB_BITS-(nebits-hn)));

  //      #else
  //      // some bits in second limb    
  //      an =
  //        (mpz_getlimbn(n, sn-1) << (nebits-hn)) |
  //        (mpz_getlimbn(n, sn-2) >> (GMP_LIMB_BITS-(nebits-hn)));
  //      #endif
  //      
  //    }

  //    // assemble denominator bits  
  //    if (debits <= hd) {
  //         #ifdef  _WIN32
  //             // all bits in highest limb    
  //            ad = geo_mpz_getlimbn(d, sd-1) >> (hd-debits);

  //         #else
  //            // all bits in highest limb    
  //            ad = mpz_getlimbn(d, sd-1) >> (hd-debits);
  //         #endif
  //     
  //    } else {
  //      assert(sd >= 2);
  //      assert(debits > hd);    
  //         #ifdef  _WIN32
  //           // some bits in second limb    
  //           ad =
  //             (geo_mpz_getlimbn(d, sd-1) << (debits-hd)) |
  //             (geo_mpz_getlimbn(d, sd-2) >> (GMP_LIMB_BITS-(debits-hd)));
  //         #else
  //           // some bits in second limb    
  //           ad =
  //             (mpz_getlimbn(d, sd-1) << (debits-hd)) |
  //             (mpz_getlimbn(d, sd-2) >> (GMP_LIMB_BITS-(debits-hd)));
  //         #endif
  //      
  //    }

  //    // pad an and ad with 0s to reach exactly bits bits

  //    int ndDiff = nbits - dbits; // original length difference l(n) - l(d)

  //    if (nebits < bits) {
  //      an <<= (bits - nebits);    // pad with 0s
  //    }
  //    if (debits < bits) {
  //      ad <<= (bits - debits);    // pad with 0s
  //    }

  //    // compute intervals
  //    using namespace boost::numeric;
  //    using namespace interval_lib;
  //    typedef interval<double> I_aux;
  //    
  //    I_aux::traits_type::rounding rnd;
  //    typedef unprotect<I_aux>::type I;

  //    // lower,upper bounds
  //    const I intn(an, an+1);
  //    const I intd(ad, ad+1);
  //    I q = intn/intd;

  //    if (x < 0) q = -q;

  //    // adjust ratio according to bit length difference
  //    while (ndDiff > 0) { q *= 2.0; ndDiff--; }
  //    while (ndDiff < 0) { q *= 0.5; ndDiff++; }  
  //    
  //    // cout << "bits: " << bits << " " << an << " " << ad << " : " << ((double)an)/ad << endl;
  //    // std::cout << "p! " <<  q.lower() << " " << q.upper() << " ratio: " << x << std::endl;
  //    // compare with ratio computed by GMP for testing
  //  #if TEST  
  //    double v = x;
  //    if (v < q.lower() || v > q.upper()) {
  //      cout << "outside! " <<  q.lower() << " " << q.upper() << " ratio: " << v << endl;
  //      exit(-1);
  //    }
  //  #endif
  //        
  //    xmin = q.lower();
  //    xmax = q.upper();
  //  }

  //
  //
  //  /* approximate 4 points using the highest bits in num/den */
  //  static void approximate( coord_t& p1x,  coord_t& p1y, 
  //                           coord_t& p2x,  coord_t& p2y,
  //                           coord_t& p3x,  coord_t& p3y, 
  //                           coord_t& px,  coord_t& py,
  //                          int bits,
  //                          ApproxPoint& apa, ApproxPoint& apb,
  //                          ApproxPoint& apc, ApproxPoint& apd)
  //  {
  //
  //    
  //    
  //    Exact* p1xe = Geometry::getVar();
  //    Exact* p1ye = Geometry::getVar();
  //    Exact* p2xe = Geometry::getVar();
  //    Exact* p2ye = Geometry::getVar();
  //    Exact* p3xe = Geometry::getVar();
  //    Exact* p3ye = Geometry::getVar();
  //    Exact* pxe = Geometry::getVar();
  //    Exact* pye = Geometry::getVar();
  //
  //
  //     
  //      RatToExact(p1x,p1xe);
  //      RatToExact(p1y,p1ye);
  //      RatToExact(p2x,p2xe);
  //      RatToExact(p2y,p2ye);
  //      RatToExact(p3x,p3xe);
  //      RatToExact(p3y,p3ye);
  //      RatToExact(px,pxe);
  //      RatToExact(py,pye);
  //  
  //
  //      approximate(*p1xe, bits, apa.xmin, apa.xmax);
  //      approximate(*p1ye, bits, apa.ymin, apa.ymax);
  //      approximate(*p2xe, bits, apb.xmin, apb.xmax);
  //      approximate(*p2ye, bits, apb.ymin, apb.ymax);
  //      approximate(*p3xe, bits, apc.xmin, apc.xmax);
  //      approximate(*p3ye, bits, apc.ymin, apc.ymax);
  //      approximate(*pxe, bits, apd.xmin, apd.xmax);
  //      approximate(*pye, bits, apd.ymin, apd.ymax);
  //       
  //}
  //


  /** approximate rational number by double interval [xmin,xmax] using
      a small number of bits to represent the numerator and denominator.
  */
  static void approximate(coord_t& x, int bits, double& xmin, double& xmax)
  {
    if (x.hasApp()) {
      xmin = x.getAppMin();
      xmax = x.getAppMax();
      return;
    }
         
    if (bits < 2 || bits > 30) { std::cout << "bits < 2 | > 30" << std::endl; exit(-1); }

    sint8 n = x.getNum();
    sint8 d = x.getDen();  

    if (n < 0) {
      n = -n;
    }

    int sn = 1;
    int sd = 1;
   
    // collect highest bits from both numbers

    typedef uint8 Limb;
    
    // cout << x << " " << GMP_LIMB_BITS << endl;
    // cout << sn << " " << sd << endl;

   
    // relevant bits in highest limb
    int hn = highestBitSet(n)+1;
    int hd = highestBitSet(d)+1;

      
    //the rational number is 0/den ;
    if (hn < 1 || hd < 1) { 
      //cout << "0 limb" << endl; exit(-1);

      //0/den -> exact
      //restore the sign 
      if (x < 0) {
        n = -n; 
      }
      xmin = xmax = n;
      x.setApp(true, xmin, xmax);
      return;
    }

    // how many relevant bits in n and d
    int nbits = hn + (sn-1)*64;
    int dbits = hd + (sd-1)*64;

    if (dbits == 1 && nbits < 50) {
      // small integer -> exact
      //restore the sign 
      if (x < 0) {
        n = -n; 
      }
	  xmin = xmax = n;
      x.setApp(true, xmin, xmax);

#if TEST
      double v = x;
      if (v != xmin) {
        ERR("not exact! " <<  v << " " << xmin << " " << mpz_get_sint8(n) << " " << mpz_get_sint8(d));
      }
#endif    
      return;
    }
      
    if (abs(nbits - dbits) >= 200) {
      ERR("huge difference " << nbits << " " << dbits);
    }

    // how many bits to extract from n and d
    int nebits = min(bits, nbits);
    int debits = min(bits, dbits);
      
    sint4 an, ad; // highest bits approximating n and d

    // assemble numerator bits
    // if (nebits <= hn) 
           
    // all bits in highest limb    
    an = n >> (hn-nebits);


    // assemble denominator bits  
    // if (debits <= hd) 
         
    // all bits in highest limb    
    ad = d >> (hd-debits);
     

    // pad an and ad with 0s to reach exactly bits bits

    int ndDiff = nbits - dbits; // original length difference l(n) - l(d)

    if (nebits < bits) {
      an <<= (bits - nebits);    // pad with 0s
    }
    if (debits < bits) {
      ad <<= (bits - debits);    // pad with 0s
    }

    // compute intervals
    using namespace boost::numeric;
    using namespace interval_lib;
    typedef interval<double> I_aux;
      
    I_aux::traits_type::rounding rnd;
    typedef unprotect<I_aux>::type I;

    // lower,upper bounds
    const I intn(an, an+1);
    const I intd(ad, ad+1);
    I q = intn/intd;

    if (x < 0) q = -q;

    // adjust ratio according to bit length difference
    while (ndDiff > 0) { q *= 2.0; ndDiff--; }
    while (ndDiff < 0) { q *= 0.5; ndDiff++; }  
      
    // cout << "bits: " << bits << " " << an << " " << ad << " : " << ((double)an)/ad << endl;
    // std::cout << "p! " <<  q.lower() << " " << q.upper() << " ratio: " << x << std::endl;
    // compare with ratio computed by GMP for testing
#if TEST  
    double v = x;
    if (v < q.lower() || v > q.upper()) {
      cout << "outside! " <<  q.lower() << " " << q.upper() << " ratio: " << v << endl;
      exit(-1);
    }
#endif
    xmin = q.lower();
    xmax = q.upper();
    x.setApp(true, q.lower(), q.upper());
  }


  /* approximate 4 points using the highest bits in num/den */
  static void approximate(coord_t& p1x, coord_t& p1y, 
                          coord_t& p2x, coord_t& p2y,
                          coord_t& p3x, coord_t& p3y, 
                          coord_t& px,  coord_t& py,
                          int bits,
                          ApproxPoint& apa, ApproxPoint& apb,
                          ApproxPoint& apc, ApproxPoint& apd)
  {
    approximate(p1x, bits, apa.xmin, apa.xmax);
    approximate(p1y, bits, apa.ymin, apa.ymax);
    approximate(p2x, bits, apb.xmin, apb.xmax);
    approximate(p2y, bits, apb.ymin, apb.ymax);
    approximate(p3x, bits, apc.xmin, apc.xmax);
    approximate(p3y, bits, apc.ymin, apc.ymax);
    approximate(px,  bits, apd.xmin, apd.xmax);
    approximate(py,  bits, apd.ymin, apd.ymax);
  }

  /** @return 0 if d lies on the circle spanned by a b c, +1 if d is
      inside, -1 if outside assuming ccw ordering of a b c. Applies
      quick test first and falls back to the slow test. ~10 times
      faster than exact computation if co-circular cases are rare
  */
  static int filterInCircle(coord_t& p1x, coord_t& p1y, 
                            coord_t& p2x, coord_t& p2y,
                            coord_t& p3x, coord_t& p3y, 
                            coord_t& px,  coord_t& py)
  {
    ApproxPoint apa, apb, apc, apd;

    approximate(p1x,p1y,p2x,p2y,p3x,p3y,px,py, 24, apa, apb, apc, apd);
    int r = approxInCircle(apa, apb, apc, apd);
  
    // cout << ((r == 0) ? "." : "!") << flush;
    
    if (r == 0) {
     
      // undecided -> slow test
   
      RatToExact(p1x,p1xe);
      RatToExact(p1y,p1ye);
      RatToExact(p2x,p2xe);
      RatToExact(p2y,p2ye);
      RatToExact(p3x,p3xe);
      RatToExact(p3y,p3ye);
      RatToExact(px,pxe);
      RatToExact(py,pye);
      r = exactInCircle(p1xe,p1ye,p2xe,p2ye,p3xe,p3ye,pxe,pye);
    
      

    }

    // appCount++;
    // decided
    
#if TEST 
   
    Exact* p1xe = Geometry::getVar();
    Exact* p1ye = Geometry::getVar();
    Exact* p2xe = Geometry::getVar();
    Exact* p2ye = Geometry::getVar();
    Exact* p3xe = Geometry::getVar();
    Exact* p3ye = Geometry::getVar();
    Exact* pxe = Geometry::getVar();
    Exact* pye = Geometry::getVar();
    RatToExact(p1x,p1xe);
    RatToExact(p1y,p1ye);
    RatToExact(p2x,p2xe);
    RatToExact(p2y,p2ye);
    RatToExact(p3x,p3xe);
    RatToExact(p3y,p3ye);
    RatToExact(px,pxe);
    RatToExact(py,pye);
    if (r != exactInCircle(*p1xe,*p1ye,*p2xe,*p2ye,*p3xe,*p3ye,*pxe,*pye)) {
      ERR("in circle ERROR");
    }
#endif
  

    return r;
  }
  /** @return true iff line segments p1-p2 and p3-p4 intersect
      assumes LimRat values are integers!
  */
  static bool exactIntersection(const IntPoint& p1, const IntPoint& p2,
                                const IntPoint& p3, const IntPoint& p4);

  /** @return true iff line segments p1-p2 and p3-p4 intersect
      not tested, not optimized
  */
  static bool exactIntersection(const Exact& p1x,const Exact& p1y, 
                                const Exact& p2x,const Exact& p2y,
                                const Exact& p3x,const Exact& p3y, 
                                const Exact& p4x,const Exact& p4y);

  /** @return true iff line segments p1-p2 and p3-p4 intersect
      not tested, not optimized
  */
  static bool exactIntersection(coord_t& p1x,coord_t& p1y, 
                                coord_t& p2x,coord_t& p2y,
                                coord_t& p3x,coord_t& p3y, 
                                coord_t& p4x,coord_t& p4y);

  /** @return true iff line segments p1-p2 and p3-p4 intersect
      not tested, not optimized
  */
  static bool exactIntersection(coord_t2& p1x,coord_t2& p1y, 
                                coord_t2& p2x,coord_t2& p2y,
                                coord_t2& p3x,coord_t2& p3y, 
                                coord_t2& p4x,coord_t2& p4y);


#if 0
  /** @return true iff line segments p1-p2 and p3-p4 intersect.
      If so, also return intersection point.
      Not optimized yet. Not tested either.
  */
  static bool exactIntersection(const ExactPoint& p1, const ExactPoint& p2,
                                const ExactPoint& p3, const ExactPoint& p4,
                                ExactPoint& ip);

  

  
  /** @return true iff line segments p1-p2 and p3-p4 intersect.
      If so, also return rational intersection point.
  */
  static bool exactIntersection(const IntPoint& p1, const IntPoint& p2,
                                const IntPoint& p3, const IntPoint& p4,
                                LimRatPoint& ip);
#endif
  /** @return +1 if pa,pb and pc are oriented in CCW order,
      -1 if they are CW, 0 if we don't know and we should fall back to exactCCW.   

      | 1  ax  ay |
      d= | 1  bx  by |
      | 1  cx  cy |

      if  dInt > 0, then d > 0,
      if  dInt < 0, then d < 0,
      if  dInt = 0, then d = 0.

      when 0 belongs to dInt, we don't know the sign of the determinant. 
  */
  static int approxCCW(const ApproxPoint& pa, const ApproxPoint& pb,
                       const ApproxPoint& pc)
  {
  
    using namespace boost::numeric;
    using namespace interval_lib;
    typedef interval<double> I_aux;

    I_aux::traits_type::rounding rnd; // switches on default rounding mode
    typedef unprotect<I_aux>::type I;

    const I pax(pa.xmin, pa.xmax);
    const I pay(pa.ymin, pa.ymax);
    const I pbx(pb.xmin, pb.xmax);
    const I pby(pb.ymin, pb.ymax);
    const I pcx(pc.xmin, pc.xmax);
    const I pcy(pc.ymin, pc.ymax);
 
    I det = ((pax*pby)-(pbx*pay)+(pbx*pcy)-(pcx*pby)+(pcx*pay)-(pax*pcy));

    //I det = pax*(pby-pcy)-pbx*(pay-pcy)-pcx*(pby-pay);
  

    using namespace compare::certain;
  
    if (det > 0.) return +1;
    if (det < 0.) return -1;
    return 0;

  }

  /** @return +1 if pa,pb and pc are oriented in CCW order,
      -1 if they are CW, 0 if they are collinear  

      | 1  ax  ay |
      d= | 1  bx  by |
      | 1  cx  cy |
  */
  static int exactCCW(const Exact&  p1x, const Exact&  p1y, 
                      const Exact&  p2x, const Exact&  p2y, 
                      const Exact&  p3x, const Exact&  p3y)
  {
  
    Exact det = ((p1x*p2y)-(p2x*p1y)+(p2x*p3y)-(p3x*p2y)+(p3x*p1y)-(p1x*p3y));
  
    if (det > 0) return +1;
    if (det < 0) return -1;
    return 0;

  }

  /** @return 0 if pa,pb and pc are collinear, +1 if they are CCW,
      -1 if they are CW. Applies quick interval test first and falls
      back to the slow test. ~X times faster than exact computation 
      if co-circular cases are rare
  */
  static int filterCCW(coord_t&  p1x, coord_t&  p1y, 
                       coord_t&  p2x, coord_t&  p2y, 
                       coord_t&  p3x, coord_t&  p3y)
  {
    ApproxPoint apa, apb, apc;
  
    approximate(p1x, 24, apa.xmin, apa.xmax);
    approximate(p1y, 24, apa.ymin, apa.ymax);
    approximate(p2x, 24, apb.xmin, apb.xmax);
    approximate(p2y, 24, apb.ymin, apb.ymax);
    approximate(p3x, 24, apc.xmin, apc.xmax);
    approximate(p3y, 24, apc.ymin, apc.ymax);
  
    int r = approxCCW(apa, apb, apc);

    // cout << ((r == 0) ? "." : "!") << flush;
   
    if (r == 0) {
   
      RatToExact(p1x,p1xe);
      RatToExact(p1y,p1ye);
      RatToExact(p2x,p2xe);
      RatToExact(p2y,p2ye);
      RatToExact(p3x,p3xe);
      RatToExact(p3y,p3ye);
     
      // undecided -> slow test
      r = exactCCW(p1xe, p1ye, p2xe, p2ye, p3xe, p3ye);
      exactCount++;

    } else {

      appCount++;

      // decided
    
#if 0 
    
      RatToExact(p1x,p1xe);
      RatToExact(p1y,p1ye);
      RatToExact(p2x,p2xe);
      RatToExact(p2y,p2ye);
      RatToExact(p3x,p3xe);
      RatToExact(p3y,p3ye);

      if (r !=exactCCW(p1xe, p1ye, p2xe, p2ye, p3xe, p3ye))  {
        exactCCW(p1xe, p1ye, p2xe, p2ye, p3xe, p3ye);
        approxCCW(apa, apb, apc);
        cerr << "ERROR" << endl;
        exit(-1);
      }
#endif
    }
  
    return r;
  }

  //inline static predicates------------------------------------------------------
  static const coord_t zero;
  

  /** convert Rational<long long> to Exact.
      @return true iff OK 
  */
  static bool RatToExact(coord_t& a, Exact& b)
  {
    if (a.getDen() == 0) {
      return false;
    }
    mpz_class& num = b.get_num();
    mpz_class& den = b.get_den();
    mpz_set_sint8(num.get_mpz_t(), a.getNum());
    mpz_set_sint8(den.get_mpz_t(), a.getDen());    
    //b.canonicalize();
 
    return true;
  }


  /** convert Exact to Rational<long long>
      @return true iff OK
  */
  static bool ExactToRat(Exact& a, coord_t& b)
  {
    if (a == 0) {
      b = coord_t( 0, 1);
      return true;
    }
  
    const mpz_class& num = a.get_num();
    const mpz_class& den = a.get_den();
  
    if (!mpz_fits_sint8_p(num.get_mpz_t()) ||
        !mpz_fits_sint8_p(den.get_mpz_t())) {
      return false;
    }
    
    b = coord_t( mpz_get_sint8(num.get_mpz_t()), mpz_get_sint8(den.get_mpz_t()));
    return true;
  }


  static coord_t dist2(const coord_t& x1, const coord_t& y1, const coord_t& x2, const coord_t& y2)
  {
    coord_t dx, dy;
    dx=x1-x2;
    dy=y1-y2;
    return dx*dx + dy*dy;
  }

  static int ccw(coord_t& p1x, coord_t& p1y, coord_t& p2x, coord_t& p2y, coord_t& p3x, coord_t& p3y)
  {
    /*Geometry::ExactPoint pa, pb, pc;
	 

      pa.x = p1x;
      pa.y = p1y;
      pb.x = p2x;
      pb.y = p2y;
      pc.x = p3x;
      pc.y = p3y;*/
    
    int r = Geometry::filterCCW(p1x, p1y, p2x, p2y, p3x, p3y);

    //Post condition: 
    //assert( (ccw_old(p1x, p1y, p2x, p2y, p3x, p3y)>0) == (r==1));

    return r ;
  }

  static int ccw_old2(const coord_t2& ax, const coord_t2& ay, const coord_t2& bx, const coord_t2& by, const coord_t2& cx, const coord_t2& cy)
  {
    coord_t2 res = ((ax*by)-(bx*ay)+(bx*cy)-(cx*by)+(cx*ay)-(ax*cy));

    if(res > 0)
      return 1;
    else if( res < 0)
      return -1;
    else 
      return 0;

    //return ((ax*by)-(bx*ay)+(bx*cy)-(cx*by)+(cx*ay)-(ax*cy));
  }


  static int ccw_old(const coord_t& ax, const coord_t& ay, const coord_t& bx, const coord_t& by, const coord_t& cx, const coord_t& cy)
  {
    coord_t res = ((ax*by)-(bx*ay)+(bx*cy)-(cx*by)+(cx*ay)-(ax*cy));

    if(res > 0)
      return 1;
    else if( res < 0)
      return -1;
    else 
      return 0;
  }


  static bool overlap(const coord_t& x1, const coord_t& y1, const coord_t& x2, const coord_t& y2)
  {
    return x1 == x2 && y1 == y2;

  }

  /*********************************************************************************************************************************************
In-segment test. Determine if the point d lies inside or outside line segment defined by the two points a, b
1.collinear test
| ax  ay   1 |
| bx  by   1 | = 0
| dx  dy   1 |

2.range test

min(ax,bx) < dx < max(ax,bx)
min(ay,by) < dy < max(ay,by)
  **********************************************************************************************************************************************/
  static bool onSegment(coord_t& x1, coord_t& y1, coord_t& x2, coord_t& y2, coord_t& x, coord_t& y)
  {
     
    //collinear test
    //bool collinear = (x1*(y2-y) + x2*(y-y1) + x*(y1-y2) == zero);

    /*Geometry::ExactPoint pa, pb, pc;
      pa.x = x1;
      pa.y = y1;
      pb.x = x2;
      pb.y = y2;
      pc.x = x;
      pc.y = y;*/
    bool collinear = ccw(x1, y1, x2, y2, x, y) == 0; 
    
    if(!collinear)
      {
        return false;
      }
    else //position test. is (x,y) between (x1,y1) and (x2,y2)?
      {
        //Exact comparison needed due to overflows in Rational<long long>
        RatToExact(x1,p1xe);
        RatToExact(y1,p1ye);
        RatToExact(x2,p2xe);
        RatToExact(y2,p2ye);
        RatToExact(x,pxe);
        RatToExact(y,pye);
          
        Exact xMax = p1xe > p2xe ?  p1xe : p2xe;

        Exact xMin = p1xe < p2xe ?  p1xe : p2xe;
        Exact yMax = p1ye > p2ye ?  p1ye : p2ye;
        Exact yMin = p1ye < p2ye ?  p1ye : p2ye;

        if( pxe<xMin || pxe>xMax)
          {
            return false;
          }
        if( pye<yMin || pye>yMax)
          {
            return false;
          }
        return true;
        /*coord_t xMax = x1 > x2 ?  x1 : x2;

          coord_t xMin = x1 < x2 ?  x1 : x2;
          coord_t yMax = y1 > y2 ?  y1 : y2;
          coord_t yMin = y1 < y2 ?  y1 : y2;

          if( x<xMin || x>xMax)
          {
          return false;
          }
          if( y<yMin || y>yMax)
          {
          return false;
          }
          return true;*/
      }
  }


  static bool inCircle(coord_t& p1x, coord_t& p1y,
                       coord_t& p2x, coord_t& p2y,
                       coord_t& p3x, coord_t& p3y,
                       coord_t& px,  coord_t& py)
  {
    int r = Geometry::filterInCircle(p1x, p1y, p2x, p2y, p3x, p3y, px, py);
     

    //std::cout<<"appro"<< Geometry::appCount<<std::endl;
    //std::cout<<"exact"<< Geometry::exactCount<<std::endl;

    //Post condition: 
    //assert(in_circle_old(p1x, p1y, p2x, p2y, p3x, p3y, px, py) == (r==1));

    return r == 1;
  }



  /*********************************************************************************************************************************************
In-circle test. Determine if the point d lies inside or outside the circle defined by the three points a, b, and c in the plane.
Assuming that a, b, c are labeled in counterclockwise order around the circle, the following determinant 
is positive if d is inside the circle, negative if d is outside the circle, and zero if all four points are cocircular.
This generalizes naturally to higher dimensions, e.g., point inside a sphere defined by 4 points. 

| ax  ay  ax^2 + ay^2  1 |
| bx  by  bx^2 + by^2  1 |
| cx  cy  cx^2 + cy^2  1 |
| dx  dy  dx^2 + dy^2  1 |
  **********************************************************************************************************************************************/
  static bool inCircleOld(const coord_t& p1x, const coord_t& p1y,
                          const coord_t& p2x, const coord_t& p2y,
                          const coord_t& p3x, const coord_t& p3y,
                          const coord_t& px,  const coord_t& py)
  {
     
    //assert ccw ordering 
    coord_t p1z = p1x*p1x + p1y*p1y;
    coord_t p2z = p2x*p2x + p2y*p2y;
    coord_t p3z = p3x*p3x + p3y*p3y;
    coord_t pz  = px*px   + py*py;

    coord_t m12 = p2x*p3y - p2y*p3x;
    coord_t m13 = p2x*p3z - p2z*p3x;
    coord_t m14 = p2x     - p3x;
    coord_t m23 = p2y*p3z - p2z*p3y;
    coord_t m24 = p2y     - p3y;
    coord_t m34 = p2z     - p3z;
    coord_t d1  = p1y*m34 - p1z*m24 + m23;
    coord_t d2  = p1x*m34 - p1z*m14 + m13;
    coord_t d3  = p1x*m24 - p1y*m14 + m12;
    coord_t d4  = p1x*m23 - p1y*m13 + p1z*m12;

    coord_t det = d4 - px*d1 + py*d2 - pz*d3;

    return det > zero;
  }

  static bool coincident_segments(coord_t& x1,coord_t& y1, coord_t& x2, coord_t& y2,coord_t& x3,coord_t& y3,coord_t& x4,coord_t& y4)
  {
    coord_t denom = (y4-y3)*(x1-x2)-(y1-y2)*(x4-x3);

    coord_t nume_a = ((y4-y3)*(x4-x2)-(x4-x3)*(y4-y2)) ;
    coord_t nume_b = ((y4-y2)*(x1-x2)-(y1-y2)*(x4-x2)) ;
    
    if ( denom == zero ) 
      {
        if(nume_a == zero && nume_b == zero)
          {
            return true;// they are coincident
          }
      }
    return false;

  }

  // a pseudo segment intersect check based on ccw checks
  static bool segments_intersect(coord_t& x1, coord_t& y1,
                                 coord_t& x2, coord_t& y2,
                                 coord_t& x3, coord_t& y3,
                                 coord_t& x4, coord_t& y4)
  {
    int a = ccw(x1,y1,x2,y2,x3,y3);
    int b = ccw(x1,y1,x2,y2,x4,y4);
    
    if ((a > 0 && b > 0) || (a < 0 && b < 0)) {

      return false;

    } else {

      int c = ccw(x3,y3,x4,y4,x1,y1);
      int d = ccw(x3,y3,x4,y4,x2,y2);

      if ((c > 0 && d > 0) || (c < 0 && d < 0)) {
        return false;
      }
    }
    
    return true;
  }

  /*
    static bool segments_intersect(coord_t& x1,coord_t& y1, coord_t& x2, coord_t& y2,coord_t& x3,coord_t& y3,coord_t& x4,coord_t& y4)
    {
    coord_t denom = (y4-y3)*(x1-x2)-(y1-y2)*(x4-x3);
    if ( denom == zero ) 
    {
    return false;// they are parallel
    }

    coord_t nume_a = ((y4-y3)*(x4-x2)-(x4-x3)*(y4-y2)) ;
    coord_t nume_b = ((y4-y2)*(x1-x2)-(y1-y2)*(x4-x2)) ;
     
    if(denom > zero)
    {          
    if ( nume_a <= zero || nume_a >= denom ) 
    {
    return false; // outside [p1,p2]
    }
         
    if ( nume_b <= zero || nume_b >= denom ) 
    {
    return false; // outside [p3,p4]
    }

    }
    else //(denom < 0)
    {
    if ( nume_a >= zero || nume_a <= denom ) 
    {
    return false; // outside [p1,p2]
    }

    if ( nume_b >= zero || nume_b <= denom ) 
    {
    return false; // outside [p3,p4]
    }

    }
  
    return true;

    }*/


  static void exactFindIntersect(coord_t& x1, coord_t& y1,
                                 coord_t& x2, coord_t& y2,
                                 coord_t& x3, coord_t& y3,
                                 coord_t& x4, coord_t& y4,
                                 coord_t& vx, coord_t& vy)
  {
    
    Exact x1e,y1e,x2e,y2e,x3e,y3e,x4e,y4e;
    RatToExact(x1,x1e);
    RatToExact(y1,y1e);
    RatToExact(x2,x2e);
    RatToExact(y2,y2e);
    RatToExact(x3,x3e);
    RatToExact(y3,y3e);
    RatToExact(x4,x4e);
    RatToExact(y4,y4e);

      
     
    Exact denom = (y4e-y3e)*(x1e-x2e)-(y1e-y2e)*(x4e-x3e);
    Exact nume_a = ((y4e-y3e)*(x4e-x2e)-(x4e-x3e)*(y4e-y2e)) ;
    Exact nume_b = ((y4e-y2e)*(x1e-x2e)-(y1e-y2e)*(x4e-x2e)) ;
    Exact px = (nume_a/denom)*(x1e-x2e) + x2e;
    Exact py = (nume_a/denom)*(y1e-y2e) + y2e;

    ExactToRat(px,vx);
    ExactToRat(py,vy);

     
  
  }

  //invoke exact version for integer grid sizes over [-1,000,000,1,000,000] 
  //static void find_intersect(coord_t& x1,coord_t& y1, coord_t& x2, coord_t& y2,coord_t& x3,coord_t& y3,coord_t& x4,coord_t& y4, coord_t& vx, coord_t& vy)
  //{
  //     exact_find_intersect(x1,y1,x2,y2,x3,y3,x4,y4,vx,vy);
  //  
  //}


  static void findIntersect(const coord_t& x1, const coord_t& y1,
                            const coord_t& x2, const coord_t& y2,
                            const coord_t& x3, const coord_t& y3,
                            const coord_t& x4, const coord_t& y4,
                            coord_t& vx, coord_t& vy)
  {
    coord_t denom = (y4-y3)*(x1-x2)-(y1-y2)*(x4-x3);
    coord_t nume_a = ((y4-y3)*(x4-x2)-(x4-x3)*(y4-y2)) ;
    coord_t nume_b = ((y4-y2)*(x1-x2)-(y1-y2)*(x4-x2)) ;
    vx = (nume_a/denom)*(x1-x2) + x2;
    vy = (nume_a/denom)*(y1-y2) + y2;
  }

};


#endif

