#ifndef RATIONAL_H
#define RATIONAL_H

#include <iostream>
#include <cmath>
#include <cstdlib>
#include <limits>
#include <cassert>

//#define OVERFLOWCHECKS

template <class T=long long>
class Rational
{

private: //data
  T _numerator;
  T _denominator;

  // floating point approximation
  bool   _hasApp;
  double _appMin, _appMax; // would floats suffice here?
        
public:

  // constructors		

  Rational(): _numerator(0), _denominator(1), _hasApp(false) {};
  Rational(T numerator): _numerator(numerator), _denominator(1), _hasApp(false) { _appMin = _appMax = 0; };
  Rational(T numerator, T denominator);

  //data access

  T getNum() const{ return _numerator; };
  T getDen() const { return _denominator; };

  //void set_num(const T& n) { _numerator = n; };
  //void set_den(const T& d) { _denominator = d; };

  bool   hasApp()          { return _hasApp; }
  double getAppMin() const { return _appMin; }
  double getAppMax() const { return _appMax; }  

  void setApp(bool has, double min, double max) {
    _hasApp = has;
    _appMin = min;
    _appMax = max;
  }

  //operators

  // The assignment operator	
  Rational<T>& operator=(const Rational<T>& r);
     	
  //Rational power(int power) const;
  Rational<T> invert() const;
     	
  // this is the pre-increment operator
  const Rational<T>& operator++();
     	
  // this is the post-increment operator
  const Rational<T> operator++(int);
     	
  // this is the pre-decrement operator
  const Rational<T>& operator--();
     	
  // this is the post-decrement operator
  const Rational<T> operator--(int);
     	
  // this is the negation operator
  const Rational<T>& operator-();
         
  //binary operators
  const Rational<T> operator-(const Rational<T>& r) const;
  const Rational<T> operator+(const Rational<T>& r) const;
  const Rational<T> operator*(const Rational<T>& r) const;
  const Rational<T> operator/(const Rational<T>& r) const;

  //comparisons
  bool operator==(const Rational<T>& r) const;
  bool operator< (const Rational<T>& r) const;
  bool operator<=(const Rational<T>& r) const;
  bool operator> (const Rational<T>& r) const;
  bool operator>=(const Rational<T>& r) const;


  // The output stream operator.  
  friend std::ostream& operator<<(std::ostream& out, const Rational<T>& r)
  {
    out << "(" << r._numerator << "/" << r._denominator << ")";
    return out ;
  }
            
  //gcd 
  static T gcd(T a, T b) ;
            
  //integral check 
  bool isInteger() const;

  //convert to double
  double getD() const;
            
private:
     	
  void normalize();
};  


using namespace std; // FIXME: avoid this in h files


/**
 * Creates a rational number with the given
 * numerator and denominator
 */
template <class T>
Rational<T>::Rational(T numerator, T denominator): _numerator(numerator), _denominator(denominator) 
{
  if(_denominator == 0) {
    printf("denominator is zero");
    exit(0);
  }
  _hasApp = false;
  normalize();
}


/**
 * The normalize function moves the sign to numerator 
 * and removes common divisors from the numerator and 
 * denominator
 */
template <class T>
inline void Rational<T>::normalize() 
{
  T zero(0);

  // Handle the case of zero separately, to avoid division by zero
  if (_numerator == zero) {
    _denominator = T(1);
    return;
  }

  T g = gcd(_numerator, _denominator);

  if (g != T(1)) {
    _numerator   /= g;
    _denominator /= g;
  }
  
  // Ensure that the denominator is positive
  if (_denominator < zero) {
    _numerator = -_numerator;
    _denominator = -_denominator;
  }
}

/**
 * Assignment operator 
 */
template <class T>
inline Rational<T>& Rational<T>::operator=(const Rational<T>& r) 
{
  _numerator = r._numerator;
  _denominator = r._denominator;
  _hasApp = r._hasApp;
  _appMax = r._appMax;
  _appMin = r._appMin;
  // normalize(); // not needed if r is already normalized 
  return *this;
}


/**
* Inversion function
*/
template <class T>
inline Rational<T> Rational<T>::invert( ) const 
{
  return Rational<T>( _denominator, _numerator);
}


/**
 * Adding two rational numbers a/b and c/d is defined as
 * (a*d + c*b) / (b*d)
 * 
 */
template <class T>
inline const Rational<T> Rational<T>::operator+(const Rational<T>& r) const 
{

#ifdef OVERFLOWCHECKS
  //overflow checks for addition 
     
  //first normalize the numerator parts
  //NOTE: omg!!! long long multiplication overflow before normalization<============================
  // FIXME: this can cause overflows already!
  Rational<T> partA( (_numerator * r._denominator), _denominator * r._denominator) ;
  Rational<T> partB( (r._numerator * _denominator), _denominator * r._denominator) ;

  T lhs = partA._numerator;
  T rhs = partB._numerator;
     
  if( (rhs>0 && lhs>0) || (rhs<0 && lhs<0)  ) {
    //either two negatives or two positives
    if(rhs < 0) {
      //two negatives
      if (lhs < std::numeric_limits<T>::min() - rhs) {
        printf("addition overflow \n");
        std::cout<<*this<<" "<<r<<"\n";
        std::cout<<rhs<<" "<<lhs<<" min: "<< std::numeric_limits<T>::min()<<"\n";
        assert(false);
      }
      // ok
    } else {
      // two positives
      if(std::numeric_limits<T>::max() - lhs < rhs) {
        printf("addition overflow \n");
        std::cout<<*this<<" "<<r<<"\n";
        std::cout<<rhs<<" "<<lhs<<" max: "<< std::numeric_limits<T>::max()<<"\n";
        assert(false);

      }
      //ok
    }
  }
  // else overflow not possible
#endif

  T num   = _numerator;
  T den   = _denominator;
  T r_num = r._numerator;
  T r_den = r._denominator;

  T g = Rational<T>::gcd(den, r_den);
  den /= g;  
  num = num * (r_den / g) + r_num * den;
  g = gcd(num, g);
  num /= g;
  den *= r_den/g;

  return Rational<T>(num,den);
	
}


/**
 * Subtracting two rational numbers a/b and c/d is defined as
 * (a*d - c*b) / (b*d)
 * or
 * r1 + -1*r2  (where r1 = a/b and r2 = c/d)
 */
template <class T>
inline const Rational<T> Rational<T>::operator-(const Rational<T>& r) const 
{
#ifdef OVERFLOWCHECKS
  //overflow checks for subtraction
  //first normalize the numerator parts

  // FIXME: this can cause overflows already!
  Rational<T> partA( (_numerator * r._denominator), _denominator * r._denominator) ;
  Rational<T> partB( (r._numerator * _denominator), _denominator * r._denominator) ;

  T lhs = partA._numerator;
  T rhs = partB._numerator;
     
  if ((rhs>0 && lhs< 0) || (rhs<0 && lhs>0)) {
    //mixed positive and negative
    //two cases  +X - -Y => X + Y   check for overflow against MaxInt()
    //            -X - +Y   check for overflow against MinInt()

    if (lhs >= 0) { //first case
      //test is X - -Y > MaxInt()
      //equivalent to X > MaxInt() - +Y
      //Y == MinInt() creates special case
      //Even 0 - MinInt() can't be done
      //note that the special case collapses into the general 
      //case, due to the fact
      //MaxInt() - MinInt() == -1, and lhs is non-negative
      if(lhs > std::numeric_limits<T>::max() + rhs) //remember that rhs is negative
        {
          printf("subtraction overflow \n");
          assert(false);
        }
      //fall through to return value
    } else {
      //second case
      //test is -X - Y < MinInt()
      //or      -X < MinInt() + Y
      //we do not have the same issues because 
      //abs(MinInt()) > MaxInt()
      if (lhs < std::numeric_limits<T>::min() + rhs) {
        printf("subtraction overflow \n");
        assert(false);
      }
      //fall through to return value
    }
  }
  // else 
  //both negative, or both positive
  //no possible overflows
#endif
    	
  T num = _numerator;
  T den = _denominator;
  T r_num = r._numerator;
  T r_den = r._denominator;

  T g = Rational<T>::gcd(den, r_den);
  den /= g;
  num = num * (r_den / g) - r_num * den;
  g = gcd(num, g);
  num /= g;
  den *= r_den/g;

  return Rational<T>(num,den);
}

/**
 * Multiplying two rational numbers a/b and c/d is defined as
 * a*c / b*d
 */
template <class T>
inline const Rational<T> Rational<T>::operator*(const Rational<T>& r) const 
{
#ifdef OVERFLOWCHECKS
  //overflow checks for multiplication
  T a = _numerator;
  T b = r._numerator;

  if (a > 0 && b > 0) {
         
    if (a > std::numeric_limits<T>::max()/b) {
      printf("multiplication overflow \n");
      assert(false);
    }

  } else if (a < 0 && b < 0) {
         
    if (a < std::numeric_limits<T>::max()/b) {
      printf("multiplication overflow \n");
      assert(false);
    }

  } else if (a > 0 && b < 0) {
         
    if (b < std::numeric_limits<T>::min()/a) {
      printf("multiplication overflow \n");
      assert(false);
    }

  } else if (a < 0 && b > 0) {
         
    if (a < std::numeric_limits<T>::min()/b) {
      printf("multiplication overflow \n");
      assert(false);
    }
  }

  //overflow checks for multiplication
  a = _denominator;
  b = r._denominator;

  if (a > 0 && b > 0) {
         
    if (a > std::numeric_limits<T>::max()/b) {
      printf("multiplication overflow \n");
      assert(false);
    }

  } else if (a < 0 && b < 0) {
         
    if (a < std::numeric_limits<T>::max()/b) {
      printf("multiplication overflow \n");
      assert(false);
    }
    
  } else if (a > 0 && b < 0) {
         
    if (a > std::numeric_limits<T>::min()/b) {
      printf("multiplication overflow \n");
      assert(false);
    }
    
  } else if (a < 0 && b > 0) {
         
    if (a < std::numeric_limits<T>::min()/b) {
      printf("multiplication overflow \n");
      assert(false);
    }
  }
#endif

  T zero(0);
  T num = _numerator;
  if (num   == zero) return *this;

  T r_num = r._numerator;
  if (r_num == zero) return r;
  
  T den   = _denominator;
  T r_den = r._denominator;

  // Avoid overflow and preserve normalization
  T gcd1 = Rational<T>::gcd(num, r_den);
  T gcd2 = Rational<T>::gcd(r_num, den);
  num = (num/gcd1) * (r_num/gcd2);
  den = (den/gcd2) * (r_den/gcd1);

  if (den < zero) {
    num = -num;
    den = -den;
  }
  
  return Rational<T>(num,den);
}

/**
 * Dividing Rational numbers a/b and c/d is defined as
 * a*d / b*c
 * or
 * r1 * r2^(-1) (mulitiplication of the inverse)
 * where r1 = a/b and r2 = c/d
 */
template <class T>
inline const Rational<T> Rational<T>::operator/(const Rational<T>& r) const 	
{
  T num = _numerator;
  T den = _denominator;
  T r_num = r._numerator;
  T r_den = r._denominator;

  // Avoid repeated construction
  T zero(0);

  // Trap division by zero
  if (r_num == zero) {
    printf("division by 0!");
    exit(0);
  }

  if (num == zero) return *this;

  // FIXME: just call multiplication with inverted r (which also checks for overflow!)
  
  // Avoid overflow and preserve normalization
  T gcd1 = Rational<T>::gcd(num, r_num);
  T gcd2 = Rational<T>::gcd(r_den, den);
  num = (num/gcd1) * (r_den/gcd2);
  den = (den/gcd2) * (r_num/gcd1);

  if (den < zero) {
    num = -num;
    den = -den;
  }

  return Rational<T>(num,den);
}

/**
 * Two rational numbers are equal if their cross product is equal.
 * That is, if the numerator of this number times the denominator
 * of r is equal to the numerator of r times the denominator of this
 * then the two rationals are equal
 */
template <class T>
inline bool Rational<T>::operator==(const Rational& r) const 
{
  // FIXME: check for overflow!
  return (_numerator * r._denominator) == (_denominator * r._numerator);
}

/**
 * Postfix unary plus operation
 */
template <class T>
inline const Rational<T> Rational<T>::operator++(int) 
{
  // store the old value
  Rational<T> store(*this);
  // FIXME: check for overflow!  
  // perform the update
  _numerator += _denominator;
  // normalize(); not necessary if already normalized    
  return store;
}
		
/**
 * Prefix unary plus operation
 */
template <class T>
inline const Rational<T>& Rational<T>::operator++() { 
  _numerator += _denominator;
  // normalize(); not necessary if already normalized  
  return *this;
}


/**
 * Postfix unary minux operation
 */
template <class T>
inline const Rational<T> Rational<T>::operator--(int) {
  // store the old value
  Rational<T> store(*this);

  // FIXME: overflow detection!  
  _numerator -= _denominator;
  // normalize(); not necessary if already normalized
  return store;
}
	
/**
 * Prefix unary minux operation
 */
template <class T>
inline const Rational<T>& Rational<T>::operator--() {
  // FIXME: overflow detection!  
  _numerator -= _denominator;
  // normalize(); not necessary if already normalized  
  return *this;
}

/**
 * uniary negation operation
 */
template <class T>
inline const Rational<T>& Rational<T>::operator-()
{
  // FIXME: overflow detection!  
  _numerator = -_numerator;
  // normalize(); not necessary if already normalized
  return *this;
}

/**
 * The less than operator
 */
template <class T>
inline bool Rational<T>::operator<(const Rational<T>& r) const 
{
  // FIXME: overflow detection!
  return _numerator * r._denominator < _denominator * r._numerator;
}

/**
 * The less than or equal operator.
 */
template <class T>
inline bool Rational<T>::operator<=(const Rational<T>& r) const {
  // FIXME: overflow detection!  
  return (_numerator * r._denominator < _denominator * r._numerator) || (_numerator * r._denominator == _denominator * r._numerator) ;
}

/**
 * The greater than operator (defined in terms of <=)
 */
template <class T>
inline bool Rational<T>::operator>(const Rational<T>& r) const 
{
  // FIXME: overflow detection!  
  return _numerator * r._denominator > _denominator * r._numerator;
}

/**
 * The greater than or equal opreator (defined in terms of <)
 */
template <class T>
inline bool Rational<T>::operator>=(const Rational<T>& r) const {
  // FIXME: overflow detection!
  return (_numerator * r._denominator > _denominator * r._numerator) || (_numerator * r._denominator == _denominator * r._numerator) ;
}
	


/**
 * Returns the greatest common divisor
 * of a and b
 */
template <class T>
inline T Rational<T>::gcd(T n, T m) 
{
  T zero(0);

  if (n < zero)
    n = -n;
  if (m < zero)
    m = -m;

  for(;;) {
    if(m == zero)
      return n;
    n %= m;
    if(n == zero)
      return m;
    m %= n;
  }
}

/**
 * Checks if the rational number is an integer
 * 
 */
template <class T>
inline bool Rational<T>::isInteger() const
{
  normalize();
  return _denominator == 1;
}


/**
 * Convert the rational number to a floating point number
 * 
 */
template <class T>
inline double Rational<T>::getD() const
{
  return (double)_numerator/(double)_denominator;
}

#endif // __RATIONAL_H__
