///  Face.h

/// Kai: Declaration of base face class 

#ifndef FACE_H
#define FACE_H

#include <list>
#include <vector>
#include "Rational.h"
#include "Vertex.h"

class Face 
{
public:

  /// typedefs
  /// typedef mpq_class coord_t;
  /// typedef double coord_t

  /**
   * Default constructor.
   * 
  */
  Face();
   
  /**
   * Default destructor.
  */
  virtual ~Face();

public: /// accessors and mutators
  

  /// vetext type for the mesh.
  typedef Vertex V; 
  
  /// edge type of the mesh
  typedef HalfEdge H;
  
  /// number type for the coordinate 
  typedef V::coord_t C;
  
  /** accessor for the number of half edges in the mesh
  */
  int getEdgeCount() const { return _edges.size(); }

  /** accessor for the adjacent half edge with index i 
  /// @param i the ith halfedge in _edges
  /// @return returns the ith halfedge
  */
  H* getEdge(int i) const { return _edges[i]; }
  
  /** mutator for setting the adjacent half edges of the face
  /// e1, e2, and e3 will be oriented in ccw order 
  /// @param e1 one of the halfedges in the face
  /// @param e2 one of the halfedges in the face
  /// @param e3 one of the halfedges in the face
  /// 
  */
  void setEdges(H* e1, H* e2, H* e3);
  
  /** mutator for setting the adjacent half edges of the face
  /// @param v vector of halfedges which contains the adjacent half edges in ccw order
  */
  void setEdges(const std::vector<H*>& v);
  
  /** mutator for the bourndary flag.
  /// the flat is set to true if the face represents the boundary face
  /// @param b boolean value setting the _boundary flag. 
  */
  void setBoundary(bool b) { _boundary = b; }
  
  /** accessor for the bourndary flag.
  /// @return returns true if the face represents the exterior face
  */
  bool isBoundary() const { return _boundary; }

  /** function checks if vertex v is adjacent to the face
  /// @param v the vertex to be cheched within the face
  */
  bool containsVertex(const Vertex* v) const;

  /** accessor for the vector index of the face stored in Mesh::_faces
  /// @return the vector index of the face
  */
  int getVecIndex() const { return vecIndex; }
  
  /** mutator for the vector index of the face stored in Mesh::_faces
  /// the vector index info. is used for vector swaps in face removal.
  /// @param i the vector index value to be set
  */
  void setVecIndex(int i) { vecIndex = i; }

  /** accessor for the marking value.
  /// the return _mark value will be check with the current global marking index for faces.
  */
  int getMark() const { return _mark; }
  
  /** mutator for the makring value. 
  /// m is set to a global marking index when begin_marking() is initiated.
  */
  void setMark(int m) { _mark = m; }
  
private: /// data

  /** stores the adjacent half edges in ccw order
  */
  std::vector<H*> _edges;
  
  /** flag for exterior or interior face
  */
  bool _boundary;

  /** used for marking. 
  /// _mark is assigned to the global marking value when mark() is called after begin_marking()
  /// the face is marked when _mark == the global marking value 
  /// the global marking value is increased when end_marking() is called  
  */
  int _mark;
  
  /**
  /// stores the sector index for sector-based point location.
  /// Note: it is not needed when using the smart sector update scheme
  /// check pointLocation() for details
  /// deprecated
  */
  std::list<std::pair<int,int> > attachedSectors;
 
  /** stores the vector position of a face in the vector Mesh::faces
  */
  int vecIndex;
};

#endif
