#ifndef HALFEDGE_H
#define HALFEDGE_H

#include <list>

class Vertex;
class Face;

class HalfEdge 
{
public:

  HalfEdge();
  ~HalfEdge();

  ///  HalfEdge(Vertex* orgItr, Vertex* destItr);

  typedef Vertex V;
  typedef Face F;
  
private: /// data

  /** store the integer left and right endpoint of a constrained edge
  /// Note: the unconstrained edges do not always have integer endpoints
  */
  std::pair<bool,V*> _leftEnd, _rightEnd;
  
  /** the origin vertex 
  */
  V* _orgItr;
  
  /** the destination vertex
    */
  V* _destItr;
  
  /** pointer to the next edge in ccw order within a face
    */
  HalfEdge* _next;
  
  /** pointer to the prior edge in ccw order within a face
    */
  HalfEdge* _prior;
  
  /** pointer to the opposite half edge. (the adjacent half edge in opposite direction)  
    */
  HalfEdge* _oppEdge;
  
  /** deprecated
    */
  HalfEdge* _itrPos;
  
  /** pointer to the adjacent face
    */
  F* _attachedFace;
  
  /** flag indicating if it is a boundary edge.
  /// Note. the boundary edge is adjacent to the exterior face 
    */
  bool _boundary;
  
  /** flag indicating if the half edge is constrained
    */
  bool _constrained;

  /** a list of constrained IDs. 
  /// Note. the half edge could have multiple IDs when several constrained edges overlap
    */
  std::list<int> _ids;
  
  /** used for marking the edges within algorithms
    */
  int _mark;
  
  /** stores the vector index of the current half edge in Mesh::_halfedges
  /// Note: the index is exchanged  in vector element swaps during edge removal
    */
  int vecIndex;

public: /// accessors and mutators

  /** mutator for the vector index of the half edge stored in Mesh::_halfedges
  /// the vector index info. is used for vector swaps in face removal.
  /// @param i the vector index value to be set
  */
  void setVecIndex(int i) { vecIndex = i; }
  
  /** accessor for the vector index of the half edge stored in Mesh::_halfedges
  /// @return the vector index of the half edge
  */
  int getVecIndex() const { return vecIndex; }

  /** mutator for the makring value. 
  /// @param m the global marking index after begin_marking() is called.
  */
  void setMark(int m) { _mark = m; }
  
   /** accessor for the marking value.
  /// @return the _mark value. the value is used to check with the current global marking index for half edges.
  */
  int getMark() const { return _mark; }
  
  
  /**
  ///
  */
  void set(V* orgItr, V* destItr);
  
  /**
  ///
  */
  V* getOrgVertex() { return _orgItr; }
  
  /**
  ///
  */
  V* getDestVertex() { return _destItr;}
  
  /**
  ///
  */
  HalfEdge* getNext() { return _next; }
  
  /**
  ///
  */
  HalfEdge* getPrior() { return _prior; }
  
  /**
  ///
  */
  HalfEdge* getOppEdge() { return _oppEdge; }
  
  /**
  ///
  */
  Face* getFace() { return _attachedFace; }
  
  /**
  ///
  */
  std::pair<bool,V*>& getLeftEnd() { return _leftEnd; }
  
  /**
  ///
  */
  std::pair<bool,V*>& getRightEnd() { return _rightEnd; }
  
  /**
  ///
  */
  void setLeftEnd(std::pair<bool,V*>& l) { _leftEnd = l; }
  
  /**
  ///
  */
  void setRightEnd(std::pair<bool,V*>& r) { _rightEnd = r; }
  
  /**
  ///
  */
  void setOrgVertex(V* vItr) { _orgItr = vItr; }
  
  /**
  ///
  */
  void setDestVertex(V* vItr) { _destItr = vItr; }
  
  /**
  ///
  */
  void setNext(HalfEdge* eItr) { _next = eItr; }
  
  /**
  ///
  */
  void setPrior(HalfEdge* eItr) { _prior = eItr; }
  
  /**
  ///
  */
  void setOppEdge(HalfEdge* eItr) { _oppEdge = eItr; }
  
  /**
  ///
  */
  void setFace(Face* fItr) { _attachedFace = fItr; }
  
  /**
  ///
  */
  void setBoundary(bool b) { _boundary = b; }
  
  /**
  ///
  */
  bool isBoundary() const { return _boundary; }
  
  /**
  ///
  */  
  void setConstrained(int id)
  {
    _constrained = true;
    _ids.push_back(id);
  }

  void setConstrained(const std::list<int>& ids)
  {
    _constrained = true;
    _ids = ids;
  }

  void setUnconstrained() { _constrained = false; } 
  const std::list<int>& getConstraintIDs() const { return _ids; }
  bool isConstrained() { return _constrained; }

  
  void removeID(int id);
  bool hasConstraintID(int id);
};

#endif
