#include "Mesh.h"

#include "Face.h"
#include "Vertex.h"
#include "HalfEdge.h"

Mesh::~Mesh()
{
  FORALL (_vertices, it) {
    delete *it;
  }

  FORALL (_faces, it) {
    delete *it;
  }

  FORALL (_halfedges, it) {
    delete *it;
  }
}

Mesh::Mesh(MeshObjFactory& mof)
  : _mof(mof)
{
  //setting a boundary face in face list.
  //this _boundaryFace will be referenced to by boundary edges.
  _boundaryFace = mof.newFace();
  _boundaryFace->setBoundary(true);
  _faces.push_back(_boundaryFace);
  _marking = false;
  _curmark = 0;

  //_index_max_value = std::numeric_limits<unsigned int>::max();  
}


Mesh::Mesh(MeshObjFactory& mof, int xmax, int ymax)
  : _mof(mof), _xmax(xmax), _ymax(ymax)
{
  std::vector<H*> faceEdges;
 
  C cxmin = 0;
  C cxmax = _xmax;  
  C cymin = 0;
  C cymax = _ymax;  
  
  V* v1 = createVertex(cxmin, cymin);
  V* v2 = createVertex(cxmax, cymin);
  V* v3 = createVertex(cxmax, cymax);
  V* v4 = createVertex(cxmin, cymax);
	
  H* e1 = createEdge(v1,v2);
  e1->setBoundary(true);
  e1->getOppEdge()->setBoundary(true);
  e1->getOppEdge()->setFace(_boundaryFace);
  faceEdges.push_back(e1);
	
  H* e2 = createEdge(v2,v3);
  e2->setBoundary(true);
  e2->getOppEdge()->setBoundary(true);
  e2->getOppEdge()->setFace(_boundaryFace);
  faceEdges.push_back(e2);
 	
  H* e3 = createEdge(v3,v4);
  e3->setBoundary(true);
  e3->getOppEdge()->setBoundary(true);
  e3->getOppEdge()->setFace(_boundaryFace);
  faceEdges.push_back(e3);

  H* e4 = createEdge(v4,v1);
  e4->setBoundary(true);
  e4->getOppEdge()->setBoundary(true);
  e4->getOppEdge()->setFace(_boundaryFace);
  faceEdges.push_back(e4);

  F* f1 = createFace(faceEdges);
  splitFace(f1,v3,v1);
	
  //setting a boundary face in face list.
  //this _boundaryFace will be referenced to by boundary edges.
  _boundaryFace = mof.newFace();
  _boundaryFace->setBoundary(true);
  _faces.push_back(_boundaryFace);

  _marking = false;
  _curmark = 0;

  //_index_max_value = std::numeric_limits<unsigned int>::max();     
}
void Mesh::init(int xmax, int ymax)
{
  _xmax = xmax;
  _ymax = ymax;
	
  std::vector<H*> faceEdges;
 
  C cxmin = 0;
  C cxmax = _xmax;  
  C cymin = 0;
  C cymax = _ymax;  
  
  V* v1 = createVertex(cxmin, cymin);
  V* v2 = createVertex(cxmax, cymin);
  V* v3 = createVertex(cxmax, cymax);
  V* v4 = createVertex(cxmin, cymax);
	
  H* e1 = createEdge(v1,v2);
  e1->setBoundary(true);
  e1->getOppEdge()->setBoundary(true);
  e1->getOppEdge()->setFace(_boundaryFace);
  faceEdges.push_back(e1);
	
  H* e2 = createEdge(v2,v3);
  e2->setBoundary(true);
  e2->getOppEdge()->setBoundary(true);
  e2->getOppEdge()->setFace(_boundaryFace);
  faceEdges.push_back(e2);
 	
  H* e3 = createEdge(v3,v4);
  e3->setBoundary(true);
  e3->getOppEdge()->setBoundary(true);
  e3->getOppEdge()->setFace(_boundaryFace);
  faceEdges.push_back(e3);

  H* e4 = createEdge(v4,v1);
  e4->setBoundary(true);
  e4->getOppEdge()->setBoundary(true);
  e4->getOppEdge()->setFace(_boundaryFace);
  faceEdges.push_back(e4);

  F* f1 = createFace(faceEdges);
  splitFace(f1,v3,v1);

}
#if 0

// fixme: is this function actually used?  .erase doesn't work (need to pass iterator)
 
void Mesh::spliceFaces(H* e1, H* e3)
{

  //assert edge e1 and e3 are in different direction
  //assert(e1->getDestVertex()->getPosX()  == e3->getOrgVertex()->getPosX() && e1->getDestVertex()->getPosY()  == e3->getOrgVertex()->getPosY());
  //assert(e1->getOrgVertex()->getPosX()  == e3->getDestVertex()->getPosX() && e1->getOrgVertex()->getPosY()  == e3->getDestVertex()->getPosY());
	
  H* e2 = e1->getOppEdge();
  H* e4 = e3->getOppEdge();

  //replace the opposite edges.
  e1->setOppEdge(e3);
  e3->setOppEdge(e1);

  _halfedges.erase(e2);
  _halfedges.erase(e4);
}

#endif
 
Mesh::F* Mesh::joinFaces(F* f1, F* f2, H* e1Itr)
{
  // FIXME: not needed? F* f;
  if (f1 == f2) {
    std::vector<H*> edges;
    H* e2Itr = e1Itr;
    e2Itr = e2Itr->getNext();

    do {
      edges.push_back(e2Itr);
      e2Itr->setFace(f1);
      e2Itr = e2Itr->getNext();
    } while (e2Itr != e1Itr->getOppEdge());

    //reuse f1
    f1->setEdges(edges);
    removeEdge(e1Itr);

  } else {
    
    std::vector<H*> edges;
    H* e2Itr = e1Itr;
    e2Itr = e2Itr->getNext();

    do {
      edges.push_back(e2Itr);
      e2Itr->setFace(f1);
      e2Itr = e2Itr->getNext();

    } while (e2Itr != e1Itr);

    H* e1OppItr = e1Itr->getOppEdge();
    H* e4Itr = e1OppItr;
    e4Itr = e4Itr->getNext();

    do {
      edges.push_back(e4Itr);
      e4Itr->setFace(f1);
      e4Itr = e4Itr->getNext();

    } while(e4Itr != e1OppItr);

    //reuse f1
    f1->setEdges(edges);
    removeEdge(e1Itr);
    removeFace(f2);
  }
  
  return f1;
}



/* split the face f with edge e(v1,v2)
   return one of the halfedge that splits the face 
*/
 
Mesh::H* Mesh::splitFace(F* f, V* v1, V* v2)
{
  H* eItr = createEdge(v1,v2);
  H* e1Itr = 0;
  H* e2Itr = 0;

  //locate the start edge originated from v1 and v2;
  for(int i = 0; i < f->getEdgeCount(); i++) {
    H* tmp = f->getEdge(i);

    if(tmp->getOrgVertex() == v1) {
      e1Itr = tmp;
    }
    if(tmp->getOrgVertex() == v2) {
      e2Itr = tmp;
    }
  }

  std::vector<H*> vec1;

  //store the edges in the half face

  H* e3Itr = e1Itr;

  do {
    vec1.push_back(e3Itr);
    e3Itr->setFace(f);
    e3Itr = e3Itr->getNext();
  } while (e3Itr != e2Itr);

  vec1.push_back(eItr->getOppEdge());
  eItr->getOppEdge()->setFace(f);

  //store the edges in the other half face
  std::vector<H*> vec2;
  H* e4Itr = e2Itr;
  do {
    vec2.push_back(e4Itr);
    e4Itr = e4Itr->getNext();
  } while(e4Itr != e1Itr);

  vec2.push_back(eItr);

  //reuse f

  f->setEdges(vec1);

  // fixme: this looks suspicious, don't we need to store the
  // face address somewhere?
  createFace(vec2);
  
  return eItr;
}




 
Mesh::V* Mesh::createVertex(C& x, C& y)
{
  V* v = _mof.newVertex();
  v->setXY(x,y);
  _vertices.push_back(v);
  v->setVecIndex(_vertices.size()-1);
  return v;
}

 
Mesh::H* Mesh::createEdge(V* org, V* dest)
{
  //construct symetrical edges and connect them
  //adding the halfedges to the vertex edges lists
  H* e1 = _mof.newHalfEdge();
  e1->set(org, dest);
  _halfedges.push_back(e1);
  e1->setVecIndex(_halfedges.size()-1);

  H* e2 = _mof.newHalfEdge();
  e2->set(dest, org);
  _halfedges.push_back(e2);
  e2->setVecIndex(_halfedges.size()-1);

  //link the vertex to one of the outgoing edges
  //Note:: we need to find a new incident outedge with operators flip() and spliceFace() which trigger removeEdge();
  org->setOutEdge(e1);
  dest->setOutEdge(e2);

  //initialize e1 pointers
  e1->setNext(e1);
  e1->setPrior(e1);
  e1->setOppEdge(e2);
   
  e2->setNext(e2);
  e2->setPrior(e2);

  e2->setOppEdge(e1);
  
  return e1;
}

 
Mesh::F* Mesh::createFace(H* e1Itr,
                          H* e2Itr,
                          H* e3Itr)
{
  F* f = _mof.newFace();
  _faces.push_back(f);
  f->setVecIndex(_faces.size()-1);

  //store the edges in ccw order
  f->setEdges(e1Itr, e2Itr, e3Itr);

  e1Itr->setFace(f);
  e2Itr->setFace(f);
  e3Itr->setFace(f);


  return f;
}

 
Mesh::F* Mesh::createFace(std::vector<H*>& edges)
{
  F* f = _mof.newFace();

  f->setEdges(edges);
  
  _faces.push_back(f);
  f->setVecIndex(_faces.size()-1);
     
  FORALL (edges, it) {
    (*it)->setFace(f);
  }

  return f;
}


 
void Mesh::removeFace(F* f)
{
  //for(std::list<std::pair<int,int>>::iterator itrI= f->attachedSectors.begin(); itrI!= f->attachedSectors.end(); ++itrI)
  //{
  //       _triangulator->sectors[itrI->first][itrI->second] = -1;
  //    

  //}

  if (_faces.empty()) ERR("can't delete face");
  
  _faces[f->getVecIndex()] = _faces.back();
  _faces[f->getVecIndex()]->setVecIndex(f->getVecIndex());
  _faces.pop_back();
  delete f;
  //_faces.erase(_faces.begin()+f->vecIndex);
}

 
void Mesh::removeEdge(H* e)
{
  if (_halfedges.empty()) ERR("can't delete halfedge");

  //find a new out edge for org and dest of e
  e->getOrgVertex()->setOutEdge(e->getOppEdge()->getNext());
  e->getDestVertex()->setOutEdge(e->getNext());

  _halfedges[e->getOppEdge()->getVecIndex()] = _halfedges.back();
  _halfedges[e->getOppEdge()->getVecIndex()]->setVecIndex(e->getOppEdge()->getVecIndex());
  _halfedges.pop_back();
  delete e->getOppEdge();

  if (_halfedges.empty()) ERR("can't delete halfedge");
  
  _halfedges[e->getVecIndex()] = _halfedges.back();
  _halfedges[e->getVecIndex()]->setVecIndex(e->getVecIndex());
  _halfedges.pop_back();
  delete e;

  //_halfedges.erase(_halfedges.begin()+e->getOppEdge()->vecIndex);
  //_halfedges.erase(_halfedges.begin()+e->vecIndex);
}



 
void Mesh::removeVertex(V* v)
{
  if (_vertices.empty()) ERR("can't delete vertex");
  
  _vertices[v->getVecIndex()] = _vertices.back();
  _vertices[v->getVecIndex()]->setVecIndex(v->getVecIndex());
  _vertices.pop_back();
  delete v;
  //_vertices.erase(_vertices.begin()+v->vecIndex);
}

 
void Mesh::removeConstraint(H* e,int id)
{
  e->removeID(id);
  e->getOppEdge()->removeID(id);
}


 
void Mesh::clear()
{
  
}



 
void Mesh::addConstraint(H* e,int id, V* left, V* right)
{
  if (!e->getLeftEnd().first) {
    std::pair<bool,V*> lf(true,left);
    e->setLeftEnd(lf);
  }
  
  if (!e->getRightEnd().first) {
    std::pair<bool,V*> rt(true,right);
    e->setRightEnd(rt);
  }

  if (!e->getOppEdge()->getLeftEnd().first) {
    std::pair<bool,V*> rt(true,right);
    e->getOppEdge()->setLeftEnd(rt);
  }

  if (!e->getOppEdge()->getRightEnd().first) {
    std::pair<bool,V*> lf(true,left);
    e->getOppEdge()->setRightEnd(lf);
  }

  e->setConstrained(id);
  e->getOppEdge()->setConstrained(id);

}

 
void Mesh::addConstraint(H* e,std::list<int>& ids, V* left, V* right)
{

    if(!e->getLeftEnd().first) 
    {
        std::pair<bool,V*> lf(true,left);
       e->setLeftEnd(lf);
        
    }
    if(!e->getRightEnd().first)
    {
        std::pair<bool,V*> rt(true,right);
        e->setRightEnd(rt);
    }

    if(!e->getOppEdge()->getLeftEnd().first) 
    {
        std::pair<bool,V*> rt(true,right);
       e->getOppEdge()->setLeftEnd(rt);
        
    }
    if(!e->getOppEdge()->getRightEnd().first)
    {
        std::pair<bool,V*> lf(true,left);
        e->getOppEdge()->setRightEnd(lf);
    }
    
     e->setConstrained(ids);
     e->getOppEdge()->setConstrained(ids);
         
}


 
Mesh::V* Mesh::getOpposedVertex(F* f1, F* f2)
{
  H* e1Itr = f1->getEdge(0);
  H* e2Itr = e1Itr;
  V* res;

  do {
    if (e1Itr->getNext()->getOppEdge()->getFace() == f2) {
      res = e1Itr->getOrgVertex();
      break;
    }

    e1Itr = e1Itr->getNext();
  } while(e1Itr != e2Itr);

  return res;
}

 
Mesh::F* Mesh::getOpposedFace(F* f, V* v, H** shareEdge)
{
  H* e1Itr = f->getEdge(0); 
  H* e2Itr = e1Itr;

  do {
    V* v1Itr = e1Itr->getOrgVertex();

    if (v1Itr == v) {
      break;
    }

    e1Itr = e1Itr->getNext();
  } while (e1Itr != e2Itr);
     
  *shareEdge = e1Itr->getNext();
  return (*shareEdge)->getOppEdge()->getFace();
     
}

//---------------------------------------------------------------

 
bool Mesh::inTriangle(V* v1,V* v2,V* v3,V* v)
{
  C x1, y1, x2, y2, x3, y3,x,y;
  
  x1 = v1->getX();
  y1 = v1->getY();
  x2 = v2->getX();
  y2 = v2->getY();
  x3 = v3->getX();
  y3 = v3->getY();
  x = v->getX();
  y = v->getY();

  return
    Geometry::ccw(x,y,x1,y1,x2,y2) >= 0 &&
    Geometry::ccw(x,y,x2,y2,x3,y3) >= 0 &&
    Geometry::ccw(x,y,x3,y3,x1,y1) >=0;
}



// fixme: is this actually used?  H* e, en, enn doesn't work!!!

bool Mesh::inTriangle(F* f, C& x, C& y, F** resultFace)
{
  C x1, y1, x2, y2, x3, y3;
  H* e = f->getEdge(0);
  H* en = e->getNext();
  H* enn = en->getNext();

  V* v1 = e->getOrgVertex(); 
  V* v2 = en->getOrgVertex(); 
  V* v3 = enn->getOrgVertex();

  x1 = v1->getX();
  y1 = v1->getY();
  x2 = v2->getX();
  y2 = v2->getY();
  x3 = v3->getX();
  y3 = v3->getY();


  if (Geometry::ccw(x,y,x1,y1,x2,y2) >= 0 &&
      Geometry::ccw(x,y,x2,y2,x3,y3) >= 0 &&
      Geometry::ccw(x,y,x3,y3,x1,y1) >=0) {
    *resultFace = f;
    return true;
  }

  return false;
}



 
bool Mesh::onVertex(F* f, C& x, C& y,V** resultVertex)
{
  C x1, y1, x2, y2, x3, y3;
        
  H* e = f->getEdge(0);
  H* en = e->getNext();
  H* enn = en->getNext();

  V* v1=e->getOrgVertex(); 
  V* v2=en->getOrgVertex(); 
  V* v3=enn->getOrgVertex();

  x1 = v1->getX();
  y1 = v1->getY();
  x2 = v2->getX();
  y2 = v2->getY();
  x3 = v3->getX();
  y3 = v3->getY();

  if (Geometry::overlap(x,y,x1,y1)) { *resultVertex = v1; return true; }
  if (Geometry::overlap(x,y,x2,y2)) { *resultVertex = v2; return true; }
  if (Geometry::overlap(x,y,x3,y3)) { *resultVertex = v3; return true; }
     
  return false;
}

 
bool Mesh::onEdge(F* f, C& x, C& y, H** resultEdge)
{
  C x1, y1, x2, y2, x3, y3;
     
  H* e = f->getEdge(0);
  H* en = e->getNext();
  H* enn = en->getNext();

  V* v1=e->getOrgVertex(); 
  V* v2=en->getOrgVertex(); 
  V* v3=enn->getOrgVertex();

  x1 = v1->getX();
  y1 = v1->getY();
  x2 = v2->getX();
  y2 = v2->getY();
  x3 = v3->getX();
  y3 = v3->getY();

  if (Geometry::onSegment(x1,y1,x2,y2,x,y)) { *resultEdge = e;   return true; }
  if (Geometry::onSegment(x2,y2,x3,y3,x,y)) { *resultEdge = en;  return true; }
  if (Geometry::onSegment(x3,y3,x1,y1,x,y)) { *resultEdge = enn; return true; }
     
  return false;
}

 
bool Mesh::isConnected(V* v1Itr, V* v2Itr)
{
  H* e1 = v1Itr->getOutEdge();
  H* e2 = e1;
  
  do {
    V* v3Itr = e1->getDestVertex();
    if(v3Itr == v2Itr) {
      return true;
    }
                                       
    e1 = e1->getOppEdge()->getNext();
  } while (e1 != e2);

  return false;
}

 
bool Mesh::isConnected(V* v1Itr, V* v2Itr, H** eItr)
{
  H* e1 = v1Itr->getOutEdge();
  H* e2 = e1;
     
  do {
    V* v3Itr = e1->getDestVertex();
    if(v3Itr == v2Itr) {
      *eItr = e1;
      return true;
    }
    
    e1 = e1->getOppEdge()->getNext();
  } while(e1 != e2);

  return false;
}



//It is used to find constrained edges
//Note:: the coords should be (int, int, limRat) 
 
bool Mesh::vertexAboveSegment(V* v1,V* v2, V* v)
{
  /*C x1,y1,x2,y2,x,y;

    x1 = v1->getX();
    y1 = v1->getY();
    x2 = v2->getX();
    y2 = v2->getY();
    x  = v->getX();
    y  = v->getY();*/

  //return Geometry::ccw_old(v1->getX(),v1->getY(),v2->getX(),v2->getY(),v->getX(),v->getY()) > 0;
  return Geometry::ccw(v1->getX(), v1->getY(),
                       v2->getX(), v2->getY(),
                       v->getX(),  v->getY()) > 0;
}

//It is used to find constrained edges
//Note:: the coords should be (int, int, limRat) 
 
bool Mesh::vertexBelowSegment(V* v1, V* v2, V* v)
{
  /*C x1,y1,x2,y2,x,y;

    x1 = v1->getX();
    y1 = v1->getY();
    x2 = v2->getX();
    y2 = v2->getY();
    x  = v->getX();
    y  = v->getY();*/

  // return Geometry::ccw_old(v1->getX(),v1->getY(),v2->getX(),v2->getY(),v->getX(),v->getY()) < 0;
  return Geometry::ccw(v1->getX(), v1->getY(),
                       v2->getX(), v2->getY(),
                       v->getX(),  v->getY()) < 0;
}

//It is used to find unconstrained  edges
//Note:: the coords can be (int, limRat, limRat) so exact compuation is needed
 
bool Mesh::vertexAboveSegment2(V* v1, V* v2, V* v)
{
  /*C x1,y1,x2,y2,x,y;

    x1 = v1->getX();
    y1 = v1->getY();
    x2 = v2->getX();
    y2 = v2->getY();
    x  = v->getX();
    y  = v->getY();*/

  return Geometry::ccw(v1->getX(), v1->getY(),
                       v2->getX(), v2->getY(),
                       v->getX(),  v->getY()) > 0;
}

 
bool Mesh::isTriangle(H* e)
{
  return e->getNext()->getNext()->getNext() == e;
}


//============================ marking =====================================

 
void Mesh::normalizeMark () // private method
{
  _curmark=1;
   
  FORALL (_halfedges, it) {
    (*it)->setMark(0);
  }
}

 
void Mesh::beginMarking ()
{
  if (_marking) {
    printf("Mesh::begin_marking() not allowed as marking is already in use!");
    exit(0);
  }

  _marking = true;

  if (_curmark == _mark_max_value) { 
    normalizeMark();
  } else
    _curmark++;
}

 
void Mesh::endMarking ()
{
  _marking = false;
}

 
bool Mesh::marked(H* e) 
{
  if (!_marking) { printf("Mesh::marked(e): marking is not active!\n"); exit(0); }
  return e->getMark() == _curmark;
}

 
void Mesh::mark(H* e) 
{ 
  if (!_marking) { printf("Mesh::mark(e): marking is not active!\n"); exit(0); }
  e->setMark(_curmark);
}

 
void Mesh::unmark(H* e) 
{ 
  if (!_marking) { printf("Mesh::unmark(e): marking is not active!\n"); exit(0); }
  e->setMark(_curmark-1);
}


 
bool Mesh::marked(F* f) 
{
  if (!_marking) { printf("Mesh::marked(f): marking is not active!\n"); exit(0); }
  return f->getMark() == _curmark;
}

 
void Mesh::mark(F* f) 
{ 
  if (!_marking) { printf("Mesh::mark(f): marking is not active!\n"); exit(0); }
  f->setMark(_curmark);
}

 
void Mesh::unmark(F* f) 
{ 
  if (!_marking) { printf("Mesh::unmark(f): marking is not active!\n"); exit(0); }
  f->setMark(_curmark-1);
}

