#ifndef MESH_H
#define MESH_H

#include <cassert>
#include <algorithm>
#include <limits.h>
#include <boost/timer.hpp>
#include <list>
#include <vector>
#include "Rational.h"
#include "Geometry.h"

#include "Vertex.h" // for coord_t
class HalfEdge;
class Face;

//class mpq_class;
//
//
//template <class T=long long>
//class Rational;
//
//

struct MeshObjFactory
{
  virtual Vertex*   newVertex()   const = 0;
  virtual HalfEdge* newHalfEdge() const = 0;
  virtual Face*     newFace()     const = 0;
  virtual ~MeshObjFactory() {}
};


class Mesh 
{

public:

  //typedefs

  //typedef mpq_class coord_t;
  typedef Vertex     V;
  typedef HalfEdge   H;
  typedef Face       F;
  typedef V::coord_t C;
  
  // constructors and destructor	

  Mesh(MeshObjFactory &mof);
  Mesh(MeshObjFactory &mof, int xmax, int ymax);
  virtual ~Mesh();

  //static query functions 
    
  static bool isTriangle(H* e);
  static bool inTriangle(V* v1, V* v2, V* v3, V* v);
  static bool inTriangle(F* f, C& x, C& y, F** resultFace);
  static bool onVertex(F* f, C& x, C& y, V** resultVertex);
  static bool onEdge(F* f, C& x, C& y, H** resultEdge);
  static bool isConnected(V* v1Itr, V* v2Itr, H** eItr);
  static bool isConnected(V* v1Itr, V* v2Itr);
    
  static bool vertexAboveSegment(V* v1,V* v2, V* v);
  static bool vertexAboveSegment2(V* v1,V* v2, V* v);
  static bool vertexBelowSegment(V* v1,V* v2, V* v);
      
  // operators
  void init(int _xmax, int _ymax);		
  V* createVertex(C& x, C& y);
  H* createEdge(V* v1, V* v2);
  F* createFace(H* e1, H* e2, H* e3);
  F* createFace(std::vector<H*>& edges);
   
  void spliceFaces(H* e1, H* e2);
  F* joinFaces(F* f1, F* f2, H* e);
  H* splitFace(F* f, V* v1, V* v2);
  void addConstraint(H* e,int id, V* left, V* right);
  void addConstraint(H* e,std::list<int>& id, V* left, V* right);
  void removeConstraint(H* e,int id);
  F* getOpposedFace(F* f , V* v, H** shareEdge);
  V* getOpposedVertex(F* f1, F* f2);

  void removeFace(F* f);
  void removeEdge(H* e);
  void removeVertex(V* v);
  void clear();

  // Marking =================================================================

  //reset indexing when max mark index is reached.
  void normalizeMark();

  /*! Allows the user to safely mark elements with mark(), marked() and unmark()
    methods. Marking elements is necessary for many algorithms.
    the HalfEdge mantains a current marker value, so that each time begin_marking()
    is called, the marker value is incremented, guaranteing that old marked
    elements have a different mark value.
    When the max marking value is reached, a global normalization is done.
    Attention: end_marking() must be called when marking operations are finished! */
  void beginMarking ();

  /*! Must be called after a begin_marking() call, when the user finishes using the 
    marking methods. Note: There is no problem if end_marking() is called without
    a previously call to begin_marking() */
  void endMarking ();

  /*! See if an HalfEdge is marked (begin_marking() should be called first) */
  bool marked(H* e);

  /*! Unmark an HalfEdge (begin_marking() should be called first) */
  void unmark(H* e);

  /*! Mark an HalfEdge (begin_marking() should be called first) */
  void mark(H* e);


  /*! See if an HalfEdge is marked (begin_marking() should be called first) */
  bool marked(F* f);

  /*! Unmark an HalfEdge (begin_marking() should be called first) */
  void unmark(F* f);

  /*! Mark an HalfEdge (begin_marking() should be called first) */
  void mark(F* f);

  // iterators and element accessors

  const std::vector<V*>& getVertices() const { return _vertices; };
  const std::vector<H*>& getHalfedges() const { return _halfedges; };
  const std::vector<F*>& getFaces()    const { return _faces; }; 
  
   
public: // accessor

  int getXMax() const { return _xmax; }
  int getYMax() const { return _ymax; }  
  int getVertexCount()   const { return _vertices.size();  }
  int getHalfEdgeCount() const { return _halfedges.size(); }
  int getFaceCount()     const { return _faces.size();     }


protected:

  // data
	
  MeshObjFactory& _mof;
  int _xmax, _ymax; // mins = 0

  bool _marking;
  int _curmark;
  int _mark_max_value;

  std::vector<F*> _faces;
  std::vector<H*> _halfedges;
  std::vector<V*> _vertices;

  F* _boundaryFace;
};


#endif
