//Abstraction.h
//From Douglas Demyen's Abstraction implementation for pathfinding 
//Modified to run under dcdtlite

//DJD: definition of Abstraction class {

#ifndef ABSTRACTION_H
#define ABSTRACTION_H

#include <cfloat>
#include "PathFace.h"

//defines an invalid entry, and "infinity", for our purposes

class PathFace;
//class which contains information about our topological abstraction

class Abstraction
{

  static const int INVALID = -1;
 
protected:
  //list of adjacent nodes
  //if this node is degree-1 in a tree, all will be 0
  //if it is degree-1 otherwise, one will be the root of the tree, the others will be 0
  //if it is degree-2 in a ring, all will be 0
  //if it is degree-2 otherwise, two will be degree-3 nodes on the ends of the corridor, the other will be 0
  //if it is degree-3, all will be degree-3 nodes on the end of the corridors
  array<PathFace*,3> m_CAdjacent;

  //sum of interior angles of triangles between this one and the corresponding adjacent one
  //INVALID if the corresponding adjacent node is 0
  array<float,3> m_dAngle;

  //width between the corresponding edge and the one counterclockwise to it, through the triangle
  array<float, 3> m_dWidth;

  //smallest width through the triangles between this one and the corresponding adjacent one
  //INVALID if the corresponding adjacent node is 0
  array<float,3> m_dChoke;

  //degree of this node
  int m_nDegree;

  //connected component to which this triangle belongs
  int m_nComponent;

  //makes sure the index is in the range [0, 2]
  bool checkBounds(int n)
  {
    return (n >= 0 && n < 3);
  }

public:
  
  //creates the abstraction with given degree
  Abstraction(int nDegree)
  {
    //starts all adjacents as 0 and other information as INVALID
    for (int i = 0; i < 3; i++) {
      m_CAdjacent[i] = 0;
      m_dAngle[i] = INVALID;
      m_dWidth[i] = INVALID;
      m_dChoke[i] = INVALID;
    }
    m_nDegree = nDegree;
    m_nComponent = INVALID;
  }

  //default destructor
  ~Abstraction() { }

  //accessor for the degree of the node
  int degree() { return m_nDegree; }

  //mutator for the degree of the node
  void degree(int nDegree) { m_nDegree = nDegree; }

  //accessor for the connected component of the node
  int component() { return m_nComponent; }

  //mutator for the connected component of the node
  void component(int nComponent) { m_nComponent = nComponent; }

  //accessor for the adjacent node with index given
  PathFace* adjacent(int n) 
  {
    return (checkBounds(n)) ? m_CAdjacent[n] : 0;
  }

  //mutator for the adjacent node with index given
  void adjacent(int n, PathFace* cAdjacent)
  {
    if (checkBounds(n)) {
      m_CAdjacent[n] = cAdjacent;
    }
  }

  //accessor for the angle with index given
  float angle(int n)
  {
    return (checkBounds(n)) ? m_dAngle[n] : INVALID;
  }

  //mutator for the angle with index given
  void angle(int n, float dAngle)
  {
    if (checkBounds(n)) {
      m_dAngle[n] = dAngle;
    }
  }

  //accessor for the width with index given
  float width(int n)
  {
    return (checkBounds(n)) ? m_dWidth[n] : INVALID;
  }

  //mutator for the width with index given
  void width(int n, float dWidth)
  {
    if (checkBounds(n)) {
      m_dWidth[n] = dWidth;
    }
  }

  //accessor for the choke point width with index given
  float choke(int n)
  {
    return (checkBounds(n)) ? m_dChoke[n] : INVALID;
  }

  //mutator for the choke point width with index given
  void choke(int n, float dChoke)
  {
    if (checkBounds(n)) {
      m_dChoke[n] = dChoke;
    }
  }
};

#endif

//definition of Abstraction class }
