#include "DCDTPath.h"

#include "HalfEdge.h"
#include "Abstraction.h"

using namespace std;


const float DCDTPath::PI = 3.1415926535897932384626433832795f; 
//function for comparing SearchNodes
 
struct check : public std::binary_function <SearchNode *, SearchNode *, bool> 
{
  bool operator()(SearchNode *&sn1, SearchNode *&sn2) const {
    return ((sn1->f() > sn2->f()) || ((sn1->f() == sn2->f()) && (sn1->h() > sn2->h())));;
  }
};


// cast Face to PathFace pointer
//#define PF(p) (assert(p == 0 || dynamic_cast<PF*>(p) != 0), static_cast<PF*>(p))
#define PF(p) ( static_cast<PF*>(p))

float DCDTPath::getShortestPath(FPolygon &path,
                                std::vector<H*>& channel,
                                float x1, float y1, float x2, float y2, float r)
{
  path.clear();

  //if there are no edges in the channel, the path is a straight line
  if (channel.size() < 1) {
    FPoint start(x1,y1);
    FPoint goal(x2,y2);
    path.push_back(start);
    path.push_back(goal);
    return length(x1, y1, x2, y2);
  }
     
  //otherwise, initialize a funnel deque
  //with the starting point
  FunnelDeque fd(x1, y1, r);

  //get the first edge in the channel
  H* s = channel[0];

  //get the reference whose sym edge
  //belongs to the next triangle in the channel
  for (int i = 0; i < 3; i++) {
    if (((channel.size() > 1) && (s->getOppEdge()->getFace() == channel[1]->getFace()))
        || (inTriangle(s->getOppEdge()->getFace(), x2, y2))) {
      break;
    }
    s = s->getNext();
  }
     
  //Kai: needs double check on correctness of left and right vertex
  //get the point on the right side of that edge
  FPoint p(s->getOrgVertex()->getX().getD(), s->getOrgVertex()->getY().getD());

  //insert it into the deque
  fd.add(p, FunnelDeque::CT_LeftTangent, path);
     
  V* right = s->getOrgVertex();

  //get the point on the left side of that edge
  s = s->getNext();
  V* left = s->getOrgVertex();
  p.set(s->getOrgVertex()->getX().getD(),s->getOrgVertex()->getY().getD());

  //insert it into the deque
  fd.add(p, FunnelDeque::CT_RightTangent, path);

  //go through the other edges in the channel
  for (unsigned int i = 1; i < channel.size() - 1; i++) {

    s = channel[i];

    //get the reference whose sym edge
    //belongs to the next triangle in the channel
    for (int j = 0; j < 3; j++) {
      if (s->getOppEdge()->getFace() == channel[i + 1]->getFace()) {
        break;
      }
      s = s->getNext();
    }
    //if that edge shares a vertex with the right side
    //of the funnel, the opposite point should be
    //added to the left side of the funnel
    if (s->getOrgVertex() == right) {

      p.set(s->getNext()->getOrgVertex()->getX().getD(), s->getNext()->getOrgVertex()->getY().getD());
      fd.add(p, FunnelDeque::CT_RightTangent, path);
      left = s->getNext()->getOrgVertex();

    } else {

      //otherwise, add it to the right side of the funnel
      p.set(s->getOrgVertex()->getX().getD(), s->getOrgVertex()->getY().getD());
      fd.add(p, FunnelDeque::CT_LeftTangent, path);
      right = s->getOrgVertex();
    }
  }

  //go to the last edge in the channel
  s = channel[channel.size() - 1];

  //find the reference across which is
  //the triangle with the goal in it
  for (int j = 0; j < 3; j++) {
    if (inTriangle(s->getOppEdge()->getFace(), x2, y2)) {
      break;
    }
    s = s->getNext();
  }
  //like above, determines whether to add
  //the point to the left or right side of the funnel
  if (s->getOrgVertex() == right) {

    p.set(s->getNext()->getOrgVertex()->getX().getD(), s->getNext()->getOrgVertex()->getY().getD());
    fd.add(p, FunnelDeque::CT_RightTangent, path);

  } else {

    p.set(s->getOrgVertex()->getX().getD(),s->getOrgVertex()->getY().getD());
    fd.add(p, FunnelDeque::CT_LeftTangent, path);

  }

  //finally, add the goal point to the funnel
  p.set(x2, y2);
  fd.add(p, FunnelDeque::CT_Point, path);

  //calculate the length of the path sand return it
  return fd.length(path);
}


//returns the current path
std::vector<Mesh::H*> DCDTPath::constructBaseChannel(SearchNode *goal)
{
  std::vector<H*> channel;
     
  //if there is only one node, return an empty path
  if (goal->back() == 0)
    return channel;
  
  //otherwise, move back in the chain
  SearchNode *current = goal->back();
  //until we are at the second-to-first node

  while (current->back() != 0) {

    //go through the edges of the triangle associated with this node
    H* s = current->triangle()->getEdge(2);

    for (int i = 0; i < 3; i++) {

      //if the previous triangle is across this edge,
      if (s->getOppEdge()->getFace() == current->back()->triangle()) {

        //add this edge/face to the channel
        channel.push_back(s);

        //and move to the next node
        break;
      }

      //go to the next edge in the triangle
      s = s->getNext();
    }

    //move to the previous node in the chain
    current = current->back();
  }

  //add the last triangle to the channel
  channel.push_back(current->triangle()->getEdge(2));

  //the channel was constructed backward, so flip it, then return it
  std::reverse(channel.begin(), channel.end());
  return channel;
}

//utility functions for pathfinding
 
float DCDTPath::pointLineDistance(float x, float y, float x1, float y1, float x2, float y2)
{
  if (x1 == x2) {
    return fabs(x - x1);
  }
  float rise = y2 - y1;
  float run = x2 - x1;
  float intercept = y1 - (rise / run) * x1;
  float a = rise;
  float b = -run;
  float c = run * intercept;
  return (fabs(a * x + b * y + c) / sqrt(a * a + b * b));
}

 
float DCDTPath::orientation(float x1, float y1, float x2, float y2, float x3, float y3)
{
  return ((x1 * y2) + (x2 * y3) + (x3 * y1) - (x1 * y3) - (x2 * y1) - (x3 * y2));
}

bool DCDTPath::isAccute(float theta)
{
  return (fabs(theta) < (M_PI * 0.5f));
}
 
bool DCDTPath::inTriangle(F* face, float x, float y)
{
         
         if(face->isBoundary())
         {
                  return false;
         }
  //get the triangle's vertices
  float x1, y1, x2, y2, x3, y3;
  triangleVertices(face->getEdge(0), x1, y1, x2, y2, x3, y3);

  //check that the point is counterclockwise of all edges
  return ((orientation(x1, y1, x2, y2, x, y) >= 0)
          && (orientation(x2, y2, x3, y3, x, y) >= 0)
          && (orientation(x3, y3, x1, y1, x, y) >= 0));
}

float DCDTPath::angleBetween(float x1, float y1, float x2, float y2, float x3, float y3)
{
  float angle1 = (float)atan2(y1 - y2, x1 - x2);
  float angle2 = (float)atan2(y3 - y2, x3 - x2);
  float diff = angle1 - angle2;
  diff += (diff <= -M_PI) ? 2.0f * M_PI : (diff > M_PI) ? -2.0f * M_PI : 0.0f;
  return diff;
}

 
bool DCDTPath::blocked(H* e)
{
  return e->isConstrained() || e->isBoundary();
}

void DCDTPath::triangleVertices(H* e, float &x1, float &y1, float &x2, float &y2, float &x3, float &y3)
{
  x1 = e->getOrgVertex()->getX().getD();
  y1 = e->getOrgVertex()->getY().getD();
  e  = e->getNext();
  x2 = e->getOrgVertex()->getX().getD();
  y2 = e->getOrgVertex()->getY().getD();
  e  = e->getNext();
  x3 = e->getOrgVertex()->getX().getD();
  y3 = e->getOrgVertex()->getY().getD();
}

 
float DCDTPath::length(float x1, float y1, float x2, float y2)
{
  float xDiff = x1 - x2;
  float yDiff = y1 - y2;
  return (sqrt(xDiff * xDiff + yDiff * yDiff));
}

//returns the closest point on the segment given, at least r in from the endpoints, to the point given
 
Point<float> DCDTPath::closestPointOn(float x1, float y1, float x2, float y2, float x, float y, float r)
{
  Point<float> point;
  float distance;

  //checks if the angles along this edge and to the point are accute
  bool accute1 = isAccute(angleBetween(x1, y1, x2, y2, x, y));
  bool accute2 = isAccute(angleBetween(x2, y2, x1, y1, x, y));

  if (accute1 && accute2) {

    //if both are, get closest point on this line
    distance = pointLineDistance(x, y, x1, y1, x2, y2);
    float theta = atan2(y2 - y1, x2 - x1);
    theta += (orientation(x, y, x1, y1, x2, y2) > 0.0f) ? M_PI / -2.0f : M_PI / 2.0f;
    theta += (theta <= -M_PI) ? 2.0f * M_PI : (theta > M_PI) ? -2.0f * M_PI : 0.0f;
    point.set(x + cos(theta) * distance, y + sin(theta) * distance);

    //if it's < r from either end, move it out to distance r
    if (length(point.getX(), point.getY(), x1, y1) < r) {
      accute2 = false;
    } else if (length(point.getX(), point.getY(), x2, y2) < r) {
      accute1 = false;
    }
  }
  if (!accute1 && accute2) {

    //get point distance r from (x2, y2) in the direction of (x1, y1)
    float theta = atan2(y1 - y2, x1 - x2);
    point.set(x2 + cos(theta) * r, y2 + sin(theta) * r);

  } else if (accute1 && !accute2) {

    //get point distance r from (x1, y1) in the direction of (x2, y2)
    float theta = atan2(y2 - y1, x2 - x1);
    point.set(x1 + cos(theta) * r, y1 + sin(theta) * r);
  }

  return point;
}


 
Point<float> DCDTPath::closestPointTo(F* face, float x, float y, float r)
{
  // if the point is in the triangle
  if (inTriangle(face, x, y)) {

    //the closest point is the point itself
    return Point<float>(x, y);
  }

  Point<float> closestPoint, point;
  float closestDistance = INFINITY_LEN, distance;

  //get the edges of the triangle and go through them
  H* s = face->getEdge(2);

  for (int i = 0; i < 3; i++) {

    //gets the vertices of the triangle
    float x1, y1, x2, y2, x3, y3;
    triangleVertices(s, x1, y1, x2, y2, x3, y3);

    //doesn't consider blocked edges
    if (blocked(s)) {
      s = s->getNext();
      continue;
    }

    //checks if the angles along this edge and to the point are accute
    bool accute1 = isAccute(angleBetween(x1, y1, x2, y2, x, y));
    bool accute2 = isAccute(angleBetween(x2, y2, x1, y1, x, y));

    if (accute1 && accute2) {

      //if both are, get closest point on this line
      distance = pointLineDistance(x, y, x1, y1, x2, y2);
      float theta = atan2(y2 - y1, x2 - x1);
      theta += (orientation(x, y, x1, y1, x2, y2) > 0.0f) ? M_PI / -2.0f : M_PI / 2.0f;
      theta += (theta <= -M_PI) ? 2.0f * M_PI : (theta > M_PI) ? -2.0f * M_PI : 0.0f;
      point.setX(x + cos(theta) * distance);
      point.setY(y + sin(theta) * distance);

      //if it's < r from either end, move it out to distance r
      if (length(point.getX(), point.getY(), x1, y1) < r) {
        accute2 = false;
      } else if (length(point.getX(), point.getY(), x2, y2) < r) {
        accute1 = false;
      }
    }
    if (!accute1 && accute2) {

      //get point distance r from (x2, y2) in the direction of (x1, y1)
      float theta = atan2(y1 - y2, x1 - x2);
      point.setX(x2 + cos(theta) * r);
      point.setY(y2 + sin(theta) * r);

    } else if (accute1 && !accute2) {

      //get point distance r from (x1, y1) in the direction of (x2, y2)
      float theta = atan2(y2 - y1, x2 - x1);
      point.setX(x1 + cos(theta) * r);
      point.setY(y1 + sin(theta) * r);
    }

    //measure the distance between this point and the one given
    distance = length(point.getX(), point.getY(), x, y);

    //if this point is closer than others,
    if (distance < closestDistance) {

      //set it as the closest point
      closestPoint = point;
      closestDistance = distance;
    }

    //move to the next edge
    s = s->getNext();
  }

  //return the closest point found
  return closestPoint;
}

 
void DCDTPath::calculateWidths(PF* face)
{
  //if the triangle hasn't been abstracted yet, return

  if (!face->link)
    return;

  //goes through the vertices of the triangle

  assert(face->getEdgeCount() == 3);
  
  for (int i = 0; i < 3; i++) {
    HalfEdge *s = face->getEdge(i);
    //calculate and set the width between those edges
    face->link->width(i, triangleWidth(s->getNext()));
  }
}

 
int DCDTPath::degree(PF* face)
{
  if (!face->link)
    return UNABSTRACTED;

  return face->link->degree();
}

 
float DCDTPath::triangleWidth(HalfEdge *s)
{
  //get the coordinates of the triangle
  float x, y, x1, y1, x2, y2;

  triangleVertices(s, x, y, x1, y1, x2, y2);

  //if either base angle isn't accute, the width of that
  //triangle is the length of the shorter edge

  if (!isAccute(angleBetween(x, y, x1, y1, x2, y2))) {

    return length(x, y, x1, y1);

  } else if (!isAccute(angleBetween(x, y, x2, y2, x1, y1))) {

    return length(x, y, x2, y2);

  } else {

    //otherwise, calculate the upper bound on the width of the triangle
    float currentWidth = min(length(x, y, x1, y1), length(x, y, x2, y2));

    //and calculate the actual value of the width
    return searchWidth(x, y, s->getNext()->getOppEdge(), currentWidth);
  }
}

 
float DCDTPath::searchWidth(float x, float y, HalfEdge *s, float currentWidth)
{
  //get the coordinates of the triangle
  float x1, y1, x2, y2, x3, y3;
  
  triangleVertices(s, x1, y1, x2, y2, x3, y3);

  //checks if the entrance edge should be considered
  if (!consider(x, y, x1, y1, x2, y2)) {

    //if not, returns the width unmodified
    return currentWidth;
  }

  //calculates the distance between this edge and the "origin" point
  float distance = pointLineDistance(x, y, x1, y1, x2, y2);

  //if it's farther than the upper bound
  if (distance >= currentWidth) {
    return currentWidth;

  } else if (blocked(s->getOppEdge())) {

    //if the edge is constrained,
    //return this distance as the new upper bound
    return distance;
  }

  //gets the distance to the vertex opposite the entrance edge
  distance = length(x, y, x3, y3);

  //if this distance is less than the current upper bound,
  if (distance <= currentWidth) {

    //return this distance as the new upper bound
    return distance;
  }

  //otherwise, searches across the other two edges for bounds
  currentWidth = searchWidth(x, y, s->getNext()->getOppEdge(), currentWidth);
  return searchWidth(x, y, s->getNext()->getNext()->getOppEdge(), currentWidth);
}


//determines if an edge should be considered when calculating the width
//(if the base angles of the triangle formed by this edge joined with the "origin" point are both accute)

bool DCDTPath::consider(float x, float y, float x1, float y1, float x2, float y2)
{
  return ((isAccute(angleBetween(x, y, x1, y1, x2, y2)))
          && (isAccute(angleBetween(x, y, x2, y2, x1, y1))));
}


//pathfinding 
//TA*
 
bool DCDTPath::searchPathBaseFast(float x1, float y1, float x2, float y2, float r)
{
  //initialize the path to empty

  currentPath.clear();
  F* startFace;
  H* startEdge;
  V* startVertex;

  Mesh::C cx1(round(x1));
  Mesh::C cy1(round(y1));
  
  //Kai: need float version of locatePoint for pathfinding.
  //find the triangle containing the start point
  DCDT::LocateResult startResult = _dcdt.locatePoint(0, cx1, cy1,
                                                     &startFace, &startEdge, &startVertex);

  //if it wasn't found, return a path could not be found
  if (startResult == DCDT::NotFound)
    return false;
      
  //create a priority queue for the search nodes
  std::priority_queue<SearchNode*, std::vector<SearchNode*>, check> q;

  //calculate the closest point to the goal in the first triangle
  FPoint p = closestPointTo(startFace, x2, y2, r);

  //use the distance between that and the goal as the heuristic in the first search node
  q.push(new SearchNode(0.0f, length(p.getX(), p.getY(), x2, y2), p, PF(startFace), 0));

  //the shortest path so far and its length and channel
  float shortestPathLength = INFINITY_LEN;
  FPolygon shortestPath;
  std::vector<H*> shortestChannel;

  float startGoalDistance = length(x1, y1, x2, y2);
  
  //continue until all nodes have been exhausted
  while (!q.empty()) {

    //gets the best search node so far
    SearchNode *current = q.top();
    q.pop();

    //if the shortest path through this node is longer than the best path we have found so far,
    //or if we have no unexplored nodes on the queue and we haven't found a path yet
    if ((current->f() >= shortestPathLength)) { // || ((unexplored <= 0) && (shortestPathLength == INFINITY_LEN)))

      //empty the primary queue, close and delete all remaining nodes
      while (!q.empty()) {
        q.top()->close();
        delete q.top();
        q.pop();
      }

      //close and delete the current node
      current->close();
      delete current;

      //exit the loop
      break;
    }
    //if we haven't explored the current node yet
    //		if (!_mesh->marked(current->Triangle()))
    //		{
    //mark it as explored now
    //			_mesh->mark(current->Triangle());
    //we are taking an unexplored node off of the queue
    //			unexplored--;
    //		}
    //if the goal is in the current triangle,

            
    if (inTriangle(current->triangle(), x2, y2)) {

      //construct the channel formed by our search
      std::vector<H*> channel = constructBaseChannel(current);

      //close and delete the current search node
      current->close();
      delete current;

      //check if the endpoints are valid
                  
      //Kai: remove width checks temporarily

      // fixme: what about this check?
      
      if (validEndpoints(channel, x1, y1, x2, y2, r))
      {
        //calculate the path and its length
        FPolygon currentShortestPath;
        float currentPathLength = getShortestPath(currentShortestPath, channel, x1, y1, x2, y2, r);
        //if this is the best path so far,
        if (currentPathLength < shortestPathLength) {
          //set this as the best path
          shortestPathLength = currentPathLength;
          shortestPath = currentShortestPath;
          shortestChannel = channel;
        }
      }

      //move to the next triangle
      continue;
    }

    //go through the edges of the current triangle
    H* s = current->triangle()->getEdge(2);

    for (int i = 0; i < 3; i++) {

      //only consider crossing edges that aren't blocked
      if (!blocked(s)) {

        //get the triangle associated with the potential child node
        PF* childFace = PF(s->getOppEdge()->getFace());

        //calculate the width through this triangle between the last triangle and the next
         float width = INFINITY_LEN;
         width = (current->back() == NULL) ? INFINITY_LEN : (s->getNext()->getOppEdge()->getFace() == current->back()->triangle()) ?
          current->triangle()->link->width(i) : current->triangle()->link->width((i + 2) % 3);
                      
        //only generate the child node if we haven't explored it yet in this path, the width through
        //the triangle is enough, and it's not the last face we explored
        if ((!current->searched(childFace)) && (width >= 2.0f * r)
            && ((current->back() == 0) || (childFace != current->back()->triangle()))) {

          //get a component of the child face
          H* s1 = childFace->getEdge(2);
          float x_1, y_1, x_2, y_2, x_3, y_3;

          //go through the edges to determine which is the entry edge
          int j;
          for (j = 0; j < 3; j++) {

            //that is the edge across which is the parent triangle
            if (s1->getOppEdge()->getFace() == s->getFace()) {

              //and get the vertices starting from that edge
              triangleVertices(s1, x_1, y_1, x_2, y_2, x_3, y_3);
              break;
            }

            //move to the next edge to check
            s1 = s1->getNext();
          }
 
          if (j >= 3) ERR("edge not found");

          //get the closest point on the entry edge to the goal
          FPoint p = closestPointOn(x_1, y_1, x_2, y_2, x2, y2, r);

          //and the start
          FPoint p2 = closestPointOn(x_1, y_1, x_2, y_2, x1, y1, r);

          //get the parent triangle's vertices
          triangleVertices(s, x_1, y_1, x_2, y_2, x_3, y_3);

          //and calculate the angle through that triangle to this one
          float theta = (current->back() == 0) ? 0.0f :
            (s->getNext()->getOppEdge()->getFace() == current->back()->triangle()) ?
            fabs(angleBetween(x_1, y_1, x_2, y_2, x_3, y_3)) :
            fabs(angleBetween(x_3, y_3, x_1, y_1, x_2, y_2));

          //the heuristic is the Euclidean distance between the goal and the closest point
          //to it in this triangle's entry edge
          float h = length(p.getX(), p.getY(), x2, y2);

          //the distance travelled is the maximum of that for the parent triangle
          //plus the maximum of the angle through the parent triangle times the
          //unit's radius, and the parent triangle's heuristic minus that of this one
          float minDistance = current->g() + ::max(theta * r, current->h() - h);

          //maxxed with the minimum distance between this triangle's entry edge and the start
          minDistance = ::max(length(x1, y1, p2.getX(), p2.getY()), minDistance);

          //maxxed with the Euclidean distance between the start and goal minus this
          //triangle's heuristic value
          minDistance = ::max(minDistance, startGoalDistance - h);

          //if this child could yield a shorter path than the best found so far,
          if ((minDistance + h) < shortestPathLength) {

            //put the child search node on the queue
            q.push(new SearchNode(minDistance, h, p, childFace, current));

            //tell the current node that it has an open child
            current->openChild();

            // fixme: what about this code?

            //if we haven't explored this triangle before,
            //						if (!_mesh->marked(childFace))
            //						{
            //							//we are putting an unexplored node on the queue
            //							unexplored++;
            //						}
          }
        }
      }

      //move to the next edge of the parent triangle
      s = s->getNext();
    }
            
    //if no children were opened, close the current node and delete it
    if (current->openChildren() <= 0) {
      current->close();
      delete current;
    }
  }

  //set the current path and channel to the best one found
  currentPath = shortestPath;
  currentChannel = shortestChannel;

  //set the start and goal points
  startPoint.setX(x1);
  startPoint.setY(y1);
  goalPoint.setX(x2);
  goalPoint.setY(y2);
  //	_mesh->end_marking();

  //return whether a path was found or not
  return shortestPathLength < INFINITY_LEN;
}

 
void DCDTPath::abstract()
{
  //Degree-1 and Degree-0 faces
  //keep track of the connected component
  int component = 0;

  Mesh& mesh(_dcdt.getMesh());
  
  //possible degree-1 triangles
  std::list<PF *> degree1;

  //possible degree-2 or 3 triangles
  std::vector<PF *> processing;

  //go through all faces in the triangulation
  FORALL (_dcdt.getMesh().getFaces(), fit) {

    PF *pf = PF(*fit);
    
    if ((!pf) || (pf->isBoundary()))
      continue;

    //keep track of how many edges are constrained
    int numConstrained = 0;

    //this is the adjacent F to a degree-1 triangle
    PF *tempFace = 0;
		
    //loop through the edges of the Face
    //count the number of constrained edges around this Face

    assert(pf->getEdgeCount() == 3);
    
    for (int i = 0; i < 3; i++) {
      HalfEdge *s = pf->getEdge(i);

      //consider a constrained edge or one bordering the outside F to be blocked
      if (blocked(s)) {

        numConstrained++;

      } else {
        
        //if there's an unconstrained edge, record the Face across it
        tempFace = PF(s->getOppEdge()->getFace());
      }
    }

    //if there are two constrained edges, this is a degree-1 triangle
    if (numConstrained == 2) {

      //abstract the F and calculate its widths
      pf->link = new Abstraction(1);
      calculateWidths(pf);

      //if the F across the unconstrained edge is already degree-1,
      if (degree(tempFace) == 1) {
        //this is a tree component
        treeAbstract(tempFace, component++);

      } else if (degree(tempFace) == UNABSTRACTED) {

        //if it's not abstracted, it might now be a degree-1 node
        
        degree1.push_back(tempFace);
      }
    } else if (numConstrained == 3) {

      //if all edges are constrained, this is a degree-0 triangle
      //abstract it and set its component
      pf->link = new Abstraction(0);
      pf->link->component(component++);

    } else {

      //if it's not degree-0 or 1, it might be degree-2 or 3
      //enqueue it for later processing
      processing.push_back(pf);
    }
		        
  }//end face loop

  //keep track of which faces we have considered for degree-1 nodes
  mesh.beginMarking();
  
  //go through the possible degree-1 nodes
  while (!degree1.empty()) {

    //get the next possible degree-1 node
    PF *currentFace = degree1.front();
    degree1.pop_front();

    //if this F has already been dealt with, skip it
    if ((mesh.marked(currentFace)) || (degree(currentFace) != UNABSTRACTED))
      continue;

    //keep track of the number of adjacent triangles
    int numAdjacent = 0;

    //the F across an unconstrained edge
    PF *tempFace = 0;
	
    //calculate the number of unabstracted faces across unconstrained edges
    for (int i = 0; i < 3; i++) {
      HalfEdge *s = currentFace->getEdge(i);

      //only consider adjacent, unabstracted edges
      if (!blocked(s) && (degree(PF(s->getOppEdge()->getFace())) == UNABSTRACTED)) {
        numAdjacent++;
        tempFace = PF(s->getOppEdge()->getFace());
      }
    }

    //if there is only one, this is a degree-1 node
    if (numAdjacent == 1) {

      //abstract this node and calculate its widths
      currentFace->link = new Abstraction(1);
      calculateWidths(currentFace);

      //the adjacent triangle is now possibly degree-1
      degree1.push_back(tempFace);

      //mark this F as dealt with
      mesh.mark(currentFace);

    } else if (numAdjacent == 0) {

      //if there aren't any, this is a tree component
      treeAbstract(currentFace, component++);

      //mark this F as dealt with
      mesh.mark(currentFace);
    }
  }

  //done with degree-1 nodes
  mesh.endMarking();

  //Degree-2 and Degree-3 faces
  //go through all possible degree-3 triangles

  for (size_t n = 0; n < processing.size(); n++) {

    //gets the current F to check
    PF *currentFace = processing[n];

    //we only care about unabstracted nodes at this point
    if (degree(currentFace) != UNABSTRACTED)
      continue;

    //the number of non-degree-1 triangles across unconstrained edges
    int numAdjacent = 0;
	
    //go through the edges of the triangle
    for (int i = 0; i < 3; i++) {

      H *s = currentFace->getEdge(i);

      //get the degree of the triangle across that edge
      int degr = degree(PF(s->getOppEdge()->getFace()));

      //keep track of non-degree-1 edges across unconstrained edges
      if (!blocked(s) && (degr != 1)) {
        numAdjacent++;
      }
    }

    //if this is not a degree-3 triangle, skip it
    if (numAdjacent < 3)
      continue;

    //stack of degree-3 nodes in this connected component
    std::list<PF *> degree3;

    //put the current triangle on the stack
    degree3.push_back(currentFace);

    //continue through all degree-3 nodes in this component
    while (!degree3.empty()) {

      //get one of the degree-3 nodes from the stack
      PF *stackFace = degree3.front();
      degree3.pop_front();

      //abstract it and calculate its widths and set its connected component
      stackFace->link = new Abstraction(3);

      calculateWidths(stackFace);
      stackFace->link->component(component);
      //access the elements of the triangle

      //go in all directions from this node
      for (int i = 0; i < 3; i++) {

        H* s = stackFace->getEdge(i);

        //get the F across this particular edge
        PF *tempFace = PF(s->getOppEdge()->getFace());

        //mark the first F and the previous F
        PF *firstFace = stackFace;
        PF *lastFace = stackFace;

        //keep track of the cumulative angle since the last degree-3 node
        float angle = 0;

        //follow this chain until another degree-3 node is encountered
        while (true) {

          //the number of adjacent faces across unconstrained edges not abstracted as degree-1
          numAdjacent = 0;
				
          //go through the edges of this triangle
          for (int j = 0; j < 3; j++) {
            H* s1 = tempFace->getEdge(j);

            //get the degree of the F across this edge
            int deg = degree(PF(s1->getOppEdge()->getFace()));

            //if it's not blocked or degree-1, it's adjacent
            if (!blocked(s1) && (deg != 1)) {
              numAdjacent++;
            }
          }

          //if the current triangle is degree-3,
          if (numAdjacent == 3) {

            //if it hasn't been abstracted yet, put it on the stack to deal with
            if (degree(tempFace) == UNABSTRACTED) {
              degree3.push_back(tempFace);
            }

            //sets the original degree-3 node adjacent to this one
            stackFace->link->adjacent(i, tempFace);

            //and sets the total angle between them
            stackFace->link->angle(i, angle);

            //if there were no degree-2 nodes in between these,
            if (lastFace == stackFace) {
							
              //go through the edges
              for (int j = 0; j < 3; j++) {

                H* s1 = tempFace->getEdge(j);

                //find the edge between this triangle and the last one
                if (PF(s1->getOppEdge()->getFace()) == lastFace) {

                  //the choke point is the length of the edge between these faces
                  float x1, y1, x2, y2, x3, y3;
                  triangleVertices(s1, x1, y1, x2, y2, x3, y3);
                  stackFace->link->choke(i, length(x1, y1, x2, y2));
                  break;
                }
              }

            } else {
              
              //if there were degree-2 nodes between them,
							
              float choke, width;

              //go through its edges
              for (int j = 0; j < 3; j++) {
                
                H* s1 = lastFace->getEdge(j);

                //find the direction of the original degree-3 node
                if (lastFace->link->adjacent(j) == stackFace) {

                  //get the choke point to this node
                  choke = lastFace->link->choke(j);

                  //get the width through this triangle between the two degree-3 nodes
                  width = lastFace->link->width(s1->getNext()->getOppEdge()->getFace() == tempFace ?
                    j : (j + 2) % 3);

                  //the choke point of the original degree-3 node is the lesser of the two
                  stackFace->link->choke(i, ::min(choke, width));
                  break;
                }
              }
            }
            //we've reached the other degree-3 node - stop following this chain
            break;

          } else if (numAdjacent == 2) {

            //if the triangle is degree-2
            //if the triangle wasn't abstracted before,
            if (degree(tempFace) == UNABSTRACTED) {

              //abstract it and set its widths and connected component
              tempFace->link = new Abstraction(2);
              calculateWidths(tempFace);
              tempFace->link->component(component);
            }
						
            //go through the edges of the triangle
            for (int j = 0; j < 3; j++) {
              H* s1 = tempFace->getEdge(j);
              
              //find the edge across which is the last triangle
              if (s1->getOppEdge()->getFace() == lastFace) {

                //set the angle back towards the original degree-3 node
                tempFace->link->angle(j, angle);

                //and set the adjacent degree-3 node in that direction
                tempFace->link->adjacent(j, firstFace);

                //if there are no nodes between this one and the degree-3 node,
                if (lastFace == firstFace) {

                  //the choke point is simply the length of the edge between them
                  float x1, y1, x2, y2, x3, y3;
                  triangleVertices(s1, x1, y1, x2, y2, x3, y3);
                  tempFace->link->choke(j, length(x1, y1, x2, y2));

                } else {

                  float choke=0, width=0;
									
                  //and go through the edges
                  for (int k = 0; k < 3; k++) {
                    H* s2 = lastFace->getEdge(k);

                    //we want the edge which leads back to the original degree-3 node
                    if (PF(s2->getOppEdge()->getFace()) == tempFace ||
                        (lastFace->link->adjacent(k) == 0)) {
                      continue;
                    }

                    //get the associated choke point width
                    choke = lastFace->link->choke(k);

                    //get the width between the two degree-3 nodes
                    if (PF(s2->getNext()->getOppEdge()->getFace()) == tempFace) {
                      width = lastFace->link->width(k);
                    } else {
                      width = lastFace->link->width((k + 2) % 3);
                    }
                    break;
                  }
                  //the choke point value is the lesser of these
                  tempFace->link->choke(j, ::min(choke, width));
                }
                break;
              }
            }
            
            //calculate the angle between edges leading to degree-3 nodes
            float currentAngle = 0;
						
            //record the next triangle to move to
            PF *nextFace = 0;

            //go through the edges of the triangle
            for (int j = 0; j < 3; j++) {  

              H* s1 = tempFace->getEdge(j);
              PF* pf = PF(s1->getOppEdge()->getFace());
                
              //get the degree of the adjacent triangle across this edge
              int deg = degree(pf);

              //if there is a tree component off this node, collapse it
              if (!blocked(s1) && (deg == 1) && (pf->link->component() == INVALID)) {
                treeCollapse(tempFace, pf, component);
              }

              //when finding the next F, record it
              if (!blocked(s1) && pf != lastFace && (deg != 1)) {
                nextFace = pf;
              }

              //determine the degree of the triangle across the next edge
              int deg2 = degree(PF(s1->getNext()->getOppEdge()->getFace()));

              //if this edge and the next are the connecting edges, record the angle between them
              if (!((deg == 0) || (deg == 1) || (deg2 == 0) || (deg2 == 1))
                  && !blocked(s1) && !blocked(s1->getNext())) {
                float x1, y1, x2, y2, x3, y3;
                triangleVertices(s1, x1, y1, x2, y2, x3, y3);
                currentAngle = abs(angleBetween(x1, y1, x2, y2, x3, y3));
              }
            }

            //increment the cumulative angle so far
            angle += currentAngle;

            //move to the next edge
            lastFace = tempFace;
            tempFace = nextFace;

          } else {
            //in other cases, report an error              
            ERR("ERROR: should only encounter degree-2 and degree-3 faces this way\n");
          }
        }
      }
    }
    //get the next component
    component++;
  }

  //Degree-2 faces (in a ring)
  //go through the processing queue again
  for (size_t i = 0; i < processing.size(); i++) {

    //get the current triangle
    PF *currentFace = processing[i];

    //only continue if it hasn't been abstracted yet
    if (degree(currentFace) != UNABSTRACTED)
      continue;

    //mark it as the first triangle in the ring

    //FIXME: not used? F *firstFace = currentFace;
    //the next triangle to visit
    PF *nextFace = 0;

    //follow the ring of unabstracted faces
    for (;;) {
			
      //go through its edges
      for (int i = 0; i < 3; i++) {
        H* s = currentFace->getEdge(i);

        //record the degree of the triangle across this edge
        int deg = degree(PF(s->getOppEdge()->getFace()));

        //if there's a tree component off this F, collapse it
        if (!blocked(s) && (deg == 1)) {
          treeCollapse(currentFace, PF(s->getOppEdge()->getFace()), component);
        }

        //when the next F is found in the ring, record it
        if (!blocked(s) && (deg == UNABSTRACTED)) {
          nextFace = PF(s->getOppEdge()->getFace());
        }
      }

      //abstract the current F, calculate its widths, and set its connected component
      currentFace->link = new Abstraction(2);
      calculateWidths(currentFace);
      currentFace->link->component(component);
			
      //go through the edges again
      for (int i = 0; i < 3; i++) {

        H* s = currentFace->getEdge(i);

        //if the adjacent triangle is part of the ring
        if (!blocked(s) && degree(PF(s->getOppEdge()->getFace())) != 1) {

          //set the angle and choke values accordingly
          currentFace->link->angle(i, INFINITY);
          currentFace->link->choke(i, INFINITY);
        }
      }

      //if the start of the ring has been reencountered, stop following it
      if (currentFace == nextFace) {
        break;
      }

      //move to the next F in the ring
      currentFace = nextFace;
    }
    
    //move to the next ring (connected component)
    component++;
  }
}



//Abstract into a tree
 
void DCDTPath::treeAbstract(PF *first, int component)
{
  //create a stack of the nodes in the tree
  std::list<PF *> tree;
  //put the initial node on the tree
  tree.push_back(first);

  //continue until the tree can't be expanded anymore
  while (!tree.empty()) {

    //get the next face off of the stack
    PF* currentFace = tree.front();
    tree.pop_front();

    //get the elements of this triangle
		
    //go through its edges
    for (int i = 0; i < 3; i++) {
      HalfEdge *s = currentFace->getEdge(i);

      //ignore triangles across contrained edges
      if (blocked(s))
        continue;

      //gets the face across that edge
      PF* tempFace = PF(s->getOppEdge()->getFace());

      //if the adjacent triangle hasn't been dealt with, put it on the stack
      if ((tempFace->link->angle(0) == INVALID) &&
          (tempFace->link->angle(1) == INVALID) &&
          (tempFace->link->angle(2) == INVALID)) {
        tree.push_back(tempFace);
      }
    }
    
    //abstract the triangle if it hasn't already been
    if (degree(currentFace) == UNABSTRACTED) {

      //if the triangle hasn't been abstracted yet, abstract it and calculate its widths
      currentFace->link = new Abstraction(1);
      calculateWidths(currentFace);
    }
		
    //move through its edges
    for (int i = 0; i < 3; i++) {
      HalfEdge *s = currentFace->getEdge(i);

      //for each unconstrained edge,
      if (!blocked(s)) {

        //set the angle and choke values accordingly
        currentFace->link->angle(i, INFINITY);
        currentFace->link->choke(i, INFINITY);
      }
    }
    //set the triangle's connected component
    currentFace->link->component(component);
  }
}

 
void DCDTPath::treeCollapse(PF* root, PF* currentFace, int component)
{
  Mesh& mesh(_dcdt.getMesh());
  
  //mark edges we have dealt with
  mesh.beginMarking();

  //starting with the root of the tree
  mesh.mark(root);

  //create a stack of triangles to be processed
  std::list<PF*> tree;

  //and a parallel one for angles of those triangles
  std::list<float> angles;

  //put the current triangle on the stack to be processed
  tree.push_back(currentFace);

  //and record that it is at angle 0 from the root
  angles.push_back(0.0f);

  //continue until all faces in the tree have been dealt with
  while (!tree.empty()) {

    //get the top face off the stack
    currentFace = tree.front();
    tree.pop_front();

    //set its connected component
    currentFace->link->component(component);

    //mark this face as dealt with
    mesh.mark(currentFace);
		
    //the angles to the other adjacent triangles
    float angle[2];
    H* s;
    int i;

    //go through the edges
    for (i = 0; i < 3; i++) {
      s = currentFace->getEdge(i);

      //the face across that edge
      PF *tempFace = PF(s->getOppEdge()->getFace());

      //if this is the entry edge,
      if (!blocked(s) && mesh.marked(tempFace)) {

        //set the angle
        currentFace->link->angle(i, angles.front());
        angles.pop_front();

        //calculate the angle to the other adjacent triangles
        float x1, y1, x2, y2, x3, y3;
        triangleVertices(s, x1, y1, x2, y2, x3, y3);
        angle[0] = abs(angleBetween(x1, y1, x2, y2, x3, y3)) + currentFace->link->angle(i);
        angle[1] = abs(angleBetween(x2, y2, x3, y3, x1, y1)) + currentFace->link->angle(i);

        //set the adjacent triangle in that direction to the root
        currentFace->link->adjacent(i, root);

        //if this is the triangle next to the root,
        if (tempFace == root) {

          //the choke value is the length of the edge between them
          currentFace->link->choke(i, length(x1, y1, x2, y2));

        } else {

          //if there are triangles between this and the root,              
					
          float width=0, choke=0;

          //go through its edges
          for (int j = 0; j < 3; j++) {
            H* s1 = tempFace->getEdge(j);

            //get the choke value in that direction
            float temp = tempFace->link->choke(j);

            //if it's a valid value,
            if ((temp != INVALID) && (temp != INFINITY)) {

              //record it
              choke = temp;

              //gets the width through that triangle between the root and current triangles
              if (s1->getNext()->getOppEdge()->getFace() == currentFace) {
                width = tempFace->link->width(j);
              } else {
                width = tempFace->link->width((j + 2) % 3);
              }
              break;
            }
          }
          //the choke value of this triangle is the lesser of these values
          currentFace->link->choke(i, ::min(width, choke));
        }
        break;
      }
    }

    //record where we left off
    int base = i + 1;

    //visit the other two adjacent triangles
    for (int i = 0; i < 2; i++) {

      //move to the next edge
      s = s->getNext();

      //only deal with ones not across a constrained edge
      if (blocked(s))
        continue;

      //set the angle and choke values acoordingly
      currentFace->link->angle((i + base) % 3, INFINITY);
      currentFace->link->choke((i + base) % 3, INFINITY);

      //put them on the tree for processing
      tree.push_back(PF(s->getOppEdge()->getFace()));

      //also put the appropriate angle on the other tree
      angles.push_back(angle[i]);
    }
  }

  //finished collapsing the tree
  mesh.endMarking();
}



//deletes the information from the abstraction
void DCDTPath::deleteAbstraction()
{
  FORALL (_dcdt.getMesh().getFaces(), fit) {

    PF *pf = PF(*fit);
    
    if(pf->link != NULL) {
      delete pf->link;
      pf->link = NULL;
    
    }    
    //if the triangle is abstracted,
    //free the abstraction information
    
  }
   
}





//determines if the given endpoint has to be checked against the current triangle
bool DCDTPath::validEndpoint(H *s, float x, float y, bool left)
{
        //gets the vertices of the current triangle
        float x1, y1, x2, y2, x3, y3, x4, y4;
        triangleVertices(s, x1, y1, x2, y2, x3, y3);
        //checks if a base angle is obtuse
        if ((abs(angleBetween(x1, y1, x2, y2, x3, y3)) >= PI / 2.0f)
                || (abs(angleBetween(x1, y1, x3, y3, x2, y2)) >= PI / 2.0f))
        {
                //if so, we don't have to worry
                return true;
        }
        //gets the angle of the opposite edge
        float theta = atan2(y3 - y2, x3 - x2);
        //turn 90 degrees clockwise
        theta += (theta <= PI / -2.0f) ? 3.0f * PI / 2.0f : PI / -2.0f;
        //create a point this angle from the "top" vertex
        x4 = x1 + cos(theta);
        y4 = y1 + sin(theta);
        //returns whether the unit doesn't have to pass through the narrowest point in this triangle
        return ((orientation(x1, y1, x4, y4, x, y) > 0) == (left));
}


//checks whether there is a valid path from the start or goal through their corresponding triangles
bool DCDTPath::validEndpoint(F *first, F *second, float x, float y, float r)
{
        //goes through the edges of the first triangle
        H *s = first->getEdge(2);
        for (int i = 0; i < 3; i++)
        {
                //when it finds the one between the first and second triangles,
                if (s->getOppEdge()->getFace() == second)
                {
                        //if it has to cross the line extending from either end vertex, checks their width
                        return (((validEndpoint(s, x, y, false)) || (PF(first)->link->width((i + 2) % 3) >= 2.0f * r))
                                && ((validEndpoint(s->getNext(), x, y, true)) || (PF(first)->link->width(i) >= 2.0f * r)));
                }
                //moves to the next edge
                s = s->getNext();
        }
        //if the second triangle was never found, return failure
        return false;
}




bool DCDTPath::validEndpoints(std::vector<H *> & path, float x1, float y1, float x2, float y2, float r)
{
        //if the path is size 1, no problem, otherwise checks that it can go from the first to second triangle,
        //and from the last to second-last triangle
        return ((path.size() < 2) || ((validEndpoint((F *)path[0]->getFace(), (F *)path[1]->getFace(), x1, y1, r))
                && (validEndpoint((F *)path[path.size() - 1]->getFace(), (F *)path[path.size() - 2]->getFace(), x2, y2, r))));
}
