//definition of the SearchNode class 

#ifndef SEARCHNODE_H
#define SEARCHNODE_H

#include "Mesh.h"
#include "Point.h"

//class used to store a state in the search functions

class SearchNode
{
protected:
  //the (minimum) distance travelled since the start node
  float m_dMinDistance;
  //the (admissible and consistent) estimate of the distance to the goal
  float m_dHeuristic;
  //the closest point in this triangle to the goal
  FPoint m_CClosestPoint;
  //the triangle (face) associated with this node
  PathFace* m_CTriangle;
  //the parent search node of this one
  SearchNode *m_CBack;
  //the number of open children
  int m_nOpenChildren;
  //the direction taken from the last node to this one
  int m_nDirection;

public:
  
  //constructor - initializes the member variables
  SearchNode(float dMinDistance, float dHeuristic, FPoint cClosestPoint,
             PathFace* cTriangle, SearchNode *cBack, int nDirection = -1)
  {
    m_dMinDistance = dMinDistance;
    m_dHeuristic = dHeuristic;
    m_CClosestPoint = cClosestPoint;
    m_CTriangle = cTriangle;
    m_CBack = cBack;
    m_nOpenChildren = 0;
    m_nDirection = nDirection;
  }

  //returns the estimated minimum cost of a path from the start to the goal through this node
  float f()
  {
    return (m_dMinDistance + m_dHeuristic);
  }

  //returns the estimated minimum cost of a path from the start to this node
  float g()
  {
    return m_dMinDistance;
  }

  //returns the estimated minimum cost of a path from this node to the goal
  float h()
  {
    return m_dHeuristic;
  }

  //returns (a copy of) the closest point in this triangle to the goal
  FPoint point()
  {
    return m_CClosestPoint;
  }

  //returns the triangle associated with this node
  PathFace* triangle()
  {
    return m_CTriangle;
  }

  //returns the parent search node of this one
  SearchNode *back()
  {
    return m_CBack;
  }

  //returns the direction taken from the last node to this one
  int direction()
  {
    return m_nDirection;
  }

  //sets the number of open children
  void openChild()
  {
    m_nOpenChildren++;
  }

  //closes a child
  bool closeChild()
  {
    return ((--m_nOpenChildren) <= 0);
  }

  //closes the current search node
  //(deleting its parent if it has no other open children)
  void close()
  {
    if ((m_CBack != NULL) && m_CBack->closeChild()) {
      m_CBack->close();
      delete m_CBack;
      m_CBack = NULL;
    }
  }

  //returns the number of open children this node has
  int openChildren()
  {
    return m_nOpenChildren;
  }

  //checks the ancestors of this node to see if a given triangle was already searched
  bool searched(PathFace* triangle)
  {
    SearchNode *current = m_CBack;
    while (current != 0) {
      if (current->triangle() == triangle) {
        return true;
      }
      current = current->back();
    }
    return false;
  }
};

#endif


