/** @file VecField.C
    @see VecField.H

    $Id: VecField.C 5279 2007-06-23 02:49:08Z mburo $
    $Source$
*/

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "Global.H"
#include "VecField.H"
#include "DrawOnTerrain.H"
#include "Game.H"
#include "Profiler.H"
#include "Options.H"

using namespace std;

//===================================================================

VecField::VecField()
{
}

//-------------------------------------------------------------------

VecField::~VecField()
{
  clear();
}

//-------------------------------------------------------------------

void VecField::add_emitter(Emitter *em)
{
  assert(em);
  emitters.insert(em);
}

void VecField::remove_emitter(Emitter*& em)
{
  if (em){
    emitters.erase(em);
    delete em;
    em = 0;
  }
}

//-------------------------------------------------------------------

void VecField::clear()
{
  FORALL (emitters, e) {
    delete *e;
  }
  emitters.clear();
}

//-------------------------------------------------------------------

Vec2<real4> VecField::field(real4 x, real4 y, real4 dx, real4 dy) const
{
  Vec2<real4> vec;
  FORALL (emitters, e) {
    vec += (*e)->field(x, y, dx, dy);
  }
  return vec;
}

//-------------------------------------------------------------------

void VecField::draw(DrawOnTerrain &dot, const Game &game, const Vec3<real4> &color) const
{
  sint4 tile_points = game.get_tile_points();

  sint4 width = game.get_map().get_width();
  sint4 height = game.get_map().get_height();

  real4 dx = 1.0 / tile_points;
  real4 dy = 1.0 / tile_points;

  {
    PROFILE("draw cells");

    dot.begin_lines();

    for (real4 y = 0.0; y <= height; y += 1.0) {
      for (real4 x = 0.0; x <= width; x += 1.0) {

        //    FORS (y, height+1) {
        //      FORS (x, width+1) {

        Vec2<real4> v = field(x*tile_points, y*tile_points);
        if (fabs(v.x) > 0.0001 || fabs(v.y) > 0.0001)
          dot.draw_line(x, y, x+v.x*dx, y+v.y*dy, color, 1.0);

      }
    }
    dot.end_lines();
  }
}

//===================================================================
//===================================================================

VecField::PointEmitter::PointEmitter(real4 x, real4 y, real4 radius_, real4 mag_) :
  Emitter(x, y),
  radius(radius_), mag(mag_)
{
  assert(fabs(x) <= 1.84391e19);
  radius_2 = square(radius);
}

//-------------------------------------------------------------------

Vec2<real4> VecField::PointEmitter::field(real4 px, real4 py, real4, real4) const
{
  Vec2<real4> delta(px - x, py - y);
  
  real4 d2 = square(delta.x) + square(delta.y);
  if (d2 > radius_2 || d2 <= 1) return Vec2<real4>(0,0);
  delta.normalize();
  return delta * mag;
}

//===================================================================

VecField::PointEmitter2::PointEmitter2(real4 x, real4 y, real4 r1_, real4 r2_, real4 mag_) :
  Emitter(x, y),
  r1(r1_), r2(r2_), mag(mag_)
{
}

//-------------------------------------------------------------------

Vec2<real4> VecField::PointEmitter2::field(real4 px, real4 py, real4, real4) const
{
  Vec2<real4> delta(px - x, py - y);
  real4 d = sqrt(square(delta.x) + square(delta.y));
  if (d > r2 || d <= 1) return Vec2<real4>(0,0);
  delta.normalize();
  if (d <= r1) {
    return delta * mag;
  } else {
    return delta * mag * (1.0 - (d - r1) / (r2 - r1));
  }
}


//===================================================================

VecField::GaussianEmitter::GaussianEmitter
(real4 x, real4 y, real4 r1, real4 r2, real4 mag, real4 decay_) :
  PointEmitter2(x, y, r1, r2, mag),
  decay(decay_)
{
}

//-------------------------------------------------------------------

Vec2<real4> VecField::GaussianEmitter::field(real4 px, real4 py, real4, real4) const
{
  Vec2<real4> delta(px - x, py - y);
  real4 d = delta.length();
  if (d > r2 || d <= 0.1) return Vec2<real4>(0,0);
  d -= r1;
  if (d < 0.0) d = 0.0;

  real4 d2 = d*d;
  d2 /= 256.0;

  real4 s = decay;
  real4 s2 = s*s;
  const real4 r2pi = 2.506628274;
  real4 v = mag / (s*r2pi) * expf(-d2 / (2*s2));

  delta.normalize();
  return delta * v;
}

//===================================================================

VecField::DirMultiplier::DirMultiplier(real4 x, real4 y, real4 dx, real4 dy, const Emitter *em_) :
  Emitter(x, y),
  em(em_), dir(dx, dy)
{
  dir.normalize();
}

//-------------------------------------------------------------------

VecField::DirMultiplier::~DirMultiplier()
{
  if (em) delete em;
}

//-------------------------------------------------------------------

Vec2<real4> VecField::DirMultiplier::field(real4 px, real4 py, real4, real4) const
{
  Vec2<real4> delta(px - x, py - y);
  if (delta.x == 0 && delta.y == 0) return em->field(px, py);
  delta.normalize();

  real4 dy = dir.dot(delta);
  Vec2<real4> orth = delta - (dir * dy);
  real4 dx = orth.length();

  real4 mult = dx + dy;
  if (mult <= 0.0) return Vec2<real4>(0, 0);

  return em->field(px, py) * mult;
}

//===================================================================

VecField::OrthoEmitter::OrthoEmitter(real4 x, real4 y, real4 r1_, real4 r2_, real4 mag_) :
  Emitter(x, y),
  r1(r1_), r2(r2_), mag(mag_)
{
}

//-------------------------------------------------------------------

Vec2<real4> VecField::OrthoEmitter::field(real4 px, real4 py, real4 dx, real4 dy) const
{
  Vec2<real4> delta(px - x, py - y);
  if (fabs(delta.x) <= 0.0001 && fabs(delta.y) <= 0.0001) return Vec2<real4>(0,0);

  Vec2<real4> dir(dx, dy);
  if (fabs(dir.x) <= 0.0001 && fabs(dir.y) <= 0.0001) return Vec2<real4>(0,0);

  real4 dist = delta.length();
  if (dist > r2) return Vec2<real4>(0,0);


  real4 dot = delta.dot(dir);
  if (dot >= 0.0) {
    //    cout << "away" << endl;
    return Vec2<real4>(0,0);
  }

  Vec2<real4> orth(dir.y, -dir.x);



  delta.normalize();
  dir.normalize();

  // determine which side the point is on, to turn it away from the main line
  real4 cross = dir.x * delta.y - dir.y * delta.x;

  //  cout << "cross = " << cross << endl;
  //  cout << "orth = " << orth.x << " " << orth.y << endl;

  if (cross < 0) {
    cross = 1.2 + cross;
  } else {
    cross = -(1.2 - cross);
  }


  dist -= r1;
  if (dist > 0.0) {
    real4 mult = 1.0 - (dist / (r2 - r1));
    orth *= mult;
    //    orth.x -= (1.0 - mult) * dx;
    //    orth.y -= (1.0 - mult) * dy;

    orth *= cross;

  } else {
    //    orth.x -= dx;
    //    orth.y -= dy;
  }

  //  orth *= cross;

  return orth * mag;
}
//===================================================================


//-------------------------------------------------------------------
VecField::OrthoLineEmitter::OrthoLineEmitter(real4 x1, real4 y1, real4 x2_, real4 y2_, real4 r1_, real4 r2_, real4 mag_) :
  Emitter(x1, y1),
  x2(x2_),
  y2(y2_),
  r1(r1_), r2(r2_),
  mag(mag_)
{
  if ((x > x2) || (x == x2 && y > y2)){
    std::swap(x, x2);
    std::swap(y, y2);
  }
}
//-------------------------------------------------------------------
Vec2<real4> VecField::OrthoLineEmitter::field(real4 px, real4 py, real4 dx, real4 dy) const{

  Vec2<real4> point = get_point_on_line(px, py);
  OrthoEmitter em(point.x, point.y, r1, r2, mag);
  
  return em.field(px, py, dx, dy);
}
  
//-------------------------------------------------------------------
Vec2<real4> VecField::OrthoLineEmitter::get_point_on_line(real4 px, real4 py) const{
  real4 line_x = x2 - x;
  real4 line_y = y2 - y;
  
  real4 v_x = px - x;
  real4 v_y = py - y;
  
  real4 a = line_x * v_x + line_y * v_y;
  real4 b = line_x * line_x + line_y * line_y;
  
  Vec2<real4> proj(x + line_x * a / b, y + line_y * a / b);

 
  if ((line_x && proj.x < x) || (!line_x && proj.y < y)){
    return Vec2<real4>(x, y);
  }
  else if((line_x && proj.x > x2) || (!line_x && proj.y > y2)){
    return Vec2<real4>(x2, y2);
  }
  else return proj;
}
//===================================================================
VecField::RectangleEmitter::RectangleEmitter(real4 x1, real4 y1, real4 x2, real4 y2, real4 r1, real4 r2, real4 mag) : Emitter(x1, y1)
{
  OrthoLineEmitter* em;
  //north
  em = new OrthoLineEmitter(x1, y1, x2, y1, r1, r2, mag);
  emitters.push_back(em);
  //south
  em = new OrthoLineEmitter(x1, y2, x2, y2, r1, r2, mag);
  emitters.push_back(em);
  //east
  em = new OrthoLineEmitter(x2, y1, x2, y2, r1, r2, mag);
  emitters.push_back(em);
  //west
  em = new OrthoLineEmitter(x1, y1, x1, y2, r1, r2, mag);
  emitters.push_back(em);

}
//-------------------------------------------------------------------
VecField::RectangleEmitter::~RectangleEmitter()
{
  FORALL (emitters, e) {
    delete *e;
  }
  emitters.clear();
}
//-------------------------------------------------------------------
Vec2<real4> VecField::RectangleEmitter::field(real4 px, real4 py, real4 dx, real4 dy) const
{
  assert(!emitters.empty());
  Vec2<real4> temp, closest_p;
  //Vector<VecField::OrthoLineEmitter*> it, closest;
  typeof(emitters.begin()) it, closest;
  closest = it = emitters.begin();
  closest_p = (*it)->get_point_on_line(px, py);
  it++;
  for(;it != emitters.end(); it++){
    temp = (*it)->get_point_on_line(px, py);
    if (Vec2<real4>(temp.x - px, temp.y - py).length()
	< Vec2<real4>(closest_p.x - px, closest_p.y - py).length()){
      closest = it;
      closest_p = temp;
    }
  }
  
  return (*closest)->field(px, py, dx, dy);

  //Vec2<real4> vec;
  //FORALL (emitters, e) {
  //  vec += (*e)->field(px, py, dx, dy);
  //}
  //return vec;    
}
//===================================================================
VecField::PolygonEmitter::PolygonEmitter(Vector<TerrainBase::Segment> segments_, Emitter* em_)
  : 
  Emitter(0,0), segments(segments_), em(em_)
{
}

//-------------------------------------------------------------------
Vec2<real4> VecField::PolygonEmitter::field(real4 px, real4 py, real4 dx, real4 dy) const
{
  if (segments.empty()) return Vec2<real4>(0, 0);

  Vec2<real4> closest, temp;
  // Vector<TerrainBase::Segment>::iterator it;
  typeof(segments.begin()) it = segments.begin();
  closest = closest_point_on_seg(*it, px, py);
  for (; it != segments.end(); it++){
    temp = closest_point_on_seg(*it, px, py);
    if (Vec2<real4>(temp.x - px, temp.y - py).length() <
	Vec2<real4>(closest.x - px, closest.y - py).length()){
      closest = temp;
    }
  }
  em->set_position(closest.x, closest.y);
  return em->field(px, py, dx, dy);
 
}
//-------------------------------------------------------------------
Vec2<real4> VecField::PolygonEmitter::closest_point_on_seg(const TerrainBase::Segment& seg, real4 px, real4 py) const
{
  real4 line_x = seg.l2.x - seg.l1.x;
  real4 line_y = seg.l2.y - seg.l1.y;
  
  real4 v_x = px - seg.l1.x;
  real4 v_y = py - seg.l1.y;
  
  real4 a = line_x * v_x + line_y * v_y;
  real4 b = line_x * line_x + line_y * line_y;
  
  Vec2<real4> proj(x + line_x * a / b, y + line_y * a / b);

  if (line_x){ //not a vertical line
    if ((proj.x <= seg.l2.x && proj.x >= seg.l1.x) || (proj.x >= seg.l2.x && proj.x <= seg.l1.x)){
      return proj;
    }
  }
  else{ //vertical line
    if ((proj.y <= seg.l2.y && proj.y >= seg.l1.y) || (proj.y >= seg.l2.y && proj.y <= seg.l1.y)){
      return proj;
    }
  }

  if (Vec2<real4>(proj.x - seg.l1.x, proj.y - seg.l1.y).length() < 
      Vec2<real4>(proj.x - seg.l2.x, proj.y - seg.l2.y).length()){
    return Vec2<real4>(seg.l1.x, seg.l1.y);
  }
  else return Vec2<real4>(seg.l2.x, seg.l2.y);

}

//===================================================================
VecField::Randomizer::Randomizer(real4 mag_r_, Emitter *em_)
  : Emitter(0,0),  em(em_), mag_r(mag_r_)
{
  sint4 seed = 0;
  ::Options::get("-seed", seed);
  random = Random(seed);
}

//-------------------------------------------------------------------
VecField::Randomizer::~Randomizer(){
  if (em)
    delete em;
}
//-------------------------------------------------------------------
Vec2<real4> VecField::Randomizer::field(real4 px, real4 py, real4 dx, real4 dy) const
{
  Vec2<real4> em_field = em->field(px, py, dx, dy);
  if (em_field.x == 0 && em_field.y == 0){
    return em_field;
  }
  real4 rand_angle = (real4)random.rand_uint4();
  Vec2<real4> rand_field(em_field.length() * mag_r * cos(rand_angle),
			 em_field.length() * mag_r * sin(rand_angle));
  return rand_field + em_field;
  
}
//-------------------------------------------------------------------
void VecField::Randomizer::set_position(real4 x, real4 y)
{
  em->set_position(x, y);
}



//===================================================================
