#ifndef VECFIELD_H
#define VECFIELD_H

// $Id: VecField.H 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "LinAlg.H"
#include "TerrainBase.H"
#include "Random.H"

class DrawOnTerrain;
class Game;

//===================================================================

class VecField
{
public:

  //==============================

  class Emitter {
  protected:
    real4 x, y;
  public:
    Emitter(real4 x_, real4 y_) : x(x_), y(y_) {}
    virtual ~Emitter() {}
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const =0;
    virtual void set_position(real4 x_, real4 y_) {x = x_; y = y_;};
  };

  //==============================

  class PointEmitter : public Emitter {
  protected:
    real4 radius, radius_2;
    real4 mag;
  public:
    PointEmitter(real4 x, real4 y, real4 radius, real4 mag);
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //------------------------------

  /** A truncated cone with radial direction. */

  class PointEmitter2 : public Emitter {
  protected:
    real4 r1, r2;
    real4 mag;
  public:
    PointEmitter2(real4 x, real4 y, real4 r1, real4 r2, real4 mag);
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //------------------------------

  //  r2    _______
  //     _--       --_
  //  __/   r1        \_
  //  |                 |

  /** Gaussian decay starting from an inner radius and cut off at an outer limit. */

  class GaussianEmitter : public PointEmitter2 {
  protected:
    real4 decay;
  public:
    GaussianEmitter(real4 x, real4 y, real4 r1, real4 r2, real4 mag, real4 decay);
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //==============================

  class OrthoEmitter : public Emitter {
  protected :
    real4 r1, r2;
    real4 mag;
  public:
    OrthoEmitter(real4 x, real4 y, real4 r1, real4 r2, real4 mag);
    Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //==============================

  /** Emitter that calculates the point on a line segment nearest to
      the point where the field is being calculated, and treats that point
      as an OrthoEmitter
  */
  class OrthoLineEmitter : public Emitter {
  protected :
    real4 x2, y2;
    real4 r1, r2;
    real4 mag;

  public:
    
    OrthoLineEmitter(real4 x1, real4 y1, real4 x2, real4 y2, real4 r1, real4 r2, real4 mag);
    Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
    Vec2<real4> get_point_on_line(real4 px, real4 py) const;
  };

//==============================
/** Emitter for axis aligned rectangle. Uses OrthoLineEmitters*/
  class RectangleEmitter : public Emitter{
  protected:
    Vector<VecField::OrthoLineEmitter*> emitters;
  public:
    RectangleEmitter(real4 x1, real4 y1, real4 x2, real4 y2, real4 r1, real4 r2,  real4 mag);
    virtual ~RectangleEmitter();
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //==============================
  class PolygonEmitter : public Emitter{
  protected:
    Vector<TerrainBase::Segment> segments;
    Emitter *em;
  public:
    PolygonEmitter(Vector<TerrainBase::Segment> segments_, Emitter* em_);
    virtual ~PolygonEmitter(){if (em) delete em;};
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
    virtual void set_position(real4, real4) {ERR("don't do this");};
    Vec2<real4> closest_point_on_seg(const TerrainBase::Segment& seg, real4 px, real4 py) const;
  };
  //==============================
  // Modifiers
  //==============================

  /** Convolves another emitter by a truncated cone of height 1. */

  class RadialDecay : public Emitter {
  protected:
    const Emitter *em;
    real4 r1; ///< 1 until this radius;
    real4 r2; ///< Linear decay from 1 to 0 from r1 to r2;
  public:
    RadialDecay(real4 x, real4 y, real4 dx, real4 dy, const Emitter *em);
    ~RadialDecay();
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  //------------------------------

  /** Restricts the field to one direction. */

  class DirMultiplier : public Emitter {
  protected:
    const Emitter *em;
    Vec2<real4> dir;
  public:
    DirMultiplier(real4 x, real4 y, real4 dx, real4 dy, const Emitter *em);
    ~DirMultiplier();
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;
  };

  class Randomizer : public Emitter{
  protected:
    Emitter *em;
    real4 mag_r;//ratio of random vector magnitude to emitter vector magnitude
    mutable Random random;
  public:
    Randomizer(real4 mag_r, Emitter *em);
    ~Randomizer();
    virtual Vec2<real4> field(real4 px, real4 py, real4 dx=0, real4 dy=0) const;        virtual void set_position(real4 x_, real4 y_);
  };
  

  //==============================

protected:
  std::set<Emitter*> emitters;

public:
  VecField();
  ~VecField();
  
  Vec2<real4> field(real4 x, real4 y, real4 dx=0, real4 dy=0) const;
  
  void add_emitter(Emitter *em);
  void remove_emitter(Emitter*& em);
  void clear();

  void draw(DrawOnTerrain &dot, const Game &game, const Vec3<real4> &color) const;
};

REGISTER_TYPEOF(2801, Vector<VecField::Emitter*>::iterator);
REGISTER_TYPEOF(2802, Vector<VecField::Emitter*>::const_iterator);

//===================================================================

#endif
