//Base.h

#ifndef BASE_H
#define BASE_H

//DJD: child classes for separating code {

#include "SDL.h"
#ifdef WIN32
#include "se_dcdt.h"
#else
#include "SE/se_dcdt.h"
#endif

#include "Abstraction.h"
#include "SearchNode.h"

#define PI 3.1415926535897932384626433832795f

typedef SeBase TraBase;
typedef SeDcdtVertex TraVertex;
typedef SeDcdtEdge TraEdge;

typedef SeMesh<TraVertex, TraEdge, TraFace> TraMesh;

class TraFace : public SeElement
{
protected:
	//the unprocessed list
	static SrArray<TraFace *> processing;
	//the global update
	static int current;
	//the index of this face in the unprocessed list
	int index;
	//the update when this face was inserted in the unprocessed list
	int update;
	//initializes the face - puts it in the unprocessed list and sets its index and update
	void Initialize() {index = processing.size(); processing.push() = this; update = current;}
public:
	//the number of new faces that haven't been dealt with
	static int Faces() {return processing.size();}
	//retrieve a particular unprocessed face
	static TraFace *Face(int n) {return ((n < 0) || (n >= processing.size())) ? NULL : processing[n];}
	//clear unprocessed face list, increment the global update
	static void Clear() {while (!processing.empty()) processing.pop(); current++;}
	//link to external information
	Abstraction *link;
//	SE_ELEMENT_CASTED_METHODS(SeLinkFace, SeDcdtSymEdge);
	//default constructor - set the link to null and initialize the face
	TraFace() : link (NULL) {Initialize();}
	//copy constructor - initialize the link to that of the other face, and initialize it
	TraFace(const TraFace& f) : SeElement(), link (f.link) {Initialize();}
	//constructor - initialize the link to the value passed and initialize the face
	TraFace(Abstraction *l) : link (l) {Initialize();}
	//destructor - deletes a non-null link, and nulls its entry in the unprocessed list if it's in there
	~TraFace() {if (link != NULL) delete link; if (update == current) processing[index] = NULL;}
	friend SrOutput& operator<<(SrOutput& out, const TraFace&) {return out;}
	friend SrInput& operator>>(SrInput& inp, TraFace&) {return inp;}
	friend int sr_compare(const TraFace*, const TraFace*) {return 0;}
};

//child classes for separating code }

//child class of DCDT, to use our new faces
class TraDcdt : public SeDcdt
{
public:
	//constructor - define the mesh and triangulator to use our faces
	TraDcdt();
	//updates the enemy influence in a particular triangle,
	//and then updates the enemy threat in adjacent triangles
	void UpdateInfluence(TraFace *face, float dUpdate);

//DJD: prototypes for functions used in abstraction {

public:
	//adds abstraction information to the triangulation
	void Abstract();
	//deletes all abstraction information from the triangulation
	void DeleteAbstraction();
private:
	//the "outside" face of the triangulation
	TraFace *outside;
	//abstract an acyclic component of the triangulation graph
	void TreeAbstract(TraFace *first, int component);
	//collapse an acyclic portion into the degree-2 root given
	void TreeCollapse(TraFace *root, TraFace *currentFace, int component);

//prototypes for functions used in abstraction }

//DJD: prototypes for functions used in width calculation {

private:
	//calculates all widths through a triangle
	void CalculateWidths(TraFace *face);
	//calculates a particular width through a triangle
	float TriangleWidth(TraBase *s);
	//search across an edge for the width through the triangle
	float SearchWidth(float x, float y, TraBase *s, float CurrentWidth);
	//checks if the position given in the face given is at least r from any obstacle
	bool ValidPosition(float x, float y, TraFace *face, float r);
	//determines if angles (x, y)-(x1, y1)-(x2, y2) and (x, y)-(x2, y2)-(x1, y1) are accute
	bool Consider(float x, float y, float x1, float y1, float x2, float y2);
	//determines the minimum distance between the point (x, y) and the line going through (x1, y1)-(x2, y2)
	float PointLineDistance(float x, float y, float x1, float y1, float x2, float y2);

//prototypes for functions used in width calculation }

//DJD: protptypes for functions used in point location {

public:
	//perform initialization for sector-based point location on the unprocessed triangles
	void InitializeSectors();
	//checks if the given point is inside the given triangle
	bool InTriangle(TraFace *face, float x, float y);
	//use sector-based point location to find the face in which the given point resides
	SeTriangulator::LocateResult LocatePoint(float x, float y, TraBase* &result);
	//use regular point location to perform the above
	SeTriangulator::LocateResult LocatePointOld(float x, float y, TraBase* &result);
	//"walk" from the given triangle to the one specified
	SeTriangulator::LocateResult LocatePoint(TraFace *start, float x, float y, TraBase* &result);
	//generate a random valid x value for testing point location
	float RandomX();
	//generate a random valid y value for testing point location
	float RandomY();
private:
	//the number of sectors in the vertical dimension
	const static int ySectors = 10;
	//the number of sectors in the horizontal dimension
	const static int xSectors = 10;
	//the width of a single sector
	float sectorWidth;
	//the height of a single sector
	float sectorHeight;
	//the array of sectors
	TraFace *sectors[ySectors][xSectors];
	//return the maximum of the 3 numbers passed
	float Max(float a, float b, float c);
	//return the maximum of the 3 numbers passed
	float Min(float a, float b, float c);
	//return a random value between those given
	float RandomBetween(float min, float max);

//protptypes for functions used in point location }

//DJD: prototypes for functions used in abstracted space searching {

public:
	//find given path with TRA*
	bool SearchPathFast(float x1, float y1, float x2, float y2, float r,
		unsigned int softLimit, unsigned int hardLimit);
	//find given path with TA*
	bool SearchPathBaseFast(float x1, float y1, float x2, float y2, float r,
		Uint32 softLimit, Uint32 hardLimit);
	//get the shortest path between the given points within the given channel of triangles
	float GetShortestPath(SrPolygon& path, SrArray<TraBase *> Channel, float x1, float y1, float x2, float y2, float r);
	//return the polygon around the current channel
	SrPolygon GetChannelBoundary();
	//return the current path
	SrPolygon &GetPath();
	//return the length of the current path
	float GetPathLength();
private:
	//search for a path in an acyclic portion of the triangulation graph
	bool Degree1Path(SrArray<TraBase *>& path, float x1, float y1, TraFace *startFace,
		float x2, float y2, TraFace *goalFace, float r);
	//search for a path along a degree-2 corridor
	bool Degree2Path(SrArray<TraBase *>& path, TraFace *startFace, TraFace *nextFace,
		TraFace *goalFace, float r);
	//search for a path along a degree-2 corridor, crossing a degree-3 node
	bool FollowLoop(SrArray<TraBase *>& path, TraFace *startFace, TraFace *nextFace,
		TraFace *goalFace, TraFace *degree3, float r);
	//walk from one face to a given adjacent face, in a direction specified
	bool WalkBetween(SrArray<TraBase *>& path, TraFace *sourceFace, TraFace *destinationFace, float r, int direction = INVALID);
	//determine if an object of given radius can cross the middle triangle between the adjacent ones given
	bool CanCross(TraBase *first, TraBase *second, TraBase *third, float r);
	bool CanCross(TraFace *middle, TraBase *end1, TraBase *end2, float r);
	//checks if a unit of radius r can enter the path from the start and goal
	bool ValidEndpoints(SrArray<TraBase *>& path, float x1, float y1, float x2, float y2, float r);
	//checks if a unit has to cross the middle of the triangle to enter the path from the point given
	bool ValidEndpoint(TraBase *s, float x, float y, bool left);
	//determines if a unit can fit through the middle of the given triangle to get to the path
	bool ValidEndpoint(TraFace *first, TraFace *second, float x, float y, float r);
	//checks if a given path is valid for a unit of radius r
	bool ValidPath(SrArray<TraBase *>& path, float x1, float y1, float x2, float y2, float r);
	//returns the closest point in a triangle to a given point for a unit of given radius
	SrPnt2 ClosestPointTo(TraFace *face, float x, float y, float r);
	//given a search node, construct the channel it represents
	SrArray<TraBase *> ConstructBaseChannel(SearchNode *goal);
	//given a search node, construct a channel, filling in degree-1 and 2 nodes between those searched
	void ConstructBasePath(SrArray<TraBase *> &path, SearchNode *goalNode, TraFace *start, TraFace *goal, int direction = INVALID);
	//the current channel
	SrArray<TraBase *> currentChannel;
	//the current path
	SrPolygon currentPath;
	//the length of the current path
	float currentPathLength;
	//the current start position
	SrPnt2 startPoint;
	//the current goal position
	SrPnt2 goalPoint;

//prototypes for functions used in abstracted space searching }

//DJD: prototypes for utility functions {

public:
	//retrieves the vertices and whether the edges are constrained, for a given triangle
	void TriangleInfo(TraFace *face, float &x1, float &y1, bool &e1, 
		float &x2, float &y2, bool &e2, float &x3, float &y3, bool &e3);
	void TriangleInfo(TraBase *s, float &x1, float &y1, bool &e1, 
		float &x2, float &y2, bool &e2, float &x3, float &y3, bool &e3);
	//revireves the vertices for a given triangle
	void TriangleVertices(TraFace *face, float &x1, float &y1,
		float &x2, float &y2, float &x3, float &y3);
	void TriangleVertices(TraBase *s, float &x1, float &y1,
		float &x2, float &y2, float &x3, float &y3);
	//retrieves whether the edges are constrained, for a given triangle
	void TriangleEdges(TraFace *face, bool &e1, bool &e2, bool &e3);
	void TriangleEdges(TraBase *s, bool &e1, bool &e2, bool &e3);
	//returns the midpoint of a given triangle
	void TriangleMidpoint(TraFace *face, float &x, float &y);
	SrPnt2 TriangleMidpoint(TraFace *face);
	//returns whether a given edge is either constrained, or borders the "outside" polygon
	bool Blocked(TraBase *s);
	//returns the degree of a triangle's node (UNABSTRACTED if it has not been mapped)
	int Degree(TraBase *s);
	int Degree(SeFace *face);
private:
	//returns whether the given angle is accute
	bool IsAccute(float theta);
	//returns whether the given angle is obtuse
	bool IsObtuse(float theta);
	//returns the length of the line segment (x1, y1)-(x2, y2)
	float Length(float x1, float y1, float x2, float y2);
	//returns the smaller of the two values passed
	float Minimum(float a, float b);
	//returns the larger of the two values passed
	float Maximum(float a, float b);
	//determines the angle between (x1, y1)-(x2, y2)-(x3, y3)
	float AngleBetween(float x1, float y1, float x2, float y2, float x3, float y3);
	//returns 0 if the points are colinear, <0 if they are in clockwise order, and >0 for counterclockwise
	float Orientation(float x1, float y1, float x2, float y2, float x3, float y3);
	//returns the minimum distance between the line segments (A1x, A1y)-(A2x, A2y) and (B1x, B1y)-(B2x, B2y)
	float SegmentDistance(float A1x, float A1y, float A2x, float A2y, float B1x, float B1y, float B2x, float B2y);
	//returns the minimum distance between the line segments (ls1x, ls1y)-(ls2x, ls2y) and (ll1x, ll1y)-(ll2x, ll2y),
	//at least r away from the segments' end points
	float SegmentDistance(float ls1x, float ls1y, float ls2x, float ls2y, float ll1x, float ll1y, float ll2x, float ll2y, float r);
	//returns the minimum distance between point (x, y) and segment (x1, y1)-(x2, y2)
	float PointSegmentDistance(float x, float y, float x1, float y1, float x2, float y2);
	//returns the closest point on the segment (x1, y1)-(x2, y2) to point (x, y),
	//at least r away from th segments' midpoints
	SrPnt2 ClosestPointOn(float x1, float y1, float x2, float y2, float x, float y, float r);

//prototypes for utility functions }
};

#endif
