//-----------------------------------------------------------------------------
/** @file bucketfringesearch.C
    @see bucketfringesearch.h

    $Id: bucketfringesearch.C 5279 2007-06-23 02:49:08Z mburo $
    $Source$
*/
//-----------------------------------------------------------------------------

#include "bucketfringesearch.h"

#include <algorithm>

using namespace std;
using namespace PathFind;

//-----------------------------------------------------------------------------

BucketFringe::BucketFringe(int numberBuckets)
    : m_headNode(numberBuckets, NO_NODE)
{
    assert(numberBuckets > 0);
    m_currentBucket = -1;
    m_numberBuckets = numberBuckets;
    m_numberNodes = 0;
}
    
bool BucketFringe::allLowerBucketsEmpty(int bucket) const
{
    assert(bucket >= 0);
    assert(bucket < m_numberBuckets);
    for (int i = 0; i < bucket; ++i)
        if (m_headNode[i] != NO_NODE)
            return false;
    return true;
}

void BucketFringe::init(int numberNodes)
{
    m_nodes.resize(numberNodes);
    m_nodeInBucket.clear();
    m_nodeInBucket.resize(numberNodes, -1);
    m_numberNodes = numberNodes;
    m_currentNode = NO_NODE;
    fill(m_headNode.begin(), m_headNode.end(), NO_NODE);
}

void BucketFringe::insertNode(int nodeId)
{
    int oldBucket = m_nodeInBucket[nodeId];
    if (oldBucket >= 0)
        removeNode(nodeId);
    m_nodeInBucket[nodeId] = m_currentBucket;
    Node& newNode = getNode(nodeId);
    int headNode = getHeadNode(m_currentBucket);
    if (headNode != NO_NODE)
    {
        Node& current = getNode(m_currentNode);
        int next = current.m_next;
        newNode.m_previous = m_currentNode;
        newNode.m_next = next;
        if (next != NO_NODE)
            getNode(next).m_previous = nodeId;
        current.m_next = nodeId;
    }
    else
    {
        setHeadNode(m_currentBucket, nodeId);
        newNode.m_previous = NO_NODE;
        newNode.m_next = NO_NODE;
    }
}

void BucketFringe::insertNode(int nodeId, int bucket)
{
    int oldBucket = m_nodeInBucket[nodeId];
    if (oldBucket >= 0)
        removeNode(nodeId);
    m_nodeInBucket[nodeId] = bucket;
    int headNode = getHeadNode(bucket);
    Node& newHead = getNode(nodeId);
    if (headNode != NO_NODE)
    {
        Node& oldHead = getNode(headNode);
        newHead.m_previous = NO_NODE;
        newHead.m_next = headNode;
        oldHead.m_previous = nodeId;
    }
    else
    {
        newHead.m_previous = NO_NODE;
        newHead.m_next = NO_NODE;
    }
    setHeadNode(bucket, nodeId);
}

void BucketFringe::nextNode()
{
    assert(m_currentNode != NO_NODE);
    m_currentNode = getNode(m_currentNode).m_next;
    assert(m_currentNode == NO_NODE || m_nodeInBucket[m_currentNode] >= 0);
}

void BucketFringe::print(ostream& o) const
{
    o << "BucketFringe {\n";
    for (int bucket = 0; bucket < m_numberBuckets; ++bucket)
    {
        int nodeId = getHeadNode(bucket);
        if (nodeId == NO_NODE)
            continue;
        o << "Bucket "<< bucket << " {\n";
        while (nodeId != NO_NODE)
        {
            o << nodeId << '\n';
            nodeId = getNode(nodeId).m_next;
        }
        o << "}\n";
    }
    o << "}\n";
}

void BucketFringe::removeNode(int nodeId)
{
    assert(nodeId != NO_NODE);
    assert(m_nodeInBucket[nodeId] >= 0);
    Node& node = getNode(nodeId);
    int previous = node.m_previous;
    int next = node.m_next;
    if (previous != NO_NODE)
        getNode(previous).m_next = next;
    else
        setHeadNode(m_nodeInBucket[nodeId], next);
    if (next != NO_NODE)
        getNode(next).m_previous = previous;
    m_nodeInBucket[nodeId] = -1;
}

void BucketFringe::startIteration(int bucket)
{
    m_currentBucket = bucket;
    assert(allLowerBucketsEmpty(bucket));
    m_currentNode = getHeadNode(bucket);
    assert(m_currentNode != NO_NODE);
}

//-----------------------------------------------------------------------------
