/** @file MapTool.C
    @see MapTool.H

    $Id: MapCliff.C 5279 2007-06-23 02:49:08Z mburo $
    $Source$
*/
// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "MapTool.H"
#include "GameConst.H"
#include "Random.H"
#include "Options.H"

using namespace std;

static bool CREATE_WATER = false;
#define TERRAN_UNITS       0
#define TERRAN_BUILDINGS   1
#define BUG_UNITS          0
#define BUG_BUILDINGS      1
#define PANTHEON_UNITS     0
#define PANTHEON_BUILDINGS 1

//===================================================================

/** Generate a random map with CLIFF tiles. */

void MapTool::generate_random_cliff_map(stringstream &world)
{
  opt.get();

  Random rand;
  sint4 seed = opt.seed;
  rand.srand(seed);

  // note: if loading a map, -nplayers values must match
  sint4 num_players = opt.nplayers;
  
  sint4 tiles_x, tiles_y;
  sint4 gtiles_x = opt.width;
  sint4 gtiles_y = opt.height;
  sint4 z_sight = 0;

  bool rotate = opt.rot;
  CREATE_WATER = opt.water;

  // if the world is rotated 45 degrees, impose somewhat artificial 45
  // degree bariers to define the playing area.  The area of
  // the playing field is the same as it would have been, but the
  // world is still an axis aligned rectangle, so the actual world may
  // be much larger

  /*
    +-----+
    |  /\ |
    | /  \|
    |/   /|
    |\  / |
    | \/  |
    +-----+
  */

  // tiles_x is the actual width of the tile array stored
  // gtiles_x is the width of the playfield and is constant under rotation

  if (rotate) {
    gtiles_x = sint4(gtiles_x * 0.70710678); // side lengths are now fractions of 45 degree lines
    gtiles_y = sint4(gtiles_y * 0.70710678);
    tiles_x = gtiles_x + gtiles_y;
    tiles_y = gtiles_x + gtiles_y;
  } else {
    tiles_x = gtiles_x;
    tiles_y = gtiles_y;
  }

  world << num_players << endl;
  world << gtiles_x << endl;
  world << gtiles_y << endl;
  world << z_sight << endl;
  if (rotate) world << 1 << endl;
  else        world << 0 << endl;

  sint4 tile_n = tiles_x * tiles_y;

  Map<Tile> map_data;
  map_data.allocate(tiles_x, tiles_y);

  sint4 h_x = tiles_x + 1;
  sint4 h_y = tiles_y + 1;  
  sint4 h_n = h_x * h_y;

  Vector<sint2> height(h_n); // height field (index is not a tile index!)
  Vector<sint2> bad(tile_n); // do not put stuff on tile if true, works on tiles!

  // first heights heights represent layers (will be multiplied by GameConst::HEIGHT_MULT later

  real8 fplat = opt.fplat;
  bool plug_holes = opt.holes;
  sint4 clump = opt.clump;
  bool smooth = opt.smooth;

  if (smooth) plug_holes = false; // smooth > plug_holes

  if (clump < 1) clump = 1;

  if (clump <= 1) {

    FORS (i, sint4(tile_n * fplat)) {
      sint4 r;
      do { r = rand.rand_uint4() % h_n; } while (height[r] == 1);
      height[r] = 1;
    }

  } else {

    FORS (i, sint4(tile_n * fplat)) {
      sint4 r;
      do { r = rand.rand_uint4() % h_n; } while (height[r] >= 3);
    
      sint4 next_to = 0;
      sint4 out_of = 4;
    
      sint4 x = r % h_x;
      sint4 y = r / h_x;
  
      if (x == 0 || x == h_x-1) out_of--;
      if (y == 0 || y == h_y-1) out_of--;
    
      sint4 h = height[r];

      if (x > 0     && height[r-1]   == h+1) next_to++;
      if (x < h_x-1 && height[r+1]   == h+1) next_to++;
      if (y > 0     && height[r-h_x] == h+1) next_to++;
      if (y < h_y-1 && height[r+h_x] == h+1) next_to++;

      const sint4 offset = 4;

      if (out_of == next_to) {
	if (rand.rand_uint4() % 3) height[r] += 1;
	else if (h <= 1) height[r] += 2;
        continue;
      }
      if (next_to * 2 >= out_of) {
        next_to += offset;
        out_of += offset;
      }
      if (next_to > 0) {
        if ((sint4)(rand.rand_uint4() % out_of) < next_to) {
	  height[r] += 1;
	  continue;
	}
      }
      if (rand.rand_uint4() % clump < 1) {
        height[r]++; // = 1;
        continue;
      }
      --i;
    }
  }

  if (plug_holes) {

    FORS (i, h_n) {
      sint4 r = i;

      sint4 next_to = 0;
      sint4 out_of = 4;

      sint4 rx = r % h_x;
      sint4 ry = r / h_x;

      sint4 h = height[r];
    
      if (rx == 0 || rx == h_x-1) out_of--;
      if (ry == 0 || ry == h_y-1) out_of--;

      if (rx > 0     && height[r-1]   > h) next_to++;
      if (rx < h_x-1 && height[r+1]   > h) next_to++;
      if (ry > 0     && height[r-h_x] > h) next_to++;
      if (ry < h_y-1 && height[r+h_x] > h) next_to++;

      if (out_of == next_to) {// && height[r] == 0) {
        //	cout << i << " " << next_to << " " << out_of << " plug" << endl;
        height[r]++;// = 1;
      }
    }
  }

  // the world is now a height field with no ramps
  // flatten the corner and add demo buildings

  if (tiles_x > 40 && tiles_y > 60) {
    FORU (i, 40) {
      FORU (j, 60) {
        height[i + j*h_x] = 0;
      }
    }
  }

  // smooth plateaus:
  //  continue filling tiles iff it has three or four filled HV neighbors

  bool changed = smooth;

  while (changed) {

    changed = false;
    
    sint4 dx[] = { 1,  0, -1, 0 };
    sint4 dy[] = { 0, -1,  0, 1 };

    FORS (i, h_n) {

      sint4 h = height[i];

      sint4 ix = i % h_x;
      sint4 iy = i / h_x;

      sint4 next_to = 0;
      
      FORS (j, 4) {
	sint4 jx = ix + dx[j];
	sint4 jy = iy + dy[j];	  
	if (jx < 0 || jx >= h_x || jy < 0 || jy >= h_y) next_to++;
	else if (height[jy*h_x + jx] > h) next_to++;
      }
      
      if (next_to >= 3) { height[i]++; changed = true; }
    }
  }


  FORS (i, h_n) height[i] *= GameConst::HEIGHT_MULT;

  // from here on: heights are multiples of GameConst::HEIGHT_MULT
  // or at half-heights (ramps)
  
  FORS (i, h_n) {

    sint4 x = i % h_x;
    sint4 y = i / h_x;

    if (x == h_x-1 || y == h_y-1) continue;

    //    cout << height[i] << " ";
    
    sint4 h[4];
    
    h[0] = height[i];
    h[1] = height[i+1];
    h[2] = height[i+h_x+1];
    h[3] = height[i+h_x];

    Tile &tile = map_data.get_tile(x, y);

    tile.set_topo(h[0], h[1], h[2], h[3], Tile::NO_SPLIT);
    tile.set_type(GROUND);

    assert(x >= 0 && x < tiles_x);	
    assert(y >= 0 && y < tiles_y);
    
    // equal heights -> regular ground tile
    
    if (h[0] == h[1] && h[0] == h[2] && h[0] == h[3]) {

      tile.set_topo(h[0], h[1], h[2], h[3], Tile::NO_SPLIT);
      tile.set_type(GROUND);
      continue;
    }

    // 3 equal heights -> split
    //  - make equal height GROUND triangle and the other CLIFF or
    //  - create ridge
    
    {
      if ((h[0] == h[1] && h[1] == h[2]) ||
          (h[0] == h[1] && h[1] == h[3]) ||
          (h[0] == h[2] && h[2] == h[3]) ||
          (h[1] == h[2] && h[2] == h[3])) {

        sint4 i;
        FOR (i, 4) {
          if (h[i] != h[(i+1)&3] && h[i] != h[(i+2)&3]) break;
        }

	if (1 || (rand.rand_uint4() & 8)) {

	  // equal height triangle
	  
	  if (i == 0 || i == 2) {  // |/|
	    tile.set_topo(h[0], h[1], h[2], h[3], Tile::BT_SPLIT);
	  } else { // |\|
	    tile.set_topo(h[0], h[1], h[2], h[3], Tile::TB_SPLIT);
	  }
      
	  if (i == 0 || i == 3) tile.set_type(CLIFF, GROUND); // west CLIFF
	  else                  tile.set_type(GROUND, CLIFF); // east CLIFF

	} else {

	  // ridge

	  if (i == 0 || i == 2) {  // |\|
	    tile.set_topo(h[0], h[1], h[2], h[3], Tile::TB_SPLIT);
	  } else { // |/|
	    tile.set_topo(h[0], h[1], h[2], h[3], Tile::BT_SPLIT);
	  }

	  tile.set_type(CLIFF, CLIFF);
	}
	assert(x >= 0 && x < tiles_x);	
	assert(y >= 0 && y < tiles_y);
	bad[y*tiles_x+x] = 1;
	continue;
      }
    }

    
    // 2 and 2 equal heights -> full CLIFF tile
    
    {
      sint4 i;
      FOR (i, 4) {
	if (h[i] == h[(i+1) & 3] && h[(i+2) & 3] == h[(i+3) & 3]) break;
      }
    
      if (i < 4) {

	tile.set_topo(h[0], h[1], h[2], h[3], Tile::NO_SPLIT);

	// small chance of ramp
	if (rand.rand_uint4() % 100 < 15) tile.set_type(GROUND);
	else                              tile.set_type(CLIFF);

	assert(x >= 0 && x < tiles_x);	
	assert(y >= 0 && y < tiles_y);
	bad[y*tiles_x+x] = 1;
	continue;
      }
    }

    // otherwise: random CLIFF/CLIFF split

    {
      tile.set_type(CLIFF, CLIFF);
   
      if (rand.rand_uint4() & 8) {
	tile.set_topo(h[0], h[1], h[2], h[3], Tile::BT_SPLIT);
      } else {
	tile.set_topo(h[0], h[1], h[2], h[3], Tile::TB_SPLIT);
      }

      assert(x >= 0 && x < tiles_x);	
      assert(y >= 0 && y < tiles_y);
      bad[y*tiles_x+x] = 1;
    }
  }

  if (rotate) {

    FORS (i, tile_n) {
      sint4 x = i % tiles_x;
      sint4 y = i / tiles_x;

      if ((x + y < gtiles_y-1) ||
          (tiles_x - x + tiles_y - y < gtiles_y+1) ||
          (tiles_x - x + y < gtiles_x) ||
          (x + tiles_y - y < gtiles_x)) {
        map_data(x,y).set_type(Tile::UNKNOWN);
      }
    }
  }

  FORS (i, tile_n) {

    sint4 x = i % tiles_x;
    sint4 y = i / tiles_x;

    Tile &t = map_data.get_tile(i);
    Tile::Type tw, te;
    Tile::Split split;
    t.get_type(tw, te);
    sint4 h0, h1, h2, h3;
    t.get_topo(h0, h1, h2, h3, split);

    if (split == Tile::NO_SPLIT && tw != CLIFF && (h0 != h3 || h0 != h1)) {

      // ramp

      if (h0 != h3) {

	// | ramp

	if (x > 0) {  // extend left
	  Tile &t = map_data.get_tile(x-1, y);
	  t.get_type(tw, te);
	  t.get_topo(h0, h1, h2, h3, split);

	  if (rand.rand_uint4() % 100 < 50 && te == CLIFF) {
	    if (split) t.set_type(tw, GROUND);
	    else       t.set_type(GROUND, GROUND);
	  }
	}
	
	if (x < tiles_x-1) {  // extend right
	  Tile &t = map_data.get_tile(x+1, y);
	  t.get_type(tw, te);
	  t.get_topo(h0, h1, h2, h3, split);

	  if (rand.rand_uint4() % 100 < 50 && tw == CLIFF) {
	    if (split) t.set_type(GROUND, te);
	    else       t.set_type(GROUND, GROUND);
	  }
	}

      } else {

	// - ramp

 	if (y > 0) {  // extend up
	  Tile &t = map_data.get_tile(x, y-1);
	  t.get_type(tw, te);
	  t.get_topo(h0, h1, h2, h3, split);

	  if (rand.rand_uint4() % 100 < 50 &&
	      ((!split && tw == CLIFF) ||                    // | |
	       (split == Tile::TB_SPLIT && tw == CLIFF) ||   // |\|
	       (split == Tile::BT_SPLIT && te == CLIFF))) {  // |/|

	    if   (!split)                     t.set_type(GROUND, GROUND);
	    else if (split == Tile::TB_SPLIT) t.set_type(GROUND, te);
	    else                              t.set_type(tw, GROUND);
	  }
	}
	
	if (y < tiles_y-1) {  // extend down
	  Tile &t = map_data.get_tile(x, y+1);
	  t.get_type(tw, te);
	  t.get_topo(h0, h1, h2, h3, split);

	  if (rand.rand_uint4() % 100 < 50 &&
	      ((!split && tw == CLIFF) ||                    // | |
	       (split == Tile::TB_SPLIT && te == CLIFF) ||   // |\|
	       (split == Tile::BT_SPLIT && tw == CLIFF))) {  // |/|

	    if   (!split   )                  t.set_type(GROUND, GROUND);
	    else if (split == Tile::TB_SPLIT) t.set_type(tw, GROUND);
	    else                              t.set_type(GROUND, te);
	  }
	}
      }
    }
  }
  
  if (CREATE_WATER) {

    // create lakes
    FORS (i, 20) {

      const sint4 D = 5;
      sint4 x, y;
      Tile::Type tw, te;
      do {
	x = rand.rand_uint4() % (tiles_x-D);
	y = rand.rand_uint4() % (tiles_y-D);
	map_data(x,y).get_type(tw, te);
      } while (!map_data(x,y).is_flat() || map_data(x,y).get_min_h() != 0 || tw == Tile::UNKNOWN);

      FORS (dx, D) {
        FORS (dy, D) {
          if (!map_data(x+dx, y+dy).is_flat()) break;
          if (map_data(x+dx, y+dy).get_min_h() != map_data(x,y).get_min_h()) break;
          map_data(x+dx, y+dy).set_type(WATER);
        }
      }
    }

    // smooth corners
    for (sint4 y = 1; y < tiles_y-2; y++) {
      for (sint4 x = 1; x < tiles_x-2; x++) {
        
        Tile::Type a, b;
        map_data(x,y).get_type(a, b);

        if (a == b && b == WATER) {
	  // sint4 h[4];
	  // Tile::Split split;

          if (map_data(x, y-1).get_typeS() != WATER &&
              map_data(x-1, y).get_typeE() != WATER &&
              map_data(x, y+1).get_typeN() == WATER &&
              map_data(x+1, y).get_typeW() == WATER)
            {
              map_data(x,y).set_type(GROUND, WATER, Tile::BT_SPLIT);
            }
          if (map_data(x, y+1).get_typeN() != WATER &&
              map_data(x+1, y).get_typeW() != WATER &&
              map_data(x, y-1).get_typeS() == WATER &&
              map_data(x-1, y).get_typeE() == WATER)
            {
              map_data(x,y).set_type(WATER, GROUND, Tile::BT_SPLIT);
            }
          if (map_data(x, y+1).get_typeN() != WATER &&
              map_data(x-1, y).get_typeE() != WATER &&
              map_data(x, y-1).get_typeS() == WATER &&
              map_data(x+1, y).get_typeW() == WATER)
            {
              map_data(x,y).set_type(GROUND, WATER, Tile::TB_SPLIT);
            }
          if (map_data(x, y-1).get_typeS() != WATER &&
              map_data(x+1, y).get_typeW() != WATER &&
              map_data(x, y+1).get_typeN() == WATER &&
              map_data(x-1, y).get_typeE() == WATER)
            {
              map_data(x,y).set_type(WATER, GROUND, Tile::TB_SPLIT);
            }
        }
      }
    }
  }

  world << "<TILES>" << endl;
  map_data.write(world);
  world << "</TILES>" << endl;

  // mark the tiles on the boundary, stop object creation there

  if (rotate) {

    FORS (i, tile_n) {
      sint4 x = i % tiles_x;
      sint4 y = i / tiles_x;

      if ((x + y <= gtiles_y-1) ||
          (tiles_x - x + tiles_y - y <= gtiles_y+1) ||
          (tiles_x - x + y <= gtiles_x) ||
          (x + tiles_y - y <= gtiles_x))
        { bad[i] = 1; continue; }
    }
  }

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  sint4 tile_points = GameConst::TILE_POINTS;  // fixme: should depend on game instance

  sint4 nobjs = opt.nobjs;

  const bool AIR_UNITS = true;
  const bool LAND_UNITS = true;
  
  if (AIR_UNITS && LAND_UNITS) nobjs /= 2;

  if (tiles_x > 40 && tiles_y > 60) {

    Vector<string> units;
#if TERRAN_BUILDINGS
    units.push_back("bunker");
    units.push_back("barracks");
    units.push_back("mechBay");
    units.push_back("airBase");
    units.push_back("extractor");
    units.push_back("academy");
    units.push_back("supplyCache");
    units.push_back("engineering");
    units.push_back("theoretical_lab");
    units.push_back("armory");
    units.push_back("turret");
    units.push_back("controlCenter");
    //Extensions have been commented out.  They will cause the game to crash if they are created without the proper base building with the current setup.
    //units.push_back("airTower");
    //units.push_back("comsat");
    //units.push_back("factory_shop");
    //units.push_back("phys_ext");
    //units.push_back("inf_ext");
    //units.push_back("nukeSilo");    
#endif

#if TERRAN_UNITS
    units.push_back("marine");
    units.push_back("spy");
    units.push_back("medic");
    units.push_back("toaster");
    units.push_back("mech");
    units.push_back("hoverbike");
    units.push_back("tank");
    units.push_back("flyingLab");
    units.push_back("dropship");
    units.push_back("striker");
#endif

#if BUG_BUILDINGS
    units.push_back("xenopsyl_nest");
    units.push_back("wyrmhole");
    units.push_back("mutator");
    units.push_back("reservoir");
    units.push_back("snare_pit");
    units.push_back("nursery");
    units.push_back("mound");
    units.push_back("tor");
    units.push_back("citadel");
    units.push_back("pillar");
    units.push_back("obelisk");
#endif

#if BUG_UNITS
    units.push_back("chalcix");
    units.push_back("hercantus");
    units.push_back("hylecoete");
    units.push_back("invicor");
    units.push_back("larva");
    units.push_back("myrmec");
    units.push_back("pupa");
    units.push_back("nephilata");
    units.push_back("vector");
    units.push_back("xenopsyl");
    units.push_back("crenator");
    units.push_back("versutum");
    units.push_back("acherops");
    units.push_back("veneniferum");
    units.push_back("infensum");
#endif

#if PANTHEON_BUILDINGS
    units.push_back("pantheon_acropolis");
    units.push_back("pantheon_troop_gate");
    units.push_back("pantheon_rift_gate");
    units.push_back("pantheon_rift_amplifier");
    units.push_back("pantheon_sentinel");
    units.push_back("pantheon_emitter");
    units.push_back("pantheon_energy_tap");
    units.push_back("pantheon_sanctum");
    units.push_back("pantheon_well");
    units.push_back("pantheon_foundry");
    units.push_back("pantheon_aerodrome");
    units.push_back("pantheon_watchtower");
    units.push_back("pantheon_arena");
    units.push_back("pantheon_aerie");
    units.push_back("pantheon_constructor");
    units.push_back("pantheon_workshop");
#endif

#if PANTHEON_UNITS
    units.push_back("alekto");
    units.push_back("megaera");
    units.push_back("despina");
    units.push_back("dactyl");
    units.push_back("sycorax");
    units.push_back("atropos");
    units.push_back("aether");
    units.push_back("algos");
    units.push_back("aello");
    units.push_back("okypetes");
    units.push_back("podarge");
    units.push_back("hydra");
    units.push_back("fang");
    units.push_back("panoptes");
    units.push_back("atlas");
#endif

    sint4 sx = 3;
    sint4 sy = 3;
    FORU (i, units.size()) {
      string t = units[i];
      place_unit(world, t, 0, sx, sy);
      FORU (j, 6) {
        FORU (k, 6) {
          bad[(sx-2+j) + tiles_x*(sy-2+k)] = 16;
        }
      }
      sx += 6;
      if (sx > 37) {
        sy += 6;
        sx = 3;
      }
    }
  }

#if 0
  // create a mineral patch close to the control center
  if (tiles_x > 40 && tiles_y > 60) {
    sint4 x0,y0;
    x0 = 24;
    y0 = 24;

    FORS (j, 10) {
      sint4 x = x0 - 2 + rand.rand_uint4() % 5;
      sint4 y = y0 - 2 + rand.rand_uint4() % 5;
      if (x < 1 || x >= tiles_x-1) continue;
      if (y < 1 || y >= tiles_y-1) continue;

      FORU (u, 3) {
        FORU (v, 3) {
          bad[(x-1+u) + tiles_x*(y-1+v)] = 2;
        }
      }

      world << "mineral" << endl;
      world << "<INT_ATTR>" << endl;
      world << "owner " << num_players << endl;
      world << "x " << x * tile_points - 6 + rand.rand_uint4() % 13 << endl;
      world << "y " << y * tile_points - 6 + rand.rand_uint4() % 13 << endl;
      world << "</INT_ATTR>" << endl;
    }

    FORU (i, 16) {
      FORU (j, 16) {
        bad[i + tiles_x*j] = 32; // freespace
      }
    }
  }
#endif


  // create random clusters of minerals

#if 1
  FORS (i, 7) {
    sint4 x0, y0;
    sint4 t = rand.rand_uint4() % tile_n;
    x0 = t % tiles_x;
    y0 = t / tiles_y;
    if (bad[t] & ~2) { --i; continue; }

    FORS (j, 10) {
      sint4 x = x0 - 2 + rand.rand_uint4() % 5;
      sint4 y = y0 - 2 + rand.rand_uint4() % 5;
      if (x < 1 || x >= tiles_x-1) continue;
      if (y < 1 || y >= tiles_y-1) continue;
      bool can_use = true;
      FORU (u, 3) {
        FORU (v, 3) {
          if (bad[(x-1+u) + tiles_x*(y-1+v)] & ~2) can_use = false;
        }
      }

      if (!can_use) continue;

      FORU (u, 3) {
        FORU (v, 3) {
          bad[(x-1+u) + tiles_x*(y-1+v)] = 2;
        }
      }

      world << "mineral" << endl;
      world << "<INT_ATTR>" << endl;
      world << "owner " << num_players << endl;
      world << "x " << x * tile_points - 6 + rand.rand_uint4() % 13 << endl;
      world << "y " << y * tile_points - 6 + rand.rand_uint4() % 13 << endl;
      world << "</INT_ATTR>" << endl;
    }
  }
#endif

  // create random geysers
#if 1
  FORS (i, 10) {
    sint4 x0, y0;
    sint4 t = rand.rand_uint4() % tile_n;
    x0 = t % tiles_x;
    y0 = t / tiles_y;
    if (bad[t] & ~4) continue;

    FORS (j, 10) {
      sint4 x = x0 - 2 + rand.rand_uint4() % 5;
      sint4 y = y0 - 2 + rand.rand_uint4() % 5;
      if (x < 1 || x >= tiles_x-1) continue;
      if (y < 1 || y >= tiles_y-1) continue;
      bool can_use = true;
      FORU (u, 3) {
        FORU (v, 3) {
          if (bad[(x-1+u) + tiles_x*(y-1+v)]) can_use = false;
        }
      }
      if (!can_use) continue;
      FORU (u, 3) {
        FORU (v, 3) {
          bad[(x-1+u) + tiles_x*(y-1+v)] = 4;
        }
      }
      world << "geyser" << endl;
      world << "<INT_ATTR>" << endl;
      world << "owner " << num_players << endl;
      world << "x " << x * tile_points - 6 + rand.rand_uint4() % 13 << endl;
      world << "y " << y * tile_points - 6 + rand.rand_uint4() % 13 << endl;
      world << "</INT_ATTR>" << endl;
    }
  }
#endif

  FORU (p, num_players) {

    if (AIR_UNITS) {

      FORS (i, nobjs) {
        sint4 r;
        do { r = rand.rand_uint4() % tile_n; } while (bad[r]);
	bad[r] = 1;

        place_unit(world, "destroyer", p, r % tiles_x, r / tiles_x);
      }
    }

    if (LAND_UNITS) { 

      FORS (i, nobjs) {
        sint4 r;
        do { r = rand.rand_uint4() % tile_n; } while (bad[r]);
	bad[r] = 1;

        place_unit(world, "worker", p, r % tiles_x, r / tiles_x);
      }
    }
  }

  // create random clusters of trees
#if 1
  sint4 tree1 = opt.tree1;
  sint4 tree2 = opt.tree2;

  FORS (i, tree1) {
    sint4 x0,y0;
    sint4 t = rand.rand_uint4() % tile_n;
    x0 = t % tiles_x;
    y0 = t / tiles_y;
    if (bad[t] & ~(1 | 8)) { --i; continue; }

    FORS (j, tree2) {
      sint4 x = x0 - 3 + rand.rand_uint4() % 7;
      sint4 y = y0 - 3 + rand.rand_uint4() % 7;
      if (x < 1 || x >= tiles_x-1) continue;
      if (y < 1 || y >= tiles_y-1) continue;
      bool can_use = true;

      FORU (u, 3) {
        FORU (v, 3) {
          if (bad[(x-1+u) + tiles_x*(y-1+v)] & ~(1 | 8)) can_use = false;
        }
      }
      if (!can_use) continue;

      FORU (u, 3) {
        FORU (v, 3) {
          bad[(x-1+u) + tiles_x*(y-1+v)] = 8;
        }
      }
      world << "tree" << endl;
      world << "<INT_ATTR>" << endl;
      world << "owner " << num_players << endl;
      world << "x " << x * tile_points - 6 + rand.rand_uint4() % 13 << endl;
      world << "y " << y * tile_points - 6 + rand.rand_uint4() % 13 << endl;
      world << "</INT_ATTR>" << endl;
    }
  }
#endif

  world << "</OBJECTS>\n";
}

//===================================================================

