/** @file MapTool.C
    @see MapTool.H

    $Id: TournamentMapTool.C 10761 2011-11-30 00:35:13Z mburo $
    $Source$
*/
// This is an ORTS file (c) Michael Buro, licensed under the GPL

// map generators for the AIIDE competition

// game1: resource gathering (single player, no fow)
//   single base + resource clusters

// game2: combat             (2 player, no fow)
//   small number of bases each

// game3: RTS                (2 player, fow)
//   one base each + resource clusters
//   one cluster close to base
//   minimum distance between bases

// game4: combat             (2 player, no fow)
//   two opposing armies (marines & tanks)

// skirmish: combat (2 players)
// marines vs marines (2007 game 4)

// put in safeguards to prevent hanging
// (loops that try to fit stuff in will give up and create a new map)

#include "Global.H"
#include "MapTool.H"
#include "GameConst.H"
#include "Random.H"
#include "Options.H"

using namespace std;

const sint4 BASE_SIDE      = 7;  // must be odd
const sint4 CLUSTER_SIDE   = 4;  // must be even
const sint4 GAS_CLUSTER_SIDE = 6;
const sint4 RESOURCE_NUM   = 10; // how many resources per cluster
const sint4 GEYSER_X       = 4;
const sint4 GEYSER_Y       = 3;
const sint4 MIN_DIST       = 20; // minimal tile distance between bases

const sint4 GAME1_NUM_CLUSTERS     = 4;

const sint4 GAME2_NUM_BASES        = 5;

const sint4 GAME3_NUM_CLUSTERS_MIN = 3;
const sint4 GAME3_NUM_CLUSTERS_MAX = 5;

const sint4 GAME4_NUM_UNITS = 50;
const sint4 GAME4_NUM_TANKS = 20;

const sint4 SKIRMISH_NUM_UNITS = 50; // marines

//==============================================================================

static void rand_loc(MapData &md, sint4 &x0, sint4 &y0)
{
  sint4 t = md.rand.rand_uint4() % md.tile_n;
  x0 = t % md.tiles_x;
  y0 = t / md.tiles_y;
}


static bool unit_loc_ok(const MapData &md, sint4 x0, sint4 y0, sint4 radius=0)
{
  if (x0 < radius || x0 >= md.tiles_x-radius) return false;
  if (y0 < radius || y0 >= md.tiles_y-radius) return false;

  for (sint4 y = y0-radius; y <= y0+radius; ++y) {
    for (sint4 x = x0-radius; x <= x0+radius; ++x) {
      sint4 i = y * md.tiles_x + x;
      if (md.bad[i]) return false;
    }
  }

  return true;
}

static void mark_unit_loc(MapData &md, sint4 x, sint4 y, sint4 radius=0)
{
  sint4 x0 = max(0, x - radius);
  sint4 x1 = min(md.tiles_x-1, x + radius);
  sint4 y0 = max(0, y - radius);
  sint4 y1 = min(md.tiles_y-1, y + radius);

  for (sint4 y = y0; y <= y1; ++y) {
    for (sint4 x = x0; x <= x1; ++x) {
      sint4 i = y * md.tiles_x + x;
      md.bad[i] = 1;
    }
  }
}

static bool resource_loc_ok(const MapData &md, sint4 &x0, sint4 &y0, sint4 cluster_side = CLUSTER_SIDE)
{
  if (x0 < cluster_side/2 || x0 > md.tiles_x-cluster_side/2 ||
      y0 < cluster_side/2 || y0 > md.tiles_y-cluster_side/2)
    return false;
  
  for (sint4 u = -cluster_side/2; u < cluster_side/2; ++u)
    for (sint4 v = -cluster_side/2; v < cluster_side/2; ++v)
      if (md.bad[(x0+u) + md.tiles_x*(y0+v)]) return false;

  return true;
}


// places a gas well and mineral resources on a single square area
// assumes: resorces fit on one tile
void populate_resource_loc_gas(MapData &md, sint4 x0, sint4 y0, 
				      stringstream &world)
{
  uint4 gas_on_top = md.rand.rand_uint4() % 2;
  //if gas_on_top = 1, then geyser will be above minerals
  //if gas_on_top = 0, then geyser will be below minerals
  
  FORS (j, RESOURCE_NUM) {
      
    sint4 x = x0 - GAS_CLUSTER_SIDE/2 + md.rand.rand_uint4() % GAS_CLUSTER_SIDE;
    sint4 y = y0 - GAS_CLUSTER_SIDE/2 + gas_on_top * GEYSER_Y + md.rand.rand_uint4() % (GAS_CLUSTER_SIDE-GEYSER_Y);
    
    if (x < 0 || x > md.tiles_x-1) { --j; continue; }
    if (y < 0 || y > md.tiles_y-1) { --j; continue; }
    if (md.bad[x + md.tiles_x*y])  { --j; continue; }

    sint4 tile_points = GameConst::TILE_POINTS;
    world << "mineral" << endl;
    world << "<INT_ATTR>" << endl;
    world << "owner " << md.num_players << endl;
    world << "x " << x*tile_points+tile_points/2 - 2 + md.rand.rand_uint4() % 5 << endl;
    world << "y " << y*tile_points+tile_points/2 - 2 + md.rand.rand_uint4() % 5 << endl;
    world << "</INT_ATTR>" << endl;

    // mark region occupied

    md.bad[x + md.tiles_x*y] = 1;
  }
  
  sint4 x = x0; 
  sint4 y = y0 + GAS_CLUSTER_SIDE/2 - gas_on_top * GEYSER_Y - (GEYSER_Y+1) / 2;
  
  //these checks may be too paranoid
  assert(x >= 0 && x < md.tiles_x);
  assert(y >= 0 || y < md.tiles_y);
  assert(!md.bad[x + md.tiles_x*y]);

  sint4 tile_points = GameConst::TILE_POINTS;
  world << "geyser" << endl;
  world << "<INT_ATTR>" << endl;
  world << "owner " << md.num_players << endl;
  world << "x " << x*tile_points+tile_points/2 /* - 2 + md.rand.rand_uint4() % 5*/ << endl;
  world << "y " << y*tile_points+tile_points/2 /* - 2 + md.rand.rand_uint4() % 5*/ << endl;
  world << "</INT_ATTR>" << endl;
  
  // mark region occupied
  sint4 min_x = x -GEYSER_X /2;
  sint4 min_y = y - GEYSER_Y /2;
  for (sint4 i = min_x; i < min_x + GEYSER_X; i++){
    for(sint4 j = min_y; j < min_y + GEYSER_Y ; j++){
      md.bad[i + md.tiles_x*j] = 1;
    }
  }
}


static void populate_gas_loc(MapData &md, sint4 x0, sint4 y0,
			     stringstream &world){
  sint4 tile_points = GameConst::TILE_POINTS;
  world << "geyser" << endl;
  world << "<INT_ATTR>" << endl;
  world << "owner " << md.num_players << endl;
  world << "x " << x0*tile_points+tile_points/2 /* - 2 + md.rand.rand_uint4() % 5*/ << endl;
  world << "y " << y0*tile_points+tile_points/2 /* - 2 + md.rand.rand_uint4() % 5*/ << endl;
  world << "</INT_ATTR>" << endl;
  
  // mark region occupied
  sint4 min_x = x0 -GEYSER_X /2;
  sint4 min_y = y0 - GEYSER_Y /2;
  for (sint4 i = min_x; i < min_x + GEYSER_X; i++){
    for(sint4 j = min_y; j < min_y + GEYSER_Y ; j++){
      md.bad[i + md.tiles_x*j] = 1;
    }
  }

}

static void populate_resource_loc(MapData &md, sint4 x0, sint4 y0,
				  stringstream &world)
{
  FORS (j, RESOURCE_NUM) {
    
    sint4 x = x0 - CLUSTER_SIDE/2 + md.rand.rand_uint4() % CLUSTER_SIDE;
    sint4 y = y0 - CLUSTER_SIDE/2 + md.rand.rand_uint4() % CLUSTER_SIDE;

    if (x < 0 || x > md.tiles_x-1) { --j; continue; }
    if (y < 0 || y > md.tiles_y-1) { --j; continue; }
    if (md.bad[x + md.tiles_x*y])  { --j; continue; }

    sint4 tile_points = GameConst::TILE_POINTS;
    world << "mineral" << endl;
    world << "<INT_ATTR>" << endl;
    world << "owner " << md.num_players << endl;
    world << "x " << x*tile_points+tile_points/2 - 2 + md.rand.rand_uint4() % 5 << endl;
    world << "y " << y*tile_points+tile_points/2 - 2 + md.rand.rand_uint4() % 5 << endl;
    world << "</INT_ATTR>" << endl;

    // mark region occupied

    md.bad[x + md.tiles_x*y] = 1;
  }
}


static bool base_loc_ok(const MapData &md, sint4 x0, sint4 y0)
{
  // cout << "bounds " << BASE_SIDE/2 << " " << (md.tiles_x-BASE_SIDE/2) << endl;
  
  if (x0 < BASE_SIDE/2 || x0 >= md.tiles_x-BASE_SIDE/2) return false;
  if (y0 < BASE_SIDE/2 || y0 >= md.tiles_y-BASE_SIDE/2) return false;

  for (sint4 y = y0-BASE_SIDE/2; y <= y0+BASE_SIDE/2; ++y) {
    for (sint4 x = x0-BASE_SIDE/2; x <= x0+BASE_SIDE/2; ++x) {
      sint4 i = y * md.tiles_x + x;
      if (md.bad[i]) return false;
    }
  }

  return true;
}


static void mark_base_loc(MapData &md, sint4 x0, sint4 y0)
{
  for (sint4 y = y0-BASE_SIDE/2; y <= y0+BASE_SIDE/2; ++y) {
    for (sint4 x = x0-BASE_SIDE/2; x <= x0+BASE_SIDE/2; ++x) {
      if (x >= 0 && x < md.tiles_x && y >= 0 && y < md.tiles_y) {
	sint4 i = y * md.tiles_x + x;
	md.bad[i] = 1;
      }
    }
  }
}

void MapTool::init_map_data(MapData &md)
{
  opt.get();

  sint4 seed = opt.seed;
  md.rand.srand(seed);
  md.bad.clear();

  md.gtiles_x = opt.width;
  md.gtiles_y = opt.height;
  md.num_players = opt.nplayers; 
  md.tiles_x = md.gtiles_x;
  md.tiles_y = md.gtiles_y;
  md.tile_n  = md.tiles_x * md.tiles_y;
  md.z_sight = 0;
  
  sint4 len = opt.player_races.length();
  for (sint4 i = 0; i < md.num_players; i++){
    if (i >= len) {
      md.player_races.push_back(MapData::TERRAN);
    }
    else{
      char race = opt.player_races.at(i);
      if (race == 'B') md.player_races.push_back(MapData::BUG);
      else if (race == 'P') md.player_races.push_back(MapData::PANTHEON);
      else md.player_races.push_back(MapData::TERRAN);
    }
  }
}



#if 0
// (tiles_x+1)*(tiles_y+1) corner height index to
// (tiles_x)*(tiles_y) tile index

static sint4 height2tile(const MapData &md, sint4 index)
{
  sint4 x = index % (md.tiles_x+1); if (x >= md.tiles_x) x = md.tiles_x-1;
  sint4 y = index / (md.tiles_y+1); if (y >= md.tiles_y) y = md.tiles_y-1;
  return y*md.tiles_x + x;
}
#endif


// true iff tiles adjacent to corner are not bad

static bool corner_clear(const MapData &md, sint4 index)
{
  sint4 x0 = index % (md.tiles_x+1);
  sint4 y0 = index / (md.tiles_y+1);

  // (x0,y0) : corner coordinates
  
  for (sint4 dy = -1; dy <= 0; ++dy) {
    for (sint4 dx = -1; dx <= 0; ++dx) {    
      sint4 x = x0 + dx;
      sint4 y = y0 + dy;

      // (x,y) tile coordinate to be checked
      
      if (x < 0) x = 0; else if (x >= md.tiles_x) x = md.tiles_x-1;
      if (y < 0) y = 0; else if (y >= md.tiles_y) y = md.tiles_y-1;
      if (md.bad[y*md.tiles_x+x]) return false;
    }
  }

  return true;
}


// adds tiles to world string
// obeys md.bad[] (true <=> tile corner heights = 0)
// updates md.bad[]

void MapTool::generate_game_tiles(MapData &md, stringstream &world, bool bad_init=true)
{
  world << md.num_players << endl;
  world << md.gtiles_x << endl;
  world << md.gtiles_y << endl;
  world << md.z_sight << endl;
  world << 0 << endl; // no rotation

  Map<Tile> map_data;
  map_data.allocate(md.tiles_x, md.tiles_y);
 
  sint4 h_x = md.tiles_x + 1;
  sint4 h_y = md.tiles_y + 1;  
  sint4 h_n = h_x * h_y;

  Vector<sint2> height(h_n); // height field (index is not a tile index!)

  if (bad_init) {
    md.bad.clear();
    md.bad.reserve(md.tile_n);
    FORS (i, md.tile_n) md.bad.push_back(0);
  }

#if 0  
  FORS (y, md.tiles_y) {
    FORS (x, md.tiles_x) {
      cout << md.bad[y*md.tiles_x+x];
    }
    cout << endl;
  }
#endif
  
  // otherwise, use bad[] that has been passed and do not create
  // heights > 0 at corners of tiles with md.bad[] != 0
  
  // first heights represent layers
  // (will be multiplied by GameConst::HEIGHT_MULT later)

  real8 fplat      = opt.fplat;
  bool  plug_holes = opt.holes;
  sint4 clump      = opt.clump;
  bool  smooth     = opt.smooth;

  if (smooth) plug_holes = false; // smooth > plug_holes

  if (clump < 1) clump = 1;

  if (clump <= 1) {

    FORS (i, sint4(md.tile_n * fplat)) {
      sint4 r;
      do {
	r = md.rand.rand_uint4() % h_n;
      } while (height[r] || !corner_clear(md, r));
      height[r] = 1;
    }

  } else {

    FORS (i, sint4(md.tile_n * fplat)) {
      sint4 r;
      do {
	r = md.rand.rand_uint4() % h_n;
      } while (height[r] || !corner_clear(md, r));
    
      sint4 next_to = 0;
      sint4 out_of = 4;
    
      sint4 x = r % h_x;
      sint4 y = r / h_x;
  
      if (x == 0 || x == h_x-1) out_of--;
      if (y == 0 || y == h_y-1) out_of--;
    
      sint4 h = height[r];

      if (x > 0     && height[r-1]   == h+1) next_to++;
      if (x < h_x-1 && height[r+1]   == h+1) next_to++;
      if (y > 0     && height[r-h_x] == h+1) next_to++;
      if (y < h_y-1 && height[r+h_x] == h+1) next_to++;

      const sint4 offset = 4;

      if (out_of == next_to) {
	height[r] += 1;
        continue;
      }
      if (next_to * 2 >= out_of) {
        next_to += offset;
        out_of += offset;
      }
      if (next_to > 0) {
        if (static_cast<sint4>(md.rand.rand_uint4() % out_of) < next_to) {
	  height[r] += 1;
	  continue;
	}
      }
      if (md.rand.rand_uint4() % clump < 1) {
        height[r]++; // = 1;
        continue;
      }
      --i;
    }
  }

  if (plug_holes) {

    FORS (i, h_n) {
      sint4 r = i;

      sint4 next_to = 0;
      sint4 out_of = 4;

      sint4 rx = r % h_x;
      sint4 ry = r / h_x;

      sint4 h = height[r];
      
      if (rx == 0 || rx == h_x-1) out_of--;
      if (ry == 0 || ry == h_y-1) out_of--;

      if (rx > 0     && height[r-1]   > h) next_to++;
      if (rx < h_x-1 && height[r+1]   > h) next_to++;
      if (ry > 0     && height[r-h_x] > h) next_to++;
      if (ry < h_y-1 && height[r+h_x] > h) next_to++;

      if (out_of == next_to) {// && height[r] == 0) {
        //	cout << i << " " << next_to << " " << out_of << " plug" << endl;
	if (corner_clear(md, r)) {
	  height[r]++;// = 1;
	}
      }
    }
  }

  // smooth plateaus:
  //  continue filling tiles iff it has three or four filled HV neighbors

  bool changed = smooth;

  while (changed) {

    changed = false;
    
    sint4 dx[] = { 1,  0, -1, 0 };
    sint4 dy[] = { 0, -1,  0, 1 };

    FORS (i, h_n) {

      sint4 h = height[i];

      sint4 ix = i % h_x;
      sint4 iy = i / h_x;

      sint4 next_to = 0;
      
      FORS (j, 4) {
	sint4 jx = ix + dx[j];
	sint4 jy = iy + dy[j];	  
	if (jx < 0 || jx >= h_x || jy < 0 || jy >= h_y) next_to++;
	else if (height[jy*h_x + jx] > h) next_to++;
      }
      
      if (next_to >= 3 && corner_clear(md, i)) {
	height[i]++; changed = true;
      }
    }
  }

  FORS (i, h_n) height[i] *= GameConst::HEIGHT_MULT;

  // from here on: heights are multiples of GameConst::HEIGHT_MULT
  // or at half-heights (ramps)

  // create tiles from height map
  
  FORS (i, h_n) {

    sint4 x = i % h_x;
    sint4 y = i / h_x;

    if (x == h_x-1 || y == h_y-1) continue;

    //    cout << height[i] << " ";
    
    sint4 h[4];
    
    h[0] = height[i];
    h[1] = height[i+1];
    h[2] = height[i+h_x+1];
    h[3] = height[i+h_x];

    Tile &tile = map_data.get_tile(x, y);

    tile.set_topo(h[0], h[1], h[2], h[3], Tile::NO_SPLIT);
    tile.set_type(GROUND);

    //    assert(x >= 0 && x < tiles_x);
    //    assert(y >= 0 && y < tiles_y);

    if (h[0] != 0 || h[1] != 0 || h[2] != 0 || h[3] != 0) {
      md.bad[y*md.tiles_x+x] = 1;
    }
    
    // equal heights -> regular ground tile
    
    if (h[0] == h[1] && h[0] == h[2] && h[0] == h[3]) {

      tile.set_topo(h[0], h[1], h[2], h[3], Tile::NO_SPLIT);
      tile.set_type(GROUND);
      continue;
    }

    // 3 equal heights -> split
    //  - make equal height GROUND triangle and the other CLIFF or
    //  - create ridge
    
    {
      if ((h[0] == h[1] && h[1] == h[2]) ||
          (h[0] == h[1] && h[1] == h[3]) ||
          (h[0] == h[2] && h[2] == h[3]) ||
          (h[1] == h[2] && h[2] == h[3])) {

        sint4 i;
        FOR (i, 4) {
          if (h[i] != h[(i+1)&3] && h[i] != h[(i+2)&3]) break;
        }

	if (1 || (md.rand.rand_uint4() & 8)) {

	  // equal height triangle
	  
	  if (i == 0 || i == 2) {  // |/|
	    tile.set_topo(h[0], h[1], h[2], h[3], Tile::BT_SPLIT);
	  } else { // |\|
	    tile.set_topo(h[0], h[1], h[2], h[3], Tile::TB_SPLIT);
	  }
      
	  if (i == 0 || i == 3) tile.set_type(CLIFF, GROUND); // west CLIFF
	  else                  tile.set_type(GROUND, CLIFF); // east CLIFF

	} else {

	  // ridge

	  if (i == 0 || i == 2) {  // |\|
	    tile.set_topo(h[0], h[1], h[2], h[3], Tile::TB_SPLIT);
	  } else { // |/|
	    tile.set_topo(h[0], h[1], h[2], h[3], Tile::BT_SPLIT);
	  }

	  tile.set_type(CLIFF, CLIFF);
	}
        //	assert(x >= 0 && x < tiles_x);
        //	assert(y >= 0 && y < tiles_y);
	continue;
      }
    }

    
    // 2 and 2 equal heights -> full CLIFF tile
    
    {
      sint4 i;
      FOR (i, 4) {
	if (h[i] == h[(i+1) & 3] && h[(i+2) & 3] == h[(i+3) & 3]) break;
      }
    
      if (i < 4) {

	tile.set_topo(h[0], h[1], h[2], h[3], Tile::NO_SPLIT);

	// small chance of ramp
	if (md.rand.rand_uint4() % 100 < 15) tile.set_type(CLIFF);//no Ramps for now; was: (GROUND);
	else                              tile.set_type(CLIFF);

        //	assert(x >= 0 && x < tiles_x);
        //	assert(y >= 0 && y < tiles_y);
	continue;
      }
    }

    // otherwise: random CLIFF/CLIFF split

    {
      tile.set_type(CLIFF, CLIFF);
   
      if (md.rand.rand_uint4() & 8) {
	tile.set_topo(h[0], h[1], h[2], h[3], Tile::BT_SPLIT);
      } else {
	tile.set_topo(h[0], h[1], h[2], h[3], Tile::TB_SPLIT);
      }

      //      assert(x >= 0 && x < tiles_x);
      //      assert(y >= 0 && y < tiles_y);
    }
  }

  FORS (i, md.tile_n) {

    sint4 x = i % md.tiles_x;
    sint4 y = i / md.tiles_x;

    Tile &t = map_data.get_tile(i);
    Tile::Type tw, te;
    Tile::Split split;
    t.get_type(tw, te);
    sint4 h0, h1, h2, h3;
    t.get_topo(h0, h1, h2, h3, split);

    if (split == Tile::NO_SPLIT && tw != CLIFF && (h0 != h3 || h0 != h1)) {

      // ramp

      ERR("should not get here"); // no ramps!
      
      if (h0 != h3) {

	// | ramp

	if (x > 0) {  // extend left
	  Tile &t = map_data.get_tile(x-1, y);
	  t.get_type(tw, te);
	  t.get_topo(h0, h1, h2, h3, split);

	  if (md.rand.rand_uint4() % 100 < 50 && te == CLIFF) {
	    if (split) t.set_type(tw, GROUND);
	    else       t.set_type(GROUND, GROUND);
	  }
	}
	
	if (x < md.tiles_x-1) {  // extend right
	  Tile &t = map_data.get_tile(x+1, y);
	  t.get_type(tw, te);
	  t.get_topo(h0, h1, h2, h3, split);

	  if (md.rand.rand_uint4() % 100 < 50 && tw == CLIFF) {
	    if (split) t.set_type(GROUND, te);
	    else       t.set_type(GROUND, GROUND);
	  }
	}

      } else {

	// - ramp

 	if (y > 0) {  // extend up
	  Tile &t = map_data.get_tile(x, y-1);
	  t.get_type(tw, te);
	  t.get_topo(h0, h1, h2, h3, split);

	  if (md.rand.rand_uint4() % 100 < 50 &&
	      ((!split && tw == CLIFF) ||                    // | |
	       (split == Tile::TB_SPLIT && tw == CLIFF) ||   // |\|
	       (split == Tile::BT_SPLIT && te == CLIFF))) {  // |/|

	    if   (!split)                     t.set_type(GROUND, GROUND);
	    else if (split == Tile::TB_SPLIT) t.set_type(GROUND, te);
	    else                              t.set_type(tw, GROUND);
	  }
	}
	
	if (y < md.tiles_y-1) {  // extend down
	  Tile &t = map_data.get_tile(x, y+1);
	  t.get_type(tw, te);
	  t.get_topo(h0, h1, h2, h3, split);

	  if (md.rand.rand_uint4() % 100 < 50 &&
	      ((!split && tw == CLIFF) ||                    // | |
	       (split == Tile::TB_SPLIT && te == CLIFF) ||   // |\|
	       (split == Tile::BT_SPLIT && tw == CLIFF))) {  // |/|

	    if   (!split   )                  t.set_type(GROUND, GROUND);
	    else if (split == Tile::TB_SPLIT) t.set_type(tw, GROUND);
	    else                              t.set_type(GROUND, te);
	  }
	}
      }
    }
  }
  
  world << "<TILES>" << endl;
  map_data.write(world);
  world << "</TILES>" << endl;
}


void MapTool::generate_game1_map(stringstream &output_world)
{
  MapData md;

  init_map_data(md);

 try_again:;

  REM("generate world");
  
  stringstream world;

  generate_game_tiles(md, world);

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  sint4 loops = 0;

  FOREVER {

    // create start locations

    FORU (p, md.num_players) {
      
      // retry if hung
      if (++loops >= 1000) goto try_again;
      
      sint4 x0, y0;
      rand_loc(md, x0, y0);
      if (!base_loc_ok(md, x0, y0)) { --p; continue; }
    
      mark_base_loc(md, x0, y0);
      place_unit(world, "start_loc", p, x0, y0);
    }

    // create random resource clusters
  
    FORS (i, GAME1_NUM_CLUSTERS) {

      // retry if hung
      if (++loops >= 1000) goto try_again;

      sint4 x0, y0;

      rand_loc(md, x0, y0);
      if (!resource_loc_ok(md, x0, y0)) { --i; continue; }
      populate_resource_loc(md, x0, y0, world);
    }

    break;
  }
  world << "</OBJECTS>\n";

  output_world << world.str();
}

//==============================================================================

#if 0

// old code: arbitrary base placement => often unfair

void MapTool::generate_game2_map(stringstream &output_world)
{
  MapData md;

  init_map_data(md);

 try_again:;

  REM("generate world");
  
  stringstream world;

  generate_game_tiles(md, world);

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  sint4 loops = 0;

  FOREVER {
  
    // create start locations

    FORS (i, GAME2_NUM_BASES) {

      FORU (p, md.num_players) {

	// retry if hung
	if (++loops >= 1000) goto try_again;
      
	sint4 x0, y0;

	rand_loc(md, x0, y0);
	if (!base_loc_ok(md, x0, y0)) { --p; continue; }
    
	mark_base_loc(md, x0, y0);
	place_unit(world, "start_loc", p, x0, y0);
      }
    }

    // create random clusters of trees
#if 0
    sint4 tree1 = opt.tree1;
    sint4 tree2 = opt.tree2;

    FORS (i, tree1) {
      sint4 x0,y0;
      sint4 t = rand.rand_uint4() % tile_n;
      x0 = t % tiles_x;
      y0 = t / tiles_y;
      if (bad[t] & ~(1 | 8)) { --i; continue; }

      FORS (j, tree2) {
	sint4 x = x0 - 3 + rand.rand_uint4() % 7;
	sint4 y = y0 - 3 + rand.rand_uint4() % 7;
	if (x < 1 || x >= tiles_x-1) continue;
	if (y < 1 || y >= tiles_y-1) continue;
	bool can_use = true;

	FORU (u, 3) {
	  FORU (v, 3) {
	    if (bad[(x-1+u) + tiles_x*(y-1+v)] & ~(1 | 8)) can_use = false;
	  }
	}
	if (!can_use) continue;

	FORU (u, 3) {
	  FORU (v, 3) {
	    bad[(x-1+u) + tiles_x*(y-1+v)] = 8;
	  }
	}
	world << "tree" << endl;
	world << "<INT_ATTR>" << endl;
	world << "owner " << num_players << endl;
	world << "x " << x * tile_points - 6 + rand.rand_uint4() % 13 << endl;
	world << "y " << y * tile_points - 6 + rand.rand_uint4() % 13 << endl;
	world << "</INT_ATTR>" << endl;
      }
    }
#endif
    break;
  }

  world << "</OBJECTS>\n";

  output_world << world.str();
}

#endif


struct StartLoc {
 sint4 player, x, y;
    
 StartLoc(sint4 player_, sint4 x_, sint4 y_) :
   player(player_), x(x_), y(y_) {}
};

REGISTER_TYPEOF(10101, Vector<StartLoc>::const_iterator );
REGISTER_TYPEOF(10102, Vector<StartLoc>::iterator );

// center symmetric bases => more fair

void MapTool::generate_game2_map(stringstream &output_world)
{
  MapData md;
  stringstream world;

  init_map_data(md);

  assert(md.num_players <= 2);

 try_again:;

  REM("generate symmetric bases");  
  
  md.bad.clear();
  md.bad.reserve(md.tile_n);
  FORS (i, md.tile_n) md.bad.push_back(0);

  Vector<StartLoc> slocs;
  sint4 loops = 0;
  
  // create center symmetric start locations

  FORS (i, GAME2_NUM_BASES) {
    
    sint4 x0, y0;

    FOREVER {
      rand_loc(md, x0, y0);
      if (base_loc_ok(md, x0, y0)) break;
      if (++loops >= 1000) goto try_again; // should not happen
    }

    mark_base_loc(md, x0, y0);
    slocs.push_back(StartLoc(0, x0, y0));

    if (md.num_players > 0) {

      // create symmetric location

      x0 = md.tiles_x - 1 - x0;
      y0 = md.tiles_y - 1 - y0;

      if (!base_loc_ok(md, x0, y0)) goto try_again; // center overlap
      
      mark_base_loc(md, x0, y0);
      slocs.push_back(StartLoc(1, x0, y0));
    }
  }
  cout<<"Creating game 2" << endl;

  generate_game_tiles(md, world, false);

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  FORALL (slocs, i) {
    place_unit(world, "start_loc", (*i).player, (*i).x, (*i).y);
  }

  output_world << world.str();
}


//==============================================================================

void MapTool::generate_game3_map(stringstream &output_world)
{
  MapData md;

  init_map_data(md);

 try_again:;

  REM("generate world");
  
  stringstream world;

  generate_game_tiles(md, world);

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  // create start locations with nearby resource cluster

  sint4 loops = 0;
    
  FOREVER {

    Vector<sint4> xb, yb, xr, yr; // base and resource location next to it
    
    FORU (p, md.num_players) {

      // retry if hung
      if (++loops >= 1000) goto try_again;
    
      // look for empty base+resource location
      
      sint4 x0, y0;
      rand_loc(md, x0, y0);
      if (!base_loc_ok(md, x0, y0)) { --p; continue; }

      // points to check
      
      Point res_locs[4] = {
	{ x0+(BASE_SIDE/2+CLUSTER_SIDE/2), y0 },
	{ x0-(BASE_SIDE/2+CLUSTER_SIDE/2), y0 },
	{ x0, y0+(BASE_SIDE/2+CLUSTER_SIDE/2) },
	{ x0, y0-(BASE_SIDE/2+CLUSTER_SIDE/2) }
      };

      random_shuffle(res_locs, res_locs+4, md.rand);

      sint4 i;
	  {
      FOR (i, 4) {
	sint4 x = res_locs[i].x;
	sint4 y = res_locs[i].y;

	if (resource_loc_ok(md, x, y)) {
	  xb.push_back(x0);
	  yb.push_back(y0);
	  xr.push_back(x);
	  yr.push_back(y);
	  break;
	}
      }
	  }

      if (i >= 4) { --p; continue; }
    }

    // ensure minimal base distances

    bool ok = true;

    FORS (i, md.num_players) {

      int j;
      for (j=i+1; j < md.num_players; ++j) {
	if (sqrt(static_cast<real8>(square(xb[i]-xb[j]) + square(yb[i]-yb[j]))) < MIN_DIST) {
	  ok = false;
	  break;
	}
      }
      if (!ok) break;
    }

    if (ok) {

      FORS (p, md.num_players) {

	cout << xb[p] << " " << yb[p] << " " << xr[p] << " " << yr[p] << endl;
	mark_base_loc(md, xb[p], yb[p]);
	place_unit(world, "start_loc", p, xb[p], yb[p]);
	populate_resource_loc(md, xr[p], yr[p], world);
      }
      
      break;
    }
  }
  
  // create additional random resource clusters

  sint4 n_cluster =
    GAME3_NUM_CLUSTERS_MIN +
    md.rand.rand_uint4() % (GAME3_NUM_CLUSTERS_MAX-GAME3_NUM_CLUSTERS_MIN+1);

  FORS (i, n_cluster) {

    sint4 x0, y0;

    rand_loc(md, x0, y0);
    if (!resource_loc_ok(md, x0, y0)) { --i; continue; }
    populate_resource_loc(md, x0, y0, world);
  }
  
  world << "</OBJECTS>\n";

  output_world << world.str();
}

//==============================================================================

void MapTool::generate_game4_map(stringstream &output_world)
{
  MapData md;
  stringstream world;

  init_map_data(md);

  assert(md.num_players <= 2);

  REM("generate symmetric bases");  
  
  md.bad.clear();
  md.bad.reserve(md.tile_n);
  FORS (i, md.tile_n) md.bad.push_back(0);

  Vector<StartLoc> slocs;
  Vector<StartLoc> slocs2;
  //  sint4 loops = 0;
  
  // create center symmetric start locations

  FORS (i, GAME4_NUM_UNITS) {
    
    sint4 x0, y0;

    FOREVER {
      x0 = md.rand.rand_uint4() % (md.tiles_x / 4);
      y0 = md.rand.rand_uint4() % md.tiles_y;
      if (unit_loc_ok(md, x0, y0)) break;
    }
    
    mark_unit_loc(md, x0, y0);

    slocs.push_back(StartLoc(0, x0, y0));
    x0 = md.tiles_x - 1 - x0;
    y0 = md.tiles_y - 1 - y0;
    slocs.push_back(StartLoc(1, x0, y0));
  }

  FORS (i, GAME4_NUM_TANKS) {
    
    sint4 x0, y0;

    FOREVER {
      x0 = md.rand.rand_uint4() % (md.tiles_x / 4);
      y0 = md.rand.rand_uint4() % md.tiles_y;
      if (unit_loc_ok(md, x0, y0)) break;
    }
    
    mark_unit_loc(md, x0, y0);

    slocs2.push_back(StartLoc(0, x0, y0));
    x0 = md.tiles_x - 1 - x0;
    y0 = md.tiles_y - 1 - y0;
    slocs2.push_back(StartLoc(1, x0, y0));
  }

  cout<<"Creating game 4" << endl;

  generate_game_tiles(md, world, false);

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  FORALL (slocs, i) {
    place_unit(world, "marine", (*i).player, (*i).x, (*i).y);
  }
  FORALL (slocs2, i) {
    place_unit(world, "tank_mk2", (*i).player, (*i).x, (*i).y);
  }

  output_world << world.str();
}

//==============================================================================

void MapTool::generate_skirmish_map(stringstream &output_world)
{
  MapData md;
  stringstream world;

  init_map_data(md);

  assert(md.num_players <= 2);

  REM("generate symmetric bases");  
  
  md.bad.clear();
  md.bad.reserve(md.tile_n);
  FORS (i, md.tile_n) md.bad.push_back(0);

  Vector<StartLoc> slocs;
  Vector<StartLoc> slocs2;
  //  sint4 loops = 0;
  
  // create center symmetric start locations

  FORS (i, GAME4_NUM_UNITS) {
    
    sint4 x0, y0;

    FOREVER {
      x0 = md.rand.rand_uint4() % (md.tiles_x / 4);
      y0 = md.rand.rand_uint4() % md.tiles_y;
      if (unit_loc_ok(md, x0, y0)) break;
    }
    
    mark_unit_loc(md, x0, y0);

    slocs.push_back(StartLoc(0, x0, y0));
    x0 = md.tiles_x - 1 - x0;
    y0 = md.tiles_y - 1 - y0;
    slocs.push_back(StartLoc(1, x0, y0));
  }

  FORS (i, GAME4_NUM_TANKS) {
    
    sint4 x0, y0;

    FOREVER {
      x0 = md.rand.rand_uint4() % (md.tiles_x / 4);
      y0 = md.rand.rand_uint4() % md.tiles_y;
      if (unit_loc_ok(md, x0, y0)) break;
    }
    
    mark_unit_loc(md, x0, y0);

    slocs2.push_back(StartLoc(0, x0, y0));
    x0 = md.tiles_x - 1 - x0;
    y0 = md.tiles_y - 1 - y0;
    slocs2.push_back(StartLoc(1, x0, y0));
  }

  cout<<"Creating game 4" << endl;

  generate_game_tiles(md, world, false);

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  FORALL (slocs, i) {
    place_unit(world, "marine", (*i).player, (*i).x, (*i).y);
  }

  output_world << world.str();
}

//==============================================================================
void MapTool::generate_std_map(stringstream &output_world){
  MapData md;

  init_map_data(md);
  sint4 outer_loops = 0;
 try_again:;
  if (outer_loops++ > 20){
    ERR("Unable to Generate map!\n");
  }
  REM("generate world");
  
  stringstream world;

  generate_game_tiles(md, world);

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  // create start locations with nearby resource cluster
  
  sint4 loops = 0;
  sint4 n_cluster =
    GAME3_NUM_CLUSTERS_MIN +
    md.rand.rand_uint4() % (GAME3_NUM_CLUSTERS_MAX-GAME3_NUM_CLUSTERS_MIN+1);

    
  FOREVER {

    Vector<sint4> xb, yb, xr, yr, xg, yg; // base, resource, and gas locations
   
    FORU (p, md.num_players + n_cluster) {

      // retry if hung
      if (++loops >= 1000) goto try_again;
    
      // look for empty base+resource location
      
      sint4 x0, y0;
      rand_loc(md, x0, y0);
      if (!base_loc_ok(md, x0, y0)) { --p; continue; }

      // points to check
      
      Point res_locs[4] = {
	{ x0+(BASE_SIDE/2+CLUSTER_SIDE/2), y0 },
	{ x0-(BASE_SIDE/2+CLUSTER_SIDE/2), y0 },
	{ x0, y0+(BASE_SIDE/2+CLUSTER_SIDE/2) },
	{ x0, y0-(BASE_SIDE/2+CLUSTER_SIDE/2) }
      };

      random_shuffle(res_locs, res_locs+4, md.rand);

      sint4 i = 0;
      {
	sint4 temp_xg , temp_yg;
	temp_xg = temp_yg = -1; //to prevent warning;
	
	for (;i < 3; i++){
	  sint4 x = res_locs[i].x;
	  sint4 y = res_locs[i].y;
	  if (resource_loc_ok(md, x, y, CLUSTER_SIDE)){
	    temp_xg = x;
	    temp_yg = y;
	    break;
	  }
	}
	i++;
	if (i >= 4) {--p; continue;}
	
	for (;i < 4; i++){
	  sint4 x = res_locs[i].x;
	  sint4 y = res_locs[i].y;
	  
	  if (resource_loc_ok(md, x, y, CLUSTER_SIDE)) {
	    xb.push_back(x0);
	    yb.push_back(y0);
	    xr.push_back(x);
	    yr.push_back(y);
	    xg.push_back(temp_xg);
	    yg.push_back(temp_yg);
	    break;
	  }
	}
      }

      if (i >= 4) { --p; continue; }
    }

    // ensure minimal base distances

    bool ok = true;

    FORS (i, md.num_players) {

      int j;
      for (j=i+1; j < md.num_players; ++j) {
	if (sqrt(static_cast<real8>(square(xb[i]-xb[j]) + square(yb[i]-yb[j]))) < MIN_DIST) {
	  ok = false;
	  break;
	}
      }
      if (!ok) break;
    }

    if (ok) {

      FORS (p, md.num_players) {
        

	cout << xb[p] << " " << yb[p] << " " << xr[p] << " " << yr[p] << endl;
	mark_base_loc(md, xb[p], yb[p]);
        switch (md.player_races[p]){
	case MapData::TERRAN:
	  place_unit(world, "terran_start_loc", p, xb[p], yb[p]);
	  break;
	case MapData::BUG:
	  place_unit(world, "bug_start_loc", p, xb[p], yb[p]);
	  break;
	case MapData::PANTHEON:
	  place_unit(world, "pantheon_start_loc", p, xb[p], yb[p]);
	  break;
        default:
	  ERR("Race not recognized");
	}
	//place_unit(world, "start_loc", p, xb[p], yb[p]);
	populate_resource_loc(md, xr[p], yr[p], world);
	populate_gas_loc(md, xg[p], yg[p], world);
      }
      
      for (int i = md.num_players; i < md.num_players + n_cluster; i++){
	mark_base_loc(md, xb[i], yb[i]);
	populate_resource_loc(md, xr[i], yr[i], world);
	populate_gas_loc(md, xg[i], yg[i], world);
      }
   

      break;
    }
    
    
  }
  


//   FORS (i, n_cluster) {

//     sint4 x0, y0;

//     rand_loc(md, x0, y0);
//     if (!resource_loc_ok(md, x0, y0, GAS_CLUSTER_SIDE)) { --i; continue; }
//     populate_resource_loc_gas(md, x0, y0, world);
//   }
  
  world << "</OBJECTS>\n";

  output_world << world.str();
}


void MapTool::generate_tastar_debug_map(stringstream &output_world){
  MapData md;

  init_map_data(md);

 try_again:;

  REM("generate world");
  
  stringstream world;

  generate_game_tiles(md, world);

  string fname = opt.bp_file;
  fname = sanitize_path(fname);
  parse_map_include(world, fname);

  world << "<OBJECTS>\n";

  sint4 loops = 0;

  FOREVER {

    // create start locations

    FORU (p, md.num_players) {
      
      // retry if hung
      if (++loops >= 1000) goto try_again;
      
      sint4 x0, y0;
      x0 = 10;
      y0 = 10;
      //rand_loc(md, x0, y0);
      if (!base_loc_ok(md, x0, y0)) { --p; continue; }
    
      mark_base_loc(md, x0, y0);
      place_unit(world, "start_loc", p, x0, y0);
    }

    // create random resource clusters
  
    /*FORS (i, GAME1_NUM_CLUSTERS) {

      // retry if hung
      if (++loops >= 1000) goto try_again;

      sint4 x0, y0;

      rand_loc(md, x0, y0);
      if (!resource_loc_ok(md, x0, y0)) { --i; continue; }
      populate_resource_loc(md, x0, y0, world);
    }*/

    break;
  }
  world << "</OBJECTS>\n";

  output_world << world.str();
}
