// $Id: Decomp.H 5279 2007-06-23 02:49:08Z mburo $
 
// This is an ORTS file (c) Michael Buro, licensed under the GPL

// rectangular free-space decomposition

#ifndef Decomp_H
#define Decomp_H

#include "Global.H"

#define NODE_IDS 1 // for testing

// todo: decomposition should be quick to copy
//       store nodes in a vector and use indexes instead of pointers
//       using only POD types (-> memcpy)

class Options;

class SearchNode;

class Decomp {

public:

  typedef uint4 Data;      // associated with rectangle
  typedef uint2 GridCoor;  // grid coordinate  (should be unsigned, otherwise add <0 checks)
  
  struct Rect {

    GridCoor x0, x1, y0, y1; // x0 < x1, y0 < y1

    void write(std::ostream &os = std::cout) const {
      os << "[" << x0 << "," << x1 << "]x["
	 << y0 << "," << y1 << "]";
    }

    void set(GridCoor x0_, GridCoor x1_, GridCoor y0_, GridCoor y1_) {
      x0 = x0_;
      x1 = x1_;
      y0 = y0_;
      y1 = y1_;      
    }
    
  };  

  // todo: Node size is currently 68! (+ neighbor lists stored elsewhere)
  //       that needs to be addressed

  struct Node
  {
  public:

    typedef Vector<Node*> RList;
    
    enum Type { NONE, X_NODE, Y_NODE, R_NODE };
    enum Dir  { N=0, E, S, W }; // don't change

    sint2 type;

    union {

      // R_NODE
      struct {
	Data data; 
	Rect r;
      };

      // X/Y_NODE
      struct {
	GridCoor v;
	Node *lt, *ge;
      };
    };

    SearchNode *s_node; // abstraction
    RList nl[4]; // neighbor lists for 4 boundaries (N,E,S,W)

    void write(std::ostream &os = std::cout) const;
    bool check(const string &msg) const;
    
#if NODE_IDS    
    uint4 id;
    uint4 sub_id;
    static uint4 id0;
    static uint4 sub_id0;
#endif
    
  };

  Decomp();
  virtual ~Decomp() { free(); }
  
  Decomp(const Decomp& x) { copy(x); }

  Decomp& operator=(const Decomp& x) {
    if (this != &x) { free(); copy(x); }
    return *this;
  }

  Decomp *clone() {
    Decomp *n = new Decomp;
    *n = *this; return n;
  }

  void insert(const Rect &r, Data mask, Data data);

  Node *locate(GridCoor x, GridCoor y);

  bool check(const string &msg);
  
  static void test(Options &opt);

  struct Edge {
    Decomp::Node *a;
    Decomp::Node *b;
    Edge(Decomp::Node *a_, Decomp::Node *b_) : a(a_), b(b_) {}
  };

  Vector<Node*> nodes; // for clean-up
  void count(sint4 &r, sint4 &x, sint4 &y, sint4 &l) const;

  Node *rect0();
  
protected:

  Node *root;

  Rect r0;     // surrounding rectangle (for checks)

  Node *new_Node();
  Node *new_Node(GridCoor x0, GridCoor x1, GridCoor y0, GridCoor y1, Data d);
  Node *new_Node(Node::Type type_, Data v_, Node *lt_, Node *ge_);

  void init();
  void free();
  void copy(const Decomp& x);

  void insert(Node *p, const Rect &r, Data mask, Data data, set<Node*> &visited);

  // testing
  
  void collect_edges(Vector<Edge> &es);  
  void area(Node *p, set<Node*> &s, sint4 &a);
  void dump(Node *p, set<Node*> &s);
  void dump();
  
  // compute all relevant intervals and edge coverage
  void north_edge(Node *p, const Rect &r, Node::RList &succ, bool &cl, bool &cr, bool &ca);
  void east_edge (Node *p, const Rect &r, Node::RList &succ, bool &cl, bool &cr, bool &ca);
  void south_edge(Node *p, const Rect &r, Node::RList &succ, bool &cl, bool &cr, bool &ca);
  void west_edge (Node *p, const Rect &r, Node::RList &succ, bool &cl, bool &cr, bool &ca);  

  // delete all edges pointing to p from north,east,south,west
  void delete_edges_to(Node *p);

  // connect new rectangle to outside
  void connect_north(Node *p, Node *rnew);
  void connect_east (Node *p, Node *rnew);
  void connect_south(Node *p, Node *rnew);
  void connect_west (Node *p, Node *rnew);

  // 16 cases
  
  void fully_embedded(Node *p, const Rect &r, Data mask, Data data);

  void NW_corner(Node *p, const Rect &r, Data mask, Data data);
  void NE_corner(Node *p, const Rect &r, Data mask, Data data);
  void SE_corner(Node *p, const Rect &r, Data mask, Data data);
  void SW_corner(Node *p, const Rect &r, Data mask, Data data);

  void full_N_edge(Node *p, const Rect &r, Data mask, Data data);
  void full_E_edge(Node *p, const Rect &r, Data mask, Data data);
  void full_S_edge(Node *p, const Rect &r, Data mask, Data data);
  void full_W_edge(Node *p, const Rect &r, Data mask, Data data);

  void WE_embedded_strip(Node *p, const Rect &r, Data mask, Data data);
  void NS_embedded_strip(Node *p, const Rect &r, Data mask, Data data);

  void N_middle_section(Node *p, const Rect &r, Data mask, Data data);
  void E_middle_section(Node *p, const Rect &r, Data mask, Data data);
  void S_middle_section(Node *p, const Rect &r, Data mask, Data data);
  void W_middle_section(Node *p, const Rect &r, Data mask, Data data);
};


inline bool operator<(const Decomp::Edge &x, const Decomp::Edge &y) {
  if (x.a < y.a) return true;
  if (x.a > y.a) return false;
  return x.b < y.b;
}

inline bool operator==(const Decomp::Edge &x, const Decomp::Edge &y) {
  return x.a == y.a && x.b == y.b;
}

inline bool operator!=(const Decomp::Edge &x, const Decomp::Edge &y) {
  return !(x == y);
}


class SectorDecomp
{
public:
  
  SectorDecomp(Decomp::GridCoor w_, Decomp::GridCoor h_, Decomp::GridCoor sector_size_);
  virtual ~SectorDecomp() {}

  Decomp::Node *locate(Decomp::GridCoor x, Decomp::GridCoor y);
  void insert(const Decomp::Rect &r, Decomp::Data mask, Decomp::Data data);

  Decomp::GridCoor w, h;        // total width, height
  Decomp::GridCoor sector_size; // sector size

  sint4 sx, sy;      // sectors in x and y direction
  
  Vector<Decomp> sectors;
};

REGISTER_TYPEOF(68, Decomp::Node::RList);

#endif
