// $Id: Maze.C 5279 2007-06-23 02:49:08Z mburo $

/*
    Copyright (C) 2001 by W.C.A. Wijngaards

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// cleaned-up by Michael Buro 2003

#include "Maze.H"

// http://www.pmms.cam.ac.uk/~gjm11/programs/maze/index.html
// This URL helped me to understand this code

Maze::Maze(sint4 w, sint4 h)
{
  width = w;
  height = h;
  maze = new MazeNode [width*height];
  straightness = 0.5f;
  cyclicalness = 0.1f;
  init_maze();
}

Maze::~Maze()
{
  delete[] maze;
}

void Maze::init_maze()
{
  FORS (y, height) {
    FORS (x, width) {
      get_node(x,y).visited = false;
      FORS (i, 4) get_node(x,y).opening[i] = false;
    }
  }
}

void Maze::make_access(sint4 x, sint4 y)
{
  if (x == 0) get_node(x,y).opening[3] = true;
  if (y == 0) get_node(x,y).opening[0] = true;
  if (x == width-1) get_node(x,y).opening[1] = true;
  if (y == height-1) get_node(x,y).opening[2] = true;
}

void Maze::generate(sint4 x, sint4 y)
{
  sint4 direction = sint4(float(random()+1)/RAND_MAX*4.0);
  visit_node(x, y, direction);
}

void Maze::visit_node(sint4 x, sint4 y, sint4 direction)
{
  /// generate an order in which to test the neighbours.

  sint4 order[4];
  gen_order(order, direction);
  get_node(x,y).visited = true;

  /// take each direction and go there...

  FORS (i, 4) {
    
    sint4 nx, ny;
    get_neighbor(order[i], x, y, nx, ny);

    /// don't go outside the map
    if (nx < 0 || nx >=width || ny < 0 || ny >=height) continue;

    /// been there?

    if (get_node(nx,ny).visited) {
      // avoid the cycle?
      if ((1.0+rand())/RAND_MAX > cyclicalness) continue;
    }

    /// make opening
    make_opening(x, y, nx, ny);

    /// visit that neightbor
    if (get_node(nx, ny).visited) continue;

    visit_node(nx, ny, order[i]);
  }
}

void Maze::gen_order(sint4 order[4], sint4 direction)
{
  assert(direction >= 0 && direction < 4);
  bool done[4];

  { FORS (i, 4) done[i] = false; }

  sint4 i=0;

  if (Rand.rand_real8() < straightness) {
    i=1;
    order[0] = direction;
    done[direction] = true;
  }
 
  for (;i<4; i++) {

    /// skip from 0 ... 4-i-1
    sint4 skip = sint4(real4(4-i)*Rand.rand_real8());
    if (skip > 4-i-1) skip = 4-i-1;
    assert(skip >= 0 && skip <= 4-i-1);

    /// find it
    sint4 numskipped = 0;
    sint4 found = 0;
    FORS (k, 4) {
      if (!done[k]) {
        numskipped++;
        if (numskipped == skip+1) { found = k; break; }
      }
    }
    /// store it;
    order[i] = found;
    done[found] = true;
  }

  assert(done[0] && done[1] && done[2] && done[3]);
}

void Maze::get_neighbor(sint4 dir, sint4 x, sint4 y, sint4& nx, sint4& ny)
{
  nx = x; ny = y;
  switch(dir) {
  case 0: ny--; break;
  case 1: nx++; break;
  case 2: ny++; break;
  case 3: nx--; break;
  default: ERR("what?");
  }
}

void Maze::make_opening(sint4 x1, sint4 y1, sint4 x2, sint4 y2)
{
  if (x1 == x2-1) {
    get_node(x1,y1).opening[1] = true;
    get_node(x2,y2).opening[3] = true;
    return;
  }
  if (x1 == x2+1) {
    get_node(x1,y1).opening[3] = true;
    get_node(x2,y2).opening[1] = true;
    return;
  }
  if (y1 == y2-1) {
    get_node(x1,y1).opening[2] = true;
    get_node(x2,y2).opening[0] = true;
    return;
  }
  if (y1 == y2+1) {
    get_node(x1,y1).opening[0] = true;
    get_node(x2,y2).opening[2] = true;
    return;
  }
}

bool Maze::opening(sint4 x1, sint4 y1, sint4 x2, sint4 y2)
{
  if (x1 == x2-1) {
    assert(get_node(x1,y1).opening[1] == get_node(x2,y2).opening[3]);
    return get_node(x1,y1).opening[1];
  }
  
  if (x1 == x2+1) {
    assert(get_node(x1,y1).opening[3] == get_node(x2,y2).opening[1]);
    return get_node(x1,y1).opening[3];
  }
  
  if (y1 == y2-1) {
    assert(get_node(x1,y1).opening[2] == get_node(x2,y2).opening[0]);
    return get_node(x1,y1).opening[2];
  }
  
  if (y1 == y2+1) {
    assert(get_node(x1,y1).opening[0] == get_node(x2,y2).opening[2]);
    return get_node(x1,y1).opening[0];
  }
  /// unreachable
  assert(false);
  return false;
}

sint4 Maze::actual_height()
{
  return height*2+1;
}

sint4 Maze::actual_width()
{
  return width*2+1;
}

/// Returns true if ActualX,ActualY is solid
/// Notes: The maze might be say 10 x 10 but the real
/// size of the maze (with tickness of walls taken into
/// account is 21 x 21) - This function adjusts for those 
/// walls and reports on it. 
///   -- based on code cut from isptest.cpp

bool Maze::actual_solid(sint4 x, sint4 y)
{
  // Last row of the maze
  // The x%2 will trip on actual_height()-1
  // thus setting the SE corner to solid (true)

  if (y == actual_height()-1) {
    if ((x & 1) == 0)
      return true;
    else
      return !get_node(x/2,(y/2)-1).opening[2];
  }

  // Every other row, Starting with the 0'th row,
  // only need to check northern openings here

  if ((y & 1) == 0) {
    if (x == actual_width()-1) return true;

    if ((x & 1) == 0)
      return true;
    else
      return !get_node(x/2,y/2).opening[0];
  }

  // Every other row, starting with the 1'st row
  // only need to check western openings generally
  // but check eastern at the end of a row

  if (x == actual_width()-1)
    return !get_node((x/2)-1,y/2).opening[1];
  if ((x & 1) == 0)
    return !get_node(x/2,y/2).opening[3];	
  else
    return false;
}

/// Prints Maze - Helpful during testing

void Maze::print()
{
  FORS (y, actual_height()) {
    FORS (x, actual_width()) {
      if (actual_solid(x,y))
	printf("X");
      else
	printf(" ");
    }
    printf("\n");
  }
}
