// $Id: Quadtree.C 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

// Quadtree implementation
// generates region adjecency graph on the fly 

#include "Quadtree.H"
#include "SDLinit.H"
#include "SDL_gfxPrimitives.h"
#include "GUI.H"

const sint4 GFX_MAG = 1; // magnification

static bool gfx = true;
static uint4 updates, splits, edges;

QuadtreeNode::QuadtreeNode()
{
  pred = succ[0] = succ[1] = succ[2] = succ[3] = 0;
  xl = yu = w = 0;
}

QuadtreeNode::~QuadtreeNode()
{
  FORS (i, 4) if (succ[i]) delete succ[i];
}


const QuadtreeNode *QuadtreeNode::locate(sint4 x, sint4 y) const
{
  if (x < xl || x >= xl+w || y < yu || y >= yu+w) ERR("point outside square");

  if (!is_split()) return this;

  sint4 sx = xl + (w >> 1);
  sint4 sy = yu + (w >> 1);
  
  if (x < sx) {
    
    if (y < sy) return succ[S_UL]->locate(x, y); else return succ[S_LL]->locate(x, y);
    
  } else {
    
    if (y < sy) return succ[S_UR]->locate(x, y); else return succ[S_LR]->locate(x, y);
  }
}


bool QuadtreeNode::insert(sint4 x, sint4 y)
{
  if (x < xl || x >= xl+w || y < yu || y >= yu+w) {
    cerr << x << " " << y << " | " << xl << " " << yu << " " << w << endl;
    ERR("point outside square");
  }

  if (w > 1) {
  
    if (!is_split()) {
      split();
      if (w == 2) return true;
    }

    sint4 sx = xl + (w >> 1);
    sint4 sy = yu + (w >> 1);
    
    if (x < sx) {
      
      if (y < sy) return succ[S_UL]->insert(x, y); else return succ[S_LL]->insert(x, y);
      
    } else {
      
      if (y < sy) return succ[S_UR]->insert(x, y); else return succ[S_LR]->insert(x, y);
    }
  }

  return false;
}


// connect child with neighbor

inline static void connect(QuadtreeNode &parent, QuadtreeNode &from,
			   sint4 t, sint4 fldir, sint4 tldir)
{
  QuadtreeNode &to = *parent.succ[t]; 
  QuadtreeNode::NList &fl = from.nls[fldir];      // from-list
  QuadtreeNode::NList &tl = to.nls[tldir];        // to-list
  QuadtreeNode::Edge e;

  ++edges;
  assert(tl.empty() && fl.empty());
  
  tl.push_back(e);
  fl.push_back(e);
  tl.begin()->node     = &from;
  tl.begin()->list_pos = fl.begin();
  fl.begin()->node     = &to;
  fl.begin()->list_pos = tl.begin();
}


const sint4 HORI = 0, VERT = 1; // how children are aligned

void update_edges(QuadtreeNode *node,
		  sint4 n_dir, sint4 n_dir_c, // dir to neighbor, complement
		  sint4 align,                // HORI or VERT alignment of successors
		  sint4 succ1, sint4 succ2    // successor indexes (West/East North/S)
		  )
{    
  const sint4 w2 = node->w >> 1;  // width of children
  const sint4 sx = node->xl + w2; // midpoint
  const sint4 sy = node->yu + w2;  

  FORALL (node->nls[n_dir], i) {              // traverse through neighbor list

    ++updates;
    QuadtreeNode &n = *(*i).node;              // neighbor is n
    QuadtreeNode::NList &nl = n.nls[n_dir_c];  // neighbor list of n

    bool half[2] = { n.xl < sx, n.yu < sy };   // prepare data-driven comparison below
    // 0: true:West  false:East
    // 1: true:North false:South

    if (n.w < node->w) {

      // neighbor is smaller -> only one connection 
	
      assert(nl.size() == 1);

      // choose child to connect to
      QuadtreeNode *child = half[align] ? node->succ[succ1] : node->succ[succ2];
      QuadtreeNode::Edge e;
      QuadtreeNode::NList &cnl = child->nls[n_dir]; // child neighbor list

      // new entry in child n-list
      QuadtreeNode::NList::iterator lp = cnl.insert(cnl.end(), e);
	
      assert(nl.begin() == (*i).list_pos); // one element

      // connect neighbor to child
      nl.begin()->list_pos = lp; 
      nl.begin()->node = child;

      // connect child to neighbor
      lp->node = &n;
      lp->list_pos = (*i).list_pos;
	
    } else {

      // neighbor covers entire square -> link both children to neighbor

      QuadtreeNode::Edge e;
      QuadtreeNode *child1 = node->succ[succ1];
      QuadtreeNode *child2 = node->succ[succ2];	
      QuadtreeNode::NList &cnl1 = child1->nls[n_dir]; // child1 neighbor list
      QuadtreeNode::NList &cnl2 = child2->nls[n_dir]; // child2 neighbor list

      // add edges to child neighbor lists
      QuadtreeNode::NList::iterator lp1 = cnl1.insert(cnl1.end(), e);
      QuadtreeNode::NList::iterator lp2 = cnl2.insert(cnl2.end(), e);

      QuadtreeNode::NList::iterator nlp1 = (*i).list_pos;
      QuadtreeNode::NList::iterator nlp2 = nlp1;
      ++nlp2;
      nlp2 = nl.insert(nlp2, e);
      ++edges;
      
      // connect children -> neighbor
      lp1->node     = &n;
      lp1->list_pos = nlp1;
      lp2->node     = &n;
      lp2->list_pos = nlp2;	

      // connect neighbor -> children
      nlp1->node     = child1;
      nlp1->list_pos = lp1;
      nlp2->node     = child2;
      nlp2->list_pos = lp2;	
    }
  }
}


// split square and adjust adjecency graph

void QuadtreeNode::split()
{
  ++splits;
  
  assert(!is_split());
  assert(w > 1);
  
  FORS (i, 4) succ[i] = new QuadtreeNode;

  const sint4 w2 = w >> 1;
  const sint4 sx = xl + w2;
  const sint4 sy = yu + w2;  

  // set successor dimensions, connect

  { // S_UL
    QuadtreeNode &child = *succ[S_UL];    
    child.xl = xl; child.yu = yu; child.w = w2; child.pred = this;
    connect(*this, child, S_UR, N_E, N_W); // East  -> S_UR
  }

  { // S_UR
    QuadtreeNode &child = *succ[S_UR];
    child.xl = sx; child.yu = yu; child.w = w2; child.pred = this;
    connect(*this, child, S_LR, N_S, N_N); // South -> S_LR
  }

  { // S_LL
    QuadtreeNode &child = *succ[S_LL];
    child.xl = xl; child.yu = sy; child.w = w2; child.pred = this;
    connect(*this, child, S_UL, N_N, N_S); // North -> S_UL
  }

  { // S_LR
    QuadtreeNode &child = *succ[S_LR];
    child.xl = sx; child.yu = sy; child.w = w2; child.pred = this;
    connect(*this, child, S_LL, N_W, N_E); // West  -> S_LL
  }

  update_edges(this, N_N, N_S, HORI, S_UL, S_UR);  // North
  update_edges(this, N_S, N_N, HORI, S_LL, S_LR);  // South
  update_edges(this, N_W, N_E, VERT, S_UL, S_LL);  // West
  update_edges(this, N_E, N_W, VERT, S_UR, S_LR);  // East    
}


static void traverse(SDL_Surface *screen, QuadtreeNode *qt)
{
  if (!qt->is_split()) {

    if (gfx) {
      // cout <<  qt->xl << " " << qt->yu << " " << qt->w << endl;
      RGB rgb;

      if (qt->w == 1) {
	rgb.from(255,255,255);
      } else {
	rgb.from(rand() % 100 + 50, rand() % 100 + 50, rand() % 100 + 50);
      }
      sint4 x1 = qt->xl*GFX_MAG;
      sint4 y1 = qt->yu*GFX_MAG;
      SDL_Rect rect;
      rect.x = x1;
      rect.y = y1;
      rect.w = GFX_MAG*qt->w;
      rect.h = GFX_MAG*qt->w;
      SDL_FillRect(screen, &rect, rgb.color());

      // draw neighbor edges

      if (qt->w >= 8) {

	{
	  // North connections

	  QuadtreeNode::NList &nl = qt->nls[QuadtreeNode::N_N];
	  rgb.from(0,0,0);

	  FORALL (nl, i) {
	    QuadtreeNode &n = *(*i).node;
	    lineColor(screen, qt->xl+qt->w/2, qt->yu+qt->w/2, n.xl+n.w/2, n.yu+n.w/2,
		      rgb.color_alpha());
	  }
	}

	{
	  // East connections
	  QuadtreeNode::NList &nl = qt->nls[QuadtreeNode::N_W];
	  rgb.from(0,0,0);

	  FORALL (nl, i) {
	    QuadtreeNode &n = *(*i).node;
	    lineColor(screen, qt->xl+qt->w/2, qt->yu+qt->w/2, n.xl+n.w/2, n.yu+n.w/2,
		      rgb.color_alpha());
	  }
	}
      }
    }
    return;
  }
  
  traverse(screen, qt->succ[QuadtreeNode::S_UL]);
  traverse(screen, qt->succ[QuadtreeNode::S_UR]);
  traverse(screen, qt->succ[QuadtreeNode::S_LL]);
  traverse(screen, qt->succ[QuadtreeNode::S_LR]);    
}


void QuadtreeNode::test()
{
  const sint4 N = 4096; //1024; //2048; // 1024;
  const sint4 WX = 900, WY = 700;
  
  SDL_Surface *screen = 0;
  
  if (gfx) {
    SDLinit::video_init();
    screen = SDL_SetVideoMode(WX, WY, 0, SDL_HWSURFACE | SDL_ANYFORMAT);
    
    if (screen == NULL) {
      fprintf(stderr, "Couldn't set video mode");
      exit(1);
    }
    
    SDL_WM_SetCaption("test", 0);
  }

  cout << "size: " << sizeof(QuadtreeNode) << endl;
  cout << "dim: " << N << endl;
  //WX/GFX_MAG << " K=" << log(double(WX/GFX_MAG))/log(2.0) << endl;

  for (sint4 l=100; l <= 50000; l *= 10) {

    QuadtreeNode *qt = new QuadtreeNode;
    qt->xl = qt->yu = 0; qt->w = N;

    updates = splits = edges = 0;
  
    FORS (i, l) {
      qt->insert(rand() % qt->w, rand() % qt->w);
    }

    cout << "n=" << l << ": "
	 << updates << " updates, "
	 << splits << " splits, "
	 << 3*splits+1 << " squares, "      
	 << edges << " edges, "
	 << double(2*edges)/(3*splits+1) << " avg.degree"
	 << endl;

    if (gfx) {
      if (SDL_MUSTLOCK(screen) && SDL_LockSurface(screen) < 0 ) {
	fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
	return;
      }
    }
  
    traverse(screen, qt);

    if (gfx) {
      if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);
      SDL_UpdateRect(screen, 0, 0, 0, 0);
      cout << "wait" << endl;
      string s;
      cin >> s;
    }

    delete qt;
  }
}

