//-----------------------------------------------------------------------------
/** @file pathfindtest.cpp
  Example program using the pathfinding library.
  (Used as a test to see if library is installed correctly.)

  $Id: pathfindtest.C 5279 2007-06-23 02:49:08Z mburo $
  $Source$
    
  Based on example.cpp from pathfind library
*/

// This is an ORTS file (c) Michael Chung, licensed under the GPL

//-----------------------------------------------------------------------------

#include "Global.H"

#include "pathfind.h"

using namespace std;
using namespace PathFind;

//-----------------------------------------------------------------------------

static const int s_numberRuns = 10;

static const int s_columns = 40;

static const int s_rows = 20;

static const float s_obstaclePercentage = 0.2;

static const long long int s_nodesLimit = 100000000L;

// Parameters for BucketFringeSearch.

static const int s_numberBuckets = 256;

static const int s_deltaBucketFValue = 100;

//-----------------------------------------------------------------------------

typedef enum {
  A_STAR,
  IDA_STAR,
  FRINGE,
  BUCKETFRINGE
} SearchAlgorithm;

//-----------------------------------------------------------------------------

void printHeader(SearchAlgorithm searchAlgorithm, Tiling::Type type)
{
  cout << "==============================================================\n";
  switch (searchAlgorithm) {
  case A_STAR:
    cout << "A_STAR";
    break;
  case IDA_STAR:
    cout << "IDA_STAR";
    break;
  case FRINGE:
    cout << "FRINGE";
    break;
  case BUCKETFRINGE:
    cout << "BUCKETFRINGE";
    break;
  }
  cout << "  ";
  switch (type) {
  case Tiling::TILE:
    cout << "TILE";
    break;
  case Tiling::OCTILE:
    cout << "OCTILE";
    break;
  case Tiling::OCTILE_UNICOST:
    cout << "OCTILE_UNICOST";
    break;
  case Tiling::HEX:
    cout << "HEX";
    break;
  }
  cout << '\n';
  cout << "==============================================================\n";
}

void runExperiment(SearchAlgorithm searchAlgorithm, Tiling::Type type)
{
  printHeader(searchAlgorithm, type);
  auto_ptr<Search> search;
  switch (searchAlgorithm) {
  case A_STAR:
    search.reset(new AStar<MarkerFastClear,
		 AStarOpenBuckets<MarkerFastClear,AStarCompareF>,
		 AStarClosedLookup<MarkerFastClear> >());
    break;
  case IDA_STAR:
    search.reset(new IDAStar(IDAStar::f_CACHING));
    break;
  case FRINGE:
    search.reset(new FringeSearch<>());
    break;
  case BUCKETFRINGE:
    search.reset(new BucketFringeSearch<>(s_numberBuckets,
					  s_deltaBucketFValue));
    break;
  }
  search->setNodesLimit(s_nodesLimit);
  SearchUtils searchUtils;
  StatisticsCollection statistics = search->createStatistics();
  for (int runIndex = 0; runIndex < s_numberRuns; ++runIndex) {
    // Create a tiling search environment
    Tiling tiling(type, s_rows, s_columns);
    tiling.setObstacles(s_obstaclePercentage, false);
        
    // Choose random start and target
    int start, target;
    searchUtils.findRandomStartTarget(tiling, start, target);
    cout << "Run " << runIndex << ":\n";
    tiling.printFormatted(cout, start, target);
    cout << '\n';
        
    // Do the search
    search->findPath(tiling, start, target);
    cout << '\n';
    const vector<int>& path = search->getPath();
    tiling.printFormatted(cout, path);
    cout << '\n';
    tiling.printPathAndLabels(cout, path, search->getVisitedNodes());
    cout << '\n';
        
    // Accumulate statistics
    const StatisticsCollection& searchStatistics = search->getStatistics();
    searchStatistics.print(cout);
    cout << '\n';
    statistics.add(searchStatistics);
    statistics.print(cout);
    cout << '\n';
  }
  statistics.print(cout);
}

int main()
{
  cout << "PathFindTest running..." << endl;

  try {
    //srand(time(0));        
    runExperiment(A_STAR, Tiling::TILE);
    runExperiment(A_STAR, Tiling::OCTILE);
    runExperiment(A_STAR, Tiling::OCTILE_UNICOST);
    runExperiment(A_STAR, Tiling::HEX);
    runExperiment(IDA_STAR, Tiling::TILE);
    runExperiment(IDA_STAR, Tiling::OCTILE);
    runExperiment(IDA_STAR, Tiling::OCTILE_UNICOST);
    runExperiment(IDA_STAR, Tiling::HEX);
    runExperiment(FRINGE, Tiling::TILE);
    runExperiment(BUCKETFRINGE, Tiling::TILE);
  }
  catch (const exception& e) {
    cerr << "Error: " << e.what() << '\n';
    return -1;
  }
  
  cout << "Test complete." << endl;
  
  return 0;
}

//-----------------------------------------------------------------------------
