/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
  $Header$
  $Log$
  Revision 1.1  2004/10/11 09:09:18  orts_mburo
  Initial revision

  Revision 1.1.1.1  2003/06/11 22:04:54  orts_mburo
  import

*/

#include "RTS_Game.H"
#include "Element.H"
#include "RTS_regact.H"
#include "RT_GameParam.H"
#include "BoolMatrix.H"
#include "CycleCounter.H"
#include <algorithm>
#include <vector>
#include <SDL.h>
#include "SDL_gfxPrimitives.h"

static sint4 t_n = 0;
static sint8 t_comp_act = 0;
static sint8 t_exe_act = 0;
static sint8 t_save = 0;

// fixme: currently only 2 player games supported

static sint4 intlog2(sint4 x) {

  if (x <= 0) ERR("x <= 0 in intlog2");
  
  sint4 i = 0, m = 1;

  while (m < x) { i++; m <<= 1; }
  return i;
}

RTS_PlayerInfo::RTS_PlayerInfo() : idf(RTS_Game::MAX_ELEM_NUM) {}

void RTS_PlayerInfo::delete_cid(ObjId cid)
{
  // delete map entries
  
  IdMap::const_iterator mit = cid2sid.find(cid);
  if (mit == cid2sid.end()) ERR("cid not found");
  cid2sid.erase(cid);
  sid2cid.erase((*mit).second); // sid

  // free cid

  if (idf.free(cid)) ERR("cid can't be freed");
}

ObjId RTS_PlayerInfo::add_sid(ObjId sid)
{
  ObjId cid = idf.next();
  if (!cid) ERR("no more ids");
  sid2cid[sid] = cid;
  cid2sid[cid] = sid;
  return cid;
}


ObjId RTS_PlayerInfo::get_cid(ObjId sid)
{
  IdMap::const_iterator mit = sid2cid.find(sid);
  if (mit == sid2cid.end()) return 0;
  return (*mit).second;
}

void RTS_PlayerInfo::check() const
{
  // maps must be bijecive

  if (sid2cid.size() != cid2sid.size()) ERR("maps have different sizes");

  {
    set<ObjId> range;
    FORALL_CONST(IdMap, cid2sid, it) range.insert((*it).second);
    if (range.size() != cid2sid.size()) ERR("cid2sid not injective");
  }
  
  {
    set<ObjId> range;
    FORALL_CONST (IdMap, sid2cid, it) range.insert((*it).second);
    if (range.size() != sid2cid.size()) ERR("sid2cid not injective");
  }
}


static void read_player_info(RTS_Game &g,
			     const ConstrMap &act_map,
			     Array<CompActCtor> &constrs,
			     vector<Prop> &features,
			     vector<string> &p_files,
			     const string &file_name,
			     bool save_history,
			     sint4 index)
{
  cout << "reading player " << (index+1) << " file '" << file_name << "'" << endl;
  ifstream ifs(file_name.c_str());
  if (!ifs) { ERR("file not found"); }

  // 1. line: action
  string action_name;
  ifs >> action_name;
  if (!ifs) { ERR("action name expected"); }
  cout << " action:" << action_name << endl;
  MapConstIt it = act_map.find(action_name);
  if (it == act_map.end()) ERR("action not registered");
  constrs[index] = (*it).second;

  // 2. line: parameter file
  ifs >> p_files[index];
  if (!ifs) { ERR("parameter file name expected"); }
  cout << " parameters:" << p_files[index] << endl;

  // 3. line: history file
  string hfn;
  ifs >> hfn;
  if (!ifs) { ERR("history file name expected"); }

  cout << " history:" << hfn << endl;
  
  if (save_history) {
    g.hist_os[index] = fopen(hfn.c_str(), "w");
    if (!g.hist_os[index]) ERR("can't open file");
  }

  // rest: features
  if (features[index].read_init_values(ifs)) { ERR("file corrupt"); }
  cout << " OK" << endl;
}



static sint4 scale(sint4 x, sint4 pix) {
  if (pix == 0) return x;
  if (pix > 0) return x << pix;
  return x >> (-pix);
}

RTS_Game::RTS_Game(const RTS_Options &opt_) : opt(opt_), idf(MAX_ELEM_NUM)
{

  ConstrMap act_map;
  Array<CompActCtor> constrs;
  vector<Prop> props;
  vector<string>   p_files;

  rts_regact(act_map);
  
  FORS (i, 2) {

    hist_os[i] = 0;
    CompActCtor N = 0;
    constrs.push_back(N);

    Prop f;
    props.push_back(f);

    string s;
    p_files.push_back(s);
  }

  string p1, p2;
  opt.get("-p1", p1);
  opt.get("-p2", p2);  

  bool save_history;
  opt.get("-save", save_history);

  read_player_info(*this, act_map, constrs, props, p_files, p1, save_history, 0);
  read_player_info(*this, act_map, constrs, props, p_files, p2, save_history, 1);
  
  // ----------------------------------------------
  
  screen = 0;

  opt.get("-x",   dim_x);
  opt.get("-y",   dim_y);
  opt.get("-pix", pixel_scale);

  // if (pixel_scale < 1) ERR("-pix < 1");
  
  sectors = 0;
  sector_num = 0;
  
  s_x = scale(dim_x, pixel_scale)+1;
  s_y = scale(dim_y, pixel_scale)+1;

  sub_grid_res_bits = intlog2(2*(dim_x + dim_y));
  sub_grid_res = 1 << sub_grid_res_bits;

  bool display;
  opt.get("-disp", display);

  string mode;
  opt.get("-m", mode);

  sint4 pnum;
  opt.get("-pnum", pnum);  

  if (pnum > 2 || pnum <= 0) ERR("pnum > 2 || pnum <= 0");
  
  // create shared view and friend status matrices
  
  x_shares_view_with_y      = new BoolMatrix(pnum, pnum);
  x_is_friend_of_y          = new BoolMatrix(pnum, pnum);
  prev_x_shares_view_with_y = new BoolMatrix(pnum, pnum);
  prev_x_is_friend_of_y     = new BoolMatrix(pnum, pnum);

  // create player infos
  
  players.reserve(pnum);
  
  FORS (i, pnum) {
    RTS_PlayerInfo pinf;
    String(pinf.name).form("player %d", i);
    players.push_back(pinf);
  }
  
  if (display) {
    
    // open screen
    
    // Have a preference for 8-bit, but accept any depth
    screen = SDL_SetVideoMode(s_x, s_y, 8, SDL_SWSURFACE|SDL_ANYFORMAT);
    if (screen == NULL) {
      fprintf(stderr, "Couldn't set %dx%dx8 video mode: %s\n",
	      s_x, s_y, SDL_GetError());
      exit(1);
    }
    
    printf("Set %dx%d at %d bits-per-pixel mode\n",
	   s_x, s_y,
	   screen->format->BitsPerPixel);

    sint4 cpind;
    opt.get("-cpind", cpind);

    ostringstream oss;
    oss << "ORTS " << mode;
    if (mode == "client") oss << " " << cpind;
    SDL_WM_SetCaption(oss.str().c_str(), 0);
    // UNFREEZE(oss);
  }
  
  // populate world
  
  //World *world = new World();
  //sb.new_root_container(world);
  elems.clear();
   
  if (mode != "client") {
    
    sint4 s = props[0].sv_radius;
    s = ::max(s, sint4(props[1].sv_radius));

    real8 sec_size;
    opt.get("-sec", sec_size);
    s = sint4(ceil(sec_size*2*s));  // sector size
  
    sint4 n = 0;
    vector<Pos> positions;
    sint4 n_max;
    opt.get("-nmax", n_max);
  
    for (sint4 y = s/2; y < dim_y - s/2; y += s) {
      for (sint4 x = s/2; x < dim_x - s/2; x += s) {
	Pos pos(x, y);
	positions.push_back(pos);
	n++;
	if (n_max >= 0 && n >= n_max) break;
      }
      if (n_max >= 0 && n >= n_max) break;      
    }
  
    if (n & 1) { n--; positions.pop_back(); }

    // color assignment
  
    vector<sint4> col(n);

    real8 freq;
    opt.get("-freq", freq);

    if (freq >= -1.0 && freq <= 1.0) freq = 0.5;
    else if (freq > 1) {
      freq = freq/(freq+1.0);
    } else {
      freq = 1/(1.0-freq);
    }

    if (freq > 1.0 || freq < 0.0) ERR("freq?");
    sint4 t = my_round(n*freq);

    for (sint4 i=0; i < t; i++) col[i] = 0;
    for (sint4 i=t; i < n; i++) col[i] = 1;
    random_shuffle(col.begin(), col.end());

    FORS (i, n) {
      
      Element *e = new Element();
      ostringstream oserr;
      sint4 index = col[i];
      
      e->comp_act = (*constrs[index])();

      sint4 t_pos_x = e->sv_pos_x; // save pos
      sint4 t_pos_y = e->sv_pos_y;      
      *(Prop*)e = props[index];
      e->sv_pos_x = t_pos_x;       // restore pos
      e->sv_pos_y = t_pos_y;      
      e->sv_id    = idf.next();
      
      ifstream ifs(p_files[index].c_str());
      if (!ifs) ERR("can't read parameter file");
      e->comp_act->read_params(ifs);
      ifs.close();
      
      e->sv_pos_x = positions[i].x;
      e->sv_pos_y = positions[i].y;      
      e->sv_owner = col[i];
      elems.push_back(e);
      
      // world->elements.link(e);

    }

    cout << n << " elements" << endl;
  }

  refresh_id2index();
}


RTS_Game::~RTS_Game()
{
  free_sectors();
  free_elems();

  delete x_shares_view_with_y;
  delete prev_x_shares_view_with_y;
  delete x_is_friend_of_y;
  delete prev_x_is_friend_of_y;
  
  bool display;
  opt.get("-disp", display);
  if (display) SDL_Quit();
}


void RTS_Game::alloc_sectors(sint4 n)
{
  if (n <= sector_num) return;
  free_sectors();
  sector_num = n;
  sectors  = (SectorInfo*) ::malloc(sizeof(SectorInfo) * n);

  if (!sectors) ERR("out of memory");

  FORS (i, n) {
    sectors[i].init();
  }
}

void RTS_Game::free_elems()
{
  FORALL (vector<Element*>, elems, i) if (*i) delete *i;
  elems.clear();
}


void RTS_Game::free_sectors() {
  if (sectors) {
    FORS (i, sector_num) sectors[i].free();
    ::free(sectors);
    sectors = 0;
  }
}

// display entire scene

void RTS_Game::display(uint4 ind1, uint4 ind2)
{
  Uint32 col_bg = 0xff;         // black
  // Uint32 col_fg = 0xffffffff;   // white
  // Uint32 col_mark = 0x00ffffff;

  /* Lock the screen for direct access to the pixels */

  if (SDL_MUSTLOCK(screen)) {
    if (SDL_LockSurface(screen) < 0 ) {
      fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
      return;
    }
  }

  // clear screen

  boxColor(screen, 0, 0, s_x, s_y, col_bg);
  
  // draw all objects

  sint4 track_id;
  opt.get("-track", track_id);

  bool fow;
  opt.get("-fow", fow);

  string mode;
  bool client_mode;
  opt.get("-m", mode);
  client_mode = (mode == "client");

  sint4 cpind;
  opt.get("-cpind", cpind);

  uint4 n = elems.size();

  if (client_mode && fow) {

    // fog of war
    
    Uint32 color_sight;

    if (cpind == 0) {
      color_sight = 0x004000ff;
    } else if (cpind == 1) {
      color_sight = 0x580000ff;
    } else {
      color_sight = 0x555500ff;
    }

    FORU (i, n) {
      const Element &e = *elems[i];
      if (e.sv_owner == cpind) {

	Pos pos(e.sv_pos_x, e.sv_pos_y);
	sint4 x = scale(pos.x, pixel_scale);
	sint4 y = scale(pos.y, pixel_scale);
	sint4 r = scale(e.sv_sight_range, pixel_scale) - 1;
	if (r <= 0) r = 1;
	filledCircleColor(screen, x, y, r, color_sight);
      }
    }
  }
  
  FORU (i, n) {

    const Element &e = *elems[i];
    Pos pos(e.sv_pos_x, e.sv_pos_y);
    sint4 x = scale(pos.x, pixel_scale);
    sint4 y = scale(pos.y, pixel_scale);
    Uint32 color, color_sight;

    if (e.sv_owner == 0) {
      color = 0x00f000ff; color_sight = 0x005000ff;
    } else if (e.sv_owner == 1) {
      color = 0xff0000ff; color_sight = 0x680000ff;
    } else {
      color = 0xaaaa00ff; color_sight = 0x555500ff;
    }
    sint4 r;

    if (track_id >= 0) {

      bool mark = false;

      if (i == uint4(track_id)) {
	
	r = scale(e.sv_sight_range, pixel_scale);
	if (r <= 0) r = 1;
	circleColor(screen, x, y, r, 0xffffffff);
	
      } else {

	if ((uint4)track_id < n) {

	  const Element &track_elem = *elems[track_id];
	  
	  FORS (j, track_elem.neighbors.size()) {
	    if (track_elem.neighbors[j] == (sint4)i) { mark = true; break; }
	  }
	}
      }
      
      if (mark) {
	r = scale(e.sv_radius, pixel_scale) + 2;
	circleColor(screen, x, y, r, 0xffffffff);
      }
    }

    r = scale(e.sv_radius, pixel_scale) - 1;
    if (r <= 0) r = 1;
    circleColor(screen, x, y, r, color);

    if (e.sv_hit_points < 0) {

      // element s about to die

      color = 0xffffffff;
      lineColor(screen, x-5, y, x+5, y, color);
      lineColor(screen, x, y+5, x, y-5, color);      
    }

    // display action
    
    const Action &act = e.last_action;

    switch (act.action) {

    case Action::NO_ACTION:
      break;

    case Action::STOP:
      color = 0xffffffff;
      circleColor(screen, x, y, 3, color);
      break;

    case Action::ATTACK:
      {
	IdMap::const_iterator mit = id2index.find(act.target_id);
	if (mit != id2index.end()) {
	  Element &target = *elems[(*mit).second];
	  sint4 tx = scale(target.sv_pos_x, pixel_scale);
	  sint4 ty = scale(target.sv_pos_y, pixel_scale);
	  color = 0xffffffff;
	  lineColor(screen, x, y, tx, ty, color);
	} 
      }
      break;
	
    case Action::MOVE:
      {
	real8 angle = (2.0*M_PI*act.move_direction)/Action::DIR_N;
	real8 dx = ::cos(angle); if (fabs(dx) < .001) dx = 0.0;
	real8 dy = ::sin(angle); if (fabs(dy) < .001) dy = 0.0;
	color = 0xffffffff;
	lineColor(screen, x, y,
		  x+scale(my_round(dx*6), pixel_scale),
		  y+scale(my_round(dy*6), pixel_scale),
		  color);
      }
      break;
    default:
      ERR("unknown action");
      break;	
    }
  }

  if (SDL_MUSTLOCK(screen)) { SDL_UnlockSurface(screen); }

  SDL_Flip( screen );
}


bool RTS_Game::finished(sint4 &winner)
{
  //SharedCont::SuccList objects;
  //World *pw = dynamic_cast<World*>(sb.get_root());

  uint4 n = elems.size();
  winner = -1;
  
  if (n == 0) return true;

  const sint4 pn = players.size();
  sint4 count[pn], newc = 0;
  
  FORS (i, pn) count[i] = 0;

  FORU (i, n) {
    Element &e = *elems[i];
    assert(e.sv_owner >= 0 && e.sv_owner < pn);
    if (!count[e.sv_owner]) newc++;
    if (newc >= 2) return false;
    count[e.sv_owner]++;
  }

  FORS (i, pn) if (count[i]) { winner = i; break; }
  return true;
}


bool RTS_Game::check_setup(ostream &oserr) const
{
  sint4 x, y;
  opt.get("-x", x);
  opt.get("-x", y);
  if (x < 1 || y < 1) { oserr << "x < 1 || y < 1" << endl; return true; }

  sint4 pnum;
  opt.get("-pnum", pnum);  
  if (pnum < 1) { oserr << "pnum < 1" << endl; return true; }

  return false;
}


bool RTS_Game::init()
{
  return false;
}


void RTS_Game::exec_actions()
{
  uint4 n = elems.size();
  sint4 dimx, dimy;
  opt.get("-x", dimx);
  opt.get("-y", dimy);

  bool save_history;
  opt.get("-save", save_history);

  *prev_x_shares_view_with_y = *x_shares_view_with_y;
  
  FORU (i, n) {
    
    Element &e = *elems[i];
    Action &act = e.last_action;
      
    if (save_history) {

      // write vicinity ids + action

      FILE *fp = hist_os[e.sv_owner];
      
      vector<Prop> &vic = e.last_vicinity;
      uint4 s = vic.size();

      FORU (i, s) fprintf(fp, "%d ", vic[i].sv_id);
      fprintf(fp, ": ");
      act.write(fp);
      fputs("\n", fp);
      if (ferror(fp)) ERR("history write error");
    }

    // execute

    switch (act.action) {

    case Action::NO_ACTION:
      break;

    case Action::STOP:

      elem_stop(&e);
      break;
	
    case Action::ATTACK:

      elem_stop(&e);
	
      {
	IdMap::const_iterator mit = id2index.find(act.target_id);
	if (mit == id2index.end()) {
	  cerr << "illegal target " << act.target_id << endl;
	  break;
	}
	Element &target = *elems[(*mit).second];

	//cout << "attack: from player: " << e.sv_owner << " to obj: " << target.sv_id << endl;

	Pos epos(e.sv_pos_x, e.sv_pos_y);
	Pos tpos(target.sv_pos_x, target.sv_pos_y);
	sint8 r2 = tpos.l2_dist_sq(epos);
	sint8 maxd = e.sv_max_attack_range + target.sv_radius;
	sint8 mind = e.sv_min_attack_range - target.sv_radius;
	if (mind < 0) mind = 0;

	// cout << r2 << " " << (mind*mind) << " " << (maxd*maxd) << endl;
	if (r2 >= square(mind) && r2 <= square(maxd))
	  // target within range => decrease hitpoints
	  target.sv_hit_points = target.sv_hit_points-e.sv_damage;
	else {
	  cout << ">>>>>>>>>>>>>>>>> invalid attack\a" << endl;
	}
      }
      break;

    case Action::MOVE:

      {
	if (act.move_direction < 0 || act.move_direction >= Action::DIR_N) {
	  cerr << "illegal move direction " << act.move_direction
	       << "(" << Action::DIR_N << ")" << endl;
	  break;
	}	

	elem_stop(&e);
	real8 angle = (2.0*M_PI*act.move_direction)/Action::DIR_N;
	real8 dx = ::cos(angle); if (fabs(dx) < .001) dx = 0.0;
	real8 dy = ::sin(angle); if (fabs(dy) < .001) dy = 0.0;
	Scalar r = e.sv_radius;
	Scalar x = e.sv_pos_x;
	Scalar y = e.sv_pos_y;      
	const Scalar EPS = 2;
	  
	if (x < r || x + r >= dimx ||
	    y < r || y + r >= dimy) {
	  cout << x << " " << y << " " << r << " " << dimx << " " << dimy << endl;
	  ERR("obj out of bounds");
	}

	if (x <= r+EPS && dx < 0) break;
	if (x >= dimx-r-EPS && dx > 0) break;

	if (y <= r+EPS && dy < 0) break;
	if (y >= dimy-r-EPS && dy > 0) break;

	r += EPS;

	Scalar tx, ty;

	if (dx > 0) tx = dimx-r; else tx = r;
	if (dy > 0) ty = dimy-r; else ty = r;

	real8 tmin = 1e200;
	  
	if (dx != 0.0) tmin = (tx-x)/dx;
	if (dy != 0.0) tmin = ::min(tmin, (ty-y)/dy);

	assert(tmin >= 0);
	tx = x + Scalar(tmin * dx);
	ty = y + Scalar(tmin * dy);
	Pos pos(tx, ty);
	elem_move_to(&e, pos);
      }
      break;      

    default:
      ERR("unknown action");
    }
  }
}


// remove all dead elements

void RTS_Game::remove_dead()
{
  uint4 n = elems.size();

  FORU (i, n) {
    
    Element &e = *elems[i];

    if (e.sv_hit_points < 0) {

      idf.free(elems[i]->sv_id);
      delete elems[i];
      elems[i] = elems[n-1];
      elems.pop_back();
      n--; i--;
    }
  }

  refresh_id2index();
}


bool RTS_Game::compute_actions()
{
  uint4 n = elems.size();

  FORU (i, n) {

    Element &e = *elems[i];
    vector<Prop> &vic = e.last_vicinity;
    vic.clear();
    vic.push_back(e);

    // compute features for all elements in e's vicinity

    uint4 s = e.neighbors.size();
    vic.reserve(s+1);

    FORU (j, s) {
      Element &e_vis = *elems[e.neighbors[j]];
      vic.push_back(e_vis);
    }

    // compute action
      
    e.last_action = e.comp_act->action(vic);
  }

  return false;
}


bool RTS_Game::simulation_round(sint4 tick)
{
  bool save_history;

  opt.get("-save", save_history);

  move_elements();
  compute_visible_elements();

  // compute features

  uint4 num[2];
  num[0] = num[1] = 0;
    
  uint4 n = elems.size();

  FORU (i, n) num[elems[i]->sv_owner]++;

  if (save_history) {

    CycleCounter start;
    
    // write element description to both files
    // must conform to TickInfo::write
    
    fprintf(hist_os[0], "tick: %d n: %d owner: 0 actions: %d\n", tick, n, num[0]);
    fprintf(hist_os[1], "tick: %d n: %d owner: 1 actions: %d\n", tick, n, num[1]);
    FORU (i, n) {
      elems[i]->Prop::write(hist_os[0]); fputs("\n", hist_os[0]);
      elems[i]->Prop::write(hist_os[1]); fputs("\n", hist_os[1]);
    }

    CycleCounter end;
    t_save += end.count()-start.count();
  }

  sint4 dimx, dimy;
  opt.get("-x", dimx);
  opt.get("-y", dimy);
    
  CycleCounter start;

  compute_actions();

  CycleCounter end;
  t_comp_act += end.count() - start.count();

  exec_actions();

  t_n++;

  bool stat;
  opt.get("-stat", stat);
  
  if (stat) { 
    form(cout, "comp_action: %10lld\n", t_comp_act/t_n);
    form(cout, "exe_action:  %10lld\n", t_exe_act/t_n);
    form(cout, "save:        %10lld\n", t_save/t_n);  
  }
  
  bool d;
  opt.get("-disp", d);

  if (d) display();

  remove_dead();
  // ts.tick();

  sint4 winner;
  return finished(winner);
}


void RTS_Game::simulation()
{
  sint4 ticks = 0;
  sint4 max_ticks;
  bool stop;

  opt.get("-tmax", max_ticks);
  opt.get("-stop", stop);  
  
  FOREVER {

#if TEST_OUT    
    cout << endl << "tick " << tick << endl;
#endif

    ticks++;
    simulation_round(ticks);
    
    sint4 w;
    opt.get("-wait", w);
    if (w > 0) usleep(w);
    else if (w < 0) {
      char s[20];
      cin.getline(s,10);
    }

    sint4 winner;
    if ((max_ticks > 0 && ticks >= max_ticks) || (stop && finished(winner))) break;
  }

  FORS (i, 2) {
    if (hist_os[i]) {
      if (ferror(hist_os[i])) ERR("write error hist-file");
      fclose(hist_os[i]);
    }
  }
}


