#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <stack>
#include <list>
#include <set>

#include <Windows.h>

std::string globalRoot;

class XMLFile
{
  std::ostream & out;
  std::stack<std::string> tagStack;

  void NextLine()
  {
    out << "\n\r" << std::string(tagStack.size(), '\t');
  }
public:
  XMLFile(std::ostream & out) : out(out) 
  {
    out << "<?xml version=\"1.0\" encoding=\"Windows-1252\"?>";
  }
  void OpenTag(const std::string & tag)
  {
    NextLine();
    out << "<" << tag;
    tagStack.push(tag);
  }

  void AddProperty(const std::string & prop, const std::string & value)
  {
    NextLine();
    out << prop << "=\"" << value << "\"";
  }

  void CloseTag()
  {
    out << ">";
  }

  void CloseEndTag()
  {
    out << "/>";
    tagStack.pop();
  }

  void BeginTag(const std::string & tag)
  {
    OpenTag(tag);
    CloseTag();
  }

  void EndTag()
  {
    std::string tag = tagStack.top(); tagStack.pop();
    NextLine();
    out << "</" << tag << ">";
  }
};

void BeginFilter(XMLFile & file, const std::string & filter)
{
  file.OpenTag("Filter");
  file.AddProperty("Name", filter);
  file.CloseTag();
}

void EndFilter(XMLFile & file)
{
  file.EndTag();
}

void WriteFile(XMLFile & file, const std::string & root, const std::string & filepath)
{
  std::string fullPath(root); fullPath += filepath;

  std::istringstream filepathStream(filepath);
  std::string fileBase, fileExt;
  getline(filepathStream, fileBase, '.');
  getline(filepathStream, fileExt);

  std::string objectName(fileBase);
  for(size_t i=0; i<objectName.size(); i++)
  {
    if(objectName[i] == '\\') objectName[i] = '.';
  }
  objectName += ".obj";

  file.OpenTag("File");
  file.AddProperty("RelativePath", fullPath);
  file.CloseTag();

  if(fileExt == "C" || fileExt == "cpp" || fileExt == "c")
  {
    file.OpenTag("FileConfiguration");
    file.AddProperty("Name", "Debug|Win32");
    file.CloseTag();

    {
      std::string objectPath(globalRoot);
      objectPath += "obj\\dbg\\";
      objectPath += objectName;

      file.OpenTag("Tool");
      file.AddProperty("Name", "VCCLCompilerTool");
      file.AddProperty("ObjectFile", objectPath);
      file.CloseEndTag();
    }

    file.EndTag();
    file.OpenTag("FileConfiguration");
    file.AddProperty("Name", "Release|Win32");
    file.CloseTag();

    {
      std::string objectPath(globalRoot);
      objectPath += "obj\\opt\\";
      objectPath += objectName;
      file.OpenTag("Tool");
      file.AddProperty("Name", "VCCLCompilerTool");
      file.AddProperty("ObjectFile", objectPath);
      file.CloseEndTag();
    }

    file.EndTag();
  }

  file.EndTag();
}



class Directory
{
	std::string root;
	std::string name;
	std::list<Directory*> subDirs;
	std::set<std::string> fileNames;

	void ReportFiles(XMLFile & out, const std::string & path) const
	{
		for(std::set<std::string>::const_iterator it = fileNames.begin(); it != fileNames.end(); it++)
		{
			std::ostringstream oss;
			oss << path << "src\\" << (*it);

			std::cout << "Using " << oss.str() << std::endl;

			WriteFile(out, root, oss.str());
		}
	}

	void EnumFiles(const std::string & path, const std::string & mask)
	{
		std::ostringstream oss;
		oss << root << path << "src\\" << mask;

		HANDLE fileHandle;
		WIN32_FIND_DATA findData;

		fileHandle = FindFirstFile( oss.str().c_str(), &findData );
		if( fileHandle != INVALID_HANDLE_VALUE )
		{
			fileNames.insert(std::string(findData.cFileName));

			// loop on all remeaining entries in dir
			while( FindNextFile( fileHandle, &findData ) )
			{
				fileNames.insert(std::string(findData.cFileName));
			}
		}

		FindClose( fileHandle );
	}
public:
	Directory(const std::string & root, const std::string & name) : root(root)
	{
		std::istringstream iss(name);
		getline(iss, this->name, '/');
		std::string rem;
		getline(iss, rem);
		if(!rem.empty())
		{
			AddSubDirectory(root, name);
		}
	}

	~Directory()
	{
		for(std::list<Directory*>::iterator it = subDirs.begin(); it != subDirs.end(); it++)
		{
			delete (*it);
		}
		subDirs.clear();
	}

	bool AddSubDirectory(const std::string & dirRoot, const std::string & dir)
	{
		std::istringstream iss(dir);
		std::string token;
		getline(iss, token, '/');
		if(token == name)
		{
			getline(iss, token);
			for(std::list<Directory*>::iterator it = subDirs.begin(); it != subDirs.end(); it++)
			{
				if((*it)->AddSubDirectory(dirRoot, token))
				{
					return true;
				}
			}
			subDirs.push_back(new Directory(dirRoot, token));
			return true;
		}
		return false;
	}

	void WriteFiles(XMLFile & out, const std::string & path)
	{
		BeginFilter(out, name);

		std::string myPath(path); myPath += name; myPath += "\\"; 

		EnumFiles(myPath, "*.C");
		EnumFiles(myPath, "*.cpp");
		EnumFiles(myPath, "*.c");
		EnumFiles(myPath, "*.H");
		EnumFiles(myPath, "*.hpp");
		EnumFiles(myPath, "*.h");
		ReportFiles(out, myPath);

		for(std::list<Directory*>::iterator it = subDirs.begin(); it != subDirs.end(); it++)
		{
			(*it)->WriteFiles(out, myPath);
		}

		EndFilter(out);
	}

	void WriteDirectories(std::ostream & out, const std::string & path) const
	{
		std::string myPath(path); myPath += name; myPath += "\\"; 
		out << "&quot;" << root << myPath << "src&quot;;";

		for(std::list<Directory*>::const_iterator it = subDirs.begin(); it != subDirs.end(); it++)
		{
			(*it)->WriteDirectories(out, myPath);
		}
	}
};

std::string globalName;
std::vector<std::string> roots;
Directory * dirApps;
Directory * dirLibs;

void WriteGUID(XMLFile & file)
{
	// TODO: Implement proper GUID generation
	static const char hexChars[]("0123456789ABCDEF");
	srand(GetTickCount());
	std::string guid("{");
	for(size_t i=0; i<8; i++) guid += hexChars[rand()%16];
	guid += '-';
	for(size_t i=0; i<4; i++) guid += hexChars[rand()%16];
	guid += '-';
	for(size_t i=0; i<4; i++) guid += hexChars[rand()%16];
	guid += '-';
	for(size_t i=0; i<4; i++) guid += hexChars[rand()%16];
	guid += '-';
	for(size_t i=0; i<12; i++) guid += hexChars[rand()%16];
	guid += '}';
	file.AddProperty("ProjectGUID", guid.c_str());
}

void WritePlatforms(XMLFile & file)
{
	file.BeginTag("Platforms");

	file.OpenTag("Platform");
	file.AddProperty("Name", "Win32");
	file.CloseEndTag();

	file.EndTag();
}

void WriteEmptyTool(XMLFile & file, const std::string & name)
{
	file.OpenTag("Tool");
	file.AddProperty("Name", name);
	file.CloseEndTag();
}

enum CONFIG { DEBUG, RELEASE };
void WriteCompilerTool(XMLFile & file, CONFIG config)
{
	file.OpenTag("Tool");
	file.AddProperty("Name", "VCCLCompilerTool");
	file.AddProperty("AdditionalOptions", "/TP /D _WIN32");

	if(config == DEBUG)
	{
		file.AddProperty("Optimization", "0");
	}

	std::ostringstream oss;
	dirApps->WriteDirectories(oss, "");
	dirLibs->WriteDirectories(oss, "");
	oss << globalRoot << "win32\\zlib;";
	oss << globalRoot << "win32\\include;";
	oss << globalRoot << "win32\\SDL-1.2\\include;";
	oss << globalRoot << "win32\\glew\\include;";
	oss << globalRoot << "config;";
	oss << "$(NOINHERIT)";
	file.AddProperty("AdditionalIncludeDirectories", oss.str());

	if(config == DEBUG)
	{
		file.AddProperty("PreprocessorDefinitions", "WIN32;_DEBUG;_CONSOLE;");
		file.AddProperty("MinimalRebuild", "true");
		file.AddProperty("BasicRuntimeChecks", "0");
		file.AddProperty("RuntimeLibrary", "3");
	}
	else
	{
		file.AddProperty("PreprocessorDefinitions", "WIN32;NDEBUG;_CONSOLE;TRANSFORM_MOUSE_Y=1;");
		file.AddProperty("RuntimeLibrary", "2");
	}

	if(config == DEBUG)
	{
		file.AddProperty("BufferSecurityCheck", "false");
		file.AddProperty("RuntimeTypeInfo", "true");
	}

	file.AddProperty("UsePrecompiledHeader", "0");

	if(config == DEBUG)
	{
		std::string pch(globalRoot); pch += "libs\\kernel\\src\\Global.H";
		file.AddProperty("PrecompiledHeaderThrough", pch);
	}

	file.AddProperty("WarningLevel", "3");
	file.AddProperty("DebugInformationFormat", (config == DEBUG) ? "4" : "3");
	file.AddProperty("CompileAs", "0");
	file.AddProperty("DisableSpecificWarnings", "4996,4267,4099,4800,4244,4309,4267,4355,4305");

	file.CloseEndTag();
}

void WriteLinkerTool(XMLFile & file, CONFIG config)
{
	file.OpenTag("Tool");
	file.AddProperty("Name", "VCLinkerTool");

	std::ostringstream oss;
	oss << globalRoot << "win32\\lib\\SDL.lib ";
	oss << globalRoot << "win32\\lib\\SDL_net.lib ";
	oss << globalRoot << "win32\\lib\\SDLmain.lib ";
	oss << globalRoot << "win32\\lib\\glew32.lib ";
	oss << globalRoot << "win32\\lib\\glew32s.lib ";
	oss << globalRoot << "win32\\lib\\glut32.lib";
	file.AddProperty("AdditionalDependencies", oss.str());

	file.AddProperty("OutputFile", globalRoot + "bin\\" + globalName + ".exe");
	file.AddProperty("LinkIncremental", "2");
	file.AddProperty("GenerateDebugInformation", "true");
	file.AddProperty("SubSystem", "1");

	if(config == RELEASE)
	{
		file.AddProperty("OptimizeReferences", "2");
		file.AddProperty("EnableCOMDATFolding", "2");
	}

	file.AddProperty("RandomizedBaseAddress", "1");
	file.AddProperty("DataExecutionPrevention", "0");
	file.AddProperty("TargetMachine", "1");

	file.CloseEndTag();
}

void WriteConfiguration(XMLFile & file, CONFIG config)
{
	file.OpenTag("Configuration");
	file.AddProperty("Name", (config == DEBUG) ? "Debug|Win32" : "Release|Win32");

	std::ostringstream oss;
	oss << globalRoot << "bin\\$(ConfigurationName)\\$(ProjectName)";
	file.AddProperty("OutputDirectory", oss.str());

	oss.str("");
	oss << globalRoot << "obj\\$(ConfigurationName)\\$(ProjectName)";
	file.AddProperty("IntermediateDirectory", oss.str());
	file.AddProperty("ConfigurationType", "1");
	file.CloseTag();

	WriteEmptyTool(file, "VCPreBuildEventTool");
	WriteEmptyTool(file, "VCCustomBuildTool");
	WriteEmptyTool(file, "VCXMLDataGeneratorTool");
	WriteEmptyTool(file, "VCWebServiceProxyGeneratorTool");
	WriteEmptyTool(file, "VCMIDLTool");
	WriteCompilerTool(file, config);
	WriteEmptyTool(file, "VCManagedResourceCompilerTool");
	WriteEmptyTool(file, "VCResourceCompilerTool");
	WriteEmptyTool(file, "VCPreLinkEventTool");
	WriteLinkerTool(file, config);
	WriteEmptyTool(file, "VCALinkTool");
	WriteEmptyTool(file, "VCManifestTool");
	WriteEmptyTool(file, "VCXDCMakeTool");
	WriteEmptyTool(file, "VCBscMakeTool");
	WriteEmptyTool(file, "VCFxCopTool");
	WriteEmptyTool(file, "VCAppVerifierTool");
	WriteEmptyTool(file, "VCPostBuildEventTool");

	file.EndTag();
}

void WriteConfigurations(XMLFile & file)
{
	file.BeginTag("Configurations");
	WriteConfiguration(file, DEBUG);
	WriteConfiguration(file, RELEASE);
	file.EndTag();
}

void WriteReferences(XMLFile & file)
{
	file.BeginTag("References");
	file.EndTag();
}

void WriteFiles(XMLFile & file)
{
	file.BeginTag("Files");

	dirApps->WriteFiles(file, "");
	dirLibs->WriteFiles(file, "");

	BeginFilter(file, "config");
	WriteFile(file, globalRoot, "config\\All.H");
	EndFilter(file);

	BeginFilter(file, "win32");
	BeginFilter(file, "zlib");
	WriteFile(file, globalRoot, "win32\\zlib\\adler32.c");
	WriteFile(file, globalRoot, "win32\\zlib\\compress.c");
	WriteFile(file, globalRoot, "win32\\zlib\\crc32.c");
	WriteFile(file, globalRoot, "win32\\zlib\\crc32.h");
	WriteFile(file, globalRoot, "win32\\zlib\\deflate.c");
	WriteFile(file, globalRoot, "win32\\zlib\\deflate.h");
	WriteFile(file, globalRoot, "win32\\zlib\\gzio.c");
	WriteFile(file, globalRoot, "win32\\zlib\\infback.c");
	WriteFile(file, globalRoot, "win32\\zlib\\inffast.c");
	WriteFile(file, globalRoot, "win32\\zlib\\inffast.h");
	WriteFile(file, globalRoot, "win32\\zlib\\inffixed.h");
	WriteFile(file, globalRoot, "win32\\zlib\\inflate.c");
	WriteFile(file, globalRoot, "win32\\zlib\\inflate.h");
	WriteFile(file, globalRoot, "win32\\zlib\\inftrees.c");
	WriteFile(file, globalRoot, "win32\\zlib\\inftrees.h");
	WriteFile(file, globalRoot, "win32\\zlib\\trees.c");
	WriteFile(file, globalRoot, "win32\\zlib\\trees.h");
	WriteFile(file, globalRoot, "win32\\zlib\\uncompr.c");
	WriteFile(file, globalRoot, "win32\\zlib\\Windows.h");
	WriteFile(file, globalRoot, "win32\\zlib\\zconf.h");
	WriteFile(file, globalRoot, "win32\\zlib\\zlib.h");
	WriteFile(file, globalRoot, "win32\\zlib\\zutil.c");
	WriteFile(file, globalRoot, "win32\\zlib\\zutil.h");
	EndFilter(file);
	EndFilter(file);

	file.EndTag();
}

void WriteGlobals(XMLFile & file)
{
	file.BeginTag("Globals");
	file.EndTag();
}

bool CheckDirectoryRoot(size_t index, const std::string & directory)
{
	std::ostringstream oss;
	oss << roots[index] << directory << "/src";

	HANDLE fileHandle;
	WIN32_FIND_DATA findData;

	fileHandle = FindFirstFile( oss.str().c_str(), &findData );
	bool result = (fileHandle != INVALID_HANDLE_VALUE);
	FindClose( fileHandle );

	return result;
}

size_t LookForDirectoryRoot(const std::string & directory)
{
	for(size_t i = 0; i < roots.size(); i++)
	{
		if(CheckDirectoryRoot(i, directory))
		{
			return i;
		}
	}
	return roots.size();
}

void ParseAppMkFile(std::istream & mk)
{
	std::string token;
	while( mk )
	{
		mk >> token;
		if(token == "APP_DIR")
		{
			mk >> token;
			if(token == ":=")
			{
				mk >> globalName;
				std::string dir("apps/"); dir += globalName;
				for(size_t i=0; i<dir.size(); i++)
				{
					if(dir[i] == '.') dir[i] = '/';
				}

				size_t index = LookForDirectoryRoot(dir);
				if(index < roots.size())
				{
					dirApps->AddSubDirectory(roots[index], dir);
				}
				else
				{
					std::cout << "Warning! Could not find directory " << dir << std::endl;
				}
			}
		}
		else if(token == "APP_LIBS")
		{
			mk >> token;
			if(token == ":=")
			{
				std::string line;
				lbreak:
				getline(mk, line);
				std::istringstream iss(line);
				while( iss )
				{
					iss >> token;
					if (token == "\\") goto lbreak;
					std::string dir("libs/"); dir += token;

					size_t index = LookForDirectoryRoot(dir);
					if(index < roots.size())
					{  
						dirLibs->AddSubDirectory(roots[index], dir);
					}
					else
					{
						std::cout << "Warning! Could not find directory " << dir << std::endl;
					}
				}
			}
		}
	}
}

int main(int argc, char *argv[])
{
	std::string app;
  
	if (argc > 1)
	{
		globalRoot = argv[1];
		for (int i = 1; i < argc-1; ++i) 
		{
			roots.push_back(std::string(argv[i]));
		}
		app = std::string(argv[argc-1]);
	}
	else
	{
		std::cout << "Enter primary ORTS repository location: ";
		std::cin >> globalRoot;

		while(true)
		{
			std::string root;
			std::cout << "Enter secondary ORTS repository location (q to quit): ";
			std::cin >> root;
			if(root == "q")
			{
				break;
			}
			roots.push_back(root);
		}

		std::cout << "Enter application designation: ";
		std::cin >> app;

		std::cout << "\nSettings\n========\n" << std::endl;
		std::cout << "[Primary repository]   : " << globalRoot << std::endl;
		for(size_t i = 0; i < roots.size(); i++)
		{
			std::cout << "[Secondary repository] : " << roots[i] << std::endl;
		}
		std::cout << "[Application]          : " << app << std::endl;

		roots.push_back(globalRoot);
	}

	for(size_t i = 0; i < app.size(); i++)
	{
		if(app[i] == '.') app[i] = '\\';
	}

	dirApps = new Directory(globalRoot, "apps");
	dirLibs = new Directory(globalRoot, "libs");

	for(std::vector<std::string>::iterator it = roots.begin(); it != roots.end(); it++)
	{
		std::string mkfile( (*it) );
		mkfile += "apps\\";
		mkfile += app;
		mkfile += "\\src\\app.mk";

		std::ifstream in(mkfile.c_str());
		if(in)
		{
			ParseAppMkFile(in);

			std::ostringstream oss;
			oss << globalName << ".vcproj";
			std::ofstream out(oss.str().c_str());
			XMLFile file(out);

			file.OpenTag("VisualStudioProject");
			file.AddProperty("ProjectType", "Visual C++");
			file.AddProperty("Version", "9.00");
			file.AddProperty("Name", globalName);
			WriteGUID(file);
			file.AddProperty("Keyword", "Win32Proj");
			file.CloseTag();

			WritePlatforms(file);
			WriteConfigurations(file);
			WriteReferences(file);
			WriteFiles(file);
			WriteGlobals(file);

			file.EndTag();

			return 0;
		}
	}

	std::cerr << "Error! app.mk not found!" << std::endl;

	return 0;
}
