// $Id: AI_Plan.H 5279 2007-06-23 02:49:08Z mburo $

// This is an ORTS file (c) Michael Chung, licensed under the GPL

#ifndef AI_Plan_H
#define AI_Plan_H

#include <map>
#include <vector>

#include "AI_Common.H"

class AI_Board;

// different AI plans:
#define AI_PLAN_UNKNOWN (0)
#define AI_PLAN_STAND   (1)
#define AI_PLAN_RANDOM  (2)
#define AI_PLAN_FLAG    (3)
#define AI_PLAN_EXPLORE (4)
#define AI_PLAN_MC      (5)
#define AI_PLAN_ALL     (7)
#define AI_PLAN_GROUP   (8)
#define AI_PLAN_DEFEND  (9)
#define AI_PLAN_RANDSIM (10)


typedef std::vector<int> int_vector;

/** a path is part of a plan - it's a plan for a single unit */
class AI_Path {
  
public:
  // constructors
  AI_Path() {}
  AI_Path(int_vector path, sint4 width)
  :
    m_path(path),
    m_width(width)
  {
  }

  void set(int_vector path, sint4 width);
  
  Point id_to_point(node_id_type node_id) const
  {
    return Point(node_id % m_width, node_id / m_width);
  }
  
  bool get_point(Point &pt, uint4 index) const
  {
    if (index >= m_path.size()) return false;
    
    pt = id_to_point(m_path[index]);
    
    return true;
  }
  
  bool get_next_waypoint(Point &pt) const;
  void pop_next_waypoint();
  
  sint4 length() const
  {
    return m_path.size();
  }
  
private:
  int_vector m_path;

  sint4 m_width; // used to convert id from/to Point
};

/** Description
    A plan that the AI Client uses to direct its actions.

*/
class AI_Plan 
{
public:
  // Plan ID functions
  static std::string get_str_from_id(sint4 id);
  static sint4 get_id_from_str(const std::string &s);

public:
  AI_Plan();

  void clear();

  bool is_empty() const { return m_paths.empty(); }
  
  bool plan_path(sint4 obj_id, const AI_Board &bd, Point pt_start, Point pt_dest);

  /** get (pointer to) the path for the object id given */
  const AI_Path *get_path_c(sint4 obj_id) const;
  AI_Path *get_path(sint4 obj_id);
  
  Point get_dest(sint4 obj_id) const;

  /** check if this object is in the plan */
  bool contains(sint4 obj_id) const {
    return !(m_paths.find(obj_id) == m_paths.end());
  }
  
  /** plan start time */
  void set_time(sint4 t) { m_n_time = t; }
  sint4 get_time() { return m_n_time; }
  
  bool is_completed(const AI_Board &bd) const;
  
  //-------------------------------------------------------------------
  // Plan Generation Functions
  //-------------------------------------------------------------------
  
  void generate_stand_plan();
  
  /** generates a completely random plan 
      @param bd        board-should be already updated for current game
      @param max_dist  maximum distance a unit is allowed to move
      @param side  side - the side to plan for
  */
  void generate_random_plan(const AI_Board &bd, sint4 max_dist, sint4 side);
  
  /** generates a capture-the-flag plan 
  
      if not in possession of flag, have units go after enemy flag
      otherwise return to base with enemy flag
      
      (if we don't know where flag is, then return error)
  
      @param bd    board - should be already updated for current game
      @param side  side - the side to plan for
  
      @return true if a plan is successfully generated, else false
  */
  bool generate_flag_plan(const AI_Board &bd, sint4 side);
  
  bool generate_explore_plan(const AI_Board &bd);
  
  bool generate_defend_plan(const AI_Board &bd, sint4 side);
  
  bool generate_group_plan(const AI_Board &bd);
  
  bool generate_plan(sint4 plan_id, const AI_Board &bd, sint4 side);
  
  sint4 get_plan_id() const { return m_plan_id; }
  
private:
  void add_path(sint4 id, const AI_Path &path) { m_paths[id] = path; }
  
  /** generate a random (valid) destination point 
     @param d  -  max_dist parameter
  */
  Point get_random_dest(const Point &pt, const AI_Board &bd, sint4 d);

private:
  // map from unit ID's to planned paths
  // the map should be complete (have a path for all units)
  std::map<sint4, AI_Path> m_paths;
  
  sint4 m_n_time;

  bool m_b_completed;

  sint4 m_plan_id;  // identifies the plan type
};


/** store information about the recent move commands so that we can 
	required to detect and handle collisions 
*/
class AI_Move_Command {
public:
  AI_Move_Command()
  :
	m_obj_id(-1),
	m_move_time(-1),
	m_move_origin(-1, -1),
	m_move_destination(-1, -1)
  {}

public:
  sint4 m_obj_id;

  // move information
  sint4 m_move_time;
  Point m_move_origin;
  Point m_move_destination;
};

#endif
